﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     FsSampleStreaming.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46404 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include "demo.h"

#include "../Common/FsSampleCommon.h"
#include "FsSampleStreaming.h"

using namespace nn;

/*
 * Overview
 *
 * This sample illustrates an implementation for file access that demands realtime characteristics, such as for streaming playback of video or audio.
 * 
 *
 * This sample demonstrates real-time streaming. It parses drawings made on the touch panel as raw binary bitmaps, and saves and plays them frame by frame.
 * 
 * 
 */

/*
 * Basic Flow
 *
 * Opens FileStream, and sets nn::fs::PRIORITY_APP_REALTIME using the TrySetPriority function.
 * Prepares one thread with a higher priority than the main thread, and performs reading/writing of an open file in that thread.
 * 
 * Note that the content of the buffer allocated for writing should not be changed during the write.
 */

/*
 * Implementing this sample
 *
 * All of the above basic flow is performed with RealTimeFileReader and RealTimeFileWriter.
 * With this class, to receive or send data via nn::os::BlockingQueue, the main thread does not pay attention to the behavior of the read/write thread.
 * 
 *
 * On the main thread side, prepared packets are received from BlockingQueue, and when the packet process is finished in the main thread, processed packets are packed into the BlockingQueue for returns.
 * 
 * 
 *
 * On the read/write thread side, when a sufficient amount of buffers have been packed into Queue, data is read/written in batch.
 * 
 */

namespace sample { namespace fs { namespace streaming {

// Saves data in the background.
Result BgSave(const s32 data)
{
    // When mounting an archive, nn::fs::PRIORITY_APP_NORMAL is applied if no priority has been specified in particular.
    RETURN_IF_FAILED(nn::fs::MountSaveData(SAVE_DATA_ARCHIVE));

    SaveData allData;
    allData.data[0] = data;
    {
        // When opening a file, the archive priority is inherited if no priority has been specified in particular.
        nn::fs::FileOutputStream file(DATA_FILE_NAME, false);
        file.Write(&allData, sizeof(allData), true);
    }

    Result result = nn::fs::CommitSaveData(SAVE_DATA_ARCHIVE);

    nn::fs::Unmount(SAVE_DATA_ARCHIVE);

    return result;
}

// Load the data in the background.
Result BgLoad(s32 &data)
{
    // When mounting an archive, nn::fs::PRIORITY_APP_NORMAL is applied if no priority has been specified in particular.
    RETURN_IF_FAILED(nn::fs::MountSaveData(SAVE_DATA_ARCHIVE));

    SaveData allData;
    {
        // When opening a file, the archive priority is inherited if no priority has been specified in particular.
        nn::fs::FileInputStream file(DATA_FILE_NAME);
        file.Read(&allData, sizeof(allData));
    }

    nn::fs::Unmount(SAVE_DATA_ARCHIVE);

    data = allData.data[0];
    return ResultSuccess();
}

}}}
