/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     demo_Models.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/dbg.h>
#include <nn/assert.h>

#include <cmath>

#include "demo_Models.h"

namespace demo {

namespace {
    inline u8* GetBufferOffset(size_t size)
    {
        return (u8*)NULL + size;
    }
} // namespace (unnamed)

/* ------------------------------------------------------------------------
        ModelBase class (abstract class)
   ------------------------------------------------------------------------ */
ModelBase::ModelBase() : m_AttributeFlag(0), m_IndexSize(0)
{
    m_DataSize[0] = m_DataSize[1] = m_DataSize[2] = m_DataSize[3] = 0;
    m_Initialized = false;
}

void ModelBase::FinalizeImpl()
{
    if (!m_Initialized)
    {
        return;
    }
    
    glDeleteBuffers(1, &m_ArrayBufferObjectId);
    glDeleteBuffers(1, &m_ElementArrayBufferObjectId);
    m_Initialized = false;
}

void ModelBase::Draw()
{
    if (!m_Initialized)
    {
        NN_PANIC("Model is not initialized.\n");
    }
    
    glBindBuffer(GL_ARRAY_BUFFER, m_ArrayBufferObjectId);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_ElementArrayBufferObjectId);
    
    int attrNum = 0;
    size_t offset = 0;
    
    if (m_AttributeFlag & ATTR_POSITION)
    {
        glVertexAttribPointer(attrNum, detail::ATTR_NUM_POS, GL_FLOAT, GL_FALSE, 0, 0);
        ++attrNum;
        offset += m_DataSize[attrNum-1];
    }
    
    if (m_AttributeFlag & ATTR_NORMAL)
    {
        glVertexAttribPointer(attrNum, detail::ATTR_NUM_NORMAL, GL_FLOAT, GL_FALSE, 0, GetBufferOffset(offset));
        ++attrNum;
        offset += m_DataSize[attrNum-1];
    }
    
    if (m_AttributeFlag & ATTR_VERTEX_COLOR)
    {
        glVertexAttribPointer(attrNum, detail::ATTR_NUM_COLOR, GL_FLOAT, GL_FALSE, 0, GetBufferOffset(offset));
        ++attrNum;
        offset += m_DataSize[attrNum-1];
    }
    
    if (m_AttributeFlag & ATTR_TEX_COORDINATE)
    {
        glVertexAttribPointer(attrNum, detail::ATTR_NUM_TEX, GL_FLOAT, GL_FALSE, 0, GetBufferOffset(offset));
        ++attrNum;
    }
    
    for (int i=0; i<attrNum; ++i)
    {
        glEnableVertexAttribArray(i);
    }
    
    glDrawElements(GL_TRIANGLES, m_IndexSize/sizeof(u16), GL_UNSIGNED_SHORT, 0);
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    
    for (int j=0; j<attrNum; ++j)
    {
        glDisableVertexAttribArray(j);
    }
}

void ModelBase::InitializeCore(u16* indices, f32* attr1, f32* attr2, f32* attr3, f32* attr4)
{
    if (indices == NULL || attr1 == NULL)
    {
        NN_PANIC("illegal input.\n");
    }
    
    glGenBuffers(1, &m_ArrayBufferObjectId);
    glBindBuffer(GL_ARRAY_BUFFER, m_ArrayBufferObjectId);
    glBufferData(GL_ARRAY_BUFFER, m_DataSize[0] + m_DataSize[1] + m_DataSize[2] + m_DataSize[3], 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, m_DataSize[0], attr1);
    
    // If there are more than two vertex attributes, specify them in order
    if ( attr2 != NULL )
    {
        glBufferSubData(GL_ARRAY_BUFFER, m_DataSize[0], m_DataSize[1], attr2);
        if ( attr3 != NULL )
        {
            glBufferSubData(GL_ARRAY_BUFFER, m_DataSize[0] + m_DataSize[1], m_DataSize[2], attr3);
            if ( attr4 != NULL )
            {
                glBufferSubData(GL_ARRAY_BUFFER, m_DataSize[0] + m_DataSize[1] + m_DataSize[2], m_DataSize[3], attr4);
            }
        }
    }
    
    glGenBuffers(1, &m_ElementArrayBufferObjectId);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_ElementArrayBufferObjectId);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, m_IndexSize, indices, GL_STATIC_DRAW);
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    
    m_Initialized = true;
}

/* ------------------------------------------------------------------------
        Cube class
   ------------------------------------------------------------------------ */
void CubeModel::Initialize(u8 flag, f32 r, f32 g, f32 b, f32 alpha)
{
    if (m_Initialized)
    {
        return;
    }
    
    m_DataSize[0] = m_DataSize[1] = m_DataSize[2] = m_DataSize[3] = 0;
    m_AttributeFlag = flag;
    
    // Define in counterclockwise order the vertices of the triangles that form the surfaces
    f32 positions[] =
    {
         0.5f, -0.5f,  0.5f,   0.5f, -0.5f, -0.5f,   0.5f,  0.5f, -0.5f,   0.5f,  0.5f,  0.5f, // +X
        -0.5f, -0.5f, -0.5f,  -0.5f, -0.5f,  0.5f,  -0.5f,  0.5f,  0.5f,  -0.5f,  0.5f, -0.5f, // -X
        -0.5f,  0.5f,  0.5f,   0.5f,  0.5f,  0.5f,   0.5f,  0.5f, -0.5f,  -0.5f,  0.5f, -0.5f, // +Y
        -0.5f, -0.5f, -0.5f,   0.5f, -0.5f, -0.5f,   0.5f, -0.5f,  0.5f,  -0.5f, -0.5f,  0.5f, // -Y
        -0.5f, -0.5f,  0.5f,   0.5f, -0.5f,  0.5f,   0.5f,  0.5f,  0.5f,  -0.5f,  0.5f,  0.5f, // +Z
         0.5f, -0.5f, -0.5f,  -0.5f, -0.5f, -0.5f,  -0.5f,  0.5f, -0.5f,   0.5f,  0.5f, -0.5f  // -Z
    };
    
    f32 len = ::std::sqrt(0.5*0.5+0.5*0.5+0.5*0.5);
    
    f32 normals[] =
    {
         0.5f/len, -0.5f/len,  0.5f/len,   0.5f/len, -0.5f/len, -0.5f/len,   0.5f/len,  0.5f/len, -0.5f/len,   0.5f/len,  0.5f/len,  0.5f/len, // +X
        -0.5f/len, -0.5f/len, -0.5f/len,  -0.5f/len, -0.5f/len,  0.5f/len,  -0.5f/len,  0.5f/len,  0.5f/len,  -0.5f/len,  0.5f/len, -0.5f/len, // -X
        -0.5f/len,  0.5f/len,  0.5f/len,   0.5f/len,  0.5f/len,  0.5f/len,   0.5f/len,  0.5f/len, -0.5f/len,  -0.5f/len,  0.5f/len, -0.5f/len, // +Y
        -0.5f/len, -0.5f/len, -0.5f/len,   0.5f/len, -0.5f/len, -0.5f/len,   0.5f/len, -0.5f/len,  0.5f/len,  -0.5f/len, -0.5f/len,  0.5f/len, // -Y
        -0.5f/len, -0.5f/len,  0.5f/len,   0.5f/len, -0.5f/len,  0.5f/len,   0.5f/len,  0.5f/len,  0.5f/len,  -0.5f/len,  0.5f/len,  0.5f/len, // +Z
         0.5f/len, -0.5f/len, -0.5f/len,  -0.5f/len, -0.5f/len, -0.5f/len,  -0.5f/len,  0.5f/len, -0.5f/len,   0.5f/len,  0.5f/len, -0.5f/len  // -Z
    };
    
    f32 colors[] =
    {
        r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,
        r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,
        r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,
        r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,
        r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,
        r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha,  r, g, b, alpha
    };
    
    f32 texcoords[] =
    {
        0.0f, 0.0f,   1.0f, 0.0f,   1.0f, 1.0f,   0.0f, 1.0f,  // +X
        0.0f, 0.0f,   1.0f, 0.0f,   1.0f, 1.0f,   0.0f, 1.0f,  // -X
        0.0f, 0.0f,   1.0f, 0.0f,   1.0f, 1.0f,   0.0f, 1.0f,  // +Y
        0.0f, 0.0f,   1.0f, 0.0f,   1.0f, 1.0f,   0.0f, 1.0f,  // -Y
        0.0f, 0.0f,   1.0f, 0.0f,   1.0f, 1.0f,   0.0f, 1.0f,  // +Z
        0.0f, 0.0f,   1.0f, 0.0f,   1.0f, 1.0f,   0.0f, 1.0f,  // -Z
    };
    
    u16 indices[] =
    {
         0,  1,  2,  0,  2,  3,
         4,  5,  6,  4,  6,  7,
         8,  9, 10,  8, 10, 11,
        12, 13, 14, 12, 14, 15,
        16, 17, 18, 16, 18, 19,
        20, 21, 22, 20, 22, 23
    };
    m_IndexSize = sizeof(indices);

    int attrNum = 0;
    f32* attrBufPtr[4] = {NULL, NULL, NULL, NULL};
    if (flag & ATTR_POSITION)
    {
        attrBufPtr[attrNum] = positions;
        m_DataSize[attrNum] = sizeof(positions);
        ++attrNum;
    }
    
    if (flag & ATTR_NORMAL)
    {
        attrBufPtr[attrNum] = normals;
        m_DataSize[attrNum] = sizeof(normals);
        ++attrNum;
    }
    
    if (flag & ATTR_VERTEX_COLOR)
    {
        attrBufPtr[attrNum] = colors;
        m_DataSize[attrNum] = sizeof(colors);
        ++attrNum;
    }
    
    if (flag & ATTR_TEX_COORDINATE)
    {
        attrBufPtr[attrNum] = texcoords;
        m_DataSize[attrNum] = sizeof(texcoords);
        ++attrNum;
    }
    
    InitializeCore(indices, attrBufPtr[0], attrBufPtr[1], attrBufPtr[2], attrBufPtr[3]);
}

/* ------------------------------------------------------------------------
   ------------------------------------------------------------------------ */

}

