﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ShadowKnotCube.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */
 
/*
 * Comments
 * --------
 *
 * This sample uses omnidirectional shadow mapping (cube shadow mapping).
 * The first pass creates a shadow texture for six sides. For this reason, rendering for six scenes is needed in the first pass.
 * 
 */

#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/math.h>
#include <nn/init.h>
#include <nn/os.h>
#include <nn/applet.h>
#include "Util.h"
#include "Loader.h"

#include "demo.h"

#include <string.h>
#include <math.h>

#include "Memory.h"

#define APP_NAME "ShadowKnotCube"
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

/* shadow setting */
static const float fSZBiasScale[] = {1.7f, 1.3f, 1.3f};
/* shadow artifact suppression SETTINGS 0 - lowest 2 - highest */ 
/* Parameters for suppressing artifacts. The larger the number, the greater the effect of the suppression.*/
#define SDW_SETTINGS 0
/* Shadow texture size.
 * A shadow texture used with an omnidirectional shadow map must have a square size, just like a cube map.*/
#define SBWIDTHHEIGHT 128

/* GL object name */
/* program object */
static GLuint s_LrProgName, s_SdwProgName;
/* shader object */
static GLuint s_ShadersName[2];
/* texture object */
static GLuint s_ShadowTexName;
/* frame buffer object */
static GLuint s_FboName;

/* obj loader class object */
static dat_t knot, cube;
static dat_t* objects[] = {&knot, &cube, 0};

/* 1st pass (shadow accumulation pass) frustum argument */
/* When an omnidirectional shadow map is used, the camera used in the first pass must have a field-of-view angle of 90 degrees. The same is true when a dynamic cube map is used.
 * 
 * For this reason, the left(r) and top(t) values for the near face must be the same.*/
static GLfloat n = 0.2f, f = 10.f, r = 0.2f, t = 0.2f;

/* Light position */
static GLfloat lpos0[] = {0.9f, 0.f, 0.f, 1.f};

/* frame number */
static unsigned int s_FrameNo = 0;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
const u32 s_GxHeapSize = 0x400000;
const u32 s_HeapSize = 0x200000;

demo::RenderSystem s_RenderSystem;

// Generates a standard OpenGL projection transformation matrix
nn::math::Matrix44*
MTX44FrustumGl(nn::math::Matrix44* pOut, f32 l, f32 r, f32 b, f32 t, f32 n, f32 f)
{
    f32 (*const m)[4] = pOut->m;
    f32 tmp     =  1.0f / (r - l);
    m[0][0] =  (2*n) * tmp;
    m[0][1] =  0.0f;
    m[0][2] =  (r + l) * tmp;
    m[0][3] =  0.0f;

    tmp     =  1.0f / (t - b);
    m[1][0] =  0.0f;
    m[1][1] =  (2*n) * tmp;
    m[1][2] =  (t + b) * tmp;
    m[1][3] =  0.0f;

    m[2][0] =  0.0f;
    m[2][1] =  0.0f;

    tmp = 1.0f / (n - f);

    m[2][2] = (f + n) * tmp;
    m[2][3] = 2.0f * f * n * tmp;

    m[3][0] =  0.0f;
    m[3][1] =  0.0f;
    m[3][2] = -1.0f;
    m[3][3] =  0.0f;
    
    return pOut;
}

static void RenderObjectsShadow(dat_t& _objs)
{
    /* Because vertex processing in the first pass only converts vertex coordinates, the only vertex attributes to enable are the vertex coordinates.
     * */
    glEnableVertexAttribArray(0);

    for (int i = 0; i < _objs.obj_num; i++)
    {
        glBindBuffer(GL_ARRAY_BUFFER, _objs.posVB);
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].vtx_offset);
        for (unsigned j = _objs.obj[i].patch_offset; j < _objs.obj[i].patch_size + _objs.obj[i].patch_offset; j++)
        {
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _objs.idxVB);
            glDrawElements(GL_TRIANGLES, _objs.patch[j].elm_size,
                GL_UNSIGNED_SHORT, (GLvoid*)(_objs.patch[j].elm_offset + _objs.obj[i].elm_offset));
        }
    }

    glDisableVertexAttribArray(0);
}

static void RenderObjectsShadowed(dat_t& _objs)
{
    /* Normal vectors are needed to perform fragment lighting in the second pass.
     * The vertex attributes to enable are vertex coordinates and normals.*/
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    for (int i = 0; i < _objs.obj_num; i++)
    {
        glBindBuffer(GL_ARRAY_BUFFER, _objs.posVB);
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].vtx_offset);
        glBindBuffer(GL_ARRAY_BUFFER, _objs.normVB);
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].nrm_offset);
        for (unsigned j = _objs.obj[i].patch_offset; j < _objs.obj[i].patch_size + _objs.obj[i].patch_offset; j++)
        {
            GLfloat ma[4] = {0.f, 0.f, 0.f, 1.f};
            GLfloat md[4] = {0.f, 0.f, 0.f, 1.f};
            GLfloat ms[4] = {0.f, 0.f, 0.f, 1.f};
            for (int col = 0; col < 3; col++)
            {
                ma[col] = _objs.patch[j].ambient[col];
                md[col] = _objs.patch[j].diffuse[col];
                ms[col] = _objs.patch[j].specular[col];
            }

            glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_FragmentMaterial.ambient"), 1, ma);
            glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_FragmentMaterial.diffuse"), 1, md);
            glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_FragmentMaterial.specular0"), 1, ms);
            glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_FragmentMaterial.specular1"), 1, ms);

            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _objs.idxVB);
            glDrawElements(GL_TRIANGLES, _objs.patch[j].elm_size,
                GL_UNSIGNED_SHORT, (GLvoid*)(_objs.patch[j].elm_offset + _objs.obj[i].elm_offset));
        }
    }

    glDisableVertexAttribArray(0);
    glDisableVertexAttribArray(1);
}

static void RenderShadowBuffer(dat_t** _objs, nn::math::Vector3 lpos)
{
    nn::math::Matrix44 m;
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos, camUp, target;

    glUseProgram(s_SdwProgName);

    /* When using omnidirectional shadow maps, rendering for six sides is needed in the first pass.
     * Because we render the scene as it appears from the light position, note that the position of the light will become the position of the viewpoint.
     * */

    // rendering +x cube
    /* Arguments 4, 5, 6 of the lookAt function indicate the target position seen from the viewpoint.
     * In other words, lpos[0]+1.f, the fourth argument, indicates that the viewer faces toward the positive x-axis.*/
    camPos = nn::math::Vector3(lpos[0], lpos[1], lpos[2]);
    camUp = nn::math::Vector3(0.f, -1.f, 0.f);
    target = nn::math::Vector3(lpos[0] + 1.f, lpos[1], lpos[2]);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProgName, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    /* Attaches a shadow texture to the framebuffer object, but note that this specifies a different target for each surface.
     * */
    glFramebufferTexture2D(
        GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_CUBE_MAP_POSITIVE_X, s_ShadowTexName, 0);
    glCheckFramebufferStatus(GL_FRAMEBUFFER);

    glClear(GL_COLOR_BUFFER_BIT);

    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering -x cube
    target = nn::math::Vector3(lpos[0] - 1.f, lpos[1], lpos[2]);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProgName, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    glFramebufferTexture2D(
        GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_CUBE_MAP_NEGATIVE_X, s_ShadowTexName, 0);

    glClear(GL_COLOR_BUFFER_BIT);

    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering +y cube
    camUp = nn::math::Vector3(0.f, 0.f, 1.f);
    target = nn::math::Vector3(lpos[0], lpos[1] + 1.f, lpos[2]);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProgName, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    glFramebufferTexture2D(
        GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_CUBE_MAP_POSITIVE_Y, s_ShadowTexName, 0);

    glClear(GL_COLOR_BUFFER_BIT);

    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering -y cube
    camUp = nn::math::Vector3(0.f, 0.f, -1.f);
    target = nn::math::Vector3(lpos[0], lpos[1] - 1.f, lpos[2]);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProgName, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    glFramebufferTexture2D(
        GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, s_ShadowTexName, 0);

    glClear(GL_COLOR_BUFFER_BIT);

    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering +z cube
    camUp = nn::math::Vector3(0.f, -1.f, 0.f);
    target = nn::math::Vector3(lpos[0], lpos[1], lpos[2] + 1.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProgName, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    glFramebufferTexture2D(
        GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_CUBE_MAP_POSITIVE_Z, s_ShadowTexName, 0);

    glClear(GL_COLOR_BUFFER_BIT);

    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering -z cube
    target = nn::math::Vector3(lpos[0], lpos[1], lpos[2] - 1.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProgName, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    glFramebufferTexture2D(
        GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, s_ShadowTexName, 0);

    glClear(GL_COLOR_BUFFER_BIT);

    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }
}

static void RenderLight(dat_t& _objs)
{
    glUseProgram(s_LrProgName);

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    /* Renders light source using constant color, and not the fragment lighting results.*/
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_FragmentLighting.enabled"), GL_FALSE);
    glUniform3i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].srcRgb"), GL_CONSTANT, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].srcAlpha"), GL_CONSTANT, GL_CONSTANT, GL_CONSTANT);

    glDisable(GL_CULL_FACE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    for (int i = 0; i < _objs.obj_num; i++)
    {
        glBindBuffer(GL_ARRAY_BUFFER, _objs.posVB);
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].vtx_offset);
        glBindBuffer(GL_ARRAY_BUFFER, _objs.normVB);
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].nrm_offset);
        for (unsigned j = _objs.obj[i].patch_offset; j < _objs.obj[i].patch_size + _objs.obj[i].patch_offset; j++)
        {
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _objs.idxVB);
            glDrawElements(GL_TRIANGLES, _objs.patch[j].elm_size,
                GL_UNSIGNED_SHORT, (GLvoid*)(_objs.patch[j].elm_offset + _objs.obj[i].elm_offset));
        }
    }

    glEnable(GL_CULL_FACE);
    glDisable(GL_BLEND);

    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_FragmentLighting.enabled"), GL_TRUE);
    glUniform3i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].srcRgb"), GL_FRAGMENT_PRIMARY_COLOR_DMP, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].srcAlpha"), GL_FRAGMENT_PRIMARY_COLOR_DMP, GL_CONSTANT, GL_CONSTANT);

    glDisableVertexAttribArray(0);
    glDisableVertexAttribArray(1);
}

static int Setup1stPsssShader()
{
    nn::math::Matrix44 proj;

    s_SdwProgName = glCreateProgram();

    glAttachShader(s_SdwProgName, s_ShadersName[0]);
    glAttachShader(s_SdwProgName, GL_DMP_FRAGMENT_SHADER_DMP);
    glBindAttribLocation(s_SdwProgName, 0, "aPosition");

    glLinkProgram(s_SdwProgName);
    glUseProgram(s_SdwProgName);

    /* Performs rendering for first pass hard shadows.
     * For this reason, textures and fragment lighting are not used.
     * Because hard shadows must output as black in color (0.0, 0.0, 0.0, 0.0), a constant color is specified as the 0'th input operand to the texture combiner, GL_REPLACE is specified as the mode, and the color is output unchanged.
     * 
     * */
    glUniform1i(glGetUniformLocation(s_SdwProgName, "dmp_Texture[0].samplerType"), GL_FALSE);
    glUniform1i(glGetUniformLocation(s_SdwProgName, "dmp_FragmentLighting.enabled"), GL_FALSE);
    glUniform1i(glGetUniformLocation(s_SdwProgName, "dmp_FragmentLightSource[0].enabled"), GL_FALSE);

    GLfloat color[] = {0.f, 0.f, 0.f, 0.f};
    glUniform4fv(glGetUniformLocation(s_SdwProgName, "dmp_TexEnv[0].constRgba"), 1, color);
    glUniform1i(glGetUniformLocation(s_SdwProgName, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_SdwProgName, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_SdwProgName, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_SdwProgName, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_SdwProgName, "dmp_TexEnv[0].srcRgb"), GL_CONSTANT, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_SdwProgName, "dmp_TexEnv[0].srcAlpha"), GL_CONSTANT, GL_CONSTANT, GL_CONSTANT);

    /* Note that when fragment operations are in shadow accumulation mode, the depth value that is stored in the shadow texture must not be divided by w.
     * 
     * The scaling coefficient is set to the camera's 'far' value.
     *  For details, see DMP shadows in the programming guide.*/
    glUniform1f(glGetUniformLocation(s_SdwProgName, "dmp_FragOperation.wScale"), 1.f / f);
    glUniform1i(glGetUniformLocation(s_SdwProgName, "dmp_FragOperation.mode"), GL_FRAGOP_MODE_SHADOW_DMP);

    /* Because r = 0.1, t = 0.1, the frustum's vertical and horizontal field-of-view angles must be 90 degrees.
     * */
    nn::math::MTX44Frustum(&proj, -r, r, -t, t, n, f);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProgName, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glValidateProgram(s_SdwProgName);

    return 1;
}

static int Setup2ndPsssShader()
{
    nn::math::Matrix44 proj;

    s_LrProgName = glCreateProgram();

    glAttachShader(s_LrProgName, s_ShadersName[1]);
    glAttachShader(s_LrProgName, GL_DMP_FRAGMENT_SHADER_DMP);
    glBindAttribLocation(s_LrProgName, 0, "aPosition");
    glBindAttribLocation(s_LrProgName, 1, "aNormal");

    glLinkProgram(s_LrProgName);
    glUseProgram(s_LrProgName);

    /* For details on fragment lighting settings, see DMP fragment lighting in the programming guide and the fragment lighting sample programs.
     * 
     * */
    GLfloat ld0[] = {1.f, 1.f, 1.f, 1.f};
    GLfloat ls0[] = {0.f, 0.f, 0.f, 0.f};
    GLfloat ls1[] = {1.f, 1.f, 1.f, 1.f};
    GLfloat la0[] = {1.f, 1.f, 1.f, 1.f};

    glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_FragmentLightSource[0].diffuse"), 1, ld0);
    glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_FragmentLightSource[0].specular0"), 1, ls0);
    glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_FragmentLightSource[0].specular1"), 1, ls1);
    glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_FragmentLightSource[0].ambient"), 1, la0);

    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_FragmentLighting.enabled"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_FragmentLightSource[0].enabled"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_FragmentLightSource[0].shadowed"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_LightEnv.lutInputRR"), GL_LIGHT_ENV_NH_DMP);
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_LightEnv.lutEnabledRefl"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_LightEnv.config"), GL_LIGHT_ENV_LAYER_CONFIG1_DMP);

    /* Add shadow contribution to the primary color (diffuse component) of fragment lighting.
     * */
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_LightEnv.shadowPrimary"), GL_TRUE);

    GLuint lutids[3];

    glGenTextures(3, lutids);
    glBindTexture(GL_TEXTURE_COLLECTION_DMP, lutids[0]);
    
    GLfloat lut[512];
    memset(lut, 0, sizeof(lut));

    int j;

    for (j = 1; j < 128; j++)
    {
        lut[j] = powf( (float)j / 127.f, 30.f);
        lut[j + 255] = lut[j] - lut[j - 1];
    }

    glBindTexture(GL_LUT_TEXTURE0_DMP, lutids[1]);
    glTexImage1D(GL_LUT_TEXTURE0_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, lut);
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_FragmentMaterial.samplerRR"), 0);

    /* The texture's sampling type is set to omnidirectional shadow.*/
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_Texture[0].samplerType"), GL_TEXTURE_SHADOW_CUBE_DMP);

     /* The shadowZBias argument is not always used to suppress artifacts, but making the bias deeper does help to suppress artifacts.
     * */
    glUniform1f(glGetUniformLocation(s_LrProgName, "dmp_Texture[0].shadowZBias"), n * fSZBiasScale[SDW_SETTINGS] / (f - n));

    /* When rendering to shadow textures and perspective projection is used as the frustum, perspectiveShadow must be enabled. (This is the default.)
     * */
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_Texture[0].perspectiveShadow"), GL_TRUE);

    /* The constant color set here is used when rendering the light source.*/
    GLfloat color[] = {1.f, 1.f, 0.f, 0.5f};
    glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].constRgba"), 1, color);

    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].srcRgb"), GL_FRAGMENT_PRIMARY_COLOR_DMP, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_LrProgName, "dmp_TexEnv[0].srcAlpha"), GL_FRAGMENT_PRIMARY_COLOR_DMP, GL_CONSTANT, GL_CONSTANT);

    /* Fragment operation is set to the mode that performs the same operations normally performed in OpenGL, and the depth value stored in the depth buffer is divided by w as per normal OpenGL operations.
     * */
    glUniform1f(glGetUniformLocation(s_LrProgName, "dmp_FragOperation.wScale"), 0.f);
    glUniform1i(glGetUniformLocation(s_LrProgName, "dmp_FragOperation.mode"), GL_FRAGOP_MODE_GL_DMP);

    nn::math::MTX44Frustum(&proj, -0.1f, 0.1f, -0.1f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH,
                  0.1f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 10.f);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProgName, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glValidateProgram(s_LrProgName);

    return 1;
}

static int SetupShader()
{
    /* The shader setup process with DMPGL2.0 uses the same mechanism as OpenGLES2.0.
     * Because DMPGL 2.0 only allows the vertex shader to be user-defined, only a vertex shader object is created.
     * 
     * This sample creates two shader objects because it performs different vertex processing in the first and second passes.
     * */
    s_ShadersName[0] = glCreateShader(GL_VERTEX_SHADER);
    s_ShadersName[1] = glCreateShader(GL_VERTEX_SHADER);

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(2, s_ShadersName, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    Setup1stPsssShader();
    Setup2ndPsssShader();

    return 1;
}

static void LoadObjects()
{
    loadDAT( "rom:/resources/knot.dat", &knot);
    loadDAT( "rom:/resources/rcube.dat", &cube);
}

/* unload objects */
static void UnloadObjects(void)
{
    unloadDAT(&knot);
    unloadDAT(&cube);
}

static void Draw1stPass()
{
    glBindFramebuffer(GL_FRAMEBUFFER, s_FboName);
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

    glViewport(0, 0, SBWIDTHHEIGHT, SBWIDTHHEIGHT);

    RenderShadowBuffer(objects, nn::math::Vector3(lpos0));

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

static void  Draw2ndPass()
{
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(
        3.f * sinf(static_cast<f32>(s_FrameNo) * static_cast<f32>(M_PI) / 180.f),
        1.f,
        3.f * cosf(static_cast<f32>(s_FrameNo) * static_cast<f32>(M_PI) / 180.f));
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();   

    glUseProgram(s_LrProgName);

    /* In the second pass, the shadow texture created in the first pass is accessed.*/
    glBindTexture(GL_TEXTURE_CUBE_MAP, s_ShadowTexName);

    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Matrix44 m(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProgName, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    nn::math::Vector4 vlpos0(lpos0);
    nn::math::Vector4 mv0(mv.m[0][0], mv.m[0][1], mv.m[0][2], mv.m[0][3]);
    nn::math::Vector4 mv1(mv.m[1][0], mv.m[1][1], mv.m[1][2], mv.m[1][3]);
    nn::math::Vector4 mv2(mv.m[2][0], mv.m[2][1], mv.m[2][2], mv.m[2][3]);
    nn::math::Vector4 lpos(nn::math::VEC4Dot(&mv0, &vlpos0), nn::math::VEC4Dot(&mv1, &vlpos0), nn::math::VEC4Dot(&mv2, &vlpos0), vlpos0.w);
    glUniform4fv(glGetUniformLocation(s_LrProgName, "dmp_FragmentLightSource[0].position"), 1, static_cast<f32*>(lpos));

    /* The appropriate texture matrix must be set in the second pass.
     * For details, see the chapter on DMP shadows in the DMPGL 2.0 Programming Guide.*/
    nn::math::Matrix34 tex;
    nn::math::Vector3 scale(1.f / (f - n), 1.f / (f - n), 1.f / (f - n));
    nn::math::Vector3 trans(-lpos0[0], -lpos0[1], -lpos0[2]);
    nn::math::MTX34Scale(&tex, &scale);
    nn::math::MTX34MultTranslate(&tex, &tex, &trans);
    
    m = nn::math::Matrix44(tex);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProgName, "uTexture"), 1, GL_TRUE, static_cast<f32*>(m));

    RenderObjectsShadowed(cube);
    RenderObjectsShadowed(knot);

    /* Renders the light source.*/
    trans = nn::math::Vector3(lpos0[0], lpos0[1], lpos0[2]);
    scale = nn::math::Vector3( 0.02f, 0.02f, 0.02f );
    
    nn::math::MTX34MultTranslate(&mv, &mv, &trans);
    nn::math::MTX34MultScale(&mv, &mv, &scale);
    m = nn::math::Matrix44(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProgName, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    RenderLight(cube);

    s_RenderSystem.SwapBuffers();
}

int DrawFrame(void)
{
    /* Light position */
    lpos0[0] = 0.9f;
    lpos0[1] = 0.5f * sinf(s_FrameNo * 5.f * (float)M_PI / 180.f);
    lpos0[2] = 0.5f * cosf(s_FrameNo * 5.f * (float)M_PI / 180.f);
    lpos0[3] = 1.f;

    /* shadow accumulation pass */
    Draw1stPass();
    /* shadow test pass */
    Draw2ndPass();

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

    s_FrameNo++;

    return !glGetError();
}

int Initialize()
{
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    /* Initialize display */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);

    /* Create heap for malloc*/
    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    LoadObjects();

    if (!SetupShader())
        return -1;

    /* Because a color (1.f, 1.f, 1.f, 1.f) must be used to clear the shadow texture in the first pass, the color buffer's clear color is set as follows.
     * */
    glClearColor(1.f, 1.f, 1.f, 1.f);

    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);

    glGenTextures(1, &s_ShadowTexName);
    glBindTexture(GL_TEXTURE_CUBE_MAP, s_ShadowTexName);

    /* When we access the shadow texture, the addressing mode is GL_CLAMP_TO_BORDER.
     * The border color must be set to white (1.f, 1.f, 1.f, 1.f) at this time.
     * This guarantees that the rendering result will not be affected by shadows if the object is outside the shadow texture's boundaries.
     * */
    GLfloat bcolor[] = {1.f, 1.f, 1.f, 1.f};
    glTexParameterfv(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_BORDER_COLOR, bcolor);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

    /* Note that shadow textures do not support mipmaps.
     * */
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_LOD, 0);

    for (unsigned int face = 0; face < 6; face++)
        glTexImage2D(
            GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_SHADOW_DMP, SBWIDTHHEIGHT, SBWIDTHHEIGHT,
            0, GL_SHADOW_DMP, GL_UNSIGNED_INT, 0);

    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

    glGenFramebuffers(1, &s_FboName);

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    /* initialization */
    if (Initialize() >= 0)
    {
        /* Enter loop */
        while (1)
        {
            (void)DrawFrame();
        }
    }
    
    UnloadObjects();
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}
