/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */

#include <nn.h>
#include <nn/fs.h>
#include <nn/gx.h>
#include <nn/math.h>
#include <nn/applet.h>

#include "demo.h"

#include <string.h>
#include <math.h>

#include "Loader.h"
#include "Util.h"
#include "Memory.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#define DMP_PI  (3.1415926f)
#define REV_PI  (1.0f/DMP_PI)

namespace {

/* program id */
GLuint s_ProgramID;

/* shader id */
GLuint s_ShaderID[2];

/* panda data */
dat_t s_Panda;

/* diffuse texture name */
GLuint s_DiffTexName;

nn::fnd::ExpHeap s_AppHeap;
uptr s_AddrForGxHeap;
uptr s_AddrForMemMalloc;
const u32 s_GxHeapSize = 0x400000;
const u32 s_HeapSize = 0x400000;

demo::RenderSystem s_RenderSystem;

} // Namespace

static void LoadObjects(void)
{
    bool use_alpha;

    loadDAT( "rom:/resources/pandasub.dat", &s_Panda);

    /* load texture */
    glGenTextures(1, &s_DiffTexName);
    glActiveTexture(GL_TEXTURE0);
    /* stored in the texture collection */
    glBindTexture(GL_TEXTURE_2D, s_DiffTexName);
    /* limit texture sampling to the top level only */
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, 0);
    loadTexture( "rom:/resources/panda_d.tga", GL_TEXTURE_2D, 0, use_alpha, true);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

static void UnloadObjects(void)
{
    unloadDAT(&s_Panda);
    
    return;
}

void DrawFrame(void)
{
    static int f = 0;
    float step = 1.f;

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();
        
    nn::math::Matrix34 modelview;
    nn::math::Vector3 camPos(1.5f * sinf(-step * f * (float)M_PI / 180.f), 0.f, 1.5f * cosf(-step * f * (float)M_PI / 180.f));
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::Vector3 scale(0.8f, 0.8f, 0.8f);

    /* modelview setting */
    nn::math::MTX34LookAt(&modelview, &camPos, &camUp, &target);
    nn::math::MTX34MultScale(&modelview, &modelview, &scale);
    
    nn::math::Matrix44 m(modelview);
    glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    glDrawElements(GL_GEOMETRY_PRIMITIVE_DMP, s_Panda.patch[0].elm_size,
        GL_UNSIGNED_SHORT, (GLvoid*)(s_Panda.patch[0].elm_offset + s_Panda.obj[0].elm_offset));
    glFinish();

    s_RenderSystem.SwapBuffers();

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

    f++;
}

/* initialization */
static int Initialize(void)
{
    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    /* Initialize display */
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_AddrForGxHeap = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    s_RenderSystem.Initialize(s_AddrForGxHeap, s_GxHeapSize);

    s_AddrForMemMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_AddrForMemMalloc, s_HeapSize);

    s_ProgramID = glCreateProgram();
    s_ShaderID[0] = glCreateShader(GL_VERTEX_SHADER);
    s_ShaderID[1] = glCreateShader(GL_GEOMETRY_SHADER_DMP);

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(2, s_ShaderID, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    glAttachShader(s_ProgramID, s_ShaderID[0]);
    glAttachShader(s_ProgramID, s_ShaderID[1]);
    glAttachShader(s_ProgramID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_ProgramID, 0, "aPosition");
    glBindAttribLocation(s_ProgramID, 1, "aTexcoord0");
    glBindAttribLocation(s_ProgramID, 2, "aValence");

    glLinkProgram(s_ProgramID);
    glValidateProgram(s_ProgramID);
    glUseProgram(s_ProgramID);

    LoadObjects();

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
    glEnableVertexAttribArray(2);

    glClearColor(0.36f, 0.42f, 0.5f, 1.0f);
    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);

    nn::math::Matrix44 proj;
    
    /* Projection settings */
    nn::math::MTX44Frustum(&proj, -0.08f, 0.08f, -0.08f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH,
                         0.08f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 1000.f);

    glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glBindBuffer(GL_ARRAY_BUFFER, s_Panda.posVB);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, s_Panda.texVB);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, s_Panda.midxVB);
    glVertexAttribPointer(2, 4, GL_UNSIGNED_BYTE, GL_FALSE, 0, 0);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_Panda.idxVB);

    glBindTexture(GL_TEXTURE_2D, s_DiffTexName);
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_FragmentLighting.enabled"), GL_FALSE);
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].srcRgb"), GL_TEXTURE0, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].srcAlpha"), GL_TEXTURE0, GL_CONSTANT, GL_CONSTANT);

    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].srcRgb"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].srcAlpha"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);

    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_Texture[0].samplerType"), GL_TEXTURE_2D);

    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_Subdivision.fragmentLightingEnabled"), GL_FALSE);
    glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_Subdivision.level"), 2.f);

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    /* initialization */
    if (Initialize() >= 0)
    {
        while (1)
        {
            DrawFrame();
        }
    }
    
    UnloadObjects();
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_AddrForGxHeap));
    s_AppHeap.Free(reinterpret_cast<void*>(s_AddrForMemMalloc));
    s_AppHeap.Finalize();
}
