/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SubdivisionSimple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */

/*
 * Comments
 * --------
 *
 * This sample illustrates an adaptive tessellation of a Catmull-Clark subdivision
 * surface described as a set of subdivision primitives.
 * Subdivision primitives are generated from the original quadrangle mesh with at most one irregular
 * vertex per quadrangle face ("irregular" means there are not 4 adjacent edges).
 * The difference between the original and subdivided mesh is most noticeable at
 * the silhouette. Different vertex normals are preserved with respect to different adjacent faces
 * (not averaged), so normal (and other parameter) discontinuities are kept in the subdivided mesh.
 */

#include <nn.h>
#include <nn/fnd.h>
#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include "demo.h"

#include <string.h>
#include <math.h>

#include "Memory.h"

template <typename TYPE> class simple_vector
{
public:
    simple_vector()
    {
        element_size = 512;
        current_size = 0;
        element = (TYPE*)malloc(element_size * sizeof(TYPE));
    }
    ~simple_vector()
    {
        free(element);
    }
    int size(){ return current_size; }
    void clear(){ current_size = 0; }
    void push_back(TYPE d)
    {
        element[current_size++] = d;
        if (current_size == element_size)
        {
            TYPE* new_element = (TYPE*)malloc(2 * element_size * sizeof(TYPE));
            memcpy(new_element, element, sizeof(TYPE) * element_size);
            element_size *= 2;
            free(element);
            element = new_element;
        }
    }
    TYPE& operator[](int idx) { return element[idx]; }
    
    typedef TYPE*   iterator;
    iterator begin() { return element; }
    iterator end() { return &element[current_size]; }
    

protected:
    TYPE*   element;
    int     current_size;
    int     element_size;
};  // class simple_vector

namespace {

/* program id */
GLuint s_ProgramID;

/* shader id */
GLuint s_ShaderID[2];

struct object_info_t
{
    simple_vector<unsigned short> idx;
    simple_vector<float> v;
    simple_vector<float> n;
    simple_vector<float> c;
    simple_vector<float> t;
};

struct object_info_t* s_SimpleInfo;

GLuint s_ArrayBufferID;
GLuint s_ElementArrayBufferID;

nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
u32 s_GxHeapSize = 0x400000;
u32 s_HeapSize = 0x400000;

demo::RenderSystem s_RenderSystem;

} // Namespace

/* generate simple subdivision object */
void LoadObjects(void)
{   
    for (unsigned int i = 0; i < 7; i++)
    {
        for (unsigned int j = 0; j < 7; j++)
        {

            float x = -3.f + (float)i;
            float y = -3.f + (float)j;
            float z = 0.5f*(x*x + y*y);
            s_SimpleInfo->v.push_back(x);
            s_SimpleInfo->v.push_back(y);
            s_SimpleInfo->v.push_back(z);
            /*
            s_SimpleInfo->c.push_back(( (j*7+i)%3==0 )? 1.f: 0.f);
            s_SimpleInfo->c.push_back(( (j*7+i)%3==1 )? 1.f: 0.f);
            s_SimpleInfo->c.push_back(( (j*7+i)%3==2 )? 1.f: 0.f);
            */
            s_SimpleInfo->c.push_back(( (j*7+i)%2==0 )? 1.f: 0.f);
            s_SimpleInfo->c.push_back(( (j*7+i)%2==0 )? 1.f: 0.f);
            s_SimpleInfo->c.push_back(( (j*7+i)%2==0 )? 1.f: 0.f);
        }
    }

    for (unsigned int i = 0; i < 4; i++)
    {
        for (unsigned int j = 0; j < 4; j++)
        {
            s_SimpleInfo->idx.push_back(16);
            s_SimpleInfo->idx.push_back(( j+2 )*7 + ( i+2 ));
            s_SimpleInfo->idx.push_back(( j+2 )*7 + ( i+1 ));
            s_SimpleInfo->idx.push_back(( j+1 )*7 + ( i+1 ));
            s_SimpleInfo->idx.push_back(( j+1 )*7 + ( i+2 ));
            s_SimpleInfo->idx.push_back(( j+3 )*7 + ( i+2 ));
            s_SimpleInfo->idx.push_back(( j+3 )*7 + ( i+1 ));
            s_SimpleInfo->idx.push_back(( j+3 )*7 + ( i+0 ));
            s_SimpleInfo->idx.push_back(( j+2 )*7 + ( i+0 ));
            s_SimpleInfo->idx.push_back(( j+1 )*7 + ( i+0 ));
            s_SimpleInfo->idx.push_back(( j+0 )*7 + ( i+0 ));
            s_SimpleInfo->idx.push_back(( j+0 )*7 + ( i+1 ));
            s_SimpleInfo->idx.push_back(( j+0 )*7 + ( i+2 ));
            s_SimpleInfo->idx.push_back(( j+0 )*7 + ( i+3 ));
            s_SimpleInfo->idx.push_back(( j+1 )*7 + ( i+3 ));
            s_SimpleInfo->idx.push_back(( j+2 )*7 + ( i+3 ));
            s_SimpleInfo->idx.push_back(( j+3 )*7 + ( i+3 ));
        }
    }

    glGenBuffers(1, &s_ArrayBufferID);
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    
    glGenBuffers(1, &s_ElementArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    
    glBufferData(GL_ARRAY_BUFFER, sizeof(float) * (s_SimpleInfo->v.size() + s_SimpleInfo->c.size()), 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(float) * s_SimpleInfo->v.size(), &s_SimpleInfo->v[0]);
    glBufferSubData(GL_ARRAY_BUFFER, sizeof(float) * s_SimpleInfo->v.size(), sizeof(float) * s_SimpleInfo->c.size(), &s_SimpleInfo->c[0]);

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)(sizeof(float) * s_SimpleInfo->v.size()));
    
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
    
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned short) * s_SimpleInfo->idx.size(), &s_SimpleInfo->idx[0], GL_STATIC_DRAW);
}

void DrawFrame(void)
{
    static int f = 0;

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();

    nn::math::Matrix34 mv, rot;
    nn::math::Vector3 camPos(0.f, 0.f, 15.f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::Vector3 trans(0.f, 0.f, 10*sinf(static_cast<f32>(f) / 36.f));
    
    /* Projection settings */
    /* modelview setting */
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::MTX34MultTranslate(&mv, &mv, &trans);
    
    nn::math::MTX34RotXYZDeg(&rot, 0.f, -static_cast<f32>(f), 0.f);
    nn::math::MTX34Mult(&mv, &mv, &rot);
    
    nn::math::Matrix44 m(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    glDrawElements(GL_GEOMETRY_PRIMITIVE_DMP, (GLsizei)s_SimpleInfo->idx.size(), GL_UNSIGNED_SHORT, 0);
    glFinish();

    s_RenderSystem.SwapBuffers();

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

    f++;
}

/* initialization */
int Initialize(void)
{
    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    /* Initialize display */
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);

    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    s_SimpleInfo = new object_info_t();

    s_ProgramID = glCreateProgram();
    s_ShaderID[0] = glCreateShader(GL_VERTEX_SHADER);
    s_ShaderID[1] = glCreateShader(GL_GEOMETRY_SHADER_DMP);

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(2, s_ShaderID, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    glAttachShader(s_ProgramID, s_ShaderID[0]);
    glAttachShader(s_ProgramID, s_ShaderID[1]);
    glAttachShader(s_ProgramID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_ProgramID, 0, "aPosition");
    glBindAttribLocation(s_ProgramID, 1, "aColor");

    glLinkProgram(s_ProgramID);
    glValidateProgram(s_ProgramID);
    glUseProgram(s_ProgramID);

    glClearColor(0.36f, 0.42f, 0.5f, 1.0f);
    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glDisable(GL_CULL_FACE);

    LoadObjects();

    /* set subdivision level */
    glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_Subdivision.level"), 1);
    
    /* set fragment lighting disabling */
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_Subdivision.fragmentLightingEnabled"), GL_FALSE);
    
    nn::math::Matrix44 proj;
    
    /* Projection settings */
    nn::math::MTX44Frustum(&proj, -0.07f, 0.07f, -0.07f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH,
                 0.07f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);
    glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].srcRgb"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].srcAlpha"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    /* initialization */
    if (Initialize() >= 0)
    {
        /* Enter loop */
        while (1)
        {
            DrawFrame();
        }
    }
    
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Finalize();
}
