/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     EarlyDepthTestSample.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/ulcd.h>
#include <nn/font.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"
#include "demo.h"

#include "Util.h"

#include "Object.h"

#include <nn/gd.h>

// Does not wait for VSync in order to make early depth test effects more easily perceptible.
//#define WAIT_VSYNC
#define USE_EARLY_DEPTH_TEST
#define USE_DOUBLE_COMMAND_LIST
// Displays the number of fragments input into the per-fragment operation module if enabled.
//#define GET_PROFILING_RESULT

// Display buffer
u32 s_Displaybuffer0[2];
u32 s_Displaybuffer1[2];
s32 s_CurrentDisplaybuffer0 = 0;
s32 s_CurrentDisplaybuffer1 = 0;

// Depth stencil state
static nn::gd::DepthStencilState* s_pDepthStencil = 0;

// Frame Buffer
static nn::gd::Texture2DResource*  s_pMainColorBufferResource        = 0;
static nn::gd::Texture2DResource*  s_pMainDepthStencilBufferResource = 0;
static nn::gd::RenderTarget*       s_pMainRenderTarget               = 0;
static nn::gd::DepthStencilTarget* s_pMainDepthStencilTarget         = 0;

// Block 32 mode must be set when using an early depth test.
// A buffer having the same size as the screen cannot be used for this reason.
#ifdef USE_EARLY_DEPTH_TEST
static u32 s_Disp0Width  = 256;
static u32 s_Disp0Height = 416;
static u32 s_Disp1Width  = 256;
static u32 s_Disp1Height = 320;
static nn::gd::Memory::MemoryLayout s_BlockSize = nn::gd::Memory::LAYOUT_BLOCK_32;
#else
static u32 s_Disp0Width  = 240;
static u32 s_Disp0Height = 400;
static u32 s_Disp1Width  = 240;
static u32 s_Disp1Height = 320;
static nn::gd::Memory::MemoryLayout s_BlockSize = nn::gd::Memory::LAYOUT_BLOCK_8;
#endif

// Shader
static nn::gd::ShaderBinary*   s_pShaderBinary   = 0;
static nn::gd::ShaderPipeline* s_pShaderPipeline = 0;
static nn::gd::Shader*         s_pVertexShader   = 0;

static nn::gd::UniformLocation s_UniformLocationProj;
static nn::gd::UniformLocation s_UniformLocationView;

// Input layout
static nn::gd::InputLayout* s_pInputLayout = 0;

// Vertex buffer
static nn::gd::VertexBufferResource* s_pCubeBufferPos      = 0;
static nn::gd::VertexBufferResource* s_pCubeBufferNrm      = 0;
static nn::gd::VertexBufferResource* s_pCubeBufferTexCoord = 0;
static nn::gd::VertexBufferResource* s_pCubeIdx            = 0;

// Combiner
static nn::gd::CombinerState* s_pCombinerState = 0;

// Texture
static nn::gd::Texture2DResource* s_pTexture2DResource = 0;
static nn::gd::Texture2D*         s_pTexture2D         = 0;
static nn::gd::SamplerState*      s_pSamplerState      = 0;

// Matrix
static nn::math::MTX44  s_ProjMtx;
static nn::math::MTX34  s_ViewMtx;

// ExpHeap for app.
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

// FPS display
#define FPS_AVG_FRAME_COUNT 100
static u32 s_CurrentFpsCount = 0;
static f64 s_TmpFpsSum = 0.0;

#ifdef USE_DOUBLE_COMMAND_LIST
    u32 s_CommandList[2];
    u32 s_CommandIdx = 0;
#else
    u32 s_CommandList;
#endif

#define PRINT(__msg) NN_LOG __msg

void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

s64 GetTimeMicroSeconds(void)
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

static void DrawCube(void)
{
    nn::gd::VertexBufferResource* buffers[] = { s_pCubeBufferPos, s_pCubeBufferNrm, s_pCubeBufferTexCoord };
    u32 offsets[] = { 0, 0, 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 3, buffers, offsets);

    nn::gd::VertexInputStage::SetIndexBuffer(s_pCubeIdx, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
    nn::gd::System::DrawIndexed(36, 0);
}

s32 DrawFrame(void)
{
    static s32 f = 0;
    s64 startTime = GetTimeMicroSeconds();

    // Gets display information.
    static s32 disp0Format = 0, disp0Width = 0, disp0Height = 0;
    if (disp0Width == 0)
    {
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH,  &disp0Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0Height);
    }

    // Sets render targets.
    nn::gd::OutputStage::SetRenderTarget(s_pMainRenderTarget);
    nn::gd::OutputStage::SetDepthStencilTarget(s_pMainDepthStencilTarget);

    // Viewport settings
    nn::gd::Viewport viewPort(0, 0, nn::gx::DISPLAY0_WIDTH * 2, nn::gx::DISPLAY0_HEIGHT * 2);
    nn::gd::RasterizerStage::SetViewport(viewPort);

   // Model view matrix
    nn::math::Vector3 camPos( 0.f, 0.f, 14.f );
    nn::math::Vector3 camUp( 0.f, 1.f, 0.f );
    nn::math::Vector3 target( 0.f, 0.f, 0.f );
    nn::math::MTX34LookAt( &s_ViewMtx, &camPos, &camUp, &target );

    // Display 0
    {

        u8 clearColor[] = {80, 100, 125, 255};

        nn::gd::Memory::ClearTargets(s_pMainRenderTarget, s_pMainDepthStencilTarget, clearColor, 1.f, 0);

#ifdef USE_EARLY_DEPTH_TEST
        nn::gd::RasterizerStage::SetEarlyDepthTest(GD_TRUE, nn::gd::RasterizerStage::EARLYDEPTH_FUNCTION_LESS);
        nn::gd::RasterizerStage::ClearEarlyDepth(1.f);
#endif // USE_EARLY_DEPTH_TEST

        // If an early depth test has been enabled, be sure to begin rendering objects from the front toward the back of the screen.
        // Performance is not improved by drawing objects in the foreground after objects in the background because the early depth buffer does not update.
        // 
        static f32 frontPos = 12.0f;
        if( frontPos <= -60.0f )frontPos = 12.0f;

        // Object in screen foreground.
        for(int i = 0; i < 16; i++)
        {
            // Model view matrix
            nn::math::Matrix34 trans, vTmp;
            nn::math::Vector3 transVec( 0.f, frontPos + ( i * 3.0f ), 7.2f );
            nn::math::MTX34Translate( &trans, &transVec );
            nn::math::MTX34Mult(&vTmp, &s_ViewMtx, &trans);

            nn::math::Matrix44 tmp(vTmp);
            nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipeline, s_UniformLocationView, static_cast<f32*>(tmp));
            // Rendering a cube
            DrawCube();
        }

        frontPos -= 0.05f;

        // The fragment load has been increased to the extreme in order to clearly show the effect of the early depth test.
        nn::gd::LightingStage::light[1].EnableLight(GD_TRUE);
        nn::gd::LightingStage::light[2].EnableLight(GD_TRUE);
        nn::gd::LightingStage::light[3].EnableLight(GD_TRUE);
        nn::gd::LightingStage::SetLayerConfiguration(nn::gd::LightingStage::LAYER_CONFIGURATION_7);
        nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

        static f32 backPos = -12.0f;
        if( backPos >= -3.0f )backPos = -12.0f;

        // Object in screen background.
        for(int i = 0; i < 8; i++)
        {
            // Model view matrix
            nn::math::Matrix34 trans, vTmp;
            nn::math::Vector3 transVec( 0.f, backPos + ( i * 3.0f ), 5.2f );
            nn::math::MTX34Translate( &trans, &transVec );
            nn::math::MTX34Mult(&vTmp, &s_ViewMtx, &trans);

            nn::math::Matrix44 tmp(vTmp);
            nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipeline, s_UniformLocationView, static_cast<f32*>(tmp));
            // Rendering a cube
            DrawCube();
        }

        backPos += 0.05f;

        nn::gd::LightingStage::light[1].EnableLight(GD_FALSE);
        nn::gd::LightingStage::light[2].EnableLight(GD_FALSE);
        nn::gd::LightingStage::light[3].EnableLight(GD_FALSE);
        nn::gd::LightingStage::SetLayerConfiguration(nn::gd::LightingStage::LAYER_CONFIGURATION_0);
        nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

        // Send the contents in the frame buffer to the display buffer
        s32 dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_Displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_pMainColorBufferResource,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0Width,                                            //dstWidth
            disp0Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_2X2,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

#ifdef USE_DOUBLE_COMMAND_LIST

    // Wait for termination of commands for the previous frame
    nngxBindCmdlist( s_CommandList[1 - s_CommandIdx] );
    nngxWaitCmdlistDone();
    // Clear commands for the previous frame
    nngxClearCmdlist();
#else
    nngxWaitCmdlistDone();
    nngxClearCmdlist();
#endif

    // Swap buffers
#ifdef USE_DOUBLE_COMMAND_LIST
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);
#endif

    nngxBindDisplaybuffer(s_Displaybuffer0[s_CurrentDisplaybuffer0]);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_Displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);

#ifndef USE_DOUBLE_COMMAND_LIST
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);
#endif

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

#ifdef GET_PROFILING_RESULT
    static u32 resFragment = 0, oldResFragment = 0;
    nngxGetProfilingResult( NN_GX_PROFILING_FRAGMENT, &resFragment );
    NN_LOG("Fragment Count: %d pixels\n", resFragment - oldResFragment);
    oldResFragment = resFragment;
#endif

#ifdef USE_DOUBLE_COMMAND_LIST

    // Run commands on the command list.
    nngxBindCmdlist( s_CommandList[s_CommandIdx] );
    nngxRunCmdlist();

    // Swap command lists.
    s_CommandIdx = 1 - s_CommandIdx;
    nngxBindCmdlist( s_CommandList[s_CommandIdx] );

#else
    nngxRunCmdlist();
#endif // USE_DOUBLE_COMMAND_LIST

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    // FPS display
    s_TmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_CurrentFpsCount == FPS_AVG_FRAME_COUNT)
    {
        f64 avgtime = s_TmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.5f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_CurrentFpsCount = 0;
        s_TmpFpsSum = 0.0;
    }

    f++;

    return 0;
}

u32 convertGLFormatTypeToGDFormat(unsigned int format, unsigned int type)
{
    u32 res = 1000;
    if (type != GL_UNSIGNED_BYTE) return 1000;
    switch (format)
    {
        case GL_RGB:    res = nn::gd::Resource::FORMAT_RGB_888;      break;
        case GL_RGBA:   res = nn::gd::Resource::FORMAT_RGBA_8888;    break;
        default: return 1000;
    }
    return res;
}

static void InitShader(void)
{
    // Load the shader.
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    // Create a shader binary.
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &s_pShaderBinary);
    NN_UTIL_PANIC_IF_FAILED(res);
    // Create shader.
    res = nn::gd::ShaderStage::CreateShader(s_pShaderBinary, 0, &s_pVertexShader);
    NN_UTIL_PANIC_IF_FAILED(res);
    // Create shader pipeline.
    res = nn::gd::ShaderStage::CreateShaderPipeline(s_pVertexShader, NULL, &s_pShaderPipeline);
    NN_UTIL_PANIC_IF_FAILED(res);


    // Set shader pipeline.
    nn::gd::ShaderStage::SetShaderPipeline(s_pShaderPipeline);

    s_UniformLocationProj = nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipeline, "uProjection");
    s_UniformLocationView = nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipeline, "uModelView");

    NN_ASSERT(s_UniformLocationProj.IsValid());
    NN_ASSERT(s_UniformLocationView.IsValid());

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void)
{
    // Set cube vertex buffer.
    nnResult res;
    nn::gd::VertexBufferResourceDescription descCube;
    descCube.m_ByteSize = sizeof(s_CubePos);
    descCube.m_MemLocation = nn::gd::Memory::FCRAM;
    res = nn::gd::Resource::CreateVertexBufferResource(&descCube, s_CubePos, &s_pCubeBufferPos);
    NN_UTIL_PANIC_IF_FAILED(res);

    descCube.m_ByteSize = sizeof(s_CubeNrm);
    res = nn::gd::Resource::CreateVertexBufferResource(&descCube, s_CubeNrm, &s_pCubeBufferNrm);
    NN_UTIL_PANIC_IF_FAILED(res);

    descCube.m_ByteSize = sizeof(s_CubeTexCoord);
    res = nn::gd::Resource::CreateVertexBufferResource(&descCube, s_CubeTexCoord, &s_pCubeBufferTexCoord);
    NN_UTIL_PANIC_IF_FAILED(res);

    descCube.m_ByteSize = sizeof(s_CubeIdx);
    res = nn::gd::Resource::CreateVertexBufferResource(&descCube, s_CubeIdx, &s_pCubeIdx);
    NN_UTIL_PANIC_IF_FAILED(res);

    // Input layout
    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition",  nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 1, "aNormal",    nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 2, "aTexCoord",  nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
    };

    u32 strides[] = { sizeof(f32) * 3,  sizeof(f32) * 3,  sizeof(f32) * 3};
    res = nn::gd::VertexInputStage::CreateInputLayout(descs, 3, strides, s_pVertexShader, &s_pInputLayout);
    nn::gd::VertexInputStage::SetInputLayout(s_pInputLayout);

    // Load texture.
    u32 width, height, format, type;
    nn::gd::Resource::Format imgFormat;
    u8* pixels = 0;
    nn::gd::Resource::NativeFormat nativeFormat;

    pixels = GetTextureData_TGA("rom:/resources/knight.tga", width, height, format, type, &s_AppHeap);
    NN_NULL_ASSERT(pixels);
    imgFormat = (nn::gd::Resource::Format)convertGLFormatTypeToGDFormat(format, type);
    res = nn::gd::Resource::Helper::ConvertTextureResourceToNativeFormat(imgFormat, width, height, pixels, pixels, &nativeFormat);
    NN_UTIL_PANIC_IF_FAILED(res);
    nn::gd::Texture2DResourceDescription Text2DResDesc = {width, height, 1, nativeFormat, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, pixels, GD_FALSE, &s_pTexture2DResource);
    NN_UTIL_PANIC_IF_FAILED(res);
    nn::gd::Texture2DDescription tex2DDesc = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(s_pTexture2DResource, &tex2DDesc, &s_pTexture2D);
    NN_UTIL_PANIC_IF_FAILED(res);
    s_AppHeap.Free(pixels);

    // Set sampler
    nn::gd::SamplerStateDescription descSampler_0;
    descSampler_0.SetMinFilter(nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR);
    descSampler_0.SetMagFilter(nn::gd::TextureStage::SAMPLER_MAG_FILTER_NEAREST);
    descSampler_0.SetMinLod(0);
    nn::gd::TextureStage::CreateSamplerState(&descSampler_0, &s_pSamplerState);

    // Set texture
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_pTexture2D);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, s_pSamplerState);
}

static void InitializeGd(void)
{
    // Depth stencil state
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &s_pDepthStencil);
    nn::gd::OutputStage::SetDepthStencilState(s_pDepthStencil, 0);

    // Projection matrix
    f32 aspect = static_cast<f32>(nn::gx::DISPLAY0_HEIGHT) / static_cast<f32>(nn::gx::DISPLAY0_WIDTH);
    nn::math::MTX44Frustum(&s_ProjMtx, -0.04f, 0.04f, -0.04f * aspect, 0.04f * aspect, 0.22, 100.0f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipeline, s_UniformLocationProj, static_cast<f32*>(s_ProjMtx));

    // Set combiner
    if (!s_pCombinerState)
    {
        nn::gd::CombinerDescription combiner;
        combiner.ToDefault();
        combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_CONSTANT);
        combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_CONSTANT);
        combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
        combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
        combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MODULATE);
        combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
        combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
        nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_pCombinerState);
    }
    nn::gd::CombinerStage::SetTextureCombinerState(s_pCombinerState);

    // Light settings
    nn::gd::LightingStage::light[0].EnableLight(GD_TRUE);
    nn::gd::LightingStage::light[0].SetColorDiffuse(140, 140, 140);
    nn::gd::LightingStage::light[0].SetPosition(130, 10, 100);

    nn::gd::LightingStage::light[1].SetColorDiffuse(60, 220, 40);
    nn::gd::LightingStage::light[1].SetPosition(50, 90, 60);

    nn::gd::LightingStage::light[2].SetColorDiffuse(40, 40, 220);
    nn::gd::LightingStage::light[2].SetPosition( -60, -120, 20);

    nn::gd::LightingStage::light[3].SetColorDiffuse(220, 40, 40);
    nn::gd::LightingStage::light[3].SetPosition( 0, 0, -80);

    nn::gd::LightingStage::SetGlobalColorAmbient( 20, 20, 20);

}

s32 Initialize(void)
{
    // Initialize the file system
    nn::fs::Initialize();

    // Memory initialization
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    // Allocate memory from device memory.
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);

    // Initialize nngx
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    // Initialize command list
#ifdef USE_DOUBLE_COMMAND_LIST
    nngxGenCmdlists( 2, &s_CommandList[0] );
    nngxBindCmdlist( s_CommandList[0] );
    nngxCmdlistStorage( 0x40000, 128 );
    nngxSetCmdlistParameteri( NN_GX_CMDLIST_RUN_MODE, NN_GX_CMDLIST_SERIAL_RUN );
    nngxClearCmdlist();

    nngxBindCmdlist(s_CommandList[1]);
    nngxCmdlistStorage(0x40000, 128);
    nngxSetCmdlistParameteri( NN_GX_CMDLIST_RUN_MODE, NN_GX_CMDLIST_SERIAL_RUN );
    nngxClearCmdlist();

    nngxBindCmdlist( s_CommandList[0] );
#else
    nngxGenCmdlists(1, &s_CommandList);
    nngxBindCmdlist(s_CommandList);
    nngxCmdlistStorage(0x40000, 128);

    nngxRunCmdlist();
#endif

    // Initialize display buffer
    nngxGenDisplaybuffers(2, s_Displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_Displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, s_Disp0Width, s_Disp0Height, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_Displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, s_Disp0Width, s_Disp0Height, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_Displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_Displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, s_Disp1Width, s_Disp1Height, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_Displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, s_Disp1Width, s_Disp1Height, NN_GX_MEM_FCRAM);

    // Initialize GD
    nnResult res = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(res) ) return -1;
    // Error handling
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    // Initialize framebuffer
    // Allocate a buffer four times the usual size in order to apply anti-aliasing.
    nn::gd::Texture2DResourceDescription tex2DResDescColorBuffer = {s_Disp0Width * 2, s_Disp0Height * 2, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, s_BlockSize, nn::gd::Memory::VRAMA};
    res = nn::gd::Resource::CreateTexture2DResource(&tex2DResDescColorBuffer, 0, GD_FALSE, &s_pMainColorBufferResource);
    NN_UTIL_PANIC_IF_FAILED(res);

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_pMainColorBufferResource, &descRenderTarget, &s_pMainRenderTarget);
    NN_UTIL_PANIC_IF_FAILED(res);
    // Depth stencil buffer
    nn::gd::Texture2DResourceDescription tex2DResDescDepthStencilBuffer = {s_Disp0Width * 2, s_Disp0Height * 2, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, s_BlockSize, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&tex2DResDescDepthStencilBuffer, 0, GD_FALSE, &s_pMainDepthStencilBufferResource);
    NN_UTIL_PANIC_IF_FAILED(res);

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_pMainDepthStencilBufferResource, &descDepthStencilTarget, &s_pMainDepthStencilTarget);
    NN_UTIL_PANIC_IF_FAILED(res);

    // Viewport settings
    nn::gd::TargetProperties renderTargetProperty;
    nn::gd::OutputStage::GetRenderTargetProperties( s_pMainRenderTarget, &renderTargetProperty );
    nn::gd::Viewport viewPort(0, 0, renderTargetProperty.m_Width, renderTargetProperty.m_Height);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    // Initialize the shader
    InitShader();
    // Initialize GD
    InitializeGd();
    // Initialize the object
    InitObjects();

    return 0;
}

static void Finalize(void)
{
    // Release the Gd object.
    if (s_pCombinerState)      nn::gd::CombinerStage::ReleaseTextureCombinerState(s_pCombinerState);

    if (s_pDepthStencil)       nn::gd::OutputStage::ReleaseDepthStencilState(s_pDepthStencil);

    if (s_pTexture2DResource)  nn::gd::Resource::ReleaseTexture2DResource(s_pTexture2DResource);
    if (s_pTexture2D)          nn::gd::TextureStage::ReleaseTexture2D(s_pTexture2D);

    if (s_pSamplerState)       nn::gd::TextureStage::ReleaseSamplerState(s_pSamplerState);

    if (s_pCubeBufferPos)      nn::gd::Resource::ReleaseVertexBufferResource(s_pCubeBufferPos);
    if (s_pCubeBufferTexCoord) nn::gd::Resource::ReleaseVertexBufferResource(s_pCubeBufferTexCoord);
    if (s_pCubeBufferNrm)      nn::gd::Resource::ReleaseVertexBufferResource(s_pCubeBufferNrm);
    if (s_pCubeIdx)            nn::gd::Resource::ReleaseVertexBufferResource(s_pCubeIdx);
    if (s_pInputLayout)        nn::gd::VertexInputStage::ReleaseInputLayout(s_pInputLayout);

    if (s_pShaderPipeline)     nn::gd::ShaderStage::ReleaseShaderPipeline(s_pShaderPipeline);
    if (s_pVertexShader)       nn::gd::ShaderStage::ReleaseShader(s_pVertexShader);
    if (s_pShaderBinary)       nn::gd::ShaderStage::ReleaseShaderBinary(s_pShaderBinary);

    if (s_pMainRenderTarget)               nn::gd::OutputStage::ReleaseRenderTarget(s_pMainRenderTarget);
    if (s_pMainColorBufferResource)        nn::gd::Resource::ReleaseTexture2DResource(s_pMainColorBufferResource);
    if (s_pMainDepthStencilTarget)         nn::gd::OutputStage::ReleaseDepthStencilTarget(s_pMainDepthStencilTarget);
    if (s_pMainDepthStencilBufferResource) nn::gd::Resource::ReleaseTexture2DResource(s_pMainDepthStencilBufferResource);

    // Finalize the GD library
    nn::gd::System::Finalize();

    // Release the display buffer
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_Displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_Displaybuffer1);
    s_Displaybuffer0[0] = s_Displaybuffer0[1] = 0;
    s_Displaybuffer1[0] = s_Displaybuffer1[1] = 0;
    
    // Release the command list
#ifdef USE_DOUBLE_COMMAND_LIST
    nngxDeleteCmdlists(2, s_CommandList);
#else
    nngxDeleteCmdlists(1, &s_CommandList);
#endif

    nngxFinalize();

    // Release memory
    demo::FinalizeMemoryManager();
}

void nnMain(void)
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }


    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}
