﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     meshUtil.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <wchar.h>

#include <nn/fs.h>
#include <nn/gx.h>
#include "meshUtil.h"
#include "binFileUtil.h"

#include "demo.h"


CSampleMesh* loadMesh(const MeshDeclaration &mdecl, const char* szDataPath)
{
    if (!szDataPath)   return NULL;

    CSampleMesh *mesh = 0;
    char path[256];
    sprintf(path, "%s%s", szDataPath, mdecl.fileName);
    NN_LOG("\nLoading sample mesh [%s]...\n", path );

    CBinalyFile file;
    if(file.load(path))
    {
        mesh = new CSampleMesh;
        mesh->createMesh( mdecl, file.getData());
    }
    else
        NN_LOG("ERROR: couldn't load the mesh data from the binary file %s\n", path );

    return mesh;
}


//-----------------------------------------------------------------------------
//      Class CSampleMesh
//-----------------------------------------------------------------------------
CSampleMesh::CSampleMesh(void)
{
    // reset VBO names
    memset(&m_vbo1, 0, sizeof(m_vbo1));
}

CSampleMesh::~CSampleMesh(void)
{
    // Free up the buffers
    for(int a = 0; a < 8; a++)
    {
        if(m_vbo1[a] != 0)
        {
            nn::gd::Resource::ReleaseVertexBufferResource(m_vbo1[a]);
        }
    }
}

bool CSampleMesh::createMesh(const MeshDeclaration &mdecl, const void *meshData)
{
    const char *meshBytes = reinterpret_cast<const char*>(meshData);

    // Mesh declaration structure
    m_mdecl = mdecl;

    // For platforms that cannot use the index array, the vertex data is decompressed in index order.
    // 
    unsigned int vertexDataSize = mdecl.numVertices * mdecl.stride;
    const char *vertexData = & meshBytes[0];

    NN_LOG("  Mesh info = %i vertices, %i indices\n", mdecl.numVertices, mdecl.numIndices );
    NN_LOG("  Vertex data size = %i (Bytes)\n", vertexDataSize );

    // Read index data
    const unsigned int *indexData =
        reinterpret_cast<const unsigned int*>(& meshBytes[vertexDataSize]);

    // Converts the format from vertex data with packed attributes to vertex data that has an array for each attribute.
    // At the same time, the vertex data is decompressed in index order and converted to a primitive without indices.
    // 
    struct VertexDecl
    {
        int offset;
        unsigned int size;
    };
    const VertexDecl vertexDecl[8] =
    {
        { mdecl.position,   3 * sizeof(float) },
        { mdecl.normal,     3 * sizeof(float) },
        { mdecl.color,      4 * sizeof(unsigned char) },
        { mdecl.texcoord,   2 * sizeof(float) },
        { mdecl.tangent,    4 * sizeof(float) },
        { mdecl.binormal,   3 * sizeof(float) },
        { mdecl.index,      4 * sizeof(unsigned char) },
        { mdecl.weight,     4 * sizeof(float) },
    };

    for(int a = 0; a < 8; a++)
    {
        // If the mesh doesn't contain a particular attribute (offset = -1), we skip it
        const VertexDecl &vd = vertexDecl[a];
        if(vd.offset < 0) continue;

        // Format the data before sending them to GPU memory
        char *bufferData = new char [vd.size * mdecl.numIndices];

        char *d = bufferData;

        for(unsigned int i = 0; i < mdecl.numIndices; i++)
        {
            const char *s = vertexData + vd.offset + indexData[i] * mdecl.stride;
            memcpy(d, s, vd.size);
            d += vd.size;
        }
   
        nn::gd::VertexBufferResourceDescription desc;

        desc.m_ByteSize = vd.size * mdecl.numIndices;
        desc.m_MemLocation = nn::gd::Memory::FCRAM;
        nnResult res = nn::gd::Resource::CreateVertexBufferResource(&desc, bufferData, &m_vbo1[a]);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        delete [] bufferData;
    }
    return true;
}

//-----------------------------------------------------------------------------
//      Rendering function
//-----------------------------------------------------------------------------
struct VertexAttribs
{
    GLint       index;
    GLint       size;
    GLenum      type;
    GLboolean   normalize;
    NN_PADDING3;
};


void CSampleMesh::drawMesh(GLint positionAttr, GLint normalAttr, GLint textAttr,
                           GLint colorAttr, GLint tangentAttr, GLint binormalAttr,
                           GLint skinningIndexAttr, GLint skinningWeightAttr)
{
    // Set the attributes we want to use
    const VertexAttribs vertexAttribs[8] =
    {
        { positionAttr, 3, GL_FLOAT,        GL_FALSE },     // position
        { normalAttr,   3, GL_FLOAT,        GL_FALSE },     // normal
        { colorAttr,    4, GL_UNSIGNED_BYTE,GL_TRUE },      // Vertex color
        { textAttr, 2, GL_FLOAT,        GL_FALSE },         // Text coords
        { tangentAttr,  4, GL_FLOAT,        GL_FALSE },     // tangent
        { binormalAttr, 3, GL_FLOAT,        GL_FALSE },     // Binormal
        { skinningIndexAttr,    4, GL_UNSIGNED_BYTE,GL_FALSE },     // index
        { skinningWeightAttr,   4, GL_FLOAT,        GL_FALSE },     // weight
    };

    nn::gd::VertexBufferResource* buffers[8] =
                      {0, 0, 0, 0, 0, 0, 0, 0};
    u32 offsets[] = { 0, 0, 0, 0, 0, 0, 0, 0 };
    int i=0;

    for(int a = 0; a < 8; a++)
    {
        const VertexAttribs &va = vertexAttribs[a];
        if((va.index < 0) || (m_vbo1[a] == 0)) continue;

        buffers[i++] = m_vbo1[a];
    }

    nn::gd::VertexInputStage::SetVertexBuffers(0, i, buffers, offsets);
    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
    nn::gd::System::Draw(m_mdecl.numIndices, 0);
}


////-----------------------------------------------------------------------------
//// Execute rendering for morphing.
////-----------------------------------------------------------------------------
///*static*/
//bool CSampleMesh::drawMorphMesh(
//  ASPContext /*dev*/, CSampleMesh *_target[3])
//{
//  int numActiveAttrib = 0;
//  GLint activeAttribs[4*8];
//
//  // Reset the OpenGL error.
//  GLenum error = glGetError();
//
//  // Check the vertex attribute that requires the ASP shader.
//  const ASPShaderInputs *attrib = aspGetShaderInputs();
//  CSampleMesh *target[4] = { this, _target[0], _target[1], _target[2] };
//  int numPosition         = attrib->numPosition;
//  const GLint *position   = attrib->paramPosition;
//  int numNormal           = attrib->numNormal;
//  const GLint *normal     = attrib->paramNormal;
//  GLint color             = attrib->paramColor;
//  GLint texcoord          = attrib->paramTexCoord[0];
//  int numTangent          = attrib->getNumTangentAndBinormal();
//  const GLint *tangent    = attrib->paramTangent;
//  const GLint *binormal   = attrib->paramBinormal;
//  GLint index             = attrib->paramBlendIndex;
//  GLint weight            = attrib->paramBlendWeight;
//
//  // Process each morph target in a loop.
//  for(int t = 0; t < 4; t++)
//  {
//      // If there is no morph target, stop there.
//      if(target[t] == NULL) break;
//
//      // Bind the VBO that sets the vertex's attribute data
//      // Look at the input attribute count, and determine whether to bind this morph target VBO.
//      // 
//      const VertexAttribs vertexAttribs[8] =
//      {
//          { ((t < numPosition)? position[t]: -1), 3, GL_FLOAT, GL_FALSE },
//          { ((t < numNormal)? normal[t]: -1),     3, GL_FLOAT, GL_FALSE },
//          { ((t == 0)? color: -1),                4, GL_UNSIGNED_BYTE,GL_TRUE },
//          { ((t == 0)? texcoord: -1),             2, GL_FLOAT, GL_FALSE },
//          { ((t < numTangent)? tangent[t]: -1),   4, GL_FLOAT, GL_FALSE },
//          { ((t < numTangent)? binormal[t]: -1),  3, GL_FLOAT, GL_FALSE },
//          { ((t == 0)? index: -1),                4, GL_UNSIGNED_BYTE,GL_FALSE },
//          { ((t == 0)? weight: -1),               4, GL_FLOAT, GL_FALSE },
//      };
//      for(int a = 0; a < 8; a++)
//      {
//          const VertexAttribs &va = vertexAttribs[a];
//          if((va.index < 0) || (target[t]->m_vbo[a] == 0)) continue;
//
//          glBindBuffer(GL_ARRAY_BUFFER, target[t]->m_vbo[a]);
//          glVertexAttribPointer(va.index, va.size, va.type, va.normalize, 0, 0);
//          glEnableVertexAttribArray(va.index);
//
//          activeAttribs[numActiveAttrib++] = va.index;
//
//          // Error check
//          error = glGetError();
//          if(error != GL_NO_ERROR)
//          {
//              printf("ERROR: glVertexAttribPointer() was failed.\n");
//              break;
//          }
//      }
//      if(error != GL_NO_ERROR) break;
//  }
//
//  // Render mesh.
//  if(error == GL_NO_ERROR)
//  {
//      glDrawArrays(GL_TRIANGLES, 0, target[0]->m_mdecl.numIndices);
//  }
//
//  // Disable the VBO.
//  for(int a = 0; a < numActiveAttrib; a++)
//  {
//      glDisableVertexAttribArray(activeAttribs[a]);
//  }
//  glBindBuffer(GL_ARRAY_BUFFER, 0);
//
//  return true;
//}

////-----------------------------------------------------------------------------
//void CMotion::init(const void *data, unsigned int size, int numInstances)
//{
//  m_numInstances = numInstances;
//  m_numStep = size / (numInstances * sizeof(float16));
//  m_data = static_cast<const float16*>(data);
//}
//
////-----------------------------------------------------------------------------
//const ASPMatrix4 CMotion::getMatrix(int index, int step, int loopStep) const
//{
//  assert(index < m_numInstances);
//  assert(m_data);
//
//  // Process the loop playback.
//  if(loopStep <= step)
//  {
//      // If the playback position is past m_numStep, use loopStep as the loop point
//      // Playback loop.
//      const int loopLength = m_numStep - loopStep;
//      step = loopStep + ((step - loopStep) % loopLength);
//  }
//
//  return m_data[index * m_numStep + step];
//}
