﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     MultiShaders.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"
#include "demo.h"

#include <string.h>
#include <math.h>

#include <nn/gd.h>

#define WAIT_VSYNC

#define DMP_PI  (3.1415926f)
#define REV_PI  (1.0f/DMP_PI)
#ifndef PI
#define PI DMP_PI
#endif

enum primitive_type_t
{
    TRIANGLES,
    TRIANGLE_STRIP,
    LINE_STRIP,
    SUBD_PRIM_DMP,
    TRIANGLES_SILHOUETTE_DMP,
    TRIANGLE_STRIP_SILHOUETTE_DMP,
    LINES,
    POINTS
};

/* definitions to specify use_arrays parameter */
#define USE_ARRAY_POSITION          0x0001
#define USE_ARRAY_COLOR             0x0002
#define USE_ARRAY_COLOR_BYTE        0x0004
#define USE_ARRAY_TEXCOORD0         0x0008
#define USE_ARRAY_TEXCOORD1         0x0010
#define USE_ARRAY_TEXCOORD2         0x0020
#define USE_ARRAY_TEXCOORD          (USE_ARRAY_TEXCOORD0 | USE_ARRAY_TEXCOORD1 | USE_ARRAY_TEXCOORD2)
#define USE_ARRAY_NORMAL            0x0040
#define USE_ARRAY_TANGENT           0x0080
#define USE_ARRAY_PALETTE           0x0100
#define USE_ARRAY_TEXCOORD0_3COMP   0x0200

u32 s_IndexCount;

/* definitions to specify array_type */
typedef enum tag_array_type_t
{
    INDEPENDENT_ARRAY,
    INTERLEAVED_ARRAY
} array_type_t;

typedef struct tag_position_t
{
    GLfloat     x, y, z;
} position_t;

typedef struct tag_color_t
{
    GLfloat     r, g, b, a;
} color_t;

typedef struct tag_byte_color_t
{
    GLubyte     r, g, b, a;
} byte_color_t;

typedef struct tag_texcoord_t
{
    GLfloat     u, v;
} texcoord_t;

typedef struct tag_texcoord3_t
{
    GLfloat     u, v, w;
} texcoord3_t;

typedef struct tag_normal_t
{
    GLfloat     x, y, z;
} normal_t;

typedef struct tag_tangent_t
{
    GLfloat     x, y, z;
} tangent_t;

typedef struct tag_idx_t
{
    GLubyte     i[4];
} idx_t;

typedef struct tag_weight_t
{
    GLfloat     w[4];
} weight_t;


//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;
u8 s_clearColor[4];

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;
static nn::gd::Shader* geometryShader = 0;
static nn::gd::ShaderPipeline* shaderPipelineL = 0;
static nn::gd::Shader* vertexShaderL = 0;
static nn::gd::Shader* geometryShaderL = 0;

//Vertex input
static nn::gd::InputLayout* inputLayout = 0;
static nn::gd::VertexBufferResource* bufferCoord = 0;
static nn::gd::VertexBufferResource* bufferIndex = 0;
static nn::gd::InputLayout* inputLayoutL = 0;
static nn::gd::VertexBufferResource* bufferCoordL = 0;
static nn::gd::VertexBufferResource* bufferColorL = 0;
static nn::gd::VertexBufferResource* bufferIndexL = 0;

static nn::gd::VertexBufferResource* buffers[1];
static nn::gd::VertexBufferResource* buffersL[2];
static u32 offsets[] = { 0 };
static u32 offsetsL[] = { 0, 0 };

static nn::gd::UniformLocation uniformLoc_projection;
static nn::gd::UniformLocation uniformLoc_modelview;
static nn::gd::UniformLocation uniformLoc_projectionL;
static nn::gd::UniformLocation uniformLoc_modelviewL;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;


GLuint s_CmdlistID;

//#define ENABLE_LOG
#define PRINT(__msg) NN_LOG __msg

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

static void MakeTorus(GLenum geometry_mode, unsigned use_arrays, array_type_t array_type, int palette_size,
            int horizontal_polygon_count, int virtical_polygon_count, u32* s_IndexCount)
{

    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0 },
        { 0, "aColor", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, sizeof(position_t) },
        { 0, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, sizeof(position_t) + sizeof(color_t) },
    };
    u32 strides[] = { sizeof(position_t) + sizeof(color_t) + sizeof(normal_t) };
    nn::gd::VertexInputStage::CreateInputLayout(descs, 3, strides, vertexShader, &inputLayout);

    int vx, vy;
    int vxnum = horizontal_polygon_count;
    int vynum = virtical_polygon_count;
    
    int vertex_size = 0;
    unsigned char* vertex;
    int pos_offs, col_offs, tex_offs[3], nrm_offs, tang_offs, idx_offs, wgt_offs;

    pos_offs = col_offs = tex_offs[0] = tex_offs[1] = tex_offs[2] = nrm_offs = tang_offs = idx_offs = wgt_offs = 0;

    /* check vertex size and array offset */
    if (use_arrays & USE_ARRAY_POSITION)
    {
        pos_offs = vertex_size;
        vertex_size += sizeof(position_t);
    }
    if (use_arrays & USE_ARRAY_COLOR)
    {
        col_offs = vertex_size;
        vertex_size += sizeof(color_t);
    }
    else if (use_arrays & USE_ARRAY_COLOR_BYTE)
    {
        col_offs = vertex_size;
        vertex_size += sizeof(byte_color_t);
    }
    if (use_arrays & USE_ARRAY_NORMAL)
    {
        nrm_offs = vertex_size;
        vertex_size += sizeof(normal_t);
    }
    if (use_arrays & USE_ARRAY_TANGENT)
    {
        tang_offs = vertex_size;
        vertex_size += sizeof(tangent_t);
    }
    if (use_arrays & USE_ARRAY_PALETTE)
    {
        idx_offs = vertex_size;
        vertex_size += sizeof(idx_t);
        wgt_offs = vertex_size;
        vertex_size += sizeof(weight_t) / 4 * palette_size;
    }
    if (use_arrays & USE_ARRAY_TEXCOORD0)
    {
        tex_offs[0] = vertex_size;
        vertex_size += sizeof(texcoord_t);
    }
    if (use_arrays & USE_ARRAY_TEXCOORD1)
    {
        tex_offs[1] = vertex_size;
        vertex_size += sizeof(texcoord_t);
    }
    if (use_arrays & USE_ARRAY_TEXCOORD2)
    {
        tex_offs[2] = vertex_size;
        vertex_size += sizeof(texcoord_t);
    }
    
    /* make torus arrays */
    vertex = static_cast<unsigned char*>(s_AppHeap.Allocate(vertex_size * (vxnum + 1) * (vynum + 1)));
    memset(vertex, 0, vertex_size * (vxnum + 1) * (vynum + 1));
    {
        const float rad_x = 0.4f;
        const float rad_y = 1.f;
        
        for (vy = 0; vy <= vynum; vy++)
        {
            float cosx, cosy, sinx, siny;
            
            cosy = (float)cos((2 * PI / vynum) * (vy % vynum));
            siny = (float)sin((2 * PI / vynum) * (vy % vynum));

            for (vx = 0; vx <= vxnum; vx++)
            {
                cosx = (float)cos((2 * PI / vxnum) * (vx % vxnum));
                sinx = (float)sin((2 * PI / vxnum) * (vx % vxnum));

                if (array_type == INTERLEAVED_ARRAY)
                {
                    int offset = vertex_size * (vy * (vxnum + 1) + vx);
                    
                    /* position */
                    if (use_arrays & USE_ARRAY_POSITION)
                    {
                        ((position_t*)&(vertex[offset]))->x = cosy * (rad_y + cosx * rad_x);
                        ((position_t*)&(vertex[offset]))->y = siny * (rad_y + cosx * rad_x);
                        ((position_t*)&(vertex[offset]))->z = - sinx * rad_x;
                        offset += sizeof(position_t);
                    }
                    
                    /* color */
                    if (use_arrays & USE_ARRAY_COLOR)
                    {
                        ((color_t*)&(vertex[offset]))->r = 0.5f + cosx * 0.5f;
                        ((color_t*)&(vertex[offset]))->g = 0.5f - cosx * 0.5f;
                        ((color_t*)&(vertex[offset]))->b = 0.5f + cosy * 0.5f;
                        ((color_t*)&(vertex[offset]))->a = 1.f;
                        offset += sizeof(color_t);
                    }
                    else if (use_arrays & USE_ARRAY_COLOR_BYTE)
                    {
                        ((byte_color_t*)&(vertex[offset]))->r = (GLubyte)(255 * (0.5f + cosx * 0.5f));
                        ((byte_color_t*)&(vertex[offset]))->g = (GLubyte)(255 * (0.5f - cosx * 0.5f));
                        ((byte_color_t*)&(vertex[offset]))->b = (GLubyte)(255 * (0.5f + cosy * 0.5f));
                        ((byte_color_t*)&(vertex[offset]))->a = 255;
                        offset += sizeof(byte_color_t);
                    }
                    
                    /* normal */
                    if (use_arrays & USE_ARRAY_NORMAL)
                    {
                        float norm[4];
                        norm[0] = cosy * (rad_y + cosx * rad_x) - rad_y * cosy;
                        norm[1] = siny * (rad_y + cosx * rad_x) - rad_y * siny;
                        norm[2] = - sinx * rad_x;
                        norm[3] = (float)sqrt(norm[0] * norm[0] + norm[1] * norm[1] + norm[2] * norm[2]);
                        ((normal_t*)&(vertex[offset]))->x = norm[0] / norm[3];
                        ((normal_t*)&(vertex[offset]))->y = norm[1] / norm[3];
                        ((normal_t*)&(vertex[offset]))->z = norm[2] / norm[3];
                        offset += sizeof(normal_t);
                    }
                    
                    /* tangent */
                    if (use_arrays & USE_ARRAY_TANGENT)
                    {
                        ((tangent_t*)&(vertex[offset]))->x = -siny;
                        ((tangent_t*)&(vertex[offset]))->y = cosy;
                        ((tangent_t*)&(vertex[offset]))->z = 0;
                        offset += sizeof(tangent_t);
                    }
                    
                    if (use_arrays & USE_ARRAY_PALETTE)
                    {
                        float ratio;
                        
                        /* palette index */
                        ((idx_t*)&(vertex[offset]))->i[0] = 0;
                        ((idx_t*)&(vertex[offset]))->i[1] = 1;
                        ((idx_t*)&(vertex[offset]))->i[2] = 2;
                        ((idx_t*)&(vertex[offset]))->i[3] = 3;
                        offset += sizeof(idx_t);
                        
                        /* palette weight */
                        
                        ratio = (float)vy / (float)vynum;
                        switch (palette_size)
                        {
                            case 4:
                                ((weight_t*)&(vertex[offset]))->w[0] = 0.01f;
                                ((weight_t*)&(vertex[offset]))->w[1] = 0.01f;
                                ((weight_t*)&(vertex[offset]))->w[2] = 0.01f;
                                ((weight_t*)&(vertex[offset]))->w[3] = 0.01f;
                                if (ratio <= 0.25f)
                                {
                                    ((weight_t*)&(vertex[offset]))->w[0] = cosy - 0.01f;
                                    ((weight_t*)&(vertex[offset]))->w[1] = 1 - cosy - 0.01f;
                                }
                                else if (ratio <= 0.5f)
                                {
                                    ((weight_t*)&(vertex[offset]))->w[1] = 1 + cosy - 0.01f;
                                    ((weight_t*)&(vertex[offset]))->w[2] = -cosy - 0.01f;
                                }
                                else if (ratio <= 0.75f)
                                {
                                    ((weight_t*)&(vertex[offset]))->w[2] = -cosy - 0.01f;
                                    ((weight_t*)&(vertex[offset]))->w[3] = 1 + cosy - 0.01f;
                                }
                                else
                                {
                                    ((weight_t*)&(vertex[offset]))->w[3] = 1 - cosy - 0.01f;
                                    ((weight_t*)&(vertex[offset]))->w[0] = cosy - 0.01f;
                                }
                                break;
                            case 3:
                                if (ratio < 0.3333f)
                                {
                                    ((weight_t*)&(vertex[offset]))->w[0] = 1.f - ratio * 3.f - 0.005f;
                                    ((weight_t*)&(vertex[offset]))->w[1] = ratio * 3.f - 0.005f;
                                }
                                else if (ratio < 0.6666f)
                                {
                                    ratio -= 0.3333f;
                                    ((weight_t*)&(vertex[offset]))->w[1] = 1.f - ratio * 3.f - 0.005f;
                                    ((weight_t*)&(vertex[offset]))->w[2] = ratio * 3.f - 0.005f;
                                }
                                else
                                {
                                    ratio -= 0.6666f;
                                    ((weight_t*)&(vertex[offset]))->w[2] = 1.f - ratio * 3.f - 0.005f;
                                    ((weight_t*)&(vertex[offset]))->w[0] = ratio * 3.f - 0.005f;
                                }
                                break;
                            case 2:
                                ((weight_t*)&(vertex[offset]))->w[0] = (cosy + 1.f) / 2.f;
                                ((weight_t*)&(vertex[offset]))->w[1] = 1.f - (cosy + 1.f) / 2.f;
                                break;
                            case 1:
                                ((weight_t*)&(vertex[offset]))->w[0] = 1.f;
                                break;
                        }
                        if (vy == 0 || vy == vynum)
                        {
                            switch (palette_size)
                            {
                                case 4:
                                    ((weight_t*)&(vertex[offset]))->w[3] = 0.f;
                                case 3:
                                    ((weight_t*)&(vertex[offset]))->w[2] = 0.f;
                                case 2:
                                    ((weight_t*)&(vertex[offset]))->w[1] = 0.f;
                                case 1:
                                    ((weight_t*)&(vertex[offset]))->w[0] = 1.f;
                                    break;
                            }
                        }
                        offset += sizeof(weight_t) / 4 * palette_size;
                    }

                    /* texture coord */
                    if (use_arrays & USE_ARRAY_TEXCOORD)
                    {
                        if (use_arrays & USE_ARRAY_TEXCOORD0)
                        {
                            ((texcoord_t*)&(vertex[offset]))->u = (float)vx * (float)(1.f / vxnum);
                            ((texcoord_t*)&(vertex[offset]))->v = (float)vy * (float)(1.f / vynum);
                            offset += sizeof(texcoord_t);
                        }
                        if (use_arrays & USE_ARRAY_TEXCOORD1)
                        {
                            ((texcoord_t*)&(vertex[offset]))->u = (float)vx * (float)(1.f / vxnum);
                            ((texcoord_t*)&(vertex[offset]))->v = (float)vy * (float)(1.f / vynum);
                            offset += sizeof(texcoord_t);
                        }
                        if (use_arrays & USE_ARRAY_TEXCOORD2)
                        {
                            ((texcoord_t*)&(vertex[offset]))->u = (float)vx * (float)(1.f / vxnum);
                            ((texcoord_t*)&(vertex[offset]))->v = (float)vy * (float)(1.f / vynum);
                            offset += sizeof(texcoord_t);
                        }
                    }
                    
                }
                else if (array_type == INDEPENDENT_ARRAY)
                {
                    int offset;
                    
                    /* position */
                    if (use_arrays & USE_ARRAY_POSITION)
                    {
                        offset = (vy * (vxnum + 1) + vx) * sizeof(position_t) + pos_offs * (vxnum + 1) * (vynum + 1);
                        ((position_t*)&(vertex[offset]))->x = cosy * (rad_y + cosx * rad_x);
                        ((position_t*)&(vertex[offset]))->y = siny * (rad_y + cosx * rad_x);
                        ((position_t*)&(vertex[offset]))->z = - sinx * rad_x;
                    }
                    
                    /* color */
                    if (use_arrays & USE_ARRAY_COLOR)
                    {
                        offset = (vy * (vxnum + 1) + vx) * sizeof(color_t) + col_offs * (vxnum + 1) * (vynum + 1);
                        ((color_t*)&(vertex[offset]))->r = 0.5f + cosx * 0.5f;
                        ((color_t*)&(vertex[offset]))->g = 0.5f - cosx * 0.5f;
                        ((color_t*)&(vertex[offset]))->b = 0.5f + cosy * 0.5f;
                        ((color_t*)&(vertex[offset]))->a = 1.f;
                    }
                    else if (use_arrays & USE_ARRAY_COLOR_BYTE)
                    {
                        offset = (vy * (vxnum + 1) + vx) * sizeof(byte_color_t) + col_offs * (vxnum + 1) * (vynum + 1);
                        ((byte_color_t*)&(vertex[offset]))->r = (GLubyte)(255 * (0.5f + cosx * 0.5f));
                        ((byte_color_t*)&(vertex[offset]))->g = (GLubyte)(255 * (0.5f - cosx * 0.5f));
                        ((byte_color_t*)&(vertex[offset]))->b = (GLubyte)(255 * (0.5f + cosy * 0.5f));
                        ((byte_color_t*)&(vertex[offset]))->a = 255;
                    }
                    
                    /* normal */
                    if (use_arrays & USE_ARRAY_NORMAL)
                    {
                        float norm[4];
                        norm[0] = cosy * (rad_y + cosx * rad_x) - rad_y * cosy;
                        norm[1] = siny * (rad_y + cosx * rad_x) - rad_y * siny;
                        norm[2] = - sinx * rad_x;
                        norm[3] = (float)sqrt(norm[0] * norm[0] + norm[1] * norm[1] + norm[2] * norm[2]);
                        offset = (vy * (vxnum + 1) + vx) * sizeof(normal_t) + nrm_offs * (vxnum + 1) * (vynum + 1);
                        ((normal_t*)&(vertex[offset]))->x = norm[0] / norm[3];
                        ((normal_t*)&(vertex[offset]))->y = norm[1] / norm[3];
                        ((normal_t*)&(vertex[offset]))->z = norm[2] / norm[3];
                    }
                    
                    /* tangent */
                    if (use_arrays & USE_ARRAY_TANGENT)
                    {
                        offset = (vy * (vxnum + 1) + vx) * sizeof(tangent_t) + tang_offs * (vxnum + 1) * (vynum + 1);
                        ((tangent_t*)&(vertex[offset]))->x = -siny;
                        ((tangent_t*)&(vertex[offset]))->y = cosy;
                        ((tangent_t*)&(vertex[offset]))->z = 0;
                    }
                    
                    if (use_arrays & USE_ARRAY_PALETTE)
                    {
                        float ratio;
                        
                        /* palette index */
                        offset = (vy * (vxnum + 1) + vx) * (sizeof(idx_t) / 4 * palette_size) + idx_offs * (vxnum + 1) * (vynum + 1);
                        switch (palette_size)
                        {
                            case 4:
                                ((idx_t*)&(vertex[offset]))->i[3] = 3;
                            case 3:
                                ((idx_t*)&(vertex[offset]))->i[2] = 2;
                            case 2:
                                ((idx_t*)&(vertex[offset]))->i[1] = 1;
                            case 1:
                                ((idx_t*)&(vertex[offset]))->i[0] = 0;
                                break;
                        }
                        
                        /* palette weight */
                        offset = (vy * (vxnum + 1) + vx) * (sizeof(weight_t) / 4 * palette_size) + wgt_offs * (vxnum + 1) * (vynum + 1);
                        ratio = (float)vy / (float)vynum;
                        switch (palette_size)
                        {
                            case 4:
                                ((weight_t*)&(vertex[offset]))->w[0] = 0.01f;
                                ((weight_t*)&(vertex[offset]))->w[1] = 0.01f;
                                ((weight_t*)&(vertex[offset]))->w[2] = 0.01f;
                                ((weight_t*)&(vertex[offset]))->w[3] = 0.01f;
                                if (ratio <= 0.25f)
                                {
                                    ((weight_t*)&(vertex[offset]))->w[0] = cosy - 0.01f;
                                    ((weight_t*)&(vertex[offset]))->w[1] = 1 - cosy - 0.01f;
                                }
                                else if (ratio <= 0.5f)
                                {
                                    ((weight_t*)&(vertex[offset]))->w[1] = 1 + cosy - 0.01f;
                                    ((weight_t*)&(vertex[offset]))->w[2] = -cosy - 0.01f;
                                }
                                else if (ratio <= 0.75f)
                                {
                                    ((weight_t*)&(vertex[offset]))->w[2] = -cosy - 0.01f;
                                    ((weight_t*)&(vertex[offset]))->w[3] = 1 + cosy - 0.01f;
                                }
                                else
                                {
                                    ((weight_t*)&(vertex[offset]))->w[3] = 1 - cosy - 0.01f;
                                    ((weight_t*)&(vertex[offset]))->w[0] = cosy - 0.01f;
                                }
                                break;
                            case 3:
                                if (ratio < 0.3333f)
                                {
                                    ((weight_t*)&(vertex[offset]))->w[0] = 1.f - ratio * 3.f - 0.005f;
                                    ((weight_t*)&(vertex[offset]))->w[1] = ratio * 3.f - 0.005f;
                                }
                                else if (ratio < 0.6666f)
                                {
                                    ratio -= 0.3333f;
                                    ((weight_t*)&(vertex[offset]))->w[1] = 1.f - ratio * 3.f - 0.005f;
                                    ((weight_t*)&(vertex[offset]))->w[2] = ratio * 3.f - 0.005f;
                                }
                                else
                                {
                                    ratio -= 0.6666f;
                                    ((weight_t*)&(vertex[offset]))->w[2] = 1.f - ratio * 3.f - 0.005f;
                                    ((weight_t*)&(vertex[offset]))->w[0] = ratio * 3.f - 0.005f;
                                }
                                break;
                            case 2:
                                ((weight_t*)&(vertex[offset]))->w[0] = (cosy + 1.f) / 2.f;
                                ((weight_t*)&(vertex[offset]))->w[1] = 1.f - (cosy + 1.f) / 2.f;
                                break;
                            case 1:
                                ((weight_t*)&(vertex[offset]))->w[0] = 1.f;
                                break;
                        }
                        if (vy == 0 || vy == vynum)
                        {
                            switch (palette_size)
                            {
                                case 4:
                                    ((weight_t*)&(vertex[offset]))->w[3] = 0.f;
                                case 3:
                                    ((weight_t*)&(vertex[offset]))->w[2] = 0.f;
                                case 2:
                                    ((weight_t*)&(vertex[offset]))->w[1] = 0.f;
                                case 1:
                                    ((weight_t*)&(vertex[offset]))->w[0] = 1.f;
                                    break;
                            }
                        }
                    }
                    
                    /* texture coord */
                    if (use_arrays & USE_ARRAY_TEXCOORD)
                    {
                        if (use_arrays & USE_ARRAY_TEXCOORD0)
                        {
                            offset = (vy * (vxnum + 1) + vx) * sizeof(texcoord_t) + tex_offs[0] * (vxnum + 1) * (vynum + 1);
                            ((texcoord_t*)&(vertex[offset]))->u = (float)vx * (float)(1.f / vxnum);
                            ((texcoord_t*)&(vertex[offset]))->v = (float)vy * (float)(1.f / vynum);
                        }
                        if (use_arrays & USE_ARRAY_TEXCOORD1)
                        {
                            offset = (vy * (vxnum + 1) + vx) * sizeof(texcoord_t) + tex_offs[1] * (vxnum + 1) * (vynum + 1);
                            ((texcoord_t*)&(vertex[offset]))->u = (float)vx * (float)(1.f / vxnum);
                            ((texcoord_t*)&(vertex[offset]))->v = (float)vy * (float)(1.f / vynum);
                        }
                        if (use_arrays & USE_ARRAY_TEXCOORD2)
                        {
                            offset = (vy * (vxnum + 1) + vx) * sizeof(texcoord_t) + tex_offs[2] * (vxnum + 1) * (vynum + 1);
                            ((texcoord_t*)&(vertex[offset]))->u = (float)vx * (float)(1.f / vxnum);
                            ((texcoord_t*)&(vertex[offset]))->v = (float)vy * (float)(1.f / vynum);
                        }
                    }
                }
            }
        }
    }
    /************************************************************/
    /* Update vertex data */
    //glBufferData(GL_ARRAY_BUFFER, vertex_size * (vxnum + 1) * (vynum + 1), vertex, GL_STATIC_DRAW);
    nn::gd::VertexBufferResourceDescription desc;
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    desc.m_ByteSize = vertex_size * (vxnum + 1) * (vynum + 1);

    nn::gd::Resource::CreateVertexBufferResource(&desc, vertex, &bufferCoord);
    /************************************************************/

    s_AppHeap.Free(reinterpret_cast<void*>(vertex));
    

    /* Make triangle index */
    {
        int idxnum = 0;
        unsigned short* index_array = 0;
        
        /* Make triangles index */
        if (geometry_mode == TRIANGLES)
        {
            index_array = static_cast<unsigned short*>(s_AppHeap.Allocate(sizeof(unsigned short) * vxnum * vynum * 6));
            memset(index_array, 0, sizeof(unsigned short) * vxnum * vynum * 6);
            for (vy = 0; vy < vynum; vy++)
            {
                for (vx = 0; vx < vxnum; vx++)
                {
                    unsigned short rect[4];
                    rect[0] = vx + vy * (vxnum + 1);
                    rect[1] = (vx + 1) + vy * (vxnum + 1);
                    rect[2] = vx + (vy + 1) * (vxnum + 1);
                    rect[3] = (vx + 1) + (vy + 1) * (vxnum + 1);
                    
                    index_array[idxnum++] = rect[2];
                    index_array[idxnum++] = rect[0];
                    index_array[idxnum++] = rect[3];
                    index_array[idxnum++] = rect[3];
                    index_array[idxnum++] = rect[0];
                    index_array[idxnum++] = rect[1];
                }
            }
        }

        /* Make triangle strip index */
        else if (geometry_mode == TRIANGLE_STRIP || geometry_mode == LINE_STRIP)
        {
            index_array = static_cast<unsigned short*>(s_AppHeap.Allocate(sizeof(unsigned short) * (vxnum + 1) * 2 * vynum));
            memset(index_array, 0, sizeof(unsigned short) * (vxnum + 1) * 2 * vynum);
            for (vy = vynum - 1; vy >= 0; vy--)
            {
                for (vx = 0; vx <= vxnum; vx++)
                {
                    index_array[idxnum++] = vx + (vy + 1) * (vxnum + 1);
                    index_array[idxnum++] = vx + vy * (vxnum + 1);
                }
            }
        }

        /* Make subdivision index */
        else if (geometry_mode == SUBD_PRIM_DMP)
        {
            index_array = static_cast<unsigned short*>(s_AppHeap.Allocate(sizeof(unsigned short) * vxnum * vynum * 17));
            memset(index_array, 0, sizeof(unsigned short) * vxnum * vynum * 17);
            for (vy = 0; vy < vynum; vy++)
            {
                for (vx = 0; vx < vxnum; vx++)
                {
                    index_array[idxnum +  0] = 16;
                    index_array[idxnum +  1] = vy * (vxnum + 1) + vx;
                    index_array[idxnum +  2] = vy * (vxnum + 1) + vx + 1;
                    index_array[idxnum +  3] = (vy + 1) * (vxnum + 1) + vx + 1;
                    index_array[idxnum +  4] = (vy + 1) * (vxnum + 1) + vx;
                    index_array[idxnum +  5] = ((vy + vynum -1) % vynum) * (vxnum + 1) + vx;
                    index_array[idxnum +  6] = ((vy + vynum -1) % vynum) * (vxnum + 1) + vx + 1;
                    index_array[idxnum +  7] = ((vy + vynum -1) % vynum) * (vxnum + 1) + ((vx + 2 + vxnum) % vxnum);
                    index_array[idxnum +  8] = vy * (vxnum + 1) + ((vx + 2 + vxnum) % vxnum);
                    index_array[idxnum +  9] = (vy + 1) * (vxnum + 1) + ((vx + 2 + vxnum) % vxnum);
                    index_array[idxnum + 10] = ((vy + 2 + vynum) % vynum) * (vxnum + 1) + ((vx + 2 + vxnum) % vxnum);
                    index_array[idxnum + 11] = ((vy + 2 + vynum) % vynum) * (vxnum + 1) + vx + 1;
                    index_array[idxnum + 12] = ((vy + 2 + vynum) % vynum) * (vxnum + 1) + vx;
                    index_array[idxnum + 13] = ((vy + 2 + vynum) % vynum) * (vxnum + 1) + ((vx - 1 + vxnum) % vxnum);
                    index_array[idxnum + 14] = (vy + 1) * (vxnum + 1) + ((vx - 1 + vxnum) % vxnum);
                    index_array[idxnum + 15] = vy * (vxnum + 1) + ((vx - 1 + vxnum) % vxnum);
                    index_array[idxnum + 16] = ((vy + vynum -1) % vynum) * (vxnum + 1) + ((vx - 1 + vxnum) % vxnum);
                    idxnum += 17;
                }
            }
        }

        /* Make silhouette triangle index */
        else if (geometry_mode == TRIANGLES_SILHOUETTE_DMP)
        {
            index_array = static_cast<unsigned short*>(s_AppHeap.Allocate(sizeof(unsigned short) * vxnum * vynum * 12));
            memset(index_array, 0, sizeof(unsigned short) * vxnum * vynum * 12);
            for (vy = 0; vy < vynum; vy++)
            {
                for (vx = 0; vx < vxnum; vx++)
                {
                    int rect[8];
                    rect[0] = vy * (vxnum + 1) + vx;
                    rect[1] = vy * (vxnum + 1) + vx + 1;
                    rect[2] = (vy + 1) * (vxnum + 1) + vx;
                    rect[3] = (vy + 1) * (vxnum + 1) + vx + 1;
                    rect[4] = ((vy + vynum -1) % vynum) * (vxnum + 1) + vx;
                    rect[5] = (vy + 1) * (vxnum + 1) + ((vx + 2 + vxnum) % vxnum);
                    rect[6] = ((vy + 2 + vynum) % vynum) * (vxnum + 1) + vx + 1;
                    rect[7] = vy * (vxnum + 1) + ((vx - 1 + vxnum) % vxnum);
                    
                    index_array[idxnum++] = rect[0];
                    index_array[idxnum++] = rect[1];
                    index_array[idxnum++] = rect[4];
                    index_array[idxnum++] = rect[3];
                    index_array[idxnum++] = rect[2];
                    index_array[idxnum++] = rect[5];
                    
                    index_array[idxnum++] = rect[0];
                    index_array[idxnum++] = rect[3];
                    index_array[idxnum++] = rect[1];
                    index_array[idxnum++] = rect[2];
                    index_array[idxnum++] = rect[7];
                    index_array[idxnum++] = rect[6];
                }
            }
        }

        /* Make silhouette strip index */
        else if (geometry_mode == TRIANGLE_STRIP_SILHOUETTE_DMP)
        {
            index_array = static_cast<unsigned short*>(s_AppHeap.Allocate(sizeof(unsigned short) * (vxnum * vynum * 4 + 4)));
            memset(index_array, 0, sizeof(unsigned short) * (vxnum * vynum * 4 + 4));
            for (vy = vynum - 1; vy >= 0; vy--)
            {
                for (vx = 0; vx < vxnum; vx++)
                {
                    if (vx == 0 && (vy == (vynum - 1)))
                    {
                        index_array[idxnum++] = (vy + 1) * (vxnum + 1) + vx;
                        index_array[idxnum++] = vy * (vxnum + 1) + vx;
                        index_array[idxnum++] = vy * (vxnum + 1) + ((vx - 1 + vxnum) % vxnum);
                        index_array[idxnum++] = (vy + 1) * (vxnum + 1) + vx + 1;
                        index_array[idxnum++] = ((vy + 2 + vynum) % vynum) * (vxnum + 1) + vx + 1;
                        index_array[idxnum++] = vy * (vxnum + 1) + vx + 1;
                        index_array[idxnum++] = ((vy + vynum -1) % vynum) * (vxnum + 1) + vx;
                        index_array[idxnum++] = (vy + 1) * (vxnum + 1) + ((vx + 2 + vxnum) % vxnum);
                    }
                    else if (vx == vxnum - 2)
                    {
                        index_array[idxnum++] = ((vy + 2 + vynum) % vynum) * (vxnum + 1) + vx + 1; /* 11 */
                        index_array[idxnum++] = vy * (vxnum + 1) + vx + 1; /* 2 */
                        index_array[idxnum++] = ((vy + vynum -1) % vynum) * (vxnum + 1) + vx; /* 5 */
                        index_array[idxnum++] = vy * (vxnum + 1) + ((vx + 2 + vxnum) % vxnum); /* 8 */
                    }
                    else if (vx == vxnum - 1)
                    {
                        index_array[idxnum++] = (vy + 1) * (vxnum + 1) + vx + 1; /* 3 */
                        index_array[idxnum++] = ((vy + vynum -1) % vynum) * (vxnum + 1) + vx + 1; /* 6 */
                        index_array[idxnum++] = ((vy + vynum -1) % vynum) * (vxnum + 1) + vx; /* 5 */
                        index_array[idxnum++] = vy * (vxnum + 1) + ((vx + 2 + vxnum) % vxnum); /* 8 */
                    }
                    else
                    {
                        index_array[idxnum++] = ((vy + 2 + vynum) % vynum) * (vxnum + 1) + vx + 1; /* 11 */
                        index_array[idxnum++] = vy * (vxnum + 1) + vx + 1; /* 2 */
                        index_array[idxnum++] = ((vy + vynum -1) % vynum) * (vxnum + 1) + vx; /* 5 */
                        index_array[idxnum++] = (vy + 1) * (vxnum + 1) + ((vx + 2 + vxnum) % vxnum); /* 9 */
                    }
                }
            }
        }

        /* Make lines index (wireframe) */
        else if (geometry_mode == LINES)
        {
            index_array = static_cast<unsigned short*>(s_AppHeap.Allocate(sizeof(unsigned short) * vxnum * vynum * 6 * 2));
            memset(index_array, 0, sizeof(unsigned short) * vxnum * vynum * 6 * 2);
            for (vy = 0; vy < vynum; vy++)
            {
                for (vx = 0; vx < vxnum; vx++)
                {
                    unsigned short rect[4];
                    rect[0] = vx + vy * (vxnum + 1);
                    rect[1] = (vx + 1) + vy * (vxnum + 1);
                    rect[2] = vx + (vy + 1) * (vxnum + 1);
                    rect[3] = (vx + 1) + (vy + 1) * (vxnum + 1);
                    
                    index_array[idxnum++] = rect[0];
                    index_array[idxnum++] = rect[1];
                    index_array[idxnum++] = rect[1];
                    index_array[idxnum++] = rect[3];
                    index_array[idxnum++] = rect[3];
                    index_array[idxnum++] = rect[0];
                    index_array[idxnum++] = rect[0];
                    index_array[idxnum++] = rect[3];
                    index_array[idxnum++] = rect[3];
                    index_array[idxnum++] = rect[2];
                    index_array[idxnum++] = rect[2];
                    index_array[idxnum++] = rect[0];
                }
            }
        }

        else if (geometry_mode == POINTS)
        {
            index_array = static_cast<unsigned short*>(s_AppHeap.Allocate(sizeof(unsigned short) * vxnum * vynum));
            memset(index_array, 0, sizeof(unsigned short) * vxnum * vynum);
            for (vy = 0; vy < vynum; vy++)
            {
                for (vx = 0; vx < vxnum; vx++)
                {
                    index_array[idxnum++] = vx + vy * (vxnum + 1);
                }
            }
        }

        /* No other primitive */
        else
            NN_ASSERT(0);

        *s_IndexCount = idxnum;
        //glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned short) * idxnum, index_array, GL_STATIC_DRAW);
        
        desc.m_ByteSize = sizeof(unsigned short) * (idxnum);
        nn::gd::Resource::CreateVertexBufferResource(&desc, index_array, &bufferIndex);


        buffers[0] = bufferCoord;
        
        

        s_AppHeap.Free(reinterpret_cast<void*>(index_array));
    }

    return;
}   /* void MakeTorus(GLenum geometry_mode, unsigned use_arrays, array_type_t array_type,
            int horizontal_polygon_count, int virtical_polygon_count, GLint* s_IndexCount) */

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);                           if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);                           if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 1, &vertexShaderL);                          if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 2, &geometryShader);                         if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 3, &geometryShaderL);                        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, geometryShader, &shaderPipeline);    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShaderL, geometryShaderL, &shaderPipelineL); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    uniformLoc_projection = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection");
    uniformLoc_modelview = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");
    uniformLoc_projectionL = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipelineL, "uProjection");
    uniformLoc_modelviewL = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipelineL, "uModelView");

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void)
{
    //Silhouette
    MakeTorus(TRIANGLE_STRIP_SILHOUETTE_DMP,
    USE_ARRAY_POSITION | USE_ARRAY_COLOR | USE_ARRAY_NORMAL, INTERLEAVED_ARRAY, 0, 40, 40, &s_IndexCount);

    //Line
    f32 coords[] = {
         0.5f, 0.0f, 0.f, 1.f,
        -0.5f, 0.5f, 0.f, 1.f,
        -0.5f,-0.5f, 0.f, 1.f
    };
    f32 color[] = {
        1.f, 0.0f, 0.0f,
        0.f, 1.0f, 0.0f,
        0.f, 0.0f, 1.0f
    };

    u16 idxs[] = {0, 1, 2, 0};

    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0 },
        { 1, "aColor", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0 },
    };
    u32 strides[] = { sizeof(float)*4,  sizeof(float)*3 };
    nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShaderL, &inputLayoutL);

    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize =  sizeof(coords);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    nn::gd::Resource::CreateVertexBufferResource(&desc, coords, &bufferCoordL);

    desc.m_ByteSize =  sizeof(color);
    nn::gd::Resource::CreateVertexBufferResource(&desc, color, &bufferColorL);

    desc.m_ByteSize = sizeof(idxs);
    nn::gd::Resource::CreateVertexBufferResource(&desc, idxs, &bufferIndexL);

    buffersL[0] = bufferCoordL;
    buffersL[1] = bufferColorL;

    //nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffersL, offsetsL);

    //nn::gd::VertexInputStage::SetInputLayout(inputLayout);
        
    //nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
    //nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);

    //NN_LOG("INIT done! ");
}

void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

    //Clear color
    s_clearColor[0] = s_clearColor[1] = 25; 
    s_clearColor[2] = 122; 
    s_clearColor[3] = 255;

    //Texture combiner
    nn::gd::CombinerDescription combiner;
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);
    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

    // Some uniforms for GS
    GLfloat silhwidth[2];
    silhwidth[0] = 0.15f;
    silhwidth[1] = silhwidth[0] * (float)nn::gx::DISPLAY0_WIDTH / (float)nn::gx::DISPLAY0_HEIGHT;
    GLfloat silhcolor[4] = {0.f, 0.f, 1.f, 1.f};

    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_Silhouette.width"), static_cast<f32*>(silhwidth));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_Silhouette.color"), static_cast<f32*>(silhcolor));
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_Silhouette.frontFaceCCW"), 1, 1);
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_Silhouette.acceptEmptyTriangles"), 0, 1);
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_Silhouette.scaleByW"), 1, 1);

    GLfloat v[4];
    GLfloat lwidth = 1.f;
    v[0] = (GLfloat)nn::gx::DISPLAY0_WIDTH / lwidth;
    v[1] = (GLfloat)nn::gx::DISPLAY0_HEIGHT / lwidth;
    v[2] = (GLfloat)nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT;
    v[3] = 2.f / lwidth;

    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipelineL, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipelineL, "dmp_Line.width"), static_cast<f32*>(v));
}

#ifdef ENABLE_LOG
//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;
#endif

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);
#ifdef ENABLE_LOG
    s64 startTime = GetTimeMicroSeconds();
#endif
    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Set the viewport
        nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
        nn::gd::RasterizerStage::SetViewport(viewPort);

        //Projection matrix
        nn::math::Matrix44 proj;
        nn::math::MTX44Frustum(&proj, -0.02f, 0.02f, -0.02f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.02f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.1f, 100.f);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_projection, static_cast<f32*>(proj));
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipelineL, uniformLoc_projectionL, static_cast<f32*>(proj));
    
    }

    //static int firstTimeModelView = true;
    //if (firstTimeModelView)
    {
        //Modelview
        nn::math::Matrix34 mv, rot;
        nn::math::Vector3 camPos(0.f, 0.4f, 7.5f);
        nn::math::Vector3 camUp(0.f, 1.f, 0.f);
        nn::math::Vector3 target(0.f, 0.f, 0.f);
        nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
        nn::math::MTX34RotXYZDeg(&rot, 0.f, -6.f*f, 0.f);
        nn::math::MTX34Mult(&mv, &mv, &rot);
        nn::math::Matrix44 m(mv);

        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_modelview, static_cast<f32*>(m));
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipelineL, uniformLoc_modelviewL, static_cast<f32*>(m));
        //firstTimeModelView = false;
    }

    /********************************************************/
    /********************** Display 0 ***********************/
    {
        //Clear the render buffers
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, s_clearColor, 1.f, 0);

        // ************ SILHOUETTE *****************
        nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);
        nn::gd::VertexInputStage::SetVertexBuffers(0, 1, buffers, offsets);
        nn::gd::VertexInputStage::SetInputLayout(inputLayout);
        nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
        nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);
        //Draw
        nn::gd::System::DrawIndexed(s_IndexCount, 0);
        //nngxDumpCmdBufRegistries(0x000, 0xfff, false);
        //nngxDumpCmdBufHistory(0x000, 0xfff, true);


        // ************ STRIP LINE *****************
        nn::gd::ShaderStage::SetShaderPipeline(shaderPipelineL);
        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffersL, offsetsL);
        nn::gd::VertexInputStage::SetInputLayout(inputLayoutL);
        nn::gd::VertexInputStage::SetIndexBuffer(bufferIndexL, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
        nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);
        //Draw
        nn::gd::System::DrawIndexed(4, 0);


        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();


    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

#ifdef ENABLE_LOG
    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }
#endif
    return 0;
}

static void Finalize()
{
    //Release GD objects
    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (bufferCoord) nn::gd::Resource::ReleaseVertexBufferResource(bufferCoord);
    if (bufferIndex) nn::gd::Resource::ReleaseVertexBufferResource(bufferIndex);
    if (bufferCoordL) nn::gd::Resource::ReleaseVertexBufferResource(bufferCoordL);
    if (bufferColorL) nn::gd::Resource::ReleaseVertexBufferResource(bufferColorL);
    if (bufferIndexL) nn::gd::Resource::ReleaseVertexBufferResource(bufferIndexL);
    if (inputLayout) nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout);

    if (shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader) nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (geometryShader) nn::gd::ShaderStage::ReleaseShader(geometryShader);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (shaderPipelineL) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipelineL);
    if (vertexShaderL) nn::gd::ShaderStage::ReleaseShader(vertexShaderL);
    if (geometryShaderL) nn::gd::ShaderStage::ReleaseShader(geometryShaderL);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Init command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Init GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Init main frame buffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    InitShader();
    InitObjects();
    Initialize_GD();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}
