﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ProjectionTexture.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <math.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"
#include "demo.h"

#include "Util.h"
#include "Loader.h"

#include <nn/gd.h>

#define WAIT_VSYNC

#define DMP_PI  (3.1415926f)
#define REV_PI  (1.0f/DMP_PI)

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;

//Vertex input
static nn::gd::InputLayout* s_inputLayout = 0;
static nn::gd::VertexBufferResource* s_cubeBufferCoord = 0;
static nn::gd::VertexBufferResource* s_cubeBufferNorm = 0;
static nn::gd::VertexBufferResource* s_cubeIndex = 0;

static nn::gd::UniformLocation s_shaderVariable_uModelView;
static nn::gd::UniformLocation s_shaderVariable_uTexModelViewProjection;

//textures
static nn::gd::Texture2DResource* s_texture2DResource;
static nn::gd::Texture2D* s_texture2D = 0;

//Sampler states
static nn::gd::SamplerState* s_samplerDefault = 0;

/* obj loader class object */
typedef struct tag_datGD_t
{
    nn::gd::VertexBufferResource*          posVB;
    nn::gd::VertexBufferResource*          normVB;
    nn::gd::VertexBufferResource*          texVB;
    nn::gd::VertexBufferResource*          tangVB;
    nn::gd::VertexBufferResource*          wghtVB;
    nn::gd::VertexBufferResource*          midxVB;
    nn::gd::VertexBufferResource*          idxVB;
    dat_obj_t*      obj;
    dat_patch_t*    patch;
    int             obj_num;
} datGD_t;
static datGD_t robot;

void loadDAT_GD(const char* _name, datGD_t* _dat);
void unloadDAT_GD(datGD_t* dat);

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

GLuint s_CmdlistID;

#define PRINT(__msg) NN_LOG __msg


//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, NULL, &shaderPipeline); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    s_shaderVariable_uModelView = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");
    s_shaderVariable_uTexModelViewProjection = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uTexModelViewProjection");

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void)
{
    //Load robot data
    loadDAT_GD( "rom:/resources/robot1.dat", &robot);

    //cube data
    /* load cube model */
    GLfloat pos[] = {
        -1.f,-1.f,1.f,  1.f,1.f,1.f, -1.f,1.f,1.f, 1.f,-1.f,1.f,        // Front
        -1.f,-1.f,-1.f,  1.f,1.f,-1.f, -1.f,1.f,-1.f, 1.f,-1.f,-1.f,    // Back
        1.f,-1.f,-1.f,  1.f,1.f,1.f, 1.f,-1.f,1.f, 1.f,1.f,-1.f,        // Right
        -1.f,-1.f,-1.f,  -1.f,1.f,1.f, -1.f,-1.f,1.f, -1.f,1.f,-1.f,    // Left
        -1.f,-1.f,-1.f, 1.f,-1.f,1.f, -1.f,-1.f,1.f, 1.f,-1.f,-1.f,     // Bottom
        -1.f,1.f,-1.f, 1.f,1.f,1.f, -1.f,1.f,1.f, 1.f,1.f,-1.f          // Top
    };
    GLfloat nrm[] = {
        0.f,0.f,1.f, 0.f,0.f,1.f, 0.f,0.f,1.f, 0.f,0.f,1.f,
        0.f,0.f,-1.f, 0.f,0.f,-1.f, 0.f,0.f,-1.f, 0.f,0.f,-1.f,
        1.f,0.f,0.f, 1.f,0.f,0.f, 1.f,0.f,0.f, 1.f,0.f,0.f, 
        -1.f,0.f,0.f, -1.f,0.f,0.f, -1.f,0.f,0.f, -1.f,0.f,0.f, 
        0.f,-1.f,0.f, 0.f,-1.f,0.f, 0.f,-1.f,0.f, 0.f,-1.f,0.f, 
        0.f,1.f,0.f, 0.f,1.f,0.f, 0.f,1.f,0.f, 0.f,1.f,0.f
    };
    GLushort idx[] = {
        0,1,2, 0,3,1,
        4,6,5, 4,5,7,
        8,9,10, 8,11,9,
        12,14,13, 12,13,15,
        16,17,18, 16,19,17,
        20,22,21, 20,21,23
    };

    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize = sizeof(pos);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    nnResult res = nn::gd::Resource::CreateVertexBufferResource(&desc, pos, &s_cubeBufferCoord);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(nrm);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, nrm, &s_cubeBufferNorm);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(idx);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, idx, &s_cubeIndex);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}


    //Create vertex input layout
    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
    };

    u32 strides[] = { sizeof(float) * 3, sizeof(float) * 3};
    res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShader, &s_inputLayout);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Load texture
    u32 width, height, textureWidth, textureHeight;
    u8* pixel = GetTextureData_BMP(L"rom:/resources/knight.bmp", width, height, textureWidth, textureHeight, false, &s_AppHeap);
    if ( !pixel ) {NN_PANIC("\n");}

    nn::gd::Texture2DResourceDescription text2DResDesc = {textureWidth, textureHeight, 1, nn::gd::Resource::NATIVE_FORMAT_RGB_888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
    res = nn::gd::Resource::CreateTexture2DResource(&text2DResDesc, pixel, GD_FALSE, &s_texture2DResource);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::Texture2DDescription tex2ddesc = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(s_texture2DResource, &tex2ddesc, &s_texture2D);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    s_AppHeap.Free(pixel);

    //Sampler states
    nn::gd::SamplerStateDescription descDefault;
    descDefault.ToDefault();
    descDefault.SetMinFilter(nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR);
    descDefault.SetMagFilter(nn::gd::TextureStage::SAMPLER_MAG_FILTER_LINEAR);
    descDefault.SetWrapS(nn::gd::TextureStage::SAMPLER_WRAP_CLAMP_TO_BORDER);
    descDefault.SetWrapT(nn::gd::TextureStage::SAMPLER_WRAP_CLAMP_TO_BORDER);
    nn::gd::TextureStage::CreateSamplerState(&descDefault, &s_samplerDefault);

    //Texture unit 0
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, s_samplerDefault);
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D);
    nn::gd::TextureStage::SetTexture2DProjectionForUnit0(GD_TRUE);
}

void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

    //Projection matrix
    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.06f, 0.06f, -0.06f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.06f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);
    //proj.Transpose();
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection"), static_cast<f32*>(proj));
    //proj.Transpose();

    //Texture combiner
    nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0,nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0,nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0,nn::gd::CombinerStage::COMBINE_RGB_MULT_ADD);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0,nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);

    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

    //Initialize lighting
    nn::gd::LightingStage::light[0].EnableLight(GD_TRUE);

    nn::gd::LightingStage::SetGlobalColorAmbient(26, 26, 26);
        
    nn::gd::LightingStage::light[0].SetColorDiffuse(102, 102, 102);
    nn::gd::LightingStage::light[0].SetLightType(nn::gd::Light::SOURCE_TYPE_DIRECTIONAL);
}

void DrawScene()
{
    static int f = 0;
    nn::math::Matrix44 proj, tmp44;
    nn::math::Matrix34 mv, view, texview, model, tmp34;
    nn::math::Vector3 vec3;

    // setup projection matrix for projection texture
    nn::math::MTX44Frustum(&proj, -0.04f, 0.04f, -0.04f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.04f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);
    vec3.Set(0.5f, 0.5f, 1.f);
    nn::math::MTX34Scale(&tmp34, &vec3);
    tmp44 = nn::math::Matrix44(tmp34);
    nn::math::MTX44Mult(&proj, &tmp44, &proj);
    vec3.Set(0.5f, 0.5f, 0.f);
    nn::math::MTX34Translate(&tmp34, &vec3);
    tmp44 = nn::math::Matrix44(tmp34);
    nn::math::MTX44Mult(&proj, &tmp44, &proj);

    nn::math::Vector3 camPos(2.5f, 0.5f, 6.f);
    nn::math::Vector3 camUp(-1.f, 0.f, 0.f);
    nn::math::Vector3 target(-0.2f, 0.f, 0.f);
    nn::math::MTX34LookAt(&view, &camPos, &camUp, &target);

    camPos.Set(0.f, 0.f, 4.f);
    camUp.Set(0.f, 1.f, 0.f);
    target.Set(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&texview, &camPos, &camUp, &target);

    // setup light direction
    nn::math::Vector4 lpos0(0.f, 0.f, 1.f, 0.f);
    nn::math::Vector4 view0(view.m[0][0], view.m[0][1], view.m[0][2], view.m[0][3]);
    nn::math::Vector4 view1(view.m[1][0], view.m[1][1], view.m[1][2], view.m[1][3]);
    nn::math::Vector4 view2(view.m[2][0], view.m[2][1], view.m[2][2], view.m[2][3]);
    nn::math::Vector4 lpos(nn::math::VEC4Dot(&view0, &lpos0), nn::math::VEC4Dot(&view1, &lpos0), nn::math::VEC4Dot(&view2, &lpos0), lpos0.w);
    nn::gd::LightingStage::light[0].SetPosition(lpos.x, lpos.y, lpos.z);

    /* draw robot */
    {
        /* setup modelview matrix */
        int aniframe = f % 360;
        int anitype = (f / 360) % 3;

        vec3.Set(0.f, 0.f, 2.f);
        nn::math::MTX34Translate(&model, &vec3);
        switch (anitype)
        {
        case 0: /* rotation */
            nn::math::MTX34RotXYZDeg(&tmp34, 0.f, static_cast<f32>(-aniframe), 0.f);
            nn::math::MTX34Mult(&model, &model, &tmp34);
            break;
        case 1: /* translate to front and back */
            vec3.Set(0.f, 0.f, 0.6f * sin(DMP_PI/180.f*static_cast<f32>(aniframe)));
            nn::math::MTX34Translate(&tmp34, &vec3);
            nn::math::MTX34Mult(&model, &model, &tmp34);
            break;
        case 2: /* translate to right and left */
            vec3.Set(0.6f * sin(DMP_PI/180.f*static_cast<f32>(aniframe)), 0.f, 0.f);
            nn::math::MTX34Translate(&tmp34, &vec3);
            nn::math::MTX34Mult(&model, &model, &tmp34);
            break;
        }
        vec3.Set(0.65f, 0.65f, 0.65f);
        nn::math::MTX34Scale(&tmp34, &vec3);
        nn::math::MTX34Mult(&model, &model, &tmp34);

        nn::math::MTX34Mult(&mv, &view, &model);
        nn::math::Matrix44 m(mv);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_uModelView, static_cast<f32*>(m));

        /* setup modelview projection matrix for projection texture */
        nn::math::MTX34Mult(&mv, &texview, &model);
        nn::math::Matrix44 mvp(mv);
        nn::math::MTX44Mult(&mvp, &proj, &mvp);

        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_uTexModelViewProjection, static_cast<f32*>(mvp));

        for (int i = 0; i < robot.obj_num; i++)
        {
            nn::gd::VertexInputStage::SetInputLayout(s_inputLayout);

            nn::gd::VertexBufferResource* buffers[] = { robot.posVB, robot.normVB };
            u32 offsets[] = { 0, 0 };
            nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);
            for (unsigned j = robot.obj[i].patch_offset; j < robot.obj[i].patch_size + robot.obj[i].patch_offset; j++)
            {
                nn::gd::VertexInputStage::SetIndexBuffer(robot.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

                nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
                nn::gd::System::DrawIndexed(robot.patch[j].elm_size, 0);
            }
        }


    }

    //draw the cube
    {
        // setup modelview matrix
        vec3.Set(0.f, 0.f, -3.5f);
        nn::math::MTX34Translate(&model, &vec3);
        vec3.Set(1.f, 1.f, 0.f);
        nn::math::MTX34RotAxisFIdx(&tmp34, &vec3, static_cast<f32>(-f)/2.f);
        nn::math::MTX34Mult(&model, &model, &tmp34);
        vec3.Set(1.75f, 1.75f, 1.75f);
        nn::math::MTX34Scale(&tmp34, &vec3);
        nn::math::MTX34Mult(&model, &model, &tmp34);

        nn::math::MTX34Mult(&mv, &view, &model);
        nn::math::Matrix44 m(mv);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_uModelView, static_cast<f32*>(m));

        // setup modelview projection matrix for projection texture
        nn::math::MTX34Mult(&mv, &texview, &model);
        nn::math::Matrix44 mvp(mv);
        nn::math::MTX44Mult(&mvp, &proj, &mvp);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_uTexModelViewProjection, static_cast<f32*>(mvp));

        nn::gd::VertexInputStage::SetInputLayout(s_inputLayout);

        nn::gd::VertexBufferResource* buffers[] = { s_cubeBufferCoord, s_cubeBufferNorm };
        u32 offsets[] = { 0, 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);

        nn::gd::VertexInputStage::SetIndexBuffer(s_cubeIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

        nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
        nn::gd::System::DrawIndexed(36, 0);
    }

    f++;
}

//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);
    s64 startTime = GetTimeMicroSeconds();

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Set the viewport
        nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
        nn::gd::RasterizerStage::SetViewport(viewPort);
    }

    /********************************************************/
    /********************** Display 0 ***********************/
    {
        //Clear the render buffers
        u8 clearColor[] = {80, 100, 125, 255};
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, clearColor, 1.f, 0);

        //Draw
        DrawScene();

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    return 0;
}

static void Finalize()
{
    //Release GD objects
    if (s_samplerDefault) nn::gd::TextureStage::ReleaseSamplerState(s_samplerDefault);

    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (s_inputLayout) nn::gd::VertexInputStage::ReleaseInputLayout(s_inputLayout);
    if (s_cubeBufferCoord) nn::gd::Resource::ReleaseVertexBufferResource(s_cubeBufferCoord);
    if (s_cubeBufferNorm) nn::gd::Resource::ReleaseVertexBufferResource(s_cubeBufferNorm);
    if (s_cubeIndex) nn::gd::Resource::ReleaseVertexBufferResource(s_cubeIndex);

    if (shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader) nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    if (s_texture2DResource) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource);
    if (s_texture2D) nn::gd::TextureStage::ReleaseTexture2D(s_texture2D);


    //release data
    unloadDAT_GD(&robot);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    InitShader();
    InitObjects();
    Initialize_GD();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

void loadDAT_GD(const char* _name, datGD_t* _dat)
{
    if (!_name || !_dat)
        return;

    char* fname = 0;
    wchar_t* wfname = 0;
    unsigned char* dat = 0;
    long long int filesize;
    nn::fs::FileReader file;

    do
    {
        fname = (char*)s_AppHeap.Allocate(strlen(_name) + 1);
        if (!fname)
            break;

        int dotpos = 0;
        strcpy(fname, _name);
        for (int i = (int)strlen(fname) - 1; i >= 0; i--)
        {
            if (fname[i] == '.')
            {
                dotpos = i;
                break;
            }
        }
        if (strncmp(&fname[dotpos + 1], "dat", 3) != 0)
            break;

        wfname = static_cast<wchar_t*>(s_AppHeap.Allocate( (strlen(fname)+1)*2 ));
        if (!wfname)
            break;
        (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
        file.Initialize(wfname);
        filesize = file.GetSize();
        dat = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
        if (!dat)
            break;
        filesize = file.Read(dat, filesize);

        dat_header_t* header = (dat_header_t*)dat;
        dat_obj_t* obj = (dat_obj_t*)&dat[sizeof(dat_header_t)];
        dat_patch_t* patch = (dat_patch_t*)&dat[sizeof(dat_header_t) + sizeof(dat_obj_t) * header->obj_num];

        nn::fs::FileReader attribfile;
        unsigned char* attribdata = 0;
        nnResult res;
        nn::gd::VertexBufferResourceDescription desc;
        memset(_dat, 0, sizeof(dat_t));

        if (header->vtx_size)
        {
            memcpy(&fname[dotpos + 1], "vtx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->vtx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->posVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->nrm_size)
        {
            memcpy(&fname[dotpos + 1], "nrm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->nrm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->normVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tex_size)
        {
            memcpy(&fname[dotpos + 1], "tex", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tex_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->texVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tgt_size)
        {
            memcpy(&fname[dotpos + 1], "tgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->tangVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->wgt_size)
        {
            memcpy(&fname[dotpos + 1], "wgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->wgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->wghtVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->idx_size)
        {
            memcpy(&fname[dotpos + 1], "idx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->idx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->midxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->elm_size)
        {
            memcpy(&fname[dotpos + 1], "elm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->elm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->idxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }

        _dat->obj = (dat_obj_t*)s_AppHeap.Allocate(sizeof(dat_obj_t) * header->obj_num);
        _dat->patch = (dat_patch_t*)s_AppHeap.Allocate(sizeof(dat_patch_t) * header->patch_num);
        if (!_dat->obj || !_dat->patch)
        {
            if (_dat->obj)
                s_AppHeap.Free(_dat->obj);
            if (_dat->patch)
                s_AppHeap.Free(_dat->patch);
            break;
        }
        memcpy(_dat->obj, obj, sizeof(dat_obj_t) * header->obj_num);
        memcpy(_dat->patch, patch, sizeof(dat_patch_t) * header->patch_num);
        _dat->obj_num = header->obj_num;
    } while (0);

    if (fname)
        s_AppHeap.Free(fname);
    if (wfname)
        s_AppHeap.Free(wfname);
    if (dat)
        s_AppHeap.Free(dat);
    file.Finalize();

    return;
}

void unloadDAT_GD(datGD_t* dat)
{
    if (dat)
    {
        if (dat->posVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->posVB);
        if (dat->normVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->normVB);
        if (dat->texVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->texVB);
        if (dat->tangVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->tangVB);
        if (dat->wghtVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->wghtVB);
        if (dat->midxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->midxVB);
        if (dat->idxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->idxVB);
        if (dat->obj)
            s_AppHeap.Free(dat->obj);
        if (dat->patch)
            s_AppHeap.Free(dat->patch);
    }

    return;
}
