﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     RenderToTarget.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"
#include "demo.h"

#include <nn/gd.h>

#define WAIT_VSYNC

//Choose one of the following in order to test some GD memory transfer features
#define TEST_NONE                                       0
#define TEST_COPYING_SUBRESOURCE                        1
#define TEST_COPYING_SUBRESOURCE_BLOCK_TO_LINEAR        2
#define DISPLAY_BOTTOM_MODE TEST_NONE

//This field lets us test the creation and use of an incomplete InputLayout: the inputlayout's slots don't match all the vertex shader's input attributes.
//In such a case an undefined value will be used for the vertex shader's input attribute which is not defined by the input layout
#define TEST_USING_INCOMPLETE_INPUT_LAYOUT 0

#define TEST_DRAW_USING_IMMEDIAT_MODE 0

//#define TEST_COPY_VERTEXBUFFER

//Test casting a texture2DResource to another format in order to display the resource from the DepthStencilTarget as a texture
//#define TEST_CASTING_TEXTURE2DRESOURCE

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;

//Depth/stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//2D textures
static nn::gd::Texture2D* s_texture2D_BMP = 0;
static nn::gd::Texture2D* s_texture2D_BMP2 = 0;
static nn::gd::Texture2D* s_texture2D_FromCastedResource = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

static int s_RenderTextureWidth = 256;
static int s_RenderTextureHeight = 256;
static nn::gd::Texture2DResource* s_texture2DResource_ColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_DepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_renderTarget = 0;
static nn::gd::DepthStencilTarget* s_depthStencilTarget = 0;
static nn::gd::Texture2D* s_texture2D_colorBuffer = 0;

//Sampler states
static nn::gd::SamplerState* s_samplerDefault = 0;
static nn::gd::SamplerState* s_samplerLinear = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;

//Vertex input
static nn::gd::InputLayout* inputLayout = 0;
static nn::gd::VertexBufferResource* bufferCoord = 0;
static nn::gd::VertexBufferResource* bufferCoord_Bis = 0;
static nn::gd::VertexBufferResource* bufferTexCoord = 0;
static nn::gd::VertexBufferResource* bufferIndex = 0;

static nn::gd::UniformLocation s_shaderVariable_proj;
static nn::gd::UniformLocation s_shaderVariable_view;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

GLuint s_CmdlistID;

//#define ENABLE_LOG
#define PRINT(__msg) NN_LOG __msg

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);

    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, NULL, &shaderPipeline);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    s_shaderVariable_proj = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection");
    s_shaderVariable_view = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");
    NN_ASSERT(s_shaderVariable_proj.IsValid());
    NN_ASSERT(s_shaderVariable_view.IsValid());

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    file.Finalize();
    s_AppHeap.Free(buf);
}

float coords[] = {
        -1.0f,-1.0f, 1.0f,1.0f,
        1.0f,-1.0f, 1.0f,1.0f,
        -1.0f, 1.0f, 1.0f,1.0f,
        1.0f, 1.0f, 1.0f,1.0f,
        -1.0f,-1.0f, -1.0f,1.0f,
        1.0f,-1.0f, -1.0f,1.0f,
        -1.0f, 1.0f, -1.0f,1.0f,
        1.0f, 1.0f, -1.0f,1.0f,
        -1.0f, 1.0f,-1.0f,1.0f,
        1.0f, 1.0f,-1.0f,1.0f,
        -1.0f, 1.0f, 1.0f,1.0f,
        1.0f, 1.0f, 1.0f,1.0f,
        -1.0f, -1.0f,-1.0f,1.0f,
        1.0f, -1.0f,-1.0f,1.0f,
        -1.0f, -1.0f, 1.0f,1.0f,
        1.0f, -1.0f, 1.0f,1.0f,
        1.0f,-1.0f,-1.0f,1.0f,
        1.0f,-1.0f, 1.0f,1.0f,
        1.0f, 1.0f,-1.0f,1.0f,
        1.0f, 1.0f, 1.0f,1.0f,
        -1.0f,-1.0f,-1.0f,1.0f,
        -1.0f,-1.0f, 1.0f,1.0f,
        -1.0f, 1.0f,-1.0f,1.0f,
        -1.0f, 1.0f, 1.0f,1.0f
};

float coords_bis[] = {
    -1.0f,-0.5f, 1.0f,1.f,
    1.0f,-0.5f, 1.0f,1.f,
    -1.0f, 0.5f, 1.0f,1.f,
    1.0f, 0.5f, 1.0f,1.f,
    -1.0f,-0.5f, -1.0f,1.f,
    1.0f,-0.5f, -1.0f,1.f,
    -1.0f, 0.5f, -1.0f,1.f,
    1.0f, 0.5f, -1.0f,1.f,
    -1.0f, 0.5f,-1.0f,1.f,
    1.0f, 0.5f,-1.0f,1.f,
    -1.0f, 0.5f, 1.0f,1.f,
    1.0f, 0.5f, 1.0f,1.f,
    -1.0f, -0.5f,-1.0f,1.f,
    1.0f, -0.5f,-1.0f,1.f,
    -1.0f, -0.5f, 1.0f,1.f,
    1.0f, -0.5f, 1.0f,1.f,
    1.0f,-0.5f,-1.0f,1.f,
    1.0f,-0.5f, 1.0f,1.f,
    1.0f, 0.5f,-1.0f,1.f,
    1.0f, 0.5f, 1.0f,1.f,
    -1.0f,-0.5f,-1.0f,1.f,
    -1.0f,-0.5f, 1.0f,1.f,
    -1.0f, 0.5f,-1.0f,1.f,
    -1.0f, 0.5f, 1.0f,1.f
};

float texcoords[] = {
    1.f, 0.f, 0.f,
    1.f, 1.f, 0.f,
    0.f, 0.f, 0.f,
    0.f, 1.f, 0.f,
    0.f, 1.f, 0.f,
    0.f, 0.f, 0.f,
    1.f, 1.f, 0.f,
    1.f, 0.f, 0.f,
    1.f, 0.f, 0.f,
    1.f, 1.f, 0.f,
    0.f, 0.f, 0.f,
    0.f, 1.f, 0.f,
    0.f, 1.f, 0.f,
    0.f, 0.f, 0.f,
    1.f, 1.f, 0.f,
    1.f, 0.f, 0.f,
    1.f, 0.f, 0.f,
    1.f, 1.f, 0.f,
    0.f, 0.f, 0.f,
    0.f, 1.f, 0.f,
    0.f, 1.f, 0.f,
    0.f, 0.f, 0.f,
    1.f, 1.f, 0.f,
    1.f, 0.f, 0.f
};
u16 idxs[] = {0, 1, 2, 2, 1, 3, 4, 5, 6, 6, 5, 7, 8, 9, 10, 10, 9, 11, 12, 13, 14 ,14, 13, 15, 16, 17, 18, 18, 17, 19, 20, 21, 22, 22, 21, 23};

static void InitObjects(void)
{
    nnResult res;

    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0},
        { 1, "aTexCoord", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
    };

    u32 strides[] = { sizeof(float) * 4, sizeof(float) * 3};

#if TEST_USING_INCOMPLETE_INPUT_LAYOUT == 1
    // Test the creation and use of an incomplete InputLayout: the inputlayout's slots don't match all the vertex shader's input attributes
    // In such a case an undefined value will be used for the vertex shader's input attribute which is not defined by the input layout
    res = nn::gd::VertexInputStage::CreateInputLayout(descs, 1, strides, vertexShader, &inputLayout);
#else
    res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShader, &inputLayout);
#endif
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize = sizeof(coords);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, coords, &bufferCoord);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, coords_bis, &bufferCoord_Bis);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(texcoords);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, texcoords, &bufferTexCoord);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(idxs);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, idxs, &bufferIndex);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::VertexBufferResource* buffers[] = { bufferCoord, bufferTexCoord };
    u32 offsets[] = { 0, 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);

    nn::gd::VertexInputStage::SetInputLayout(inputLayout);
        
    nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
}

void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

    nnResult res;
    //2D textures
    {
        wchar_t* bmpRomFilename = L"rom:/resources/rainbow.bmp";

        u32 width = 0; u32 height = 0; u32 textureWidth = 0; u32 textureHeight = 0;
        //Get the texture data in native format
        u8* textureDataPtr = GetTextureData_BMP(bmpRomFilename, width, height, textureWidth, textureHeight, false, &s_AppHeap);

        nn::gd::Texture2DResource* texture2DResource;
        nn::gd::Texture2DResourceDescription Text2DResDesc = {textureWidth, textureHeight, 1, nn::gd::Resource::NATIVE_FORMAT_RGB_888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
        res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, textureDataPtr, GD_FALSE, &texture2DResource);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
        s_AppHeap.Free(textureDataPtr);

        nn::gd::Texture2DDescription tex2DDesc = {0, 0};
        res = nn::gd::TextureStage::CreateTexture2D(texture2DResource, &tex2DDesc, &s_texture2D_BMP);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
        nn::gd::Resource::ReleaseTexture2DResource(texture2DResource);
    }

    {
        wchar_t* bmpRomFilename = L"rom:/resources/knight.bmp";

        u32 width = 0; u32 height = 0; u32 textureWidth = 0; u32 textureHeight = 0;
        //Get the texture data in native format
        u8* textureDataPtr = GetTextureData_BMP(bmpRomFilename, width, height, textureWidth, textureHeight, false, &s_AppHeap);

        nn::gd::Texture2DResource* texture2DResource;
        nn::gd::Texture2DResourceDescription Text2DResDesc = {textureWidth, textureHeight, 1, nn::gd::Resource::NATIVE_FORMAT_RGB_888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
        res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, textureDataPtr, GD_FALSE, &texture2DResource);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
        s_AppHeap.Free(textureDataPtr);

        nn::gd::Texture2DDescription tex2DDesc = {0, 0};
        res = nn::gd::TextureStage::CreateTexture2D(texture2DResource, &tex2DDesc, &s_texture2D_BMP2);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
        nn::gd::Resource::ReleaseTexture2DResource(texture2DResource);
    }

    //Texture combiner
    nn::gd::CombinerDescription combiner;
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
    combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_REPLACE);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);
    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

    //Sampler states
    nn::gd::SamplerStateDescription descDefault;
    descDefault.ToDefault();
    nn::gd::TextureStage::CreateSamplerState(&descDefault, &s_samplerDefault);

    nn::gd::SamplerStateDescription descSamplerState;
    descSamplerState.SetMinFilter(nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR_MIPMAP_LINEAR);
    descSamplerState.SetMagFilter(nn::gd::TextureStage::SAMPLER_MAG_FILTER_LINEAR);
    descSamplerState.SetWrapS(nn::gd::TextureStage::SAMPLER_WRAP_REPEAT);
    descSamplerState.SetWrapT(nn::gd::TextureStage::SAMPLER_WRAP_REPEAT);
    descSamplerState.SetBorderColor(0.0f, 1.0f, 0.0f, 1.0f);
    nn::gd::TextureStage::CreateSamplerState(&descSamplerState, &s_samplerLinear);

    //Additional Framebuffer
    //Color buffer (with 4 mipmaps)
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {s_RenderTextureWidth, s_RenderTextureHeight, 4, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_ColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_ColorBuffer, &descRenderTarget, &s_renderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {s_RenderTextureWidth, s_RenderTextureHeight, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_DepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_DepthStencilBuffer, &descDepthStencilTarget, &s_depthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Create a texture from the color buffer
    nn::gd::Texture2DDescription tex2DDesc_colorBuffer = {0, -1};
    res = nn::gd::TextureStage::CreateTexture2D(s_texture2DResource_ColorBuffer, &tex2DDesc_colorBuffer, &s_texture2D_colorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

#ifdef TEST_CASTING_TEXTURE2DRESOURCE
    {
        nn::gd::Texture2DResource* texture2DResource;
        res = nn::gd::Resource::CreateTexture2DResourceCastFrom(s_texture2DResource_DepthStencilBuffer, nn::gd::Resource::NATIVE_FORMAT_RGB_888, nn::gd::Memory::LAYOUT_BLOCK_8, &texture2DResource);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        nn::gd::Texture2DDescription tex2DDesc = {0, 0};
        res = nn::gd::TextureStage::CreateTexture2D(texture2DResource, &tex2DDesc, &s_texture2D_FromCastedResource);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
        nn::gd::Resource::ReleaseTexture2DResource(texture2DResource);
    }
#endif
}

void RenderToTexture()
{
    static nn::gd::TargetProperties renderTargetProperty;
    static bool firstTime = true;
    if (firstTime)
    {
        nn::gd::OutputStage::GetRenderTargetProperties(s_renderTarget, &renderTargetProperty);
        firstTime = false;
    }

    //Projection
    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.02f, 0.02f, -0.02f*s_RenderTextureHeight/s_RenderTextureWidth, 0.02f*s_RenderTextureHeight/s_RenderTextureWidth, 0.2f, 100.f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_proj, static_cast<f32*>(proj));

    //Clear the render targets
    u8 color[] = {0, 0, 0, 255};
    nn::gd::Memory::ClearTargets(s_renderTarget, s_depthStencilTarget, color, 1.f, 0);

    //Set the color/depth/stencil targets
    nn::gd::OutputStage::SetRenderTarget(s_renderTarget);
    nn::gd::OutputStage::SetDepthStencilTarget(s_depthStencilTarget);

    //Viewport
    nn::gd::Viewport viewPort(0, 0, renderTargetProperty.m_Width, renderTargetProperty.m_Height);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    //nn::gd::OutputStage::SetDepthStencilState(NULL, 0);

    //Texture unit 0
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D_BMP);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, s_samplerDefault);
  
#if TEST_DRAW_USING_IMMEDIAT_MODE == 1
    {
        void* tabImmediatBuffers[] = {(void*)coords, texcoords};
        u32 offsets[] = { 0, 0 };

        nn::gd::System::DrawImmediateIndexed(18, 0, tabImmediatBuffers, offsets, idxs, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

        nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D_BMP2);
        nn::gd::System::DrawImmediateIndexed(18, 18, tabImmediatBuffers, offsets, idxs, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
    }
#else
    //DRAW
    nn::gd::System::DrawIndexed(18, 0);

    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D_BMP2);
    nn::gd::System::DrawIndexed(18, 18);
#endif

}

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

#ifdef ENABLE_LOG
//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;
#endif

int DrawFrame(void)
{
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);
#ifdef ENABLE_LOG
    s64 startTime = GetTimeMicroSeconds();
#endif
    nnResult res;

#ifdef TEST_COPY_VERTEXBUFFER
    if (f == 200)
    {
        nn::gd::VertexBufferResourceProperties properties;
        nn::gd::Resource::GetVertexBufferResourceProperties(bufferCoord_Bis, &properties);
        nn::gd::Memory::CopyVertexBufferSubResource(bufferCoord_Bis, 0, properties.m_ByteSize, bufferCoord, 0);
    }
#endif

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);
    }

    //static int firstTimeModelView = true;
    //if (firstTimeModelView)
    {
        //Modelview
        int angle = f;
        nn::math::Matrix34 eye, rot1, rot2;
        nn::math::Vector3 camPos(0.f, 0.0f, 15.0f );
        nn::math::Vector3 camUp(0.f, 1.f, 0.f);
        nn::math::Vector3 target(0.f, 0.f, 0.f);
        nn::math::Vector3 rotAxis1(0.f, 1.f, 0.f);
        nn::math::Vector3 rotAxis2(1.f, 0.f, 0.f);
        nn::math::MTX34Identity(&eye);
        nn::math::MTX34LookAt(&eye, &camPos, &camUp, &target);
        nn::math::MTX34RotAxisDeg(&rot1, &rotAxis1, -1.f * angle);
        nn::math::MTX34RotAxisDeg(&rot2, &rotAxis2, -3.f * angle);
        nn::math::MTX34Mult(&eye, &eye, &rot1);
        nn::math::MTX34Mult(&eye, &eye, &rot2);
        nn::math::Matrix44 mv(eye);

        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_view, static_cast<f32*>(mv));
        //firstTimeModelView = false;
    }

    //Render to texture
    RenderToTexture();

    // test mapping the texture2DResource
    /*static nn::gd::Texture2DResource* s_texture2DResource_ResourceTestMapping = 0;
    {
        //we cannot directly write in a ressource allocated in the VRAM, so we first need to copy it in the FCRAM
        if (s_texture2DResource_ResourceTestMapping == 0)
        {
            nn::gd::Texture2DResourceDescription Text2DResDesc = {s_RenderTextureWidth, s_RenderTextureHeight, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
            res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, 0, GD_FALSE, &s_texture2DResource_ResourceTestMapping);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
        }

        nn::gd::Memory::Rect memRect(0, 0, -1, -1);
        res = nn::gd::Memory::CopyTextureSubResource(s_texture2DResource_ColorBuffer, 0, memRect, s_texture2DResource_ResourceTestMapping, 0, 0, 0);

        u8* data;
        nn::gd::Resource::MapTexture2DResource(s_texture2DResource_ResourceTestMapping, 0, nn::gd::Resource::MAP_READ_WRITE, (void**)&data);

        //We need to wait for all previously command requests to be done before acceding to the data
        nngxWaitCmdlistDone();

        int k=0;
        for (int y=0; y<s_RenderTextureHeight*s_RenderTextureWidth>>1; y++)
        {
            data[k++] = 255;
            data[k++] = 255;
            data[k++] = 255;
            data[k++] = 255;
        }

        nn::gd::Resource::UnmapTexture2DResource(s_texture2DResource_ResourceTestMapping);

        res = nn::gd::Memory::CopyTextureSubResource(s_texture2DResource_ResourceTestMapping, 0, memRect, s_texture2DResource_ColorBuffer, 0, 0, 0);
    }*/


    //Generate mipmaps
    nn::gd::Memory::GenerateMipMaps(s_texture2DResource_ColorBuffer, 0, -1);

    /********************************************************/
    /********************** Display 0 ***********************/
    {
        //Clear the render buffers
        u8 color[] = {25, 25, 122, 255};
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, color, 1.f, 0);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Set the viewport
        nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
        nn::gd::RasterizerStage::SetViewport(viewPort);

        //Projection matrix
        nn::math::Matrix44 proj;
        nn::math::MTX44Frustum(&proj, -0.02f, 0.02f, -0.02f*nn::gx::DISPLAY0_HEIGHT/nn::gx::DISPLAY0_WIDTH, 0.02f*nn::gx::DISPLAY0_HEIGHT/nn::gx::DISPLAY0_WIDTH, 0.2f, 100.f);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_proj, static_cast<f32*>(proj));

        //Texture unit 0
        nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, s_samplerLinear);
#ifdef TEST_CASTING_TEXTURE2DRESOURCE
        nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D_FromCastedResource);
#else
        nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D_colorBuffer);
#endif
        //nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D_BMP);

        //Draw
        nn::gd::System::DrawIndexed(36, 0);

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_565,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );

    }

    /********************************************************/
    /********************** Display 1 ***********************/
#if DISPLAY_BOTTOM_MODE == TEST_NONE
    //Simply clear display 1
    {
        //Clear the color buffer
        u8 color[] = {25, 25, 122, 255};
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, color, 1.f, 0);

        nn::gd::Memory::Rect memRect(0, 0, -1, -1);
        res = nn::gd::Memory::CopyTextureSubResource(s_texture2DResource_ColorBuffer, 0, memRect, s_texture2DResource_MainColorBuffer, 0, 0, (disp1_Height - 256) / 2);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        //Transfer the framebuffer to the display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        //Transfer the framebuffer to the display buffer
        res = nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
                s_texture2DResource_MainColorBuffer,                    //source
                0,                                                      //sourceMipLevelIndex
                0,                                                      //srcOffsetY
                (u8*)dstAddr,                                           //dstAddr
                disp1_Width,                                            //dstWidth
                disp1_Height,                                           //dstHeight
                nn::gd::Resource::NATIVE_FORMAT_RGB_565,                //dstFormat
                nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
                GD_FALSE                                                //yFlip
            );
    }
#elif DISPLAY_BOTTOM_MODE == TEST_COPYING_SUBRESOURCE
    //Test copying subresources
    {
        static nn::gd::Texture2DResource* s_texture2DResource_ResourceTestCopy = 0;
        if (s_texture2DResource_ResourceTestCopy == 0)
        {
            nn::gd::Texture2DResourceDescription Text2DResDesc = {nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
            res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, 0, GD_FALSE, &s_texture2DResource_ResourceTestCopy);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
        }

        //Clear the resource
        u8 color[] = {10, 10, 10, 255};
        res = nn::gd::Memory::ClearTexture2DResource(s_texture2DResource_ResourceTestCopy, 0, 0, color);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        int offsx = ((f/50)%25) - 5;
        int offsy = ((f/30)%35) - 5;
        nn::gd::Memory::Rect memRect(offsx*8, offsy*8, 96 + (offsx/2) * 8, 96 + (offsy/2) * 8);
        //nn::gd::Memory::Rect memRect(0, 0, -1, -1);

        res = nn::gd::Memory::CopyTextureSubResource(s_texture2DResource_MainColorBuffer, 0, memRect, s_texture2DResource_ResourceTestCopy, 0, offsx*8, offsy*8);

        //Transfer the resource to the display
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        res = nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
                s_texture2DResource_ResourceTestCopy,                   //source
                0,                                                      //sourceMipLevelIndex
                0,                                                      //srcOffsetY
                (u8*)dstAddr,                                           //dstAddr
                disp1_Width,                                            //dstWidth
                disp1_Height,                                           //dstHeight
                nn::gd::Resource::NATIVE_FORMAT_RGB_565,                //dstFormat
                nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
                GD_FALSE                                                //yFlip
            );

        //TODO: If we clear and transfer the resource to the display buffer one more time, the displays will show the 2 transferred data sets interleaved (need to investigate)
        /*
        res = nn::gd::Memory::ClearTexture2DResource(s_texture2DResource_ResourceTestCopy, 0, 0, color);

        res = nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
                s_texture2DResource_ResourceTestCopy,                   //source
                0,                                                      //sourceMipLevelIndex
                0,                                                      //srcOffsetY
                (u8*)dstAddr,                                           //dstAddr
                disp1_Width,                                            //dstWidth
                disp1_Height,                                           //dstHeight
                nn::gd::Resource::NATIVE_FORMAT_RGB_565,                //dstFormat
                nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
                GD_FALSE                                                //yFlip
            );
        */
    }
#else
    //Test copying resources with block-to-linear and linear-to-block features
    {
        static nn::gd::Texture2DResource* s_texture2DResource_ResourceTestBlock = 0;
        static nn::gd::Texture2DResource* s_texture2DResource_ResourceTestLinear = 0;
        static nn::gd::Texture2DResource* s_texture2DResource_ResourceTestBlockBis = 0;
        if (s_texture2DResource_ResourceTestBlock == 0)
        {
            nn::gd::Texture2DResourceDescription Text2DResDesc = {nn::gx::DISPLAY1_WIDTH*2, nn::gx::DISPLAY1_HEIGHT*2, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
            res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, 0, GD_FALSE, &s_texture2DResource_ResourceTestBlock);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

            nn::gd::Texture2DResourceDescription Text2DResDescL = {nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGB_565,  nn::gd::Memory::LAYOUT_LINEAR, nn::gd::Memory::VRAMA};
            res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDescL, 0, GD_FALSE, &s_texture2DResource_ResourceTestLinear);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

            nn::gd::Texture2DResourceDescription Text2DResDescBb = {nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGB_565,  nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
            res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDescBb, 0, GD_FALSE, &s_texture2DResource_ResourceTestBlockBis);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
        }

        //Clear the resource
        u8 color[] = {255, 0, 0, 255};
        res = nn::gd::Memory::ClearTexture2DResource(s_texture2DResource_ResourceTestBlock, 0, 0, color);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        //int offsx = ((f/50)%25);
        int offsy = ((f/20)%25);
        nn::gd::Memory::Rect memRect(0, 0, -1, -1);

        res = nn::gd::Memory::CopyTextureSubResource(s_texture2DResource_MainColorBuffer, 0, memRect, s_texture2DResource_ResourceTestBlock, 0, 0, 0);

        //We could also directly transfer s_texture2DResource_ResourceTestBlock to the display buffer, but we just want to test the function below
        res = nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
                s_texture2DResource_ResourceTestBlock,                   //source
                0,                                                      //sourceMipLevelIndex
                0,                                                      //srcOffsetY
                -1,                                                     //srcCountRow
                s_texture2DResource_ResourceTestLinear,                 //dest
                0,                                                      //destMipLevelIndex
                0,                                              //destOffsetY
                nn::gd::Memory::DOWNSCALING_2X2,                        //DownScalingMode
                GD_FALSE                                                //yFlip
            );

        //With format 24 bits, Width must be a multiple of 32 so we cannot transfer the display buffer back  O_o
        res = nn::gd::Memory::CopyTexture2DResourceLinearToBlock(
            s_texture2DResource_ResourceTestLinear,         //source
            0,                                              //srcMipLevelIndex
            offsy * 3,                                              //srcOffsetY
            320 - offsy*8,                                             //srcCountRow
            s_texture2DResource_ResourceTestBlockBis,       //dstAddr
            0,                                              //dstMipLevelIndex
            0                                              //dstOffsetY
            );

        //Transfer s_texture2DResource_ResourceTestLinear to display 1
        //(s_texture2DResource_ResourceTestLinear has already a linear layout so the result will appear broken)
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        res = nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
                s_texture2DResource_ResourceTestBlockBis,               //source
                0,                                                      //sourceMipLevelIndex
                0,                                                      //srcOffsetY
                (u8*)dstAddr,                                           //dstAddr
                disp1_Width,                                            //dstWidth
                disp1_Height,                                           //dstHeight
                nn::gd::Resource::NATIVE_FORMAT_RGB_565,                //dstFormat
                nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
                GD_FALSE                                                //yFlip
            );


       /* //If format is 24 bits, Width must be multiple of 32 so we cannot transfer the display buffer back  O_o
        res = nn::gd::Memory::CopyTexture2DResourceLinearToBlock(
                (u8*)dstAddr,
                disp1_Width,
                disp1_Height,
                s_texture2DResource_ResourceTestLinear,
                0,
                0
            );
          */

    }
#endif

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

#ifdef ENABLE_LOG
    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }
#endif
    //if (f==100) return -1;
    return 0;
}

static void Finalize()
{
    nngxStopCmdlist();

    //Release GD objects
    if (s_samplerDefault) nn::gd::TextureStage::ReleaseSamplerState(s_samplerDefault);
    if (s_samplerLinear) nn::gd::TextureStage::ReleaseSamplerState(s_samplerLinear);

    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (s_texture2D_BMP) nn::gd::TextureStage::ReleaseTexture2D(s_texture2D_BMP);
    if (s_texture2D_BMP2) nn::gd::TextureStage::ReleaseTexture2D(s_texture2D_BMP2);
    if (s_texture2D_FromCastedResource) nn::gd::TextureStage::ReleaseTexture2D(s_texture2D_FromCastedResource);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (bufferCoord) nn::gd::Resource::ReleaseVertexBufferResource(bufferCoord);
    if (bufferCoord_Bis) nn::gd::Resource::ReleaseVertexBufferResource(bufferCoord_Bis);
    if (bufferTexCoord) nn::gd::Resource::ReleaseVertexBufferResource(bufferTexCoord);
    if (bufferIndex) nn::gd::Resource::ReleaseVertexBufferResource(bufferIndex);
    if (inputLayout) nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout);

    if (shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader) nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_texture2D_colorBuffer) nn::gd::TextureStage::ReleaseTexture2D(s_texture2D_colorBuffer);
    if (s_renderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_renderTarget);
    if (s_texture2DResource_ColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_ColorBuffer);
    if (s_depthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_depthStencilTarget);
    if (s_texture2DResource_DepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_DepthStencilBuffer);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();
    
    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Initialize memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Initialize display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB565, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB565, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB565, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB565, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main frame buffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

//Will corrupt the texture data
//nngxStopCmdlist();
    InitShader();
    InitObjects();
    Initialize_GD();
//nngxRunCmdlist();
    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

