﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ShadowKnot.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/math.h>
#include <nn/init.h>
#include <nn/os.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>
#include "Util.h"
#include "Loader.h"

#include "demo.h"

#include <nn/gd.h>

#include <string.h>
#include <math.h>

#include "Memory.h"

#include <cstdlib>
#include <nn/dbg.h>

#define WAIT_VSYNC

// Performs orthogonal projection shadow mapping if enabled.
//#define SHADOW_ORTHO
#define SOFT_SHADOW

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

// Shadow map parameters
const u32 SHADOW_TEXTURE_WIDTH  = 512;
const u32 SHADOW_TEXTURE_HEIGHT = 512;

// The bias and wScale values differ between projections and orthogonal projections
#ifdef SHADOW_ORTHO
    const f32 PROJECTION_N = 0.1f, PROJECTION_F = 10.f, PROJECTION_R = 0.8f, PROJECTION_T = 0.8f;
    const f32 SHADOW_BIAS = 0.01f;
    const f32 SHADOW_WSCALE = 1.0f;
    const f32 SILHOUETTE_WIDTH = 0.15f;
#else
    const f32 PROJECTION_N = 0.1f, PROJECTION_F = 10.f, PROJECTION_R = 0.1f, PROJECTION_T = 0.1f;
    const f32 SHADOW_BIAS = PROJECTION_N * 1.2f / (PROJECTION_F - PROJECTION_N);
    const f32 SHADOW_WSCALE = PROJECTION_F;
    const f32 SILHOUETTE_WIDTH = 0.05f;
#endif // SHADOW_ORTHO

// Light position
const f32 LIGHT_POS[] = {1.0f, 2.0f, -1.0f, 1.0f};

// Frame count
static u32 s_FrameNo = 0;

// Display buffer
static u32 s_Displaybuffer0[2];
static u32 s_Displaybuffer1[2];
static u32 s_CurrentDisplaybuffer0 = 0;
static u32 s_CurrentDisplaybuffer1 = 0;

// Depth stencil
static nn::gd::DepthStencilState* s_pDepthStencilEnabled = 0;

// Texture combiner
static nn::gd::CombinerState* s_pCombinerShadowRender = 0;
static nn::gd::CombinerState* s_pCombinerHardShadow   = 0;
static nn::gd::CombinerState* s_pCombinerSoftShadow   = 0;

// Frame Buffer
static nn::gd::Texture2DResource*  s_pMainColorBufferResource        = 0;
static nn::gd::Texture2DResource*  s_pMainDepthStencilBufferResource = 0;
static nn::gd::RenderTarget*       s_pMainRenderTarget               = 0;
static nn::gd::DepthStencilTarget* s_pMainDepthStencilTarget         = 0;
static nn::gd::TargetProperties    s_MainRenderTargetProperty;

// Shadow buffer
static nn::gd::Texture2DResource* s_pShadowTextureResource = 0;
static nn::gd::Texture2D*         s_pShadowTexture         = 0;
static nn::gd::RenderTarget*      s_pShadowRenderTarget    = 0;

// Shader
static nn::gd::ShaderBinary*   s_pShaderBinary                 = 0;

static nn::gd::ShaderPipeline* s_pShaderPipelineHardShadow     = 0;
static nn::gd::ShaderPipeline* s_pShaderPipelineSoftShadow     = 0;
static nn::gd::ShaderPipeline* s_pShaderPipelineShadowRender   = 0;

static nn::gd::Shader*         s_pVertexShaderHardShadowPass   = 0;
static nn::gd::Shader*         s_pVertexShaderSoftShadowPass   = 0;
static nn::gd::Shader*         s_pGeometryShaderSoftShadowPass = 0;
static nn::gd::Shader*         s_pVertexShaderShadowRenderPass = 0;

// Input layout
static nn::gd::InputLayout* s_pInputLayoutShadowRender = 0;
static nn::gd::InputLayout* s_pInputLayoutHardShadow   = 0;
static nn::gd::InputLayout* s_pInputLayoutSoftShadow   = 0;

// Sampler
nn::gd::SamplerState* s_pSamplerShadow = 0;

// Uniform location
static nn::gd::UniformLocation s_LocationHardShadowProjection;
static nn::gd::UniformLocation s_LocationHardShadowModelView;
static nn::gd::UniformLocation s_LocationSoftShadowProjection;
static nn::gd::UniformLocation s_LocationSoftShadowModelView;
static nn::gd::UniformLocation s_LocationShadowRenderProjection;
static nn::gd::UniformLocation s_LocationShadowRenderModelView;
static nn::gd::UniformLocation s_LocationShadowRenderTexture;

// ExpHeap for app.
static nn::fnd::ExpHeap s_AppHeap;
static uptr s_HeapForGx;
static const u32 s_GxHeapSize = 0x400000;

static u32 s_CommandList;

static u8* s_pShadowTexturedata = 0; 

// FPS display
#define FPS_AVG_FRAME_COUNT 300
static u32 s_CurrentFpsCount = 0;
static f64 s_TmpFpsSum = 0.0;

#define PRINT(__msg) NN_LOG __msg

struct DatGd
{
    nn::gd::VertexBufferResource*          posVB;
    nn::gd::VertexBufferResource*          normVB;
    nn::gd::VertexBufferResource*          texVB;
    nn::gd::VertexBufferResource*          tangVB;
    nn::gd::VertexBufferResource*          wghtVB;
    nn::gd::VertexBufferResource*          midxVB;
    nn::gd::VertexBufferResource*          idxVB;
    dat_obj_t*      obj;
    dat_patch_t*    patch;
    u32             numObj;
};

static DatGd  s_Knot, s_KnotSilh, s_Plane;

void LoadDatGd(const char* fileName, DatGd* datGd);
void UnloadDatGd(DatGd* dat);

void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL)
    {
        NN_LOG("Return from \"%s\"\n",  funcname);
    }
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL)
    {
        NN_LOG("Error: %s\n", str);
    }

    NN_ASSERTMSG( 0, "callBack Error function called");
}

s64 GetTimeMicroSeconds(void)
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

static void SetupLightingLut(void)
{
    f32 lut[512];
    memset(lut, 0, sizeof(lut));

    for (int i = 1; i < 128; i++)
    {
        lut[i] = powf( (f32)i / 127.f, 30.f);
        lut[i + 255] = lut[i] - lut[i - 1];
    }
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_RR, 0, lut, lut+256, 256);

    for (int i = 1; i < 128; i++)
    {
        lut[i] = powf( (f32)i / 127.f, 2.f);
        lut[i + 255] = lut[i] - lut[i - 1];
    }
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_FR, 0, lut, lut+256, 256);
}

static void LoadObjects(void)
{
    LoadDatGd( "rom:/resources/knot.dat", &s_Knot);
    LoadDatGd( "rom:/resources/knotsilh.dat", &s_KnotSilh);
    LoadDatGd( "rom:/resources/plane.dat", &s_Plane);
}

static void UnloadObjects(void)
{
    UnloadDatGd(&s_Knot);
    UnloadDatGd(&s_KnotSilh);
    UnloadDatGd(&s_Plane);
}

static void RenderObjectsShadow(DatGd& objs)
{
    for (int i = 0; i < objs.numObj; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(s_pInputLayoutHardShadow);

        nn::gd::VertexBufferResource* buffers[] = { objs.posVB };
        u32 offsets[] = { 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 1, buffers, offsets);
        for (unsigned j = objs.obj[i].patch_offset; j < objs.obj[i].patch_size + objs.obj[i].patch_offset; j++)
        {
            nn::gd::VertexInputStage::SetIndexBuffer(objs.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

            nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
            nn::gd::System::DrawIndexed(objs.patch[j].elm_size, 0);
        }
    }
}

static void RenderObjectsShadowSilhouette(DatGd& objs)
{
    for (int i = 0; i < objs.numObj; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(s_pInputLayoutSoftShadow);

        nn::gd::VertexBufferResource* buffers[] = { objs.posVB, objs.normVB };
        u32 offsets[] = { 0, 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);
        for (unsigned j = objs.obj[i].patch_offset; j < objs.obj[i].patch_size + objs.obj[i].patch_offset; j++)
        {
            nn::gd::VertexInputStage::SetIndexBuffer(objs.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

            nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);
            nn::gd::System::DrawIndexed(objs.patch[j].elm_size, 0);
        }
    }
}

static void RenderObjectsShadowed(DatGd& objs)
{
    for (int i = 0; i < objs.numObj; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(s_pInputLayoutShadowRender);

        nn::gd::VertexBufferResource* buffers[] = { objs.posVB, objs.normVB };
        u32 offsets[] = { objs.obj[i].vtx_offset, objs.obj[i].nrm_offset };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);

        for (unsigned j = objs.obj[i].patch_offset; j < objs.obj[i].patch_size + objs.obj[i].patch_offset; j++)
        {
            f32 ma[4] = {0.f, 0.f, 0.f, 1.f};
            f32 md[4] = {0.f, 0.f, 0.f, 1.f};
            f32 ms[4] = {0.f, 0.f, 0.f, 1.f};

            for (int col = 0; col < 3; col++)
            {
                ma[col] = objs.patch[j].ambient[col];
                md[col] = objs.patch[j].diffuse[col];
                ms[col] = objs.patch[j].specular[col];
            }

            f32 ld0[] = {1.f, 1.f, 1.f, 1.f};  // Diffuse
            f32 ls0[] = {0.f, 0.f, 0.f, 0.f};  // specular0
            f32 ls1[] = {1.f, 1.f, 1.f, 1.f};  // specular1
            f32 la0[] = {1.f, 1.f, 1.f, 1.f};  // Ambient

            u8 Ambinet[3];
            Ambinet[0] = nn::gd::Utils::FloatToUnsignedByte(ma[0] * la0[0]);
            Ambinet[1] = nn::gd::Utils::FloatToUnsignedByte(ma[1] * la0[1]);
            Ambinet[2] = nn::gd::Utils::FloatToUnsignedByte(ma[2] * la0[2]);
            nn::gd::LightingStage::light[0].SetColorAmbient(Ambinet[0], Ambinet[1], Ambinet[2]);

            u8 Diffuse[3];
            Diffuse[0] = nn::gd::Utils::FloatToUnsignedByte(ld0[0] * md[0]);
            Diffuse[1] = nn::gd::Utils::FloatToUnsignedByte(ld0[1] * md[1]);
            Diffuse[2] = nn::gd::Utils::FloatToUnsignedByte(ld0[2] * md[2]);
             nn::gd::LightingStage::light[0].SetColorDiffuse(Diffuse[0], Diffuse[1], Diffuse[2]);

            u8 Specular0[3];
            Specular0[0] = nn::gd::Utils::FloatToUnsignedByte(ms[0] * ls0[0]);
            Specular0[1] = nn::gd::Utils::FloatToUnsignedByte(ms[1] * ls0[1]);
            Specular0[2] = nn::gd::Utils::FloatToUnsignedByte(ms[2] * ls0[2]);
            nn::gd::LightingStage::light[0].SetColorSpecular0(Specular0[0], Specular0[1], Specular0[2]);

            nn::gd::LightingStage::light[0].SetColorSpecular1(nn::gd::Utils::FloatToUnsignedByte(ls1[0]), nn::gd::Utils::FloatToUnsignedByte(ls1[1]), nn::gd::Utils::FloatToUnsignedByte(ls1[2]));
            nn::gd::VertexInputStage::SetIndexBuffer(objs.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

            nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
            nn::gd::System::DrawIndexed(objs.patch[j].elm_size, 0);
        }
    }
}

static void Draw1stPass(void)
{
    // Render target
    nn::gd::OutputStage::SetRenderTarget(s_pShadowRenderTarget);

    // Viewport
    nn::gd::Viewport viewPort(0, 0, SHADOW_TEXTURE_WIDTH, SHADOW_TEXTURE_HEIGHT);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    // Clear target
    u8 color[] = {255, 255, 255, 255};
    nn::gd::Memory::ClearTargets(s_pShadowRenderTarget, NULL, color, 1.0f, 0);

    // Shader
    nn::gd::ShaderStage::SetShaderPipeline(s_pShaderPipelineHardShadow);

    // Shadow parameters
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, NULL);
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_SHADOW_MAP);
    nn::gd::OutputStage::SetDepthRangeMode(nn::gd::OutputStage::DEPTHBUFFER_LINEAR, 0.0f, SHADOW_WSCALE, 0);

    // Turn light source OFF
    nn::gd::LightingStage::light[0].EnableLight(GD_FALSE);

    // Set combiner
    nn::gd::CombinerStage::SetTextureCombinerState(s_pCombinerHardShadow);

    // Create projection matrix
    nn::math::Matrix44 proj;
#ifdef SHADOW_ORTHO
    nn::math::MTX44Ortho(&proj, -PROJECTION_R, PROJECTION_R, -PROJECTION_T, PROJECTION_T, PROJECTION_N, PROJECTION_F);
#else
    nn::math::MTX44Frustum(&proj, -PROJECTION_R, PROJECTION_R, -PROJECTION_T, PROJECTION_T, PROJECTION_N, PROJECTION_F);
#endif
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineHardShadow, s_LocationHardShadowProjection, static_cast<f32*>(proj));

    // Create view matrix
    nn::math::Matrix34 mv, rot;
    nn::math::Vector3 camPos(LIGHT_POS[0], LIGHT_POS[1], LIGHT_POS[2]);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::Vector3 axis(1.f, 1.f, 1.f);
    
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Matrix44 tmp(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineHardShadow, s_LocationHardShadowModelView, static_cast<f32*>(tmp));

    RenderObjectsShadow(s_Plane);

    nn::math::MTX34RotAxisDeg(&rot, &axis, static_cast<f32>(s_FrameNo));
    nn::math::MTX34Mult(&mv, &mv, &rot);
    nn::math::Matrix44 tmp2(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineHardShadow, s_LocationHardShadowModelView, static_cast<f32*>(tmp2));

    RenderObjectsShadow(s_Knot);

    // Soft shadow rendering
#ifdef SOFT_SHADOW
    nn::gd::ShaderStage::SetShaderPipeline(s_pShaderPipelineSoftShadow);

    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineSoftShadow, s_LocationSoftShadowProjection, static_cast<f32*>(proj));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineSoftShadow, s_LocationSoftShadowModelView, static_cast<f32*>(tmp2));

    // Uniform settings for the silhouette shader 
    f32 silhwidth[2];
    silhwidth[0] = SILHOUETTE_WIDTH;
    silhwidth[1] = silhwidth[0] * static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT);
    f32 silhcolor[4] = {1.f, 1.f, 1.f, 1.f};
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineSoftShadow, nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineSoftShadow, "dmp_Silhouette.width"), static_cast<f32*>(silhwidth));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineSoftShadow, nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineSoftShadow, "dmp_Silhouette.color"), static_cast<f32*>(silhcolor));
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(s_pShaderPipelineSoftShadow, nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineSoftShadow, "dmp_Silhouette.frontFaceCCW"), 1, 1);
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(s_pShaderPipelineSoftShadow, nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineSoftShadow, "dmp_Silhouette.acceptEmptyTriangles"), 0, 1);
#ifdef SHADOW_ORTHO
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(s_pShaderPipelineSoftShadow, nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineSoftShadow, "dmp_Silhouette.scaleByW"), 0, 1);
#else
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(s_pShaderPipelineSoftShadow, nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineSoftShadow, "dmp_Silhouette.scaleByW"), 1, 1);
#endif // SHADOW_ORTHO
    // Texture combiner
    nn::gd::CombinerStage::SetTextureCombinerState(s_pCombinerSoftShadow);

    // Shadow parameters
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);
    nn::gd::OutputStage::SetPenumbraScaleBias(0.0f, 1.0f);
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_SHADOW_MAP);
    nn::gd::OutputStage::SetDepthRangeMode(nn::gd::OutputStage::DEPTHBUFFER_LINEAR, 0.0f, SHADOW_WSCALE, 0);

    RenderObjectsShadowSilhouette(s_KnotSilh);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);
#endif // SOFT_SHADOW

}

static void Draw2ndPass(void)
{
    // Render target
    nn::gd::OutputStage::SetRenderTarget(s_pMainRenderTarget);
    nn::gd::OutputStage::SetDepthStencilTarget(s_pMainDepthStencilTarget);

    // Viewport
    nn::gd::Viewport viewPort(0, 0, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    // Clear target
    u8 color[] = {92, 107, 128, 255};
    nn::gd::Memory::ClearTargets(s_pMainRenderTarget, s_pMainDepthStencilTarget, color, 1.f, 0);

    // Shader
    nn::gd::ShaderStage::SetShaderPipeline(s_pShaderPipelineShadowRender);

    // Use textures generated over cumulative passes
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_pShadowTexture);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, s_pSamplerShadow);

    // Combiner settings
    nn::gd::CombinerStage::SetTextureCombinerState(s_pCombinerShadowRender);

    // Shadow bias setting
    nn::gd::TextureStage::SetShadowZBias(SHADOW_BIAS);

#ifdef SHADOW_ORTHO
    nn::gd::TextureStage::SetPerspectiveShadow(GD_FALSE);
#else
    nn::gd::TextureStage::SetPerspectiveShadow(GD_TRUE);
#endif
    nn::gd::OutputStage::SetDepthRangeMode(nn::gd::OutputStage::DEPTHBUFFER_RECIPROCAL_FUNCTION, 0.f, 1.f, 0);
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_FRAGMENT_COLOR);

    // Light source settings
    nn::gd::LightingStage::light[0].EnableLight(GD_TRUE);
    nn::gd::LightingStage::light[0].EnableShadowed(GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_RR, nn::gd::LightingStage::INPUT_VALUE_NH);
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_FR, nn::gd::LightingStage::INPUT_VALUE_LN);
    nn::gd::LightingStage::EnableLookUpTableReflection(GD_TRUE);
    nn::gd::LightingStage::SetLayerConfiguration(nn::gd::LightingStage::LAYER_CONFIGURATION_1);
    nn::gd::LightingStage::SetFresnelSelector(nn::gd::LightingStage::FRESNEL_SELECTOR_TYPE_PRIMARY_SECONDARY_ALPHA_FRESNEL);
    nn::gd::LightingStage::SetShadowAlpha(GD_TRUE);
    nn::gd::LightingStage::SetInvertShadow(GD_TRUE);

    // Create projection matrix
    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.1f, 0.1f, -0.1f, 0.1f, 0.2f, 10.f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineShadowRender, s_LocationShadowRenderProjection, static_cast<f32*>(proj));

    // Create view matrix
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(3.f * sinf(s_FrameNo * static_cast<f32>(M_PI)/180.f), 1.f, 3.f * cosf(s_FrameNo * static_cast<f32>(M_PI)/180.f));
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Matrix44 tmp(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineShadowRender, s_LocationShadowRenderModelView, static_cast<f32*>(tmp));

    // Light source settings
    nn::math::Vector4 mv0(mv.m[0][0], mv.m[0][1], mv.m[0][2], mv.m[0][3]);
    nn::math::Vector4 mv1(mv.m[1][0], mv.m[1][1], mv.m[1][2], mv.m[1][3]);
    nn::math::Vector4 mv2(mv.m[2][0], mv.m[2][1], mv.m[2][2], mv.m[2][3]);

    nn::math::Vector4 vlpos0(LIGHT_POS);

    nn::gd::LightingStage::light[0].SetPosition(nn::math::VEC4Dot(&mv0, &vlpos0), nn::math::VEC4Dot(&mv1, &vlpos0), nn::math::VEC4Dot(&mv2, &vlpos0));

    // Create texture transform matrix for referencing shadow textures.
    nn::math::Matrix34 textmp;
#ifdef SHADOW_ORTHO
    nn::math::MTX34ShadowProjectionOrtho(&textmp, -PROJECTION_R, PROJECTION_R, -PROJECTION_T, PROJECTION_T, PROJECTION_N, PROJECTION_F);
#else
    nn::math::MTX34ShadowProjectionFrustum(&textmp, -PROJECTION_R, PROJECTION_R, -PROJECTION_T, PROJECTION_T, PROJECTION_N, PROJECTION_F);
#endif
    nn::math::Matrix44 tex(textmp);

    nn::math::Matrix34 tmpmv;
    nn::math::Vector3 camPos2(LIGHT_POS[0], LIGHT_POS[1], LIGHT_POS[2]);
    nn::math::MTX34LookAt(&tmpmv, &camPos2, &camUp, &target);
    nn::math::Matrix44 eye(tmpmv);
    nn::math::MTX44Mult(&tex, &tex, &eye);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineShadowRender, s_LocationShadowRenderTexture, static_cast<f32*>(tex));

    RenderObjectsShadowed(s_Plane);

    nn::math::Matrix34 rot;
    nn::math::Vector3 axis(1.f, 1.f, 1.f);
    
    nn::math::MTX34RotAxisDeg(&rot, &axis, static_cast<f32>(s_FrameNo));
    nn::math::MTX34Mult(&mv, &mv, &rot);

    tmp = nn::math::Matrix44(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineShadowRender, s_LocationShadowRenderModelView, static_cast<f32*>(tmp));

    tmp = nn::math::Matrix44(rot);
    nn::math::MTX44Mult(&tex, &tex, &tmp);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_pShaderPipelineShadowRender, s_LocationShadowRenderTexture, static_cast<f32*>(tex));

    RenderObjectsShadowed(s_Knot);
}

int DrawFrame(void)
{
    s64 startTime = GetTimeMicroSeconds();

    static int disp0Format = 0, disp0Width = 0, disp0Height = 0;
    static int disp1Format = 0, disp1Width = 0, disp1Height = 0;
    if (disp0Width == 0)
    {
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_pMainRenderTarget, &s_MainRenderTargetProperty);
    }

    /********************** Display 0 ***********************/
    {
        // Cumulative shadow passes
        Draw1stPass();
        // Shadow access pass
        Draw2ndPass();

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_Displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_pMainColorBufferResource,                             //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0Width,                                             //dstWidth
            disp0Height,                                            //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_Displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);


#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (s_FrameNo == 0)
    {
        nngxStartLcdDisplay();
    }

    // FPS display
    s_TmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_CurrentFpsCount == FPS_AVG_FRAME_COUNT)
    {
        f64 avgtime = s_TmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.5f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_CurrentFpsCount = 0;
        s_TmpFpsSum = 0.0;
    }

    s_FrameNo++;

    return 0;
}
static void InitShader(void)
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;

    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &s_pShaderBinary);
    NN_UTIL_PANIC_IF_FAILED(res);
    res = nn::gd::ShaderStage::CreateShader(s_pShaderBinary, 0, &s_pVertexShaderHardShadowPass);
    NN_UTIL_PANIC_IF_FAILED(res);
    res = nn::gd::ShaderStage::CreateShader(s_pShaderBinary, 1, &s_pVertexShaderSoftShadowPass);
    NN_UTIL_PANIC_IF_FAILED(res);
    res = nn::gd::ShaderStage::CreateShader(s_pShaderBinary, 3, &s_pGeometryShaderSoftShadowPass);
    NN_UTIL_PANIC_IF_FAILED(res);
    res = nn::gd::ShaderStage::CreateShader(s_pShaderBinary, 2, &s_pVertexShaderShadowRenderPass);
    NN_UTIL_PANIC_IF_FAILED(res);

    res = nn::gd::ShaderStage::CreateShaderPipeline(s_pVertexShaderHardShadowPass, NULL, &s_pShaderPipelineHardShadow);
    NN_UTIL_PANIC_IF_FAILED(res);
    res = nn::gd::ShaderStage::CreateShaderPipeline(s_pVertexShaderSoftShadowPass, s_pGeometryShaderSoftShadowPass, &s_pShaderPipelineSoftShadow);
    NN_UTIL_PANIC_IF_FAILED(res);
    res = nn::gd::ShaderStage::CreateShaderPipeline(s_pVertexShaderShadowRenderPass, NULL, &s_pShaderPipelineShadowRender);
    NN_UTIL_PANIC_IF_FAILED(res);

    // Shader uniform location
    s_LocationHardShadowProjection = nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineHardShadow, "uProjection");
    s_LocationHardShadowModelView  = nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineHardShadow, "uModelView");

    s_LocationSoftShadowProjection = nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineSoftShadow, "uProjection");
    s_LocationSoftShadowModelView  = nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineSoftShadow, "uModelView");

    s_LocationShadowRenderProjection = nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineShadowRender, "uProjection");
    s_LocationShadowRenderModelView  = nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineShadowRender, "uModelView");
    s_LocationShadowRenderTexture    = nn::gd::ShaderStage::GetShaderUniformLocation(s_pShaderPipelineShadowRender, "uTexture");

    file.Finalize();
    s_AppHeap.Free(buf);


}
static void InitObjects(void)
{
    // Input layout (HardShadow)
    {
        nn::gd::InputElementDescription descs[] = { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0};
        u32 strides[] = {3 * sizeof(float)} ;
        nnResult res = nn::gd::VertexInputStage::CreateInputLayout(descs, 1, strides, s_pVertexShaderHardShadowPass, &s_pInputLayoutHardShadow);
        NN_UTIL_PANIC_IF_FAILED(res);
    }

    // Input layout (SoftShadow)
    {
        nn::gd::InputElementDescription descs[] = 
        {
            { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
            { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
        };
        u32 strides[] = {3 * sizeof(float), 3 * sizeof(float)} ;
        nnResult res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, s_pVertexShaderSoftShadowPass, &s_pInputLayoutSoftShadow);
        NN_UTIL_PANIC_IF_FAILED(res);
    }

    // Input layout (ShadowRender)
    {
        nn::gd::InputElementDescription descs[] = 
        {
            { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
            { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
        };
        u32 strides[] = {3 * sizeof(float), 3 * sizeof(float)} ;
        nnResult res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, s_pVertexShaderShadowRenderPass, &s_pInputLayoutShadowRender);
        NN_UTIL_PANIC_IF_FAILED(res);
    }

    // Load objects
    LoadObjects();

}
static void InitializeGd(void)
{
    // Depth stencil
    nn::gd::DepthStencilStateDescription DSDesc;
    DSDesc.ToDefault();
    DSDesc.m_DepthEnable = GD_TRUE;
    DSDesc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&DSDesc, &s_pDepthStencilEnabled);
    nn::gd::OutputStage::SetDepthStencilState(s_pDepthStencilEnabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

    // Shadow texture resources
    s_pShadowTexturedata = reinterpret_cast<u8*>( s_AppHeap.Allocate(4 * SHADOW_TEXTURE_WIDTH * SHADOW_TEXTURE_HEIGHT) );
    memset(s_pShadowTexturedata, 255, 4 * SHADOW_TEXTURE_WIDTH * SHADOW_TEXTURE_HEIGHT);

    nn::gd::Texture2DResourceDescription Text2DResDesc = {SHADOW_TEXTURE_WIDTH, SHADOW_TEXTURE_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_SHADOW, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, s_pShadowTexturedata, GD_FALSE, &s_pShadowTextureResource);
    NN_UTIL_PANIC_IF_FAILED(res);

    // Create a shadow texture from resources
    nn::gd::Texture2DDescription desc = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(s_pShadowTextureResource, &desc, &s_pShadowTexture);
    NN_UTIL_PANIC_IF_FAILED(res);

    // Create shadow render target
    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_pShadowTextureResource, &descRenderTarget, &s_pShadowRenderTarget);
    NN_UTIL_PANIC_IF_FAILED(res);

    // Set shadow sampler
    nn::gd::SamplerStateDescription samplerDesc;
    samplerDesc.SetShadow();
    samplerDesc.m_BorderColor [0] = samplerDesc.m_BorderColor [1] = samplerDesc.m_BorderColor [2] = samplerDesc.m_BorderColor [3] = 255;
    res = nn::gd::TextureStage::CreateSamplerState(&samplerDesc, &s_pSamplerShadow);
    NN_UTIL_PANIC_IF_FAILED(res);

    nn::gd::CombinerDescription combinerShadowPass;
    combinerShadowPass.ToDefault();
    combinerShadowPass.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_ADD_MULT);
    combinerShadowPass.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combinerShadowPass.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_FRAGMENT_SECONDARY_COLOR, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR);
    combinerShadowPass.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_FRAGMENT_SECONDARY_COLOR, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combinerShadowPass.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_ONE_MINUS_SRC_ALPHA);
    combinerShadowPass.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
    combinerShadowPass.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);

    nn::gd::CombinerStage::CreateTextureCombinerState(&combinerShadowPass, &s_pCombinerShadowRender);

    // Texture combiner (HardShadow)
    nn::gd::CombinerDescription combinerHardShadow;
    combinerHardShadow.ToDefault();
    combinerHardShadow.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combinerHardShadow, &s_pCombinerHardShadow);

    // Texture combiner (SoftShadow)
    nn::gd::CombinerDescription combinerSoftShadow;
    combinerSoftShadow.ToDefault();
    combinerSoftShadow.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combinerSoftShadow.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combinerSoftShadow.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combinerSoftShadow, &s_pCombinerSoftShadow);

    // Light source settings
    nn::gd::LightingStage::SetGlobalColorAmbient(0, 0, 0);

    // Create LUT for light source
    SetupLightingLut();
}
static int Initialize(void)
{
    nn::fs::Initialize();

    // Memory Initialization
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    // Initialize command list
    nngxGenCmdlists(1, &s_CommandList);
    nngxBindCmdlist(s_CommandList);
    nngxCmdlistStorage(0x40000, 128);

    nngxRunCmdlist();

    // Initialize display buffer
    nngxGenDisplaybuffers(2, s_Displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_Displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_Displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_Displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_Displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_Displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    // Initialize GD
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    // Initialize framebuffer
    // Color buffer
    nn::gd::Texture2DResourceDescription ResourceDescColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&ResourceDescColorBuffer, 0, GD_FALSE, &s_pMainColorBufferResource);
    NN_UTIL_PANIC_IF_FAILED(res);

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_pMainColorBufferResource, &descRenderTarget, &s_pMainRenderTarget);
    NN_UTIL_PANIC_IF_FAILED(res);

    // Depth stencil buffer
    nn::gd::Texture2DResourceDescription ResourceDescDepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&ResourceDescDepthStencilBuffer, 0, GD_FALSE, &s_pMainDepthStencilBufferResource);
    NN_UTIL_PANIC_IF_FAILED(res);

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_pMainDepthStencilBufferResource, &descDepthStencilTarget, &s_pMainDepthStencilTarget);
    NN_UTIL_PANIC_IF_FAILED(res);

    InitShader();
    InitObjects();
    InitializeGd();

    return 0;
}

static void Finalize()
{
    nngxStopCmdlist();

    // Release model data
    UnloadObjects();

    // Release the GD object.
    if (s_pShadowTexturedata) s_AppHeap.Free(s_pShadowTexturedata);

    if (s_pDepthStencilEnabled) nn::gd::OutputStage::ReleaseDepthStencilState(s_pDepthStencilEnabled);

    if (s_pCombinerShadowRender) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_pCombinerShadowRender);
    if (s_pCombinerHardShadow)   nn::gd::CombinerStage::ReleaseTextureCombinerState(s_pCombinerHardShadow);
    if (s_pCombinerSoftShadow)   nn::gd::CombinerStage::ReleaseTextureCombinerState(s_pCombinerSoftShadow);

    if (s_pMainColorBufferResource)         nn::gd::Resource::ReleaseTexture2DResource(s_pMainColorBufferResource);
    if (s_pMainDepthStencilBufferResource)  nn::gd::Resource::ReleaseTexture2DResource(s_pMainDepthStencilBufferResource);
    if (s_pMainRenderTarget)                nn::gd::OutputStage::ReleaseRenderTarget(s_pMainRenderTarget);
    if (s_pMainDepthStencilTarget)          nn::gd::OutputStage::ReleaseDepthStencilTarget(s_pMainDepthStencilTarget);

    if (s_pShadowTextureResource) nn::gd::Resource::ReleaseTexture2DResource(s_pShadowTextureResource);
    if (s_pShadowTexture)         nn::gd::TextureStage::ReleaseTexture2D(s_pShadowTexture);
    if (s_pShadowRenderTarget)    nn::gd::OutputStage::ReleaseRenderTarget(s_pShadowRenderTarget);
    if (s_pSamplerShadow)         nn::gd::TextureStage::ReleaseSamplerState(s_pSamplerShadow);

    if (s_pShaderPipelineHardShadow)     nn::gd::ShaderStage::ReleaseShaderPipeline(s_pShaderPipelineHardShadow);
    if (s_pShaderPipelineSoftShadow)     nn::gd::ShaderStage::ReleaseShaderPipeline(s_pShaderPipelineSoftShadow);
    if (s_pShaderPipelineShadowRender)   nn::gd::ShaderStage::ReleaseShaderPipeline(s_pShaderPipelineShadowRender);
    if (s_pVertexShaderHardShadowPass)   nn::gd::ShaderStage::ReleaseShader(s_pVertexShaderHardShadowPass);
    if (s_pVertexShaderSoftShadowPass)   nn::gd::ShaderStage::ReleaseShader(s_pVertexShaderSoftShadowPass);
    if (s_pGeometryShaderSoftShadowPass) nn::gd::ShaderStage::ReleaseShader(s_pGeometryShaderSoftShadowPass);
    if (s_pVertexShaderShadowRenderPass) nn::gd::ShaderStage::ReleaseShader(s_pVertexShaderShadowRenderPass);
    if (s_pShaderBinary)                 nn::gd::ShaderStage::ReleaseShaderBinary(s_pShaderBinary);

    if (s_pInputLayoutShadowRender) nn::gd::VertexInputStage::ReleaseInputLayout(s_pInputLayoutShadowRender);
    if (s_pInputLayoutHardShadow)   nn::gd::VertexInputStage::ReleaseInputLayout(s_pInputLayoutHardShadow);
    if (s_pInputLayoutSoftShadow)   nn::gd::VertexInputStage::ReleaseInputLayout(s_pInputLayoutSoftShadow);

    // Release the GD library
    nn::gd::System::Finalize();

    // Release the display buffer
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_Displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_Displaybuffer1);
    s_Displaybuffer0[0] = s_Displaybuffer0[1] = 0;
    s_Displaybuffer1[0] = s_Displaybuffer1[1] = 0;

    // Release the command list
    nngxDeleteCmdlists(1, &s_CommandList);

    nngxFinalize();

    // Release memory
    demo::FinalizeMemoryManager();
}

void nnMain(void)
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    // shutdown_display 
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

void LoadDatGd(const char* fileName, DatGd* datGd)
{
    if (!fileName || !datGd)
        return;

    char* fname = 0;
    wchar_t* wfname = 0;
    unsigned char* dat = 0;
    long long int filesize;
    nn::fs::FileReader file;

    do
    {
        fname = (char*)s_AppHeap.Allocate(strlen(fileName) + 1);
        if (!fname)
            break;

        int dotpos = 0;
        strcpy(fname, fileName);
        for (int i = (int)strlen(fname) - 1; i >= 0; i--)
        {
            if (fname[i] == '.')
            {
                dotpos = i;
                break;
            }
        }
        if (strncmp(&fname[dotpos + 1], "dat", 3) != 0)
            break;

        wfname = static_cast<wchar_t*>(s_AppHeap.Allocate( (strlen(fname)+1)*2 ));
        if (!wfname)
            break;
        (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
        file.Initialize(wfname);
        filesize = file.GetSize();
        dat = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
        if (!dat)
            break;
        filesize = file.Read(dat, filesize);

        dat_header_t* header = (dat_header_t*)dat;
        dat_obj_t* obj = (dat_obj_t*)&dat[sizeof(dat_header_t)];
        dat_patch_t* patch = (dat_patch_t*)&dat[sizeof(dat_header_t) + sizeof(dat_obj_t) * header->obj_num];

        nn::fs::FileReader attribfile;
        unsigned char* attribdata = 0;
        nnResult res;
        nn::gd::VertexBufferResourceDescription desc;
        memset(datGd, 0, sizeof(dat_t));

        if (header->vtx_size)
        {
            memcpy(&fname[dotpos + 1], "vtx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->vtx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &datGd->posVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->nrm_size)
        {
            memcpy(&fname[dotpos + 1], "nrm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->nrm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &datGd->normVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tex_size)
        {
            memcpy(&fname[dotpos + 1], "tex", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tex_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &datGd->texVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tgt_size)
        {
            memcpy(&fname[dotpos + 1], "tgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &datGd->tangVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->wgt_size)
        {
            memcpy(&fname[dotpos + 1], "wgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->wgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &datGd->wghtVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->idx_size)
        {
            memcpy(&fname[dotpos + 1], "idx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->idx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &datGd->midxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->elm_size)
        {
            memcpy(&fname[dotpos + 1], "elm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->elm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &datGd->idxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }

        datGd->obj = (dat_obj_t*)s_AppHeap.Allocate(sizeof(dat_obj_t) * header->obj_num);
        datGd->patch = (dat_patch_t*)s_AppHeap.Allocate(sizeof(dat_patch_t) * header->patch_num);
        if (!datGd->obj || !datGd->patch)
        {
            if (datGd->obj)
                s_AppHeap.Free(datGd->obj);
            if (datGd->patch)
                s_AppHeap.Free(datGd->patch);
            break;
        }
        memcpy(datGd->obj, obj, sizeof(dat_obj_t) * header->obj_num);
        memcpy(datGd->patch, patch, sizeof(dat_patch_t) * header->patch_num);
        datGd->numObj = header->obj_num;
    } while (0);

    if (fname)
        s_AppHeap.Free(fname);
    if (wfname)
        s_AppHeap.Free(wfname);
    if (dat)
        s_AppHeap.Free(dat);
    file.Finalize();

    return;
}

void UnloadDatGd(DatGd* dat)
{
    if (dat)
    {
        if (dat->posVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->posVB);
        if (dat->normVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->normVB);
        if (dat->texVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->texVB);
        if (dat->tangVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->tangVB);
        if (dat->wghtVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->wghtVB);
        if (dat->midxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->midxVB);
        if (dat->idxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->idxVB);
        if (dat->obj)
            s_AppHeap.Free(dat->obj);
        if (dat->patch)
            s_AppHeap.Free(dat->patch);
    }

    return;
}
