﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ShadowSimple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "demo.h"

#include <nn/gd.h>

#define WAIT_VSYNC

//Shadow texture parameters
#define SBWIDTH 512
#define SBHEIGHT 512
#define SZBIASSCALE 1.3f
#define SZSCALE 1.f
const float s_lightPosition[] = {0.f, 0.f, 0.15f, 1.f};

// 1st pass (shadow accumulation pass) frustum argument
const float frustrum_n = 0.1f, frustrum_f = 10.f, frustrum_sw2 = 0.5f, frustrum_sh2 = 0.5f;

//Display buffers
u32 s_displaybuffer0[2];
u32 s_displaybuffer1[2];
u32 s_CurrentDisplaybuffer0 = 0;
u32 s_CurrentDisplaybuffer1 = 0;
u8 s_clearColor[4];

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_tc_shadowPass = 0;
static nn::gd::CombinerState* s_tc_shadowHard = 0;
static nn::gd::CombinerState* s_tc_shadowSoft = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;
static nn::gd::TargetProperties mainRenderTargetProperty;

//shadow frame buffer
static nn::gd::Texture2DResource* s_texture2DRes_shadow = 0;
static nn::gd::Texture2D* s_tex_shadow = 0;
static nn::gd::RenderTarget* s_renderTarget_shadow = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline_hardShadow = 0;
static nn::gd::ShaderPipeline* shaderPipeline_softShadow = 0;
static nn::gd::ShaderPipeline* shaderPipeline_shadowPass = 0;

static nn::gd::Shader* vertexShader_hardShadowPass = 0;
static nn::gd::Shader* vertexShader_softShadowPass = 0;
static nn::gd::Shader* geometryShader_softShadowPass = 0;
static nn::gd::Shader* vertexShader_shadowRendering = 0;

//Vertex input
static nn::gd::VertexBufferResource* bufferCoord = 0;
static nn::gd::VertexBufferResource* bufferColor = 0;
static nn::gd::VertexBufferResource* bufferNormal = 0;
static nn::gd::VertexBufferResource* bufferIndex = 0;
static nn::gd::VertexBufferResource* bufferIndexSilhouette = 0;
static nn::gd::InputLayout* inputLayout_shadowRendering = 0;
static nn::gd::InputLayout* inputLayout_shadowHard = 0;
static nn::gd::InputLayout* inputLayout_shadowSoft = 0;

//sampler
nn::gd::SamplerState* s_sampler_shadow = 0;

//static nn::gd::UniformLocation uniformLoc_projection;
//static nn::gd::UniformLocation uniformLoc_modelview;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

u32 s_CmdlistID;

static u8* ShadowTexturedata = 0; 


#define PRINT(__msg) NN_LOG __msg

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);                     if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader_hardShadowPass);      if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 1, &vertexShader_softShadowPass);      if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 3, &geometryShader_softShadowPass);    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 2, &vertexShader_shadowRendering);     if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader_hardShadowPass, NULL, &shaderPipeline_hardShadow);      if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader_softShadowPass, geometryShader_softShadowPass, &shaderPipeline_softShadow);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader_shadowRendering, NULL, &shaderPipeline_shadowPass);     if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //uniformLoc_projection = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection");
    //uniformLoc_modelview = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");

    //nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void)
{
    float coords[] = {
        -0.4f, -0.3f, -0.39f,
        -0.3f, 0.3f, -0.1f,
        0.0f, 0.3f, -0.0f,
        -0.75f, 0.75f, -0.4f,
        0.75f, 0.75f, -0.4f,
        0.75f, -0.75f, -0.4f,
        -0.75f, -0.75f, -0.4f,

        -0.4f, -0.3f, -0.4f,
        -0.3f, 0.3f, -0.11f,
        0.0f, 0.3f, -0.01f,
    };
    float colors[] = {
        0.7f, 0.7f, 0.7f, 1.f,
        0.7f, 0.7f, 0.7f, 1.f,
        0.7f, 0.7f, 0.7f, 1.f,
        0.0f, 0.0f, 1.0f, 1.f,
        0.0f, 0.0f, 1.0f, 1.f,
        0.0f, 0.0f, 1.0f, 1.f,
        0.0f, 0.0f, 1.0f, 1.f,
    };
    float normals[] =
    {
        -0.0f,-1.622f,-0.906f,
        -0.902f, 0.89f, 0.114f,
        0.8f, 0.732f, 0.092f,
        0.f, 0.f, 0.f,
        0.f, 0.f, 0.f,
        0.f, 0.f, 0.f,
        0.f, 0.f, 0.f,
        0.f, 0.f, 0.f,
        0.f, 0.f, 0.f,
        0.f, 0.f, 0.f,
    };
    unsigned short originIndex[] = {  2, 1, 0, 6, 5, 3, 4, };
    unsigned short silhIndex[] = { 2, 1, 7, 0, 8, 9, };

    //Buffer resources
    nnResult res;
    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize = sizeof(coords);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, coords, &bufferCoord);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(colors);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, colors, &bufferColor);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(normals);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, normals, &bufferNormal);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(originIndex);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, originIndex, &bufferIndex);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(silhIndex);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, silhIndex, &bufferIndexSilhouette);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}


    //Input Layout
    {
        nn::gd::InputElementDescription descs[] =
        {
            { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
            { 1, "aColor", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0}
        };
        u32 strides[] = { sizeof(float) * 3, sizeof(float) * 4};
        res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShader_shadowRendering, &inputLayout_shadowRendering);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    }

    {
        nn::gd::InputElementDescription descs[] =
        {
            { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
        };

        u32 strides[] = { sizeof(float) * 3};
        res = nn::gd::VertexInputStage::CreateInputLayout(descs, 1, strides, vertexShader_hardShadowPass, &inputLayout_shadowHard);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    }

    {
        nn::gd::InputElementDescription descs[] =
        {
            { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
            { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
        };

        u32 strides[] = { sizeof(float) * 3, sizeof(float) * 3 };
        res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShader_softShadowPass, &inputLayout_shadowSoft);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    }
}

void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription DSdesc;
    DSdesc.ToDefault();
    DSdesc.m_DepthEnable = GD_TRUE;
    DSdesc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&DSdesc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

    //Clear color
    s_clearColor[0] = 80; s_clearColor[1] = 100; s_clearColor[2] = 127; s_clearColor[3] = 255;

    //The shadow texture resource
    ShadowTexturedata = reinterpret_cast<u8*>( s_AppHeap.Allocate(4 * SBWIDTH * SBHEIGHT) );
    memset(ShadowTexturedata, 255, 4 * SBWIDTH * SBHEIGHT);

    nn::gd::Texture2DResourceDescription Text2DResDesc = {SBWIDTH, SBHEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_SHADOW, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, ShadowTexturedata, GD_FALSE, &s_texture2DRes_shadow);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!\n");}

    //The shadow texture
    nn::gd::Texture2DDescription desc = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(s_texture2DRes_shadow, &desc, &s_tex_shadow);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!\n");}

    //renderTarget to the shadow texture
    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DRes_shadow, &descRenderTarget, &s_renderTarget_shadow);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //sampler desc
    nn::gd::SamplerStateDescription samplerdesc;
    samplerdesc.SetShadow();
    samplerdesc.m_BorderColor [0] = samplerdesc.m_BorderColor [1] = samplerdesc.m_BorderColor [2] = samplerdesc.m_BorderColor [3] = 255;
    res = nn::gd::TextureStage::CreateSamplerState(&samplerdesc, &s_sampler_shadow);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Texture combiner
    nn::gd::CombinerDescription combinerShadowPass;
    combinerShadowPass.ToDefault();
    combinerShadowPass.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MODULATE);
    combinerShadowPass.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_MODULATE);
    combinerShadowPass.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combinerShadowPass.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combinerShadowPass.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combinerShadowPass, &s_tc_shadowPass);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::CombinerDescription combinerHard;
    combinerHard.ToDefault();
    combinerHard.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combinerHard, &s_tc_shadowHard);

    nn::gd::CombinerDescription combinerSoftShadow;
    combinerSoftShadow.ToDefault();
    combinerSoftShadow.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combinerSoftShadow.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combinerSoftShadow.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combinerSoftShadow, &s_tc_shadowSoft);
}

static void Draw1stPass()
{
    //Set the color targets
    nn::gd::OutputStage::SetRenderTarget(s_renderTarget_shadow);

    //Clear targets
    u8 color[] = {255, 255, 255, 255};
    nn::gd::Memory::ClearTargets(s_renderTarget_shadow, NULL, color, 1.f, 0);

    //Set viewport
    nn::gd::Viewport viewPort(0, 0, SBWIDTH, SBHEIGHT);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    //Set the shader
    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline_hardShadow);

    //modelview shader uniform
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(s_lightPosition);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Matrix44 m(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline_hardShadow, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_hardShadow, "uModelView"), static_cast<f32*>(m));

    //Projection shader uniform
    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -frustrum_sw2, frustrum_sw2, -frustrum_sh2, frustrum_sh2, frustrum_n, frustrum_f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline_hardShadow, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_hardShadow, "uProjection"), static_cast<f32*>(proj));

    //Set texture combiner
    nn::gd::CombinerStage::SetTextureCombinerState(s_tc_shadowHard);

    //Shadow parameters
    nn::gd::OutputStage::SetDepthRangeMode(nn::gd::OutputStage::DEPTHBUFFER_LINEAR, 0.f, frustrum_f, 0);
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_SHADOW_MAP);

    //Set InputLayout
    nn::gd::VertexInputStage::SetInputLayout(inputLayout_shadowHard);

    //Set index and vertex buffers
    nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

    nn::gd::VertexBufferResource* buffersHardPass[] = { bufferCoord };
    u32 offsetsHardPass[] = { 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 1, buffersHardPass, offsetsHardPass);

    //Draw
    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
    nn::gd::System::DrawIndexed(3, 0);

    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);
    nn::gd::System::DrawIndexed(4, 3);

    //Set soft shadow shader
    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline_softShadow);

    //modelview
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline_softShadow, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_softShadow, "uModelView"), static_cast<f32*>(m));

    //Projection
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline_softShadow, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_softShadow, "uProjection"), static_cast<f32*>(proj));

    //Geometry shader uniforms
    f32 silhwidth[2];
    silhwidth[0] = 0.05f;
    silhwidth[1] = silhwidth[0] * (float)nn::gx::DISPLAY0_WIDTH / (float)nn::gx::DISPLAY0_HEIGHT;
    f32 silhcolor[4] = {1.f, 1.f, 1.f, 1.f};
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline_softShadow, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_softShadow, "dmp_Silhouette.width"), static_cast<f32*>(silhwidth));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline_softShadow, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_softShadow, "dmp_Silhouette.color"), static_cast<f32*>(silhcolor));
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(shaderPipeline_softShadow, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_softShadow, "dmp_Silhouette.frontFaceCCW"), 1, 1);
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(shaderPipeline_softShadow, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_softShadow, "dmp_Silhouette.acceptEmptyTriangles"), 0, 1);
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(shaderPipeline_softShadow, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_softShadow, "dmp_Silhouette.scaleByW"), 0, 1);

    //Texture combiner
    nn::gd::CombinerStage::SetTextureCombinerState(s_tc_shadowSoft);

    //Shadow parameters
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);
    nn::gd::OutputStage::SetPenumbraScaleBias(0.0f, 1.0f);
    nn::gd::OutputStage::SetDepthRangeMode(nn::gd::OutputStage::DEPTHBUFFER_LINEAR, 0.f, frustrum_f, 0);
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_SHADOW_MAP);
    
    //Set InputLayout
    nn::gd::VertexInputStage::SetInputLayout(inputLayout_shadowSoft);

    //Set index and vertex buffers
    nn::gd::VertexInputStage::SetIndexBuffer(bufferIndexSilhouette, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

    nn::gd::VertexBufferResource* buffersSoftPass[] = { bufferCoord, bufferNormal };
    u32 offsetsSoftPass[] = { 0, 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffersSoftPass, offsetsSoftPass);

    //Draw
    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);
    //nn::gd::VertexInputStage::debugFunc();
    nn::gd::System::DrawIndexed(6, 0);

    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);
}

static void Draw2ndPass()
{
    //Set the color/depth/stencil targets
    nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
    nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

    //Set the viewport
    nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    //Set the shader
    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline_shadowPass);

    //Set the Texture combiner
    nn::gd::CombinerStage::SetTextureCombinerState(s_tc_shadowPass);

    //Set the Texture shadow and sampler state
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_tex_shadow);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, s_sampler_shadow);
    nn::gd::TextureStage::SetPerspectiveShadow(GD_TRUE);

    //Shadow parameters
    f32 shadowBias = frustrum_n * SZBIASSCALE / (frustrum_f - frustrum_n);
    nn::gd::TextureStage::SetShadowZBias(shadowBias);

    nn::gd::OutputStage::SetDepthRangeMode(nn::gd::OutputStage::DEPTHBUFFER_RECIPROCAL_FUNCTION, 0.f, 1.f, 0);
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_FRAGMENT_COLOR);

    //modelview shader uniform
    nn::math::Matrix34 mv, mvl, tmp[2];
    nn::math::Vector3 trans(0.f, 0.f, -0.1f);
    nn::math::MTX34RotXYZDeg(&mv, 0.f, 0.f, 270.f);
    nn::math::MTX34Translate(&tmp[0], &trans);
    nn::math::MTX34RotXYZDeg(&tmp[1], -35.f, 0.f, 0.f);
    nn::math::MTX34Mult(&mvl, &mv, &tmp[0]);
    nn::math::MTX34Mult(&mvl, &mvl, &tmp[1]);
    nn::math::Matrix44 m(mvl);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline_shadowPass, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_shadowPass, "uModelView"), static_cast<f32*>(m));

    //Projection shader uniform
    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.7f, 0.7f, -0.7f, 0.7f, 0.2f, 10.f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline_shadowPass, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_shadowPass, "uProjection"), static_cast<f32*>(proj));

    //Texture shader uniform
    nn::math::Matrix34 textmp;
    nn::math::MTX34ShadowProjectionFrustum(&textmp,  -frustrum_sw2, frustrum_sw2, -frustrum_sh2, frustrum_sh2, frustrum_n, frustrum_f );
    nn::math::Matrix44 tex(textmp);
    nn::math::Vector3 camPos(s_lightPosition);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&tmp[1], &camPos, &camUp, &target);
    nn::math::Matrix44 eye(tmp[1]);
    nn::math::MTX44Mult(&tex, &tex, &eye);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline_shadowPass, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline_shadowPass, "uTexture"), static_cast<f32*>(tex));

    //Set InputLayout
    nn::gd::VertexInputStage::SetInputLayout(inputLayout_shadowRendering);

    //Set index and vertex buffers
    nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

    nn::gd::VertexBufferResource* buffers[] = { bufferCoord, bufferColor };
    u32 offsets[] = { 0, 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);

    //Draw
    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
    nn::gd::System::DrawIndexed(3, 0);

    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);
    nn::gd::System::DrawIndexed(4, 3);
}


//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);
    s64 startTime = GetTimeMicroSeconds();

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);
    }

    /********************************************************/
    /********************** Display 0 ***********************/
    {
        /********* Main pass (2nd pass) **********/
        //Clear the main frame buffers
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, s_clearColor, 1.f, 0);

        Draw1stPass();
        Draw2ndPass();

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);


#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    return 0;
}

static void Finalize()
{
    nngxStopCmdlist();

    //Release GD objects
    if (ShadowTexturedata) s_AppHeap.Free(ShadowTexturedata);

    if (s_tc_shadowPass) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_tc_shadowPass);
    if (s_tc_shadowHard) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_tc_shadowHard);
    if (s_tc_shadowSoft) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_tc_shadowSoft);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (bufferCoord) nn::gd::Resource::ReleaseVertexBufferResource(bufferCoord);
    if (bufferColor) nn::gd::Resource::ReleaseVertexBufferResource(bufferColor);
    if (bufferNormal) nn::gd::Resource::ReleaseVertexBufferResource(bufferNormal);
    if (bufferIndex) nn::gd::Resource::ReleaseVertexBufferResource(bufferIndex);
    if (bufferIndexSilhouette) nn::gd::Resource::ReleaseVertexBufferResource(bufferIndexSilhouette);

    if (inputLayout_shadowRendering) nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout_shadowRendering);
    if (inputLayout_shadowHard) nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout_shadowHard);
    if (inputLayout_shadowSoft) nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout_shadowSoft);

    if (shaderPipeline_hardShadow) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline_hardShadow);
    if (shaderPipeline_softShadow) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline_softShadow);
    if (shaderPipeline_shadowPass) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline_shadowPass);
    if (vertexShader_hardShadowPass) nn::gd::ShaderStage::ReleaseShader(vertexShader_hardShadowPass);
    if (vertexShader_softShadowPass) nn::gd::ShaderStage::ReleaseShader(vertexShader_softShadowPass);
    if (geometryShader_softShadowPass) nn::gd::ShaderStage::ReleaseShader(geometryShader_softShadowPass);
    if (vertexShader_shadowRendering) nn::gd::ShaderStage::ReleaseShader(vertexShader_shadowRendering);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    if (s_renderTarget_shadow) nn::gd::OutputStage::ReleaseRenderTarget(s_renderTarget_shadow);
    if (s_tex_shadow) nn::gd::TextureStage::ReleaseTexture2D(s_tex_shadow);
    if (s_texture2DRes_shadow) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DRes_shadow);

    if (s_sampler_shadow) nn::gd::TextureStage::ReleaseSamplerState(s_sampler_shadow);


    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    InitShader();
    InitObjects();
    Initialize_GD();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}
