﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SubdivisionLoopSimple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"
#include "demo.h"

#include <string.h>
#include <math.h>
#include "Memory.h"

#include <nn/gd.h>

#define WAIT_VSYNC

template <typename TYPE> class simple_set
{
public:
    simple_set()
    {
        element_size = 128;
        current_size = 0;
        element = (TYPE*)malloc(element_size * sizeof(TYPE));
    }
    ~simple_set()
    {
        free(element);
    }
    int size(){ return current_size; }
    void clear(){ current_size = 0; }
    void insert(TYPE d)
    {
        int i;
        for (i = 0; i < current_size; i++)
        {
            if (element[i] == d)
                break;
        }
        if (i == current_size)
        {
            element[current_size++] = d;
            if (current_size == element_size)
            {
                TYPE* new_element = (TYPE*)malloc(2 * element_size * sizeof(TYPE));
                memcpy(new_element, element, sizeof(TYPE) * element_size);
                element_size *= 2;
                free(element);
                element = new_element;
            }
        }
    }
    
    typedef TYPE*   iterator;
    iterator begin() { return element; }
    iterator end() { return &element[current_size]; }

protected:
    TYPE*   element;
    int     current_size;
    int     element_size;
};

template <typename TYPE> class simple_vector
{
public:
    simple_vector()
    {
        element_size = 512;
        current_size = 0;
        element = (TYPE*)malloc(element_size * sizeof(TYPE));
    }
    ~simple_vector()
    {
        free(element);
    }
    int size(){ return current_size; }
    void clear(){ current_size = 0; }
    void push_back(TYPE d)
    {
        element[current_size++] = d;
        if (current_size == element_size)
        {
            TYPE* new_element = (TYPE*)malloc(2 * element_size * sizeof(TYPE));
            memcpy(new_element, element, sizeof(TYPE) * element_size);
            element_size *= 2;
            free(element);
            element = new_element;
        }
    }
    TYPE& operator[](int idx) { return element[idx]; }
    
    typedef TYPE*   iterator;
    iterator begin() { return element; }
    iterator end() { return &element[current_size]; }
    

protected:
    TYPE*   element;
    int     current_size;
    int     element_size;
};

u32 s_NumIndex;
u32 s_NumSubdivisionIndex;
nn::gd::Viewport viewport1(0, 0, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT/2);
nn::gd::Viewport viewport2(0, nn::gx::DISPLAY0_HEIGHT/2, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT/2);
static nn::gd::UniformLocation s_modelViewMatrixGD;
static nn::gd::UniformLocation s_subDivLvlGD;

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;
u8 s_clearColor[4];

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;
static nn::gd::Shader* geometryShader = 0;

//Vertex input
static nn::gd::InputLayout* inputLayout = 0;
static nn::gd::VertexBufferResource* bufferPos = 0;
static nn::gd::VertexBufferResource* bufferColor = 0;
static nn::gd::VertexBufferResource* bufferValence = 0;
static nn::gd::VertexBufferResource* bufferIndex = 0;

static nn::gd::UniformLocation uniformLoc_projection;
static nn::gd::UniformLocation uniformLoc_modelview;
static nn::gd::UniformLocation uniformLoc_subDiv;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
u32 s_GxHeapSize = 0x400000;
u32 s_HeapSize = 0x400000;

GLuint s_CmdlistID;

#define PRINT(__msg) NN_LOG __msg

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}


static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);                        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);                        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 1, &geometryShader);                      if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, geometryShader, &shaderPipeline); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    uniformLoc_projection = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection");
    uniformLoc_modelview = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");
    uniformLoc_subDiv = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_Subdivision.level");

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    file.Finalize();
    s_AppHeap.Free(buf);
}

static GLushort* make_loopsubdivision_info(GLushort* tri_index, GLuint num_index, GLuint* num_subd_index, GLfloat* valence_array)
{
    simple_vector<GLushort> loop_subd_index;
    simple_set<GLushort> e0s, e1s, e2s;
    
    for (unsigned i = 0; i < num_index; i += 3)
    {
        GLushort v0, v1, v2;
        GLushort e00, e01, e0p, e10, e11, e1p, e20, e21, e2p;
        
        e00 = e01 = e0p = e10 = e11 = e1p = e20 = e21 = e2p = 0;    /* Avoid warning. */
        e0s.clear();
        e1s.clear();
        e2s.clear();
        
        /* get main vertices */
        v0 = tri_index[i];
        v1 = tri_index[i + 1];
        v2 = tri_index[i + 2];
        
        if (v0 == v1 || v1 == v2 || v2 == v0)
            continue;   /* this module doesn't support degenerate triangle. */
        
        for (unsigned j = 0; j < num_index; j += 3)
        {
            unsigned match_mask = 0;
            bool find_v[3] = {false};
            
            if (tri_index[j] == tri_index[j + 1] || tri_index[j + 1] == tri_index[j + 2] || tri_index[j + 2] == tri_index[j])
                continue;   /* this module doesn't support degenerate triangle. */
            
            /* find vertex sharing edge with v0, v1, v2 */
            for (int k = 0; k < 3; k++)
            {
                if (tri_index[j + k] == v0)
                {
                    e0s.insert(tri_index[j + (k + 1) % 3]);
                    e0s.insert(tri_index[j + (k + 2) % 3]);
                    find_v[0] = true;
                    match_mask |= 1 << k;
                }
                if (tri_index[j + k] == v1)
                {
                    e1s.insert(tri_index[j + (k + 1) % 3]);
                    e1s.insert(tri_index[j + (k + 2) % 3]);
                    find_v[1] = true;
                    match_mask |= 1 << k;
                }
                if (tri_index[j + k] == v2)
                {
                    e2s.insert(tri_index[j + (k + 1) % 3]);
                    e2s.insert(tri_index[j + (k + 2) % 3]);
                    find_v[2] = true;
                    match_mask |= 1 << k;
                }
            }
            
            /* find ei0 */
            if (find_v[0] && !find_v[1] && find_v[2])
            {
                switch (match_mask)
                {
                    case 0x6: e00 = tri_index[j + 0]; break;
                    case 0x5: e00 = tri_index[j + 1]; break;
                    case 0x3: e00 = tri_index[j + 2]; break;
                }
            }
            else if (find_v[0] && find_v[1] && !find_v[2])
            {
                switch (match_mask)
                {
                    case 0x6: e10 = tri_index[j + 0]; break;
                    case 0x5: e10 = tri_index[j + 1]; break;
                    case 0x3: e10 = tri_index[j + 2]; break;
                }
            }
            else if (!find_v[0] && find_v[1] && find_v[2])
            {
                switch (match_mask)
                {
                    case 0x6: e20 = tri_index[j + 0]; break;
                    case 0x5: e20 = tri_index[j + 1]; break;
                    case 0x3: e20 = tri_index[j + 2]; break;
                }
            }
        }

        for (unsigned j = 0; j < num_index; j += 3)
        {
            if (tri_index[j] == tri_index[j + 1] || tri_index[j + 1] == tri_index[j + 2] || tri_index[j + 2] == tri_index[j])
                continue;   /* this module doesn't support degenerate triangle. */
            
            for (unsigned m = 0; m < 3; m++)
            {
                unsigned match_mask = 0;
                bool find_v[3] = {false};
                
                GLushort vi0, vi1, ei1_0, ei0_p, ei1_1;
                vi0 = vi1 = ei1_0 = ei0_p = ei1_1 = 0;  /* Avoid warning. */
                
                switch (m)
                {
                    case 0: vi0 = v0; vi1 = v1; ei1_0 = e10; break;
                    case 1: vi0 = v1; vi1 = v2; ei1_0 = e20; break;
                    case 2: vi0 = v2; vi1 = v0; ei1_0 = e00; break;
                }
                
                for (int k = 0; k < 3; k++)
                {
                    if (tri_index[j + k] == vi0)
                    {
                        find_v[0] = true;
                        match_mask |= 1 << k;
                    }
                    if (tri_index[j + k] == vi1)
                    {
                        find_v[1] = true;
                        match_mask |= 1 << k;
                    }
                    if (tri_index[j + k] == ei1_0)
                    {
                        find_v[2] = true;
                        match_mask |= 1 << k;
                    }
                }
                
                /* find eip and ei1 */
                if (find_v[0] && !find_v[1] && find_v[2])
                {
                    switch (match_mask)
                    {
                        case 0x6: ei0_p = tri_index[j + 0]; break;
                        case 0x5: ei0_p = tri_index[j + 1]; break;
                        case 0x3: ei0_p = tri_index[j + 2]; break;
                    }
                    switch (m)
                    {
                        case 0: e0p = ei0_p; break;
                        case 1: e1p = ei0_p; break;
                        case 2: e2p = ei0_p; break;
                    }
                }
                else if (!find_v[0] && find_v[1] && find_v[2])
                {
                    switch (match_mask)
                    {
                        case 0x6: ei1_1 = tri_index[j + 0]; break;
                        case 0x5: ei1_1 = tri_index[j + 1]; break;
                        case 0x3: ei1_1 = tri_index[j + 2]; break;
                    }
                    switch (m)
                    {
                        case 0: e11 = ei1_1; break;
                        case 1: e21 = ei1_1; break;
                        case 2: e01 = ei1_1; break;
                    }
                }
            }
        }
        
        /* make index array */
        loop_subd_index.push_back((GLushort)(e0s.size() + e1s.size() + e2s.size() + 3));
        loop_subd_index.push_back(v0);
        loop_subd_index.push_back(v1);
        loop_subd_index.push_back(v2);
        for (simple_set<GLushort>::iterator iter = e0s.begin(); iter != e0s.end(); ++iter)
            loop_subd_index.push_back(*iter);
        for (simple_set<GLushort>::iterator iter = e1s.begin(); iter != e1s.end(); ++iter)
            loop_subd_index.push_back(*iter);
        for (simple_set<GLushort>::iterator iter = e2s.begin(); iter != e2s.end(); ++iter)
            loop_subd_index.push_back(*iter);
        loop_subd_index.push_back(12);
        loop_subd_index.push_back(v0);
        loop_subd_index.push_back(v1);
        loop_subd_index.push_back(v2);
        loop_subd_index.push_back(e00);
        loop_subd_index.push_back(e10);
        loop_subd_index.push_back(e20);
        loop_subd_index.push_back(e01);
        loop_subd_index.push_back(e11);
        loop_subd_index.push_back(e21);
        loop_subd_index.push_back(e0p);
        loop_subd_index.push_back(e1p);
        loop_subd_index.push_back(e2p);
        
        valence_array[v0] = (GLfloat)e0s.size();
        valence_array[v1] = (GLfloat)e1s.size();
        valence_array[v2] = (GLfloat)e2s.size();
    }
    *num_subd_index = (GLuint)loop_subd_index.size();
    GLushort* ret = (GLushort*)malloc(sizeof(GLushort) * loop_subd_index.size());
    memcpy(ret, &loop_subd_index[0], sizeof(GLushort) * loop_subd_index.size());
    
    #if 0
    /* dump index */
    {
        for (GLuint i = 0; i < *num_subd_index; )
        {
            int patch_size = ret[i];
            printf("%d: ", ret[i]);
            i++;
            for (int j = 0; j < patch_size; j++)
            {
                printf("%d ", ret[i]);
                i++;
            }
            printf("\n");
        }
    }
    #endif

    return ret;
}

static void InitObjects(void)
{

    f32 vertex[] =
    {
        -1.2f,   0.0f,   1.2f,
         1.2f,   0.0f,   1.2f,
         1.2f,   0.0f,  -1.2f,
        -1.2f,   0.0f,  -1.2f,
         0.0f,   1.2f,   0.0f,
         0.0f,  -1.2f,   0.0f,
    };

    f32 color[] =
    {
        1.f, 0.f, 0.f, 1.f,
        0.f, 1.f, 0.f, 1.f,
        0.f, 0.f, 1.f, 1.f,
        1.f, 1.f, 0.f, 1.f,
        1.f, 1.f, 1.f, 1.f,
        0.f, 0.f, 0.f, 1.f,
    };

    u16 index[] =
    {
        0, 1, 4,
        1, 2, 4,
        2, 3, 4,
        3, 0, 4,
        1, 0, 5,
        2, 1, 5,
        3, 2, 5,
        0, 3, 5,
    };
    
    f32* valence_array;
    u16* loopsubd_index_array;
    
    valence_array = (f32*)malloc(sizeof(f32) * (sizeof(vertex) / (sizeof(f32) * 3)));
    s_NumIndex = sizeof(index) / sizeof(u16);
    loopsubd_index_array = make_loopsubdivision_info(index, s_NumIndex, &s_NumSubdivisionIndex, valence_array);

    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0 },
        { 1, "aColor",    nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0 },
        { 2, "aValence",  nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 1, 0 },
    };

    u32 strides[] = { sizeof(f32)*3,  sizeof(f32)*4, sizeof(f32) };
    nn::gd::VertexInputStage::CreateInputLayout(descs, 3, strides, vertexShader, &inputLayout);


    nn::gd::VertexBufferResourceDescription desc;
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    desc.m_ByteSize = sizeof(vertex);
    nn::gd::Resource::CreateVertexBufferResource(&desc, vertex, &bufferPos);

    desc.m_ByteSize = sizeof(color);
    nn::gd::Resource::CreateVertexBufferResource(&desc, color, &bufferColor);

    desc.m_ByteSize = (sizeof(vertex) / 3);
    nn::gd::Resource::CreateVertexBufferResource(&desc, valence_array, &bufferValence);

    desc.m_ByteSize = sizeof(u16) * s_NumSubdivisionIndex;
    nn::gd::Resource::CreateVertexBufferResource(&desc, loopsubd_index_array, &bufferIndex);


    nn::gd::VertexBufferResource* buffers[] = { bufferPos, bufferColor, bufferValence };
    u32 offsets[] = { 0, 0, 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 3, buffers, offsets);

    nn::gd::VertexInputStage::SetInputLayout(inputLayout);
        
    nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);

    free(valence_array);
    free(loopsubd_index_array);
}

void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

    //Clear color
    s_clearColor[0] = s_clearColor[1] = 25; 
    s_clearColor[2] = 122; 
    s_clearColor[3] = 255;

    //Texture combiner
    nn::gd::CombinerDescription combiner;
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);
    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

    //Some uniforms
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_Subdivision.fragmentLightingEnabled"), 0, 1);

}

//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);
    s64 startTime = GetTimeMicroSeconds();

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Projection matrix
        nn::math::Matrix44 proj;
        nn::math::MTX44Frustum(&proj, -0.03f, 0.03f, -0.03f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.03f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection"), static_cast<f32*>(proj));
    }

    //static int firstTimeModelView = true;
    //if (firstTimeModelView)
    {
        //Modelview
        nn::math::Matrix34 mv, rot;
        nn::math::Vector3 camPos(0.f, 0.f, 7.f);
        nn::math::Vector3 camUp(0.f, 1.f, 0.f);
        nn::math::Vector3 target(0.f, 0.f, 0.f);
        nn::math::Vector3 axis(1.f, 1.f, 1.f);
    
        /* Projection settings */
        /* modelview setting */
        nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
        nn::math::MTX34RotAxisDeg(&rot, &axis, -static_cast<f32>(f));
        nn::math::MTX34Mult(&mv, &mv, &rot);
        nn::math::MTX44 m(mv);

        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_modelview, static_cast<f32*>(m));
        //firstTimeModelView = false;
    }

    /********************************************************/
    /********************** Display 0 ***********************/
    {
        //Clear the render buffers
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, s_clearColor, 1.f, 0);

        //1st pass
        nn::gd::RasterizerStage::SetViewport(viewport1);

        f32 subDiv = 0;
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_subDiv, &subDiv);
        nn::gd::System::DrawIndexed(s_NumSubdivisionIndex, 0);

        //2nd pass
        nn::gd::RasterizerStage::SetViewport(viewport2);
        subDiv = 2.f;
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_subDiv, &subDiv);
        nn::gd::System::DrawIndexed(s_NumSubdivisionIndex, 0);

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    return 0;
}

static void Finalize()
{
    nngxStopCmdlist();

    //Release GD objects
    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (bufferPos) nn::gd::Resource::ReleaseVertexBufferResource(bufferPos);
    if (bufferColor) nn::gd::Resource::ReleaseVertexBufferResource(bufferColor);
    if (bufferValence) nn::gd::Resource::ReleaseVertexBufferResource(bufferValence);
    if (bufferIndex) nn::gd::Resource::ReleaseVertexBufferResource(bufferIndex);
    if (inputLayout) nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout);

    if (shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader) nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (geometryShader) nn::gd::ShaderStage::ReleaseShader(geometryShader);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    InitShader();
    InitObjects();
    Initialize_GD();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

