﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SubdivisionSimple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"
#include "demo.h"

#include <string.h>
#include <math.h>
#include "Memory.h"

#include <nn/gd.h>

#define WAIT_VSYNC

template <typename TYPE> class simple_vector
{
public:
    simple_vector()
    {
        element_size = 512;
        current_size = 0;
        element = (TYPE*)malloc(element_size * sizeof(TYPE));
    }
    ~simple_vector()
    {
        free(element);
    }
    int size(){ return current_size; }
    void clear(){ current_size = 0; }
    void push_back(TYPE d)
    {
        element[current_size++] = d;
        if (current_size == element_size)
        {
            TYPE* new_element = (TYPE*)malloc(2 * element_size * sizeof(TYPE));
            memcpy(new_element, element, sizeof(TYPE) * element_size);
            element_size *= 2;
            free(element);
            element = new_element;
        }
    }
    TYPE& operator[](int idx) { return element[idx]; }
    
    typedef TYPE*   iterator;
    iterator begin() { return element; }
    iterator end() { return &element[current_size]; }
    

protected:
    TYPE*   element;
    int     current_size;
    int     element_size;
};  // class simple_vector


struct object_info_t
{
    simple_vector<unsigned short> idx;
    simple_vector<float> v;
    simple_vector<float> n;
    simple_vector<float> c;
    simple_vector<float> t;
};

struct object_info_t* s_SimpleInfo;

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;
u8 s_clearColor[4];

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;
static nn::gd::Shader* geometryShader = 0;

//Vertex input
static nn::gd::InputLayout* inputLayout = 0;
static nn::gd::VertexBufferResource* bufferCoord = 0;
static nn::gd::VertexBufferResource* bufferColor = 0;
static nn::gd::VertexBufferResource* bufferIndex = 0;

static nn::gd::UniformLocation uniformLoc_projection;
static nn::gd::UniformLocation uniformLoc_modelview;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;
uptr s_HeapForMalloc;
u32 s_HeapSize = 0x400000;




GLuint s_CmdlistID;

#define PRINT(__msg) NN_LOG __msg

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}


static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);                        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);                        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 1, &geometryShader);                      if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, geometryShader, &shaderPipeline); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    uniformLoc_projection = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection");
    uniformLoc_modelview = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void)
{
    s_SimpleInfo = new object_info_t();

     for (unsigned int i = 0; i < 7; i++)
    {
        for (unsigned int j = 0; j < 7; j++)
        {

            float x = -3.f + (float)i;
            float y = -3.f + (float)j;
            float z = 0.5f*(x*x + y*y);
            s_SimpleInfo->v.push_back(x);
            s_SimpleInfo->v.push_back(y);
            s_SimpleInfo->v.push_back(z);
            /*
            s_SimpleInfo->c.push_back(( (j*7+i)%3==0 )? 1.f: 0.f);
            s_SimpleInfo->c.push_back(( (j*7+i)%3==1 )? 1.f: 0.f);
            s_SimpleInfo->c.push_back(( (j*7+i)%3==2 )? 1.f: 0.f);
            */
            s_SimpleInfo->c.push_back(( (j*7+i)%2==0 )? 1.f: 0.f);
            s_SimpleInfo->c.push_back(( (j*7+i)%2==0 )? 1.f: 0.f);
            s_SimpleInfo->c.push_back(( (j*7+i)%2==0 )? 1.f: 0.f);
        }
    }

    for (unsigned int i = 0; i < 4; i++)
    {
        for (unsigned int j = 0; j < 4; j++)
        {
            s_SimpleInfo->idx.push_back(16);
            s_SimpleInfo->idx.push_back(( j+2 )*7 + ( i+2 ));
            s_SimpleInfo->idx.push_back(( j+2 )*7 + ( i+1 ));
            s_SimpleInfo->idx.push_back(( j+1 )*7 + ( i+1 ));
            s_SimpleInfo->idx.push_back(( j+1 )*7 + ( i+2 ));
            s_SimpleInfo->idx.push_back(( j+3 )*7 + ( i+2 ));
            s_SimpleInfo->idx.push_back(( j+3 )*7 + ( i+1 ));
            s_SimpleInfo->idx.push_back(( j+3 )*7 + ( i+0 ));
            s_SimpleInfo->idx.push_back(( j+2 )*7 + ( i+0 ));
            s_SimpleInfo->idx.push_back(( j+1 )*7 + ( i+0 ));
            s_SimpleInfo->idx.push_back(( j+0 )*7 + ( i+0 ));
            s_SimpleInfo->idx.push_back(( j+0 )*7 + ( i+1 ));
            s_SimpleInfo->idx.push_back(( j+0 )*7 + ( i+2 ));
            s_SimpleInfo->idx.push_back(( j+0 )*7 + ( i+3 ));
            s_SimpleInfo->idx.push_back(( j+1 )*7 + ( i+3 ));
            s_SimpleInfo->idx.push_back(( j+2 )*7 + ( i+3 ));
            s_SimpleInfo->idx.push_back(( j+3 )*7 + ( i+3 ));
        }
    }

    //NN_LOG("INIT done! ");

    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0 },
        { 1, "aColor", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0 },
    };
    u32 strides[] = { sizeof(float)*3,  sizeof(float)*3 };
    nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShader, &inputLayout);

    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize =  sizeof(float) * s_SimpleInfo->v.size();
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    nn::gd::Resource::CreateVertexBufferResource(&desc, &s_SimpleInfo->v[0], &bufferCoord);

    desc.m_ByteSize =  sizeof(float) * s_SimpleInfo->c.size();
    nn::gd::Resource::CreateVertexBufferResource(&desc, &s_SimpleInfo->c[0], &bufferColor);

    desc.m_ByteSize = sizeof(unsigned short) * s_SimpleInfo->idx.size();
    nn::gd::Resource::CreateVertexBufferResource(&desc, &s_SimpleInfo->idx[0], &bufferIndex);

    nn::gd::VertexBufferResource* buffers[] = { bufferCoord, bufferColor };
    u32 offsets[] = { 0, 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);

    nn::gd::VertexInputStage::SetInputLayout(inputLayout);
        
    nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);
}

void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

    //Clear color
    s_clearColor[0] = s_clearColor[1] = 25; 
    s_clearColor[2] = 122; 
    s_clearColor[3] = 255;

    //Texture combiner
    nn::gd::CombinerDescription combiner;
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);
    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

    // Some uniforms for subdivision
    f32 subDiv = 1;
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_Subdivision.level"), &subDiv);
    nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_Subdivision.fragmentLightingEnabled"), 0, 1);

}

//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);
    s64 startTime = GetTimeMicroSeconds();

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Set the viewport
        nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
        nn::gd::RasterizerStage::SetViewport(viewPort);

        //Projection matrix
        nn::math::Matrix44 proj;
        nn::math::MTX44Frustum(&proj, -0.07f, 0.07f, -0.07f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.07f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_projection, static_cast<f32*>(proj));
    }

    //static int firstTimeModelView = true;
    //if (firstTimeModelView)
    {
        //Modelview
        nn::math::Matrix34 mv, rot;
        nn::math::Vector3 camPos(0.f, 0.f, 15.f);
        nn::math::Vector3 camUp(0.f, 1.f, 0.f);
        nn::math::Vector3 target(0.f, 0.f, 0.f);
        nn::math::Vector3 trans(0.f, 0.f, 10*sinf(static_cast<f32>(f) / 36.f));
    
        nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
        nn::math::MTX34MultTranslate(&mv, &mv, &trans);
    
        nn::math::MTX34RotXYZDeg(&rot, 0.f, -static_cast<f32>(f), 0.f);
        nn::math::MTX34Mult(&mv, &mv, &rot);
    
        nn::math::Matrix44 m(mv);

        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_modelview, static_cast<f32*>(m));
        //firstTimeModelView = false;
    }

    /********************************************************/
    /********************** Display 0 ***********************/
    {
        //Clear the render buffers
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, s_clearColor, 1.f, 0);

        //Draw
        nn::gd::System::DrawIndexed(s_SimpleInfo->idx.size(), 0);

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    return 0;
}

static void Finalize()
{
    nngxStopCmdlist();

    //Release GD objects
    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (bufferCoord) nn::gd::Resource::ReleaseVertexBufferResource(bufferCoord);
    if (bufferColor) nn::gd::Resource::ReleaseVertexBufferResource(bufferColor);
    if (bufferIndex) nn::gd::Resource::ReleaseVertexBufferResource(bufferIndex);
    if (inputLayout) nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout);

    if (shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader) nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (geometryShader) nn::gd::ShaderStage::ReleaseShader(geometryShader);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    InitShader();
    InitObjects();
    Initialize_GD();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}
