﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_CommandBufferJump.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 48260 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/fs.h>
#include <nn/tpl.h>

#include <nn/gr/CTR/gr_Utility.h>

#include "gr_Framework.h"
#include "gr_CommandBufferJump.h"

/* Specifies whether to make a subroutine of those drawing commands in the DrawPrimitive sample that configure lookup table settings.
  */
#define USE_SUBROUTINE_LUT 1

namespace
{
    // Texture package library
    wchar_t* s_TexturePackageName = L"rom:/sample_ctpk.ctpk";
    const int s_TexNum            = 1;

    // BumpTexture
    const u32 s_BumpTextureIndex  = 0;
    const char* s_BumpTextureName = "bump.tga";

    const char* s_TextureNames[ s_TexNum ] =
    {
        s_BumpTextureName
    };

    nn::tpl::CTR::CtrTextureInfo s_TextureInfos[ s_TexNum ];

    u8* s_TextureAddr[ s_TexNum ];

#if USE_SUBROUTINE_LUT
    u32* s_SubCommand;
    size_t s_SubCommandSize;
#endif
}

PicaDataTextureFormat GetPicaDataTextureFormat( const nn::tpl::CTR::CtrTexFormat ctrTextureFormat )
{
    if ( ctrTextureFormat == nn::tpl::CTR::TEXFMT_HL8 )
    {
        return PICA_DATA_TEXTURE_FORMAT_HILO8_DMP_UNSIGNED_BYTE;
    }
    else
    {
        NN_LOG("Not implemented. \n");
        NN_ASSERT( 0 );

        return PICA_DATA_TEXTURE_FORMAT_HILO8_DMP_UNSIGNED_BYTE;
    }
}

//------------------------------------------------------------------------------
GfxAllocator * GrAllocator( MemoryMgr::AllocatorDevice *heap )
{   
    return new GfxAllocator( heap );
}

//------------------------------------------------------------------------------
GraphicSystem::GfxAllocatorFactory  GetGraphicsMemoryMgrFunctor( void )
{
    return &GrAllocator;
}

//------------------------------------------------------------------------------
IScene * CreateScene( void )
{
    return new DrawPrimitiveScene();
}

//------------------------------------------------------------------------------
DrawPrimitiveScene::DrawPrimitiveScene()
    : allocator( NULL ),
      m_Counter( 0 ),
      m_LightType( LIGHT_TYPE_SPOT )
{
}


//------------------------------------------------------------------------------
DrawPrimitiveScene::~DrawPrimitiveScene()
{
}

//------------------------------------------------------------------------------
bool DrawPrimitiveScene::Initialize()
{
    bool is_success;

    NN_UNUSED_VAR( is_success );
    
    allocator = MEMMGR->GetDefaultAllocatorDevice();

    { 
        // Initialize the shader

        // Load shader binary
        nn::fs::FileReader shader_file_reader( L"rom:/shader.shbin" );
        u32 shader_file_size = static_cast<u32>( shader_file_reader.GetSize() );
        shaderBinary = static_cast<u8*>( allocator->Alloc( shader_file_size, 128 ) );
        shader_file_reader.Read( shaderBinary, shader_file_size );
        shader_file_reader.Finalize();

        // Get size of the buffer needed for a ShaderLite class from the shader binary
        u32 size = shader.GetRequiredBufferSize( shaderBinary );

        shaderBuffer = static_cast<u32*>( allocator->Alloc( size ) );

        // Specify shader binary and initialize
        // Enable the executable image at position 0 as the vertex shader
        shader.SetupBinary( shaderBinary, 0, -1, shaderBuffer );
        
#if 1
        // Search for the Uniform using the string
        is_success = shader.SearchBindSymbol( &bindSymbolWorldMtx, "user" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolProjMtx, "proj" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolViewMtx, "view" );
        NN_GR_ASSERT( is_success );
#else
        // Search for the Uniform using the symbol index
        //
        // Symbol indices are allocated by bind_symbol function in order from the top
        // #pragma bind_symbol(proj,    c0, c4)  // 0
        // #pragma bind_symbol(view,    c4, c6)  // 1
        // #pragma bind_symbol(user,    c8, c10) // 2

        is_success = shader.SearchBindSymbol( &bindSymbolProjMtx, static_cast<u8>( 0 ) );
        NN_GR_ASSERT( is_success );
       
        is_success = shader.SearchBindSymbol( &bindSymbolViewMtx, static_cast<u8>( 1 ) );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolWorldMtx, static_cast<u8>( 2 ) );
        NN_GR_ASSERT( is_success );
#endif

    }

    {
        // Initialize primitives
        
        u32 physicalAddr = 0;
        nn::gr::CTR::BindSymbolVSInput symbolPos, symbolTex, symbolClr, symbolNrm;

        {
            // About Box

            m_PrimitiveBox = static_cast< PrimitiveBox* >( allocator->Alloc( sizeof( PrimitiveBox ), 128 ) );
            m_PrimitiveBox->SetupVtxStream();

#if 1
            // Search for location using the string      
            is_success = shader.SearchBindSymbol( &symbolPos, "Position" );
            NN_GR_ASSERT( is_success );

            is_success = shader.SearchBindSymbol( &symbolTex, "TexCoord0" );
            NN_GR_ASSERT( is_success );

            is_success = shader.SearchBindSymbol( &symbolClr, "Color" );
            NN_GR_ASSERT( is_success );

            is_success = shader.SearchBindSymbol( &symbolNrm, "Normal" );
            NN_GR_ASSERT( is_success );
#else
            // Search for "location" using the symbol index
            // 
            // Symbol indices are allocated by bind_symbol function in order from the top
            // #pragma bind_symbol(proj,    c0, c4)  // 0
            // #pragma bind_symbol(view,    c4, c6)  // 1
            // #pragma bind_symbol(user,    c8, c10) // 2
            //
            // #pragma bind_symbol(Position.xyz,   v0,v0) // 3
            // #pragma bind_symbol(TexCoord0.xy,   v1,v1) // 4
            // #pragma bind_symbol(Color.xyzw,     v2,v2) // 5
            // #pragma bind_symbol(Normal,         v3,v3) // 6
            
            is_success = shader.SearchBindSymbol( &symbolPos, static_cast<u8>( 3 ) );
            NN_GR_ASSERT( is_success );

            is_success = shader.SearchBindSymbol( &symbolTex, static_cast<u8>( 4 ) );
            NN_GR_ASSERT( is_success );
            
            is_success = shader.SearchBindSymbol( &symbolClr, static_cast<u8>( 5 ) );
            NN_GR_ASSERT( is_success );

            is_success = shader.SearchBindSymbol( &symbolNrm, static_cast<u8>( 6 ) );
            NN_GR_ASSERT( is_success );
#endif

            // Vertex stream (Position) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_PrimitiveBox->streamPos ) );
            m_VertexBox.EnableAttrAsArray( symbolPos, physicalAddr, PICA_DATA_SIZE_3_FLOAT );


#if 1
            // Vertex stream (TexCoord0) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_PrimitiveBox->streamTex ) );
            m_VertexBox.EnableAttrAsArray( symbolTex, physicalAddr, PICA_DATA_SIZE_2_FLOAT );
#else
            // Fixed vertex attribute (TexCoord0) setting
            f32 texcoord[] = { 0.0f, 0.0f };
            m_VertexBox.EnableAttrAsConst( symbolTex, 2, texcoord );
#endif


#if 1
            // Vertex stream (Color) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_PrimitiveBox->streamClr ) );
            m_VertexBox.EnableAttrAsArray( symbolClr, physicalAddr, PICA_DATA_SIZE_4_FLOAT );                        
#else       
            // Fixed vertex attribute (Color) setting
            f32 color[] = { 1.0f, 0.0f, 0.0f, 0.0f };
            m_VertexBox.EnableAttrAsConst( symbolClr, 4, color );
#endif



#if 1
            // Vertex stream (Normal) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_PrimitiveBox->streamNrm ) );
            m_VertexBox.EnableAttrAsArray( symbolNrm, physicalAddr, PICA_DATA_SIZE_3_FLOAT );
#else
            // Fixed vertex attribute (Normal) setting
            f32 normal[] = { 0.0f, 0.0f, 1.0f };
            m_VertexBox.EnableAttrAsConst( symbolNrm, 3, normal );
#endif


            // Index setting
            m_IndexStreamBox.physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_PrimitiveBox->streamIdx ) );
            m_IndexStreamBox.drawVtxNum = m_PrimitiveBox->VERTEX_NUM;
            m_IndexStreamBox.isUnsignedByte = true; // Index is defined as u8.
        }

        { // About Sphere

            m_PrimitiveInterleavedSphere = static_cast< PrimitiveInterleavedSphere* >( allocator->Alloc( sizeof( PrimitiveInterleavedSphere ) ) );
            m_PrimitiveInterleavedSphere->SetupVtxStream();
            
            // Search for "location"
            is_success = shader.SearchBindSymbol( &symbolPos, "Position" );
            NN_GR_ASSERT( is_success );
            is_success = shader.SearchBindSymbol( &symbolTex, "TexCoord0" );
            NN_GR_ASSERT( is_success );
            is_success = shader.SearchBindSymbol( &symbolClr, "Color" );
            NN_GR_ASSERT( is_success );
            is_success = shader.SearchBindSymbol( &symbolNrm, "Normal" );
            NN_GR_ASSERT( is_success );

            nn::gr::CTR::Vertex::InterleaveInfo interleave_info;
            interleave_info.dataNum     = 4;
            interleave_info.symbol[0]   = &symbolTex;
            interleave_info.dataType[0] = PICA_DATA_SIZE_2_FLOAT;
            interleave_info.symbol[1]   = &symbolNrm;
            interleave_info.dataType[1] = PICA_DATA_SIZE_3_FLOAT;
            interleave_info.symbol[2]   = &symbolClr;
            interleave_info.dataType[2] = PICA_DATA_SIZE_4_FLOAT;
            interleave_info.symbol[3]   = &symbolPos;
            interleave_info.dataType[3] = PICA_DATA_SIZE_3_FLOAT;
            
            // Vertex stream setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_PrimitiveInterleavedSphere->stream ) );
            m_VertexSphere.EnableInterleavedArray( interleave_info, physicalAddr );


#if 0
            // When setting fixed vertex attribute (TexCoord0), write as follows.
            f32 texcoord[] = { 0.0f, 0.0f };
            m_VertexSphere.EnableAttrAsConst( symbolClr, 2, texcoord );
#endif


#if 0
            // When setting fixed vertex attribute (Color), write as follows.
            f32 color[] = { 0.0f, 1.0f, 0.0f, 1.0f };
            m_VertexSphere.EnableAttrAsConst( symbolClr, 4, color );
#endif


#if 0
            // When setting fixed vertex attribute (Normal), write as follows.         
            f32 normal[] = { 0.0f, 0.0f, 1.0f };
            m_VertexSphere.EnableAttrAsConst( symbolNrm, 3, normal );
#endif
            
            // Index setting
            m_IndexStreamSphere.physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_PrimitiveInterleavedSphere->streamIdx ) );
            m_IndexStreamSphere.drawVtxNum = m_PrimitiveInterleavedSphere->VERTEX_NUM;
            m_IndexStreamSphere.isUnsignedByte = false; // Index is defined as u16.            
        }        
    }    

    { 
        // Texture initialization using texture package library

        // Get the size of the texture package file
        nn::fs::FileReader texturePackageFile( s_TexturePackageName );
        size_t texturePackageFileSize = texturePackageFile.GetSize();        
        if ( texturePackageFileSize == 0 )
        {
            NN_LOG( "Failed to open %s.\n", s_TexturePackageName );
            NN_ASSERT( 0 );

            return false;
        }
       
        // Load the texture package file
        u8* texturePackageFileBuffer = (u8*) allocator->Alloc( texturePackageFileSize, 128 );
        if ( texturePackageFileBuffer == NULL )
        {
            NN_LOG( "Allocator returned NULL.\n");
            NN_ASSERT( 0 );

            return false;
        }

        s32 texturePackageFileReadSize = texturePackageFile.Read( texturePackageFileBuffer, texturePackageFileSize );
        if ( texturePackageFileReadSize == 0 )
        {
            NN_LOG( "Failed to open %s.\n", s_TexturePackageName );
            NN_ASSERT( 0 );

            return false;
        }
            
        // Get the texture index
        int textureIndex = nn::tpl::CTR::GetTextureIndex( texturePackageFileBuffer, s_TextureNames[ s_BumpTextureIndex ] );
        if( textureIndex == -1 )
        {
            NN_LOG("Failed to get texture's index.\n");
            texturePackageFile.Finalize();
            allocator->Free( texturePackageFileBuffer );
            NN_ASSERT( 0 );

            return false;
        }

        // Get the data pointer of the texture
        u8* textureDataBuffer = NULL;
        s32 mipLevel = 0;
        u32 mipmapSize = 0;
        textureDataBuffer = (u8*) nn::tpl::CTR::GetTexture( &mipLevel, &mipmapSize, texturePackageFileBuffer, textureIndex );
        if( ! textureDataBuffer )
        {
            NN_LOG("Failed to get texture data pointer.\n");
            texturePackageFile.Finalize();
            allocator->Free( texturePackageFileBuffer );
            NN_ASSERT( 0 );
            
            return false;
        }

        // Get the texture information        
        nn::tpl::CTR::CtrTextureInfo& s_BumpTextureInfo = s_TextureInfos[ s_BumpTextureIndex ];
        bool result = nn::tpl::CTR::GetTextureInfo(&s_BumpTextureInfo, texturePackageFileBuffer, textureIndex);        
        if( ! result )
        {
            NN_LOG("Failed to get texture information.\n");
            texturePackageFile.Finalize();
            allocator->Free( texturePackageFileBuffer );
            NN_ASSERT( 0 );
                 
            return false;
        }
        
        // Copy the texture data
        s_TextureAddr[ s_BumpTextureIndex ] = 
            static_cast<u8*>( allocator->Alloc( s_BumpTextureInfo.texDataSize, 128 ) );        
        std::memcpy( s_TextureAddr[ s_BumpTextureIndex ], textureDataBuffer, s_BumpTextureInfo.texDataSize );        

        // Finalization of the texture package
        texturePackageFile.Finalize();
        allocator->Free( texturePackageFileBuffer );
    }

    { 
        // Light initialization
        m_Light.source[0].diffuseR = 255;
        m_Light.source[0].diffuseG =   0;
        m_Light.source[0].diffuseB =   0;
        m_Light.source[0].ambientR =  16;
        m_Light.source[0].ambientG =   0;
        m_Light.source[0].ambientB =   0;
        m_Light.source[0].isEnableGeomFactor1 = true;
        
        m_Light.source[1].diffuseR =   0;
        m_Light.source[1].diffuseG = 255;
        m_Light.source[1].diffuseB =   0;
        m_Light.source[1].ambientR =   0;
        m_Light.source[1].ambientG =  16;
        m_Light.source[1].ambientB =   0;
        m_Light.source[1].isEnableGeomFactor1 = true;

        m_Light.source[2].diffuseR =   0;
        m_Light.source[2].diffuseG =   0;
        m_Light.source[2].diffuseB = 255;
        m_Light.source[2].ambientR =   0;
        m_Light.source[2].ambientG =   0;
        m_Light.source[2].ambientB =  16;
        m_Light.source[2].isEnableGeomFactor1 = true;
        m_Light.source[2].SetDistAttnStartEnd( 20.f, 40.f );

        // Use texture unit 0 as BUMP
        m_Light.bumpMode = PICA_DATA_FRAG_LIGHT_ENV_BUMP_AS_BUMP_DMP;
        m_Light.isEnableBumpRenorm = true;

        // Layer config 0
        m_Light.layerConfig = PICA_DATA_FRAG_LIGHT_ENV_LAYER_CONFIG0;

        // Enable distribution 0
        m_Light.isEnableLutD0 = true;

        // Set distribution 0 lookup table
        m_Light.lutConfigD0.input = PICA_DATA_FRAG_LIGHT_ENV_NH_DMP;
        m_Light.lutConfigD0.isAbs = true;

        // Set spot lookup table
        m_Light.lutConfigSP.input = PICA_DATA_FRAG_LIGHT_ENV_SP_DMP;
        m_Light.lutConfigSP.isAbs = true;
    }

    {
        // Lookup table (LUT) initialization

        // Save LUT curve in form that can be sent to GPU.
        f32 table[ 256 ];

        for ( int i = 0; i < 256; ++i )
        {
            table[ i ] = ( 256 - i ) / 256.f;
        }

        u32 lastIndex     = 255;
        f32 lastDiffValue = ( 256 - ( lastIndex + 1 )) / 256.f - 
                            ( 256 -   lastIndex     )  / 256.f;
        // Distance attenuation
        m_DistAttnTable.SetTable( table, lastDiffValue );


        for ( int i = 0; i < 256; ++i )
        {
            table[ i ] = i / 256.f;
            table[ i ] *= table[ i ];
            table[ i ] *= table[ i ];
            table[ i ] *= table[ i ];
        }
        
        lastIndex     = 255;
        f32 lastValue = 1.0f;
        lastDiffValue = lastValue - table[ lastIndex ];
        
        // Spot
        m_SpotTable.SetTable( table, lastDiffValue );
        

        for ( int i = 0; i < 256; ++i )
        {
            table[ i ] *= table[ i ];
            table[ i ] *= table[ i ];
            table[ i ] *= table[ i ];
            table[ i ] *= table[ i ];
        }
        lastIndex     = 255;
        lastValue     = 1.0f;
        lastDiffValue = lastValue - table[ lastIndex ];

        // Distribution 0
        m_SpecularTable.SetTable( table, lastDiffValue );
    }

    {
        // Fog initialization
        f32 table[ 128 ];

        for ( int i = 0; i < 128; ++i )
        {
            table[ i ] = ( 128 - i ) / 128.f;
        }

        u32 lastIndex     = 255;
        f32 lastDiffValue = ( 128 - ( lastIndex + 1 ) ) / 128.f - 
                            ( 128 -   lastIndex       ) / 128.f;
        m_Fog.SetTable( table, lastDiffValue );
        
        m_Fog.isEnable = false;
        // m_Fog.isEnable = true;
        m_Fog.isEnableZFlip = false;
        m_Fog.colorR = 255;
        m_Fog.colorG = 255;
        m_Fog.colorB = 255;
    }
       
    // Get allocator and deallocator
    nngxGetAllocator( &m_GraphicsMemoryAllocator, &m_GraphicsMemoryDeallocator );

    {
        // Specify color buffer format       
        PicaDataColor colorBufferFormat = PICA_DATA_COLOR_RGBA8_OES;
        // If you are using the following formats, reconfigure the format for the display buffer.
        //
        // PicaDataColor colorBufferFormat = PICA_DATA_COLOR_RGBA4;
        // PicaDataColor colorBufferFormat = PICA_DATA_COLOR_RGB5_A1;
        // PicaDataColor colorBufferFormat = PICA_DATA_COLOR_RGB565;

        u32 colorBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        if ( colorBufferFormat != PICA_DATA_COLOR_RGBA8_OES )
        {
            colorBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 2;
        }

        // Specify depth buffer format
        PicaDataDepth depthBufferFormat = PICA_DATA_DEPTH24_STENCIL8_EXT;
        // PicaDataDepth depthBufferFormat = PICA_DATA_DEPTH_COMPONENT16;
        // PicaDataDepth depthBufferFormat = PICA_DATA_DEPTH_COMPONENT24_OES;
        
        u32 depthBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        if ( depthBufferFormat == PICA_DATA_DEPTH_COMPONENT16 )
        {
            depthBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 2;
        }
        else if ( depthBufferFormat == PICA_DATA_DEPTH_COMPONENT24_OES )
        {
            depthBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 3;
        }

        // Allocate color buffer
        renderBuffer[ 0 ] = 1;
        GLenum area = NN_GX_MEM_VRAMA;
        GLenum aim  = NN_GX_MEM_RENDERBUFFER;
        colorBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 0 ], colorBufferByteSize ) );
        
        // Allocate depth buffer
        renderBuffer[ 1 ] = 2;
        area = NN_GX_MEM_VRAMB;
        depthBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 1 ], depthBufferByteSize ) );

        renderTarget = 1;

        // Initialize framebuffer settings
        m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

        // Initialize color buffer settings
        m_FrameBuffer.colorBuffer.virtualAddr         = colorBufferVirtualAddress;
        m_FrameBuffer.colorBuffer.format              = colorBufferFormat;
        m_FrameBuffer.colorBuffer.blockSize           = nn::gr::FrameBuffer::BLOCK_SIZE8;
        m_FrameBuffer.colorBuffer.width               = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.colorBuffer.height              = nn::gx::DISPLAY0_HEIGHT;  
        m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;
       
        // Initialize depth buffer settings
        m_FrameBuffer.depthStencilBuffer.virtualAddr  = depthBufferVirtualAddress;
        m_FrameBuffer.depthStencilBuffer.format       = depthBufferFormat;
        m_FrameBuffer.depthStencilBuffer.width        = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.depthStencilBuffer.height       = nn::gx::DISPLAY0_HEIGHT;
        m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
        m_FrameBuffer.depthStencilBuffer.clearStencil = 0;
    }

    m_Scissor.SetBufferSize( nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT );

    // Initialize using a disable command
    {
        u32* command = NULL;
        nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
            reinterpret_cast<GLint*>( &command ) );
        const u32* start_addr = command;

        command = nn::gr::CTR::MakeDisableAllCommand( command );

        nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );
    }
    
    // Create a command buffer to set the lookup table
    CreateCommandBuffer();
    
    return true;
}


//------------------------------------------------------------------------------
void DrawPrimitiveScene::Finalize( void )
{
    m_VertexBox.DisableAll();
    m_VertexSphere.DisableAll();

    allocator->Free( m_PrimitiveBox );
    allocator->Free( m_PrimitiveInterleavedSphere );

    if ( s_TextureAddr[ s_BumpTextureIndex ] != NULL )
    {
        allocator->Free( s_TextureAddr[ s_BumpTextureIndex ] );
        s_TextureAddr[ s_BumpTextureIndex ] = NULL;
    }

    allocator->Free( shaderBinary );
    allocator->Free( shaderBuffer );

    GLenum area = NN_GX_MEM_VRAMA;
    GLenum aim  = NN_GX_MEM_RENDERBUFFER;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 0 ], reinterpret_cast<void*>( colorBufferVirtualAddress ) );
    colorBufferVirtualAddress = 0;

    area = NN_GX_MEM_VRAMB;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 1 ], reinterpret_cast<void*>( depthBufferVirtualAddress ) );
    depthBufferVirtualAddress = 0;
    
    DestroyCommandBuffer();
}

void DrawPrimitiveScene::CreateCommandBuffer(void)
{
#if USE_SUBROUTINE_LUT
    u32* command;

    s_SubCommand = static_cast<u32*>( allocator->Alloc(0x1500, 16) );

    // Load the lookup table
    command = s_SubCommand;
    command = m_SpotTable.MakeSpotTableCommand( command, 0 ); // Use light 0 and this spot table
    command = m_SpotTable.MakeSpotTableCommand( command, 1 ); // Use light 1 and this spot table
    command = m_SpotTable.MakeSpotTableCommand( command, 2 ); // Use light 2 and this spot table
    command = m_DistAttnTable.MakeDistAttnTableCommand( command, 2 ); // Distance attenuation

    // Correct so that the buffer size is a multiple of 16
    // Considering that the kick command to be added later is 8 bytes
    command = nn::gr::AddDummyDataForCommandBuffer( command, (command - s_SubCommand) * sizeof(u32) + 8 );
    // The specified channel must be different than the channel executing this subroutine
    command = nn::gr::MakeChannelKickCommand( command, nn::gr::CTR::CMD_BUF_CH1 );

    s_SubCommandSize = (command - s_SubCommand) * sizeof(u32);

    nngxUpdateBuffer( s_SubCommand, s_SubCommandSize );
#endif
}

void DrawPrimitiveScene::DestroyCommandBuffer(void)
{
#if USE_SUBROUTINE_LUT
    allocator->Free( s_SubCommand );
#endif
}

//------------------------------------------------------------------------------
void DrawPrimitiveScene::Enter( void )
{
    stereo.Initialize();
    nngxSetDisplayMode( NN_GX_DISPLAYMODE_STEREO );
    Calc();
}


//------------------------------------------------------------------------------
bool DrawPrimitiveScene::Calc( void )
{
    nn::math::SinCosFIdx( &m_Sin,   &m_Cos,   m_Counter );
    nn::math::SinCosFIdx( &m_Sin2,  &m_Cos2,  m_Counter * 2 );
    nn::math::SinCosFIdx( &m_Sin01, &m_Cos01, m_Counter * 0.1f );
    
    UpdateScene_();
    ++m_Counter;
    
    m_LightType = ( m_Counter / 1200 ) % LIGHT_TYPE_NUM;
    
    return true;
}

//------------------------------------------------------------------------------
void DrawPrimitiveScene::PreDraw( void )
{
}

//------------------------------------------------------------------------------
GLuint DrawPrimitiveScene::DrawTop( void )
{
    m_Viewport.Set( 0, 0, 240, 400 );
    m_Scissor.Set( 0, 0, 240, 400 );
  
    m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
    m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    { // The camera matrix
        f32 r = 60.f;
        viewMtx.SetupLookAt(
            nn::math::VEC3( r * m_Sin01,  0.0f, r * m_Cos01 ),
            nn::math::VEC3( 0.0f,         1.0f, 0.0f ),
            nn::math::VEC3( 0.0f,         0.0f, 0.0f ) );
    }

    // Projection matrix
    projectionMtx.SetupPerspective(
        0.66f,          // fovy
        400.f / 240.f,  // aspect
        10.f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &viewMtx );
    stereo.SetBaseFrustum( &projectionMtx );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f );

    RenderSceneGr_( projL, viewL );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint DrawPrimitiveScene::DrawExt( void )
{
    m_Viewport.Set( 0, 0, 240, 400 );
    m_Scissor.Set( 0, 0, 240, 400 );

    m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
    m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &viewMtx );
    stereo.SetBaseFrustum( &projectionMtx );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f );

    RenderSceneGr_( projR, viewR );

    return renderTarget;
}


//------------------------------------------------------------------------------
GLuint DrawPrimitiveScene::DrawBtm( void )
{
    m_Viewport.Set( 0, 0, 240, 320 );
    m_Scissor.Set( 0, 0, 240, 320 );

    m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = nn::gx::DISPLAY1_WIDTH;
    m_FrameBuffer.height                          = nn::gx::DISPLAY1_HEIGHT;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    { // The camera matrix

        f32 r = 120.f;

        viewMtx.SetupLookAt(
            nn::math::VEC3( r * m_Sin01, r * m_Sin01, r * m_Cos01 ),
            nn::math::VEC3( 0.0f,        1.0f,        0.0f ),
            nn::math::VEC3( 0.0f,        0.0f,        0.0f ) );
    }

    // Projection matrix
    projectionMtx.SetupPerspective(
        0.66f,          // fovy
        320.f / 240.f,  // aspect
        0.1f,           // near
        20000.f,        // far
        nn::math::PIVOT_UPSIDE_TO_TOP );

#if 1
    // Change render position every 256 frames (Viewport/Scissor test)
    if ( ( m_Counter % 512 ) < 256 )
    {
        u32 c = ( m_Counter % ( 512 * 4 ) ) / 512;
        s32 x = 120 * ( c % 2 );
        s32 y = 160 * ( c / 2 );
        m_Viewport.Set( x, y, 120, 160 );
        m_Scissor.Set( x, y, 120, 160 );
    }
    else
    {
        m_Viewport.Set( 0, 0, 240, 320 );
        m_Scissor.Set( 0, 0, 240, 320 );
    }
#endif

    RenderSceneGr_( projectionMtx, viewMtx );

    return renderTarget;
}

//------------------------------------------------------------------------------
void DrawPrimitiveScene::PostDraw( void )
{
    GLint tmpTopAddr, tmpBufSize;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_TOP_BUFADDR, &tmpTopAddr );
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_USED_BUFSIZE, &tmpBufSize );

    // In this sample, all rendering command requests (Top, Ext, Btm) are generated using the nngxFlush3DCommandPartially function, which does not flush the cache for command buffers, so efficiency is improved by flushing the cache for all of the command buffers in one go.
    //
    //
    nngxUpdateBuffer( reinterpret_cast<void*>(tmpTopAddr), tmpBufSize );
}

//------------------------------------------------------------------------------
nn::gr::CTR::FrameBuffer::ColorBuffer* DrawPrimitiveScene::GetGrColorBufferPointer( void )
{
    return & m_FrameBuffer.colorBuffer;
}

//------------------------------------------------------------------------------
void DrawPrimitiveScene::Clear( const u32 bufferBit )
{
    u32* command = 0;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
        reinterpret_cast<GLint*>( &command ) );

    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Framebuffer settings
    command = m_FrameBuffer.MakeCommand( command, bufferBit );

    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );

    // Add a load complete command at the end of the 3D command buffer
    bool isAddSplitDrawCmdlist = true;

    // Generate a request to clear the color buffer and depth buffer
    m_FrameBuffer.MakeClearRequest( bufferBit, isAddSplitDrawCmdlist );
}

//------------------------------------------------------------------------------
void DrawPrimitiveScene::UpdateScene_( void )
{
    { // Update the model matrix
        
        nn::math::MTX34Identity( &mtx );

        const nn::math::VEC3 t = nn::math::VEC3( m_Cos2, 0.f, m_Sin2 ) * 15.f;
        const nn::math::VEC3 s = nn::math::VEC3( 1.f, 1.f, 1.f ) * 5.f;

        const f32 opt1 = m_Cos2;
        const f32 opt2 = m_Sin2 * m_Sin2;
        const f32 opt3 = 0;

        const f32 tmp00 = s.x * (m_Cos2);
        const f32 tmp10 = 0;
        const f32 tmp20 = s.x * (-m_Sin2);

        const f32 tmp01 = s.y * ((opt2) - (opt3));
        const f32 tmp11 = s.y * ((opt1));
        const f32 tmp21 = s.y * ((m_Sin2 * m_Cos2));

        const f32 tmp02 = s.z * (opt1 * m_Sin2);
        const f32 tmp12 = s.z * ((opt3 * m_Sin2) - (m_Sin2));
        const f32 tmp22 = s.z * ((m_Cos2 * m_Cos2));

        mtx.f._00 = tmp00;
        mtx.f._01 = tmp01;
        mtx.f._02 = tmp02;
        mtx.f._03 = t.x;

        mtx.f._10 = tmp10;
        mtx.f._11 = tmp11;
        mtx.f._12 = tmp12;
        mtx.f._13 = t.y;

        mtx.f._20 = tmp20;
        mtx.f._21 = tmp21;
        mtx.f._22 = tmp22;
        mtx.f._23 = t.z;
    }
}


//------------------------------------------------------------------------------
void DrawPrimitiveScene::RenderSceneGr_(const nn::math::MTX44& proj, const nn::math::MTX34& view )    
{
    bit32* command;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );

    // Specify the command buffer starting address.
    nn::gr::CommandBufferJumpHelper cbjh(command);

    // Viewport settings
    cbjh.Put(m_Viewport.MakeCommand(cbjh.GetCurrentPtr()));

    // Scissor settings
    cbjh.Put(m_Scissor.MakeCommand(cbjh.GetCurrentPtr()));

    // Shader binary settings ( glUseProgram() )
    static bool isFirst = true;
    if ( isFirst )
    {
        cbjh.Put(shader.MakeFullCommand(cbjh.GetCurrentPtr()));
        isFirst = false;
    }
    else
    {
        const bool isMakePrepareCommand = false;
        cbjh.Put(shader.MakeShaderCommand(cbjh.GetCurrentPtr(), isMakePrepareCommand));
    }

    // Render state settings
    // Cull, Depth, Stencil, AlphaTest, Blend, etc.
    nn::gr::RenderState render_state;
    render_state.wBuffer.wScale = m_Fog.isEnable ? 0.005f : 0.f;  // w-buffer settings
    cbjh.Put(render_state.MakeCommand(cbjh.GetCurrentPtr()));

    // Texture settings
    nn::gr::Texture texture;
    u32 tex_index = s_BumpTextureIndex;
    texture.unit0.texType = PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_2D;
    texture.unit0.physicalAddr = nngxGetPhysicalAddr( reinterpret_cast<uptr>( s_TextureAddr[ tex_index ] ) );
    texture.unit0.width  = s_TextureInfos[ tex_index ].width;
    texture.unit0.height = s_TextureInfos[ tex_index ].height;
    texture.unit0.format = GetPicaDataTextureFormat( s_TextureInfos[ tex_index ].texFormat );
    cbjh.Put(texture.MakeCommand(cbjh.GetCurrentPtr()));

    // Texture combiner settings
    nn::gr::Combiner combiner;
    // Add the fragment primary color and secondary color
    combiner.stage[ 0 ].rgb.combine       = PICA_DATA_TEX_ENV_COMBINE_ADD;
    combiner.stage[ 0 ].rgb.source[0]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_PRIMARY_COLOR_DMP;
    combiner.stage[ 0 ].rgb.source[1]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
    cbjh.Put(combiner.MakeCommand(cbjh.GetCurrentPtr()));

    // Load the lookup table
    cbjh.Put(m_SpecularTable.MakeCommand(cbjh.GetCurrentPtr(), PICA_DATA_SAMPLER_D0));

#if USE_SUBROUTINE_LUT
    // Specify the starting address and the size of the buffer for the commands you want to run as subroutines.
    cbjh.Jump(reinterpret_cast<uptr>(s_SubCommand), s_SubCommandSize);
#else
    cbjh.Put(m_SpotTable.MakeSpotTableCommand(cbjh.GetCurrentPtr(), 0)); // Use light 0 and this spot table
    cbjh.Put(m_SpotTable.MakeSpotTableCommand(cbjh.GetCurrentPtr(), 1)); // Use light 1 and this spot table
    cbjh.Put(m_SpotTable.MakeSpotTableCommand(cbjh.GetCurrentPtr(), 2)); // Use light 2 and this spot table
    cbjh.Put(m_DistAttnTable.MakeDistAttnTableCommand(cbjh.GetCurrentPtr(), 2)); // Distance attenuation
#endif

    // Fog settings
    cbjh.Put(m_Fog.MakeAllCommand(cbjh.GetCurrentPtr()));

    // Light settings
    switch ( m_LightType )
    {
    case LIGHT_TYPE_SPOT :
        { // Spotlight settings
            f32 r = 20.f;
            nn::math::VEC3 pos[3] =
            {
                nn::math::VEC3( r*m_Sin, r*m_Cos, 0.f ),
                nn::math::VEC3( r*m_Cos, 0.f, r*m_Sin ),
                nn::math::VEC3( 0.f, r*m_Sin, r*m_Cos ),
            };
            nn::math::VEC3 dir[3] =
            {
                nn::math::VEC3( -m_Sin,  -m_Cos,  0.f ),
                nn::math::VEC3(  -m_Cos, 0.f,  -m_Sin ),
                nn::math::VEC3( 0.f,  -m_Sin,  -m_Cos ),
            };

            // Convert to view coordinates and set the lights
            for ( int i = 0; i < 3; ++i )
            {
                m_Light.isEnable[i]         = true;  // Enable light
                m_Light.isEnableSpot[i]     = true;  // Enable angle attenuation
                m_Light.isEnableDistAttn[i] = false; // Disable distance attenuation
                nn::math::VEC3Transform( &pos[i], &view, &pos[i] );
                m_Light.source[i].SetPosition( pos[i], false );

                nn::math::VEC3TransformNormal( &dir[i], &view, &dir[i] );
                m_Light.source[i].SetSpotDirection( dir[i] );
            }
        }
        break;
    case LIGHT_TYPE_POINT :
        {
            m_Light.isEnable[0]     = false; // disable light 0
            m_Light.isEnable[1]     = false; // disable light 1
            m_Light.isEnable[2]     = true;  // Enable light 2
            m_Light.isEnableSpot[2] = false; // Disable angle attenuation
            m_Light.isEnableDistAttn[2] = true; //Enable distance attenuation

            nn::math::VEC3 pos( mtx.f._03, mtx.f._13, mtx.f._23 );
            nn::math::VEC3Transform( &pos, &view, &pos );
            m_Light.source[2].SetPosition( pos, false ); // false : Do not jump to infinite distance
        }
        break;
    case LIGHT_TYPE_DIRECTIONAL :
        {
            m_Light.isEnable[0] = false;

            m_Light.isEnable[1] = true;
            m_Light.isEnableSpot[1] = false;
            nn::math::VEC3 dir( -1.f, -1.f, -1.f );
            nn::math::VEC3TransformNormal( &dir, &view, &dir );

            // Directional light = point light positioned at infinite distance
            // Set (light direction) x (-1) as the position
            m_Light.source[1].SetPosition( -dir, true ); // true : Jump to infinite distance

            m_Light.isEnable[2] = false;
        }
        break;
    }

    // Issue the light command
    cbjh.Put(m_Light.MakeAllCommand(cbjh.GetCurrentPtr()));

    // Set model matrix
    {
        // Set Shader's Uniform
        cbjh.Put(bindSymbolWorldMtx.MakeUniformCommand( cbjh.GetCurrentPtr(), mtx ));
        cbjh.Put(bindSymbolProjMtx.MakeUniformCommand( cbjh.GetCurrentPtr(), proj ));
        cbjh.Put(bindSymbolViewMtx.MakeUniformCommand( cbjh.GetCurrentPtr(), view ));
    }

    // Switch primitives based on time
    u32 counter = m_Counter % 3000;
    u32 counter_d = counter / 5;

    nn::gr::Vertex*              vertex       = counter < 1500 ? &m_VertexBox      : &m_VertexSphere;
    nn::gr::Vertex::IndexStream* index_stream = counter < 1500 ? &m_IndexStreamBox : &m_IndexStreamSphere;

    // Disable all vertex attributes before starting to use vertices
    cbjh.Put(nn::gr::Vertex::MakeDisableCommand(cbjh.GetCurrentPtr()));

    // Change texture combiner settings
    #if 1
        // Add the fragment primary color and secondary color
        combiner.stage[ 0 ].rgb.combine       = PICA_DATA_TEX_ENV_COMBINE_ADD;
        combiner.stage[ 0 ].rgb.source[0]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_PRIMARY_COLOR_DMP;
        combiner.stage[ 0 ].rgb.source[1]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
        cbjh.Put(combiner.MakeCommand(cbjh.GetCurrentPtr()));
    #else
        // Replace with primary color
        combiner.stage[ 0 ].rgb.combine       = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
        combiner.stage[ 0 ].rgb.source[0]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        combiner.stage[ 0 ].rgb.source[1]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
        cbjh.Put(combiner.MakeCommand(cbjh.GetCurrentPtr()));
    #endif

    // Render the primitive that is rotating in the center

    // glBindBuffer( GL_ARRAY_BUFFER )
    cbjh.Put(vertex->MakeEnableAttrCommand(cbjh.GetCurrentPtr()));
    // glBindBuffer( GL_ELEMENT_ARRAY_BUFFER ) + glDrawElements()
    cbjh.Put(vertex->MakeDrawCommand(cbjh.GetCurrentPtr(), *index_stream));

    // Change texture combiner settings
    #if 1
        // Add the fragment primary color and secondary color
        combiner.stage[ 0 ].rgb.combine       = PICA_DATA_TEX_ENV_COMBINE_ADD;
        combiner.stage[ 0 ].rgb.source[0]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_PRIMARY_COLOR_DMP;
        combiner.stage[ 0 ].rgb.source[1]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
        cbjh.Put(combiner.MakeCommand(cbjh.GetCurrentPtr()));
    #else
        // Replace with primary color
        combiner.stage[ 0 ].rgb.combine       = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
        combiner.stage[ 0 ].rgb.source[0]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        combiner.stage[ 0 ].rgb.source[1]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;    
        cbjh.Put(combiner.MakeCommand(cbjh.GetCurrentPtr()));
    #endif

    // Render other primitives

    // Register is set ahead of time by vertex->MakeEnableAttrCommand(); so use this setting as-is until primitives change.
    //
    nn::math::MTX34 mtx34( mtx );
    for ( int i = 0; i < 300; ++i )
    {
        if ( counter_d == i )
        {
            // Change the primitive to Sphere
            if ( vertex != &m_VertexSphere )
            {
                // Disable the enabled vertex attributes before the primitive is switched to Box
                cbjh.Put(vertex->MakeDisableAttrCommand(cbjh.GetCurrentPtr()));
                vertex       = &m_VertexSphere;
                index_stream = &m_IndexStreamSphere;

                // Generate enabled vertex attribute command
                cbjh.Put(vertex->MakeEnableAttrCommand(cbjh.GetCurrentPtr()));
            }
        }
        else if ( counter_d == 300 + i )
        {
            // Change the primitive to Box
            if ( vertex != &m_VertexBox )
            {
                // Disable the enabled vertex attributes before the primitive is switched to Box
                cbjh.Put(vertex->MakeDisableAttrCommand(cbjh.GetCurrentPtr()));
                vertex       = &m_VertexBox;
                index_stream = &m_IndexStreamBox;

                // Generate enabled vertex attribute command
                cbjh.Put(vertex->MakeEnableAttrCommand(cbjh.GetCurrentPtr()));
            }
        }
        else
        {
            // No change to primitive
        }

        {
            // Render primitive (upper part)
            mtx34.f._03 = ( i % 100 / 10 - 4.5f ) * 10.f;
            mtx34.f._13 = 15.f + ( ( 299 - i ) / 100 ) * 10.f;
            mtx34.f._23 = ( i % 10 -  4.5f ) * 10.f;

            // Set model matrix and render
            cbjh.Put(bindSymbolWorldMtx.MakeUniformCommand( cbjh.GetCurrentPtr(), mtx34 ));
            // glDrawElements()
            cbjh.Put(vertex->MakeDrawCommand( cbjh.GetCurrentPtr(), *index_stream ));
        }

        {
            // Render primitive (lower part)
            mtx34.f._13 = - mtx34.f._13;

            // Set model matrix and render
            cbjh.Put(bindSymbolWorldMtx.MakeUniformCommand( cbjh.GetCurrentPtr(), mtx34 ));
            // glDrawElements()
            cbjh.Put(vertex->MakeDrawCommand( cbjh.GetCurrentPtr(), *index_stream ));
        }
    }

    // Disable all vertex attributes when finished using vertex
    cbjh.Put(nn::gr::Vertex::MakeDisableCommand(cbjh.GetCurrentPtr()));

    // Advance the pointer to the render command buffer
    nngxMoveCommandbufferPointer( cbjh.GetSize() );

#if USE_SUBROUTINE_LUT

    // Add a split command (specify the buffer size up to the first kick command)
    nngxFlush3DCommandPartially(cbjh.GetSizeToFirstJump());

    // Write the command buffer size from the last kick command to here
    bit32* endAddress;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>(&endAddress) );
    cbjh.FinalizeJump(endAddress);

    // The cache for the entire created command region needs to be flushed (because nngxFlush3DCommandPartially does not flush the cache), so in this sample the cache is flushed for the entire command buffer region all at once inside the PostDraw function.
    //
    //
#endif
}

