﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_DisplayBuffer.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_DISPLAY_BUFFER_H_
#define GR_DISPLAY_BUFFER_H_

#include <nn/gr.h>

class IAllocator;

class DisplayBuffer
{
public:
    struct Description
    {
        IAllocator*                 pAllocator;      //
        ScreenKind                  screenKind;      //
        DisplayBufferFormat         format;          //
        NN_PADDING2;

        s32                         width;                 //
        s32                         height;                //

        BufferTransferMode          transferMode;    //
        bool                        isTransferFlipX; //
        NN_PADDING2;

        s32                         transferOffsetX; //
        s32                         transferOffsetY; //
        s32                         displayOffsetX;  //
        s32                         displayOffsetY;  //

        //
        Description()
            : pAllocator( NULL )
            , screenKind( UPPER_SCREEN )
            , format( DISPLAY_BUFFER_RGB8_OES )
            , width( NN_GX_DISPLAY0_HEIGHT )
            , height( NN_GX_DISPLAY0_WIDTH )
            , transferMode( TRANSFER_MODE_ANTIALIASE_NOT_USED )
            , isTransferFlipX( false )
            , transferOffsetX( 0 )
            , transferOffsetY( 0 )
            , displayOffsetX( 0 )
            , displayOffsetY( 0 )
        {}
    };

    //
    class Builder
    {
    public:
        Builder &bufferDescription( const Description &desc_ )
        {
            m_Desc = desc_;

            return *this;
        }

        Builder &pAllocator( IAllocator* pAllocator )
        {
            m_Desc.pAllocator = pAllocator;

            return *this;
        }

        Builder &screenKind( ScreenKind kind )
        {
            m_Desc.screenKind = kind;

            return *this;
        }

        Builder &format( DisplayBufferFormat format )
        {
            m_Desc.format = format;

            return *this;
        }

        Builder &bufferSize( s32 width, s32 height )
        {
            m_Desc.width  = width;
            m_Desc.height = height;

            return *this;
        }

        Builder &transferMode( BufferTransferMode mode )
        {
            m_Desc.transferMode = mode;

            return *this;
        }

        Builder &isTransferFlipX( bool isTransferFlipX )
        {
            m_Desc.isTransferFlipX = isTransferFlipX;

            return *this;
        }

        Builder &transferOffset( s32 x, s32 y )
        {
            m_Desc.transferOffsetX = x;
            m_Desc.transferOffsetY = y;

            return *this;
        }

        Builder &displayOffset( s32 x, s32 y )
        {
            m_Desc.displayOffsetX = x;
            m_Desc.displayOffsetY = y;

            return *this;
        }

        DisplayBuffer* Create( void );

    private:
        Description m_Desc;
    };

    void TransferBuffer( const nn::gr::CTR::FrameBuffer::ColorBuffer* pColorBuffer,
                         const bool isAddSplitDrawCmdlist = true );
    void SwapBuffer( void );

    const Description &GetDescription( void ) const
    {
        return m_Description;
    }

    s32 GetDisplayBufferSize( void ) const
    {
        s32 pixelSize = 3;
       
        DisplayBufferFormat format = m_Description.format;
        if ( format == DISPLAY_BUFFER_RGB8_OES )
        {
            pixelSize = 3;
        }
        else if ( ( format == DISPLAY_BUFFER_RGBA4 ) ||
                  ( format == DISPLAY_BUFFER_RGB5_A1 ) ||
                  ( format == DISPLAY_BUFFER_RGB565 ) )
        {
            pixelSize = 2;
        }
        else
        {
            NN_GR_ASSERT( 0 );
        }

        return pixelSize * m_Description.width * m_Description.height;
    }

    uptr GetDisplayBufferRawVirtualAddress( void ) const
    {
        return m_DisplayBufferVirtualAddressArray[ m_CurrentDisplay ];
    }

    uptr GetDisplayBufferVirtualAddress( void ) const
    {
        s32 pixelSize = 3;

        GLenum format = GetFormat();
        if ( format == GL_RGB8_OES )
        {
            pixelSize = 3;
        }
        else if ( ( format == GL_RGBA4 ) ||
                  ( format == GL_RGB5_A1 ) ||
                  ( format == GL_RGB565 ) )
        {
            pixelSize = 2;
        }
        else
        {
            NN_GR_ASSERT( 0 );
        }

        s32 displayModeHeight = 400;
        if ( ( m_Display == NN_GX_DISPLAY0 ) || 
             ( m_Display == NN_GX_DISPLAY0_EXT ) )
        {
            displayModeHeight = 400;
        }
        else if ( m_Display == NN_GX_DISPLAY1 )
        {
            displayModeHeight = 320;
        }
        else
        {
            NN_GR_ASSERT( 0 );
        }
        
        // Equivalent to nngxDisplaybufferStorage( format, m_Description.height, m_Description.width, area )
        s32 width            = m_Description.height;
        s32 height           = m_Description.width;

        // Equivalent to nngxDisplayEnv( m_Description.displayOffsetY, m_Description.displayOffsetX )
        s32 offsetX          = m_Description.displayOffsetY;
        s32 offsetY          = m_Description.displayOffsetX;

        uptr virtualAddress  = 
            GetDisplayBufferRawVirtualAddress();

        // Add an offset to the display buffer's virtual address
        virtualAddress += 
            pixelSize * ( width * ( height - displayModeHeight - offsetY ) + offsetX );

        return virtualAddress;
    }

    GLenum GetFormat( void ) const
    {
        return m_Format;
    }

private:
    DisplayBuffer( const Description &desc );

protected:
    ~DisplayBuffer();

private:
    Description     m_Description;
    u32             m_CurrentDisplay;
    uptr            m_DisplayBufferVirtualAddressArray[ 2 ];
    GLenum          m_Display;
    GLenum          m_TransferMode;
    GLenum          m_Format;
};



#endif // GR_DISPLAY_BUFFER_H_
