﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_DisplayBuffer.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include "gr_Inner.h"
#include "gr_Common.h"
#include "gr_DisplayBuffer.h"
#include "gr_MemoryMgr.h"

//------------------------------------------------------------------------------
DisplayBuffer *DisplayBuffer::Builder::Create( void )
{
    IAllocator* pAllocator = m_Desc.pAllocator;
    NN_GR_ASSERT( pAllocator != NULL );

    DisplayBuffer* pDisplayBuffer =
        reinterpret_cast<DisplayBuffer*>( pAllocator->Alloc( sizeof( DisplayBuffer ) ) );
    new( pDisplayBuffer ) DisplayBuffer( m_Desc );

    return pDisplayBuffer;
}

//------------------------------------------------------------------------------
DisplayBuffer::DisplayBuffer( const Description &desc )
    : m_Description( desc )
    , m_CurrentDisplay( 0 )
    , m_Display( NN_GX_DISPLAY0 )
    , m_TransferMode( NN_GX_ANTIALIASE_NOT_USED )
    , m_Format( GL_RGB8_OES )
{
    for ( u32 index = 0; index < 2; index++ )
    {
        m_DisplayBufferVirtualAddressArray[ index ] = NULL;
    }

    {
        const s32 ALIGNMENT_DISPLAY_BUFFER = 16;
        s32 displayBufferSize = GetDisplayBufferSize();

        IAllocator* pAllocator = m_Description.pAllocator;
        NN_GR_ASSERT( pAllocator != NULL );

        uptr pDisplayBuffer0VirtualAddress =
            reinterpret_cast<uptr>( pAllocator->Alloc( displayBufferSize, ALIGNMENT_DISPLAY_BUFFER ) );

        uptr pDisplayBuffer1VirtualAddress =
            reinterpret_cast<uptr>( pAllocator->Alloc( displayBufferSize, ALIGNMENT_DISPLAY_BUFFER ) );

        NN_GR_ASSERT( pDisplayBuffer0VirtualAddress );
        NN_GR_ASSERT( pDisplayBuffer1VirtualAddress );

        m_DisplayBufferVirtualAddressArray[ 0 ] = pDisplayBuffer0VirtualAddress;
        m_DisplayBufferVirtualAddressArray[ 1 ] = pDisplayBuffer1VirtualAddress;
    }

    if( m_Description.screenKind == UPPER_SCREEN )
    {
        m_Display = NN_GX_DISPLAY0;
    }
    else if( m_Description.screenKind == LOWER_SCREEN )
    {
        m_Display = NN_GX_DISPLAY1;
    }
    else if( m_Description.screenKind == EXTENSION_SCREEN )
    {
        m_Display = NN_GX_DISPLAY0_EXT;
    }
    else
    {
        NN_ASSERT( 0 );
    }

    if( m_Description.transferMode == TRANSFER_MODE_ANTIALIASE_NOT_USED )
    {
        m_TransferMode = NN_GX_ANTIALIASE_NOT_USED;
    }
    else if( m_Description.transferMode == TRANSFER_MODE_ANTIALIASE_2x1 )
    {
        m_TransferMode = NN_GX_ANTIALIASE_2x1;
    }
    else if( m_Description.transferMode == TRANSFER_MODE_ANTIALIASE_2x2 )
    {
        m_TransferMode = NN_GX_ANTIALIASE_2x2;
    }
    else
    {
        NN_ASSERT( 0 );
    }

    if ( m_Description.format == DISPLAY_BUFFER_RGB8_OES )
    {
        m_Format = GL_RGB8_OES;
    }
    else if ( m_Description.format == DISPLAY_BUFFER_RGBA4 )
    {
        m_Format = GL_RGBA4;
    }
    else if ( m_Description.format == DISPLAY_BUFFER_RGB5_A1 )
    {
        m_Format = GL_RGB5_A1;
    }
    else if ( m_Description.format == DISPLAY_BUFFER_RGB565 )
    {
        m_Format = GL_RGB565;
    }
    else
    {
        NN_ASSERT( 0 );
    }
    
    m_CurrentDisplay = 0;
}

//------------------------------------------------------------------------------

DisplayBuffer::~DisplayBuffer( )
{
    for ( u32 index = 0; index < 2; index++ )
    {
        uptr pDisplayBuffer = m_DisplayBufferVirtualAddressArray[ index ];
        if ( pDisplayBuffer != NULL )
        {
            m_Description.pAllocator->Free( reinterpret_cast<GLvoid*>( pDisplayBuffer ) );
        }
        m_DisplayBufferVirtualAddressArray[ index ] = NULL;
    }
}

//------------------------------------------------------------------------------
void DisplayBuffer::TransferBuffer( const nn::gr::CTR::FrameBuffer::ColorBuffer* pColorBuffer,
                                    const bool isAddSplitDrawCmdlist )
{
    NN_GR_ASSERT( pColorBuffer != NULL );

    // Get the virtual address of the display buffer to transfer data to
    uptr displayBufferVirtualAddress = GetDisplayBufferVirtualAddress();

    GLenum colorBufferFormat = GL_RGBA8_OES;
    if ( pColorBuffer->format == PICA_DATA_COLOR_RGB5_A1 )
    {
        colorBufferFormat = GL_RGB5_A1;
    }
    else if ( pColorBuffer->format == PICA_DATA_COLOR_RGB565 )
    {
        colorBufferFormat = GL_RGB565;
    }
    else if ( pColorBuffer->format == PICA_DATA_COLOR_RGBA4 )
    {
        colorBufferFormat = GL_RGBA4;
    }

    GLsizei blockSize = 8;
    if ( pColorBuffer->blockSize == nn::gr::CTR::FrameBuffer::BLOCK_SIZE32 )
    {
        blockSize = 32;
    }

    if ( isAddSplitDrawCmdlist )
    {
        nngxFlush3DCommand();
    }

    nngxAddB2LTransferCommand(
       reinterpret_cast<GLvoid*>( pColorBuffer->virtualAddr ),
       pColorBuffer->width,
       pColorBuffer->height,
       colorBufferFormat,
       reinterpret_cast<GLvoid*>( displayBufferVirtualAddress ),
       m_Description.height,
       m_Description.width,
       m_Format,
       m_TransferMode,
       m_Description.isTransferFlipX,
       blockSize );
}

//------------------------------------------------------------------------------

void DisplayBuffer::SwapBuffer( void )
{
    m_CurrentDisplay = 1 - m_CurrentDisplay;
}
