﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_GfxAllocator.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include "gr_Inner.h"
#include "gr_Common.h"
#include "gr_MemoryMgr.h"
#include "gr_GfxAllocator.h"

namespace
{

    //------------------------------------------------------------------------------
    const s32 ALIGNMENT_SYSTEM_BUFFER     = 4;
    const s32 ALIGNMENT_VERTEX            = 4;
    const s32 ALIGNMENT_TEXTURE           = 128;
    const s32 ALIGNMENT_RENDER_BUFFER     = 64;    // 24-bit format (RGB8, D24) buffer must be aligned to 96 bytes
    const s32 ALIGNMENT_DISPLAY_BUFFER    = 16;
    const s32 ALIGNMENT_3D_COMMAND_BUFFER = 16;

    //------------------------------------------------------------------------------
    uptr MathRoundup( uptr x, s32 base )
    {
        return ( ( x ) + ( ( base ) - 1 )) & ~( ( base ) - 1 );
    }

}

//------------------------------------------------------------------------------
void GfxAllocator::Initialize( void )
{
    m_CurrentAddrVramA = nn::gx::GetVramStartAddr( nn::gx::MEM_VRAMA );
    m_CurrentAddrVramB = nn::gx::GetVramStartAddr( nn::gx::MEM_VRAMB );
      
    m_Initialized = true;
}


//------------------------------------------------------------------------------
void GfxAllocator::Finalize( void )
{
}

//------------------------------------------------------------------------------
void *GfxAllocator::Allocate( GLenum area, GLenum aim, GLuint id, GLsizei size )
{
    NN_UNUSED_VAR( id );
    
    if( size == 0 )
    {
        return NULL;
    }
    
    int     addrAlign   = 8;
    void *  resultAddr  = NULL;

    // Take into account the alignment restrictions on each data set's position
    switch( aim )
    {
        case NN_GX_MEM_SYSTEM:          
            addrAlign = ALIGNMENT_SYSTEM_BUFFER;
            break;
        case NN_GX_MEM_TEXTURE:         
            addrAlign = ALIGNMENT_TEXTURE;
            break;
        case NN_GX_MEM_VERTEXBUFFER:
            addrAlign = ALIGNMENT_VERTEX;
            break;
        case NN_GX_MEM_RENDERBUFFER:
            addrAlign = ALIGNMENT_RENDER_BUFFER;
            break;
        case NN_GX_MEM_DISPLAYBUFFER:
            addrAlign = ALIGNMENT_DISPLAY_BUFFER;
            break;
        case NN_GX_MEM_COMMANDBUFFER:
            addrAlign = ALIGNMENT_3D_COMMAND_BUFFER;
            break;
        default:
            NN_ASSERTMSG( 0,"invalid parameter. (0x%X)\n", aim );
            break;
    }
            
    switch( area )
    {
        case NN_GX_MEM_FCRAM:
            // When using FCRAM, no need to manage this because it is allocated from the expanded heap that is supported by the SDK
            resultAddr = m_AllocatorDev->Alloc( size, addrAlign );
            break;

        case NN_GX_MEM_VRAMA:
            {
                if( MathRoundup( m_CurrentAddrVramA, addrAlign ) + size > nn::gx::GetVramEndAddr( nn::gx::MEM_VRAMA ) )
                {
                    NN_ASSERTMSG( 0,"Lack of resources in VRAM-A.\n" );
                }

                m_CurrentAddrVramA = MathRoundup( m_CurrentAddrVramA, addrAlign );
                resultAddr = reinterpret_cast<void*>( m_CurrentAddrVramA );
                m_CurrentAddrVramA += size;
            }
            break;

        case NN_GX_MEM_VRAMB:
            {
                if( MathRoundup( m_CurrentAddrVramB, addrAlign ) + size > nn::gx::GetVramEndAddr( nn::gx::MEM_VRAMB ) )
                {
                    NN_ASSERTMSG( 0,"Lack of resources in VRAM-B.\n" );
                }

                m_CurrentAddrVramB = MathRoundup( m_CurrentAddrVramB, addrAlign );
                resultAddr = reinterpret_cast<void*>( m_CurrentAddrVramB );                    
                m_CurrentAddrVramB += size;
            }
            break;

        default:
            NN_ASSERTMSG( 0, "invalid parameter. (0x%X)\n", area );
            break;
    }
    
    NN_ASSERT( resultAddr );
    
    return resultAddr;
}

//------------------------------------------------------------------------------
void GfxAllocator::Deallocate( GLenum area, GLenum aim, GLuint id, void* addr )
{
    NN_UNUSED_VAR( aim );
    NN_UNUSED_VAR( id );
        
    switch( area )
    {
    case NN_GX_MEM_FCRAM:
        m_AllocatorDev->Free( addr );
        break;
    }
}
