﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_GraphicSystem.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include "gr_Inner.h"
#include "gr_Common.h"
#include "gr_MemoryMgr.h"
#include "gr_GfxAllocator.h"
#include "gr_DisplayBuffer.h"
#include "gr_GraphicSystem.h"

//------------------------------------------------------------------------------
GfxAllocator * GraphicSystem::s_GfxAllocator    = NULL;
bool           GraphicSystem::s_GetGlSystemFlag = false;



//------------------------------------------------------------------------------
void *GraphicSystem::AllocateGraphicsMemory( GLenum area, GLenum aim, GLuint id, GLsizei size )
{
    void * buffer = s_GfxAllocator->Allocate( area, aim, id, size );
    return buffer;
}

//------------------------------------------------------------------------------
void GraphicSystem::DeallocateGraphicsMemory( GLenum area, GLenum aim, GLuint id, void* addr )
{
    s_GfxAllocator->Deallocate( area, aim, id, addr );
}

//------------------------------------------------------------------------------
void GraphicSystem::Create( size_t commandBufferSize, u32 reqCount )
{
    s_GetGlSystemFlag = true;
    if( nngxInitialize( AllocateGraphicsMemory, DeallocateGraphicsMemory ) == GL_FALSE )
    {
        NN_ASSERTMSG(0,"nngxInitialize failed.\n");
    }
    s_GetGlSystemFlag = false;

    // Command list
    nngxGenCmdlists( 2, &m_CommandList[0] );
    nngxBindCmdlist( m_CommandList[0] );
    nngxCmdlistStorage( commandBufferSize, reqCount );
    nngxSetCmdlistParameteri( NN_GX_CMDLIST_RUN_MODE, NN_GX_CMDLIST_SERIAL_RUN );
    nngxClearCmdlist();

    nngxBindCmdlist( m_CommandList[1] );
    nngxCmdlistStorage( commandBufferSize, reqCount );
    nngxSetCmdlistParameteri( NN_GX_CMDLIST_RUN_MODE, NN_GX_CMDLIST_SERIAL_RUN );
    nngxClearCmdlist();

    m_CommandIndex = 0;

    nngxBindCmdlist( m_CommandList[m_CommandIndex] );
}


//------------------------------------------------------------------------------
void GraphicSystem::Destroy()
{   
    nngxFinalize();
}


//------------------------------------------------------------------------------
void GraphicSystem::SetDisplayBuffer( DisplayBuffer *top, DisplayBuffer *bottom, DisplayBuffer *extension )
{
    m_UpperSwapper     = top;
    m_LowerSwapper     = bottom;
    m_ExtensionSwapper = extension;
}

//------------------------------------------------------------------------------
void GraphicSystem::WaitVSync( s32 screenKind )
{
    static bool isFirst = true;
    
    switch( screenKind )
    {
        case EXTENSION_SCREEN | UPPER_SCREEN | LOWER_SCREEN:
        case EXTENSION_SCREEN | LOWER_SCREEN:
        case UPPER_SCREEN | LOWER_SCREEN:
            nngxWaitVSync(NN_GX_DISPLAY_BOTH);
            break;

        case EXTENSION_SCREEN | UPPER_SCREEN:
        case UPPER_SCREEN:
        case EXTENSION_SCREEN:
            nngxWaitVSync(NN_GX_DISPLAY0);
            break;

        case LOWER_SCREEN:
            nngxWaitVSync(NN_GX_DISPLAY1);
            break;
    }
    
    // Calling this function = Assuming the first image is in a ready state and turning on the LCD
    if ( isFirst )
    {
        nngxStartLcdDisplay();
        isFirst = false;
    }
}

//------------------------------------------------------------------------------
void GraphicSystem::TransferBuffer( const s32 screenKind,
                                    const nn::gr::CTR::FrameBuffer::ColorBuffer* pColorBuffer,
                                    const bool isAddSplitDrawCmdlist )
{
    NN_GR_ASSERT( pColorBuffer != NULL );

    switch( screenKind )
    {
        case EXTENSION_SCREEN | UPPER_SCREEN | LOWER_SCREEN:
            m_UpperSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            m_LowerSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            m_ExtensionSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            break;

        case EXTENSION_SCREEN | UPPER_SCREEN:
            m_UpperSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            m_ExtensionSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            break;

        case EXTENSION_SCREEN | LOWER_SCREEN:
            m_LowerSwapper    ->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            m_ExtensionSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            break;

        case EXTENSION_SCREEN:
            m_ExtensionSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            break;

        case UPPER_SCREEN | LOWER_SCREEN:
            m_UpperSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            m_LowerSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            break;

        case UPPER_SCREEN:
            m_UpperSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            break;

        case LOWER_SCREEN:
            m_LowerSwapper->TransferBuffer( pColorBuffer, isAddSplitDrawCmdlist );
            break;
    }
}

//------------------------------------------------------------------------------
void GraphicSystem::RunCommandList()
{
    if( m_DoubleCommandList )
    {
        nngxBindCmdlist( m_CommandList[m_CommandIndex] );
        nngxRunCmdlist();
    }
    else
    {
        nngxRunCmdlist();
    }
}

//------------------------------------------------------------------------------
void GraphicSystem::StopCommandList()
{
    nngxStopCmdlist();
}

//------------------------------------------------------------------------------
void GraphicSystem::SwapBuffer( s32 screenKind )
{
    uptr display0VirtualAddress          = NULL;
    uptr display0ExtensionVirtualAddress = NULL;
    uptr display1VirtualAddress          = NULL;

    // 240
    s32 display0Width                    = NN_GX_DISPLAY0_WIDTH;
    // 240
    s32 display1Width                    = NN_GX_DISPLAY1_WIDTH;

    GLenum display0Format                = GL_RGB8_OES;
    GLenum display1Format                = GL_RGB8_OES;

    if ( screenKind & EXTENSION_SCREEN )
    {
      if (m_DoubleCommandList)
      {
          m_ExtensionSwapper->SwapBuffer();
      }

        display0ExtensionVirtualAddress = m_ExtensionSwapper->GetDisplayBufferVirtualAddress();
        display0Width                   = m_ExtensionSwapper->GetDescription().height;
        display0Format                  = m_ExtensionSwapper->GetFormat();

      if (!m_DoubleCommandList)
      {
          m_ExtensionSwapper->SwapBuffer();
      }
    }


    if ( screenKind & UPPER_SCREEN )
    {
      if (m_DoubleCommandList)
      {
          m_UpperSwapper->SwapBuffer();
      }
       
      display0VirtualAddress = m_UpperSwapper->GetDisplayBufferVirtualAddress();
      display0Width          = m_UpperSwapper->GetDescription().height;
      display0Format         = m_UpperSwapper->GetFormat();

      if (!m_DoubleCommandList)
      {
          m_UpperSwapper->SwapBuffer();
      }
    }

    if ( screenKind & LOWER_SCREEN )
    {
        if (m_DoubleCommandList)
        {
            m_LowerSwapper->SwapBuffer();
        }

        display1VirtualAddress = m_LowerSwapper->GetDisplayBufferVirtualAddress();
        display1Width          = m_LowerSwapper->GetDescription().height;
        display1Format         = m_LowerSwapper->GetFormat();

        if (!m_DoubleCommandList)
        {
            m_LowerSwapper->SwapBuffer();
        }
    }


    if ( ( screenKind & UPPER_SCREEN ) || 
         ( screenKind & EXTENSION_SCREEN ) )
    {
        if ( screenKind & UPPER_SCREEN )
        {
            NN_GR_ASSERT( display0VirtualAddress != NULL );
        }

        if ( screenKind & EXTENSION_SCREEN )
        {
            NN_GR_ASSERT( display0ExtensionVirtualAddress != NULL );
        }

        nngxSwapBuffersByAddress( NN_GX_DISPLAY0, 
                reinterpret_cast<GLvoid*>( display0VirtualAddress ),
                reinterpret_cast<GLvoid*>( display0ExtensionVirtualAddress ),
                display0Width,
                display0Format );
    }

    if ( screenKind & LOWER_SCREEN )
    {
        NN_GR_ASSERT( display1VirtualAddress != NULL );

        nngxSwapBuffersByAddress( NN_GX_DISPLAY1,
                reinterpret_cast<GLvoid*>( display1VirtualAddress ),
                NULL,
                display1Width,
                display1Format );
    }
}


//------------------------------------------------------------------------------
void GraphicSystem::WaitCommandList( const bool isAddSplitDrawCmdlist )
{
    if( m_DoubleCommandList )
    {
        if( isAddSplitDrawCmdlist )
        {
            nngxSplitDrawCmdlist();
        }
        nngxBindCmdlist( m_CommandList[1 - m_CommandIndex] );
        nngxWaitCmdlistDone();
        nngxClearCmdlist();
    }
    else
    {
        if( isAddSplitDrawCmdlist )
        {
            nngxSplitDrawCmdlist();
        }
        nngxWaitCmdlistDone();
        nngxClearCmdlist();
    }
}

//------------------------------------------------------------------------------
void GraphicSystem::SwapCommandList()
{
    if( m_DoubleCommandList )
    {
        m_CommandIndex = 1 - m_CommandIndex;
        nngxBindCmdlist( m_CommandList[m_CommandIndex] );
    }
}

//------------------------------------------------------------------------------
void GraphicSystem::Initialize( GfxAllocatorFactory factory )
{
    // FS initialization
    nn::fs::Initialize();

    nn::hid::Initialize();

    // Initialize memory manager
    MEMMGR->Initialize();

    s32 workSize = nn::fs::GetRomRequiredMemorySize( 1, 1 );
    u8  *work    = static_cast<u8*>( MEMMGR->AllocateFCRAM( workSize, 128 ) );
    nn::fs::MountRom( 1, 1, work, workSize );
    
    // Initializes the graphics memory allocator that uses device memory.
    s_GetGlSystemFlag = false;
    s_GfxAllocator    = factory ? factory( MEMMGR->GetDefaultAllocatorDevice()) : new GfxAllocator( MEMMGR->GetDefaultAllocatorDevice() );
    s_GfxAllocator->Initialize();
}

//------------------------------------------------------------------------------
void GraphicSystem::Finalize( void )
{
    s_GfxAllocator->Finalize();
}

