/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_MemoryMgr.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include "gr_Inner.h"
#include "gr_Common.h"
#include "gr_MemoryMgr.h"

//------------------------------------------------------------------------------
SMP_SINGLETON_IMPL( MemoryMgr );

namespace
{

    template <typename T>
    inline void*
    AddOffsetToPtr( void* ptr, T offset )
    {
        return reinterpret_cast<void*>( reinterpret_cast<u32>( ptr ) + offset );
    }

    inline size_t
    GetOffsetFromPtr( const void* start, const void* end )
    {
        return static_cast<size_t>(reinterpret_cast<u32>( end ) - reinterpret_cast<u32>( start ));
    }

    template <typename ValueT>
    inline void*
    RoundUp(ValueT* x, unsigned int base)
    {
        u32 value = reinterpret_cast<u32>( x );
        u32 rounded = ( value + ( base - 1 )) & ~( base - 1 );
        
        return reinterpret_cast<void*>( rounded );
    }

    //------------------------------------------------------------------------------
    bool IsDeviceMemory(const void* memory)
    {
        static u32 start    = nn::os::GetDeviceMemoryAddress();
        static u32 end      = nn::os::GetDeviceMemoryAddress() + nn::os::GetDeviceMemorySize();
        return (start <= (u32)memory) && ((u32)memory <= end);
    }

    //------------------------------------------------------------------------------
    void *AlignMemory( void* memory, u8 align )
    {
        void* top = AddOffsetToPtr( memory, 1 );
        top = RoundUp( top, align );
        
        u8* head = reinterpret_cast<u8*>( AddOffsetToPtr( top, - 1 ) );
        
        size_t offset = GetOffsetFromPtr( memory, top );
        
        *head = static_cast<u8>( offset );
        return top;
    }

    //------------------------------------------------------------------------------
    void *UnAlignMemory( void* memory )
    {
        u8* head = reinterpret_cast<u8*>( AddOffsetToPtr( memory, -1 ) );
        
        u8 offset = *head;
        
        return AddOffsetToPtr( memory, - offset );
    }

}


//------------------------------------------------------------------------------
bool MemoryMgr::Initialize( size_t mainMemorySize )
{
    if( ! mainMemorySize )
    {
        return false;
    }

    m_MemoryBlock.Initialize( mainMemorySize );
        
    size_t  size;
    uptr    address;
        
    address = m_MemoryBlock.GetAddress();
    size    = m_MemoryBlock.GetSize();
    m_MainHeap.Initialize( address, size );

    address = nn::os::GetDeviceMemoryAddress();
    size    = nn::os::GetDeviceMemorySize();
    m_DeviceHeap.Initialize( address, size );

    // LOG output
    {
        NN_LOG( "Initialize Memory\n" );

        address = m_MemoryBlock.GetAddress();
        size    = m_MemoryBlock.GetSize();
        NN_LOG( "\tMain   : 0x%08X - 0x%08X\n", static_cast<u32>( address ), ( static_cast<u32>( address ) + size ) );

        address = nn::os::GetDeviceMemoryAddress();
        size    = nn::os::GetDeviceMemorySize();
        NN_LOG( "\tDevice : 0x%08X - 0x%08X\n", static_cast<u32>( address ), ( static_cast<u32>( address ) + size ) );
    }        

    return true;
}


//------------------------------------------------------------------------------
void *MemoryMgr::AllocateFCRAM( size_t size, u8 alignment )
{
    // Check for power of 2
    NN_ASSERT( ( alignment & ( alignment - 1 )) == 0 );
        
    if( alignment == 0 ) 
    {
        alignment = 1;
    }
        
    void *memory = AllocateFCRAM( size + alignment );
    return AlignMemory( memory, alignment );
}


//------------------------------------------------------------------------------
void *MemoryMgr::AllocateDevice( size_t size, u8 alignment )
{
    // Check for power of 2
    NN_ASSERT( ( alignment & ( alignment - 1 )) == 0 );
        
    if( alignment == 0 ) 
    {
        alignment = 1;
    }
        
    void *memory = AllocateDevice( size + alignment );
    return AlignMemory( memory, alignment );
}

//------------------------------------------------------------------------------
void MemoryMgr::Free( void *memory )
{
    memory = UnAlignMemory( memory );

    if( IsDeviceMemory( memory ) )
    {
        m_DeviceHeap.Free( memory );
    }
    else
    {
        m_MainHeap.Free( memory );
    }
}
