﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_GraphicsDrawing.h

  Copyright (C)2009-2012 Nintendo Co., Ltd./HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_DEMO_GRAPHICS_DRAWING_H_
#define GR_DEMO_GRAPHICS_DRAWING_H_

#include <nn/util.h>
#include <nn/math.h>
#include <nn/gx/CTR/gx_CommandAccess.h>

#include <nn/gr.h>
#include <nn/font.h>

class IAllocator;

class GraphicsDrawing
{
public:
    GraphicsDrawing( void );
    virtual ~GraphicsDrawing();
    
    void SetScreenSize( s32 width, s32 height ) 
    { 
        m_ScreenSize = nn::math::VEC2( static_cast<f32>(width), static_cast<f32>(height) ); 
    }

    nn::gr::CTR::ShaderLite* GetGrShaderPointer( void )
    {
        NN_NULL_ASSERT( m_pShapeShaderBinary );

        return &m_ShapeShader;
    }

    void Finalize( void );

public:
    bool InitializeShape( IAllocator* pShapeAllocator, const wchar_t* shaderPath );
    bool InitializeShape( void* pShaderBinary, size_t binarySize );

    void DestroyShape( void );

    void SetupShape( void );

    void SendShapeMatrix( void );

    void SetMaxShapeVertexCount( s32 maxShapeVertexCount ) 
    {
        m_MaxShapeVertexCount = maxShapeVertexCount;
    }

    void SetDrawMode( const PicaDataDrawMode drawMode )
    {
        m_ShapeShader.SetDrawMode( drawMode );
    }
    
    void SetLineWidth( const f32 width )
    { 
        m_LineWidth = width; 
    }
    
    void BeginDrawingShape( void );
    void EndDrawingShape( void );

    void DrawLine( const s32 x1, const s32 y1, const s32 x2, const s32 y2, const nn::util::FloatColor color )
    {
        this->DrawLine( 
            nn::math::VEC2( static_cast<f32>( x1 ), static_cast<f32>( y1 ) ), 
            nn::math::VEC2( static_cast<f32>( x2 ), static_cast<f32>( y2 ) ), 
            color 
            );
    }
        
    void DrawLine( const nn::math::VEC2& p1, const nn::math::VEC2& p2, const nn::util::FloatColor color );
    void DrawLine( const nn::math::VEC2& p1, const nn::math::VEC2& p2,
        const nn::util::FloatColor color1, const nn::util::FloatColor color2 );
    
    void DrawRectangle( const s32 posh, const s32 posv, const s32 sizeh, const s32 sizev,
                        const nn::util::FloatColor color )
    {
        this->DrawRectangle( 
            nn::math::VEC2(static_cast<f32>( posh ), static_cast<f32>( posv ) ), 
            nn::math::VEC2(static_cast<f32>( sizeh ), static_cast<f32>( sizev ) ), 
            color 
            );
    }
        
    void DrawRectangle( const nn::math::VEC2& pos, const nn::math::VEC2& size, const nn::util::FloatColor color );
    
    void FillRectangle( const s32 posh, const s32 posv, 
                        const s32 sizeh, const s32 sizev, const nn::util::FloatColor color )
    {
        this->FillRectangle( 
            nn::math::VEC2(static_cast<f32>(posh), static_cast<f32>(posv)), 
            nn::math::VEC2(static_cast<f32>(sizeh), static_cast<f32>(sizev)), 
            color 
            );
    }
        
    void FillRectangle( const nn::math::VEC2& pos, const nn::math::VEC2& size, const nn::util::FloatColor color );

    void FillTriangle( const nn::math::VEC2& pt1, const nn::math::VEC2& pt2, const nn::math::VEC2& pt3, nn::util::FloatColor color );
    
    void FillTriangles( const nn::math::VEC2* pPointArray, const s32 pointCount, const nn::util::FloatColor color );
    
    void SetMaxTextCount( const s32 maxTextCount )
    {
        m_MaxTextCount = maxTextCount;
    }

public:
    bool InitializeFont( IAllocator* pFontAllocator, const wchar_t* fontShaderPath, const wchar_t* fontPath );
    void DestroyFont( void );

    void BeginDrawingString( void );
    void EndDrawingString( void );

    f32 DrawString( const char* format, ... );
    f32 DrawString( s32 posh, s32 posv, const char* format, ... );
    f32 DrawStringArgs( s32 posh, s32 posv, const char* format, std::va_list args );    

    nn::font::TextWriter& GetWriter( void )
    { 
        return m_TextWriter; 
    }
    const nn::font::TextWriter& GetWriter( void ) const
    {
        return m_TextWriter;
    }

    bool InitializeFont( void* shaderBinary, size_t shaderSize, void* fontBinary, size_t fontSize, void* drawBuffer, void* stringBuffer );
    
private:
    // Shape-related private member functions.    
    void SetupShapeMaterial( void );
    
    s32 BuildTriangle( const nn::math::VEC2& pt1, const nn::math::VEC2& pt2, const nn::math::VEC2& pt3, nn::math::VEC2 *pVecArray );

    s32 BuildRectangle( const nn::math::VEC2& pos, const nn::math::VEC2& size, nn::math::VEC2 *pVecArray );

    s32 BuildRectangle( 
        const nn::math::VEC2& pt1, 
        const nn::math::VEC2& pt2, 
        const nn::math::VEC2& pt3, 
        const nn::math::VEC2& pt4, 
        nn::math::VEC2 *pVecArray );

    s32 BuildLine( const nn::math::VEC2& pt1, const nn::math::VEC2& pt2, f32 lineWidth, nn::math::VEC2 *pVecArray );

    void BuildColor( const s32 vertexCount, const nn::util::FloatColor color );    
    void BuildColor( const s32 vertexCount, const nn::util::FloatColor color1, const nn::util::FloatColor color2,
                     f32* pColorArray );

    // Font-related private member functions.
    bool InitializeFontResource( void* fontData, size_t fontSize );
    bool InitializeFontShader( void* shaderData, size_t shaderSize );    
    void SendFontMatrix( void );

    // Private member functions that use GR.
    void InitializeGraphicsState( void );
    void AddGraphicsStateCommand( void );

    static const s32 DEFAULT_SCREEN_WIDTH                = 320;
    static const s32 DEFAULT_SCREEN_HEIGHT               = 240;
    static const s32 DEFAULT_MAX_TEXT_COUNT              = 512;
    static const s32 DEFAULT_MAX_SHAPE_VERTEX_COUNT      = 128;

    static const s32 SHAPE_DRAW_LINE_VERTEX_COUNT        = 6;
    static const s32 SHAPE_DRAW_RECTANGLE_VERTEX_COUNT   = 4 * SHAPE_DRAW_LINE_VERTEX_COUNT;
    static const s32 SHAPE_FILL_TRIANGLE_VERTEX_COUNT    = 3;
    static const s32 SHAPE_FILL_RECTANGLE_VERTEX_COUNT   = 2 * SHAPE_FILL_TRIANGLE_VERTEX_COUNT;

    // Shape-related member variables.
    IAllocator*                    m_pShapeAllocator;

    nn::math::VEC2                 m_ScreenSize;

    f32                            m_LineWidth;

    s32                            m_MaxShapeVertexCount;
    s32                            m_ShapeVertexCount;

    void*                          m_pShapeShaderBinary;
    nn::gr::ShaderLite             m_ShapeShader;
    u32*                           m_pShaderBuffer;
    nn::gr::BindSymbolVSFloat      m_ShapeBindSymbolProjectionMatrix;
    nn::gr::BindSymbolVSFloat      m_ShapeBindSymbolModelViewMatrix;

    nn::gr::CTR::BindSymbolVSInput m_ShapeBindSymbolPos;
    nn::gr::CTR::BindSymbolVSInput m_ShapeBindSymbolColor;
        
    nn::gr::Vertex                 m_ShapeVertex;
    nn::gr::Vertex::IndexStream    m_ShapeIndexStream;

    nn::math::VEC2*                m_ShapePositionStreamArray;
    f32*                           m_ShapeColorStreamArray;
    u16*                           m_ShapeIndexStreamArray;

    // Font-related member variables.
    IAllocator*                    m_pFontAllocator;

    nn::font::ResFont              m_Font;
    nn::font::TextWriter           m_TextWriter;
    nn::font::RectDrawer           m_Drawer;
    s32                            m_MaxTextCount;    
    void*                          m_FontCommandBuffer;

    // GR member variables
    nn::gr::Viewport               m_Viewport;
    nn::gr::Scissor                m_Scissor;
    nn::gr::RenderState            m_RenderState;
    nn::gr::Combiner               m_Combiner;

    bit32*                         m_CommandStart;
    bit32*                         m_Command;
};

#endif
