﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_PrimitiveVertexStream.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_PRIMITIVE_VERTEX_STREAM_H_
#define GR_PRIMITIVE_VERTEX_STREAM_H_

#include <nn/gr.h>
#include <nn/gx/CTR/gx_CommandAccess.h>

#include "gr_Framework.h"



enum PhysicsType
{
    PHYSICS_DYNAMIC,
    PHYSICS_STATIC
};

class Body
{
public:
    Body( void ) :
    physicsType( PHYSICS_DYNAMIC ),
    m_Pos( nn::math::VEC2( 0.0f, 0.0f ) ),
    m_Vel( nn::math::VEC2( 0.0f, 0.0f ) )
    {
    }

    virtual ~Body()
    {
    }

    virtual void Update( const f32 timeStep )
    {
        if ( physicsType == PHYSICS_DYNAMIC )
        {
            nn::math::VEC2 p = m_Pos + ( timeStep * m_Vel );
        
            SetWorldPosition( p );
        }
    }

    virtual void SetWorldPosition( const nn::math::VEC2 _pos )
    {
        m_Pos = _pos;
    }

    virtual nn::math::VEC2 GetWorldPosition( void ) const
    {
        return m_Pos;
    }

    virtual void SetVelocity( const nn::math::VEC2 _vel )
    {
        m_Vel = _vel;
    }

    virtual nn::math::VEC2 GetVelocity( void ) const
    {
        return m_Vel;
    }

    PhysicsType physicsType;    
    NN_PADDING3;

protected:
    nn::math::VEC2 m_Pos;
    nn::math::VEC2 m_Vel;
};

class AABB 
{
public:
    AABB( void ) :
    isIntersect( false ),
    pos( nn::math::VEC2( 0.0f, 0.0f ) ),
    halfSize( nn::math::VEC2( 0.0f, 0.0f ) )    
    {
    }

    virtual ~AABB( void )
    {
    }

public:    
    bool isIntersect;
    NN_PADDING3;
    
    nn::math::VEC2 pos;
    nn::math::VEC2 halfSize;
};

bool Intersect( const AABB& aabb0, const AABB& aabb1, nn::math::VEC2& normal );

bool IntersectTrianglesCollision( const nn::math::VEC2 pos00, const nn::math::VEC2 pos01,
        const nn::math::VEC2 pos02, const nn::math::VEC2 pos10,
        const nn::math::VEC2 pos11, const nn::math::VEC2 pos12 );

bool IntersectSegmentsCollision( const nn::math::VEC2 pos00, const nn::math::VEC2 pos01,
                                 const nn::math::VEC2 pos10, const nn::math::VEC2 pos11 );

bool IntersectTrianglePointCollision( const nn::math::VEC2 pos00, const nn::math::VEC2 pos01,
                                      const nn::math::VEC2 pos10, const nn::math::VEC2 pos1 );

enum PrimitiveType
{
    PRIMITIVE_TRIANGLES,
    PRIMITIVE_STRIP    ,
    PRIMITIVE_FAN      
};

class PrimitiveBase : public Body
{
public:
    PrimitiveBase( void );
    virtual ~PrimitiveBase();

    virtual void Initialize( IAllocator* pAllocator, 
                             nn::gr::ShaderLite* pShader,
                             const u32 vertexNum );

    virtual void Finalize( void );
    
    virtual void BuildShape( void )
    {
        for ( s32 index = 0; index < m_VertexNum; index++ )
        {
            m_ShapePositionArray[ index ] = 
                nn::math::VEC2( 0.0f, 0.0f );
        }
    }

    virtual void BuildAABB( void );
 
    virtual void SetWorldPosition( const nn::math::VEC2 _pos )
    {
        m_Pos    = _pos;
        aabb.pos = _pos;

        for ( s32 index = 0; index < m_VertexNum; index++ )
        {
            m_PositionStreamArray[ index ] = 
                m_Pos + m_ShapePositionArray[ index ];
        }
    }

    virtual void SetShapePosition( const u32 index, const nn::math::VEC2 _pos )
    {
        NN_ASSERT( index < m_VertexNum );
        
        m_ShapePositionArray[ index ] = _pos;

        m_PositionStreamArray[ index ] = 
                m_Pos + m_ShapePositionArray[ index ];
    }

    virtual u32 GetVertexNum( void ) const
    {
        return m_VertexNum;
    }

    virtual u32 GetTrianglesNum( void ) const
    {
        return m_TrianglesNum;
    }

    virtual void SetTriangleIntersectFlag( const u32 triIndex, 
        const bool isIntersect )
    {
        NN_ASSERT( triIndex < m_TrianglesNum );

        m_TrianglesIntersectArray[ triIndex ] = 
            static_cast<u32>( isIntersect );
    }

    virtual u32 GetTriangleIntersectFlag( const u32 triIndex ) const
    {
        NN_ASSERT( triIndex < m_TrianglesNum );

        return static_cast<u32>( m_TrianglesIntersectArray[ triIndex ] );
    }

    virtual void GetTriangleShapeWorldPosition( const u32 triIndex,
        nn::math::VEC2& pos0, nn::math::VEC2& pos1, nn::math::VEC2& pos2 )
    {
        NN_UNUSED_VAR( triIndex );
        NN_UNUSED_VAR( pos0 );
        NN_UNUSED_VAR( pos1 );
        NN_UNUSED_VAR( pos2 );
    }

    virtual void SetColor( const u32 index,
                           const f32 red, const f32 green, 
                           const f32 blue, const f32 alpha )
    {
        NN_ASSERT( index < m_VertexNum );

        u32 offset                       = 4 * index;
        m_ColorStreamArray[ offset     ] = red;
        m_ColorStreamArray[ offset + 1 ] = green;
        m_ColorStreamArray[ offset + 2 ] = blue;
        m_ColorStreamArray[ offset + 3 ] = alpha;
    }

    virtual PrimitiveType GetPrimitiveType( void ) const
    {
        return m_PrimitiveType;
    }

    virtual bit32* MakeCommand( bit32* command ) const;

protected:
    IAllocator*                    m_pAllocator;
    
    PrimitiveType                  m_PrimitiveType;
    NN_PADDING3;

    s32                            m_VertexNum;
    s32                            m_TrianglesNum;

    nn::gr::ShaderLite*                m_pShader;
    
    nn::math::VEC2                 m_Position;
    nn::math::VEC2*                m_ShapePositionArray;

    nn::math::VEC2*                m_PositionStreamArray;
    f32*                           m_ColorStreamArray;
    u16*                           m_IndexStreamArray;

    u32*                           m_TrianglesIntersectArray;

    nn::gr::CTR::BindSymbolVSInput m_BindSymbolPos;
    nn::gr::CTR::BindSymbolVSInput m_BindSymbolColor;
        
    nn::gr::Vertex                 m_Vertex;
    nn::gr::Vertex::IndexStream    m_IndexStream;

public:
    AABB                           aabb;
};

class PrimitiveTriangles : public PrimitiveBase
{
public:
    PrimitiveTriangles( void );
    virtual ~PrimitiveTriangles();

    virtual void Initialize( IAllocator* pAllocator, 
                             nn::gr::ShaderLite* pShader,
                             const u32 trianglesNum );

    virtual void BuildRectangleShape( const nn::math::VEC2 size );

    virtual void GetTriangleShapeWorldPosition( const u32 triIndex,
        nn::math::VEC2& pos0, nn::math::VEC2& pos1, nn::math::VEC2& pos2 );
};

class PrimitiveFan : public PrimitiveBase
{
public:
    PrimitiveFan( void );
    virtual ~PrimitiveFan();

    virtual void Initialize( IAllocator* pAllocator, 
                             nn::gr::ShaderLite* pShader,
                             const f32 radius, const u32 trianglesNum );

    virtual void BuildShape( void );

    virtual void GetTriangleShapeWorldPosition( const u32 triIndex,
        nn::math::VEC2& pos0, nn::math::VEC2& pos1, nn::math::VEC2& pos2 );

private:
    f32             m_Radius;
};

class PrimitiveStrip : public PrimitiveBase
{
public:
    PrimitiveStrip( void );
    virtual ~PrimitiveStrip();

    virtual void Initialize( IAllocator* pAllocator, 
                             nn::gr::ShaderLite* pShader,
                             const f32 triangleEdge,
                             const u32 triangleNum );

    virtual void BuildShape( void );

    virtual void GetTriangleShapeWorldPosition( const u32 triIndex,
        nn::math::VEC2& pos0, nn::math::VEC2& pos1, nn::math::VEC2& pos2 );

private:
    f32             m_TriangleEdge;
};

#endif // GR_PRIMITIVE_VERTEX_STREAM_H_
