﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_GraphicsDrawing.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd./HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>

#include "gr_Common.h"
#include "gr_GfxAllocator.h"

#include "gr_GraphicsDrawing.h"

//--------------------------------------------------------------------------

GraphicsDrawing::GraphicsDrawing()
 : m_pShapeAllocator( NULL ),
   m_ScreenSize( f32( DEFAULT_SCREEN_WIDTH ), f32( DEFAULT_SCREEN_HEIGHT ) ),
   m_LineWidth( 2.0f ),
   m_MaxShapeVertexCount( DEFAULT_MAX_SHAPE_VERTEX_COUNT ),
   m_ShapeVertexCount( 0 ),
   m_pShapeShaderBinary( NULL ),
   m_ShapeShader(),
   m_ShapeBindSymbolProjectionMatrix(),
   m_ShapeBindSymbolModelViewMatrix(),
   m_ShapeBindSymbolPos(),
   m_ShapeBindSymbolColor(),
   m_ShapeVertex(),
   m_ShapeIndexStream(),
   m_ShapePositionStreamArray( NULL ),
   m_ShapeColorStreamArray( NULL ),
   m_ShapeIndexStreamArray( NULL ),
    m_pFontAllocator( NULL ),
   m_Font(),
   m_TextWriter(),
   m_Drawer(),
   m_MaxTextCount( DEFAULT_MAX_TEXT_COUNT ),
   m_FontCommandBuffer( NULL ),
   m_Viewport(),
   m_Scissor(),
   m_RenderState(),
   m_Combiner(),
   m_CommandStart( NULL ),
   m_Command( NULL )
{
}

//--------------------------------------------------------------------------
GraphicsDrawing::~GraphicsDrawing()
{
}

//--------------------------------------------------------------------------
void 
GraphicsDrawing::Finalize()
{
    this->DestroyShape();

    this->DestroyFont();
        
    m_pShapeAllocator = NULL;
    m_pFontAllocator = NULL;
}

//--------------------------------------------------------------------------
bool
GraphicsDrawing::InitializeShape( void* pShaderBinary, size_t binarySize )
{
    bool result = true;
    m_pShapeShaderBinary = pShaderBinary;

    NN_UNUSED_VAR( binarySize );

    u32 size = m_ShapeShader.GetRequiredBufferSize( m_pShapeShaderBinary );

    m_pShaderBuffer = static_cast<u32*>( m_pShapeAllocator->Alloc( size ) );

    // Specify shader binary and initialize
    // Enable the executable image at position 0 as the vertex shader
    m_ShapeShader.SetupBinary( pShaderBinary, 0, -1, m_pShaderBuffer );
    
    // Search for the Uniform using the string
    result = m_ShapeShader.SearchBindSymbol( &m_ShapeBindSymbolProjectionMatrix, "uProjection" );    
    NN_ASSERT( result );

    result = m_ShapeShader.SearchBindSymbol( &m_ShapeBindSymbolModelViewMatrix, "uModelView" );
    NN_ASSERT( result );
    
    // Search for location using the string      
    result = m_ShapeShader.SearchBindSymbol( &m_ShapeBindSymbolPos, "aPosition" );
    NN_ASSERT( result ); 

    result = m_ShapeShader.SearchBindSymbol( &m_ShapeBindSymbolColor, "aColor" );
    NN_ASSERT( result );

    // Vertex stream setting
    uptr physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_ShapePositionStreamArray ) );
    m_ShapeVertex.EnableAttrAsArray( m_ShapeBindSymbolPos, physicalAddr, PICA_DATA_SIZE_2_FLOAT );

    physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_ShapeColorStreamArray ) );
    m_ShapeVertex.EnableAttrAsArray( m_ShapeBindSymbolColor, physicalAddr, PICA_DATA_SIZE_4_FLOAT );
    
    // Index stream settings
    physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_ShapeIndexStreamArray ) );
    m_ShapeIndexStream.physicalAddr   = physicalAddr;
    m_ShapeIndexStream.isUnsignedByte = false;
   
    return result;
}

//--------------------------------------------------------------------------
bool
GraphicsDrawing::InitializeShape( IAllocator* pShapeAllocator, const wchar_t* shaderPath )
{
    m_pShapeAllocator = pShapeAllocator;

    nn::fs::FileReader shaderFileReader( shaderPath );
    u32 shaderFileSize = static_cast<u32>( shaderFileReader.GetSize() );
    m_pShapeShaderBinary = static_cast<void*>( m_pShapeAllocator->Alloc( shaderFileSize, 128 ) );
    shaderFileReader.Read( m_pShapeShaderBinary, shaderFileSize );
    shaderFileReader.Finalize();
    
    u32 shapePositionStreamArraySize = 2 * 4 * m_MaxShapeVertexCount;
    u32 shapeColorStreamArraySize    = 4 * 4 * m_MaxShapeVertexCount;
    u32 shapeIndexStreamArraySize    = 2 * m_MaxShapeVertexCount;

    m_ShapePositionStreamArray = static_cast<nn::math::VEC2*>( m_pShapeAllocator->Alloc( shapePositionStreamArraySize, 128 ) );
    NN_NULL_ASSERT( m_ShapePositionStreamArray );

    m_ShapeColorStreamArray    = static_cast<f32*>( m_pShapeAllocator->Alloc( shapeColorStreamArraySize, 128 ) );
    NN_NULL_ASSERT( m_ShapeColorStreamArray );

    m_ShapeIndexStreamArray    = static_cast<u16*>( m_pShapeAllocator->Alloc( shapeIndexStreamArraySize, 128 ) );
    NN_NULL_ASSERT( m_ShapeIndexStreamArray );

    for ( s32 index = 0; index < m_MaxShapeVertexCount; index++ )
    {
        m_ShapePositionStreamArray[ index ] = nn::math::VEC2( 0.0f, 0.0f );
        
        u32 colorIndex = 4 * index;
        m_ShapeColorStreamArray[ colorIndex     ] = 1.0f;
        m_ShapeColorStreamArray[ colorIndex + 1 ] = 1.0f;
        m_ShapeColorStreamArray[ colorIndex + 2 ] = 1.0f;
        m_ShapeColorStreamArray[ colorIndex + 3 ] = 1.0f;
    }

    nngxUpdateBuffer( m_ShapePositionStreamArray, shapePositionStreamArraySize );
    nngxUpdateBuffer( m_ShapeColorStreamArray, shapeColorStreamArraySize );
    
    for ( s32 index = 0; index < m_MaxShapeVertexCount; index++ )
    {
        m_ShapeIndexStreamArray[ index ] = static_cast<u16>( index );
    }

    nngxUpdateBuffer( m_ShapeIndexStreamArray, shapeIndexStreamArraySize );
    
    InitializeGraphicsState();

    return InitializeShape( m_pShapeShaderBinary, shaderFileSize );
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::DestroyShape()
{
    m_ShapeVertex.DisableAll();

    if ( m_pShapeAllocator != NULL )
    {
        m_pShapeAllocator->Free( m_pShapeShaderBinary );
        m_pShapeAllocator->Free( m_pShaderBuffer );
        
        m_pShapeAllocator->Free( m_ShapePositionStreamArray );
        m_pShapeAllocator->Free( m_ShapeColorStreamArray );
        m_pShapeAllocator->Free( m_ShapeIndexStreamArray );                
    }
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::SetupShape()
{    
    this->SetupShapeMaterial();
    
    AddGraphicsStateCommand();
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::BeginDrawingShape( void )
{
    SetDrawMode( PICA_DATA_DRAW_TRIANGLES );

    SetupShape();

    m_ShapeVertexCount = 0;

    this->SendShapeMatrix();
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::EndDrawingShape( void )
{   
    if ( m_ShapeVertexCount == 0 )
    {
        return;
    }    

    m_ShapeIndexStream.drawVtxNum    = m_ShapeVertexCount;

    u32 shapePositionStreamArraySize = 2 * 4 * m_ShapeVertexCount;
    u32 shapeColorStreamArraySize    = 4 * 4 * m_ShapeVertexCount;
    u32 shapeIndexStreamArraySize    = 2 * m_ShapeVertexCount;

    nngxUpdateBuffer( m_ShapePositionStreamArray, shapePositionStreamArraySize );
    nngxUpdateBuffer( m_ShapeColorStreamArray,    shapeColorStreamArraySize );
    nngxUpdateBuffer( m_ShapeIndexStreamArray,    shapeIndexStreamArraySize );
    
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &m_Command ) );        
    
    m_CommandStart = m_Command;

    m_Command      = nn::gr::CTR::Vertex::MakeDisableCommand( m_Command );        
    m_Command      = m_ShapeVertex.MakeEnableAttrCommand( m_Command );
    m_Command      = m_ShapeVertex.MakeDrawCommand( m_Command, m_ShapeIndexStream );
    m_Command      = m_ShapeVertex.MakeDisableAttrCommand( m_Command );
    m_Command      = nn::gr::CTR::Vertex::MakeDisableCommand( m_Command );

    nngxMoveCommandbufferPointer( ( m_Command - m_CommandStart ) * sizeof(u32) );
    
    m_ShapeVertexCount = 0;
}


//--------------------------------------------------------------------------
void 
GraphicsDrawing::DrawLine( const nn::math::VEC2& p1, const nn::math::VEC2& p2, const nn::util::FloatColor color )
{    
    NN_ASSERT( ( m_ShapeVertexCount + SHAPE_DRAW_LINE_VERTEX_COUNT ) <= m_MaxShapeVertexCount );

    s32 count = this->BuildLine( p1, p2, m_LineWidth, &m_ShapePositionStreamArray[ m_ShapeVertexCount ] );
    
    BuildColor( count, color );

    m_ShapeVertexCount += count;    
}

//--------------------------------------------------------------------------
void 
GraphicsDrawing::DrawRectangle( const nn::math::VEC2& pos, const nn::math::VEC2& size, nn::util::FloatColor color )
{      
    NN_ASSERT( ( m_ShapeVertexCount + SHAPE_DRAW_RECTANGLE_VERTEX_COUNT ) <= m_MaxShapeVertexCount );

    s32 count = 0;

    count += this->BuildLine( pos, nn::math::VEC2( pos.x + size.x, pos.y ),
        m_LineWidth, &m_ShapePositionStreamArray[ m_ShapeVertexCount + count ] );

    count += this->BuildLine(
        nn::math::VEC2( pos.x + size.x, pos.y ), pos + size,
        m_LineWidth, &m_ShapePositionStreamArray[ m_ShapeVertexCount + count ] );

    count += this->BuildLine( pos + size, nn::math::VEC2( pos.x, pos.y + size.y ),
        m_LineWidth, &m_ShapePositionStreamArray[ m_ShapeVertexCount + count ] );

    count += this->BuildLine(nn::math::VEC2( pos.x, pos.y + size.y ), pos,
        m_LineWidth, &m_ShapePositionStreamArray[ m_ShapeVertexCount + count ] );
            
    BuildColor( count, color );

    m_ShapeVertexCount += count;
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::FillRectangle( const nn::math::VEC2& pos, const nn::math::VEC2& size, const nn::util::FloatColor color )
{
    NN_ASSERT( ( m_ShapeVertexCount + SHAPE_FILL_RECTANGLE_VERTEX_COUNT ) <= m_MaxShapeVertexCount );

    s32 count = this->BuildRectangle(pos, size, &m_ShapePositionStreamArray[ m_ShapeVertexCount ]);
       
    BuildColor( count, color );

    m_ShapeVertexCount += count;        
}

//--------------------------------------------------------------------------
void 
GraphicsDrawing::FillTriangle( const nn::math::VEC2& pt1, const nn::math::VEC2& pt2, const nn::math::VEC2& pt3, nn::util::FloatColor color )
{
    NN_ASSERT( ( m_ShapeVertexCount + SHAPE_FILL_TRIANGLE_VERTEX_COUNT ) <= m_MaxShapeVertexCount );

    s32 count = this->BuildTriangle( pt1, pt2, pt3, &m_ShapePositionStreamArray[ m_ShapeVertexCount ] );
     
    BuildColor( count, color );
        
    m_ShapeVertexCount += count; 
}

//--------------------------------------------------------------------------
void 
GraphicsDrawing::FillTriangles( const nn::math::VEC2* pPointArray, s32 pointCount, nn::util::FloatColor color )
{
    NN_NULL_ASSERT( pPointArray );

    NN_ASSERT( pointCount > 0 );

    NN_ASSERT( ( m_ShapeVertexCount + pointCount ) <= m_MaxShapeVertexCount );
               
    s32 size = 2 * 4 * pointCount;
    std::memcpy( &m_ShapePositionStreamArray[ m_ShapeVertexCount ],
                 pPointArray, size );
        
    BuildColor( pointCount, color );

    m_ShapeVertexCount += pointCount;    
}

//--------------------------------------------------------------------------
bool
GraphicsDrawing::InitializeFont(
    IAllocator* allocator,
    const wchar_t* fontShaderPath,
    const wchar_t* fontPath
)
{
    m_pFontAllocator = allocator;
    
    // Loads the font resource.
    nn::fs::FileReader shaderFileReader( fontShaderPath );
    u32 shaderFileSize = static_cast<u32>( shaderFileReader.GetSize() );
    void* shaderBuffer = static_cast<void*>( m_pFontAllocator->Alloc( shaderFileSize, 128 ) );
    shaderFileReader.Read( shaderBuffer, shaderFileSize );
    shaderFileReader.Finalize();
    
    nn::fs::FileReader fontFileReader( fontPath );
    u32 fontFileSize = static_cast<u32>( fontFileReader.GetSize() );
    void* fontBuffer = static_cast<void*>( m_pFontAllocator->Alloc( fontFileSize, 128 ) );
    fontFileReader.Read( fontBuffer, fontFileSize );
    fontFileReader.Finalize();
    
    const u32 vtxBufCmdBufSize =
        nn::font::RectDrawer::GetVertexBufferCommandBufferSize( shaderBuffer, shaderFileSize );
    
    m_FontCommandBuffer = m_pFontAllocator->Alloc( vtxBufCmdBufSize );
    NN_NULL_ASSERT( m_FontCommandBuffer );

    const u32 drawBufferSize = nn::font::ResFont::GetDrawBufferSize( fontBuffer );
    void* drawBuffer = m_pFontAllocator->Alloc( drawBufferSize, 4 );
    NN_NULL_ASSERT( drawBuffer );

    const u32 stringBufferSize = nn::font::CharWriter::GetDispStringBufferSize( m_MaxTextCount );
    void *const stringBuffer = m_pFontAllocator->Alloc( stringBufferSize ); 
    NN_NULL_ASSERT( stringBuffer );

    bool result = false;
    if ( ( fontBuffer != NULL ) && ( shaderBuffer != NULL ) )
    {
        result = this->InitializeFont( shaderBuffer, shaderFileSize, 
            fontBuffer, fontFileSize, drawBuffer, stringBuffer );
    }

    if ( ! result )
    {
        m_pFontAllocator->Free( fontBuffer );
    }

    m_pFontAllocator->Free( shaderBuffer );

    InitializeGraphicsState();
    
    return result;
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::DestroyFont( void )
{
    m_Drawer.Finalize();

    if ( m_pFontAllocator != NULL )
    {
        m_pFontAllocator->Free( m_FontCommandBuffer );

        nn::font::DispStringBuffer* stringBuffer = m_TextWriter.GetDispStringBuffer();
        m_pFontAllocator->Free( stringBuffer );
    }
    
    // Destroy font
    if (! m_Font.IsManaging( NULL ) )
    {
        void* drawBuffer = m_Font.SetDrawBuffer(NULL);
        void* resource = m_Font.RemoveResource();

        if ( m_pFontAllocator != NULL )
        {
            m_pFontAllocator->Free( resource );
            m_pFontAllocator->Free( drawBuffer );
        }
    }
}

//--------------------------------------------------------------------------
void 
GraphicsDrawing::BeginDrawingString( void )
{
    AddGraphicsStateCommand();

    m_Drawer.DrawBegin();

    this->SendFontMatrix();

    m_TextWriter.SetCursor( 0, 0 );
}

//--------------------------------------------------------------------------
f32
GraphicsDrawing::DrawString( const char* format, ... )
{    
    NN_ASSERT( format );

    std::va_list vargs;
    va_start(vargs, format);

    m_TextWriter.StartPrint();
        f32 width = m_TextWriter.VPrintf(format, vargs);
    m_TextWriter.EndPrint();

    m_Drawer.BuildTextCommand(&m_TextWriter);
    m_TextWriter.UseCommandBuffer();

    va_end(vargs);

    return width;
}

//--------------------------------------------------------------------------
f32
GraphicsDrawing::DrawString( s32 posh, s32 posv, const char* format, ... )
{    
    NN_ASSERT( format );
    
    m_TextWriter.SetCursor( static_cast<f32>(posh), static_cast<f32>(posv) );

    std::va_list vargs;
    va_start(vargs, format);

    f32 width = this->DrawStringArgs( posh, posv, format, vargs);

    va_end(vargs);
    
    return width;
}

//--------------------------------------------------------------------------
f32
GraphicsDrawing::DrawStringArgs( s32 posh, s32 posv, const char* format, std::va_list args )
{    
    NN_ASSERT( format );
    
    m_TextWriter.SetCursor( static_cast<f32>(posh), static_cast<f32>( posv ) );
        
    m_TextWriter.StartPrint();
        f32 width = m_TextWriter.VPrintf(format, args);
    m_TextWriter.EndPrint();

    m_Drawer.BuildTextCommand( &m_TextWriter );
    m_TextWriter.UseCommandBuffer();

    return width;
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::EndDrawingString()
{
    m_Drawer.DrawEnd();
}

//--------------------------------------------------------------------------
bool
GraphicsDrawing::InitializeFont( 
    void* shaderBinary,
    size_t shaderSize,
    void* fontBinary,
    size_t fontSize,
    void* drawBuffer,
    void* stringBuffer
)
{
    if ( ! this->InitializeFontResource( fontBinary, fontSize ) )
    {
        return false;
    }
    
    if ( ! this->InitializeFontShader( shaderBinary, shaderSize ) )
    {
        return false;
    }

    // Sets the render buffer.
    m_Font.SetDrawBuffer(drawBuffer);
    
    m_TextWriter.SetFont( &m_Font );
    m_TextWriter.EnableFixedWidth( true );
    m_TextWriter.SetFixedWidth( 8.5f );
    nn::font::DispStringBuffer *const pDrawStringBuf = 
        nn::font::CharWriter::InitDispStringBuffer( stringBuffer, m_MaxTextCount );
    m_TextWriter.SetDispStringBuffer( pDrawStringBuf );
    
    InitializeGraphicsState();

    return true;
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::InitializeGraphicsState()
{
    // Rendering state settings
    m_RenderState.blend.isEnable          = true;
    m_RenderState.blend.eqRgb             = PICA_DATA_BLEND_EQUATION_ADD;
    m_RenderState.blend.eqAlpha           = PICA_DATA_BLEND_EQUATION_ADD;
    m_RenderState.blend.srcRgb            = PICA_DATA_BLEND_FUNC_SRC_ALPHA;
    m_RenderState.blend.srcAlpha          = PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_ALPHA;
    m_RenderState.blend.dstRgb            = PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_ALPHA;
    m_RenderState.blend.dstAlpha          = PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_ALPHA;

    m_RenderState.stencilTest.isEnable    = false;

    m_RenderState.depthTest.isEnable      = false;
    m_RenderState.depthTest.isEnableWrite = false;

    m_RenderState.cullingTest.isEnable    = false;

    // Texture combiner settings
    for ( u32 stageIndex = 0; stageIndex < nn::gr::CTR::Combiner::COMBINER_STAGE_MAX; stageIndex++ )
    {
        m_Combiner.stage[ stageIndex ].rgb.combine   = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
        m_Combiner.stage[ stageIndex ].alpha.combine = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
        
        for ( u32 sourceIndex = 0; sourceIndex < 3; sourceIndex++)
        {
            if ( ( stageIndex == 0 ) && ( sourceIndex == 0 ) )
            {        
                m_Combiner.stage[ stageIndex ].rgb.source[ sourceIndex ]   = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
                m_Combiner.stage[ stageIndex ].alpha.source[ sourceIndex ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
            }
            else
            {
                m_Combiner.stage[ stageIndex ].rgb.source[ sourceIndex ]   = PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS;
                m_Combiner.stage[ stageIndex ].alpha.source[ sourceIndex ] = PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS;
            }

            m_Combiner.stage[ stageIndex ].rgb.operand[ sourceIndex ]      = PICA_DATA_OPE_RGB_SRC_COLOR;                        
            m_Combiner.stage[ stageIndex ].alpha.operand[ sourceIndex ]    = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
        }

        m_Combiner.stage[ stageIndex ].rgb.scale   = PICA_DATA_TEX_ENV_SCALE_1;
        m_Combiner.stage[ stageIndex ].alpha.scale = PICA_DATA_TEX_ENV_SCALE_1;
    }
}

//--------------------------------------------------------------------------

void 
GraphicsDrawing::AddGraphicsStateCommand( void )
{
    m_Viewport.Set( 0, 0, m_ScreenSize.y, m_ScreenSize.x );
    m_Scissor.Set( 0, 0, m_ScreenSize.y, m_ScreenSize.x );                

    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &m_Command ) );        

    m_CommandStart = m_Command;
    m_Command      = m_Viewport.MakeCommand( m_Command );
    m_Command      = m_Scissor.MakeCommand( m_Command );
    m_Command      = m_RenderState.MakeCommand( m_Command ) ;
    m_Command      = m_Combiner.MakeCommand( m_Command );

    nngxMoveCommandbufferPointer( ( m_Command - m_CommandStart ) * sizeof(u32) );    
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::SetupShapeMaterial( void )
{
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &m_Command ) );        

    m_CommandStart = m_Command;

    m_Command = m_ShapeShader.MakeFullCommand( m_Command );    
    
    nngxMoveCommandbufferPointer( ( m_Command - m_CommandStart ) * sizeof(u32) );    
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::SendShapeMatrix( void )
{
    nn::math::MTX44 projMtx;
    nn::math::MTX44 viewMtx;

    const nn::math::VEC3 camPos( 0.0f, 0.0f, 1.0f );
    const nn::math::VEC3 aimPos( 0.0f, 0.0f, 0.0f );
    const nn::math::VEC3 camUp( 0.0f, 1.0f, 0.0f );

    const f32 left   = 0.f;
    f32 right        = m_ScreenSize.x;
    f32 bottom       = m_ScreenSize.y;
    const f32 top    = 0.f;

    const f32 znear = - 100.0f;
    const f32 zfar  =   100.0f;

    nn::math::MTX44OrthoPivot( &projMtx, left, right, bottom,
        top, znear, zfar, nn::math::PIVOT_UPSIDE_TO_TOP );

    nn::math::MTX44Identity( &viewMtx );
    nn::math::MTX34LookAt( reinterpret_cast<nn::math::MTX34*>(&viewMtx), &camPos, &camUp, &aimPos );
    
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &m_Command ) );        
    m_CommandStart = m_Command;

    // Set Shader's Uniform
    {       
        m_Command = m_ShapeBindSymbolProjectionMatrix.MakeUniformCommand( m_Command, projMtx );
        m_Command = m_ShapeBindSymbolModelViewMatrix.MakeUniformCommand( m_Command, viewMtx );
    }

    nngxMoveCommandbufferPointer( ( m_Command - m_CommandStart ) * sizeof(u32) );
}

//--------------------------------------------------------------------------
s32 
GraphicsDrawing::BuildTriangle( const nn::math::VEC2& pt1, const nn::math::VEC2& pt2, const nn::math::VEC2& pt3, nn::math::VEC2 *pVecArray )
{
    NN_NULL_ASSERT( pVecArray );
    
    pVecArray[ 0 ] = pt1;
    pVecArray[ 1 ] = pt2;
    pVecArray[ 2 ] = pt3;
    
    return 3;
}

//--------------------------------------------------------------------------
s32
GraphicsDrawing::BuildRectangle( 
    const nn::math::VEC2& pt1, 
    const nn::math::VEC2& pt2, 
    const nn::math::VEC2& pt3, 
    const nn::math::VEC2& pt4, 
    nn::math::VEC2 *pVecArray )
{
    NN_NULL_ASSERT( pVecArray );
    s32 count = 0;
    
    count += this->BuildTriangle( pt1, pt2, pt3, &pVecArray[count] );
    count += this->BuildTriangle( pt1, pt3, pt4, &pVecArray[count] );
    
    return count;
}

//--------------------------------------------------------------------------
s32 
GraphicsDrawing::BuildRectangle( const nn::math::VEC2& pos, const nn::math::VEC2& size, nn::math::VEC2 *pVecArray )
{
    NN_NULL_ASSERT( pVecArray );

    s32 count = 0;    
    count += this->BuildRectangle( 
        pos, 
        nn::math::VEC2(pos.x + size.x, pos.y), 
        pos + size, 
        nn::math::VEC2(pos.x, pos.y + size.y), 
        pVecArray );
    
    return count;
}

//--------------------------------------------------------------------------
s32
GraphicsDrawing::BuildLine( const nn::math::VEC2& pt1, const nn::math::VEC2& pt2, f32 lineWidth, nn::math::VEC2 *pVecArray )
{
    NN_ASSERT( lineWidth >= 0.f );
    NN_NULL_ASSERT( pVecArray );
        
    nn::math::VEC2 direction  = pt1 - pt2;
    nn::math::VEC2 ortho      = nn::math::VEC2( direction.y, - direction.x );

    VEC2Normalize( &ortho, &ortho );
    ortho *= lineWidth;

    s32 count = 0;
    
    count += this->BuildRectangle( pt1 - ortho / 2, pt1 + ortho / 2, pt2 + ortho / 2, pt2 - ortho / 2, pVecArray );
    
    return count;
}

//--------------------------------------------------------------------------
void
GraphicsDrawing::BuildColor( const s32 vertexCount, const nn::util::FloatColor color )
{     
    for ( s32 index = 0; index < vertexCount; index++ )    
    {        
        s32 colorIndex = 4 * m_ShapeVertexCount + 4 * index;

        m_ShapeColorStreamArray[ colorIndex     ] = color.r;
        m_ShapeColorStreamArray[ colorIndex + 1 ] = color.g;
        m_ShapeColorStreamArray[ colorIndex + 2 ] = color.b;
        m_ShapeColorStreamArray[ colorIndex + 3 ] = color.a;
    }
}

//--------------------------------------------------------------------------

void
GraphicsDrawing::BuildColor( const s32 vertexCount, const nn::util::FloatColor color1, 
                             const nn::util::FloatColor color2, f32* pColorArray )                             
{        
    for ( s32 index = 0; index < vertexCount; index++ )
    {
        s32 colorIndex = 4 * index;
        if ( ( index == 0 ) || ( index ==  1 ) || ( index == 3 ) )
        {         
            pColorArray[ colorIndex     ] = color1.r;
            pColorArray[ colorIndex + 1 ] = color1.g;
            pColorArray[ colorIndex + 2 ] = color1.b;
            pColorArray[ colorIndex + 3 ] = color1.a;
        }
        else
        {            
            pColorArray[ colorIndex     ] = color2.r;
            pColorArray[ colorIndex + 1 ] = color2.g;
            pColorArray[ colorIndex + 2 ] = color2.b;
            pColorArray[ colorIndex + 3 ] = color2.a;
        }
    }
}
//--------------------------------------------------------------------------
bool
GraphicsDrawing::InitializeFontResource( void* fontData, size_t fontSize )
{
    NN_UNUSED_VAR( fontSize );
    
    // Sets the font resource.
    return m_Font.SetResource( fontData );
}


//--------------------------------------------------------------------------
bool
GraphicsDrawing::InitializeFontShader( void* shaderBinary, size_t shaderSize )
{
    m_Drawer.Initialize(m_FontCommandBuffer, shaderBinary, shaderSize);

    return true;
}

//--------------------------------------------------------------------------
void 
GraphicsDrawing::SendFontMatrix( void )
{
    //---- Set the projection matrix to an orthogonal projection
    {
        // Sets the origin at the upper left, and the Y and Z axes in opposite directions.
        nn::math::MTX44 proj;
        const f32 znear =   0.0f;
        const f32 zfar  = - 1.0f;
        const f32 t     =      0;
        f32 b           =  static_cast<f32>( m_ScreenSize.y );
        const f32 l     =      0;
        f32 r           =  static_cast<f32>( m_ScreenSize.x );
        
        nn::math::MTX44OrthoPivot( &proj, l, r, b, t, znear, zfar, nn::math::PIVOT_UPSIDE_TO_TOP );

        m_Drawer.SetProjectionMtx( proj );
    }

    //---- Set the model view matrix to the identity matrix.
    {
        nn::math::MTX34 mv;
        nn::math::MTX34Identity(&mv);

        m_Drawer.SetViewMtxForText(mv);
    }
}
