﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_GraphicsDrawingScene.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/fs.h>

#include "gr_Framework.h"
#include "gr_GraphicsDrawingScene.h"

const wchar_t* SHAPE_2D_SHADER_FILE_NAME = L"rom:/Shape2d.shbin";
const wchar_t* FONT_SHADER_FILE_NAME     = L"rom:/FontRectDrawer.shbin";
const wchar_t* FONT_FILE_NAME            = L"rom:/tahoma.bcfnt";

GraphicsDrawing  s_GraphicsDrawing;

//------------------------------------------------------------------------------

f32 GetRandomFloat( const f32 rangeBegin, const f32 rangeEnd )
{
    const u32 RANGE_DIVISION_NUM = 1000;

    static bool s_InitializeRandomSeed = false;
    if (! s_InitializeRandomSeed )
    {
        ::std::srand(0x1234);
        s_InitializeRandomSeed = true;
    }
    
    f32 randomValue = 
        ( ::std::rand() % RANGE_DIVISION_NUM ) / static_cast<f32>( RANGE_DIVISION_NUM );

    f32 range = rangeEnd - rangeBegin;
    
    return randomValue * range + rangeBegin;
}

//------------------------------------------------------------------------------
GfxAllocator * GrAllocator( MemoryMgr::AllocatorDevice *heap )
{   
    return new GfxAllocator( heap );
}

GraphicSystem::GfxAllocatorFactory  GetGraphicsMemoryMgrFunctor( void )
{
    return &GrAllocator;
}

IScene * CreateScene( void )
{
    return new GraphicsDrawingScene();
}


//------------------------------------------------------------------------------
GraphicsDrawingScene::GraphicsDrawingScene()
    : allocator( NULL ),
      m_Counter( 0 ),
      m_PrimitiveNum( 0 )
{
}


//------------------------------------------------------------------------------
GraphicsDrawingScene::~GraphicsDrawingScene()
{
}

//------------------------------------------------------------------------------
bool GraphicsDrawingScene::Initialize()
{
    bool is_success;

    NN_UNUSED_VAR( is_success );
    
    allocator = MEMMGR->GetDefaultAllocatorDevice();

    // Performs initialization for font rendering.
    is_success = s_GraphicsDrawing.InitializeFont( allocator, 
        FONT_SHADER_FILE_NAME, FONT_FILE_NAME );
    NN_ASSERTMSG( is_success , "Fail to initialize font .");

    // Performs initialization for shape rendering.
    s_GraphicsDrawing.SetMaxShapeVertexCount( MAX_SHAPE_VERTEX_COUNT );
    is_success = s_GraphicsDrawing.InitializeShape( allocator, SHAPE_2D_SHADER_FILE_NAME );
    NN_ASSERTMSG( is_success , "Fail to initialize shape .");

    // Get allocator and deallocator
    nngxGetAllocator( &m_GraphicsMemoryAllocator, &m_GraphicsMemoryDeallocator );
    {
        // Allocate color buffer  GL_RGBA8_OES
        renderBuffer[ 0 ] = 1;
        GLenum area = NN_GX_MEM_VRAMA;
        GLenum aim  = NN_GX_MEM_RENDERBUFFER;
        u32 colorBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        colorBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 0 ], colorBufferByteSize ) );
        
        // Allocate depth buffer  GL_DEPTH24_STENCIL8_EXT
        renderBuffer[ 1 ] = 2;
        area = NN_GX_MEM_VRAMB;
        u32 depthBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        depthBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 1 ], depthBufferByteSize ) );
    }

    {
        renderTarget = 1;

        // Initialize framebuffer settings
        m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

        // Initialize color buffer settings
        m_FrameBuffer.colorBuffer.virtualAddr         = colorBufferVirtualAddress;
        m_FrameBuffer.colorBuffer.format              = PICA_DATA_COLOR_RGBA8_OES;
        m_FrameBuffer.colorBuffer.blockSize           = nn::gr::FrameBuffer::BLOCK_SIZE8;
        m_FrameBuffer.colorBuffer.width               = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.colorBuffer.height              = nn::gx::DISPLAY0_HEIGHT;  
        m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;
       
        // Initialize depth buffer settings
        m_FrameBuffer.depthStencilBuffer.virtualAddr  = depthBufferVirtualAddress;
        m_FrameBuffer.depthStencilBuffer.format       = PICA_DATA_DEPTH24_STENCIL8_EXT;
        m_FrameBuffer.depthStencilBuffer.width        = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.depthStencilBuffer.height       = nn::gx::DISPLAY0_HEIGHT;
        m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
        m_FrameBuffer.depthStencilBuffer.clearStencil = 0;
    }

    m_Scissor.SetBufferSize( nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT );

    // Initialize wall
    {
        nn::gr::ShaderLite* pShader = s_GraphicsDrawing.GetGrShaderPointer();
        const u32 triangleNum = 2;

        const f32 WALL_HALF_THICKNESS = 0.5f * WALL_THICKNESS;
        const f32 WALL_X_HALF_LENGTH  = 0.5f * WALL_X_LENGTH;
        const f32 WALL_Y_HALF_LENGTH  = 0.5f * WALL_Y_LENGTH;

        for ( u32 index = 0; index < MAX_WALL_PRIMITIVE_NUM; index++)
        {
            PrimitiveTriangles* pPrimitiveTriangles = & m_PrimitiveTrianglesWallArray[ index ];
                    
            pPrimitiveTriangles->Initialize( allocator, pShader, triangleNum );
                    
            nn::math::VEC2 halfSize( 0.0f, 0.0f );            
            if ( ( index == 0 ) || ( index == 1 ) )
            {
                halfSize.x = WALL_X_HALF_LENGTH;
                halfSize.y = WALL_HALF_THICKNESS;
            }
            else
            {
                halfSize.x = WALL_HALF_THICKNESS;
                halfSize.y = WALL_Y_HALF_LENGTH;            
            }

            pPrimitiveTriangles->BuildRectangleShape( halfSize );
            
            f32 x          = 0.0f;
            f32 y          = 0.0f;
            if ( index == 0 )
            {
                x = WALL_X_HALF_LENGTH;
                y = WALL_HALF_THICKNESS;
            }
            else if ( index == 1 )
            {
                x = WALL_X_HALF_LENGTH;
                y = WALL_Y_LENGTH - WALL_HALF_THICKNESS;
            }
            else if ( index == 2 )
            {
                x = WALL_HALF_THICKNESS;
                y = WALL_Y_HALF_LENGTH;
            }
            else
            {
                x = WALL_X_LENGTH - WALL_HALF_THICKNESS;            
                y = WALL_Y_HALF_LENGTH;
            }
            
            nn::math::VEC2 pos( x, y );
            pPrimitiveTriangles->SetWorldPosition( pos );
            
            nn::math::VEC2 vel( 0.0f, 0.0f );            
            pPrimitiveTriangles->SetVelocity( vel );

            u32 vertexNum = pPrimitiveTriangles->GetVertexNum();

            f32 red   = 0.0f;
            f32 green = 0.0f;
            f32 blue  = 0.0f;
            f32 alpha = 1.0f;

            for ( u32 vertexIndex = 0; vertexIndex < vertexNum; vertexIndex++)
            {                            
                if ( ( vertexIndex == 0 ) || ( vertexIndex == 1 ) || ( vertexIndex == 3 ) )
                {
                    red   = 0.6f;
                    green = 1.0f;
                    blue  = 1.0f;
                }
                else
                {
                    red   = 0.3f;
                    green = 0.5f;
                    blue  = 0.5f;
                }

                pPrimitiveTriangles->SetColor( vertexIndex,
                    red, green, blue, alpha );
            }
            
            pPrimitiveTriangles->physicsType = PHYSICS_STATIC;

            RegisterPrimitive( pPrimitiveTriangles );
        }
    }
   
    // Initialize rectangle
    {
        nn::gr::ShaderLite* pShader = s_GraphicsDrawing.GetGrShaderPointer();
        const u32 triangleNum   = TRIANGLES_NUM;

        for ( u32 index = 0; index < MAX_TRIANGLES_NUM; index++ )
        {
            PrimitiveTriangles* pPrimitiveTriangles = & m_PrimitiveTrianglesArray[ index ];
                    
            pPrimitiveTriangles->Initialize( allocator, pShader, triangleNum );
                           

            nn::math::VEC2 halfSize = nn::math::VEC2(
                                    GetRandomFloat( 5.0f, 25.0f ),
                                    GetRandomFloat( 5.0f, 25.0f )
                                    );
            pPrimitiveTriangles->BuildRectangleShape( halfSize );
                                    
            nn::math::VEC2 pos = nn::math::VEC2(
                                    GetRandomFloat( 40.0f, 360.0f ),
                                    GetRandomFloat( 40.0f, 200.0f )
                                    );

            pPrimitiveTriangles->SetWorldPosition( pos );
            
            nn::math::VEC2 vel( GetRandomFloat( - 5.0f, 5.0f ),
                                GetRandomFloat( - 5.0f, 5.0f ) );          
            pPrimitiveTriangles->SetVelocity( vel );
            
            f32 red   = 0.0f;
            f32 green = 0.0f;
            f32 blue  = 0.0f;
            f32 alpha = 1.0f;
            f32 grad  = 0.5f;

            f32 randomValue = GetRandomFloat( 0.0f, 1.0f );

            if ( randomValue <= 0.33f )
            {
                red   = 1.0f;
                green = 0.6f;
                blue  = 0.6f;
            }
            else if ( randomValue <= 0.66f )
            {
                red   = 0.6f;
                green = 1.0f;
                blue  = 0.6f;
            }
            else
            {
                red   = 0.6f;
                green = 0.6f;
                blue  = 1.0f;
            }

            u32 vertexNum = pPrimitiveTriangles->GetVertexNum();
            for ( u32 vertexIndex = 0; vertexIndex < vertexNum; vertexIndex++)
            {
                if ( ( vertexIndex == 0 ) || ( vertexIndex == 1 ) || ( vertexIndex == 3 ) )
                {
                    pPrimitiveTriangles->SetColor( vertexIndex,
                        red, green, blue, alpha );
                }
                else
                {
                    pPrimitiveTriangles->SetColor( vertexIndex,
                        grad * red, grad * green, grad * blue, alpha );
                }
            }
            
            pPrimitiveTriangles->physicsType = PHYSICS_DYNAMIC;
            
            RegisterPrimitive( pPrimitiveTriangles );

        }
    }
    
    // Initialize triangle fan
    {
        nn::gr::ShaderLite* pShader = s_GraphicsDrawing.GetGrShaderPointer();
        const u32 trianglesNum  = TRIANGLE_FAN_TRIANGLES_NUM;
        
        for ( u32 index = 0; index < MAX_TRIANGLE_FAN_NUM; index++)
        {
            f32 radius = GetRandomFloat( 20.0f, 60.0f );

            PrimitiveFan* pPrimitiveFan = & m_PrimitiveFanArray[ index ];
            
            pPrimitiveFan->Initialize( allocator, pShader, 
                radius, trianglesNum );    
            
            nn::math::VEC2 pos = nn::math::VEC2(
                                    GetRandomFloat( 40.0f, 360.0f ),
                                    GetRandomFloat( 40.0f, 200.0f )
                                    );
            pPrimitiveFan->SetWorldPosition( pos );

            nn::math::VEC2 vel = nn::math::VEC2(
                                    GetRandomFloat( - 5.0f, 5.0f ),
                                    GetRandomFloat( - 5.0f, 5.0f )
                                    );
            pPrimitiveFan->SetVelocity( vel );
  
            f32 red   = 0.0f;
            f32 green = 0.0f;
            f32 blue  = 0.0f;
            f32 alpha = 1.0f;
            f32 grad  = 0.3f;

            f32 randomValue = GetRandomFloat( 0.0f, 1.0f );

            if ( randomValue <= 0.33f )
            {
                red   = 1.0f;
                green = 0.6f;
                blue  = 0.6f;
            }
            else if ( randomValue <= 0.66f )
            {
                red   = 0.6f;
                green = 1.0f;
                blue  = 0.6f;
            }
            else
            {
                red   = 0.6f;
                green = 0.6f;
                blue  = 1.0f;
            }
            
            const u32 totalVertexNum  = pPrimitiveFan->GetVertexNum();
            for ( u32 vertexIndex = 0; vertexIndex < totalVertexNum; vertexIndex++)
            {
                if ( vertexIndex == 0 )
                {
                    pPrimitiveFan->SetColor( vertexIndex,
                        red, green, blue, alpha );
                }
                else
                {
                    pPrimitiveFan->SetColor( vertexIndex,
                        grad * red, grad * green, grad * blue, alpha );
                }
            }

            pPrimitiveFan->physicsType = PHYSICS_DYNAMIC;

            RegisterPrimitive( pPrimitiveFan );
        }
    }

    // Initialize triangle strip
    {
        nn::gr::ShaderLite* pShader = s_GraphicsDrawing.GetGrShaderPointer();        
        const f32 triangleEdge  = TRIANGLE_STRIP_LENGTH;
        const u32 triangleNum   = TRIANGLE_STRIP_TRIANGLES_NUM;
        
        for ( u32 index = 0; index < MAX_TRIANGLE_STRIP_NUM; index++)
        {
            PrimitiveStrip* pPrimitiveStrip = & m_PrimitiveStripArray[ index ];
            
            pPrimitiveStrip->Initialize( allocator, pShader, 
                triangleEdge, triangleNum );

            nn::math::VEC2 pos = nn::math::VEC2(
                                    GetRandomFloat( 40.0f, 360.0f ),
                                    GetRandomFloat( 40.0f, 200.0f )
                                    );
            pPrimitiveStrip->SetWorldPosition( pos );

            nn::math::VEC2 vel = nn::math::VEC2(
                                    GetRandomFloat( - 5.0f, 5.0f ),
                                    GetRandomFloat( - 5.0f, 5.0f )
                                    );
            pPrimitiveStrip->SetVelocity( vel );
        
            f32 red   =  0.0f;
            f32 green =  0.0f;
            f32 blue  =  0.0f;
            f32 alpha =  1.0f;
            f32 grad  =  0.5f;

            f32 randomValue = GetRandomFloat( 0.0f, 1.0f );

            if ( randomValue <= 0.33f )
            {
                red   = 1.0f;
                green = 0.6f;
                blue  = 0.6f;
            }
            else if ( randomValue <= 0.66f )
            {
                red   = 0.6f;
                green = 1.0f;
                blue  = 0.6f;
            }
            else
            {
                red   = 0.6f;
                green = 0.6f;
                blue  = 1.0f;
            }

            const u32 totalVertexNum  = pPrimitiveStrip->GetVertexNum();
            for ( u32 vertexIndex = 0; vertexIndex < totalVertexNum; vertexIndex++)
            {
                if ( ( vertexIndex % 2 ) == 0 )
                {
                    pPrimitiveStrip->SetColor( vertexIndex,
                        red, green, blue, alpha );
                }
                else 
                {
                    pPrimitiveStrip->SetColor( vertexIndex,
                        grad * red, grad * green, grad * blue, alpha );
                }
            }

            pPrimitiveStrip->physicsType = PHYSICS_DYNAMIC;
            
            RegisterPrimitive( pPrimitiveStrip );
        }
    }

    // Initialize using a disable command
    {
        u32* command = NULL;
        nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
            reinterpret_cast<GLint*>( &command ) );
        const u32* start_addr = command;

        command = nn::gr::CTR::MakeDisableAllCommand( command );

        nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );
    }

    return true;
}

//------------------------------------------------------------------------------
void GraphicsDrawingScene::Finalize( void )
{
    for ( u32 index = 0; index < m_PrimitiveNum; index++)
    {
        PrimitiveBase* pPrimitiveBase = m_pPrimitiveBaseArray[ index ];
        PrimitiveType primitiveType   = pPrimitiveBase->GetPrimitiveType();

        if ( primitiveType == PRIMITIVE_TRIANGLES )
        {
            PrimitiveTriangles* pTriangles = 
                static_cast<PrimitiveTriangles*>( pPrimitiveBase );
            pTriangles->Finalize();
        }
        else if ( primitiveType == PRIMITIVE_FAN )
        {
            PrimitiveFan* pFan = 
                static_cast<PrimitiveFan*>( pPrimitiveBase );
            pFan->Finalize();
        }
        else if ( primitiveType == PRIMITIVE_STRIP )
        {
            PrimitiveStrip* pStrip = 
                static_cast<PrimitiveStrip*>( pPrimitiveBase );
            pStrip->Finalize();
        }
    }

    s_GraphicsDrawing.Finalize();

    GLenum area = NN_GX_MEM_VRAMA;
    GLenum aim  = NN_GX_MEM_RENDERBUFFER;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 0 ], reinterpret_cast<void*>( colorBufferVirtualAddress ) );
    colorBufferVirtualAddress = 0;

    area = NN_GX_MEM_VRAMB;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 1 ], reinterpret_cast<void*>( depthBufferVirtualAddress ) );
    depthBufferVirtualAddress = 0;
}

//------------------------------------------------------------------------------
void GraphicsDrawingScene::Enter( void )
{
    stereo.Initialize();
    nngxSetDisplayMode( NN_GX_DISPLAYMODE_STEREO );
    Calc();
}


//------------------------------------------------------------------------------
bool GraphicsDrawingScene::Calc( void )
{
    // Wait for Draw to finish before manipulating vertices (for double command lists)
    // Here, nngxWaitCmdlistDone is used to wait for all command requests to complete, but by using nngxSetCmdlistCallback, you can wait for any specific command request to complete.
    // 
    nngxWaitCmdlistDone();

    UpdateScene_();
    ++m_Counter;
 
    return true;
}

//------------------------------------------------------------------------------
void GraphicsDrawingScene::PreDraw( void )
{
}

//------------------------------------------------------------------------------
GLuint GraphicsDrawingScene::DrawTop( void )
{
    m_Viewport.Set( 0, 0, 240, 400 );
    m_Scissor.Set( 0, 0, 240, 400 );

    s_GraphicsDrawing.SetScreenSize( 400, 240 );
  
    m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
    m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    // The camera matrix
    {        
        viewMtx.SetupLookAt(
            nn::math::VEC3( 0.0f, 0.0f, 1.0f ),
            nn::math::VEC3( 0.0f,         1.0f, 0.0f ),
            nn::math::VEC3( 0.0f,         0.0f, 0.0f ) );
    }

    // Projection matrix
    projectionMtx.SetupPerspective(
        0.66f,          // fovy
        400.f / 240.f,  // aspect
        10.f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &viewMtx );
    stereo.SetBaseFrustum( &projectionMtx );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f );
        
    RenderSceneGr_();

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint GraphicsDrawingScene::DrawExt( void )
{
    m_Viewport.Set( 0, 0, 240, 400 );
    m_Scissor.Set( 0, 0, 240, 400 );

    s_GraphicsDrawing.SetScreenSize( 400, 240 );

    m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
    m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &viewMtx );
    stereo.SetBaseFrustum( &projectionMtx );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f );
    
    RenderSceneGr_();

    return renderTarget;
}


//------------------------------------------------------------------------------
GLuint GraphicsDrawingScene::DrawBtm( void )
{
    m_Viewport.Set( 0, 0, 240, 320 );
    m_Scissor.Set( 0, 0, 240, 320 );

    s_GraphicsDrawing.SetScreenSize( 320, 240 );

    m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = nn::gx::DISPLAY1_WIDTH;
    m_FrameBuffer.height                          = nn::gx::DISPLAY1_HEIGHT;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );
    
    RenderSceneResult_();

    return renderTarget;
}

//------------------------------------------------------------------------------
void GraphicsDrawingScene::PostDraw( void )
{
}

//------------------------------------------------------------------------------
nn::gr::CTR::FrameBuffer::ColorBuffer* GraphicsDrawingScene::GetGrColorBufferPointer( void )
{    
    return & m_FrameBuffer.colorBuffer;
}

//------------------------------------------------------------------------------
void GraphicsDrawingScene::Clear( const u32 bufferBit )
{
    u32* command = 0;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
        reinterpret_cast<GLint*>( &command ) );

    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Framebuffer settings
    command = m_FrameBuffer.MakeCommand( command, bufferBit );
                                             
    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );

    // Add a load complete command at the end of the 3D command buffer
    bool isAddSplitDrawCmdlist = true;

    // Generate a request to clear the color buffer and depth buffer
    m_FrameBuffer.MakeClearRequest( bufferBit, isAddSplitDrawCmdlist );
}

//------------------------------------------------------------------------------
void GraphicsDrawingScene::UpdateScene_( void )
{
    s_AABBCollisionNum     = 0;
    s_TriangleCollisionNum = 0;

    nn::math::VEC2 pos( 0.0f, 0.0f );
    nn::math::VEC2 vel( 0.0f, 0.0f );

    for ( u32 index = 0; index < m_PrimitiveNum; index++)
    {   
        PrimitiveBase* pPrimitiveBase = m_pPrimitiveBaseArray[ index ];
        pPrimitiveBase->Update( s_TimeStep );

        pPrimitiveBase->aabb.isIntersect = false;

        u32 trianglesNum = pPrimitiveBase->GetTrianglesNum();
        for ( u32 triIndex = 0; triIndex < trianglesNum; triIndex++ )
        {        
            pPrimitiveBase->SetTriangleIntersectFlag( triIndex, false );
        }
    }

    DetectAndSolveCollision_();
}

void GraphicsDrawingScene::DetectAndSolveCollision_( void )
{
    // Broad phase
    for ( u32 index0 = 0; index0 < m_PrimitiveNum; index0++ )
    {
        PrimitiveBase* pPrimitiveBase0 = m_pPrimitiveBaseArray[ index0 ];
        AABB& aabb0                    = pPrimitiveBase0->aabb;

        for ( u32 index1 = ( index0 + 1 ); 
              index1 < m_PrimitiveNum; index1++ )
        {
            PrimitiveBase* pPrimitiveBase1 = m_pPrimitiveBaseArray[ index1 ];
            AABB& aabb1                    = pPrimitiveBase1->aabb;
                        
            nn::math::VEC2 normal;
            if ( Intersect( aabb0, aabb1, normal ) )
            {
                if ( pPrimitiveBase0->physicsType == PHYSICS_DYNAMIC )
                {
                    aabb0.isIntersect = true;
                }

                if ( pPrimitiveBase1->physicsType == PHYSICS_DYNAMIC )
                {
                    aabb1.isIntersect = true;
                }

                if ( ( pPrimitiveBase0->physicsType == PHYSICS_DYNAMIC ) &&
                     ( pPrimitiveBase1->physicsType == PHYSICS_DYNAMIC ) )
                {
                    s_AABBCollisionNum += 1;                    
                }

                SolveBroadCollisionPhase_( pPrimitiveBase0, pPrimitiveBase1, normal );

                DetectNarrowCollisionPhase_( pPrimitiveBase0, pPrimitiveBase1 );
            }            
        }     
    }    
}

void GraphicsDrawingScene::SolveBroadCollisionPhase_( PrimitiveBase* pPrimitiveBase0, PrimitiveBase* pPrimitiveBase1,
                                                      nn::math::VEC2 normal )
{
    bool solveWallCollision = false;

    AABB* pRigidBodyAABB    = NULL;
    AABB* pWallAABB         = NULL;

    PrimitiveBase* pRigidBodyPrimitiveBase = NULL;
    PrimitiveBase* pWallPrimitiveBase      = NULL;
    
    if ( ( pPrimitiveBase0->physicsType == PHYSICS_DYNAMIC ) && 
         ( pPrimitiveBase1->physicsType != PHYSICS_DYNAMIC ) )
    {
        pRigidBodyAABB = &pPrimitiveBase0->aabb;
        pWallAABB      = &pPrimitiveBase1->aabb;

        pRigidBodyPrimitiveBase = pPrimitiveBase0;
        pWallPrimitiveBase      = pPrimitiveBase1;

        solveWallCollision = true;
    }
    else if ( ( pPrimitiveBase0->physicsType != PHYSICS_DYNAMIC ) && 
              ( pPrimitiveBase1->physicsType == PHYSICS_DYNAMIC ) )
    {
        pRigidBodyAABB = &pPrimitiveBase1->aabb;
        pWallAABB      = &pPrimitiveBase0->aabb;

        pRigidBodyPrimitiveBase = pPrimitiveBase1;
        pWallPrimitiveBase      = pPrimitiveBase0;

        solveWallCollision = true;
    }

    NN_UNUSED_VAR( pWallPrimitiveBase );
    
    if ( solveWallCollision )
    {
        nn::math::VEC2 pos     = pRigidBodyPrimitiveBase->GetWorldPosition();
        nn::math::VEC2 vel     = pRigidBodyPrimitiveBase->GetVelocity();

        // Wall - X
        if ( ( normal.x > 0.0f ) && ( normal.y == 0.0f ) )
        {        
            pos.x = pRigidBodyAABB->halfSize.x + 2 * pWallAABB->halfSize.x;
            vel.x = (- vel.x ) * s_VelocityDamping;
        }            
        // Wall + X
        else if ( ( normal.x < 0.0f ) && ( normal.y == 0.0f ) )
        {
            pos.x = WALL_X_LENGTH - ( pRigidBodyAABB->halfSize.x + 2 * pWallAABB->halfSize.x );
            vel.x = (- vel.x ) * s_VelocityDamping;
        }
        // Wall - Y
        else if ( ( normal.x == 0.0f ) && ( normal.y > 0.0f ) )
        {
            pos.y = pRigidBodyAABB->halfSize.y + 2 * pWallAABB->halfSize.y;
            vel.y = (- vel.y ) * s_VelocityDamping;
        }
        // Wall + Y
        else if ( ( normal.x == 0.0f ) && ( normal.y < 0.0f ) )
        {
            pos.y = WALL_Y_LENGTH - ( pRigidBodyAABB->halfSize.y + 2 * pWallAABB->halfSize.y );
            vel.y = (- vel.y ) * s_VelocityDamping;
        }
        
        pRigidBodyPrimitiveBase->SetWorldPosition( pos );
        pRigidBodyPrimitiveBase->SetVelocity( vel );

        s_AABBCollisionNum += 1;
    }
}

//------------------------------------------------------------------------------
void GraphicsDrawingScene::DetectNarrowCollisionPhase_( PrimitiveBase* pPrimitiveBase0, PrimitiveBase* pPrimitiveBase1 )
{    
    u32 trianglesNum0               = pPrimitiveBase0->GetTrianglesNum();
    u32 trianglesNum1               = pPrimitiveBase1->GetTrianglesNum();

    PrimitiveType primitiveType0    = pPrimitiveBase0->GetPrimitiveType();
    PrimitiveType primitiveType1    = pPrimitiveBase1->GetPrimitiveType();

    PrimitiveTriangles* pTriangles0 = NULL;
    PrimitiveTriangles* pTriangles1 = NULL;

    PrimitiveFan* pFan0             = NULL;
    PrimitiveFan* pFan1             = NULL;

    PrimitiveStrip* pStrip0         = NULL;
    PrimitiveStrip* pStrip1         = NULL;

    if ( primitiveType0 == PRIMITIVE_TRIANGLES )
    {
        pTriangles0 = static_cast<PrimitiveTriangles*>( pPrimitiveBase0 );
    }
    else if ( primitiveType0 == PRIMITIVE_FAN )
    {
        pFan0 = static_cast<PrimitiveFan*>( pPrimitiveBase0 );
    }
    else if ( primitiveType0 == PRIMITIVE_STRIP )
    {
        pStrip0 = static_cast<PrimitiveStrip*>( pPrimitiveBase0 );
    }

    if ( primitiveType1 == PRIMITIVE_TRIANGLES )
    {
        pTriangles1 = static_cast<PrimitiveTriangles*>( pPrimitiveBase1 );
    }
    else if ( primitiveType1 == PRIMITIVE_FAN )
    {
        pFan1  = static_cast<PrimitiveFan*>( pPrimitiveBase1 );
    }
    else if ( primitiveType1 == PRIMITIVE_STRIP )
    {
        pStrip1 = static_cast<PrimitiveStrip*>( pPrimitiveBase1 );
    }

    nn::math::VEC2 pos00( 0.0f, 0.0f );
    nn::math::VEC2 pos01( 0.0f, 0.0f );
    nn::math::VEC2 pos02( 0.0f, 0.0f );

    nn::math::VEC2 pos10( 0.0f, 0.0f );
    nn::math::VEC2 pos11( 0.0f, 0.0f );
    nn::math::VEC2 pos12( 0.0f, 0.0f );

    // Calculates by checking all types.
    for ( u32 triIndex0 = 0; triIndex0 < trianglesNum0; triIndex0++ )
    {   
        if ( primitiveType0 == PRIMITIVE_TRIANGLES )
        {
            pTriangles0->GetTriangleShapeWorldPosition( triIndex0,
                pos00, pos01, pos02 );
        }
        else if ( primitiveType0 == PRIMITIVE_FAN )
        {
            pFan0->GetTriangleShapeWorldPosition( triIndex0,
                pos00, pos01, pos02 );
        }
        else if ( primitiveType0 == PRIMITIVE_STRIP )
        {
            pStrip0->GetTriangleShapeWorldPosition( triIndex0,
                pos00, pos01, pos02 );
        }
 
        for ( u32 triIndex1 = 0; triIndex1 < trianglesNum1; triIndex1++ )
        {   
            if ( primitiveType1 == PRIMITIVE_TRIANGLES )
            {
                pTriangles1->GetTriangleShapeWorldPosition( triIndex1,
                    pos10, pos11, pos12 );
            }
            else if ( primitiveType1 == PRIMITIVE_FAN )
            {
                pFan1->GetTriangleShapeWorldPosition( triIndex1,
                    pos10, pos11, pos12 );
            }
            else if ( primitiveType1 == PRIMITIVE_STRIP )
            {
                pStrip1->GetTriangleShapeWorldPosition( triIndex1,
                    pos10, pos11, pos12 );
            }

            bool isIntersect = 
                IntersectTrianglesCollision( pos00, pos01, pos02,
                pos10, pos11, pos12 );

            if ( isIntersect )
            {
                if ( pPrimitiveBase0->physicsType == PHYSICS_DYNAMIC )
                {
                    pPrimitiveBase0->SetTriangleIntersectFlag( triIndex0, true );
                }

                if ( pPrimitiveBase1->physicsType == PHYSICS_DYNAMIC )
                {
                    pPrimitiveBase1->SetTriangleIntersectFlag( triIndex1, true );
                }

                if ( ( pPrimitiveBase0->physicsType == PHYSICS_DYNAMIC ) && 
                     ( pPrimitiveBase1->physicsType == PHYSICS_DYNAMIC ) )
                {
                    s_TriangleCollisionNum += 1;
                }
            }
        }

    }

}

//------------------------------------------------------------------------------
void GraphicsDrawingScene::RenderSceneGr_( void )
{
    // Get a pointer to the render command buffer
    u32* command = NULL;

    {
        nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );
    
        // Get the start address (so that the pointer can be advanced at the end)
        const u32* start_addr = command;

        // Viewport settings
        command = m_Viewport.MakeCommand( command );

        // Scissor settings
        command = m_Scissor.MakeCommand( command );
        
        // Render state settings
        // Cull, Depth, Stencil, AlphaTest, Blend, etc.
        nn::gr::RenderState render_state;

        // w-buffer settings    
        command = render_state.MakeCommand( command );

        // Advance the pointer to the render command buffer
        nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );
    }

    // Renders a polygon.
    RenderSceneFillPolygons_();
       
    s_GraphicsDrawing.BeginDrawingShape();

    // Renders AABB.
    RenderSceneDrawAABBs_();

    // Renders intersecting triangles in pink
    RenderSceneFillIntersectTriangles_();

    s_GraphicsDrawing.EndDrawingShape();    
}

//------------------------------------------------------------------------------

void GraphicsDrawingScene::RenderSceneFillPolygons_( void )
{
    u32* command = NULL;

    PrimitiveType currentPrimitiveType = PRIMITIVE_TRIANGLES;
    PicaDataDrawMode drawMode          = PICA_DATA_DRAW_TRIANGLES;
    s_GraphicsDrawing.SetDrawMode( drawMode );
    s_GraphicsDrawing.SetupShape();
    s_GraphicsDrawing.SendShapeMatrix();

    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );
    u32* start_addr = command;

    for ( u32 index = 0; index < m_PrimitiveNum; index++)
    {
        PrimitiveBase* pPrimitiveBase = m_pPrimitiveBaseArray[ index ];
               
        PrimitiveType primitiveType = pPrimitiveBase->GetPrimitiveType();
        
        if ( primitiveType != currentPrimitiveType )
        {   
            if ( ( command - start_addr ) > 0 )
            {
                nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );
            }

            if ( primitiveType == PRIMITIVE_TRIANGLES )
            {            
                drawMode = PICA_DATA_DRAW_TRIANGLES;
            }
            else if ( primitiveType == PRIMITIVE_FAN )
            {                
                drawMode = PICA_DATA_DRAW_TRIANGLE_FAN;
            }
            else if ( primitiveType == PRIMITIVE_STRIP )
            {             
                drawMode = PICA_DATA_DRAW_TRIANGLE_STRIP;
            }
            
            s_GraphicsDrawing.SetDrawMode( drawMode );
            s_GraphicsDrawing.SetupShape();

            nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );
            start_addr = command;

            currentPrimitiveType = primitiveType;
        }

        command = pPrimitiveBase->MakeCommand( command );
    }

    if ( ( command - start_addr ) > 0 )
    {
        nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );
    }
}

//------------------------------------------------------------------------------

void GraphicsDrawingScene::RenderSceneDrawAABBs_( void )
{        
    for ( u32 index = 0; index < m_PrimitiveNum; index++)
    {
        PrimitiveBase* pPrimitiveBase = m_pPrimitiveBaseArray[ index ];
        AABB& aabb = pPrimitiveBase->aabb;

        nn::util::FloatColor color;

        if ( pPrimitiveBase->physicsType == PHYSICS_DYNAMIC )
        {
            if ( aabb.isIntersect )
            {
                color.r = 1.0f;
                color.g = 0.0f;
                color.b = 0.0f;
                color.a = 1.0f;
            }
            else
            {
                color.r = 0.0f;
                color.g = 1.0f;
                color.b = 0.0f;
                color.a = 1.0f;
            }

            nn::math::VEC2 pos  = aabb.pos - aabb.halfSize - AABB_DRAW_LINE_OFFSET;
            nn::math::VEC2 size = 2 * ( aabb.halfSize + AABB_DRAW_LINE_OFFSET );
            s_GraphicsDrawing.DrawRectangle( pos, size, color );
        }
    }
}

//------------------------------------------------------------------------------

void GraphicsDrawingScene::RenderSceneFillIntersectTriangles_( void )
{
    nn::util::FloatColor color;        
    color.r = 1.0f;
    color.g = 0.6f;
    color.b = 0.6f;
    color.a = 1.0f;

    nn::math::VEC2 pos0( 0.0f, 0.0f );
    nn::math::VEC2 pos1( 0.0f, 0.0f );
    nn::math::VEC2 pos2( 0.0f, 0.0f );
    
    for ( u32 index = 0; index < m_PrimitiveNum; index++)
    {
        PrimitiveBase* pPrimitiveBase = m_pPrimitiveBaseArray[ index ];

        u32 trianglesNum = pPrimitiveBase->GetTrianglesNum();

        PrimitiveType primitiveType = pPrimitiveBase->GetPrimitiveType();

        if ( primitiveType == PRIMITIVE_TRIANGLES )
        {
            PrimitiveTriangles* pTriangles = 
                static_cast<PrimitiveTriangles*>( pPrimitiveBase );

            for ( u32 triIndex = 0; triIndex < trianglesNum; triIndex++)
            {
                if ( pPrimitiveBase->GetTriangleIntersectFlag( triIndex ) )
                {
                    pTriangles->GetTriangleShapeWorldPosition( triIndex,
                        pos0, pos1, pos2 );

                    s_GraphicsDrawing.FillTriangle( pos0, pos1, pos2, color );
                }
            }
        }
        else if ( primitiveType == PRIMITIVE_FAN )
        {
            PrimitiveFan* pFan = 
                static_cast<PrimitiveFan*>( pPrimitiveBase );
            
            for ( u32 triIndex = 0; triIndex < trianglesNum; triIndex++)
            {   
                if ( pPrimitiveBase->GetTriangleIntersectFlag( triIndex ) )
                {
                    pFan->GetTriangleShapeWorldPosition( triIndex,
                        pos0, pos1, pos2 );

                    s_GraphicsDrawing.FillTriangle( pos0, pos1, pos2, color );
                }
            }
        }
        else if ( primitiveType == PRIMITIVE_STRIP )
        {
            PrimitiveStrip* pStrip = 
                static_cast<PrimitiveStrip*>( pPrimitiveBase );
            
            for ( u32 triIndex = 0; triIndex < trianglesNum; triIndex++)
            {    
                if ( pPrimitiveBase->GetTriangleIntersectFlag( triIndex ) )
                {
                    pStrip->GetTriangleShapeWorldPosition( triIndex,
                        pos0, pos1, pos2 );

                    s_GraphicsDrawing.FillTriangle( pos0, pos1, pos2, color );
                }
            }
        }
    }
}

//------------------------------------------------------------------------------
void GraphicsDrawingScene::RenderSceneResult_( void )
{
    // Get a pointer to the render command buffer
    u32* command;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );

    // Get the start address (so that the pointer can be advanced at the end)
    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );
    
    // Render state settings
    // Cull, Depth, Stencil, AlphaTest, Blend, etc.
    nn::gr::RenderState render_state;

    // w-buffer settings    
    command = render_state.MakeCommand( command );
        
    // Advance the pointer to the render command buffer
    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );

    // ---- Start font rendering.
    s_GraphicsDrawing.BeginDrawingString();

    s_GraphicsDrawing.DrawString( "AABB     collision num = %d\n", s_AABBCollisionNum );
    s_GraphicsDrawing.DrawString( "Triangle collision num = %d\n", s_TriangleCollisionNum );
    
    // ---- Finish font rendering.
    s_GraphicsDrawing.EndDrawingString();    
}

//------------------------------------------------------------------------------

void GraphicsDrawingScene::RegisterPrimitive( PrimitiveBase* pPrimitiveBase )
{
    NN_ASSERT( m_PrimitiveNum < MAX_PRIMITIVE_NUM );

    m_pPrimitiveBaseArray[  m_PrimitiveNum ]  = pPrimitiveBase;
    m_PrimitiveNum                           += 1;
}
