﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_PrimitiveVertexStream.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include "gr_PrimitiveVertexStream.h"


struct SEGMENT3
{
public:
    typedef SEGMENT3 self_type;
public:
    SEGMENT3() {}
    explicit SEGMENT3(const f32* p) : P0(p), P1(p + 3) {}
    SEGMENT3(const nn::math::VEC3& pt0, const nn::math::VEC3& pt1) : P0(pt0), P1(pt1) {}

    operator f32*() { return &P0.x; }
    operator const f32*() const { return &P0.x; }
    bool operator==(const self_type& rhs) const { return P0 == rhs.P0 && P1 == rhs.P1; }
    bool operator!=(const self_type& rhs) const { return P0 != rhs.P0 || P1 == rhs.P1; }
public:
    nn::math::VEC3 P0;
    nn::math::VEC3 P1;
};

f32 DistSqSegment3ToSegment3(const SEGMENT3* S1, const SEGMENT3* S2, f32* s, f32* t)
{
    nn::math::VEC3   u(S1->P1 - S1->P0);
    nn::math::VEC3   v(S2->P1 - S2->P0);
    nn::math::VEC3   w(S1->P0 - S2->P0);
    f32    a = nn::math::VEC3SquareLen(&u);        // always >= 0
    f32    b = nn::math::VEC3Dot(&u, &v);
    f32    c = nn::math::VEC3SquareLen(&v);        // always >= 0
    f32    d = nn::math::VEC3Dot(&u, &w);
    f32    e = nn::math::VEC3Dot(&v, &w);
    f32    D = a*c - b*b;       // always >= 0
    f32    sc, sN, sD = D;      // sc = sN / sD, default sD = D >= 0
    f32    tc, tN, tD = D;      // tc = tN / tD, default tD = D >= 0

    // Obtain the nearest point when the line segment is taken as line
    if (D < 0.0001F) { // The two line segments are parallel
        sN = 0.0;              // Take the distance between S1 starting point and S2.
        sD = 1.0;              // Measure against division operation in the code to follow
        tN = e;
        tD = c;
    }
    else {                     // Calculate the nearest point when taken as line, not line segment
        sN = (b*e - c*d);
        tN = (a*e - b*d);

        if (sN < 0.0f) {       // The side, s = 0, is "visible" if sc < 0.
            sN = 0.0f;
            tN = e;
            tD = c;
        }
        else if (sN > sD) {    // The side, s = 1, is "visible" if sc > 1.
            sN = sD;
            tN = e + b;
            tD = c;
        }
    }

    if (tN < 0.0f) {            // The side, t = 0, is "visible" if tc < 0.
        tN = 0.0f;

        // Recalculate the nearest point of sc with tc = 0.
        if (-d < 0.0f)
            sN = 0.0f;
        else if (-d > a)
            sN = sD;
        else {
            sN = -d;
            sD = a;
        }
    }
    else if (tN > tD) {        // The side, t = 1, is "visible" if tc > 1.
        tN = tD;
        // Recalculate the nearest point of sc with tc = 1.
        if ((-d + b) < 0.0f)
            sN = 0.0f;
        else if ((-d + b) > a)
            sN = sD;
        else {
            sN = (-d + b);
            sD = a;
        }
    }
    // finally do the division to get sc and tc
    sc = (nn::math::FAbs(sN) < 0.0001F ? 0.0f : sN / sD);
    tc = (nn::math::FAbs(tN) < 0.0001F ? 0.0f : tN / tD);

    if (s)
        *s = sc;
    if (t)
        *t = tc;

    // Take the difference between the nearest points.
    nn::math::VEC3   dP = w + (u * sc) - (v * tc);  // = S1(sc) - S2(tc)

    return nn::math::VEC3SquareLen(&dP);
}

//------------------------------------------------------------------------------

bool Intersect( const AABB& aabb0, const AABB& aabb1, nn::math::VEC2& normal )
{        
    f32 minDepth = 0.0f;

    for ( u32 index = 0; index < 2; index++ )
    {
        f32 posDiff    = aabb1.pos[ index ] - aabb0.pos[ index ];
        f32 aabbLength = aabb0.halfSize[ index ] + aabb1.halfSize[ index ];
        
        if ( nn::math::FAbs( posDiff ) >= aabbLength )
        {
            return false;
        }
        else
        {
            f32 depth = aabbLength - nn::math::FAbs( posDiff );

            if ( index == 0 )
            {
                if ( posDiff > 0.0f )
                {                       
                    normal.x = 1.0f;
                    normal.y = 0.0f;
                }
                else
                {
                    normal.x = - 1.0f;
                    normal.y = 0.0f;
                }

                minDepth = depth;
            }
            else if ( index == 1 )
            {                
                if ( depth < minDepth )
                {
                    if ( posDiff > 0.0f )
                    {
                        normal.x = 0.0f;
                        normal.y = 1.0f;
                    }
                    else
                    {
                        normal.x = 0.0f;
                        normal.y = - 1.0f;
                    }

                    minDepth = depth;
                }
            }
        }
    }
        
    return true;
}

//------------------------------------------------------------------------------

bool IntersectTrianglesCollision( const nn::math::VEC2 pos00, const nn::math::VEC2 pos01,
        const nn::math::VEC2 pos02, const nn::math::VEC2 pos10,
        const nn::math::VEC2 pos11, const nn::math::VEC2 pos12 )
{
    // Determine intersection of lines
    if ( IntersectSegmentsCollision( pos00, pos01, pos10, pos11) ||
         IntersectSegmentsCollision( pos00, pos01, pos10, pos12 ) ||
         IntersectSegmentsCollision( pos00, pos01, pos11, pos12 ) ||
         IntersectSegmentsCollision( pos00, pos02, pos10, pos11 ) ||
         IntersectSegmentsCollision( pos00, pos02, pos10, pos12 ) ||
         IntersectSegmentsCollision( pos00, pos02, pos11, pos12 ) ||
         IntersectSegmentsCollision( pos01, pos02, pos10, pos11 ) ||
         IntersectSegmentsCollision( pos01, pos02, pos10, pos12 ) ||
         IntersectSegmentsCollision( pos01, pos02, pos11, pos12 ) )
    {
        return true;
    }
    else
    {
        // Determine comprehensions
        if ( IntersectTrianglePointCollision( pos00, pos01, pos02, pos10 ) &&
             IntersectTrianglePointCollision( pos00, pos01, pos02, pos11 ) &&
             IntersectTrianglePointCollision( pos00, pos01, pos02, pos12 ) )
        {
            return true;
        }
        else if ( IntersectTrianglePointCollision( pos10, pos11, pos12, pos00 ) && 
                  IntersectTrianglePointCollision( pos10, pos11, pos12, pos01 ) && 
                  IntersectTrianglePointCollision( pos10, pos11, pos12, pos02 ) )
        {
            return true;
        }
        else
        {
            return false;
        }
    }                                     
}

//------------------------------------------------------------------------------

bool IntersectSegmentsCollision( const nn::math::VEC2 pos00, const nn::math::VEC2 pos01,
                                 const nn::math::VEC2 pos10, const nn::math::VEC2 pos11 )
{
    const f32 EPSILON2 = 0.01f;

    SEGMENT3 segment0;
    segment0.P0 = nn::math::VEC3( pos00.x, pos00.y, 0.0f );
    segment0.P1 = nn::math::VEC3( pos01.x, pos01.y, 0.0f );
    
    SEGMENT3 segment1;
    segment1.P0 = nn::math::VEC3( pos10.x, pos10.y, 0.0f );
    segment1.P1 = nn::math::VEC3( pos11.x, pos11.y, 0.0f );

    f32 s = 0.0f;
    f32 t = 0.0f;
    f32 distSq = DistSqSegment3ToSegment3( &segment0, &segment1, &s, &t );
    
    if ( distSq < EPSILON2 )
    {
        return true;
    }
    else
    {
        return false;
    }
}

//------------------------------------------------------------------------------

f32 CrossVec2( nn::math::VEC2* vec0, nn::math::VEC2* vec1 )
{
    nn::math::VEC3 vec3_0( vec0->x, vec0->y, 0.0f );
    nn::math::VEC3 vec3_1( vec1->x, vec1->y, 0.0f );

    nn::math::VEC3 result( 0.0f, 0.0f, 0.0f );

    nn::math::VEC3Cross( &result, &vec3_0, &vec3_1 );

    return result.z;
}

//------------------------------------------------------------------------------

bool IntersectTrianglePointCollision( const nn::math::VEC2 pos00, const nn::math::VEC2 pos01,
                                      const nn::math::VEC2 pos02, const nn::math::VEC2 pos1 )
{    
    nn::math::VEC2 vec0    = pos01 - pos00;
    nn::math::VEC2 posVec0 = pos1  - pos00;
       
    nn::math::VEC2 vec1    = pos02 - pos01;
    nn::math::VEC2 posVec1 = pos1  - pos01;

    nn::math::VEC2 vec2    = pos00 - pos02;
    nn::math::VEC2 posVec2 = pos1  - pos02;

    if ( ( CrossVec2( &vec0, &posVec0 ) >= 0.0f ) &&
         ( CrossVec2( &vec1, &posVec1 ) >= 0.0f ) &&
         ( CrossVec2( &vec2, &posVec2 ) >= 0.0f ) )
    {
        return true;
    }
    else
    {
        return false;
    }
}

//------------------------------------------------------------------------------

PrimitiveBase::PrimitiveBase( void ) :
Body(),
m_pAllocator( NULL ),
m_PrimitiveType( PRIMITIVE_TRIANGLES ),
m_VertexNum( 0 ), m_TrianglesNum( 0 ),
m_pShader( NULL ),
m_PositionStreamArray( NULL ),
m_ColorStreamArray( NULL ),
m_IndexStreamArray( NULL ),
m_BindSymbolPos(),
m_BindSymbolColor(),
m_Vertex(),
m_IndexStream()
{
}

//------------------------------------------------------------------------------

PrimitiveBase::~PrimitiveBase()
{
}

//------------------------------------------------------------------------------

void PrimitiveBase::Initialize( IAllocator* pAllocator,
                                nn::gr::ShaderLite*  pShader,
                                const u32 vertexNum )
{
    NN_NULL_ASSERT( pAllocator );
    NN_NULL_ASSERT( pShader );

    m_pAllocator = pAllocator;
    m_pShader    = pShader;
    m_VertexNum  = vertexNum;
    
    u32 positionStreamArraySize = 2 * 4 * m_VertexNum;    
    u32 colorStreamArraySize    = 4 * 4 * m_VertexNum;    
    u32 indexStreamArraySize    = 2 * m_VertexNum;
    u32 isIntersectArraySize    = 4 * m_VertexNum;

    m_PositionStreamArray = 
        static_cast<nn::math::VEC2*>( m_pAllocator->Alloc( positionStreamArraySize, 128 ) );
    NN_NULL_ASSERT( m_PositionStreamArray );

    m_ShapePositionArray = 
        static_cast<nn::math::VEC2*>( m_pAllocator->Alloc( positionStreamArraySize, 128 ) );
    NN_NULL_ASSERT( m_ShapePositionArray );

    m_ColorStreamArray    = 
        static_cast<f32*>( m_pAllocator->Alloc( colorStreamArraySize, 128 ) );
    NN_NULL_ASSERT( m_ColorStreamArray );

    m_IndexStreamArray    = 
        static_cast<u16*>( m_pAllocator->Alloc( indexStreamArraySize, 128 ) );
    NN_NULL_ASSERT( m_IndexStreamArray );
    
    m_TrianglesIntersectArray = 
        static_cast<u32*>( m_pAllocator->Alloc( isIntersectArraySize, 128 ) );
    NN_NULL_ASSERT( m_TrianglesIntersectArray );

    BuildShape();
    BuildAABB();

    for ( s32 index = 0; index < m_VertexNum; index++ )
    {
        m_PositionStreamArray[ index ] = nn::math::VEC2( 0.0f, 0.0f );

        u32 colorIndex = 4 * index;
        m_ColorStreamArray[ colorIndex     ] = 1.0f;
        m_ColorStreamArray[ colorIndex + 1 ] = 1.0f;
        m_ColorStreamArray[ colorIndex + 2 ] = 1.0f;
        m_ColorStreamArray[ colorIndex + 3 ] = 1.0f;
    }
    
    for ( s32 index = 0; index < m_VertexNum; index++ )
    {
        m_IndexStreamArray[ index ] = static_cast<u16>( index );
    }    
    nngxUpdateBuffer( m_IndexStreamArray, indexStreamArraySize );

    m_IndexStream.drawVtxNum = m_VertexNum;

    SetWorldPosition( nn::math::VEC2( 0.0f, 0.0f ) );

    bool result = false;
    NN_UNUSED_VAR( result );

    // Search for location using the string      
    result = m_pShader->SearchBindSymbol( &m_BindSymbolPos, "aPosition" );
    NN_ASSERT( result ); 

    result = m_pShader->SearchBindSymbol( &m_BindSymbolColor, "aColor" );
    NN_ASSERT( result );

    // Vertex stream setting
    uptr physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_PositionStreamArray ) );
    m_Vertex.EnableAttrAsArray( m_BindSymbolPos, physicalAddr, PICA_DATA_SIZE_2_FLOAT );

    physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_ColorStreamArray ) );
    m_Vertex.EnableAttrAsArray( m_BindSymbolColor, physicalAddr, PICA_DATA_SIZE_4_FLOAT );
    
    // Index stream settings
    physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_IndexStreamArray ) );
    m_IndexStream.physicalAddr   = physicalAddr;    
    m_IndexStream.isUnsignedByte = false;
}

//------------------------------------------------------------------------------

void PrimitiveBase::Finalize( void )
{
    m_Vertex.DisableAll();

    if ( m_pAllocator != NULL )
    {
        m_pAllocator->Free( m_PositionStreamArray );        
        m_pAllocator->Free( m_ColorStreamArray );
        m_pAllocator->Free( m_IndexStreamArray );

        m_pAllocator->Free( m_TrianglesIntersectArray );

        m_pAllocator->Free( m_ShapePositionArray );
    }
}

//------------------------------------------------------------------------------

void PrimitiveBase::BuildAABB( void )
{
    f32 minX = m_ShapePositionArray[ 0 ].x;
    f32 maxX = m_ShapePositionArray[ 0 ].x;
    f32 minY = m_ShapePositionArray[ 0 ].y;
    f32 maxY = m_ShapePositionArray[ 0 ].y;

    for ( s32 index = 1; index < m_VertexNum; index++ )
    {
        if ( m_ShapePositionArray[ index ].x < minX )
        {
            minX = m_ShapePositionArray[ index ].x;
        }

        if ( m_ShapePositionArray[ index ].x > maxX )
        {
            maxX = m_ShapePositionArray[ index ].x;
        }

        if ( m_ShapePositionArray[ index ].y < minY )
        {
            minY = m_ShapePositionArray[ index ].y;
        }

        if ( m_ShapePositionArray[ index ].y > maxY )
        {
            maxY = m_ShapePositionArray[ index ].y;
        }
    }        
    
    f32 centerX = 0.5f * ( minX + maxX );
    f32 centerY = 0.5f * ( minY + maxY );

    for ( s32 index = 0; index < m_VertexNum; index++ )
    {
         m_ShapePositionArray[ index ] -=
             nn::math::VEC2( centerX, centerY );
    }

    aabb.halfSize.x = 0.5f * ( maxX - minX );
    aabb.halfSize.y = 0.5f * ( maxY - minY );
}

//------------------------------------------------------------------------------

bit32* PrimitiveBase::MakeCommand( bit32* command ) const
{    
    s32 positionStreamArraySize = 2 * 4 * m_VertexNum;
    s32 colorStreamArraySize    = 4 * 4 * m_VertexNum;

    nngxUpdateBuffer( m_PositionStreamArray, positionStreamArraySize );
    nngxUpdateBuffer( m_ColorStreamArray, colorStreamArraySize );
            
    command = nn::gr::CTR::Vertex::MakeDisableCommand( command );
    command = m_Vertex.MakeEnableAttrCommand( command );
    command = m_Vertex.MakeDrawCommand( command, m_IndexStream );
    command = m_Vertex.MakeDisableAttrCommand( command );
    command = nn::gr::CTR::Vertex::MakeDisableCommand( command );

    return command;
}

//--------------------------------------------------------------------------

PrimitiveTriangles::PrimitiveTriangles( void ) :
PrimitiveBase()
{
    m_PrimitiveType = PRIMITIVE_TRIANGLES;
}

//------------------------------------------------------------------------------

PrimitiveTriangles::~PrimitiveTriangles()
{
}

//------------------------------------------------------------------------------

void PrimitiveTriangles::Initialize( IAllocator* pAllocator, nn::gr::ShaderLite*  pShader,
                                     const u32 trianglesNum )
{
    NN_NULL_ASSERT( pAllocator );
    NN_NULL_ASSERT( pShader );
    
    NN_ASSERT( trianglesNum >= 1 );
    
    u32 vertexNum  = 3 * trianglesNum;
    m_TrianglesNum = trianglesNum;
    
    PrimitiveBase::Initialize( pAllocator, pShader, vertexNum );        
}

//------------------------------------------------------------------------------

void PrimitiveTriangles::BuildRectangleShape( const nn::math::VEC2 halfSize )
{    
    NN_ASSERT( m_VertexNum == 6 );

    u32 index = 0;
    m_ShapePositionArray[ index ].x = ( - halfSize.x );
    m_ShapePositionArray[ index ].y = ( - halfSize.y );

    index = 1;
    m_ShapePositionArray[ index ].x =     halfSize.x;
    m_ShapePositionArray[ index ].y = ( - halfSize.y );

    index = 2;
    m_ShapePositionArray[ index ].x = ( - halfSize.x );
    m_ShapePositionArray[ index ].y = halfSize.y;

    index = 3;
    m_ShapePositionArray[ index ].x = halfSize.x;
    m_ShapePositionArray[ index ].y = ( - halfSize.y );

    index = 4;
    m_ShapePositionArray[ index ].x = halfSize.x;
    m_ShapePositionArray[ index ].y = halfSize.y;

    index = 5;
    m_ShapePositionArray[ index ].x = (- halfSize.x );
    m_ShapePositionArray[ index ].y = halfSize.y;

    aabb.halfSize                   = halfSize;

    SetWorldPosition( m_Pos );    
}

//------------------------------------------------------------------------------

void PrimitiveTriangles::GetTriangleShapeWorldPosition( const u32 triIndex,
                                                        nn::math::VEC2& pos0,
                                                        nn::math::VEC2& pos1, nn::math::VEC2& pos2 )
{
    NN_ASSERT( triIndex < m_TrianglesNum );

    u32 index = 3 * triIndex;
    pos0      = m_ShapePositionArray[ index ]    + m_Pos;
    pos1      = m_ShapePositionArray[ index + 1] + m_Pos;
    pos2      = m_ShapePositionArray[ index + 2] + m_Pos;
}

//------------------------------------------------------------------------------

PrimitiveFan::PrimitiveFan( void ) :
PrimitiveBase(),
m_Radius( 0.5f )
{
    m_PrimitiveType = PRIMITIVE_FAN;
}

//------------------------------------------------------------------------------

PrimitiveFan::~PrimitiveFan()
{
}

//------------------------------------------------------------------------------

void PrimitiveFan::Initialize( IAllocator* pAllocator,
                               nn::gr::ShaderLite*  pShader,
                               const f32 radius, const u32 trianglesNum )
{
    NN_NULL_ASSERT( pAllocator );
    NN_NULL_ASSERT( pShader );

    NN_ASSERT( radius >= 0.0f );
    NN_ASSERT( trianglesNum >= 1 );
    
    m_Radius       = radius;

    u32 vertexNum  = trianglesNum + 2;
    m_TrianglesNum = trianglesNum;
    
    PrimitiveBase::Initialize( pAllocator, pShader, vertexNum );        
}

//------------------------------------------------------------------------------

void PrimitiveFan::BuildShape( void )
{
    const f32 deltaTheta = 360.0f / ( m_VertexNum - 2 );

    for ( s32 index = 0; index < m_VertexNum; index++ )
    {
        if ( index == 0 )
        {
            m_ShapePositionArray[ index ] = nn::math::VEC2( 0.0f, 0.0f );
        }
        else
        {
            f32 sinTheta = 0.0f;
            f32 cosTheta = 0.0f;
            
            f32 theta    = deltaTheta * ( index - 1 );
            if ( index == ( m_VertexNum - 1 ) )
            {
                theta = 0.0f;
            }
            
            nn::math::SinCosDeg( &sinTheta, &cosTheta, theta );

            m_ShapePositionArray[ index ].x = m_Radius * cosTheta;
            m_ShapePositionArray[ index ].y = m_Radius * sinTheta;
        }
    }
}

//------------------------------------------------------------------------------

void PrimitiveFan::GetTriangleShapeWorldPosition( const u32 triIndex,
                                                  nn::math::VEC2& pos0, nn::math::VEC2& pos1, 
                                                  nn::math::VEC2& pos2 )
{
    NN_ASSERT( triIndex < m_TrianglesNum );

    pos0 = m_ShapePositionArray[            0 ] + m_Pos;
    pos1 = m_ShapePositionArray[ triIndex + 1 ] + m_Pos;
    pos2 = m_ShapePositionArray[ triIndex + 2 ] + m_Pos;
}

//------------------------------------------------------------------------------

PrimitiveStrip::PrimitiveStrip( void ) :
PrimitiveBase(),
m_TriangleEdge( 1.0f )
{
    m_PrimitiveType = PRIMITIVE_STRIP;
}

//------------------------------------------------------------------------------

PrimitiveStrip::~PrimitiveStrip()
{
}

//------------------------------------------------------------------------------

void PrimitiveStrip::Initialize( IAllocator* pAllocator,
                                 nn::gr::ShaderLite*  pShader,
                                 const f32 triangleEdge,
                                 const u32 triangleNum )
{
    NN_NULL_ASSERT( pAllocator );
    NN_NULL_ASSERT( pShader );

    NN_ASSERT( triangleEdge >= 0.0f );
    NN_ASSERT( triangleNum >= 1 );
        
    m_TriangleEdge = triangleEdge;

    u32 vertexNum  = triangleNum + 2;
    m_TrianglesNum = triangleNum;
    
    PrimitiveBase::Initialize( pAllocator, pShader, vertexNum );        
}

//------------------------------------------------------------------------------

void PrimitiveStrip::BuildShape( void )
{
    for ( s32 index = 0; index < m_VertexNum; index++ )
    {
        if ( ( index == 0 ) || ( index == 2 ) )
        {            
            f32 sinTheta = 0.0f;
            f32 cosTheta = 0.0f;
            f32 theta    = 30.0f;
            nn::math::SinCosDeg( &sinTheta, &cosTheta, theta );

            f32 signX = - 1;
            if ( index == 2 )
            {
                signX = 1;
            }

            m_ShapePositionArray[ index ].x = signX * sinTheta * m_TriangleEdge;                
            m_ShapePositionArray[ index ].y = cosTheta * m_TriangleEdge;
        }
        else if ( index == 1 )
        {
            m_ShapePositionArray[ index ] = nn::math::VEC2( 0.0f, 0.0f );
        }
        else
        {
            m_ShapePositionArray[ index ] = 
                m_ShapePositionArray[ index - 2 ] +                     
                    nn::math::VEC2( m_TriangleEdge, 0.0f );
        }
    }
}

//------------------------------------------------------------------------------

void PrimitiveStrip::GetTriangleShapeWorldPosition( const u32 triIndex,
                                                    nn::math::VEC2& pos0, nn::math::VEC2& pos1, 
                                                    nn::math::VEC2& pos2 )
{    
    NN_ASSERT( triIndex < m_TrianglesNum );

    pos0 = m_ShapePositionArray[ triIndex     ] + m_Pos;
    pos1 = m_ShapePositionArray[ triIndex + 1 ] + m_Pos;
    pos2 = m_ShapePositionArray[ triIndex + 2 ] + m_Pos;
}
