﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_ParticleSystemScene.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_PARTICLE_SYSTEM_SCENE_H_
#define GR_PARTICLE_SYSTEM_SCENE_H_

#include <nn/gr.h>
#include <nn/ulcd.h>
#include <nn/math.h>

#include "gr_PrimitiveVertexStream.h"

class IAllocator;

//------------------------------------------------------------------------------

enum ParticleSystemPattern
{   
    PARTICLE_EXPLOSION_PATTERN = 0,
    PARTICLE_WIND_PATTERN      = 1,
    MAX_PARTICLE_PATTERN_NUM   = 2
};

//------------------------------------------------------------------------------

// Structure for particle system
struct ParticleSystem
{
    // For vertex shader uniforms
    nn::math::Matrix44      centerMtx44;
    
    // For geometry shader uniforms
    nn::math::Matrix44      radiiMtx44;
    nn::math::Matrix44      colorMtx44;
    nn::math::VEC2          pointSizeVec2;
    nn::math::VEC2          viewportSizeVec2;
    f32                     time;
    f32                     speed;
    nn::math::VEC3          distanceAttenuationVec3;
    f32                     countMax;
    nn::math::VEC4          randSeedVec4;
    nn::math::Matrix44      aspectMtx44;
    nn::math::VEC4          randomCoreVec4;
    
    // Arguments used to advance the simulation
    f32                     timeSpan;
    s32                     frameSpan;
    f32                     deltaTime;

    void SetDisplay( const s32 display )
    {
        u32 width  = 0;
        u32 height = 0;

        if ( ( display == NN_GX_DISPLAY0 ) || ( display == NN_GX_DISPLAY0_EXT) )
        {
            width  = nn::gx::DISPLAY0_WIDTH;
            height = nn::gx::DISPLAY0_HEIGHT;
        }
        else if ( display == NN_GX_DISPLAY1 )
        {
            width  = nn::gx::DISPLAY1_WIDTH;
            height = nn::gx::DISPLAY1_HEIGHT;
        }
        else
        {
            NN_ASSERT( 0 );
        }

        viewportSizeVec2[ 0 ] = 1.0f / static_cast<f32>( width );
        viewportSizeVec2[ 1 ] = 1.0f / static_cast<f32>( height );
       
        pointSizeVec2[ 1 ]    = static_cast<f32>( height ) / 2.0f;

        distanceAttenuationVec3[ 2 ]  = 1.0f / static_cast<f32>( width * height );
    }
};

class ParticleSystemScene : public IScene
{
public:
    ParticleSystemScene( void );
    virtual ~ParticleSystemScene();
    
public:
    virtual bool Initialize( void );
    virtual void Finalize( void );
    virtual void Enter( void );
    virtual bool Calc( void );
    virtual void PreDraw( void );
    virtual GLuint DrawTop( void );
    virtual GLuint DrawBtm( void );
    virtual GLuint DrawExt( void );
    virtual void PostDraw( void );
    virtual void Clear( const u32 bufferBit );
    virtual nn::gr::CTR::FrameBuffer::ColorBuffer* GetGrColorBufferPointer( void );

private:
    void    UpdateScene_( void );
    void    RenderSceneGr_( const s32 display, 
        const nn::math::MTX44& proj, const nn::math::MTX34& view );

private:
    void InitializeParticleSystemArray( void );
    void InitializeTexture( void );

    bit32* MakeTextureCommand( bit32* command );
    bit32* MakeParticleSystemCommand( bit32* command,
        ParticleSystem* pParticleSystem,
        const nn::math::Matrix44 proj, const nn::math::Matrix44 modelView );
    bit32* MakeDrawCommand( bit32* command );

public:
    IAllocator*                     allocator;

    u8*                             shaderBinary;
    nn::gr::ShaderLite              shader;
    u32*                            shaderBuffer;

    nn::math::MTX34                 mtx;
    nn::math::MTX44                 projectionMtx44;
    nn::math::MTX34                 modelViewMtx34;
    nn::ulcd::StereoCamera          stereo;

    // Vertex shader uniforms
    nn::gr::BindSymbolVSFloat       bindSymbolProjectionMtx44;
    nn::gr::BindSymbolVSFloat       bindSymbolModelViewMtx44;
    nn::gr::BindSymbolVSFloat       bindSymbolCenterMtx44;
    nn::gr::BindSymbolVSFloat       bindSymbolRadiiMtx44;

    // Geometry shader uniforms
    nn::gr::BindSymbolGSFloat       bindSymbolColorMtx44;
    nn::gr::BindSymbolGSFloat       bindSymbolViewportVec4;
    nn::gr::BindSymbolGSFloat       bindSymbolPointSizeTimeSpeedVec4;
    nn::gr::BindSymbolGSFloat       bindSymbolDistanceAttenuationCountMaxVec4;
    nn::gr::BindSymbolGSFloat       bindSymbolRandSeedVec4;
    nn::gr::BindSymbolGSFloat       bindSymbolAspectMtx44;
    nn::gr::BindSymbolGSFloat       bindSymbolRandomCoreVec4;

    GLuint                          renderTarget;
    GLuint                          renderBuffer[ 2 ];

    uptr                            colorBufferVirtualAddress;
    uptr                            depthBufferVirtualAddress;

private:
    typedef void* (* AllocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, GLsizei size );
    typedef void  (* DeallocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, void *addr );

    AllocateGraphicsMemory          m_GraphicsMemoryAllocator;
    DeallocateGraphicsMemory        m_GraphicsMemoryDeallocator;

    nn::gr::Viewport                m_Viewport;
    nn::gr::Scissor                 m_Scissor;
    nn::gr::FrameBuffer             m_FrameBuffer;
    
    nn::gr::CTR::RenderState        m_RenderState;
    nn::gr::Combiner                m_Combiner;
    nn::gr::CTR::Texture            m_Texture;
    nn::gr::CTR::ProcedureTexture   m_ProcedureTexture;

    PrimitiveParticle*              m_pPrimitiveParticle;
    nn::gr::Vertex                  m_VertexParticle;
    nn::gr::Vertex::IndexStream     m_IndexStreamParticle;

    ParticleSystemPattern           m_ParticleSystemPattern;
    NN_PADDING3;
    ParticleSystem                  m_ParticleSystemArray[ MAX_PARTICLE_PATTERN_NUM ];
    u32                             m_Counter;
};

#endif // GR_PARTICLE_SYSTEM_SCENE_H_
