﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_ParticleSystemScene.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/fs.h>

#include "gr_Framework.h"
#include "gr_ParticleSystemScene.h"

//------------------------------------------------------------------------------
GfxAllocator * GrAllocator( MemoryMgr::AllocatorDevice *heap )
{
    return new GfxAllocator( heap );
}

//------------------------------------------------------------------------------
GraphicSystem::GfxAllocatorFactory  GetGraphicsMemoryMgrFunctor( void )
{
    return &GrAllocator;
}

//------------------------------------------------------------------------------
IScene * CreateScene( void )
{
    return new ParticleSystemScene();
}

//------------------------------------------------------------------------------
static const u32 LOOKUP_TABLE_ELEMENT_NUM       = 128;
static const u32 COLOR_LOOKUP_TABLE_ELEMENT_NUM = 256;
static const u32 PROCTEX_TEXTURE_SIZE           =   2;

//------------------------------------------------------------------------------
void InitializeBiasLUT( f32 *lutArray, f32 p )
{    
    const float dx = 2.0f / 127.0f;
    float x        = 0.0f;
    float aa       = 1.0f / p;
    float t        = 0.0f;

    for ( u32 index = 0; index < 128; index++ )
    {
        t = x;

        if ( t >= 1.0f )
        {
            lutArray[ index ] = 1.0f;
        }
        else
        {
            lutArray[ index ] = t / ( ( aa - 2.0f ) * ( 1.0f - t ) + 1.0f);
        }

        x += dx;
    }
}

//------------------------------------------------------------------------------
void InitializeColorLookUpTable( GLfloat* redLookUpTable, GLfloat* greenLookUpTable, 
                                 GLfloat* blueLookUpTable, GLfloat* alphaLookUpTable )
{
    std::memset( redLookUpTable,   0, COLOR_LOOKUP_TABLE_ELEMENT_NUM * sizeof( f32 ) );
    std::memset( greenLookUpTable, 0, COLOR_LOOKUP_TABLE_ELEMENT_NUM * sizeof( f32 ) );
    std::memset( blueLookUpTable,  0, COLOR_LOOKUP_TABLE_ELEMENT_NUM * sizeof( f32 ) );
    std::memset( alphaLookUpTable, 0, COLOR_LOOKUP_TABLE_ELEMENT_NUM * sizeof( f32 ) );

      redLookUpTable[ 0 ] = 0xff / 255.0f;
    greenLookUpTable[ 0 ] = 0xff / 255.0f;
     blueLookUpTable[ 0 ] = 0xff / 255.0f;
    alphaLookUpTable[ 0 ] = 0xff / 255.0f;

      redLookUpTable[ 1 ] = 0x00 / 255.0f;
    greenLookUpTable[ 1 ] = 0x00 / 255.0f;
     blueLookUpTable[ 1 ] = 0x00 / 255.0f;
    alphaLookUpTable[ 1 ] = 0x00 / 255.0f;
}

//------------------------------------------------------------------------------
ParticleSystemScene::ParticleSystemScene()
    : allocator( NULL ),
      shaderBinary( NULL ),
      shader(),
      mtx(), 
      projectionMtx44(),
      modelViewMtx34(),
      stereo(),
      bindSymbolProjectionMtx44(),
      bindSymbolModelViewMtx44(),
      bindSymbolCenterMtx44(),
      bindSymbolRadiiMtx44(),
      bindSymbolColorMtx44(),
      bindSymbolViewportVec4(),
      bindSymbolPointSizeTimeSpeedVec4(),
      bindSymbolDistanceAttenuationCountMaxVec4(),
      bindSymbolRandSeedVec4(),
      bindSymbolAspectMtx44(),
      bindSymbolRandomCoreVec4(),
      renderTarget( 0 ), 
      colorBufferVirtualAddress( NULL ),
      depthBufferVirtualAddress( NULL ),
      m_GraphicsMemoryAllocator(),
      m_GraphicsMemoryDeallocator(),
      m_Viewport(),
      m_Scissor(),
      m_FrameBuffer(),
      m_RenderState(),
      m_Combiner(),
      m_Texture(),
      m_ProcedureTexture(),
      m_pPrimitiveParticle( NULL ),
      m_VertexParticle(),
      m_IndexStreamParticle(),
      m_ParticleSystemPattern( PARTICLE_EXPLOSION_PATTERN ),
      m_Counter( 0 )
{
    for ( u32 index = 0; index < 2; index++)
    {
        renderBuffer[ index ] = 0;
    }
}

//------------------------------------------------------------------------------
ParticleSystemScene::~ParticleSystemScene()
{
}

//------------------------------------------------------------------------------
bool ParticleSystemScene::Initialize( void )
{
    bool is_success = false;

    NN_UNUSED_VAR( is_success );

    allocator = MEMMGR->GetDefaultAllocatorDevice();

    { 
        // Initialize the shader

        // Load shader binary
        nn::fs::FileReader shader_file_reader( L"rom:/shader.shbin" );
        u32 shader_file_size = static_cast<u32>( shader_file_reader.GetSize() );
        shaderBinary = static_cast<u8*>( allocator->Alloc( shader_file_size, 128 ) );
        shader_file_reader.Read( shaderBinary, shader_file_size );
        shader_file_reader.Finalize();

        // Get size of the buffer needed for a ShaderLite class from the shader binary
        u32 size = shader.GetRequiredBufferSize( shaderBinary );

        shaderBuffer = static_cast<u32*>( allocator->Alloc( size ) );

        // Specify shader binary and initialize
        // Enable the executable image at position 0 as the vertex shader
        // Enable the executable image at position 1 as the geometry shader
        shader.SetupBinary( shaderBinary, 0, 1, shaderBuffer );
        
        // Vertex shader uniforms
        is_success = shader.SearchBindSymbol( &bindSymbolProjectionMtx44, "uProjection" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolModelViewMtx44, "uModelView" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolCenterMtx44, "uCenter" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolRadiiMtx44, "uRadii" );
        NN_GR_ASSERT( is_success );


        // Geometry shader uniforms
        is_success = shader.SearchBindSymbol( &bindSymbolColorMtx44, "dmp_PartSys.color");
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolViewportVec4, "dmp_PartSys.viewport");
        NN_GR_ASSERT( is_success );
    
        is_success = 
            shader.SearchBindSymbol( &bindSymbolPointSizeTimeSpeedVec4, "dmp_PartSys.pointSize" );                            
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &bindSymbolDistanceAttenuationCountMaxVec4, 
                                     "dmp_PartSys.distanceAttenuation");
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolRandSeedVec4, "dmp_PartSys.randSeed");
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolAspectMtx44, "dmp_PartSys.aspect");
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolRandomCoreVec4, "dmp_PartSys.randomCore");
        NN_GR_ASSERT( is_success );
    }

    {         
        // Initialize primitives for particles
        u32 physicalAddr = 0;
        nn::gr::CTR::BindSymbolVSInput symbolPos;
                        
        m_pPrimitiveParticle = 
            static_cast< PrimitiveParticle* >( allocator->Alloc( sizeof( PrimitiveParticle ), 128 ) );
        m_pPrimitiveParticle->SetupVtxStream();

        // Search for "location"
        is_success = shader.SearchBindSymbol( &symbolPos, "attrCtrPointIndex" );
        NN_GR_ASSERT( is_success );

        // Vertex stream (Position)
        physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_pPrimitiveParticle->streamPos ) );
        NN_NULL_ASSERT( physicalAddr );
        m_VertexParticle.EnableAttrAsArray( symbolPos, physicalAddr, PICA_DATA_SIZE_1_FLOAT );
           
        // Index setting
        physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_pPrimitiveParticle->streamIdx ) );
        NN_NULL_ASSERT( physicalAddr );
        m_IndexStreamParticle.physicalAddr = physicalAddr;
        m_IndexStreamParticle.drawVtxNum   = m_pPrimitiveParticle->VERTEX_NUM;
        // Index is defined as u8.
        m_IndexStreamParticle.isUnsignedByte = true;
    }
    
    InitializeParticleSystemArray();
    InitializeTexture();
    
    // Get allocator and deallocator
    nngxGetAllocator( &m_GraphicsMemoryAllocator, &m_GraphicsMemoryDeallocator );

    {
        // Allocate color buffer  GL_RGBA8_OES
        renderBuffer[ 0 ] = 1;
        GLenum area = NN_GX_MEM_VRAMA;
        GLenum aim  = NN_GX_MEM_RENDERBUFFER;
        u32 colorBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        colorBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 0 ], colorBufferByteSize ) );

        // Allocate depth buffer  GL_DEPTH24_STENCIL8_EXT
        renderBuffer[ 1 ] = 2;
        area = NN_GX_MEM_VRAMB;
        u32 depthBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        depthBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 1 ], depthBufferByteSize ) );
    }

    {
        renderTarget = 1;

        // Initialize framebuffer settings
        m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

        // Initialize color buffer settings
        m_FrameBuffer.colorBuffer.virtualAddr         = colorBufferVirtualAddress;
        m_FrameBuffer.colorBuffer.format              = PICA_DATA_COLOR_RGBA8_OES;
        m_FrameBuffer.colorBuffer.blockSize           = nn::gr::FrameBuffer::BLOCK_SIZE8;
        m_FrameBuffer.colorBuffer.width               = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.colorBuffer.height              = nn::gx::DISPLAY0_HEIGHT;  
        m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;
       
        // Initialize depth buffer settings
        m_FrameBuffer.depthStencilBuffer.virtualAddr  = depthBufferVirtualAddress;
        m_FrameBuffer.depthStencilBuffer.format       = PICA_DATA_DEPTH24_STENCIL8_EXT;
        m_FrameBuffer.depthStencilBuffer.width        = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.depthStencilBuffer.height       = nn::gx::DISPLAY0_HEIGHT;
        m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
        m_FrameBuffer.depthStencilBuffer.clearStencil = 0;
    }

    m_Scissor.SetBufferSize( nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT );


    // Render state settings
    m_RenderState.wBuffer.wScale = 0.0f;
    
    // glEnable(GL_BLEND);
    // glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    m_RenderState.blend.isEnable = true;
    m_RenderState.blend.srcAlpha   = PICA_DATA_BLEND_FUNC_ONE;

    // glDisable(GL_DEPTH_TEST);
    m_RenderState.depthTest.isEnable      = false;
    m_RenderState.depthTest.isEnableWrite = false;

    // glDisable(GL_CULL_FACE);
    m_RenderState.cullingTest.isEnable    = false;

    {
        // Texture combiner settings
        m_Combiner.stage[ 0 ].rgb.combine         = PICA_DATA_TEX_ENV_COMBINE_MODULATE;
        m_Combiner.stage[ 0 ].rgb.source[ 0 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE3;
        m_Combiner.stage[ 0 ].rgb.source[ 1 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].rgb.source[ 2 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;

        m_Combiner.stage[ 0 ].rgb.operand[ 0 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
        m_Combiner.stage[ 0 ].rgb.operand[ 1 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
        m_Combiner.stage[ 0 ].rgb.operand[ 2 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
        
        m_Combiner.stage[ 0 ].alpha.combine     = PICA_DATA_TEX_ENV_COMBINE_MODULATE;

        m_Combiner.stage[ 0 ].alpha.source[ 0 ] = PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE3;
        m_Combiner.stage[ 0 ].alpha.source[ 1 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].alpha.source[ 2 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
            
        m_Combiner.stage[ 0 ].alpha.operand[ 0 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
        m_Combiner.stage[ 0 ].alpha.operand[ 1 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
        m_Combiner.stage[ 0 ].alpha.operand[ 2 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;

        for (u32 stageIndex = 1; stageIndex < 6; stageIndex++)
        {
            m_Combiner.stage[ stageIndex ].rgb.combine   = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
            m_Combiner.stage[ stageIndex ].alpha.combine = PICA_DATA_TEX_ENV_COMBINE_REPLACE;

            m_Combiner.stage[ stageIndex ].rgb.scale     = PICA_DATA_TEX_ENV_SCALE_1;

            for ( u32 sourceIndex = 0; sourceIndex < 3; sourceIndex++)
            {
                m_Combiner.stage[ stageIndex ].rgb.source[ sourceIndex ]    = PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS;
                m_Combiner.stage[ stageIndex ].rgb.operand[ sourceIndex ]   = PICA_DATA_OPE_RGB_SRC_COLOR;

                m_Combiner.stage[ stageIndex ].alpha.source[ sourceIndex ]  = PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS;
                m_Combiner.stage[ stageIndex ].alpha.operand[ sourceIndex ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
            }
        }
    }

    // Initialize using a disable command
    {
        u32* command = NULL;
        nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
            reinterpret_cast<GLint*>( &command ) );
        const u32* start_addr = command;

        command = nn::gr::CTR::MakeDisableAllCommand( command );

        nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );
    }

    return true;
}

void ParticleSystemScene::InitializeTexture( void )
{   
   // Texture settings
    m_Texture.unit0.texType  = PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_FALSE;
    m_Texture.unit1.texType  = PICA_DATA_TEXTURE1_SAMPLER_TYPE_FALSE;
    m_Texture.unit2.texType  = PICA_DATA_TEXTURE2_SAMPLER_TYPE_FALSE;
    m_Texture.unit3.texType  = PICA_DATA_TEXTURE3_SAMPLER_TYPE_TEXTURE_PROCEDURAL_DMP;
    m_Texture.unit3.texCoord = PICA_DATA_TEXTURE3_TEXCOORD_TEXTURE2;


    // Procedural texture settings
    nn::gr::CTR::ProcedureTexture& procedureTexture = m_ProcedureTexture;
    
    procedureTexture.texBias         = 0.5f;
    procedureTexture.texOffset       = 0;

    procedureTexture.alphaSeparate   = false;
   
    procedureTexture.rgbMap          = PICA_DATA_PROCTEX_MAX_DMP;
    procedureTexture.alphaMap        = PICA_DATA_PROCTEX_MAX_DMP;

    procedureTexture.clampU          = PICA_DATA_PROCTEX_CLAMP_TO_EDGE;
    procedureTexture.clampV          = PICA_DATA_PROCTEX_CLAMP_TO_EDGE;

    procedureTexture.shiftU          = PICA_DATA_PROCTEX_NONE_DMP;
    procedureTexture.shiftV          = PICA_DATA_PROCTEX_NONE_DMP;

    procedureTexture.minFilter       = PICA_DATA_PROCTEX_LINEAR;


    // Do not use noise
    procedureTexture.isEnableNoise   = false;
    procedureTexture.noiseUAmplitude = 0.0f;
    procedureTexture.noiseVAmplitude = 0.0f;

    procedureTexture.noiseUAmplitude = 0.0f;
    procedureTexture.noiseVAmplitude = 0.0f;

    procedureTexture.noiseUFrequency = 0.0f;
    procedureTexture.noiseVFrequency = 0.0f;

    procedureTexture.noiseUPhase     = 0.0f;
    procedureTexture.noiseVPhase     = 0.0f;


    // Set lookup tables
    f32    noiseLookUpTable[ LOOKUP_TABLE_ELEMENT_NUM ];
    f32   rgbMapLookUpTable[ LOOKUP_TABLE_ELEMENT_NUM ];
    f32 alphaMapLookUpTable[ LOOKUP_TABLE_ELEMENT_NUM ];
    
    std::memset( noiseLookUpTable, 0, LOOKUP_TABLE_ELEMENT_NUM * sizeof( f32 ) );
    InitializeBiasLUT( rgbMapLookUpTable, 0.8f );
    std::memset( alphaMapLookUpTable, 0, LOOKUP_TABLE_ELEMENT_NUM * sizeof( f32 ) );

    procedureTexture.SetNoiseLookUpTable( noiseLookUpTable );
    procedureTexture.SetRgbMapLookUpTable( rgbMapLookUpTable );
    procedureTexture.SetAlphaMapLookUpTable( alphaMapLookUpTable );


    // Set lookup tables for color    
    f32   redLookUpTable[ COLOR_LOOKUP_TABLE_ELEMENT_NUM ];
    f32 greenLookUpTable[ COLOR_LOOKUP_TABLE_ELEMENT_NUM ];
    f32  blueLookUpTable[ COLOR_LOOKUP_TABLE_ELEMENT_NUM ];
    f32 alphaLookUpTable[ COLOR_LOOKUP_TABLE_ELEMENT_NUM ];

    InitializeColorLookUpTable( redLookUpTable, greenLookUpTable, 
        blueLookUpTable, alphaLookUpTable );

    procedureTexture.SetRedLookUpTable(   redLookUpTable );
    procedureTexture.SetGreenLookUpTable( greenLookUpTable );
    procedureTexture.SetBlueLookUpTable(  blueLookUpTable );
    procedureTexture.SetAlphaLookUpTable( alphaLookUpTable );

    procedureTexture.texWidth = PROCTEX_TEXTURE_SIZE;
}

void ParticleSystemScene::InitializeParticleSystemArray( void )
{
    {
        ParticleSystem* pParticleSystem = &m_ParticleSystemArray[ PARTICLE_EXPLOSION_PATTERN ];

        // Control point 0
        pParticleSystem->centerMtx44.m[ 0 ][ 0 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 0 ][ 1 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 0 ][ 2 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 0 ][ 3 ] = 1.0f;

        // Control point 1
        pParticleSystem->centerMtx44.m[ 1 ][ 0 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 1 ][ 1 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 1 ][ 2 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 1 ][ 3 ] = 1.0f;

        // Control point 2
        pParticleSystem->centerMtx44.m[ 2 ][ 0 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 2 ][ 1 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 2 ][ 2 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 2 ][ 3 ] = 1.0f;

        // Control point 3
        pParticleSystem->centerMtx44.m[ 3 ][ 0 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 3 ][ 1 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 3 ][ 2 ] = 0.0f;
        pParticleSystem->centerMtx44.m[ 3 ][ 3 ] = 1.0f;

        // Control point 0 color
        pParticleSystem->colorMtx44.m[ 0 ][ 0 ] = 1.0f;
        pParticleSystem->colorMtx44.m[ 0 ][ 1 ] = 1.0f;
        pParticleSystem->colorMtx44.m[ 0 ][ 2 ] = 1.0f;
        pParticleSystem->colorMtx44.m[ 0 ][ 3 ] = 1.0f;

        // Control point 1 color
        pParticleSystem->colorMtx44.m[ 1 ][ 0 ] = 1.0f;
        pParticleSystem->colorMtx44.m[ 1 ][ 1 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 1 ][ 2 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 1 ][ 3 ] = 1.0f;
        
        // Control point 2 color
        pParticleSystem->colorMtx44.m[ 2 ][ 0 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 2 ][ 1 ] = 1.0f;
        pParticleSystem->colorMtx44.m[ 2 ][ 2 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 2 ][ 3 ] = 1.0f;

        // Control point 3 color
        pParticleSystem->colorMtx44.m[ 3 ][ 0 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 3 ][ 1 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 3 ][ 2 ] = 1.0f;
        pParticleSystem->colorMtx44.m[ 3 ][ 3 ] = 0.0f;
        
        // Bounding box 0
        pParticleSystem->radiiMtx44.m[ 0 ][ 0 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 0 ][ 1 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 0 ][ 2 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 0 ][ 3 ] = 1.0f;

        // Bounding box 1
        pParticleSystem->radiiMtx44.m[ 1 ][ 0 ] = 10.0f;
        pParticleSystem->radiiMtx44.m[ 1 ][ 1 ] = 10.0f;
        pParticleSystem->radiiMtx44.m[ 1 ][ 2 ] = 10.0f;
        pParticleSystem->radiiMtx44.m[ 1 ][ 3 ] = 1.0f;
        
        // Bounding box 2
        pParticleSystem->radiiMtx44.m[ 2 ][ 0 ] = 20.0f;
        pParticleSystem->radiiMtx44.m[ 2 ][ 1 ] = 20.0f;
        pParticleSystem->radiiMtx44.m[ 2 ][ 2 ] = 20.0f;
        pParticleSystem->radiiMtx44.m[ 2 ][ 3 ] = 1.0f;

        // Bounding box 3
        pParticleSystem->radiiMtx44.m[ 3 ][ 0 ] = 50.0f;
        pParticleSystem->radiiMtx44.m[ 3 ][ 1 ] = 50.0f;
        pParticleSystem->radiiMtx44.m[ 3 ][ 2 ] = 50.0f;
        pParticleSystem->radiiMtx44.m[ 3 ][ 3 ] = 1.0f;

        // +Particle aspect 0
        pParticleSystem->aspectMtx44.m[ 0 ][ 0 ] = 1.0f;
        pParticleSystem->aspectMtx44.m[ 0 ][ 1 ] = 0.0f;
        pParticleSystem->aspectMtx44.m[ 0 ][ 2 ] = 1.0f;
        pParticleSystem->aspectMtx44.m[ 0 ][ 3 ] = 1.0f;

        // +Particle aspect 1
        pParticleSystem->aspectMtx44.m[ 1 ][ 0 ] = 2.0f;
        pParticleSystem->aspectMtx44.m[ 1 ][ 1 ] = 0.0f;
        pParticleSystem->aspectMtx44.m[ 1 ][ 2 ] = 1.0f;
        pParticleSystem->aspectMtx44.m[ 1 ][ 3 ] = 1.0f;

        // +Particle aspect 2
        pParticleSystem->aspectMtx44.m[ 2 ][ 0 ] = 30.0f;
        pParticleSystem->aspectMtx44.m[ 2 ][ 1 ] = 0.0f;
        pParticleSystem->aspectMtx44.m[ 2 ][ 2 ] = 1.0f;
        pParticleSystem->aspectMtx44.m[ 2 ][ 3 ] = 0.0f;

        // +Particle aspect 3
        pParticleSystem->aspectMtx44.m[ 3 ][ 0 ] = 50.0f;
        pParticleSystem->aspectMtx44.m[ 3 ][ 1 ] = 0.0f;
        pParticleSystem->aspectMtx44.m[ 3 ][ 2 ] = 1.0f;
        pParticleSystem->aspectMtx44.m[ 3 ][ 3 ] = 0.0f;

        pParticleSystem->pointSizeVec2[ 0 ] = 1.0f;

        pParticleSystem->randSeedVec4[ 0 ] = 15.8794f;
        pParticleSystem->randSeedVec4[ 1 ] = 2.564f  + pParticleSystem->randSeedVec4[ 0 ];
        pParticleSystem->randSeedVec4[ 2 ] = 1.15f   + pParticleSystem->randSeedVec4[ 0 ];
        pParticleSystem->randSeedVec4[ 3 ] = 0.9875f + pParticleSystem->randSeedVec4[ 0 ];


        // emitter particle_speed
        pParticleSystem->distanceAttenuationVec3[ 0 ] = 0.0f;
        pParticleSystem->distanceAttenuationVec3[ 1 ] = 0.0f;

        // Core value for random
        // Algorithm is based on pseudo-random number generators: X(n+1) = (a * X n + b ) mod m
        // a
        pParticleSystem->randomCoreVec4[ 0 ] = 147.0f;
        // b
        pParticleSystem->randomCoreVec4[ 1 ] = 17.0f;
        // m
        pParticleSystem->randomCoreVec4[ 2 ] = 256847.0f;
        // 1/m 
        pParticleSystem->randomCoreVec4[ 3 ] = 1.0f / ( pParticleSystem->randomCoreVec4[ 2 ] );

        pParticleSystem->time  = 0.0f;

        //  emitter particle_speed 
        pParticleSystem->speed = 0.55f;

        // particle number
        pParticleSystem->countMax = 100.0f;

        pParticleSystem->timeSpan       = 10.0f;
        pParticleSystem->frameSpan      = 512;
        pParticleSystem->deltaTime      = pParticleSystem->timeSpan / static_cast<f32>( pParticleSystem->frameSpan );
        
        pParticleSystem->SetDisplay( NN_GX_DISPLAY0 );
    }

    {
        ParticleSystem* pParticleSystem = &m_ParticleSystemArray[ PARTICLE_WIND_PATTERN ];

        // Control point 0
        pParticleSystem->centerMtx44.m[ 0 ][ 0 ] =   0.0f;
        pParticleSystem->centerMtx44.m[ 0 ][ 1 ] = -10.0f;
        pParticleSystem->centerMtx44.m[ 0 ][ 2 ] =   0.0f;
        pParticleSystem->centerMtx44.m[ 0 ][ 3 ] =   1.0f;

        // Control point 1
        pParticleSystem->centerMtx44.m[ 1 ][ 0 ] = -10.0f;
        pParticleSystem->centerMtx44.m[ 1 ][ 1 ] = -5.0f;
        pParticleSystem->centerMtx44.m[ 1 ][ 2 ] =  0.0f;
        pParticleSystem->centerMtx44.m[ 1 ][ 3 ] =  1.0f;

        // Control point 2
        pParticleSystem->centerMtx44.m[ 2 ][ 0 ] = 10.0f;
        pParticleSystem->centerMtx44.m[ 2 ][ 1 ] =  5.0f;
        pParticleSystem->centerMtx44.m[ 2 ][ 2 ] =  0.0f;
        pParticleSystem->centerMtx44.m[ 2 ][ 3 ] =  1.0f;

        // Control point 3
        pParticleSystem->centerMtx44.m[ 3 ][ 0 ] =  0.0f;
        pParticleSystem->centerMtx44.m[ 3 ][ 1 ] = 10.0f;
        pParticleSystem->centerMtx44.m[ 3 ][ 2 ] =  0.0f;
        pParticleSystem->centerMtx44.m[ 3 ][ 3 ] =  1.0f;

        // Bounding box 0
        pParticleSystem->radiiMtx44.m[ 0 ][ 0 ] = 0.0f;
        pParticleSystem->radiiMtx44.m[ 0 ][ 1 ] = 0.0f;
        pParticleSystem->radiiMtx44.m[ 0 ][ 2 ] = 0.0f;
        pParticleSystem->radiiMtx44.m[ 0 ][ 3 ] = 0.0f;

        // Bounding box 1
        pParticleSystem->radiiMtx44.m[ 1 ][ 0 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 1 ][ 1 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 1 ][ 2 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 1 ][ 3 ] = 0.0f;

        // Bounding box 2
        pParticleSystem->radiiMtx44.m[ 2 ][ 0 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 2 ][ 1 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 2 ][ 2 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 2 ][ 3 ] = 1.0f;

        // Bounding box 3
        pParticleSystem->radiiMtx44.m[ 3 ][ 0 ] = 10.0f;
        pParticleSystem->radiiMtx44.m[ 3 ][ 1 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 3 ][ 2 ] = 1.0f;
        pParticleSystem->radiiMtx44.m[ 3 ][ 3 ] = 1.0f;

        // Control point 0 color
        pParticleSystem->colorMtx44.m[ 0 ][ 0 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 0 ][ 1 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 0 ][ 2 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 0 ][ 3 ] = 1.0f;

        // Control point 1 color
        pParticleSystem->colorMtx44.m[ 1 ][ 0 ] = 1.0f;
        pParticleSystem->colorMtx44.m[ 1 ][ 1 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 1 ][ 2 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 1 ][ 3 ] = 1.0f;
        
        // Control point 2 color
        pParticleSystem->colorMtx44.m[ 2 ][ 0 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 2 ][ 1 ] = 1.0f;
        pParticleSystem->colorMtx44.m[ 2 ][ 2 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 2 ][ 3 ] = 0.5f;

        // Control point 3 color
        pParticleSystem->colorMtx44.m[ 3 ][ 0 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 3 ][ 1 ] = 0.0f;
        pParticleSystem->colorMtx44.m[ 3 ][ 2 ] = 1.0f;
        pParticleSystem->colorMtx44.m[ 3 ][ 3 ] = 0.0f;

        pParticleSystem->pointSizeVec2[ 0 ] = 1.0f;
        
        // Application only
        pParticleSystem->time  = 0.0f;

        // emitter particle_speed  
        pParticleSystem->speed = 1.0f;
              
        pParticleSystem->distanceAttenuationVec3[ 0 ] = 0.0f;
        pParticleSystem->distanceAttenuationVec3[ 1 ] = 0.0f;
        
        // particle number
        pParticleSystem->countMax = 20.0f;
    
        pParticleSystem->randSeedVec4[ 0 ] = 45.0f;
        pParticleSystem->randSeedVec4[ 1 ] = 82.564f   + pParticleSystem->randSeedVec4[ 0 ];
        pParticleSystem->randSeedVec4[ 2 ] = 18.15f    + pParticleSystem->randSeedVec4[ 0 ];
        pParticleSystem->randSeedVec4[ 3 ] = 819.9875f + pParticleSystem->randSeedVec4[ 0 ];

        // +Particle aspect 0
        // Rotation disabled: Size, unused, unused, unused/alpha
        // Rotation enabled : Size,  Angle, Radius, unused/alpha
        pParticleSystem->aspectMtx44.m[ 0 ][ 0 ] = 10.0f;
        pParticleSystem->aspectMtx44.m[ 0 ][ 1 ] = 0.0f;
        pParticleSystem->aspectMtx44.m[ 0 ][ 2 ] = 0.9f;
        pParticleSystem->aspectMtx44.m[ 0 ][ 3 ] = 1.0f;

        // +Particle aspect 1
        pParticleSystem->aspectMtx44.m[ 1 ][ 0 ] = 30.0f;
        pParticleSystem->aspectMtx44.m[ 1 ][ 1 ] = 0.0f;
        pParticleSystem->aspectMtx44.m[ 1 ][ 2 ] = 0.9f;
        pParticleSystem->aspectMtx44.m[ 1 ][ 3 ] = 1.0f;

        // +Particle aspect 2
        pParticleSystem->aspectMtx44.m[ 2 ][ 0 ] = 30.0f;
        pParticleSystem->aspectMtx44.m[ 2 ][ 1 ] = 0.0f;
        pParticleSystem->aspectMtx44.m[ 2 ][ 2 ] = 0.9f;
        pParticleSystem->aspectMtx44.m[ 2 ][ 3 ] = 0.0f;

        // +Particle aspect 3
        pParticleSystem->aspectMtx44.m[ 3 ][ 0 ] = 10.0f;
        pParticleSystem->aspectMtx44.m[ 3 ][ 1 ] = 0.0f;
        pParticleSystem->aspectMtx44.m[ 3 ][ 2 ] = 0.9f;
        pParticleSystem->aspectMtx44.m[ 3 ][ 3 ] = 0.0f;

        // Core value for random
        // Algorithm is based on pseudo-random number generators: X(n+1) = (a*Xn+b) mod m
        // a
        pParticleSystem->randomCoreVec4[ 0 ] = 17.0f;
        // b
        pParticleSystem->randomCoreVec4[ 1 ] = 37.0f;
         // m 
        pParticleSystem->randomCoreVec4[ 2 ] = 65535.0f;
        // 1/m 
        pParticleSystem->randomCoreVec4[ 3 ] = 1.0f / (pParticleSystem->randomCoreVec4[ 2 ]);
                
        pParticleSystem->timeSpan       = 2.0f;
        pParticleSystem->frameSpan      = 256;
        pParticleSystem->deltaTime      = pParticleSystem->timeSpan / static_cast<f32>( pParticleSystem->frameSpan );

        pParticleSystem->SetDisplay( NN_GX_DISPLAY1 );
    }
}

//------------------------------------------------------------------------------
void ParticleSystemScene::Finalize( void )
{
    m_VertexParticle.DisableAll();

    allocator->Free( m_pPrimitiveParticle );

    allocator->Free( shaderBinary );
    allocator->Free( shaderBuffer );

    GLenum area = NN_GX_MEM_VRAMA;
    GLenum aim  = NN_GX_MEM_RENDERBUFFER;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 0 ], reinterpret_cast<void*>( colorBufferVirtualAddress ) );
    colorBufferVirtualAddress = 0;

    area = NN_GX_MEM_VRAMB;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 1 ], reinterpret_cast<void*>( depthBufferVirtualAddress ) );
    depthBufferVirtualAddress = 0;
}

//------------------------------------------------------------------------------
void ParticleSystemScene::Enter( void )
{
    stereo.Initialize();
    nngxSetDisplayMode( NN_GX_DISPLAYMODE_STEREO );
    Calc();
}

//------------------------------------------------------------------------------
bool ParticleSystemScene::Calc( void )
{    
    UpdateScene_();

    m_Counter++;

    return true;
}

//------------------------------------------------------------------------------
void ParticleSystemScene::PreDraw( void )
{
}

//------------------------------------------------------------------------------
GLuint ParticleSystemScene::DrawTop( void )
{
    f32 width  = nn::gx::DISPLAY0_WIDTH;
    f32 height = nn::gx::DISPLAY0_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );
      
    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;
    
    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );
 
    { 
        // The camera matrix
        // f32 r = 60.f;
        modelViewMtx34.SetupLookAt(
            // camPos
            nn::math::VEC3( 0.0f, 0.0f, 22.0f ),
            // camUp
            nn::math::VEC3( 0.0f, 1.0f, 0.0f ),
            // target
            nn::math::VEC3( 0.0f, 0.0f, 0.0f ) );
    }

    // Projection matrix    
    projectionMtx44.SetupPerspective(
        0.66f,          // fovy
        height / width, // aspect
        0.1f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR,
        15.f, 1.f, nn::math::PIVOT_UPSIDE_TO_TOP );
    
    RenderSceneGr_( NN_GX_DISPLAY0, projL, viewL );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint ParticleSystemScene::DrawExt( void )
{
    f32 width  = nn::gx::DISPLAY0_WIDTH; 
    f32 height = nn::gx::DISPLAY0_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );

    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f, nn::math::PIVOT_UPSIDE_TO_TOP);
    
    RenderSceneGr_( NN_GX_DISPLAY0_EXT, projR, viewR );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint ParticleSystemScene::DrawBtm( void )
{
    f32 width  = nn::gx::DISPLAY1_WIDTH; 
    f32 height = nn::gx::DISPLAY1_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );
 
    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.0f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );
       
    projectionMtx44.SetupPerspective(
        0.66f,          // fovy
        height / width, // aspect
        0.1f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );
            
    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f );

    RenderSceneGr_( NN_GX_DISPLAY1, projL, viewL );

    return renderTarget;
}

//------------------------------------------------------------------------------
void ParticleSystemScene::PostDraw( void )
{
}

//------------------------------------------------------------------------------
nn::gr::CTR::FrameBuffer::ColorBuffer* ParticleSystemScene::GetGrColorBufferPointer( void )
{    
    return & m_FrameBuffer.colorBuffer;
}


//------------------------------------------------------------------------------
void ParticleSystemScene::Clear( const u32 bufferBit )
{
    u32* command = 0;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
        reinterpret_cast<GLint*>( &command ) );

    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Framebuffer settings
    command = m_FrameBuffer.MakeCommand( command, bufferBit );

    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );

    // Add a load complete command at the end of the 3D command buffer
    bool isAddSplitDrawCmdlist = true;

    // Generate a request to clear the color buffer and depth buffer
    m_FrameBuffer.MakeClearRequest( bufferBit, isAddSplitDrawCmdlist );
}

//------------------------------------------------------------------------------
void ParticleSystemScene::UpdateScene_( void )
{
    for ( u32 index = 0; index < MAX_PARTICLE_PATTERN_NUM; index++)
    {
        ParticleSystem* pParticleSystem = &m_ParticleSystemArray[ index ];
        pParticleSystem->time          += pParticleSystem->deltaTime;
    }
}

//------------------------------------------------------------------------------

bit32* ParticleSystemScene::MakeTextureCommand( bit32* command )
{
    // Generate 3 dummy commands to 0x80    
    const bool isAddDummyCommand = true;

    command = m_Texture.MakeCommand( command, isAddDummyCommand );
    command = m_ProcedureTexture.MakeAllCommand( command, (! isAddDummyCommand ) );

    return command;
}

//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void ParticleSystemScene::RenderSceneGr_( const s32 display,
                                          const nn::math::MTX44& proj, const nn::math::MTX34& view )
{
    // Get a pointer to the render command buffer
    u32* command = NULL;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );

    // Get the start address (so that the pointer can be advanced at the end)
    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Shader binary setting
    command = shader.MakeFullCommand( command );
   
    // Render state settings
    command = m_RenderState.MakeCommand( command );

    // Texture combiner settings
    command = m_Combiner.MakeCommand( command );

    // Texture settings
    command = MakeTextureCommand( command );

    if ( ( display == NN_GX_DISPLAY0 ) || ( display == NN_GX_DISPLAY0_EXT ) )
    {
        m_ParticleSystemPattern = PARTICLE_EXPLOSION_PATTERN;
    }
    else if ( display == NN_GX_DISPLAY1 )
    {
        m_ParticleSystemPattern = PARTICLE_WIND_PATTERN;
    }
    else
    {
        NN_ASSERT( 0 );
    }

    // Particle system settings
    ParticleSystem* pParticleSystem = &m_ParticleSystemArray[ static_cast<u32>( m_ParticleSystemPattern ) ];
    pParticleSystem->SetDisplay( display );
    command = MakeParticleSystemCommand( command, pParticleSystem, proj, static_cast<nn::math::Matrix44>(view) );
    
    // Render kick
    command = MakeDrawCommand( command );

    // Advance the pointer to the render command buffer
    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );
}

//------------------------------------------------------------------------------

bit32* ParticleSystemScene::MakeParticleSystemCommand( bit32* command, 
                                                       ParticleSystem* pParticleSystem,
                                                       const nn::math::Matrix44 proj,
                                                       const nn::math::Matrix44 modelView )
{   
    nn::math::VEC4 vec4;

    // Sets the uniform of the vertex shader
    command = bindSymbolProjectionMtx44.MakeUniformCommand( command, proj );
    command = bindSymbolModelViewMtx44.MakeUniformCommand( command, modelView );

    command = bindSymbolCenterMtx44.MakeUniformCommand( command, pParticleSystem->centerMtx44 ); 
    command = bindSymbolRadiiMtx44.MakeUniformCommand( command, pParticleSystem->radiiMtx44 );

    // Sets the shader uniform of the geometry shader    
    command = bindSymbolColorMtx44.MakeUniformCommand( command, pParticleSystem->colorMtx44 );

    vec4[ 0 ] = pParticleSystem->viewportSizeVec2[ 0 ];
    vec4[ 1 ] = pParticleSystem->viewportSizeVec2[ 1 ];
    vec4[ 2 ] = 0.0f;
    vec4[ 3 ] = 0.0f;
    command   = bindSymbolViewportVec4.MakeUniformCommand( command, vec4 );

    vec4[ 0 ] = pParticleSystem->pointSizeVec2[ 0 ];
    vec4[ 1 ] = pParticleSystem->pointSizeVec2[ 1 ];
    vec4[ 2 ] = pParticleSystem->time;
    vec4[ 3 ] = pParticleSystem->speed; 
    command   = bindSymbolPointSizeTimeSpeedVec4.MakeUniformCommand( command, vec4 );

    vec4[ 0 ] = pParticleSystem->distanceAttenuationVec3[ 0 ];
    vec4[ 1 ] = pParticleSystem->distanceAttenuationVec3[ 1 ];
    vec4[ 2 ] = pParticleSystem->distanceAttenuationVec3[ 2 ];
    vec4[ 3 ] = pParticleSystem->countMax;
    command   = bindSymbolDistanceAttenuationCountMaxVec4.MakeUniformCommand( command, vec4 );

    command   = bindSymbolRandSeedVec4.MakeUniformCommand( command, pParticleSystem->randSeedVec4 );
    command   = bindSymbolAspectMtx44.MakeUniformCommand( command, pParticleSystem->aspectMtx44 );
    command   = bindSymbolRandomCoreVec4.MakeUniformCommand( command, pParticleSystem->randomCoreVec4 );

    return command;
}

//------------------------------------------------------------------------------

bit32* ParticleSystemScene::MakeDrawCommand( bit32* command )
{
    // Disable all vertex attributes before starting to use vertices
    command = nn::gr::Vertex::MakeDisableCommand( command );

    // glBindBuffer( GL_ARRAY_BUFFER )
    // glBindBuffer( GL_ELEMENT_ARRAY_BUFFER ) + glDrawElements()
    command = m_VertexParticle.MakeEnableAttrCommand( command );
    
    // glDrawElements( GL_GEOMETRY_PRIMITIVE_DMP, 4, GL_UNSIGNED_SHORT, 0 );
    command = m_VertexParticle.MakeDrawCommand( command, m_IndexStreamParticle );

    // Disable all vertex attributes when finished using vertex
    command = nn::gr::CTR::Vertex::MakeDisableCommand( command );

    return command;
}

