/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_PointSpriteScene.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/fs.h>
#include <nn/tpl.h>

#include "gr_Framework.h"
#include "gr_PointSpriteScene.h"

namespace
{   
    // Texture package library
    wchar_t* s_TexturePackageName = L"rom:/knight.ctpk";
    const int s_TexNum            = 1;

    // Texture
    const u32 s_TextureIndex  = 0;
    const char* s_TextureName = "knight.tga";
    
    const char* s_TextureNames[ s_TexNum ] =
    {
        s_TextureName
    };

    nn::tpl::CTR::CtrTextureInfo s_TextureInfos[ s_TexNum ];

    u8* s_TextureAddr[ s_TexNum ];
}

PicaDataTextureFormat GetPicaDataTextureFormat( const nn::tpl::CTR::CtrTexFormat ctrTextureFormat )
{
    if ( ctrTextureFormat == nn::tpl::CTR::TEXFMT_RGB888 )
    {
        return PICA_DATA_TEXTURE_FORMAT_RGB_UNSIGNED_BYTE;
    }
    else
    {
        NN_LOG("Not implemented. \n");
        NN_ASSERT( 0 );

        return PICA_DATA_TEXTURE_FORMAT_RGB_UNSIGNED_BYTE;
    }
}

//------------------------------------------------------------------------------
GfxAllocator * GrAllocator( MemoryMgr::AllocatorDevice *heap )
{
    return new GfxAllocator( heap );
}

//------------------------------------------------------------------------------
GraphicSystem::GfxAllocatorFactory  GetGraphicsMemoryMgrFunctor( void )
{
    return &GrAllocator;
}

//------------------------------------------------------------------------------
IScene * CreateScene( void )
{
    return new PointSpriteScene();
}

//------------------------------------------------------------------------------
PointSpriteScene::PointSpriteScene()
    : allocator( NULL ),
      shaderBinary( NULL ),
      shader(),
      mtx(), 
      projectionMtx44(),
      modelViewMtx34(),
      stereo(),
      bindSymbolProjectionMtx44(),
      bindSymbolModelViewMtx44(),
      twnBindSymbolViewPortVec2(),
      twnBindSymbolDistanceAttenuationBool(),
      renderTarget( 0 ), 
      colorBufferVirtualAddress( NULL ),
      depthBufferVirtualAddress( NULL ),
      m_GraphicsMemoryAllocator(),
      m_GraphicsMemoryDeallocator(),
      m_Viewport(),
      m_Scissor(),
      m_FrameBuffer(),
      m_RenderState(),
      m_Combiner(),
      m_Texture(),
      m_pPrimitivePoint( NULL ),
      m_Vertex(),
      m_IndexStream(),
      m_Counter( 0 )
{
    for ( u32 index = 0; index < 2; index++)
    {
        renderBuffer[ index ] = 0;
    }
}

//------------------------------------------------------------------------------

PointSpriteScene::~PointSpriteScene()
{
}

//------------------------------------------------------------------------------
bool PointSpriteScene::Initialize( void )
{
    bool is_success = false;

    NN_UNUSED_VAR( is_success );

    allocator = MEMMGR->GetDefaultAllocatorDevice();

    { 
        // Initialize the shader

        // Load shader binary
        nn::fs::FileReader shader_file_reader( L"rom:/shader.shbin" );
        u32 shader_file_size = static_cast<u32>( shader_file_reader.GetSize() );
        shaderBinary = static_cast<u8*>( allocator->Alloc( shader_file_size, 128 ) );
        shader_file_reader.Read( shaderBinary, shader_file_size );
        shader_file_reader.Finalize();

        u32 size = shader.GetRequiredBufferSize( shaderBinary );

        shaderBuffer = static_cast<u32*>(allocator->Alloc( size ));

        // Specify shader binary and initialize
        // Enable the executable image at position 0 as the vertex shader
        // Enable the executable image at position 1 as the geometry shader
        shader.SetupBinary( shaderBinary, 0, 1, shaderBuffer );
        
        // Vertex shader uniforms
        is_success = 
            shader.SearchBindSymbol( &bindSymbolProjectionMtx44,
            "uProjection" );
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &bindSymbolModelViewMtx44,
            "uModelView" );
        NN_GR_ASSERT( is_success );

        // Geometry shader uniforms
        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolViewPortVec2,
            "dmp_Point.viewport");
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolDistanceAttenuationBool,
            "dmp_Point.distanceAttenuation");
        NN_GR_ASSERT( is_success );
    }

    {
        // Initialize primitives
        uptr physicalAddr = NULL;
        nn::gr::CTR::BindSymbolVSInput symbolPos;
        nn::gr::CTR::BindSymbolVSInput symbolColor;
        nn::gr::CTR::BindSymbolVSInput symbolSize;

        m_pPrimitivePoint = 
            static_cast< PrimitivePoint* >( allocator->Alloc( sizeof( PrimitivePoint ), 128 ) );

        m_pPrimitivePoint->SetupVtxStream( );

        // Search for "location"
        is_success = shader.SearchBindSymbol( &symbolPos, "aPosition" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &symbolColor, "aColor" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &symbolSize, "aPointSize" );
        NN_GR_ASSERT( is_success );

        // Vertex stream (Position)  
        physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_pPrimitivePoint->streamPos ) );
        NN_NULL_ASSERT( physicalAddr );
        m_Vertex.EnableAttrAsArray( symbolPos, physicalAddr, PICA_DATA_SIZE_4_FLOAT );

        // Color
        physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_pPrimitivePoint->streamCol ) );
        NN_NULL_ASSERT( physicalAddr );
        m_Vertex.EnableAttrAsArray( symbolColor, physicalAddr, PICA_DATA_SIZE_3_FLOAT );

        // Size
        physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_pPrimitivePoint->streamSize ) );
        NN_NULL_ASSERT( physicalAddr );
        m_Vertex.EnableAttrAsArray( symbolSize, physicalAddr, PICA_DATA_SIZE_1_FLOAT );

        // Index setting
        physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_pPrimitivePoint->streamIdx ) );
        NN_NULL_ASSERT( physicalAddr );
        m_IndexStream.physicalAddr = physicalAddr;  
        m_IndexStream.drawVtxNum   = m_pPrimitivePoint->VERTEX_NUM;

        // The index is defined as u8.
        m_IndexStream.isUnsignedByte = true;
    }

    { 
        // Texture initialization using texture package library

        // Get the size of the texture package file
        nn::fs::FileReader texturePackageFile( s_TexturePackageName );
        size_t texturePackageFileSize = texturePackageFile.GetSize();        
        if ( texturePackageFileSize == 0 )
        {
            NN_LOG( "Failed to open %s.\n", s_TexturePackageName );
            NN_ASSERT( 0 );

            return false;
        }
       
        // Load the texture package file
        u8* texturePackageFileBuffer = (u8*) allocator->Alloc( texturePackageFileSize, 128 );
        if ( texturePackageFileBuffer == NULL )
        {
            NN_LOG( "Allocator returned NULL.\n");
            NN_ASSERT( 0 );

            return false;
        }

        s32 texturePackageFileReadSize = texturePackageFile.Read( texturePackageFileBuffer, texturePackageFileSize );
        if ( texturePackageFileReadSize == 0 )
        {
            NN_LOG( "Failed to open %s.\n", s_TexturePackageName );
            NN_ASSERT( 0 );

            return false;
        }
            
        // Get the texture index
        int textureIndex = nn::tpl::CTR::GetTextureIndex( texturePackageFileBuffer, s_TextureNames[ s_TextureIndex ] );
        if( textureIndex == -1 )
        {
            NN_LOG("Failed to get texture's index.\n");
            texturePackageFile.Finalize();
            allocator->Free( texturePackageFileBuffer );
            NN_ASSERT( 0 );

            return false;
        }

        // Get the data pointer of the texture
        u8* textureDataBuffer = NULL;
        s32 mipLevel = 0;
        u32 mipmapSize = 0;
        textureDataBuffer = (u8*) nn::tpl::CTR::GetTexture( &mipLevel, &mipmapSize, texturePackageFileBuffer, textureIndex );
        if( ! textureDataBuffer )
        {
            NN_LOG("Failed to get texture data pointer.\n");
            texturePackageFile.Finalize();
            allocator->Free( texturePackageFileBuffer );
            NN_ASSERT( 0 );
            
            return false;
        }

        // Get the texture information        
        nn::tpl::CTR::CtrTextureInfo& s_TextureInfo = s_TextureInfos[ s_TextureIndex ];
        bool result = nn::tpl::CTR::GetTextureInfo(&s_TextureInfo, texturePackageFileBuffer, textureIndex);        
        if( ! result )
        {
            NN_LOG("Failed to get texture information.\n");
            texturePackageFile.Finalize();
            allocator->Free( texturePackageFileBuffer );
            NN_ASSERT( 0 );
                 
            return false;
        }
        
        // Copy the texture data
        s_TextureAddr[ s_TextureIndex ] = 
            static_cast<u8*>( allocator->Alloc( s_TextureInfo.texDataSize, 128 ) );        
        std::memcpy( s_TextureAddr[ s_TextureIndex ], textureDataBuffer, s_TextureInfo.texDataSize );        

        // Finalization of the texture package
        texturePackageFile.Finalize();
        allocator->Free( texturePackageFileBuffer );
    }

    // Texture settings
    m_Texture.unit0.texType  = PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_2D;
    m_Texture.unit0.physicalAddr = nngxGetPhysicalAddr( reinterpret_cast<uptr>( s_TextureAddr[ s_TextureIndex ] ) );
    m_Texture.unit0.width  = s_TextureInfos[ s_TextureIndex ].width;
    m_Texture.unit0.height = s_TextureInfos[ s_TextureIndex ].height;
    m_Texture.unit0.format = GetPicaDataTextureFormat( s_TextureInfos[ s_TextureIndex ].texFormat );

    // Get allocator and deallocator
    nngxGetAllocator( &m_GraphicsMemoryAllocator, &m_GraphicsMemoryDeallocator );

    {
        // Allocate color buffer  GL_RGBA8_OES
        renderBuffer[ 0 ] = 1;
        GLenum area = NN_GX_MEM_VRAMA;
        GLenum aim  = NN_GX_MEM_RENDERBUFFER;
        u32 colorBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        colorBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 0 ], colorBufferByteSize ) );

        // Allocate depth buffer  GL_DEPTH24_STENCIL8_EXT
        renderBuffer[ 1 ] = 2;
        area = NN_GX_MEM_VRAMB;
        u32 depthBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        depthBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 1 ], depthBufferByteSize ) );
    }

    {
        renderTarget = 1;

        // Initialize framebuffer settings
        m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

        // Initialize color buffer settings
        m_FrameBuffer.colorBuffer.virtualAddr         = colorBufferVirtualAddress;
        m_FrameBuffer.colorBuffer.format              = PICA_DATA_COLOR_RGBA8_OES;
        m_FrameBuffer.colorBuffer.blockSize           = nn::gr::FrameBuffer::BLOCK_SIZE8;
        m_FrameBuffer.colorBuffer.width               = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.colorBuffer.height              = nn::gx::DISPLAY0_HEIGHT;  
        m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

        // Initialize depth buffer settings
        m_FrameBuffer.depthStencilBuffer.virtualAddr  = depthBufferVirtualAddress;
        m_FrameBuffer.depthStencilBuffer.format       = PICA_DATA_DEPTH24_STENCIL8_EXT;
        m_FrameBuffer.depthStencilBuffer.width        = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.depthStencilBuffer.height       = nn::gx::DISPLAY0_HEIGHT;
        m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
        m_FrameBuffer.depthStencilBuffer.clearStencil = 0;
    }

    m_Scissor.SetBufferSize( nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT );

    // Render state settings
    m_RenderState.wBuffer.wScale = 0.0f;


    // glEnable(GL_BLEND);
    // glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    m_RenderState.blend.isEnable = true;
    m_RenderState.blend.srcAlpha   = PICA_DATA_BLEND_FUNC_ONE;

    // glEnable(GL_DEPTH_TEST);
    // glDepthFunc(GL_LESS);    
    m_RenderState.depthTest.isEnable      = true;
    m_RenderState.depthTest.isEnableWrite = true;

    m_RenderState.cullingTest.isEnable  = false;

    {
        // Texture combiner settings
        m_Combiner.stage[ 0 ].rgb.combine         = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
        m_Combiner.stage[ 0 ].rgb.source[ 0 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0;
        m_Combiner.stage[ 0 ].rgb.source[ 1 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].rgb.source[ 2 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;

        m_Combiner.stage[ 0 ].rgb.operand[ 0 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
        m_Combiner.stage[ 0 ].rgb.operand[ 1 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
        m_Combiner.stage[ 0 ].rgb.operand[ 2 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;

        m_Combiner.stage[ 0 ].alpha.combine     = PICA_DATA_TEX_ENV_COMBINE_REPLACE;

        m_Combiner.stage[ 0 ].alpha.source[ 0 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].alpha.source[ 1 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].alpha.source[ 2 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;

        m_Combiner.stage[ 0 ].alpha.operand[ 0 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
        m_Combiner.stage[ 0 ].alpha.operand[ 1 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
        m_Combiner.stage[ 0 ].alpha.operand[ 2 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
    }

    return true;
}

//------------------------------------------------------------------------------
void PointSpriteScene::Finalize( void )
{
    m_Vertex.DisableAll();

    allocator->Free( m_pPrimitivePoint );

    if ( s_TextureAddr[ s_TextureIndex ] != NULL )
    {
        allocator->Free( s_TextureAddr[ s_TextureIndex ] );
        s_TextureAddr[ s_TextureIndex ] = NULL;
    }

    allocator->Free( shaderBinary );
    allocator->Free( shaderBuffer );

    GLenum area = NN_GX_MEM_VRAMA;
    GLenum aim  = NN_GX_MEM_RENDERBUFFER;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 0 ], reinterpret_cast<void*>( colorBufferVirtualAddress ) );
    colorBufferVirtualAddress = 0;

    area = NN_GX_MEM_VRAMB;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 1 ], reinterpret_cast<void*>( depthBufferVirtualAddress ) );
    depthBufferVirtualAddress = 0;
}

//------------------------------------------------------------------------------
void PointSpriteScene::Enter( void )
{
    stereo.Initialize();
    nngxSetDisplayMode( NN_GX_DISPLAYMODE_STEREO );
    Calc();
}

//------------------------------------------------------------------------------
bool PointSpriteScene::Calc( void )
{
    UpdateScene_();

    m_Counter++;

    return true;
}

//------------------------------------------------------------------------------
void PointSpriteScene::PreDraw( void )
{
}

//------------------------------------------------------------------------------
GLuint PointSpriteScene::DrawTop( void )
{
    f32 width  = nn::gx::DISPLAY0_WIDTH;
    f32 height = nn::gx::DISPLAY0_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );
      
    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;
    
    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    {
        // The camera matrix        
        modelViewMtx34.SetupLookAt(
            // camPos
            nn::math::VEC3( 0.0f, 0.4f, 7.5f ),
            // camUp
            nn::math::VEC3( 1.0f, 0.0f, 0.0f ),
            // target
            nn::math::VEC3( 0.0f, 0.0f, 0.0f ) );
    }

    // Projection matrix    
    nn::math::MTX44Frustum(&projectionMtx44, -0.02f, 0.02f, -0.02f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH,
                 0.02f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 20.f);

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR,
        6.5f, 0.2f, nn::math::PIVOT_UPSIDE_TO_TOP );

    RenderSceneGr_( NN_GX_DISPLAY0, projL, viewL );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint PointSpriteScene::DrawExt( void )
{
    f32 width  = nn::gx::DISPLAY0_WIDTH; 
    f32 height = nn::gx::DISPLAY0_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );

    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR,
        6.5f, 0.2f, nn::math::PIVOT_UPSIDE_TO_TOP);

    RenderSceneGr_( NN_GX_DISPLAY0_EXT, projR, viewR );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint PointSpriteScene::DrawBtm( void )
{
    f32 width  = nn::gx::DISPLAY1_WIDTH; 
    f32 height = nn::gx::DISPLAY1_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );
 
    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    return renderTarget;
}

//------------------------------------------------------------------------------
void PointSpriteScene::PostDraw( void )
{
}

//------------------------------------------------------------------------------
nn::gr::CTR::FrameBuffer::ColorBuffer* PointSpriteScene::GetGrColorBufferPointer( void )
{
    return & m_FrameBuffer.colorBuffer;
}


//------------------------------------------------------------------------------
void PointSpriteScene::Clear( const u32 bufferBit )
{
    u32* command = 0;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
        reinterpret_cast<GLint*>( &command ) );

    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Framebuffer settings
    command = m_FrameBuffer.MakeCommand( command, bufferBit );

    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );

    // Add a load complete command at the end of the 3D command buffer
    bool isAddSplitDrawCmdlist = true;

    // Generate a request to clear the color buffer and depth buffer
    m_FrameBuffer.MakeClearRequest( bufferBit, isAddSplitDrawCmdlist );
}

//------------------------------------------------------------------------------

void PointSpriteScene::UpdateScene_( void )
{
}

//------------------------------------------------------------------------------

bit32* PointSpriteScene::MakeTextureCommand_( bit32* command )
{
    // Generate 3 dummy commands to 0x80    
    const bool isAddDummyCommand = true;

    command = m_Texture.MakeCommand( command, isAddDummyCommand );

    return command;
}

//------------------------------------------------------------------------------
void PointSpriteScene::RenderSceneGr_( const s32 display,
                                      const nn::math::MTX44& proj, const nn::math::MTX34& view
                                       )
{
    NN_UNUSED_VAR( display );
    NN_UNUSED_VAR( proj );
    NN_UNUSED_VAR( view );

    // Get a pointer to the render command buffer
    u32* command = NULL;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );

    // Get the start address (so that the pointer can be advanced at the end)
    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Clear the framebuffer cache
    command = nn::gr::CTR::FrameBuffer::MakeClearCacheCommand( command );

    // Render state settings
    command = m_RenderState.MakeCommand( command );

    // Texture combiner settings
    command = m_Combiner.MakeCommand( command );

    // Texture settings
    command = MakeTextureCommand_( command );

    command = DrawPointSpriteCommand_( command, display, proj, view );
    
    // Advance the pointer to the render command buffer
    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof( u32 ) );
}

//------------------------------------------------------------------------------

bit32* PointSpriteScene::DrawPointSpriteCommand_( bit32* command,
                                                  const s32 display,
                                                  const nn::math::MTX44& proj,
                                                  const nn::math::MTX34& view
                                                  )
{
    nn::gr::ShaderLite* pShader = &shader;

    // Shader binary setting
    pShader->SetShaderIndex( 0, 1 );
    pShader->SetDrawMode( PICA_DATA_DRAW_GEOMETRY_PRIMITIVE );
    
    pShader->SetUniformBool( twnBindSymbolDistanceAttenuationBool, false );
    
    static bool isFirst = true;

    if ( isFirst )
    {
        command = pShader->MakeFullCommand( command );
        isFirst = false;
    }
    else
    {
        command = pShader->MakeShaderCommand( command );
    }

    u32 width  = nn::gx::DISPLAY0_WIDTH;
    u32 height = nn::gx::DISPLAY0_HEIGHT;

    if ( display == NN_GX_DISPLAY1 )
    {
        width  = nn::gx::DISPLAY1_WIDTH;
        height = nn::gx::DISPLAY1_HEIGHT;
    }

    nn::math::VEC2 vec2;
    vec2[ 0 ] = 1.f / static_cast<f32>( width );
    vec2[ 1 ] = 1.f / static_cast<f32>( height );
    command   = twnBindSymbolViewPortVec2.MakeUniformCommand( command, static_cast<nn::math::Vector4>(vec2) );

    // Rotate the model
    nn::math::Matrix34 rotMtx34;
    nn::math::Matrix34 modelViewMtx34;
    nn::math::MTX34RotXYZDeg( &rotMtx34, 0.0f, -6.0f * m_Counter, 0.0f );
    nn::math::MTX34Mult( &modelViewMtx34, &view, &rotMtx34 );
    nn::math::Matrix44 modelViewMtx44( modelViewMtx34 );

    // Index stream settings   
    {
        uptr physicalAddr = nngxGetPhysicalAddr( reinterpret_cast<uptr>( m_pPrimitivePoint->streamIdx ));
        NN_NULL_ASSERT( physicalAddr );
        m_IndexStream.physicalAddr = physicalAddr;

        m_IndexStream.drawVtxNum   = m_pPrimitivePoint->VERTEX_NUM;
    }

    // Set vertex shader uniforms
    command = bindSymbolProjectionMtx44.MakeUniformCommand( command, proj );
    command = bindSymbolModelViewMtx44.MakeUniformCommand( command, modelViewMtx44 );

    // Disable all vertex attributes before we start to use vertices
    command = nn::gr::Vertex::MakeDisableCommand( command );

    // glBindBuffer( GL_ARRAY_BUFFER )
    // glBindBuffer( GL_ELEMENT_ARRAY_BUFFER ) + glDrawElements()
    command = m_Vertex.MakeEnableAttrCommand( command );

    command = m_Vertex.MakeDrawCommand( command, m_IndexStream );

    // Disable all vertex attributes when we finish using vertices
    command = nn::gr::CTR::Vertex::MakeDisableCommand( command );

    return command;
}
