/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_ProcedureTextureData.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_PROCEDURE_TEXTURE_DATA_H_
#define GR_PROCEDURE_TEXTURE_DATA_H_

#include <nn/os.h>
#include <nn/fs.h>

#include <nn/gx/CTR/gx_CommandAccess.h>



//! The color lookup table holds 256 pieces of data.             
const u32 PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM      = 256;
const u32 PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_HALF_NUM = PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM / 2;

//! The lookup table holds 128 pieces of data. 
const u32 PROCTEX_LOOKUP_TABLE_ELEMENT_NUM            = 128;



const u32 PROCEDURE_TEXTURE_DATA_HEADER  = 0x0;
const f32 PROCEDURE_TEXTURE_DATA_VERSION = 0.1f;

/*!
    @brief Structure for procedural texture data. 
*/
struct ProcedureTextureData
{ 
    /*!
        @brief Header information.
    */
    u32 header;

    /*!
        @brief Version number.
    */
    f32 version;

    /*!
        @brief RGB mapping function mode.
               The type is PicaDataProcTexMap.
               Equivalent to dmp_Texture[3].ptRgbMap uniform. 
    */
    PicaDataProcTexMap rgbMap;

    /*!
        @brief Alpha mapping function mode.
               The type is PicaDataProcTexMap.
               Equivalent to dmp_Texture[3].ptAlphaMap uniform. 
    */
    PicaDataProcTexMap alphaMap;

    /*!
        @brief Whether alpha is treated as independent in the mapping function.
               Type is bool.
               Equivalent to dmp_Texture[3].ptAlphaSeparate uniform. 
    */
    bool alphaSeparate;

    /*!
        @brief Clamp mode of the texture's U direction.
               The type is PicaDataProcTexClamp.
               Equivalent to dmp_Texture[3].ptClamp uniform. 
    */
    PicaDataProcTexClamp clampU;

    /*!
        @brief Clamp mode of the texture's V direction.
               The type is PicaDataProcTexClamp.
               Equivalent to dmp_Texture[3].ptClampV uniform. 
    */
    PicaDataProcTexClamp clampV;

    /*!
        @brief Shift calculation mode of the texture's U direction.
               The type is PicaDataProcTexShift.
               Equivalent to dmp_Texture[3].ptShiftU uniform. 
    */
    PicaDataProcTexShift shiftU;

    /*!
        @brief Shift calculation mode of the texture's V direction.
               The type is PicaDataProcTexShift.
               Equivalent to dmp_Texture[3].ptShiftV uniform. 
    */
    PicaDataProcTexShift shiftV;

    /*!
        @brief The texture's MinFilter setting.
               The type is PicaDataProcTexFilter.
               Equivalent to dmp_Texture[3].ptMinFilter uniform. 
    */
    PicaDataProcTexFilter minFilter;

    /*!
        @brief LOD bias of the color lookup table. 
               The type is f32 and the value is [0.0, 6.0]. The default value is 0.5f. 
               Equivalent to dmp_Texture[3].texBias uniform. 
    */
    f32 texBias;

    /*!
        @brief Specifies the starting element in the partial array being used. 
               The type is u8 and the value is [0, 128]. The default value is 0. 
               Equivalent to dmp_Texture[3].ptTexOffset  uniform. 
    */
    u8  texOffset;

    /*!
        @brief Specifies the data width of the lookup table
               The type is u8 and the value is [0, 128]. The default value is 128. 
               Equivalent to dmp_Texture[3].ptTexWidth uniform. 
    */
    u8  texWidth;
    NN_PADDING2;
  
    /*!
        @brief RGB mapping function lookup table.
    */
    f32 rgbMapLookUpTable[ PROCTEX_LOOKUP_TABLE_ELEMENT_NUM ];

    /*!
        @brief Alpha mapping function lookup table.
    */
    f32 alphaMapLookUpTable[ PROCTEX_LOOKUP_TABLE_ELEMENT_NUM ];
        
    /*!
        @brief Red color lookup table.
    */
    f32 redLookUpTable[ PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM ];

    /*!
        @brief Green color lookup table.
    */
    f32 greenLookUpTable[ PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM ];

    /*!
        @brief Blue color lookup table.
    */
    f32 blueLookUpTable[ PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM ];

    /*!
        @brief Alpha color lookup table.
    */
    f32 alphaLookUpTable[ PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM ];

    /*!
        @brief Enable noise. 
               The type is bool. The default value is "false".
               Equivalent to dmp_Texture[3].ptNoiseEnable uniform. 
    */
    bool isEnableNoise;
    bool isEnableNoiseU;
    bool isEnableNoiseV;
    NN_PADDING1;
    
    /*!
        @brief Sets the noise amplitude in the texture's U direction.
               The type is f32. The default value is 0.0f.
               Equivalent to dmp_Texture[3].ptNoiseU[2] uniform. 
    */
    f32 noiseUAmplitude;

    /*!
        @brief Sets the noise amplitude in the texture's V direction.
               The type is f32. The default value is 0.0f.
               Equivalent to dmp_Texture[3].ptNoiseV[2] uniform. 
    */
    f32 noiseVAmplitude;

    /*!
        @brief Sets the noise frequency in the texture's U direction.
               The type is f32. The default value is 0.0f.
               Equivalent to dmp_Texture[3].ptNoiseU[0] uniform. 
    */
    f32 noiseUFrequency;

    /*!
        @brief Sets the noise frequency in the texture's V direction.
               The type is f32. The default value is 0.0f.
               Equivalent to dmp_Texture[3].ptNoiseV[0] uniform. 
    */
    f32 noiseVFrequency;

    /*!
        @brief Sets the noise phase in the texture's U direction.
               The type is f32. The default value is 0.0f.
               Equivalent to dmp_Texture[3].ptNoiseU[1] uniform. 
    */
    f32 noiseUPhase;

    /*!
        @brief Sets the noise phase in the texture's V direction.
               The type is f32. The default value is 0.0f.
               Equivalent to dmp_Texture[3].ptNoiseV[1] uniform. 
    */
    f32 noiseVPhase;

    /*!
        @brief Sets the noise phase step in the texture's U direction.
               The type is f32. The default value is 0.0f.        
    */
    f32 noiseUPhaseStep;

    /*!
        @brief Sets the noise phase step in the texture's U direction.
               The type is f32. The default value is 0.0f.        
    */
    f32 noiseVPhaseStep;

    /*!
        @brief Lookup table for noise-related continuous functions. 
    */
    f32 noiseLookUpTable[ PROCTEX_LOOKUP_TABLE_ELEMENT_NUM ];

    /*!
        @brief Background color.
    */
    f32 clearColor[ 4 ];

   /*!
        @brief Texture coordinates.
    */
    f32 texCoordU[ 2 ];
    f32 texCoordV[ 2 ];
};

bool ReadProcedureTextureData( const wchar_t* fileName, ProcedureTextureData& procedureTextureData )
{
    nn::fs::FileReader file( fileName );

    size_t fileSize = file.GetSize();
    
    if ( fileSize == 0 )
    {
        NN_LOG("Error : Cannot find procedureTextureData . \n");

        return false;
    }
        
    s32 readSize = file.Read( &procedureTextureData, fileSize );
    file.Finalize();

    if ( readSize == 0 )
    {
        NN_LOG("Error : Read file size of procedureTextureData is 0 .\n");

        return false;
    }

    if ( procedureTextureData.header != PROCEDURE_TEXTURE_DATA_HEADER )
    {
        NN_LOG("Error : ProcedureTextureData header mismatch. \n");
        
        return false;
    }
   
    if ( procedureTextureData.version != PROCEDURE_TEXTURE_DATA_VERSION )
    {
        NN_LOG("Error : ProcedureTextureData version mismatch. \n");

        return false;
    }

    return true;
}

#endif
