﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_ProcedureTextureScene.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/fs.h>

#include "gr_Framework.h"
#include "gr_ProcedureTextureScene.h"

//------------------------------------------------------------------------------
GfxAllocator * GrAllocator( MemoryMgr::AllocatorDevice *heap )
{
    return new GfxAllocator( heap );
}

GraphicSystem::GfxAllocatorFactory  GetGraphicsMemoryMgrFunctor( void )
{
    return &GrAllocator;
}

IScene * CreateScene( void )
{
    return new ProcedureTextureScene();
}

//------------------------------------------------------------------------------
ProcedureTextureScene::ProcedureTextureScene()
    : allocator( NULL ),
      shaderBinary( NULL ), shader(),
      mtx(), projectionMtx(), viewMtx(), stereo(),
      bindSymbolWorldMtx(), bindSymbolProjMtx(), bindSymbolViewMtx(),
      renderTarget( 0 ), 
      m_Viewport(), m_Scissor(), 
      m_UpperProcedurePatternIndex(  PROCTEX_FIRE_PATTERN ),
      m_UpperPrimitiveBox( NULL ), m_UpperVertexBox(), 
      m_UpperIndexStreamBox(),
      m_LowerProcedurePatternIndex( PROCTEX_SEA_PATTERN ),
      m_LowerPrimitiveBox( NULL ), m_LowerVertexBox(),
      m_LowerIndexStreamBox(),
      m_Pos(), m_EulerAngleDeg0(), m_EulerAngleDeg1(),
      m_Counter( 0 )
{
    for (u32 patternIndex = 0; patternIndex < PROCTEX_PATTERN_NUM; patternIndex++ )
    {
        m_IsEnableNoiseUArray[ patternIndex ] = false;
        m_IsEnableNoiseVArray[ patternIndex ] = false;

        m_NoiseUPhaseStepArray[ patternIndex ] = 0.0f;
        m_NoiseVPhaseStepArray[ patternIndex ] = 0.0f;

        for (u32 coordIndex = 0; coordIndex < 2; coordIndex++)
        {
            m_TexCoordUArray[ patternIndex ][ coordIndex ] = 0.0f;
            m_TexCoordVArray[ patternIndex ][ coordIndex ] = 0.0f;
        }
        
        for (u32 colorIndex = 0; colorIndex < 4; colorIndex++)
        {
            m_ClearColorArray[ patternIndex ][ colorIndex ] = 0.0f;
        }
    }
}

//------------------------------------------------------------------------------
ProcedureTextureScene::~ProcedureTextureScene()
{
}

//------------------------------------------------------------------------------
bool ProcedureTextureScene::Initialize( void )
{
    bool is_success;
    NN_UNUSED_VAR( is_success );

    allocator = MEMMGR->GetDefaultAllocatorDevice();

    ::std::srand(0x1234);
    { 
        // Initialize the shader

        // Load shader binary
        nn::fs::FileReader shader_file_reader( L"rom:/shader.shbin" );
        u32 shader_file_size = static_cast<u32>( shader_file_reader.GetSize() );
        shaderBinary = static_cast<u8*>( allocator->Alloc( shader_file_size, 128 ) );
        shader_file_reader.Read( shaderBinary, shader_file_size );
        shader_file_reader.Finalize();

        // Get size of the buffer needed for a ShaderLite class from the shader binary
        u32 size = shader.GetRequiredBufferSize( shaderBinary );

        shaderBuffer = static_cast<u32*>( allocator->Alloc( size ) );

        // Specify shader binary and initialize
        // Enable the executable image at position 0 as the vertex shader
        shader.SetupBinary( shaderBinary, 0, -1, shaderBuffer );
        
        // Uniform search
        is_success = shader.SearchBindSymbol( &bindSymbolWorldMtx, "user" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolProjMtx, "proj" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &bindSymbolViewMtx, "view" );
        NN_GR_ASSERT( is_success );
    }

    { 
        // Initialize primitives
        
        u32 physicalAddr = 0;
        nn::gr::CTR::BindSymbolVSInput symbolPos, symbolTex, symbolClr, symbolNrm;
        
        m_Pos.x =  0.0f;
        m_Pos.y =  0.0f;
        m_Pos.z =  0.0f;

        m_EulerAngleDeg0.x =  0.0f;
        m_EulerAngleDeg0.y =  0.0f;
        m_EulerAngleDeg0.z =  0.0f;

        m_EulerAngleDeg1.x =  0.0f;
        m_EulerAngleDeg1.y =  0.0f;
        m_EulerAngleDeg1.z =  0.0f;

        { 
            // About Box rendered on the upper screen
            m_UpperPrimitiveBox = static_cast< PrimitiveBox* >( allocator->Alloc( sizeof( PrimitiveBox ), 128 ) );
            m_UpperPrimitiveBox->SetupVtxStream();

            // Search for "location"
            is_success = shader.SearchBindSymbol( &symbolPos, "Position" );
            NN_GR_ASSERT( is_success );
            is_success = shader.SearchBindSymbol( &symbolTex, "TexCoord0" );
            NN_GR_ASSERT( is_success );
            is_success = shader.SearchBindSymbol( &symbolClr, "Color" );
            NN_GR_ASSERT( is_success );
            is_success = shader.SearchBindSymbol( &symbolNrm, "Normal" );
            NN_GR_ASSERT( is_success );

            // Vertex stream (Position) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_UpperPrimitiveBox->streamPos ) );
            m_UpperVertexBox.EnableAttrAsArray( symbolPos, physicalAddr, PICA_DATA_SIZE_3_FLOAT );
           
            // Vertex stream (TexCoord0) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_UpperPrimitiveBox->streamTex ) );
            m_UpperVertexBox.EnableAttrAsArray( symbolTex, physicalAddr, PICA_DATA_SIZE_2_FLOAT );

            // Vertex stream (Color) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_UpperPrimitiveBox->streamClr ) );
            m_UpperVertexBox.EnableAttrAsArray( symbolClr, physicalAddr, PICA_DATA_SIZE_4_FLOAT );

            // Vertex stream (Normal) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_UpperPrimitiveBox->streamNrm ) );
            m_UpperVertexBox.EnableAttrAsArray( symbolNrm, physicalAddr, PICA_DATA_SIZE_3_FLOAT );

            // Index setting
            m_UpperIndexStreamBox.physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_UpperPrimitiveBox->streamIdx ) );
            m_UpperIndexStreamBox.drawVtxNum = m_UpperPrimitiveBox->VERTEX_NUM;
            m_UpperIndexStreamBox.isUnsignedByte = true; // Index is defined as u8.
        }

        {
            // About Box rendered on the lower screen         
            m_LowerPrimitiveBox = static_cast< PrimitiveBox* >( allocator->Alloc( sizeof( PrimitiveBox ), 128 ) );
            m_LowerPrimitiveBox->SetupVtxStream();

            // Search for "location"
            is_success = shader.SearchBindSymbol( &symbolPos, "Position" );
            NN_GR_ASSERT( is_success );
            is_success = shader.SearchBindSymbol( &symbolTex, "TexCoord0" );
            NN_GR_ASSERT( is_success );
            is_success = shader.SearchBindSymbol( &symbolClr, "Color" );
            NN_GR_ASSERT( is_success );
            is_success = shader.SearchBindSymbol( &symbolNrm, "Normal" );
            NN_GR_ASSERT( is_success );

            // Vertex stream (Position) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_LowerPrimitiveBox->streamPos ) );
            m_LowerVertexBox.EnableAttrAsArray( symbolPos, physicalAddr, PICA_DATA_SIZE_3_FLOAT );
           
            // Vertex stream (TexCoord0) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_LowerPrimitiveBox->streamTex ) );
            m_LowerVertexBox.EnableAttrAsArray( symbolTex, physicalAddr, PICA_DATA_SIZE_2_FLOAT );

            // Vertex stream (Color) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_LowerPrimitiveBox->streamClr ) );
            m_LowerVertexBox.EnableAttrAsArray( symbolClr, physicalAddr, PICA_DATA_SIZE_4_FLOAT );

            // Vertex stream (Normal) setting
            physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_LowerPrimitiveBox->streamNrm ) );
            m_LowerVertexBox.EnableAttrAsArray( symbolNrm, physicalAddr, PICA_DATA_SIZE_3_FLOAT );

            // Index setting
            m_LowerIndexStreamBox.physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_LowerPrimitiveBox->streamIdx ) );
            m_LowerIndexStreamBox.drawVtxNum = m_LowerPrimitiveBox->VERTEX_NUM;
            m_LowerIndexStreamBox.isUnsignedByte = true; // Index is defined as u8.
        }
    }
    
    InitializeProcedureTextureArray();
    
    // Get allocator and deallocator
    nngxGetAllocator( &m_GraphicsMemoryAllocator, &m_GraphicsMemoryDeallocator );

    {
        // Allocate color buffer  GL_RGBA8_OES
        renderBuffer[ 0 ] = 1;
        GLenum area = NN_GX_MEM_VRAMA;
        GLenum aim  = NN_GX_MEM_RENDERBUFFER;
        u32 colorBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        colorBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 0 ], colorBufferByteSize ) );

        // Allocate depth buffer  GL_DEPTH24_STENCIL8_EXT
        renderBuffer[ 1 ] = 2;
        area = NN_GX_MEM_VRAMB;
        u32 depthBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        depthBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 1 ], depthBufferByteSize ) );
    }

    {
        renderTarget = 1;

        // Initialize framebuffer settings
        m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

        // Initialize color buffer settings
        m_FrameBuffer.colorBuffer.virtualAddr         = colorBufferVirtualAddress;
        m_FrameBuffer.colorBuffer.format              = PICA_DATA_COLOR_RGBA8_OES;
        m_FrameBuffer.colorBuffer.blockSize           = nn::gr::FrameBuffer::BLOCK_SIZE8;
        m_FrameBuffer.colorBuffer.width               = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.colorBuffer.height              = nn::gx::DISPLAY0_HEIGHT;  
        m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;
       
        // Initialize depth buffer settings
        m_FrameBuffer.depthStencilBuffer.virtualAddr  = depthBufferVirtualAddress;
        m_FrameBuffer.depthStencilBuffer.format       = PICA_DATA_DEPTH24_STENCIL8_EXT;
        m_FrameBuffer.depthStencilBuffer.width        = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.depthStencilBuffer.height       = nn::gx::DISPLAY0_HEIGHT;
        m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
        m_FrameBuffer.depthStencilBuffer.clearStencil = 0;
    }

    m_Scissor.SetBufferSize( nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT );

    // Initialize using a disable command
    {
        u32* command = NULL;
        nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
            reinterpret_cast<GLint*>( &command ) );
        const u32* start_addr = command;

        command = nn::gr::CTR::MakeDisableAllCommand( command );

        nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );
    }

    return true;
}

void ProcedureTextureScene::InitializeProcedureTextureArray( void )
{
    for ( u32 patternIndex = 0; patternIndex < PROCTEX_PATTERN_NUM; patternIndex++)
    {
        wchar_t* fileName = s_ProcedureTextureDataFileNameArray[ patternIndex ];

        ProcedureTextureData* procedureTextureDataPtr = 
            static_cast<ProcedureTextureData*>( allocator->Alloc( sizeof( ProcedureTextureData ), 128 ) );

        bool flag = ReadProcedureTextureData( fileName, (*procedureTextureDataPtr) );
        if (! flag )
        {
            NN_LOG("Cannot open *.pctxbin .\n");
            NN_ASSERT( 0 );
            return;
        }

        nn::gr::CTR::ProcedureTexture& procedureTexture = m_ProcedureTextureArray[ patternIndex ];

        procedureTexture.texBias   = 0.5f;
        procedureTexture.texOffset = 0;
        procedureTexture.texWidth = 128;

        procedureTexture.alphaSeparate = false;

        procedureTexture.rgbMap   = procedureTextureDataPtr->rgbMap;
        procedureTexture.clampU = procedureTextureDataPtr->clampU;
        procedureTexture.clampV = procedureTextureDataPtr->clampV;
        procedureTexture.shiftU = procedureTextureDataPtr->shiftU;
        procedureTexture.shiftV = procedureTextureDataPtr->shiftV;

        procedureTexture.isEnableNoise      = procedureTextureDataPtr->isEnableNoise;
        m_IsEnableNoiseUArray[ patternIndex ] = procedureTextureDataPtr->isEnableNoiseU;
        m_IsEnableNoiseVArray[ patternIndex ] = procedureTextureDataPtr->isEnableNoiseV;

        procedureTexture.noiseUAmplitude = procedureTextureDataPtr->noiseUAmplitude;
        procedureTexture.noiseVAmplitude = procedureTextureDataPtr->noiseVAmplitude;

        if ( ! m_IsEnableNoiseUArray[ patternIndex ] )
        {
            procedureTexture.noiseUAmplitude = 0.0f;
        }

        if ( ! m_IsEnableNoiseVArray[ patternIndex ] )
        {
            procedureTexture.noiseVAmplitude = 0.0f;
        }

        procedureTexture.noiseUFrequency = procedureTextureDataPtr->noiseUFrequency;
        procedureTexture.noiseVFrequency = procedureTextureDataPtr->noiseVFrequency;

        procedureTexture.noiseUPhase     = procedureTextureDataPtr->noiseUPhase;
        procedureTexture.noiseVPhase     = procedureTextureDataPtr->noiseVPhase;

        m_NoiseUPhaseStepArray[ patternIndex ] = procedureTextureDataPtr->noiseUPhaseStep;
        m_NoiseVPhaseStepArray[ patternIndex ] = procedureTextureDataPtr->noiseVPhaseStep;

        // Set lookup tables
        // Set the last difference value.
        const f32 lastDiffValue = 0.0f;
        procedureTexture.SetNoiseLookUpTable( procedureTextureDataPtr->noiseLookUpTable, lastDiffValue );
        procedureTexture.SetRgbMapLookUpTable( procedureTextureDataPtr->rgbMapLookUpTable, lastDiffValue );
        procedureTexture.SetAlphaMapLookUpTable( procedureTextureDataPtr->alphaMapLookUpTable, lastDiffValue );

        // Set lookup tables for color
        f32 redLookUpTable[ PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM ];
        f32 greenLookUpTable[ PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM ];
        f32 blueLookUpTable[ PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM ];
        f32 alphaLookUpTable[ PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM ];
       
        for (u32 tableIndex = 0; tableIndex < PROCTEX_COLOR_LOOKUP_TABLE_ELEMENT_NUM; tableIndex++ )
        {     
            redLookUpTable[ tableIndex ]   = procedureTextureDataPtr->redLookUpTable[ tableIndex ];
            greenLookUpTable[ tableIndex ] = procedureTextureDataPtr->greenLookUpTable[ tableIndex ];
            blueLookUpTable[ tableIndex ]  = procedureTextureDataPtr->blueLookUpTable[ tableIndex ];
            alphaLookUpTable[ tableIndex ] = procedureTextureDataPtr->alphaLookUpTable[ tableIndex ];
        }
        
        // No need to set the last difference value in the color lookup table
        procedureTexture.SetRedLookUpTable( redLookUpTable );
        procedureTexture.SetGreenLookUpTable( greenLookUpTable );
        procedureTexture.SetBlueLookUpTable( blueLookUpTable );
        procedureTexture.SetAlphaLookUpTable( alphaLookUpTable );

        for (u32 coordIndex = 0; coordIndex < 2; coordIndex++)
        {
            m_TexCoordUArray[ patternIndex ][ coordIndex ] = 
                procedureTextureDataPtr->texCoordU[ coordIndex ];
            m_TexCoordVArray[ patternIndex ][ coordIndex ] =
                procedureTextureDataPtr->texCoordV[ coordIndex ];
        }

        for (u32 colorIndex = 0; colorIndex < 4; colorIndex++)
        {
            m_ClearColorArray[ patternIndex ][ colorIndex ] = 
                procedureTextureDataPtr->clearColor[ colorIndex ];
        }
    
        allocator->Free( procedureTextureDataPtr );
    }
}

//------------------------------------------------------------------------------
void ProcedureTextureScene::Finalize( void )
{
    allocator->Free( m_UpperPrimitiveBox );
    allocator->Free( m_LowerPrimitiveBox );

    allocator->Free( shaderBinary );
    allocator->Free( shaderBuffer );

    GLenum area = NN_GX_MEM_VRAMA;
    GLenum aim  = NN_GX_MEM_RENDERBUFFER;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 0 ], reinterpret_cast<void*>( colorBufferVirtualAddress ) );
    colorBufferVirtualAddress = 0;

    area = NN_GX_MEM_VRAMB;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 1 ], reinterpret_cast<void*>( depthBufferVirtualAddress ) );
    depthBufferVirtualAddress = 0;
}

//------------------------------------------------------------------------------
void ProcedureTextureScene::Enter( void )
{
    stereo.Initialize();
    nngxSetDisplayMode( NN_GX_DISPLAYMODE_STEREO );
    Calc();
}


//------------------------------------------------------------------------------
bool ProcedureTextureScene::Calc( void )
{    
    UpdateScene_();

    m_Counter++;
        
    return true;
}

//------------------------------------------------------------------------------
void ProcedureTextureScene::PreDraw( void )
{
}

//------------------------------------------------------------------------------
GLuint ProcedureTextureScene::DrawTop( void )
{
    m_Viewport.Set( 0, 0, 240, 400 );
    m_Scissor.Set( 0, 0, 240, 400 );
  
    for (u32 index = 0; index < 4; index++)
    {
        m_FrameBuffer.colorBuffer.clearColor[ index ] = 
            m_ClearColorArray[ m_UpperProcedurePatternIndex ][ index ];
    }

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
    m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );
 
    { 
        // The camera matrix
        // f32 r = 60.f;
        viewMtx.SetupLookAt(
            // camPos
            nn::math::VEC3( 0.0f, 0.0f, 22.0f ),
            // camUp
            nn::math::VEC3( 0.0f, 1.0f, 0.0f ),
            // target
            nn::math::VEC3( 0.0f, 0.0f, 0.0f ) );
    }

    // Projection matrix
    projectionMtx.SetupPerspective(
        0.66f,          // fovy
        400.f / 240.f,  // aspect
        0.1f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &viewMtx );
    stereo.SetBaseFrustum( &projectionMtx );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f, nn::math::PIVOT_UPSIDE_TO_TOP );
    
    RenderSceneGr_( NN_GX_DISPLAY0, projL, viewL );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint ProcedureTextureScene::DrawExt( void )
{
  
    m_Viewport.Set( 0, 0, 240, 400 );
    m_Scissor.Set( 0, 0, 240, 400 );

    for (u32 index = 0; index < 4; index++)
    {
        m_FrameBuffer.colorBuffer.clearColor[ index ] = 
            m_ClearColorArray[ m_UpperProcedurePatternIndex ][ index ];
    }

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
    m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &viewMtx );
    stereo.SetBaseFrustum( &projectionMtx );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f, nn::math::PIVOT_UPSIDE_TO_TOP);
    
    RenderSceneGr_( NN_GX_DISPLAY0_EXT, projR, viewR );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint ProcedureTextureScene::DrawBtm( void )
{
    m_Viewport.Set( 0, 0, 240, 320 );
    m_Scissor.Set( 0, 0, 240, 320 );
 
    for (u32 index = 0; index < 4; index++)
    {
        m_FrameBuffer.colorBuffer.clearColor[ index ] = 
            m_ClearColorArray[ m_LowerProcedurePatternIndex ][ index ];
    }

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = nn::gx::DISPLAY1_WIDTH;
    m_FrameBuffer.height                          = nn::gx::DISPLAY1_HEIGHT;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );
   
    projectionMtx.SetupPerspective(
        0.66f,          // fovy
        320.f / 240.f,  // aspect
        0.1f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &viewMtx );
    stereo.SetBaseFrustum( &projectionMtx );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f );

    RenderSceneGr_( NN_GX_DISPLAY1, projR, viewR );

    return renderTarget;
}

//------------------------------------------------------------------------------
void ProcedureTextureScene::PostDraw( void )
{
}

//------------------------------------------------------------------------------
nn::gr::CTR::FrameBuffer::ColorBuffer* ProcedureTextureScene::GetGrColorBufferPointer( void )
{    
    return & m_FrameBuffer.colorBuffer;
}


//------------------------------------------------------------------------------
void ProcedureTextureScene::Clear( const u32 bufferBit )
{
    u32* command = 0;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
        reinterpret_cast<GLint*>( &command ) );
    
    const u32* start_addr = command;
    
    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Framebuffer settings
    command = m_FrameBuffer.MakeCommand( command, bufferBit );
                                             
    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );

    // Add a load complete command at the end of the 3D command buffer
    bool isAddSplitDrawCmdlist = true;

    // Generate a request to clear the color buffer and depth buffer
    m_FrameBuffer.MakeClearRequest( bufferBit, isAddSplitDrawCmdlist );
}

//------------------------------------------------------------------------------
void ProcedureTextureScene::UpdateScene_( void )
{
    UpdateProcedureTextureArray();
}

//------------------------------------------------------------------------------
void ProcedureTextureScene::RenderSceneGr_( const s32 display,
                                            const nn::math::MTX44& proj, const nn::math::MTX34& view )
{
    // Get a pointer to the render command buffer
    u32* command = NULL;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );

    // Get the start address (so that the pointer can be advanced at the end)
    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Shader binary settings ( glUseProgram() )
    command = shader.MakeFullCommand( command );

    nn::gr::RenderState render_state;
    render_state.wBuffer.wScale = 0.0f;
    render_state.blend.isEnable = true;
    render_state.blend.eqRgb = PICA_DATA_BLEND_EQUATION_ADD;
    render_state.blend.eqAlpha  = PICA_DATA_BLEND_EQUATION_ADD;
    render_state.blend.srcRgb   = PICA_DATA_BLEND_FUNC_SRC_ALPHA;
    render_state.blend.dstRgb   = PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_ALPHA;
    render_state.blend.srcAlpha = PICA_DATA_BLEND_FUNC_SRC_ALPHA;
    render_state.blend.dstAlpha = PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_ALPHA;;

    command = render_state.MakeCommand( command );

    command = DrawProcedureTexture( command, display, proj, view );  
         
    // Advance the pointer to the render command buffer
    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );
}

void UpdateEulerAngle( const u32 direction, nn::math::VEC3& eulerAngleDeg )
{
    const f32 deg_step = 2.0f;
     
    if ( direction == 0 )
    {
        eulerAngleDeg.y -= deg_step;
    }
    else if ( direction == 1 )
    {
        eulerAngleDeg.y += deg_step;
    }
    else if ( direction == 2 )
    {
        eulerAngleDeg.x -= deg_step;
    }
    else if ( direction == 3 )
    {
        eulerAngleDeg.x += deg_step;
    }
    else if ( direction == 4 )
    {
        eulerAngleDeg.z -= deg_step;
    }
    else if ( direction == 5 )
    {
        eulerAngleDeg.z += deg_step;
    }

    if ( ( eulerAngleDeg.x == ( - 360.0f ) ) || ( eulerAngleDeg.x == 360.0f ) )
    {
        eulerAngleDeg.x = 0.0f;
    }
     
    if ( ( eulerAngleDeg.y == ( - 360.0f ) ) || ( eulerAngleDeg.y == 360.0f ) )
    {
        eulerAngleDeg.y = 0.0f;
    }

    if ( ( eulerAngleDeg.z == ( - 360.0f ) ) || ( eulerAngleDeg.z == 360.0f ) )
    {        
        eulerAngleDeg.z = 0.0f;
    }
}

void ProcedureTextureScene::UpdateProcedureTextureArray( void )
{
    static u32 counter = 0;
    const u32 thre     = 240;

    const f32 deg_step = 2.0f;
    const f32 deg_count = 90 / deg_step;
    const u32 deg_thre = 140 - deg_count;

    static u32 rotateDirection0 = 0;
    static u32 rotateDirection1 = 0;

    bool resetEulerAngle = false;

    // Randomly determine the rotation direction.
    if ( ( counter % thre ) == deg_thre )
    {
        rotateDirection0 = ::std::rand() % 2;        
        rotateDirection1 = ::std::rand() % 2;        
    }

    // Rotates the cuboid.
    if ( ( ( counter % thre ) >= deg_thre ) &&
         ( ( counter % thre ) < ( deg_thre + deg_count ) ) )
    {   
        UpdateEulerAngle( rotateDirection0, m_EulerAngleDeg0 );
        UpdateEulerAngle( rotateDirection1, m_EulerAngleDeg1 );
    }

    // Change the procedural texture pattern.  
    if ( counter == ( thre - 1 ) )
    {
        m_UpperProcedurePatternIndex = PROCTEX_CLOUD_PATTERN;
        m_LowerProcedurePatternIndex = PROCTEX_MUD_PATTERN;

        resetEulerAngle = true;
    }
    else if ( counter == ( 2 * thre - 1 ) )
    {
        m_UpperProcedurePatternIndex = PROCTEX_THUNDER_PATTERN;
        m_LowerProcedurePatternIndex = PROCTEX_FIRE_PATTERN;
                      
        resetEulerAngle = true;
    }
    else if ( counter == ( 3 * thre - 1 ) )
    {
        m_UpperProcedurePatternIndex = PROCTEX_SEA_PATTERN;
        m_LowerProcedurePatternIndex = PROCTEX_CLOUD_PATTERN;

        resetEulerAngle = true;
    }
    else if ( counter == ( 4 * thre - 1 ) )
    {
        m_UpperProcedurePatternIndex = PROCTEX_MUD_PATTERN;
        m_LowerProcedurePatternIndex = PROCTEX_THUNDER_PATTERN;

        resetEulerAngle = true;
    }
    else if ( counter == ( 5 * thre - 1 ) )
    {
        m_UpperProcedurePatternIndex = PROCTEX_FIRE_PATTERN;
        m_LowerProcedurePatternIndex = PROCTEX_SEA_PATTERN;

        resetEulerAngle = true;

        counter = 0;
    }

    counter += 1;

    if ( resetEulerAngle )
    {
        m_EulerAngleDeg0.x = 0.0f;
        m_EulerAngleDeg0.y = 0.0f;
        m_EulerAngleDeg0.z = 0.0f;

        m_EulerAngleDeg1.x = 0.0f;
        m_EulerAngleDeg1.y = 0.0f;
        m_EulerAngleDeg1.z = 0.0f;
    }
       
    u32 phaseCounter = counter % thre;

    for (u32 patternIndex = 0; patternIndex < PROCTEX_PATTERN_NUM; patternIndex++)
    {        
        if ( ( m_UpperProcedurePatternIndex != patternIndex ) &&
             ( m_LowerProcedurePatternIndex != patternIndex ) )
        {
            continue;
        }        

        nn::gr::CTR::ProcedureTexture& procedureTexture = m_ProcedureTextureArray[ patternIndex ];
        if ( ! procedureTexture.isEnableNoise )
        {
            continue;
        }

        if ( m_IsEnableNoiseUArray[ patternIndex ] )
        {
            procedureTexture.noiseUPhase =
                m_NoiseUPhaseStepArray[ patternIndex ] * phaseCounter;
        }

        if ( m_IsEnableNoiseVArray[ patternIndex ] )
        {
            procedureTexture.noiseVPhase =
                m_NoiseVPhaseStepArray[ patternIndex ] * phaseCounter;
        }
    }
}

bit32* ProcedureTextureScene::DrawProcedureTexture( bit32* command, const s32 display,
                                                    const nn::math::MTX44& proj, const nn::math::MTX34& view )
{
    nn::math::MTX34Identity( &mtx );
    mtx.f._03 = m_Pos.x;
    mtx.f._13 = m_Pos.y;
    mtx.f._23 = m_Pos.z;

    nn::math::MTX34 objRot;
    nn::math::MTX34Identity( &objRot );

    if ( ( display == NN_GX_DISPLAY0 ) || ( display == NN_GX_DISPLAY0_EXT ) )
    {
        nn::math::MTX34RotXYZDeg( &objRot, 
            m_EulerAngleDeg0.x, m_EulerAngleDeg0.y, m_EulerAngleDeg0.z );
    }
    else
    {
        nn::math::MTX34RotXYZDeg( &objRot, 
            m_EulerAngleDeg1.x, m_EulerAngleDeg1.y, m_EulerAngleDeg1.z );
    }

    nn::math::MTX34Mult(&mtx, &mtx, &objRot);

    // Set model matrix
    {
        // Set Shader's Uniform
        command = bindSymbolWorldMtx.MakeUniformCommand( command, mtx );
        command = bindSymbolProjMtx.MakeUniformCommand( command, proj );
        command = bindSymbolViewMtx.MakeUniformCommand( command, view );
    }

    // Replace the texture combiner settings using a procedural texture

    nn::gr::Combiner combiner;

    // glUniform1i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    combiner.stage[ 0 ].rgb.combine         = PICA_DATA_TEX_ENV_COMBINE_REPLACE;

    // glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcRgb"), 
    //    GL_TEXTURE3, GL_CONSTANT, GL_CONSTANT);
    combiner.stage[ 0 ].rgb.source[ 0 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE3;
    combiner.stage[ 0 ].rgb.source[ 1 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT;
    combiner.stage[ 0 ].rgb.source[ 2 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT;
    
    // glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].operandRgb"), 
    //    GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    combiner.stage[ 0 ].rgb.operand[ 0 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
    combiner.stage[ 0 ].rgb.operand[ 1 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
    combiner.stage[ 0 ].rgb.operand[ 2 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;

    // glUniform1i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    combiner.stage[ 0 ].alpha.combine     = PICA_DATA_TEX_ENV_COMBINE_REPLACE;

    // glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcAlpha"), 
    //    GL_TEXTURE3, GL_CONSTANT, GL_CONSTANT);
    combiner.stage[ 0 ].alpha.source[ 0 ] = PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE3;
    combiner.stage[ 0 ].alpha.source[ 1 ] = PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT;
    combiner.stage[ 0 ].alpha.source[ 2 ] = PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT;

    // glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].operandAlpha"), 
    //    GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    combiner.stage[ 0 ].alpha.operand[ 0 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
    combiner.stage[ 0 ].alpha.operand[ 1 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
    combiner.stage[ 0 ].alpha.operand[ 2 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;

    command = combiner.MakeCommand( command );

    // Texture settings
    nn::gr::Texture texture;
    texture.unit3.texType = PICA_DATA_TEXTURE3_SAMPLER_TYPE_TEXTURE_PROCEDURAL_DMP;
    texture.unit3.texCoord = PICA_DATA_TEXTURE3_TEXCOORD_TEXTURE0;

    // Generate 3 dummy commands to 0x80    
    const bool isAddDummyCommand = true;
    command = texture.MakeCommand( command, isAddDummyCommand );

    u32                          patternIndex = 0;
    PrimitiveBox*                primitiveBox = NULL;
    nn::gr::Vertex*              vertex       = NULL;
    nn::gr::Vertex::IndexStream* index_stream = NULL;

    if ( ( display == NN_GX_DISPLAY0 ) || ( display == NN_GX_DISPLAY0_EXT ) )
    {
        patternIndex = m_UpperProcedurePatternIndex;
        primitiveBox = m_UpperPrimitiveBox;
        vertex       = &m_UpperVertexBox;
        index_stream = &m_UpperIndexStreamBox;
    }
    else
    {
        patternIndex = m_LowerProcedurePatternIndex;
        primitiveBox = m_LowerPrimitiveBox;
        vertex       = &m_LowerVertexBox;
        index_stream = &m_LowerIndexStreamBox;
    }

    // Issue the procedural texture command
    nn::gr::CTR::ProcedureTexture& procedureTexture = m_ProcedureTextureArray[ patternIndex ];
    // Dummy command to 0x80 is not needed because it was just generated.
    command = procedureTexture.MakeAllCommand( command, (! isAddDummyCommand ) );

    // Disable all vertex attributes before starting to use vertices
    command = nn::gr::Vertex::MakeDisableCommand( command );

    // Wait for Draw to finish before manipulating vertices (for double command lists)
    if( display == NN_GX_DISPLAY0 )
    {
        // Here, nngxWaitCmdlistDone is used to wait for all command requests to complete, but by using nngxSetCmdlistCallback, you can wait for any specific command request to complete.
        // 
        nngxWaitCmdlistDone();
    }

    // Adjust Box scale
    primitiveBox->SetPosScale( 10.0f, 10.0f, 10.0f );
    primitiveBox->SetTexCoord( m_TexCoordUArray[ patternIndex ][ 0 ],
        m_TexCoordVArray[ patternIndex ][ 0 ],
        m_TexCoordUArray[ patternIndex ][ 1 ],
        m_TexCoordVArray[ patternIndex ][ 1 ] );

    // glBindBuffer( GL_ARRAY_BUFFER )
    command = vertex->MakeEnableAttrCommand( command );
    // glBindBuffer( GL_ELEMENT_ARRAY_BUFFER ) + glDrawElements()
    command = vertex->MakeDrawCommand( command, *index_stream );

    // Disable all vertex attributes when finished using vertex
    command = nn::gr::CTR::Vertex::MakeDisableCommand( command );

    return command;
}
