﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_PrimitiveVertexStream.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include "gr_PrimitiveVertexStream.h"

//------------------------------------------------------------------------------
void PrimitiveModel::InitializeTorus( IAllocator* _allocator,
                                      const GLenum _geometryMode, 
                                      const VertexAttribute vertexAttribute,
                                      const ArrayType arrayType,
                                      const u32 horizontalPolygonCount,
                                      const u32 verticalPolygonCount )
{
    NN_NULL_ASSERT( _allocator );
    
    NN_ASSERT( vertexAttribute & POSITION_ATTRIBUTE );
    NN_ASSERT( vertexAttribute & COLOR_ATTRIBUTE );
    NN_ASSERT( vertexAttribute & NORMAL_ATTRIBUTE );
    
    NN_ASSERT( arrayType == INDEPENDENT_ARRAY );

    NN_ASSERT( ( _geometryMode == TRIANGLES ) || 
               ( _geometryMode == TRIANGLES_SILHOUETTE_DMP ) );

    allocator        = _allocator;
     
    u32 vertexSize   = 0;   
    u32 posOffset    = 0;
    u32 colorOffset  = 0;
    u32 normalOffset = 0;

    u32 posSize      = sizeof( Position3f );
    u32 colorSize    = sizeof( Color4f );
    u32 normalSize   = sizeof( Normal3f );

    geometryMode = _geometryMode;

    if ( vertexAttribute & POSITION_ATTRIBUTE )
    {
        posOffset    = vertexSize;
        vertexSize  += posSize;
    }

    if ( vertexAttribute & COLOR_ATTRIBUTE )
    {
        colorOffset  = vertexSize;
        vertexSize  += colorSize;
    }

    if ( vertexAttribute & NORMAL_ATTRIBUTE )
    {
        normalOffset = vertexSize;
        vertexSize  += normalSize;
    }
    
    s32 vertexXNum = horizontalPolygonCount;
    s32 vertexYNum = verticalPolygonCount;

    u32 vertexArraySize   = vertexSize   * ( vertexXNum + 1 ) * (vertexYNum + 1 );
    u32 positionArraySize = posSize  * ( vertexXNum + 1 ) * (vertexYNum + 1 );
    u32 colorArraySize    = colorOffset  * ( vertexXNum + 1 ) * (vertexYNum + 1 );
    u32 normalArraySize   = normalOffset * ( vertexXNum + 1 ) * (vertexYNum + 1 );

    pVertexArray = 
        static_cast<unsigned char*>( allocator->Alloc( vertexArraySize ) );

    pPositionArray = reinterpret_cast<uptr>( pVertexArray );
    pColorArray    = reinterpret_cast<uptr>( pVertexArray ) + colorArraySize;
    pNormalArray   = reinterpret_cast<uptr>( pVertexArray ) + normalArraySize;

    {
        const f32 radX = 0.4f;
        const f32 radY = 1.f;

        for ( u32 vy = 0; vy <= vertexYNum; vy++ )
        {
            f32 siny = 0.0f;
            f32 cosy = 0.0f;

            nn::math::SinCosDeg( &siny, &cosy, 
                ( 360.0f / vertexYNum ) * ( vy % vertexYNum ) );

            for ( u32 vx = 0; vx <= vertexXNum; vx++ )
            {
                f32 sinx = 0.0f;
                f32 cosx = 0.0f;

                nn::math::SinCosDeg( &sinx, &cosx, 
                    ( 360.0f / vertexXNum ) * ( vx % vertexXNum ) );

                if ( arrayType == INDEPENDENT_ARRAY )
                {
                    u32 offset = 0;

                    if ( vertexAttribute & POSITION_ATTRIBUTE )
                    {
                        offset = 
                            ( vy * ( vertexXNum + 1 ) + vx ) * sizeof( Position3f ) + 
                                 posOffset * ( vertexXNum + 1 ) * ( vertexYNum + 1 );

                        Position3f* pPosition =
                            reinterpret_cast<Position3f*>( &(pVertexArray[ offset ] ) );

                        pPosition->x = cosy * ( radY + cosx * radX );
                        pPosition->y = siny * ( radY + cosx * radX );
                        pPosition->z = - sinx * radX;
                    }

                    if ( vertexAttribute & COLOR_ATTRIBUTE )
                    {
                        offset = 
                            ( vy * ( vertexXNum + 1 ) + vx ) * sizeof( Color4f ) +
                                colorOffset * ( vertexXNum + 1 ) * ( vertexYNum + 1 );

                        Color4f* pColor =
                            reinterpret_cast<Color4f*>( &(pVertexArray[ offset ] ) );

                        pColor->r = 0.5f + cosx * 0.5f;
                        pColor->g = 0.5f - cosx * 0.5f;
                        pColor->b = 0.5f + cosy * 0.5f;
                        pColor->a = 1.0f;
                    }

                    if ( vertexAttribute & NORMAL_ATTRIBUTE )
                    {
                        f32 norm[ 3 ];
                        norm[ 0 ] = cosy * ( radY + cosx * radX ) - radY * cosy;
                        norm[ 1 ] = siny * ( radY + cosx * radX ) - radY * siny;
                        norm[ 2 ] = - sinx * radX;

                        f32 length = 
                            nn::math::FSqrt( norm[ 0 ] * norm[ 0 ] + norm[ 1 ] * norm[ 1 ] + norm[ 2 ] * norm[ 2 ] );

                        offset = 
                            ( vy * ( vertexXNum + 1 ) + vx ) * sizeof( Normal3f ) + 
                                normalOffset * ( vertexXNum + 1 ) * ( vertexYNum + 1 );

                        Normal3f* pNormal = 
                            reinterpret_cast<Normal3f*>( &(pVertexArray[ offset ] ) );
                        pNormal->x = norm[ 0 ] / length;
                        pNormal->y = norm[ 1 ] / length;
                        pNormal->z = norm[ 2 ] / length;
                    }
                }
            }
        }
    }



    u32 indexArraySize = 0;
    u32 twnIndexArraySize = 0;
    {
        indexCount = 0;

        // For geometryMode == TRIANGLES
        {
            indexArraySize = sizeof( u16 ) * vertexXNum * vertexYNum * 6;
            pIndexArray = 
                static_cast<u16*>( allocator->Alloc( indexArraySize ) );
            std::memset( pIndexArray, 0, indexArraySize );

            for ( u32 vy = 0; vy < vertexYNum; vy++ )
            {
                for ( u32 vx = 0; vx < vertexXNum; vx++ )
                {
                    u16 rect[ 4 ];
                    rect[ 0 ] = vx + vy * ( vertexXNum + 1 );
                    rect[ 1 ] = ( vx + 1 ) + vy * ( vertexXNum + 1 );
                    rect[ 2 ] = vx + (vy + 1) * ( vertexXNum + 1 );
                    rect[ 3 ] = ( vx + 1 ) + ( vy + 1 ) * ( vertexXNum + 1 );
                    
                    pIndexArray[ indexCount++ ] = rect[ 2 ];
                    pIndexArray[ indexCount++ ] = rect[ 0 ];
                    pIndexArray[ indexCount++ ] = rect[ 3 ];
                    pIndexArray[ indexCount++ ] = rect[ 3 ];
                    pIndexArray[ indexCount++ ] = rect[ 0 ];
                    pIndexArray[ indexCount++ ] = rect[ 1 ];
                }
            }
        }

        // For geometryMode == TRIANGLES_SILHOUETTE_DMP
        {
            twnIndexArraySize = sizeof( u16 ) * vertexXNum * vertexYNum * 12;
            twnIndexCount = 0;
            pTwnIndexArray        = static_cast<u16*>( allocator->Alloc( twnIndexArraySize ) );
            std::memset( pTwnIndexArray, 0, twnIndexArraySize );

            for ( u32 vy = 0; vy < vertexYNum; vy++ )
            {
                for ( u32 vx = 0; vx < vertexXNum; vx++ )
                {
                    s32 rect[ 8 ];
                    rect[ 0 ] = vy * ( vertexXNum + 1 ) + vx;
                    rect[ 1 ] = vy * ( vertexXNum + 1 ) + vx + 1;
                    rect[ 2 ] = ( vy + 1 ) * ( vertexXNum + 1 ) + vx;
                    rect[ 3 ] = ( vy + 1 ) * ( vertexXNum + 1 ) + vx + 1;
                    rect[ 4 ] = ( ( vy + vertexYNum -1 ) % vertexYNum ) * ( vertexXNum + 1 ) + vx;
                    rect[ 5 ] = ( vy + 1 ) * ( vertexXNum + 1) + ( ( vx + 2 + vertexXNum ) % vertexXNum );
                    rect[ 6 ] = ( ( vy + 2 + vertexYNum ) % vertexYNum ) * ( vertexXNum + 1 ) + vx + 1;
                    rect[ 7 ] = vy * ( vertexXNum + 1 ) + ( ( vx - 1 + vertexXNum ) % vertexXNum );

                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 0 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 1 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 4 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 3 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 2 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 5 ];

                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 0 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 3 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 1 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 2 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 7 ];
                    pTwnIndexArray[ twnIndexCount++ ] = rect[ 6 ];
                }
            }
        }
    }

    nngxUpdateBuffer( reinterpret_cast<const void*>( pPositionArray ), positionArraySize );
    nngxUpdateBuffer( reinterpret_cast<const void*>( pColorArray ), colorArraySize );
    nngxUpdateBuffer( reinterpret_cast<const void*>( pNormalArray ), normalArraySize );
    nngxUpdateBuffer( reinterpret_cast<const void*>( pIndexArray ), indexArraySize );
    nngxUpdateBuffer( reinterpret_cast<const void*>( pTwnIndexArray ), twnIndexArraySize );
}

void PrimitiveModel::Finalize( void )
{
    NN_NULL_ASSERT( allocator );

    allocator->Free( pVertexArray );

    if ( pIndexArray != NULL )
    {
        allocator->Free( pIndexArray );

        pIndexArray = NULL;
    }

    if ( pTwnIndexArray != NULL )
    {
        allocator->Free( pTwnIndexArray );

        pTwnIndexArray = NULL;
    }
}
