﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_ShadowScene.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/fs.h>

#include <string.h>
#include <math.h>

#include "gr_Framework.h"
#include "gr_ShadowScene.h"

//------------------------------------------------------------------------------
GfxAllocator * GrAllocator( MemoryMgr::AllocatorDevice *heap )
{
    return new GfxAllocator( heap );
}

//------------------------------------------------------------------------------
GraphicSystem::GfxAllocatorFactory  GetGraphicsMemoryMgrFunctor( void )
{
    return &GrAllocator;
}

//------------------------------------------------------------------------------
IScene * CreateScene( void )
{
    return new ShadowScene();
}

//------------------------------------------------------------------------------
ShadowScene::ShadowScene()
    : allocator( NULL ),
      shaderBinary( NULL ),
      shader(),
      mtx(), 
      projectionMtx44(),
      modelViewMtx34(),
      stereo(),
      bindSymbolProjectionMtx44(),
      bindSymbolModelViewMtx44(),
      bindSymbolTextureMtx44(),
      bindSymbolShadowProjectionMtx44(),
      bindSymbolShadowModelViewMtx44(),
      bindSymbolTexProjectionMtx44(),
      bindSymbolTexModelViewMtx44(),
      renderTarget( 0 ), 
      colorBufferVirtualAddress( NULL ),
      depthBufferVirtualAddress( NULL ),
      shadowColorBufferVirtualAddress( NULL ),
      shadowDepthBufferVirtualAddress( NULL ),
      m_Near( .1f ),
      m_Far( 10.f ),
      m_GraphicsMemoryAllocator(),
      m_GraphicsMemoryDeallocator(),
      m_Viewport(),
      m_Scissor(),
      m_FrameBuffer(),
      m_ShadowBuffer(),
      m_RenderState(),
      m_Combiner(),
      m_Light(),
      m_Table(),
      m_Texture(),
      m_pPrimitiveModel( NULL ),
      m_Vertex(),
      m_IndexStream(),
      m_Counter( 0 )
{
    for ( u32 index = 0; index < 2; index++)
    {
        renderBuffer[ index ] = 0;
    }
}

//------------------------------------------------------------------------------

ShadowScene::~ShadowScene()
{
}

//------------------------------------------------------------------------------
bool ShadowScene::Initialize( void )
{
    bool is_success = false;

    NN_UNUSED_VAR( is_success );

    allocator = MEMMGR->GetDefaultAllocatorDevice();

    { 
        // Initialize the shader

        // Load shader binary
        nn::fs::FileReader shader_file_reader( L"rom:/shader.shbin" );
        u32 shader_file_size = static_cast<u32>( shader_file_reader.GetSize() );
        shaderBinary = static_cast<u8*>( allocator->Alloc( shader_file_size, 128 ) );
        shader_file_reader.Read( shaderBinary, shader_file_size );
        shader_file_reader.Finalize();

        // Get size of the buffer needed for a ShaderLite class from the shader binary
        u32 size = shader.GetRequiredBufferSize( shaderBinary );

        shaderBuffer = static_cast<u32*>( allocator->Alloc( size ) );

        // Specify shader binary and initialize
        // Enable the executable image at position 0 as the vertex shader
        shader.SetupBinary( shaderBinary, 1, -1, shaderBuffer );
        
        // Vertex shader uniforms
        is_success = 
            shader.SearchBindSymbol( &bindSymbolProjectionMtx44,
            "uProjection" );
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &bindSymbolModelViewMtx44,
            "uModelView" );
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &bindSymbolTextureMtx44,
            "uTexture" );
        NN_GR_ASSERT( is_success );
    }

    {
        // Initialize primitives
        uptr physicalAddr = NULL;
        nn::gr::CTR::BindSymbolVSInput symbolPos;
        nn::gr::CTR::BindSymbolVSInput symbolNormal;

        m_pPrimitiveModel = 
            static_cast< PrimitiveModel* >( allocator->Alloc( sizeof( PrimitiveModel ), 128 ) );
       
       VertexAttribute vertexAttribute = 
            static_cast<VertexAttribute>(
                static_cast<u32>( POSITION_ATTRIBUTE ) |
                static_cast<u32>( COLOR_ATTRIBUTE ) | 
                static_cast<u32>( NORMAL_ATTRIBUTE ) );

        ArrayType arrayType = INDEPENDENT_ARRAY;

        u32 torusHorizontalPolygonCount = 40;
        u32 torusVerticalPolygonCount   = 40;

        NN_UNUSED_VAR( torusHorizontalPolygonCount );
        NN_UNUSED_VAR( torusVerticalPolygonCount );

        GLenum geometryMode = TRIANGLES;

        m_pPrimitiveModel->InitializeTorus( allocator,
            geometryMode,
            vertexAttribute,
            arrayType,
            torusHorizontalPolygonCount,
            torusVerticalPolygonCount );

        // Search for "location"
        is_success = shader.SearchBindSymbol( &symbolPos, "aPosition" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &symbolNormal, "aNormal" );
        NN_GR_ASSERT( is_success );

        // Vertex stream (Position)   
        physicalAddr = nngxGetPhysicalAddr( m_pPrimitiveModel->pPositionArray );
        NN_NULL_ASSERT( physicalAddr );
        m_Vertex.EnableAttrAsArray( symbolPos, physicalAddr, PICA_DATA_SIZE_3_FLOAT );

        // Normal
        physicalAddr = nngxGetPhysicalAddr( m_pPrimitiveModel->pNormalArray );
        NN_NULL_ASSERT( physicalAddr );
        m_Vertex.EnableAttrAsArray( symbolNormal, physicalAddr, PICA_DATA_SIZE_3_FLOAT );

        // Index setting        
        physicalAddr = nngxGetPhysicalAddr( (uptr) m_pPrimitiveModel->pIndexArray );
        NN_NULL_ASSERT( physicalAddr );
        m_IndexStream.physicalAddr = physicalAddr;  
        m_IndexStream.drawVtxNum   = m_pPrimitiveModel->indexCount;

        // Index is defined as u16.
        m_IndexStream.isUnsignedByte = false;
    }

    { 
        // Initialize the shader (shadow pass)

        // Enable the executable image at position 0 as the vertex shader
        shader.SetShaderIndex( 0, -1 );
        
        // Vertex shader uniforms
        is_success = 
            shader.SearchBindSymbol( &bindSymbolShadowProjectionMtx44,
            "uProjection" );
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &bindSymbolShadowModelViewMtx44,
            "uModelView" );
        NN_GR_ASSERT( is_success );
    }
    {
        uptr physicalAddr = NULL;
        nn::gr::CTR::BindSymbolVSInput symbolPos;

        // Search for "location"
        is_success = shader.SearchBindSymbol( &symbolPos, "aPosition" );
        NN_GR_ASSERT( is_success );

        // Vertex stream (Position)   
        physicalAddr = nngxGetPhysicalAddr( m_pPrimitiveModel->pPositionArray );
        NN_NULL_ASSERT( physicalAddr );
        m_Vertex.EnableAttrAsArray( symbolPos, physicalAddr, PICA_DATA_SIZE_3_FLOAT );
    }

    // Get allocator and deallocator
    nngxGetAllocator( &m_GraphicsMemoryAllocator, &m_GraphicsMemoryDeallocator );

    {
        // Allocate color buffer  GL_RGBA8_OES
        renderBuffer[ 0 ] = 1;
        GLenum area = NN_GX_MEM_VRAMA;
        GLenum aim  = NN_GX_MEM_RENDERBUFFER;
        u32 colorBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        colorBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 0 ], colorBufferByteSize ) );

        // Allocate depth buffer  GL_DEPTH24_STENCIL8_EXT
        renderBuffer[ 1 ] = 2;
        area = NN_GX_MEM_VRAMB;
        u32 depthBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        depthBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 1 ], depthBufferByteSize ) );
    }

    {
        renderTarget = 1;

        // Initialize framebuffer settings
        m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

        // Initialize color buffer settings
        m_FrameBuffer.colorBuffer.virtualAddr         = colorBufferVirtualAddress;
        m_FrameBuffer.colorBuffer.format              = PICA_DATA_COLOR_RGBA8_OES;
        m_FrameBuffer.colorBuffer.blockSize           = nn::gr::FrameBuffer::BLOCK_SIZE8;
        m_FrameBuffer.colorBuffer.width               = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.colorBuffer.height              = nn::gx::DISPLAY0_HEIGHT;  
        m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

        // Initialize depth buffer settings
        m_FrameBuffer.depthStencilBuffer.virtualAddr  = depthBufferVirtualAddress;
        m_FrameBuffer.depthStencilBuffer.format       = PICA_DATA_DEPTH24_STENCIL8_EXT;
        m_FrameBuffer.depthStencilBuffer.width        = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.depthStencilBuffer.height       = nn::gx::DISPLAY0_HEIGHT;
        m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
        m_FrameBuffer.depthStencilBuffer.clearStencil = 0;
    }

    {
        // Shadow buffer

        // Allocate color buffer  GL_RGBA8_OES
        GLenum area = NN_GX_MEM_VRAMA;
        GLenum aim  = NN_GX_MEM_RENDERBUFFER;
        u32 colorBufferByteSize = SHADOW_TEXTURE_WIDTH * SHADOW_TEXTURE_HEIGHT * 4;
        shadowColorBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    3, colorBufferByteSize ) );

        // Allocate depth buffer  GL_DEPTH24_STENCIL8_EXT
        area = NN_GX_MEM_VRAMB;
        u32 depthBufferByteSize = SHADOW_TEXTURE_WIDTH * SHADOW_TEXTURE_HEIGHT * 4;
        shadowDepthBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    4, depthBufferByteSize ) );
    }

    {
        // Shadow buffer

        // Initialize framebuffer settings
        m_ShadowBuffer.width                           = SHADOW_TEXTURE_WIDTH;
        m_ShadowBuffer.height                          = SHADOW_TEXTURE_HEIGHT;

        // Initialize color buffer settings
        m_ShadowBuffer.colorBuffer.virtualAddr         = shadowColorBufferVirtualAddress;
        m_ShadowBuffer.colorBuffer.format              = PICA_DATA_COLOR_RGBA8_OES;
        m_ShadowBuffer.colorBuffer.blockSize           = nn::gr::FrameBuffer::BLOCK_SIZE8;
        m_ShadowBuffer.colorBuffer.width               = SHADOW_TEXTURE_WIDTH;
        m_ShadowBuffer.colorBuffer.height              = SHADOW_TEXTURE_HEIGHT;  
        m_ShadowBuffer.colorBuffer.clearColor[ 0 ]     = 1.0f;
        m_ShadowBuffer.colorBuffer.clearColor[ 1 ]     = 1.0f;
        m_ShadowBuffer.colorBuffer.clearColor[ 2 ]     = 1.0f;
        m_ShadowBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

        // Initialize depth buffer settings
        m_ShadowBuffer.depthStencilBuffer.virtualAddr  = shadowDepthBufferVirtualAddress;
        m_ShadowBuffer.depthStencilBuffer.format       = PICA_DATA_DEPTH24_STENCIL8_EXT;
        m_ShadowBuffer.depthStencilBuffer.width        = SHADOW_TEXTURE_WIDTH;
        m_ShadowBuffer.depthStencilBuffer.height       = SHADOW_TEXTURE_HEIGHT;
        m_ShadowBuffer.depthStencilBuffer.clearDepth   = 1.0f;
        m_ShadowBuffer.depthStencilBuffer.clearStencil = 0;
    }

    // Texture settings
    m_Texture.unit0.texType = PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_SHADOW_2D_DMP;
    m_Texture.unit0.width  = SHADOW_TEXTURE_WIDTH;
    m_Texture.unit0.height = SHADOW_TEXTURE_HEIGHT;
    m_Texture.unit0.format = PICA_DATA_TEXTURE_FORMAT_SHADOW_UNSIGNED_INT;
    m_Texture.unit0.wrapT = PICA_DATA_TEXTURE_WRAP_CLAMP_TO_BORDER;
    m_Texture.unit0.wrapS = PICA_DATA_TEXTURE_WRAP_CLAMP_TO_BORDER;
    m_Texture.unit0.magFilter = PICA_DATA_TEXTURE_MAG_FILTER_LINEAR;
    m_Texture.unit0.minFilter = PICA_DATA_TEXTURE_MIN_FILTER_LINEAR;
    m_Texture.unit0.borderColorR = 255;
    m_Texture.unit0.borderColorG = 255;
    m_Texture.unit0.borderColorB = 255;
    m_Texture.unit0.borderColorA = 255;
    m_Texture.unit0.physicalAddr = nngxGetPhysicalAddr( m_ShadowBuffer.colorBuffer.virtualAddr );

    m_Scissor.SetBufferSize( nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT );

    // Render state settings
    m_RenderState.wBuffer.wScale = 0.0f;

    // Shadow settings
    m_RenderState.shadowMap.zBias = m_Near * 2.4f / (m_Far - m_Near);

    // glEnable(GL_DEPTH_TEST);
    // glDepthFunc(GL_LESS);    
    m_RenderState.depthTest.isEnable      = true;
    m_RenderState.depthTest.isEnableWrite = true;

    m_RenderState.cullingTest.isEnable  = true;

    { 
        // Light initialization
        m_Light.source[0].diffuseR = 140;
        m_Light.source[0].diffuseG = 210;
        m_Light.source[0].diffuseB = 227;
        m_Light.source[0].specular0R = 0;
        m_Light.source[0].specular0G = 0;
        m_Light.source[0].specular0B = 0;
        m_Light.source[0].ambientR =  75;
        m_Light.source[0].ambientG =  75;
        m_Light.source[0].ambientB =  75;
        m_Light.isEnable[0] = true;

        nn::math::VEC3 vec = nn::math::VEC3(LIGHT_POS[0], LIGHT_POS[1], LIGHT_POS[2]);
        m_Light.source[0].SetPosition(vec, true);

        // Layer config 0
        m_Light.layerConfig = PICA_DATA_FRAG_LIGHT_ENV_LAYER_CONFIG0;

        // Configure a table for shadow attenuation
        f32 lut[256];
        memset(lut, 0, sizeof(lut));

        for (int i = 1; i < 256; i++)
        {
            if( i < 151 )lut[i] = powf((float)i / 150.f, 2.0f);
            else lut[i] = 0.f;
        }

        // Set the table
        m_Table.SetTable( lut );
    }

    {
        // Initialize the texture combiner(s)

        m_Combiner.stage[ 0 ].constColorR = 0;
        m_Combiner.stage[ 0 ].constColorG = 0;
        m_Combiner.stage[ 0 ].constColorB = 0;
        m_Combiner.stage[ 0 ].constColorA = 255;

        m_Combiner.stage[ 0 ].rgb.combine         = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
        m_Combiner.stage[ 0 ].rgb.source[ 0 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].rgb.source[ 1 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].rgb.source[ 2 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;

        m_Combiner.stage[ 0 ].rgb.operand[ 0 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
        m_Combiner.stage[ 0 ].rgb.operand[ 1 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
        m_Combiner.stage[ 0 ].rgb.operand[ 2 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;

        m_Combiner.stage[ 0 ].alpha.combine     = PICA_DATA_TEX_ENV_COMBINE_REPLACE;

        m_Combiner.stage[ 0 ].alpha.source[ 0 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].alpha.source[ 1 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].alpha.source[ 2 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;

        m_Combiner.stage[ 0 ].alpha.operand[ 0 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
        m_Combiner.stage[ 0 ].alpha.operand[ 1 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
        m_Combiner.stage[ 0 ].alpha.operand[ 2 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
    }

    return true;
}

//------------------------------------------------------------------------------
void ShadowScene::Finalize( void )
{
    m_Vertex.DisableAll();

    m_pPrimitiveModel->Finalize();

    allocator->Free( m_pPrimitiveModel );

    allocator->Free( shaderBinary );
    allocator->Free( shaderBuffer );

    GLenum area = NN_GX_MEM_VRAMA;
    GLenum aim  = NN_GX_MEM_RENDERBUFFER;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 0 ], reinterpret_cast<void*>( colorBufferVirtualAddress ) );
    colorBufferVirtualAddress = 0;

    area = NN_GX_MEM_VRAMB;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 1 ], reinterpret_cast<void*>( depthBufferVirtualAddress ) );
    depthBufferVirtualAddress = 0;

    area = NN_GX_MEM_VRAMA;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 0 ], reinterpret_cast<void*>( shadowColorBufferVirtualAddress ) );
    colorBufferVirtualAddress = 0;

    area = NN_GX_MEM_VRAMB;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 1 ], reinterpret_cast<void*>( shadowDepthBufferVirtualAddress ) );
    depthBufferVirtualAddress = 0;
}

//------------------------------------------------------------------------------
void ShadowScene::Enter( void )
{
    stereo.Initialize();
    nngxSetDisplayMode( NN_GX_DISPLAYMODE_STEREO );
    Calc();
}

//------------------------------------------------------------------------------
bool ShadowScene::Calc( void )
{
    UpdateScene_();

    m_Counter++;

    return true;
}

//------------------------------------------------------------------------------
void ShadowScene::PreDraw( void )
{
    f32 width  = SHADOW_TEXTURE_WIDTH;
    f32 height = SHADOW_TEXTURE_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );

    m_ShadowBuffer.colorBuffer.clearColor[ 0 ] = 1.f;
    m_ShadowBuffer.colorBuffer.clearColor[ 1 ] = 1.f;
    m_ShadowBuffer.colorBuffer.clearColor[ 2 ] = 1.f;
    m_ShadowBuffer.colorBuffer.clearColor[ 3 ] = 1.f;

    m_ShadowBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_ShadowBuffer.depthStencilBuffer.clearStencil = 0;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit, 1 );

    m_RenderState.wBuffer.wScale = 1.0f / m_Far;

    m_RenderState.shadowMap.isEnable = true;

    // Matrix settings
    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.1f, 0.1f, -0.1f, 0.1f, m_Near, m_Far);

    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(LIGHT_POS[0], LIGHT_POS[1], LIGHT_POS[2]);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);

    // Texture settings
    m_Texture.unit0.texType = PICA_DATA_TEXTURE0_SAMPLER_TYPE_FALSE;

    // light settings
    m_Light.isEnable[0]  = false;
    m_Light.isShadowed[0] = false;

    // Texture combiner settings
    m_Combiner.stage[ 0 ].rgb.combine         = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
    m_Combiner.stage[ 0 ].rgb.source[ 0 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT;
    m_Combiner.stage[ 0 ].rgb.source[ 1 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
    m_Combiner.stage[ 0 ].rgb.source[ 2 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;

    m_Combiner.stage[ 0 ].rgb.operand[ 0 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
    m_Combiner.stage[ 0 ].rgb.operand[ 1 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
    m_Combiner.stage[ 0 ].rgb.operand[ 2 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;

    m_Combiner.stage[ 0 ].alpha.combine       = PICA_DATA_TEX_ENV_COMBINE_REPLACE;

    m_Combiner.stage[ 0 ].alpha.source[ 0 ]   = PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT;
    m_Combiner.stage[ 0 ].alpha.source[ 1 ]   = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
    m_Combiner.stage[ 0 ].alpha.source[ 2 ]   = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;

    m_Combiner.stage[ 0 ].alpha.operand[ 0 ]  = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
    m_Combiner.stage[ 0 ].alpha.operand[ 1 ]  = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
    m_Combiner.stage[ 0 ].alpha.operand[ 2 ]  = PICA_DATA_OPE_ALPHA_SRC_ALPHA;

    RenderShadowTexture_( proj, mv );
}

//------------------------------------------------------------------------------
GLuint ShadowScene::DrawTop( void )
{
    f32 width  = nn::gx::DISPLAY0_WIDTH;
    f32 height = nn::gx::DISPLAY0_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );
      
    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;
    
    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    // Render state settings
    m_RenderState.wBuffer.wScale = 0.f;
    m_RenderState.shadowMap.isEnable = false;

    // Combiner settings
    m_Combiner.stage[ 0 ].rgb.combine         = PICA_DATA_TEX_ENV_COMBINE_ADD_MULT_DMP;
    m_Combiner.stage[ 0 ].rgb.source[ 0 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_PRIMARY_COLOR_DMP;
    m_Combiner.stage[ 0 ].rgb.source[ 1 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
    m_Combiner.stage[ 0 ].rgb.source[ 2 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_PRIMARY_COLOR_DMP;

    m_Combiner.stage[ 0 ].rgb.operand[ 0 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
    m_Combiner.stage[ 0 ].rgb.operand[ 1 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
    m_Combiner.stage[ 0 ].rgb.operand[ 2 ]    = PICA_DATA_OPE_RGB_ONE_MINUS_SRC_ALPHA;

    m_Combiner.stage[ 0 ].alpha.combine       = PICA_DATA_TEX_ENV_COMBINE_REPLACE;

    m_Combiner.stage[ 0 ].alpha.source[ 0 ]   = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
    m_Combiner.stage[ 0 ].alpha.source[ 1 ]   = PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT;
    m_Combiner.stage[ 0 ].alpha.source[ 2 ]   = PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT;

    m_Combiner.stage[ 0 ].alpha.operand[ 0 ]  = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
    m_Combiner.stage[ 0 ].alpha.operand[ 1 ]  = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
    m_Combiner.stage[ 0 ].alpha.operand[ 2 ]  = PICA_DATA_OPE_ALPHA_SRC_ALPHA;

    // Texture settings
    m_Texture.unit0.texType = PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_SHADOW_2D_DMP;

    // Light settings
    m_Light.isEnable[0]   = true;
    m_Light.isShadowed[0] = true;
    m_Light.lutConfigRR.input = PICA_DATA_FRAG_LIGHT_ENV_NH_DMP;
    m_Light.lutConfigFR.input = PICA_DATA_FRAG_LIGHT_ENV_LN_DMP;
    m_Light.layerConfig = PICA_DATA_FRAG_LIGHT_ENV_LAYER_CONFIG1;
    m_Light.fresnelSelector = PICA_DATA_FRAG_LIGHT_ENV_PRI_SEC_ALPHA_FRESNEL;
    m_Light.isEnableShadowAlpha = true;
    m_Light.isInvertShadow      = true;

    {
        // The camera matrix        
        modelViewMtx34.SetupLookAt(
            // camPos
            nn::math::VEC3( 0.0f, 0.0f, 7.5f ),
            // camUp
            nn::math::VEC3( 0.0f, 1.0f, 0.0f ),
            // target
            nn::math::VEC3( 0.0f, 0.0f, 0.0f ) );
    }

    // Projection matrix    
    projectionMtx44.SetupPerspective(
        0.66f,          // fovy
        height / width, // aspect
        0.1f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR,
        6.5f, 1.f, nn::math::PIVOT_UPSIDE_TO_TOP );

    RenderSceneGr_( NN_GX_DISPLAY0, projL, viewL );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint ShadowScene::DrawExt( void )
{
    f32 width  = nn::gx::DISPLAY0_WIDTH; 
    f32 height = nn::gx::DISPLAY0_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );

    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR,
        6.5f, 1.f, nn::math::PIVOT_UPSIDE_TO_TOP);
    
    RenderSceneGr_( NN_GX_DISPLAY0_EXT, projR, viewR );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint ShadowScene::DrawBtm( void )
{
    f32 width  = nn::gx::DISPLAY1_WIDTH; 
    f32 height = nn::gx::DISPLAY1_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );
 
    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    return renderTarget;
}

//------------------------------------------------------------------------------
void ShadowScene::PostDraw( void )
{
}

//------------------------------------------------------------------------------
nn::gr::CTR::FrameBuffer::ColorBuffer* ShadowScene::GetGrColorBufferPointer( void )
{
    return & m_FrameBuffer.colorBuffer;
}


//------------------------------------------------------------------------------
void ShadowScene::Clear( const u32 bufferBit, const s32 index )
{
    u32* command = 0;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
        reinterpret_cast<GLint*>( &command ) );

    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Framebuffer settings
    if(index == 0)
    {
        command = m_FrameBuffer.MakeCommand( command, bufferBit );
    }
    else
    {
        command = m_ShadowBuffer.MakeCommand( command, bufferBit );
    }

    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );

    // Add a load complete command at the end of the 3D command buffer
    bool isAddSplitDrawCmdlist = true;

    // Generate a request to clear the color buffer and depth buffer
    if(index == 0)
    {
        m_FrameBuffer.MakeClearRequest( bufferBit, isAddSplitDrawCmdlist );
    }
    else
    {
        m_ShadowBuffer.MakeClearRequest( bufferBit, isAddSplitDrawCmdlist );
    }
}

//------------------------------------------------------------------------------

void ShadowScene::UpdateScene_( void )
{
}

//------------------------------------------------------------------------------

void ShadowScene::RenderShadowTexture_( const nn::math::MTX44& proj, const nn::math::MTX34& view )
{
    NN_UNUSED_VAR( proj );
    NN_UNUSED_VAR( view );

    // Get a pointer to the render command buffer
    u32* command = NULL;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );

    // Get the start address (so that the pointer can be advanced at the end)
    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Clear the framebuffer cache
    command = nn::gr::CTR::FrameBuffer::MakeClearCacheCommand( command );

    // Render state settings
    command = m_RenderState.MakeCommand( command );

    // Issue the light command
    command = m_Light.MakeAllCommand( command );

    // Issue the table command(s)
    command = m_Table.MakeCommand( command, PICA_DATA_SAMPLER_FR );

    // Texture combiner settings
    command = m_Combiner.MakeCommand( command );

    // Texture settings
    command = MakeTextureCommand_( command );

    static bool isFirst = true;

    command = MakeShadowCommand_( command, proj, view, isFirst );

    isFirst = false;

    // Advance the pointer to the render command buffer
    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof( u32 ) );

    nngxSplitDrawCmdlist();
}

//------------------------------------------------------------------------------

bit32* ShadowScene::MakeTextureCommand_( bit32* command )
{
    // Generate 3 dummy commands to 0x80    
    const bool isAddDummyCommand = true;

    command = m_Texture.MakeCommand( command, isAddDummyCommand );

    return command;
}

//------------------------------------------------------------------------------
void ShadowScene::RenderSceneGr_( const s32 display,
                                      const nn::math::MTX44& proj, const nn::math::MTX34& view )
{
    NN_UNUSED_VAR( display );
    NN_UNUSED_VAR( proj );
    NN_UNUSED_VAR( view );

    // Get a pointer to the render command buffer
    u32* command = NULL;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );

    // Get the start address (so that the pointer can be advanced at the end)
    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Clear the framebuffer cache
    command = nn::gr::CTR::FrameBuffer::MakeClearCacheCommand( command );

    // Render state settings
    command = m_RenderState.MakeCommand( command );

    // Issue the light command
    command = m_Light.MakeAllCommand( command );

    // Issue the table command(s)
    command = m_Table.MakeCommand( command, PICA_DATA_SAMPLER_FR );

    // Texture combiner settings
    command = m_Combiner.MakeCommand( command );

    // Texture settings
    command = MakeTextureCommand_( command );

    static bool isFirst = true;

    command = MakeDrawModelCommand_( command, proj, view, isFirst );

    isFirst = false;

    // Advance the pointer to the render command buffer
    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof( u32 ) );
}
//------------------------------------------------------------------------------
bit32* ShadowScene::MakeShadowCommand_( bit32* command,
                                               const nn::math::MTX44& proj,
                                               const nn::math::MTX34& view,
                                               const bool isMakeFullShaderCommand )
{
    nn::gr::ShaderLite* pShader = &shader;

    // Shader binary setting
    pShader->SetShaderIndex( 0, -1 );
    pShader->SetDrawMode( PICA_DATA_DRAW_TRIANGLES ); 

    if ( isMakeFullShaderCommand )
    {
        command = pShader->MakeFullCommand( command );
    }
    else
    {
        command = pShader->MakeShaderCommand( command );
    }

    // Index stream settings
    {
        uptr physicalAddr = nngxGetPhysicalAddr( (uptr) m_pPrimitiveModel->pIndexArray );
        NN_NULL_ASSERT( physicalAddr );
        m_IndexStream.physicalAddr = physicalAddr;  

        m_IndexStream.drawVtxNum   = m_pPrimitiveModel->indexCount;
    }

    // Rotate the model
    nn::math::Matrix34 rotMtx34, transMtx34;
    nn::math::Matrix34 modelViewMtx34;
    nn::math::MTX34RotXYZDeg( &rotMtx34, 2.0f * m_Counter + 90.0f, 0.0f, 0.0f );
    nn::math::MTX34Translate( &transMtx34, nn::math::VEC3( 0.5f, 0.0f, 0.0f));
    nn::math::MTX34Mult( &modelViewMtx34, &view, &rotMtx34 );
    nn::math::MTX34Mult( &modelViewMtx34, &modelViewMtx34, &transMtx34 );
    nn::math::Matrix44 modelViewMtx44( modelViewMtx34 );

    // Sets the uniform of the vertex shader
    command = bindSymbolShadowProjectionMtx44.MakeUniformCommand( command, proj );
    command = bindSymbolShadowModelViewMtx44.MakeUniformCommand( command, modelViewMtx44 );

    // Render kick
    command = MakeDrawCommand( command );

    // Render the second model
    nn::math::MTX34RotXYZDeg( &rotMtx34, 2.0f * m_Counter, 0.0f, 0.0f );
    nn::math::MTX34Translate( &transMtx34, nn::math::VEC3( -0.5f, 0.0f, 0.0f));
    nn::math::MTX34Mult( &modelViewMtx34, &view, &rotMtx34 );
    nn::math::MTX34Mult( &modelViewMtx34, &modelViewMtx34, &transMtx34 );
    modelViewMtx44 = nn::math::MTX44( modelViewMtx34 );

    // Sets the uniform of the vertex shader
    command = bindSymbolShadowProjectionMtx44.MakeUniformCommand( command, proj );
    command = bindSymbolShadowModelViewMtx44.MakeUniformCommand( command, modelViewMtx44 );

    // Render kick
    command = MakeDrawCommand( command );

    return command;
}

//------------------------------------------------------------------------------
bit32* ShadowScene::MakeDrawModelCommand_( bit32* command,
                                               const nn::math::MTX44& proj,
                                               const nn::math::MTX34& view,
                                               const bool isMakeFullShaderCommand )
{
    nn::gr::ShaderLite* pShader = &shader;

    // Shader binary setting
    pShader->SetShaderIndex( 1, -1 );
    pShader->SetDrawMode( PICA_DATA_DRAW_TRIANGLES ); 

    if ( isMakeFullShaderCommand )
    {
        command = pShader->MakeFullCommand( command );
    }
    else
    {
        command = pShader->MakeShaderCommand( command );
    }
            
    // Rotate the model
    nn::math::Matrix34 rotMtx34, transMtx34;
    nn::math::Matrix34 modelViewMtx34;
    nn::math::MTX34RotXYZDeg( &rotMtx34, 2.0f * m_Counter + 90.0f, 0.0f, 0.0f );
    nn::math::MTX34Translate( &transMtx34, nn::math::VEC3( 0.5f, 0.0f, 0.0f));
    nn::math::MTX34Mult( &modelViewMtx34, &view, &rotMtx34 );
    nn::math::MTX34Mult( &modelViewMtx34, &modelViewMtx34, &transMtx34 );
    nn::math::Matrix44 modelViewMtx44( modelViewMtx34 );

    // Shadow texture matrix
    nn::math::Matrix34 textmp;
    nn::math::MTX34ShadowProjectionFrustum(&textmp, -0.1f, 0.1f, -0.1f, 0.1f, m_Near, m_Far);
    nn::math::Matrix44 tex(textmp);

    nn::math::Vector3 camPos(LIGHT_POS[0], LIGHT_POS[1], LIGHT_POS[2]);
    nn::math::Vector3 camUp( 0.0f, 1.0f, 0.0f );
    nn::math::Vector3 target( 0.0f, 0.0f, 0.0f );

    nn::math::Matrix34 tmpmv;
    nn::math::MTX34LookAt(&tmpmv, &camPos, &camUp, &target);
    nn::math::MTX34Mult(&tmpmv, &tmpmv, &rotMtx34);
    nn::math::MTX34Mult(&tmpmv, &tmpmv, &transMtx34);
    nn::math::Matrix44 tmp(tmpmv);
    nn::math::MTX44Mult(&tex, &tex, &tmp);

    // Index stream settings
    {
        uptr physicalAddr = nngxGetPhysicalAddr( (uptr) m_pPrimitiveModel->pIndexArray );
        NN_NULL_ASSERT( physicalAddr );
        m_IndexStream.physicalAddr = physicalAddr;  

        m_IndexStream.drawVtxNum   = m_pPrimitiveModel->indexCount;
    }

    // Sets the uniform of the vertex shader
    command = bindSymbolProjectionMtx44.MakeUniformCommand( command, proj );
    command = bindSymbolModelViewMtx44.MakeUniformCommand( command, modelViewMtx44 );
    command = bindSymbolTextureMtx44.MakeUniformCommand( command, tex );

    // Render kick
    command = MakeDrawCommand( command );

    // Render the second model
    nn::math::MTX34RotXYZDeg( &rotMtx34, 2.0f * m_Counter, 0.0f, 0.0f );
    nn::math::MTX34Translate( &transMtx34, nn::math::VEC3( -0.5f, 0.0f, 0.0f));
    nn::math::MTX34Mult( &modelViewMtx34, &view, &rotMtx34 );
    nn::math::MTX34Mult( &modelViewMtx34, &modelViewMtx34, &transMtx34 );
    modelViewMtx44 = nn::math::MTX44( modelViewMtx34 );

    tex = nn::math::Matrix44( textmp );

    nn::math::MTX34LookAt(&tmpmv, &camPos, &camUp, &target);
    nn::math::MTX34Mult(&tmpmv, &tmpmv, &rotMtx34);
    nn::math::MTX34Mult(&tmpmv, &tmpmv, &transMtx34);
    tmp = nn::math::Matrix44(tmpmv);
    nn::math::MTX44Mult(&tex, &tex, &tmp);

    // Sets the uniform of the vertex shader
    command = bindSymbolProjectionMtx44.MakeUniformCommand( command, proj );
    command = bindSymbolModelViewMtx44.MakeUniformCommand( command, modelViewMtx34 );
    command = bindSymbolTextureMtx44.MakeUniformCommand( command, tex );

    // Render kick
    command = MakeDrawCommand( command );

    return command;
}

//------------------------------------------------------------------------------

bit32* ShadowScene::MakeDrawCommand( bit32* command )
{
    // Disable all vertex attributes before starting to use vertices
    command = nn::gr::Vertex::MakeDisableCommand( command );

    // glBindBuffer( GL_ARRAY_BUFFER )
    // glBindBuffer( GL_ELEMENT_ARRAY_BUFFER ) + glDrawElements()
    command = m_Vertex.MakeEnableAttrCommand( command );

    command = m_Vertex.MakeDrawCommand( command, m_IndexStream );

    // Disable all vertex attributes when finished using vertex
    command = nn::gr::CTR::Vertex::MakeDisableCommand( command );

    return command;
}
