﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_SilhouetteScene.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/fs.h>

#include "gr_Framework.h"
#include "gr_SilhouetteScene.h"

//------------------------------------------------------------------------------
GfxAllocator * GrAllocator( MemoryMgr::AllocatorDevice *heap )
{
    return new GfxAllocator( heap );
}

//------------------------------------------------------------------------------
GraphicSystem::GfxAllocatorFactory  GetGraphicsMemoryMgrFunctor( void )
{
    return &GrAllocator;
}

//------------------------------------------------------------------------------
IScene * CreateScene( void )
{
    return new SilhouetteScene();
}

//------------------------------------------------------------------------------
SilhouetteScene::SilhouetteScene()
    : allocator( NULL ),
      shaderBinary( NULL ),
      shader(),
      mtx(), 
      projectionMtx44(),
      modelViewMtx34(),
      stereo(),
      bindSymbolProjectionMtx44(),
      bindSymbolModelViewMtx44(),
      twnBindSymbolWidthOpenEdgeDepthBiasVec3(),
      twnBindSymbolColorVec4(),
      twnBindSymbolOpenEdgeColorVec4(),
      twnBindSymbolOpenEdgeWidthVec4(),
      twnBindSymbolAcceptEmptyTrianglesBool(),
      twnBindSymbolScaleByWBool(),
      twnBindSymbolFrontFaceCCWBool(),
      twnBindSymbolOpenEdgeWidthScaleByWBool(),
      twnBindSymbolOpenEdgeDepthBiasScaleByWBool(),
      renderTarget( 0 ), 
      colorBufferVirtualAddress( NULL ),
      depthBufferVirtualAddress( NULL ),
      m_GraphicsMemoryAllocator(),
      m_GraphicsMemoryDeallocator(),
      m_Viewport(),
      m_Scissor(),
      m_FrameBuffer(),
      m_RenderState(),
      m_Combiner(),
      m_Texture(),
      m_pPrimitiveModel( NULL ),
      m_Vertex(),
      m_IndexStream(),
      m_Counter( 0 )
{
    for ( u32 index = 0; index < 2; index++)
    {
        renderBuffer[ index ] = 0;
    }
}

//------------------------------------------------------------------------------

SilhouetteScene::~SilhouetteScene()
{
}

//------------------------------------------------------------------------------
bool SilhouetteScene::Initialize( void )
{
    bool is_success = false;

    NN_UNUSED_VAR( is_success );

    allocator = MEMMGR->GetDefaultAllocatorDevice();

    { 
        // Initialize the shader

        // Load shader binary
        nn::fs::FileReader shader_file_reader( L"rom:/shader.shbin" );
        u32 shader_file_size = static_cast<u32>( shader_file_reader.GetSize() );
        shaderBinary = static_cast<u8*>( allocator->Alloc( shader_file_size, 128 ) );
        shader_file_reader.Read( shaderBinary, shader_file_size );
        shader_file_reader.Finalize();

        // Get size of the buffer needed for a ShaderLite class from the shader binary
        u32 size = shader.GetRequiredBufferSize( shaderBinary );

        shaderBuffer = static_cast<u32*>( allocator->Alloc( size ) );

        // Specify shader binary and initialize
        // Enable the executable image at position 0 as the vertex shader               
        // shader.SetupBinary( shaderBinary, 0, -1, shaderBuffer );

        // Specify shader binary and initialize
        // Enable the executable image at position 0 as the vertex shader
        // Enable the executable image at position 1 as the geometry shader
        shader.SetupBinary( shaderBinary, 0, 1, shaderBuffer );
        
        // Vertex shader uniforms
        is_success = 
            shader.SearchBindSymbol( &bindSymbolProjectionMtx44,
            "uProjection" );
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &bindSymbolModelViewMtx44,
            "uModelView" );
        NN_GR_ASSERT( is_success );

        // Geometry shader uniforms
        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolWidthOpenEdgeDepthBiasVec3,
            "dmp_Silhouette.width");
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolColorVec4,
            "dmp_Silhouette.color");
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolOpenEdgeColorVec4,
            "dmp_Silhouette.openEdgeColor");
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolOpenEdgeWidthVec4,
            "dmp_Silhouette.openEdgeWidth");
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolAcceptEmptyTrianglesBool,
            "dmp_Silhouette.acceptEmptyTriangles");
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolScaleByWBool,
            "dmp_Silhouette.scaleByW");
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolFrontFaceCCWBool,
            "dmp_Silhouette.frontFaceCCW");
        NN_GR_ASSERT( is_success );

        is_success = 
            shader.SearchBindSymbol( &twnBindSymbolOpenEdgeWidthScaleByWBool,
            "dmp_Silhouette.openEdgeWidthScaleByW");
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &twnBindSymbolOpenEdgeDepthBiasScaleByWBool,
            "dmp_Silhouette.openEdgeDepthBiasScaleByW");
        NN_GR_ASSERT( is_success );
    }

    {
        // Initialize primitives
        uptr physicalAddr = NULL;
        nn::gr::CTR::BindSymbolVSInput symbolPos;
        nn::gr::CTR::BindSymbolVSInput symbolColor;
        nn::gr::CTR::BindSymbolVSInput symbolNormal;

        m_pPrimitiveModel = 
            static_cast< PrimitiveModel* >( allocator->Alloc( sizeof( PrimitiveModel ), 128 ) );
       
       VertexAttribute vertexAttribute = 
            static_cast<VertexAttribute>(
                static_cast<u32>( POSITION_ATTRIBUTE ) |
                static_cast<u32>( COLOR_ATTRIBUTE ) | 
                static_cast<u32>( NORMAL_ATTRIBUTE ) );

        ArrayType arrayType = INDEPENDENT_ARRAY;

        u32 torusHorizontalPolygonCount = 40;
        u32 torusVerticalPolygonCount   = 40;

        NN_UNUSED_VAR( torusHorizontalPolygonCount );
        NN_UNUSED_VAR( torusVerticalPolygonCount );

        GLenum geometryMode = TRIANGLES_SILHOUETTE_DMP;

        m_pPrimitiveModel->InitializeTorus( allocator,
            geometryMode,
            vertexAttribute,
            arrayType,
            torusHorizontalPolygonCount,
            torusVerticalPolygonCount );

        // Search for "location"
        is_success = shader.SearchBindSymbol( &symbolPos, "aPosition" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &symbolColor, "aColor" );
        NN_GR_ASSERT( is_success );

        is_success = shader.SearchBindSymbol( &symbolNormal, "aNormal" );
        NN_GR_ASSERT( is_success );

        // Vertex stream (Position)   
        physicalAddr = nngxGetPhysicalAddr( m_pPrimitiveModel->pPositionArray );
        NN_NULL_ASSERT( physicalAddr );
        m_Vertex.EnableAttrAsArray( symbolPos, physicalAddr, PICA_DATA_SIZE_3_FLOAT );

        // Color
        physicalAddr = nngxGetPhysicalAddr( m_pPrimitiveModel->pColorArray );
        NN_NULL_ASSERT( physicalAddr );
        m_Vertex.EnableAttrAsArray( symbolColor, physicalAddr, PICA_DATA_SIZE_4_FLOAT );

        // Normal
        physicalAddr = nngxGetPhysicalAddr( m_pPrimitiveModel->pNormalArray );
        NN_NULL_ASSERT( physicalAddr );
        m_Vertex.EnableAttrAsArray( symbolNormal, physicalAddr, PICA_DATA_SIZE_3_FLOAT );

        // Index setting        
        physicalAddr = nngxGetPhysicalAddr( (uptr) m_pPrimitiveModel->pIndexArray );
        NN_NULL_ASSERT( physicalAddr );
        m_IndexStream.physicalAddr = physicalAddr;  
        m_IndexStream.drawVtxNum   = m_pPrimitiveModel->indexCount;

        // Index is defined as u16.
        m_IndexStream.isUnsignedByte = false;
    }

    // Texture settings
    m_Texture.unit0.texType  = PICA_DATA_TEXTURE0_SAMPLER_TYPE_FALSE;
    m_Texture.unit1.texType  = PICA_DATA_TEXTURE1_SAMPLER_TYPE_FALSE;
    m_Texture.unit2.texType  = PICA_DATA_TEXTURE2_SAMPLER_TYPE_FALSE;
    m_Texture.unit3.texType  = PICA_DATA_TEXTURE3_SAMPLER_TYPE_FALSE;

    // Get allocator and deallocator
    nngxGetAllocator( &m_GraphicsMemoryAllocator, &m_GraphicsMemoryDeallocator );

    {
        // Allocate color buffer  GL_RGBA8_OES
        renderBuffer[ 0 ] = 1;
        GLenum area = NN_GX_MEM_VRAMA;
        GLenum aim  = NN_GX_MEM_RENDERBUFFER;
        u32 colorBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        colorBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 0 ], colorBufferByteSize ) );

        // Allocate depth buffer  GL_DEPTH24_STENCIL8_EXT
        renderBuffer[ 1 ] = 2;
        area = NN_GX_MEM_VRAMB;
        u32 depthBufferByteSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 4;
        depthBufferVirtualAddress = 
            reinterpret_cast<uptr>( m_GraphicsMemoryAllocator( area, aim,
                                    renderBuffer[ 1 ], depthBufferByteSize ) );
    }

    {
        renderTarget = 1;

        // Initialize framebuffer settings
        m_FrameBuffer.width                           = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.height                          = nn::gx::DISPLAY0_HEIGHT;

        // Initialize color buffer settings
        m_FrameBuffer.colorBuffer.virtualAddr         = colorBufferVirtualAddress;
        m_FrameBuffer.colorBuffer.format              = PICA_DATA_COLOR_RGBA8_OES;
        m_FrameBuffer.colorBuffer.blockSize           = nn::gr::FrameBuffer::BLOCK_SIZE8;
        m_FrameBuffer.colorBuffer.width               = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.colorBuffer.height              = nn::gx::DISPLAY0_HEIGHT;  
        m_FrameBuffer.colorBuffer.clearColor[ 0 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 1 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 2 ]     = 0.2f;
        m_FrameBuffer.colorBuffer.clearColor[ 3 ]     = 1.0f;

        // Initialize depth buffer settings
        m_FrameBuffer.depthStencilBuffer.virtualAddr  = depthBufferVirtualAddress;
        m_FrameBuffer.depthStencilBuffer.format       = PICA_DATA_DEPTH24_STENCIL8_EXT;
        m_FrameBuffer.depthStencilBuffer.width        = nn::gx::DISPLAY0_WIDTH;
        m_FrameBuffer.depthStencilBuffer.height       = nn::gx::DISPLAY0_HEIGHT;
        m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
        m_FrameBuffer.depthStencilBuffer.clearStencil = 0;
    }

    m_Scissor.SetBufferSize( nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT );

    // Render state settings
    m_RenderState.wBuffer.wScale = 0.0f;


    // glEnable(GL_BLEND);
    // glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    m_RenderState.blend.isEnable = true;
    m_RenderState.blend.srcAlpha   = PICA_DATA_BLEND_FUNC_ONE;

    // glEnable(GL_DEPTH_TEST);
    // glDepthFunc(GL_LESS);    
    m_RenderState.depthTest.isEnable      = true;
    m_RenderState.depthTest.isEnableWrite = true;

    m_RenderState.cullingTest.isEnable  = false;

    {
        // Texture combiner settings
        m_Combiner.stage[ 0 ].rgb.combine         = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
        m_Combiner.stage[ 0 ].rgb.source[ 0 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].rgb.source[ 1 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].rgb.source[ 2 ]     = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;

        m_Combiner.stage[ 0 ].rgb.operand[ 0 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
        m_Combiner.stage[ 0 ].rgb.operand[ 1 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;
        m_Combiner.stage[ 0 ].rgb.operand[ 2 ]    = PICA_DATA_OPE_RGB_SRC_COLOR;

        m_Combiner.stage[ 0 ].alpha.combine     = PICA_DATA_TEX_ENV_COMBINE_REPLACE;

        m_Combiner.stage[ 0 ].alpha.source[ 0 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].alpha.source[ 1 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
        m_Combiner.stage[ 0 ].alpha.source[ 2 ] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;

        m_Combiner.stage[ 0 ].alpha.operand[ 0 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
        m_Combiner.stage[ 0 ].alpha.operand[ 1 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
        m_Combiner.stage[ 0 ].alpha.operand[ 2 ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;

        for ( u32 stageIndex = 1; stageIndex < 6; stageIndex++ )
        {
            m_Combiner.stage[ stageIndex ].rgb.combine   = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
            m_Combiner.stage[ stageIndex ].alpha.combine = PICA_DATA_TEX_ENV_COMBINE_REPLACE;

            m_Combiner.stage[ stageIndex ].rgb.scale     = PICA_DATA_TEX_ENV_SCALE_1;

            for ( u32 sourceIndex = 0; sourceIndex < 3; sourceIndex++)
            {
                m_Combiner.stage[ stageIndex ].rgb.source[ sourceIndex ]    = PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS;
                m_Combiner.stage[ stageIndex ].rgb.operand[ sourceIndex ]   = PICA_DATA_OPE_RGB_SRC_COLOR;

                m_Combiner.stage[ stageIndex ].alpha.source[ sourceIndex ]  = PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS;
                m_Combiner.stage[ stageIndex ].alpha.operand[ sourceIndex ] = PICA_DATA_OPE_ALPHA_SRC_ALPHA;
            }
        }
    }

    return true;
}

//------------------------------------------------------------------------------
void SilhouetteScene::Finalize( void )
{
    m_Vertex.DisableAll();

    m_pPrimitiveModel->Finalize();

    allocator->Free( m_pPrimitiveModel );

    allocator->Free( shaderBinary );
    allocator->Free( shaderBuffer );

    GLenum area = NN_GX_MEM_VRAMA;
    GLenum aim  = NN_GX_MEM_RENDERBUFFER;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 0 ], reinterpret_cast<void*>( colorBufferVirtualAddress ) );
    colorBufferVirtualAddress = 0;

    area = NN_GX_MEM_VRAMB;
    m_GraphicsMemoryDeallocator( area, aim,
        renderBuffer[ 1 ], reinterpret_cast<void*>( depthBufferVirtualAddress ) );
    depthBufferVirtualAddress = 0;
}

//------------------------------------------------------------------------------
void SilhouetteScene::Enter( void )
{
    stereo.Initialize();
    nngxSetDisplayMode( NN_GX_DISPLAYMODE_STEREO );
    Calc();
}

//------------------------------------------------------------------------------
bool SilhouetteScene::Calc( void )
{
    UpdateScene_();

    m_Counter++;

    return true;
}

//------------------------------------------------------------------------------
void SilhouetteScene::PreDraw( void )
{
}

//------------------------------------------------------------------------------
GLuint SilhouetteScene::DrawTop( void )
{
    f32 width  = nn::gx::DISPLAY0_WIDTH;
    f32 height = nn::gx::DISPLAY0_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );
      
    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;
    
    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    {
        // The camera matrix        
        modelViewMtx34.SetupLookAt(
            // camPos
            nn::math::VEC3( 0.0f, 0.0f, 7.5f ),
            // camUp
            nn::math::VEC3( 0.0f, 1.0f, 0.0f ),
            // target
            nn::math::VEC3( 0.0f, 0.0f, 0.0f ) );
    }

    // Projection matrix    
    projectionMtx44.SetupPerspective(
        0.66f,          // fovy
        height / width, // aspect
        0.1f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR,
        6.5f, 1.f, nn::math::PIVOT_UPSIDE_TO_TOP );

    const PrimitiveType primitiveType = static_cast<PrimitiveType>( TRIANGLES_SILHOUETTE_DMP );

    RenderSceneGr_( NN_GX_DISPLAY0, projL, viewL, primitiveType );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint SilhouetteScene::DrawExt( void )
{
    f32 width  = nn::gx::DISPLAY0_WIDTH; 
    f32 height = nn::gx::DISPLAY0_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );

    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR,
        6.5f, 1.f, nn::math::PIVOT_UPSIDE_TO_TOP);
    
    const PrimitiveType primitiveType = static_cast<PrimitiveType>( TRIANGLES_SILHOUETTE_DMP );

    RenderSceneGr_( NN_GX_DISPLAY0_EXT, projR, viewR, primitiveType );

    return renderTarget;
}

//------------------------------------------------------------------------------
GLuint SilhouetteScene::DrawBtm( void )
{
    f32 width  = nn::gx::DISPLAY1_WIDTH; 
    f32 height = nn::gx::DISPLAY1_HEIGHT;

    m_Viewport.Set( 0, 0, width, height );
    m_Scissor.Set( 0, 0, width, height );
 
    m_FrameBuffer.colorBuffer.clearColor[ 0 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 1 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 2 ] = 0.2f;
    m_FrameBuffer.colorBuffer.clearColor[ 3 ] = 1.0f;

    m_FrameBuffer.depthStencilBuffer.clearDepth   = 1.0f;
    m_FrameBuffer.depthStencilBuffer.clearStencil = 0;

    m_FrameBuffer.width                           = width;
    m_FrameBuffer.height                          = height;

    // Clear color buffer and depth buffer
    u32 bufferBit = nn::gr::FrameBuffer::COLOR_BUFFER_BIT | 
                    nn::gr::FrameBuffer::DEPTH_BUFFER_BIT;
    Clear( bufferBit );

    projectionMtx44.SetupPerspective(
        0.66f,          // fovy
        height / width, // aspect
        0.1f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &modelViewMtx34 );
    stereo.SetBaseFrustum( &projectionMtx44 );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR,
        6.5f, 1.f );

    const PrimitiveType primitiveType = TRIANGLES;
    RenderSceneGr_( NN_GX_DISPLAY1, projL, viewL, primitiveType );

    return renderTarget;
}

//------------------------------------------------------------------------------
void SilhouetteScene::PostDraw( void )
{
}

//------------------------------------------------------------------------------
nn::gr::CTR::FrameBuffer::ColorBuffer* SilhouetteScene::GetGrColorBufferPointer( void )
{
    return & m_FrameBuffer.colorBuffer;
}


//------------------------------------------------------------------------------
void SilhouetteScene::Clear( const u32 bufferBit )
{
    u32* command = 0;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, 
        reinterpret_cast<GLint*>( &command ) );

    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Framebuffer settings
    command = m_FrameBuffer.MakeCommand( command, bufferBit );

    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof(u32) );

    // Add a load complete command at the end of the 3D command buffer
    bool isAddSplitDrawCmdlist = true;

    // Generate a request to clear the color buffer and depth buffer
    m_FrameBuffer.MakeClearRequest( bufferBit, isAddSplitDrawCmdlist );
}

//------------------------------------------------------------------------------

void SilhouetteScene::UpdateScene_( void )
{
}

//------------------------------------------------------------------------------

bit32* SilhouetteScene::MakeTextureCommand_( bit32* command )
{
    // Generate 3 dummy commands to 0x80    
    const bool isAddDummyCommand = true;

    command = m_Texture.MakeCommand( command, isAddDummyCommand );

    return command;
}

//------------------------------------------------------------------------------
void SilhouetteScene::RenderSceneGr_( const s32 display,
                                      const nn::math::MTX44& proj, const nn::math::MTX34& view,
                                      const PrimitiveType primitiveType )
{
    NN_UNUSED_VAR( display );
    NN_UNUSED_VAR( proj );
    NN_UNUSED_VAR( view );

    // Get a pointer to the render command buffer
    u32* command = NULL;
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &command ) );

    // Get the start address (so that the pointer can be advanced at the end)
    const u32* start_addr = command;

    // Viewport settings
    command = m_Viewport.MakeCommand( command );

    // Scissor settings
    command = m_Scissor.MakeCommand( command );

    // Clear the framebuffer cache
    command = nn::gr::CTR::FrameBuffer::MakeClearCacheCommand( command );

    // Render state settings
    command = m_RenderState.MakeCommand( command );

    // Texture combiner settings
    command = m_Combiner.MakeCommand( command );

    // Texture settings
    command = MakeTextureCommand_( command );

    static bool isFirst = true;

    if ( primitiveType & TRIANGLES )
    {
        command = MakeDrawModelCommand_( command, proj, view, isFirst );
        
        if ( isFirst )
        {
            isFirst = false;
        }
    }
    
    if ( primitiveType & TRIANGLES_SILHOUETTE_DMP )
    {
        command = MakeDrawSilhouetteModelCommand_( command, display, proj, view, isFirst );

        if ( isFirst )
        {
            isFirst = false;
        }
    }
    
    // Advance the pointer to the render command buffer
    nngxMoveCommandbufferPointer( ( command - start_addr ) * sizeof( u32 ) );
}

//------------------------------------------------------------------------------
bit32* SilhouetteScene::MakeDrawModelCommand_( bit32* command,
                                               const nn::math::MTX44& proj,
                                               const nn::math::MTX34& view,
                                               const bool isMakeFullShaderCommand )
{
    nn::gr::ShaderLite* pShader = &shader;

    // Shader binary setting
    pShader->SetShaderIndex( 0, -1 );
    pShader->SetDrawMode( PICA_DATA_DRAW_TRIANGLES ); 

    if ( isMakeFullShaderCommand )
    {
        command = pShader->MakeFullCommand( command );
    }
    else
    {
        command = pShader->MakeShaderCommand( command );
    }
            
    // Rotate the model
    nn::math::Matrix34 rotMtx34;
    nn::math::Matrix34 modelViewMtx34;
    nn::math::MTX34RotXYZDeg( &rotMtx34, 0.0f, -6.0f * m_Counter, 0.0f );
    nn::math::MTX34Mult( &modelViewMtx34, &view, &rotMtx34 );
    nn::math::Matrix44 modelViewMtx44( modelViewMtx34 );

    // Index stream settings
    {
        uptr physicalAddr = nngxGetPhysicalAddr( (uptr) m_pPrimitiveModel->pIndexArray );
        NN_NULL_ASSERT( physicalAddr );
        m_IndexStream.physicalAddr = physicalAddr;  

        m_IndexStream.drawVtxNum   = m_pPrimitiveModel->indexCount;
    }

    // Sets the uniform of the vertex shader
    command = bindSymbolProjectionMtx44.MakeUniformCommand( command, proj );
    command = bindSymbolModelViewMtx44.MakeUniformCommand( command, modelViewMtx44 );

    // Render kick
    command = MakeDrawCommand( command );

    return command;
}

//------------------------------------------------------------------------------

bit32* SilhouetteScene::MakeDrawSilhouetteModelCommand_( bit32* command,
                                                         const s32 display,
                                                         const nn::math::MTX44& proj,
                                                         const nn::math::MTX34& view,
                                                         const bool isMakeFullShaderCommand )
{
    nn::gr::ShaderLite* pShader = &shader;

    // Shader binary setting
    pShader->SetShaderIndex( 0, 1 );
    pShader->SetDrawMode( PICA_DATA_DRAW_GEOMETRY_PRIMITIVE );
    
    pShader->SetUniformBool( twnBindSymbolAcceptEmptyTrianglesBool, false );
    pShader->SetUniformBool( twnBindSymbolScaleByWBool, true );
        
    pShader->SetUniformBool( twnBindSymbolFrontFaceCCWBool, true );
    pShader->SetUniformBool( twnBindSymbolOpenEdgeWidthScaleByWBool, false );
    pShader->SetUniformBool( twnBindSymbolOpenEdgeDepthBiasScaleByWBool, false );
    
    if ( isMakeFullShaderCommand )
    {
        command = pShader->MakeFullCommand( command );
    }
    else
    {
        command = pShader->MakeShaderCommand( command );
    }

    u32 width  = nn::gx::DISPLAY0_WIDTH;
    u32 height = nn::gx::DISPLAY0_HEIGHT;

    if ( display == NN_GX_DISPLAY1 )
    {
        width  = nn::gx::DISPLAY1_WIDTH;
        height = nn::gx::DISPLAY1_HEIGHT;
    }

    const f32 silhouetteWidth = 0.15f;

    nn::math::VEC3 vec3;
    vec3[ 0 ] = silhouetteWidth;
    vec3[ 1 ] = vec3[ 0 ] * static_cast<float>( width ) / static_cast<float>( height );
    vec3[ 2 ] = 0.0f;
    command   = twnBindSymbolWidthOpenEdgeDepthBiasVec3.MakeUniformCommand( command, static_cast<nn::math::Vector4>(vec3) );

    nn::math::VEC4 vec4;
    vec4[ 0 ] = 0.0f;
    vec4[ 1 ] = 0.0f;
    vec4[ 2 ] = 1.0f;   
    vec4[ 3 ] = 1.0f;
    command   = twnBindSymbolColorVec4.MakeUniformCommand( command, vec4 );

    vec4[ 0 ] = 0.0f;    
    vec4[ 1 ] = 0.0f;
    vec4[ 2 ] = 1.0f;
    vec4[ 3 ] = 1.0f;
    command   = twnBindSymbolOpenEdgeColorVec4.MakeUniformCommand( command, vec4 );

    vec4[ 0 ] = width  / silhouetteWidth;
    vec4[ 1 ] = height / silhouetteWidth;
    vec4[ 2 ] = width  * height;
    vec4[ 3 ] = 2.0f   / silhouetteWidth;
    command   = twnBindSymbolOpenEdgeWidthVec4.MakeUniformCommand( command, vec4 );    

    // Rotate the model
    nn::math::Matrix34 rotMtx34;
    nn::math::Matrix34 modelViewMtx34;
    nn::math::MTX34RotXYZDeg( &rotMtx34, 0.0f, -6.0f * m_Counter, 0.0f );
    nn::math::MTX34Mult( &modelViewMtx34, &view, &rotMtx34 );
    nn::math::Matrix44 modelViewMtx44( modelViewMtx34 );

    // Index stream settings    
    {
        uptr physicalAddr = nngxGetPhysicalAddr( (uptr) m_pPrimitiveModel->pTwnIndexArray );
        NN_NULL_ASSERT( physicalAddr );
        m_IndexStream.physicalAddr = physicalAddr;  

        m_IndexStream.drawVtxNum   = m_pPrimitiveModel->twnIndexCount;
    }

    // Sets the uniform of the vertex shader
    command = bindSymbolProjectionMtx44.MakeUniformCommand( command, proj );
    command = bindSymbolModelViewMtx44.MakeUniformCommand( command, modelViewMtx44 );

    // Render kick
    command = MakeDrawCommand( command );

    return command;
}

//------------------------------------------------------------------------------

bit32* SilhouetteScene::MakeDrawCommand( bit32* command )
{
    // Disable all vertex attributes before starting to use vertices
    command = nn::gr::Vertex::MakeDisableCommand( command );

    // glBindBuffer( GL_ARRAY_BUFFER )
    // glBindBuffer( GL_ELEMENT_ARRAY_BUFFER ) + glDrawElements()
    command = m_Vertex.MakeEnableAttrCommand( command );

    command = m_Vertex.MakeDrawCommand( command, m_IndexStream );

    // Disable all vertex attributes when finished using vertex
    command = nn::gr::CTR::Vertex::MakeDisableCommand( command );

    return command;
}
