﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     AppRecommendedFlow.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47455 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <demo.h>
#include <nn/applet.h>
//
#include <nn/hid/CTR/hid_ExtraPadReader.h>
#include <nn/hid/CTR/hid_ExtraPadStatus.h>

#include "Application.h"
#include "SaveData.h"
#include "scenes/SceneId.h"
#include "scenes/MySceneFactory.h"
#include "ExtraPadDetection.h"

namespace
{
    // Device memory size
    const size_t DEVICE_MEMORY_SIZE = 4 * 1024 * 1024;

    // Renderer
    demo::RenderSystemDrawing *p_renderSystem;

    // Buffer used to send/receive for Circle Pad Pro
    u8 extraPadWorkingMemory[4096] NN_ATTRIBUTE_ALIGN(4096);

    // Sleep
    class SleepHandler
    {
    private:
        nn::os::LightEvent m_AwakeEvent;

        static nn::applet::AppletQueryReply SleepQueryCallback(uptr arg)
        {
            SleepHandler *handler = reinterpret_cast<SleepHandler *>(arg);
            handler->m_AwakeEvent.ClearSignal();
            if ( nn::applet::IsActive() )
            {
                // Always prohibit Sleep Mode when active
                return nn::applet::REPLY_LATER;
            }
            else
            {
                return nn::applet::REPLY_ACCEPT;
            }
        }
        static void AwakeCallback(uptr arg)
        {
            SleepHandler *handler = reinterpret_cast<SleepHandler *>(arg);
            handler->m_AwakeEvent.Signal();
        }
    public:
        void SetupSleep()
        {
            m_AwakeEvent.Initialize(true);
            m_AwakeEvent.Signal();
            nn::applet::SetSleepQueryCallback(SleepQueryCallback, reinterpret_cast<uptr>(this));
            nn::applet::SetAwakeCallback(AwakeCallback, reinterpret_cast<uptr>(this));
            nn::applet::EnableSleep(true);
        }
        void ProcessSleep()
        {
            if(nn::applet::IsExpectedToReplySleepQuery())
            {
                // Halt detection thread
                SampleDemos::hid::ExtraPad::AppRecommendedFlow::StopDetectExtraPadThread();

                // Stop sampling
                nn::hid::ExtraPad::StopSampling();

                NN_LOG( "\n<< Sleep >>");

                nn::applet::ReplySleepQuery( nn::applet::REPLY_ACCEPT );

                // Wait until Sleep Mode-recovery event occurs
                m_AwakeEvent.Wait();

                NN_LOG( "\n<< Awake >>");

                // Before sleep, because the detection thread calls the detection thread and the reopening is called at the proper timing
                // Does not reopen when recovering from sleep

                nn::gx::StartLcdDisplay();
            }
        }
    };
}

void Initialize();
void Finalize();

void InitializeForApplication();

void InitializeSystemRender();
void FinalizeSystemRender();

void nnMain()
{
    NN_LOG("\n----- Start AppRecommendedFlow -----");

    Initialize();

    {
        // Scene manager
        scene::SceneManager sceneManager;
        // Scene factory
        SampleDemos::hid::ExtraPad::AppRecommendedFlow::MySceneFactory sceneFactory;
        // Configure initial scene
        sceneManager.Initialize(&sceneFactory,
            SampleDemos::hid::ExtraPad::AppRecommendedFlow::SCENE_ID_EP_SELECT_SAVE_DATA);

        // Application
        SampleDemos::hid::ExtraPad::AppRecommendedFlow::Application* pApp =
            SampleDemos::hid::ExtraPad::AppRecommendedFlow::Application::GetInstance();

        // Sleep handler
        SleepHandler sleep;
        sleep.SetupSleep();

        // Mount save data region
        pApp->m_saveData.Initialize();

        // Initialize flag indicating if this is the first connection on the Circle Pad Pro first-time use scene
        pApp->m_IsFirstConnect = true;

        // Title settings
        wchar_t title[] = L"Extra Pad Demo / AppRecommendedFlow";
        scene::SetTitle(title, sizeof(title) / sizeof(wchar_t));

        // Register the render objects used by the GUI library
        scene::RegisterDrawers(pApp->GetRenderSystem(), pApp->GetTextWriter());

        // Main loop
        while (1)
        {
            // The HOME Button
            if (nn::applet::IsExpectedToProcessHomeButton())
            {
                nn::applet::ProcessHomeButton();
                nn::applet::WaitForStarting();

                // Check if return from waiting completes
                if (nn::applet::IsExpectedToCloseApplication())
                {
                    break;
                }

                // Recover the GPU register settings
                nngxUpdateState(NN_GX_STATE_ALL);
                nngxValidateState(NN_GX_STATE_ALL, GL_TRUE);
            }

            // Power Button
            if (nn::applet::IsExpectedToProcessPowerButton())
            {
                nn::applet::ProcessPowerButton();
                nn::applet::WaitForStarting();
                break;
            }

            // Sleep
            sleep.ProcessSleep();

            // Determining the end of each frame
            if (nn::applet::IsExpectedToCloseApplication())
            {
                break;
            }

            // Scene processing
            if (!sceneManager.IsExpectedToFinish())
            {
                sceneManager.Update();
                sceneManager.Draw();
            }
        }
    }NN_LOG("----- Stop AppRecommendedFlow -----\n");
    Finalize();
}

void Initialize()
{
    InitializeForApplication();

    nn::hid::Initialize();

    // Initializing Circle Pad Pro
    nn::hid::ExtraPad::Initialize(extraPadWorkingMemory);

    // Applet
    nn::applet::Enable();

    // End determination
    if (nn::applet::IsExpectedToCloseApplication())
    {
        Finalize();
        nn::applet::PrepareToCloseApplication();
        nn::applet::CloseApplication();
    }

    InitializeSystemRender();
}


void InitializeSystemRender()
{
    demo::RenderBufferDescription colorBufferDesc;
    {
        colorBufferDesc.m_Format = GL_RGBA8_OES;
        colorBufferDesc.m_Width = nn::gx::DISPLAY0_WIDTH;
        colorBufferDesc.m_Height = nn::gx::DISPLAY0_HEIGHT;
        colorBufferDesc.m_Area = NN_GX_MEM_VRAMA;
        colorBufferDesc.m_Attachment = GL_COLOR_ATTACHMENT0;
    }
    demo::RenderBufferDescription depthStencilBufferDesc;
    {
        depthStencilBufferDesc.m_Format = GL_DEPTH24_STENCIL8_EXT;
        depthStencilBufferDesc.m_Width = nn::gx::DISPLAY0_WIDTH;
        depthStencilBufferDesc.m_Height = nn::gx::DISPLAY0_HEIGHT;
        depthStencilBufferDesc.m_Area = NN_GX_MEM_VRAMB;
        depthStencilBufferDesc.m_Attachment = GL_DEPTH_STENCIL_ATTACHMENT;
    }

    demo::FrameBufferDescription frameBufferDesc;
    {
        frameBufferDesc.m_RenderBufferDescriptionArray[demo::DEMO_COLOR_BUFFER_INDEX] =
            colorBufferDesc;
        frameBufferDesc.m_RenderBufferDescriptionArray[demo::DEMO_DEPTH_STENCIL_BUFFER_INDEX] =
            depthStencilBufferDesc;
    }

    demo::DisplayBuffersDescription display0BuffersDesc;
    {
        display0BuffersDesc.m_TargetDisplay = NN_GX_DISPLAY0;
        display0BuffersDesc.m_DisplayBufferNum = demo::DEMO_DISPLAY_BUFFER_NUM;
        display0BuffersDesc.m_Format = GL_RGB8_OES;
        display0BuffersDesc.m_Width = nn::gx::DISPLAY0_WIDTH;
        display0BuffersDesc.m_Height = nn::gx::DISPLAY0_HEIGHT;
        display0BuffersDesc.m_Area = NN_GX_MEM_FCRAM;
    }

    demo::DisplayBuffersDescription display1BuffersDesc;
    {
        display1BuffersDesc.m_TargetDisplay = NN_GX_DISPLAY1;
        display1BuffersDesc.m_DisplayBufferNum = demo::DEMO_DISPLAY_BUFFER_NUM;
        display1BuffersDesc.m_Format = GL_RGB8_OES;
        display1BuffersDesc.m_Width = nn::gx::DISPLAY1_WIDTH;
        display1BuffersDesc.m_Height = nn::gx::DISPLAY1_HEIGHT;
        display1BuffersDesc.m_Area = NN_GX_MEM_FCRAM;
    }

    SampleDemos::hid::ExtraPad::AppRecommendedFlow::Application::AcquireInstance();
    p_renderSystem =
        SampleDemos::hid::ExtraPad::AppRecommendedFlow::Application::GetInstance()->GetRenderSystem();
}

void InitializeForApplication()
{
    scene::SetupMemory(DEVICE_MEMORY_SIZE);

    // File system (for creating save data)
    nn::fs::Initialize();

    // Circle Pad Pro detection thread
    SampleDemos::hid::ExtraPad::AppRecommendedFlow::InitializeExtraPadDetection(false); // Do not run thread when initializing
}

void Finalize()
{
    FinalizeSystemRender();

    // Exit application
    nn::applet::PrepareToCloseApplication();
    nn::applet::CloseApplication();

    // Circle Pad Pro detection thread
    SampleDemos::hid::ExtraPad::AppRecommendedFlow::FinalizeExtraPadDetection();
}

void FinalizeSystemRender()
{
    p_renderSystem->Finalize();
}

