﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     Application.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <string.h>
#include <cstdio>
#include "Application.h"

//////////////////////////////////////////////////////////////////////////////////////////
// Constant definitions
//////////////////////////////////////////////////////////////////////////////////////////

// Renderer buffer size
#define GX_BUFFER_SIZE (2 * 1024 * 1024)

//////////////////////////////////////////////////////////////////////////////////////////
// Variable Definitions
//////////////////////////////////////////////////////////////////////////////////////////

namespace
{
    SampleDemos::hid::ExtraPad::AppRecommendedFlow::Application* s_pApplication = NULL;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Function definitions
//////////////////////////////////////////////////////////////////////////////////////////

namespace
{
    // Get color buffer settings
    void GetDescColorBuffer(demo::RenderBufferDescription& desc)
    {
        desc.m_Format     = GL_RGBA8_OES;
        desc.m_Width      = NN_GX_DISPLAY0_WIDTH;
        desc.m_Height     = NN_GX_DISPLAY0_HEIGHT;
        desc.m_Area       = NN_GX_MEM_VRAMA;
        desc.m_Attachment = GL_COLOR_ATTACHMENT0;
    }

    // Get depth buffer settings
    void GetDescDepthBuffer(demo::RenderBufferDescription& desc)
    {
        desc.m_Format     = GL_DEPTH24_STENCIL8_EXT;
        desc.m_Width      = NN_GX_DISPLAY0_WIDTH;
        desc.m_Height     = NN_GX_DISPLAY0_HEIGHT;
        desc.m_Area       = NN_GX_MEM_VRAMB;
        desc.m_Attachment = GL_DEPTH_STENCIL_ATTACHMENT;
    }

    // Get framebuffer settings
    void GetDescFrameBuffer(demo::FrameBufferDescription& desc, demo::RenderBufferDescription& color, demo::RenderBufferDescription& depth)
    {
        GetDescColorBuffer(color);
        GetDescDepthBuffer(depth);

        desc.m_RenderBufferDescriptionArray[demo::DEMO_COLOR_BUFFER_INDEX] = color;
        desc.m_RenderBufferDescriptionArray[demo::DEMO_DEPTH_STENCIL_BUFFER_INDEX] = depth;
    }

    // Get display buffer settings for the upper screen
    void GetDescDisplayBuffer0(demo::DisplayBuffersDescription& desc)
    {
        desc.m_TargetDisplay    = NN_GX_DISPLAY0;
        desc.m_DisplayBufferNum = 2;
        desc.m_Format           = GL_RGB8_OES;
        desc.m_Width            = NN_GX_DISPLAY0_WIDTH;
        desc.m_Height           = NN_GX_DISPLAY0_HEIGHT;
        desc.m_Area             = NN_GX_MEM_VRAMA;
    }

#if 0

    // Get display buffer (EXT) settings for the upper screen
    void GetDescDisplayBuffer0Ext(demo::DisplayBuffersDescription& desc)
    {
        desc.m_TargetDisplay    = NN_GX_DISPLAY0_EXT;
        desc.m_DisplayBufferNum = 2;
        desc.m_Format           = GL_RGB8_OES;
        desc.m_Width            = NN_GX_DISPLAY0_WIDTH;
        desc.m_Height           = NN_GX_DISPLAY0_HEIGHT;
        desc.m_Area             = NN_GX_MEM_VRAMA;
    }

#endif

    // Get display buffer settings for the lower screen
    void GetDescDisplayBuffer1(demo::DisplayBuffersDescription& desc)
    {
        desc.m_TargetDisplay    = NN_GX_DISPLAY1;
        desc.m_DisplayBufferNum = 2;
        desc.m_Format           = GL_RGB8_OES;
        desc.m_Width            = NN_GX_DISPLAY1_WIDTH;
        desc.m_Height           = NN_GX_DISPLAY1_HEIGHT;
        desc.m_Area             = NN_GX_MEM_VRAMB;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////

namespace SampleDemos {
namespace hid {
namespace ExtraPad {
namespace AppRecommendedFlow {

// Create an instance.
void Application::AcquireInstance()
{
    NN_ASSERT(s_pApplication == NULL);
    s_pApplication = new Application();
}

// Release an instance.
void Application::ReleaseInstance()
{
    NN_ASSERT(s_pApplication != NULL);

    delete s_pApplication;

    s_pApplication = NULL;
}

// Gets an instance.
Application* Application::GetInstance()
{
    return s_pApplication;
}

// Get the renderer.
demo::RenderSystemDrawing* Application::GetRenderSystem()
{
    return &m_renderSystem;
}

// Get the text writer.
scene::TextWriter* Application::GetTextWriter()
{
    return &m_textWriter;
}

// Constructor.
Application::Application()
{
    m_pGxBuffer = (uptr)scene::Allocate(GX_BUFFER_SIZE);
    if (!m_pGxBuffer)
    {
        NN_LOG("Not enough memory.\n");
        NN_LOG("Required memory size = %d Bytes.\n", GX_BUFFER_SIZE);
        NN_PANIC("");
    }

    {
        // Render buffer settings
        demo::RenderBufferDescription colorBufferDesc;
        demo::RenderBufferDescription depthBufferDesc;
        // Framebuffer settings
        demo::FrameBufferDescription frameBufferDesc;
        // Display buffer settings
        demo::DisplayBuffersDescription display0BuffersDesc;
        demo::DisplayBuffersDescription display1BuffersDesc;
        // Command buffer size
        const u32 commandBufferSize = 0x40000;
        // Number of command requests
        const u32 requestNum = 128;

        GetDescFrameBuffer(frameBufferDesc, colorBufferDesc, depthBufferDesc);
        GetDescDisplayBuffer0(display0BuffersDesc);
        GetDescDisplayBuffer1(display1BuffersDesc);

        // Initialize the renderer
        m_renderSystem.Initialize(m_pGxBuffer, GX_BUFFER_SIZE, commandBufferSize, requestNum, true, display0BuffersDesc, display1BuffersDesc, frameBufferDesc);
    }

    // Initialize the text writer
    m_textWriter.Initialize(scene::GetHeap());

}

// Destructor
Application::~Application()
{
    if (m_pGxBuffer)
    {
        // Finalize the text writer
        m_textWriter.Finalize();

        // Finalize the renderer
        m_renderSystem.Finalize();
        scene::Free(reinterpret_cast<void*>(m_pGxBuffer));
    }
}

}}}}
