﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ExtraPadDetection.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/


#include <nn.h>
#include "ExtraPadDetection.h"

//////////////////////////////////////////////////////////////////////////////////////////
// Constant definitions
//////////////////////////////////////////////////////////////////////////////////////////

namespace
{
    // For Circle Pad Pro detection thread
    nn::os::Thread            g_DetectExtraPadThread;
    nn::os::StackBuffer<1024> g_DetectExtraPadStack;
    nn::os::LightEvent        g_DetectExtraPadEvent;
    bool                      g_DetectExtraPadThreadIsAlive;
}

//////////////////////////////////////////////////////////////////////////////////////////

namespace SampleDemos {
namespace hid {
namespace ExtraPad {
namespace AppRecommendedFlow {

// Performs initialization.
void InitializeExtraPadDetection(bool isStartThread)
{
    // Events for Circle Pad Pro detection thread
    g_DetectExtraPadEvent.Initialize(true); // Initialize by manually resetting

    // Set flag for determining thread existence
    g_DetectExtraPadThreadIsAlive = true;

    // Start Circle Pad Pro detection thread
    nn::Result result = g_DetectExtraPadThread.TryStart(DetectExtraPadThreadFunc, g_DetectExtraPadStack);
    if(result.IsFailure())
    {
        NN_ERR_THROW_FATAL_ALL(result);
    }

    if(isStartThread)
    {
        StartDetectExtraPadThread();
    }
}

// Finalizes.
void FinalizeExtraPadDetection()
{
        // Finalize detection thread
        g_DetectExtraPadThreadIsAlive = false;
        g_DetectExtraPadEvent.Signal();
        g_DetectExtraPadThread.Join();
        g_DetectExtraPadThread.Finalize();

        nn::hid::CTR::ExtraPad::StopSampling();
}

// Start detection thread
void StartDetectExtraPadThread()
{
    // Start thread if not connected
    if(nn::hid::ExtraPad::GetConnectionState() != nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
    {
        // If detection process not run yet
        if(!IsAliveDetectExtraPadThread())
        {
            // Set detection flag
            g_DetectExtraPadEvent.Signal();
            NN_LOG("\n[Start] Extra pad detection.");
        }
    }
}

// Stop detection thread
void StopDetectExtraPadThread()
{
    if(IsAliveDetectExtraPadThread())
    {
        // Clear detection flag
        g_DetectExtraPadEvent.ClearSignal();
        NN_LOG("\n[Stop] Extra pad detection.");
    }
}

// Confirm whether detection thread has been run
// Returns true if running, false otherwise.
bool IsAliveDetectExtraPadThread()
{
    return g_DetectExtraPadEvent.TryWait();
}


// Circle Pad Pro detection thread function.
void DetectExtraPadThreadFunc()
{
    u32 counter = 1; // For NN_LOG
    while(true)
    {
        // Wait until signal comes
        g_DetectExtraPadEvent.Wait();

        // If IsAlive is false, break and stop thread function
        if(!g_DetectExtraPadThreadIsAlive) break;

        // If not connected, try sampling every second
        if(nn::hid::ExtraPad::GetConnectionState() != nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
        {
            if(counter%30==0){NN_LOG("\n");}else{NN_LOG(".");}
            counter++;

            // Start Circle Pad Pro sampling
            nn::Result result = nn::hid::CTR::ExtraPad::StartSampling(0, 8);

            // Connection successful
            if(result.IsSuccess())
            {
                NN_LOG("\n[Detected] Extra Pad detected!");
                StopDetectExtraPadThread();
                counter = 1;
            }
            // Connection failure
            else
            {
                // Stop thread for one second
                nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromSeconds(1));
            }
        }
    }
    return;
}


}}}}
