/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SaveData.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/fs.h>
#include <nn/Result.h>

#include "SaveData.h"

namespace SampleDemos {
namespace hid {
namespace ExtraPad {
namespace AppRecommendedFlow {

const char* SaveData::SAVEDATA_ARCHIVE  = "data:";
const char* SaveData::SAVEDATA_FILEPATH[] = {
    "data:/data1.txt",
    "data:/data2.txt",
    "data:/data3.txt",
    "data:/data4.txt"
};

const wchar_t* SaveData::SAVEDATA_FILENAME[] = {
    L"Save data 1",
    L"Save data 2",
    L"Save data 3",
    L"Save data 4"
};

void SaveData::Initialize()
{
    // Flag for save data region
    m_IsMounted      = true;

    // Default value for new save data
    m_SaveDataSetDefault.isUseExtraPad  = IS_USE_EXTRA_PAD_DEFAULT;
    m_SaveDataSetDefault.isNewSaveData  = IS_NEW_SAVE_DATA_DEFAULT;
    m_SaveDataSetDefault.playTimeTick   = nn::os::Tick(0);

    m_SaveDataSetDefault.objectPositionX  = 160;
    m_SaveDataSetDefault.objectPositionY  = 120;
    m_SaveDataSetDefault.objectLineLength =  70;

    MountSaveDataArchive();
}

void SaveData::Finalize()
{
    if (m_IsMounted)
    {
        // Since it is duplicated, commit the save data region
        nn::Result result = nn::fs::CommitSaveData();
        if(result.IsFailure())
        {
            NN_LOG("[SaveData::Finalize] Cannot commit save data!\n");
        }
        UnmountSaveDataArchive();
    }
}

// Mount save data region
void SaveData::MountSaveDataArchive()
{
    nn::Result result = nn::fs::MountSaveData(SAVEDATA_ARCHIVE);

    if(result.IsFailure())
    {
        NN_LOG("[SaveData::MountSaveDataArchive] result.IsFailure()\n");
        if(result <= nn::fs::ResultNotFormatted()) //Save data region is in an unformatted state.
        {
            NN_LOG("[SaveData::MountSaveDataArchive] result = ResultNotFormatted\n");
            result = FormatSaveDataArchive();
        }
        else if(result <= nn::fs::ResultBadFormat()) //File format is in an invalid state.
        {
            NN_LOG("[SaveData::MountSaveDataArchive] result = ResultBadFormat\n");
            result = FormatSaveDataArchive();
        }
        else if(result <= nn::fs::ResultVerificationFailed()) //Save data is corrupted or has been tampered with.
        {
            NN_LOG("[SaveData::MountSaveDataArchive] result = ResultVerificationFailed\n");
            result = FormatSaveDataArchive();
        }
        else if(result <= nn::fs::ResultOperationDenied()) // Now mounting.
        {
            // The system handles this error in CTR-SDK versions 2.1 and later. Handle as an unexpected error.
            NN_LOG("[SaveData::MountSaveDataArchive] result = ResultOperationDenied\n");
            NN_ERR_THROW_FATAL_ALL(result);
        }
        else
        {
            // Any other error is unexpected.
            NN_LOG("[SaveData::MountSaveDataArchive] It is an unexpected error. \n");
            NN_ERR_THROW_FATAL_ALL(result);
        }

        // Mount if formatting failed
        if(!result.IsFailure())
        {
            result = nn::fs::MountSaveData();
            if(result.IsFailure())
            {
                m_IsMounted = false;
            }
        }
    }
}

// Format save data region
nn::Result SaveData::FormatSaveDataArchive()
{
    const size_t maxFiles        = SAVE_DATA_MAX;    // Maximum number of files
    const size_t maxDirectories  = 1;                // Maximum number of directories
    const bool   isDuplicateAll  = true;             // Duplicates the entire save data region

    nn::Result result = nn::fs::FormatSaveData(maxFiles, maxDirectories, isDuplicateAll);
    if(result.IsFailure()) // Format failed
    {
        NN_LOG("[SaveData::FormatSaveDataArchive] Cannot format save data!\n");
        NN_ERR_THROW_FATAL_ALL(result);
    }
    return result;
}


// Unmount save data region
void SaveData::UnmountSaveDataArchive()
{
    nn::Result result = nn::fs::Unmount(SAVEDATA_ARCHIVE);
    if(result.IsFailure())
    {
        NN_LOG("[SaveData::UnmountSaveDataArchive] Cannot unmount save data!\n");
        NN_DBG_PRINT_RESULT(result);
    }
}


// Load data from save data file
void SaveData::ReadSaveData(SaveDataStruct* saveDataSet, SaveDataId saveDataId)
{
    // Open
    OpenSaveDataFile(saveDataId);

    // Specify location to read (start)
    nn::Result result = m_SaveDataFile.TrySeek(0, nn::fs::POSITION_BASE_BEGIN);
    if(result.IsFailure())
    {
        NN_LOG("[SaveData::ReadSaveData] TrySeek error\n");
        NN_DBG_PRINT_RESULT(result);
    }

    // Read
    s32 readSize;
    result = m_SaveDataFile.TryRead(&readSize, saveDataSet, sizeof(SaveDataStruct));
    if(result.IsFailure())
    {
        NN_LOG("[SaveData::ReadSaveData] TryRead error\n");
        NN_DBG_PRINT_RESULT(result);
    }

    // Close the file
    m_SaveDataFile.Finalize();
    m_IsFileOpen = false;

    return;
}

// Write data to save data file
void SaveData::WriteSaveData(SaveDataStruct* saveDataSet, SaveDataId saveDataId)
{
    // Open if file not already open
    if(!m_IsFileOpen)
    {
        OpenSaveDataFile(saveDataId);
    }

    // Specify location to write (start)
    nn::Result result = m_SaveDataFile.TrySeek(0, nn::fs::POSITION_BASE_BEGIN);
    if(result.IsFailure())
    {
        NN_LOG("[SaveData::WriteSaveData] TrySeek error\n");
        NN_DBG_PRINT_RESULT(result);
        return;
    }

    // Write
    s32 writeSize;
    result = m_SaveDataFile.TryWrite(&writeSize, saveDataSet, sizeof(SaveDataStruct), true);
    if(result.IsFailure())
    {
        NN_LOG("[SaveData::WriteSaveData] TryWrite error\n");
        return;
    }

    // Commit written content
    result = nn::fs::CommitSaveData();
    if(result.IsFailure())
    {
        NN_LOG("[SaveData::WriteSaveData] CommitSaveData error\n");
        return;
    }
    // Close the file
    m_SaveDataFile.Finalize();
    m_IsFileOpen = false;

    return;
}


// Open save data file
void SaveData::OpenSaveDataFile(SaveDataId saveDataId)
{
    // Open file (creates a new one if one does not exist)
    nn::Result result = m_SaveDataFile.TryInitialize(
        SAVEDATA_FILEPATH[saveDataId],
        nn::fs::OPEN_MODE_READ | nn::fs::OPEN_MODE_WRITE | nn::fs::OPEN_MODE_CREATE
    );
    // Change state to "file exists"
    m_IsFileOpen = result.IsSuccess();

    // If new data, format with default values
    if(m_SaveDataFile.GetSize() == 0)// File size is 0 (= new file)
    {
        WriteSaveData(&m_SaveDataSetDefault, saveDataId);
        // Reopen
        result = m_SaveDataFile.TryInitialize(
                SAVEDATA_FILEPATH[saveDataId],
                nn::fs::OPEN_MODE_READ | nn::fs::OPEN_MODE_WRITE | nn::fs::OPEN_MODE_CREATE
            );
    }

    // After creating with TryInitialize, performs SetSize since the size is 0
    m_SaveDataFile.TrySetSize(SAVEDATA_SIZE);

    return;
}


void SaveData::DeleteSaveDataFile(SaveDataId saveDataId)
{
    // Delete save data file
    nn::Result result = nn::fs::TryDeleteFile(SAVEDATA_FILEPATH[saveDataId]);

    if(result.IsFailure())
    {
        if(result <= nn::fs::ResultNotFound())
        {
            // Specified path not found.
            NN_LOG("[SaveData::DeleteSaveDataFile] ResultNotFound \"%s\"\n",SAVEDATA_FILEPATH[saveDataId]);
        }
        else if(result <= nn::fs::ResultAlreadyExists())
        {
            // Specified path already exists.
            NN_LOG("[SaveData::DeleteSaveDataFile] ResultAlreadyExists \"%s\"\n",SAVEDATA_FILEPATH[saveDataId]);
        }
        else if(result <= nn::fs::ResultVerificationFailed())
        {
            // You must delete and re-create this file (directory).
            NN_LOG("[SaveData::DeleteSaveDataFile] ResultVerificationFailed \"%s\"\n",SAVEDATA_FILEPATH[saveDataId]);
        }
        else if(result <= nn::fs::ResultOperationDenied())
        {
            if(result <= nn::fs::ResultMediaAccessError())
            {
                // The system handles this error in CTR-SDK versions 2.1 and later. Handle as an unexpected error.
                NN_LOG("[SaveData::DeleteSaveDataFile] ResultMediaAccessError \"%s\"\n",SAVEDATA_FILEPATH[saveDataId]);
                NN_ERR_THROW_FATAL_ALL(result);
            }
            else
            {
                // Access denied for some reason.
                NN_LOG("[SaveData::DeleteSaveDataFile] ResultOperationDenied \"%s\"\n",SAVEDATA_FILEPATH[saveDataId]);
            }
        }
        else if(result <= nn::fs::ResultNotEnoughSpace())
        {
            // Not enough space available in archive or on media.
            NN_LOG("[SaveData::DeleteSaveDataFile] ResultNotEnoughSpace \"%s\"\n",SAVEDATA_FILEPATH[saveDataId]);
        }
        else
        {
            // Any other error is unexpected.
            NN_LOG("[SaveData::DeleteSaveDataFile] It is an unexpected error. \n");
            NN_ERR_THROW_FATAL_ALL(result);
        }
    }
    else if(result.IsSuccess()) // Success
    {
        m_IsFileOpen  = !result.IsSuccess();
    }
}

}}}}
