﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SceneConfirmDetection.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include "SceneConfirmDetection.h"


//////////////////////////////////////////////////////////////////////////////////////////
// Constant Definitions
//////////////////////////////////////////////////////////////////////////////////////////

namespace
{
    enum State
    {
        STATE_ASK_TO_USE,         // Displaying use confirmation
        STATE_SETTING_ON,         // Changing use setting
        STATE_DETECTING,          // Detecting
        STATE_CONNECTION_SUCCESS, // Connection successful
        STATE_CANT_DETECT,        // Not detected
        STATE_REMOVED,            // Removed
        STATE_LOW_BATTERY,        // Low battery
        STATE_CHECK_BATTERY,      // Battery level check
        STATE_CHANGE_SCENE,       // Switch scenes

        STATE_MAX
    };

    enum ControlId
    {
        CONTROL_ID_OK,
        CONTROL_ID_CANCEL,
        CONTROL_ID_YES,
        CONTROL_ID_NO,

        CONTROL_ID_MAX
    };

}

//////////////////////////////////////////////////////////////////////////////////////////

namespace SampleDemos {
namespace hid {
namespace ExtraPad {
namespace AppRecommendedFlow {

const wchar_t* SceneConfirmDetection::MESSAGES_ASK_TO_USE[] = {
    L"Confirmed Circle Pad Pro.",
    L"Use Circle Pad Pro?"
};

const wchar_t* SceneConfirmDetection::MESSAGES_CONNECTION_SUCCESS[] = {
    L"Confirmed connection to Circle Pad Pro."
};

const wchar_t* SceneConfirmDetection::MESSAGES_SETTING_ON[] = {
    L"Enabling Circle Pad Pro.",
    L"You can enable/disable the",
    L"Circle Pad Pro from the Options menu."
};

const wchar_t* SceneConfirmDetection::MESSAGES_DETECTING[] = {
    L"Confirming connection to Circle Pad Pro.",
    L"If in the standby state, press the ZR",
    L"Button or the ZL Button.",
};

const wchar_t* SceneConfirmDetection::MESSAGES_CANT_DETECT[] = {
    L"Cannot confirm Circle Pad Pro. It could",
    L"be that it is not connected correctly",
    L"or that the battery has run out."
};

const wchar_t* SceneConfirmDetection::MESSAGES_REMOVED[] = {
    L"Circle Pad Pro removed.",
    L"You can enable/disable the",
    L"Circle Pad Pro from the Options menu."
};

const wchar_t* SceneConfirmDetection::MESSAGES_LOW_BATTERY[] = {
    L"The Circle Pad Pro battery is low. ",
    L"Replace the battery.",
    L"You can enable/disable the",
    L"Circle Pad Pro from the Options menu."
};


// Initialize a scene.
bool SceneConfirmDetection::Initialize()
{
    // Control manager
    m_ControlManager.RegisterControlEventCallback(MyControlEventCallback, this);

    // Application
    Application* pApp = Application::GetInstance();

    // Initialize counter
    m_Counter = 0;

    // When connecting
    if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
    {
        // When enabled
        if(pApp->m_saveDataSet.isUseExtraPad)
        {
            m_State = STATE_CHECK_BATTERY;
        }
        // When disabled
        else
        {
            m_State = STATE_ASK_TO_USE;
        }
    }
    // When not connected
    else
    {
        // When enabled
        if(pApp->m_saveDataSet.isUseExtraPad)
        {
            m_State = STATE_DETECTING;
            // Get tick from when detection started
            m_StartDetectionTick = nn::os::Tick::GetSystemCurrent();
        }
        // When disabled
        else
        {
            m_State = STATE_CHANGE_SCENE;
        }
    }


    return true;
}

// Finalize the scene.
void SceneConfirmDetection::Finalize()
{
}

// Update the scene.
void SceneConfirmDetection::Update()
{
    // Check for presence of control button, place one if there isn't one
    switch (m_State)
    {
    case STATE_ASK_TO_USE:
        if(!m_ControlManager.FindBy(CONTROL_ID_NO) && !m_ControlManager.FindBy(CONTROL_ID_YES))
        {
            m_ControlManager.Clear();
            // YES / NO Buttons
            scene::AddYesNoButtons(&m_ControlManager, CONTROL_ID_NO, CONTROL_ID_YES, L"No", L"Yes");
        }
        break;

    case STATE_SETTING_ON:
    case STATE_CANT_DETECT:
    case STATE_REMOVED:
    case STATE_LOW_BATTERY:
    case STATE_CONNECTION_SUCCESS:
        if(!m_ControlManager.FindBy(CONTROL_ID_OK))
        {
            m_ControlManager.Clear();
            // OK Button
            m_ControlManager.Add(new scene::Button(CONTROL_ID_OK, 218, 208, 100, 30, L"OK"));
        }
        break;

    case STATE_DETECTING:
        // When there is sleep processing, because the thread stops,
        // start detection sequentially
        StartDetectExtraPadThread();

        // CANCEL Button
        if(!m_ControlManager.FindBy(CONTROL_ID_CANCEL))
        {
            m_ControlManager.Clear();
            m_ControlManager.Add(new scene::Button(CONTROL_ID_CANCEL, 2, 208, 100, 30, L"Cancel"));
        }

        // Display connection confirmation message for at least 1 second
        if((nn::os::Tick::GetSystemCurrent() - m_StartDetectionTick).ToTimeSpan().GetSeconds() >= 1)
        {
            // When detecting
            if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
            {
                m_State = STATE_CONNECTION_SUCCESS;
            }

            // Detection timeout
            if((nn::os::Tick::GetSystemCurrent() - m_StartDetectionTick).ToTimeSpan().GetSeconds() >= 15)
            {
                m_State = STATE_CANT_DETECT;
            }
        }
        break;

    case STATE_CHECK_BATTERY:
        // Connection
        if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
        {
            // Class that gets information from the Circle Pad Pro and the game pad on the CTR system
            nn::hid::CTR::ExtraPadReader extraPadReader;
            nn::hid::CTR::ExtraPadStatus extraPadStatus;
            nn::hid::CTR::ExtraPadStatus extraPadStatusArray[32];
            s32 extraPadReadLen = 0;

            // Class that gets information from the game pad on the CTR system
            // (PadReader does not return key information when the Circle Pad Pro is connected)
            nn::hid::PadReader padReader;
            nn::hid::PadStatus padStatus;
            nn::hid::PadStatus padStatusArray[32];
            s32 padReadLen = 0;

            // Get input from device
            extraPadReader.ReadLatest(&extraPadStatus);
            extraPadReader.Read(extraPadStatusArray,&extraPadReadLen,32);
            padReader.ReadLatest(&padStatus);
            padReader.Read(padStatusArray,&padReadLen,32);

            // Battery level is low
            if(extraPadStatus.batteryLevel == 0)
            {
                m_State = STATE_LOW_BATTERY;
            }
            else
            {
                // Scene
                SceneConfirmDetection* pScene = (SceneConfirmDetection*)this;

                // To scene using Circle Pad Pro
                pScene->SetNextSceneId(SCENE_ID_EP_USE_EXTRA_PAD);
            }
        }
        // Not connected
        else
        {
            m_State = STATE_REMOVED;
        }
        break;

    case STATE_CHANGE_SCENE:
        // Scene
        SceneConfirmDetection* pScene = (SceneConfirmDetection*)this;

        // To scene using Circle Pad Pro
        pScene->SetNextSceneId(SCENE_ID_EP_USE_EXTRA_PAD);
        break;

    } // Switch

    m_Counter++;
    m_ControlManager.Update();
}

// Render the scene.
void SceneConfirmDetection::Draw()
{
    // Application
    Application* pApp = Application::GetInstance();
    // Renderer
    demo::RenderSystemDrawing* pRenderSystem = pApp->GetRenderSystem();

    // Text writer
    scene::TextWriter* pTextWriter = pApp->GetTextWriter();

    // Upper screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY0);
    pRenderSystem->Clear();

    pTextWriter->BeginRender(NN_GX_DISPLAY0);
    {
        scene::DrawTitle(pTextWriter);

        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(2, 20, L"[Scene] Confirm detection");
    }
    pTextWriter->EndRender();

    pRenderSystem->SwapBuffers();

    // Lower screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY1);
    pRenderSystem->Clear();

    pTextWriter->BeginRender(NN_GX_DISPLAY1);
    {
        // Display separately for each state
        f32 displayHeight = 200; // Text rendering area
        f32 y;                   // Text display Y coordinate
        f32 textHeight;          // Character height
        f32 textLeft  = 10;      // Left edge of one character
        f32 lineNum;             // Number of lines in message
        f32 defaultFontScale = pTextWriter->GetFontScale(); // Default font scale

        // Configure display messages by state
        switch (m_State)
        {
        case STATE_ASK_TO_USE:
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            textHeight = pTextWriter->CalculateStringHeight(L"");
            pTextWriter->SetTextColor(255, 255, 255);
            lineNum = sizeof(MESSAGES_ASK_TO_USE)/sizeof(MESSAGES_ASK_TO_USE[0]);
            y = (displayHeight-textHeight*lineNum)/2;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_ASK_TO_USE[i]);
                y += textHeight;
            }
            break;

        case STATE_SETTING_ON:
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            textHeight = pTextWriter->CalculateStringHeight(L"");
            pTextWriter->SetTextColor(0, 255, 0); // Green
            lineNum = sizeof(MESSAGES_SETTING_ON)/sizeof(MESSAGES_SETTING_ON[0]);
            y = (displayHeight-textHeight*lineNum)/2;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_SETTING_ON[i]);
                y += textHeight;
            }
            break;

        case STATE_DETECTING:
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            textHeight = pTextWriter->CalculateStringHeight(L"");
            pTextWriter->SetTextColor(255, 255, 255);
            lineNum = sizeof(MESSAGES_DETECTING)/sizeof(MESSAGES_DETECTING[0]);
            y = (displayHeight-textHeight*lineNum)/2;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_DETECTING[i]);
                y += textHeight;
            }
            scene::DrawBusyIcon(pTextWriter, 294, 2, m_Counter);
            break;

        case STATE_CANT_DETECT:
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            textHeight = pTextWriter->CalculateStringHeight(L"");
            pTextWriter->SetTextColor(255, 0, 0); // Red
            lineNum = sizeof(MESSAGES_CANT_DETECT)/sizeof(MESSAGES_CANT_DETECT[0]);
            y = (displayHeight-textHeight*lineNum)/2;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_CANT_DETECT[i]);
                y += textHeight;
            }
            break;

        case STATE_REMOVED:
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            textHeight = pTextWriter->CalculateStringHeight(L"");
            pTextWriter->SetTextColor(255, 0, 0); // Red
            lineNum = sizeof(MESSAGES_REMOVED)/sizeof(MESSAGES_REMOVED[0]);
            y = (displayHeight-textHeight*lineNum)/2;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_REMOVED[i]);
                y += textHeight;
            }
            break;

        case STATE_LOW_BATTERY:
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            textHeight = pTextWriter->CalculateStringHeight(L"");
            pTextWriter->SetTextColor(255, 255, 0); // Yellow
            lineNum = sizeof(MESSAGES_LOW_BATTERY)/sizeof(MESSAGES_LOW_BATTERY[0]);
            y = (displayHeight-textHeight*lineNum)/2;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_LOW_BATTERY[i]);
                y += textHeight;
            }
            break;

        case STATE_CONNECTION_SUCCESS:
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            textHeight = pTextWriter->CalculateStringHeight(L"");
            pTextWriter->SetTextColor(0, 255, 0); // Green
            lineNum = sizeof(MESSAGES_CONNECTION_SUCCESS)/sizeof(MESSAGES_CONNECTION_SUCCESS[0]);
            y = (displayHeight-textHeight*lineNum)/2;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_CONNECTION_SUCCESS[i]);
                y += textHeight;
            }
            break;

        case STATE_CHANGE_SCENE:
            // No display while switching scenes
            break;

        } // Switch

        m_ControlManager.Draw();
    }
    pTextWriter->EndRender();

    pRenderSystem->SwapBuffers();
    pRenderSystem->WaitVsync(NN_GX_DISPLAY_BOTH);
}

// Event callback
void SceneConfirmDetection::MyControlEventCallback(scene::ControlEvent event, scene::ControlBase* pControl, s16 x, s16 y, void* pParam)
{
    NN_UNUSED_VAR(x);
    NN_UNUSED_VAR(y);

    // Scene
    SceneConfirmDetection* pScene = (SceneConfirmDetection*)pParam;

    // Application
    Application* pApp = Application::GetInstance();

    if (event == scene::CONTROL_EVENT_TOUCH)
    {
        switch(pControl->GetId())
        {
        case CONTROL_ID_OK:
            if(pScene->m_State == STATE_CANT_DETECT ||
               pScene->m_State == STATE_REMOVED ||
               pScene->m_State == STATE_LOW_BATTERY)
            {
                pScene->m_State = STATE_CHANGE_SCENE;
            }
            else if(pScene->m_State == STATE_SETTING_ON ||
                pScene->m_State == STATE_CONNECTION_SUCCESS)
            {
                pScene->m_State = STATE_CHECK_BATTERY;
            }
            break;

        case CONTROL_ID_YES:
            // Enable Circle Pad Pro
            pApp->m_saveDataSet.isUseExtraPad = true;
            pScene->m_State = STATE_SETTING_ON;
            break;

        case CONTROL_ID_NO:
            pScene->m_State = STATE_CHANGE_SCENE;
            break;

        case CONTROL_ID_CANCEL:
            pScene->m_State = STATE_CHANGE_SCENE;
            break;
        }
    }
}

}}}}

