﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SceneOperateSaveData.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/


#include <nn.h>
#include "SceneOperateSaveData.h"



//////////////////////////////////////////////////////////////////////////////////////////
// Constant Definitions
//////////////////////////////////////////////////////////////////////////////////////////

namespace
{
    enum State
    {
        STATE_EXTRA_PAD_SETTING, // Circle Pad Pro use confirmation
        STATE_OPERATE_SAVE_DATA, // Save data operations

        STATE_MAX
    };

    enum ControlId
    {
        CONTROL_ID_YES,
        CONTROL_ID_NO,

        CONTROL_ID_BACK,   // Return to save data selection
        CONTROL_ID_LOAD,   // Read
        CONTROL_ID_DELETE, // Delete

        CONTROL_ID_MAX
    };
}

//////////////////////////////////////////////////////////////////////////////////////////

namespace SampleDemos {
namespace hid {
namespace ExtraPad {
namespace AppRecommendedFlow {

// Initialize a scene.
bool SceneOperateSaveData::Initialize()
{
    m_ControlManager.RegisterControlEventCallback(MyControlEventCallback, this);

    Application* pApp = Application::GetInstance();

    // Load selected save data information
    pApp->m_saveData.ReadSaveData(&(pApp->m_saveDataSet), pApp->m_SelectedSaveData);

    if(pApp->m_saveDataSet.isNewSaveData)
    {
        m_State = STATE_EXTRA_PAD_SETTING;
    }
    else
    {
        m_State = STATE_OPERATE_SAVE_DATA;
        // Save data information
        SetSaveDataDisplayStatus();
    }
    return true;
}

// Configure save data content to display
void SceneOperateSaveData::SetSaveDataDisplayStatus()
{
    Application* pApp = Application::GetInstance();

    // Load selected save data information
    pApp->m_saveData.ReadSaveData(&(pApp->m_saveDataSet), pApp->m_SelectedSaveData);

    // Play time
    if(pApp->m_saveDataSet.isNewSaveData)
    {
        std::wcscpy(m_ContentLastSavedTime, L"New data");
    }
    else
    {
        // Convert ticks to play time display (hours:minutes:seconds)
        s64 playTimeSeconds = pApp->m_saveDataSet.playTimeTick.ToTimeSpan().GetSeconds();
        s64 playTimeHours   = playTimeSeconds / 3600;
        playTimeSeconds -= playTimeHours * 3600;

        s64 playTimeMinutes = playTimeSeconds / 60;
        playTimeSeconds -= playTimeMinutes * 60;

        std::swprintf(m_ContentLastSavedTime, 20, L"%4lld:%02lld:%02lld",playTimeHours, playTimeMinutes, playTimeSeconds);
    }

    // Circle Pad Pro use settings
    if(pApp->m_saveDataSet.isUseExtraPad){
        std::wcscpy(m_ContentExtraPadSetting, L"On");
    }
    else{
        std::wcscpy(m_ContentExtraPadSetting, L"Off");
    }

    return;
}


// Finalize the scene.
void SceneOperateSaveData::Finalize()
{
}


// Update the scene.
void SceneOperateSaveData::Update()
{
    switch(m_State)
    {
    case STATE_OPERATE_SAVE_DATA:
        if(!m_ControlManager.FindBy(CONTROL_ID_BACK) ||
           !m_ControlManager.FindBy(CONTROL_ID_DELETE) ||
           !m_ControlManager.FindBy(CONTROL_ID_LOAD))
        {
            m_ControlManager.Clear();
            // BACK Button
            m_ControlManager.Add(new scene::Button(CONTROL_ID_BACK, 2, 208, 78, 30, L"Back"));
            // DELETE Button
            m_ControlManager.Add(new scene::Button(CONTROL_ID_DELETE, 318-80, 2, 80, 30, L"Delete"));
            // LOAD Button
            m_ControlManager.Add(new scene::Button(CONTROL_ID_LOAD, 84, 208, 234, 30, L"Load"));
            // Configure save data information to display
            SetSaveDataDisplayStatus();
        }

        break;

    case STATE_EXTRA_PAD_SETTING:
        if(!m_ControlManager.FindBy(CONTROL_ID_NO) ||
           !m_ControlManager.FindBy(CONTROL_ID_YES))
        {
            m_ControlManager.Clear();
            // YES / NO Buttons
            scene::AddYesNoButtons(&m_ControlManager, CONTROL_ID_NO, CONTROL_ID_YES, L"No", L"Yes");
        }
        break;
    }

    m_ControlManager.Update();
}

// Render the scene.
void SceneOperateSaveData::Draw()
{
    // Application
    Application* pApp = Application::GetInstance();
    // Renderer
    demo::RenderSystemDrawing* pRenderSystem = pApp->GetRenderSystem();

    // Text writer
    scene::TextWriter* pTextWriter = pApp->GetTextWriter();

    // Upper screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY0);
    pRenderSystem->Clear();

    pTextWriter->BeginRender(NN_GX_DISPLAY0);
    {
        scene::DrawTitle(pTextWriter);

        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(2, 20, L"[Scene] Operate save data");
    }
    pTextWriter->EndRender();
    pRenderSystem->SwapBuffers();



    // Lower screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY1);
    pRenderSystem->Clear();

    pTextWriter->BeginRender(NN_GX_DISPLAY1);
    {
        pTextWriter->SetTextColor(255, 255, 255);

        f32 textHeight;     // Character height
        f32 textLeft  = 10; // Left edge of text
        f32 lineTop   = 10; // Y coordinate of top line
        f32 lineSpace = 17; // Space between lines
        f32 defaultFontScale = pTextWriter->GetFontScale(); // Default font scale

        //// Display save data status
        // Display name of selected save data
        pTextWriter->SetFontScale(defaultFontScale*1.0f);
        pTextWriter->SetTextColor(127, 127, 127);
        textHeight = pTextWriter->CalculateStringHeight(L"");
        pTextWriter->Printf(textLeft, lineTop, L"[Selected save data]");

        pTextWriter->SetFontScale(defaultFontScale*1.2f);
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(textLeft, lineTop+textHeight, SaveData::SAVEDATA_FILENAME[pApp->m_SelectedSaveData]);
        lineTop += 4*lineSpace;

        switch(m_State)
        {
        case STATE_EXTRA_PAD_SETTING:
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            pTextWriter->SetTextColor(0, 255, 0);
            textHeight = pTextWriter->CalculateStringHeight(L""); // Character height

            pTextWriter->Printf(textLeft, lineTop, L"Use the Circle Pad Pro with this");
            lineTop += textHeight;
            pTextWriter->Printf(textLeft, lineTop, L"save data?");
            lineTop += 2*textHeight;

            pTextWriter->SetFontScale(defaultFontScale*0.9f);
            pTextWriter->Printf(textLeft, lineTop, L"NOTE: This setting can be changed later");
			
            break;

        case STATE_OPERATE_SAVE_DATA:
            // Time saved
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            pTextWriter->SetTextColor(127, 127, 127);

            textHeight = pTextWriter->CalculateStringHeight(L"[Total play time]");
            pTextWriter->Printf(textLeft, lineTop, L"[Total play time]");

            pTextWriter->SetFontScale(defaultFontScale*1.2f);
            pTextWriter->SetTextColor(255, 255, 255);

            pTextWriter->Printf(textLeft, lineTop+textHeight, m_ContentLastSavedTime);
            lineTop += 4*lineSpace;

            // Circle Pad Pro settings
            pTextWriter->SetFontScale(defaultFontScale*1.0f);
            pTextWriter->SetTextColor(127, 127, 127);

            textHeight = pTextWriter->CalculateStringHeight(L"[Extra pad setting]");
            pTextWriter->Printf(textLeft, lineTop, L"[Extra pad setting]");

            pTextWriter->SetFontScale(defaultFontScale*1.2f);
            pTextWriter->SetTextColor(255, 255, 255);

            pTextWriter->Printf(textLeft, lineTop+textHeight, m_ContentExtraPadSetting);
            lineTop += 4*lineSpace;

            break;
        }

        m_ControlManager.Draw();
    }
    pTextWriter->EndRender();

    pRenderSystem->SwapBuffers();
    pRenderSystem->WaitVsync(NN_GX_DISPLAY_BOTH);
}

// Event callback
void SceneOperateSaveData::MyControlEventCallback(scene::ControlEvent event, scene::ControlBase* pControl, s16 x, s16 y, void* pParam)
{
    NN_UNUSED_VAR(x);
    NN_UNUSED_VAR(y);

    // Scene
    SceneOperateSaveData* pScene = (SceneOperateSaveData*)pParam;

    // Application
    Application* pApp = Application::GetInstance();


    if (event == scene::CONTROL_EVENT_TOUCH)
    {
        // Save data operations state
        if(pScene->m_State == STATE_OPERATE_SAVE_DATA)
        {
            // Save selected save data
            switch(pControl->GetId())
            {
            case CONTROL_ID_BACK:
                // To data selection screen
                pScene->SetNextSceneId(SCENE_ID_EP_SELECT_SAVE_DATA);
                break;

            case CONTROL_ID_LOAD:
                // Get tick when save data load began
                pApp->m_saveData.startTick = nn::os::Tick::GetSystemCurrent();

                // To connection confirmation scene
                pScene->SetNextSceneId(SCENE_ID_EP_CONFIRM_DETECTION);
                break;

            case CONTROL_ID_DELETE:
                // Delete save data
                pApp->m_saveData.DeleteSaveDataFile(pApp->m_SelectedSaveData);

                // To data selection screen
                pScene->SetNextSceneId(SCENE_ID_EP_SELECT_SAVE_DATA);
                break;
            }
        }
        // Circle Pad Pro use settings
        else if(pScene->m_State == STATE_EXTRA_PAD_SETTING)
        {
            pApp->m_saveData.ReadSaveData(&(pApp->m_saveDataSet), pApp->m_SelectedSaveData);
            switch(pControl->GetId())
            {
            case CONTROL_ID_NO:
                pApp->m_saveDataSet.isUseExtraPad = false; // Don't use Circle Pad Pro
                pApp->m_saveDataSet.isNewSaveData = false; // Clear flag for determining new data
                pScene->m_State = STATE_OPERATE_SAVE_DATA;
                break;

            case CONTROL_ID_YES:
                pApp->m_saveDataSet.isUseExtraPad = true;  // Use Circle Pad Pro
                pApp->m_saveDataSet.isNewSaveData = false; // Clear flag for determining new data
                pScene->m_State = STATE_OPERATE_SAVE_DATA;
                break;
            }
            pApp->m_saveData.WriteSaveData(&(pApp->m_saveDataSet), pApp->m_SelectedSaveData);

        }
    }
}

}}}}

