﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SceneOptionMenu.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/


#include <nn.h>
#include <stdio.h>
#include <string.h>
#include <wchar.h>
#include "SceneOptionMenu.h"

//////////////////////////////////////////////////////////////////////////////////////////
// Constant Definitions
//////////////////////////////////////////////////////////////////////////////////////////

namespace
{
    enum State
    {
        STATE_EXTRA_PAD_ON,       // Use enabled
        STATE_EXTRA_PAD_OFF,      // Use disabled
        STATE_EXTRA_PAD_APPLET,   // Connect calibration applet
        STATE_ASK_TO_CONFIRM,     // Whether to confirm connection
        STATE_CONNECTING,         // Connecting to Circle Pad Pro
        STATE_CONNECTION_SUCCESS, // Connection successful
        STATE_CONNECTION_FAILURE, // Connection failure
        STATE_DISCONNECT,         // Disconnecting after use disabled

        STATE_MAX
    };

    enum ControlId
    {
        CONTROL_ID_USE_EXTRA_PAD_ON,    // Circle Pad Pro enabled setting: Use
        CONTROL_ID_USE_EXTRA_PAD_OFF,   // Circle Pad Pro enabled setting: Not Used
        CONTROL_ID_CONNECT_EXTRA_PAD,   // Confirming connection to Circle Pad Pro
        CONTROL_ID_EXTRA_PAD_CALIBRATE, // Circle Pad Pro calibration applet
        CONTROL_ID_SAVE,                // Save
        CONTROL_ID_RETURN_EXTRA_PAD,    // Return to scene using Circle Pad Pro
        CONTROL_ID_RETURN_SELECT_DATA,  // Return to save data selection scene

        CONTROL_ID_OK,
        CONTROL_ID_CANCEL,
        CONTROL_ID_YES,
        CONTROL_ID_NO,

        CONTROL_ID_MAX
    };

    // Button size
    const u16 BUTTON_WIDTH  =  55;  // Button width
    const u16 BUTTON_HEIGHT =  22;  // Button height
    const u16 BUTTON_SPACE  =  10;  // Button spacing
    const u16 BUTTON_X      = 180;  // Button X coordinate
    const u16 BUTTON_Y_TOP  =  30;  // Top button Y coordinate
}

//////////////////////////////////////////////////////////////////////////////////////////

namespace SampleDemos {
namespace hid {
namespace ExtraPad {
namespace AppRecommendedFlow {


const wchar_t* SceneOptionMenu::MESSAGES_ASK_TO_CONFIRM[] = {
    L"Circle Pad Pro enabled.",
    L"Check connection?"
};

const wchar_t* SceneOptionMenu::MESSAGES_CONNECTING[] = {
    L"Checking connection to Circle Pad Pro.",
    L"If in the standby state, press the ZR",
    L"Button or the ZL Button."
};

const wchar_t* SceneOptionMenu::MESSAGES_CONNECTION_SUCCESS[] = {
    L"Confirmed connection to Circle Pad Pro."
};

const wchar_t* SceneOptionMenu::MESSAGES_CONNECTION_FAILURE[] = {
    L"Could not connect. Check whether the",
    L"Circle Pad Pro is connected correctly",
    L"and the battery is sufficiently charged."
};

const wchar_t* SceneOptionMenu::MESSAGES_DISCONNECT[] = {
    L"Circle Pad Pro disconnected."
};

// Initialize a scene.
bool SceneOptionMenu::Initialize()
{
    // Register control callback event
    m_ControlManager.RegisterControlEventCallback(MyControlEventCallback, this);

    // Application
    Application* pApp = Application::GetInstance();

    // Circle Pad Pro use settings
    if(pApp->m_saveDataSet.isUseExtraPad)
    {
        m_State   = STATE_EXTRA_PAD_ON;
    }
    else
    {
        m_State   = STATE_EXTRA_PAD_OFF;
    }

    m_Counter = 0;
    m_Titele  = L"OPTIONS";

    return true;
}

// Finalize the scene.
void SceneOptionMenu::Finalize()
{
}


// Update the scene.
void SceneOptionMenu::Update()
{
    // Get input from device
    while(!m_ExtraPadReader.ReadLatest(&m_ExtraPadStatus))
    {
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(10));
    }
    m_ExtraPadReader.Read(m_ExtraPadStatusArray,&m_ExtraPadReadLen,32);
    m_PadReader.ReadLatest(&m_PadStatus);
    m_PadReader.Read(m_PadStatusArray,&m_PadReadLen,32);

    // To settings changing scene after pressing START Button
    if(m_ExtraPadStatus.release & (nn::hid::BUTTON_START))
    {
        // Scene
        SceneOptionMenu* pScene = (SceneOptionMenu*)this;

        StopDetectExtraPadThread();
        // To scene using Circle Pad Pro
        pScene->SetNextSceneId(SCENE_ID_EP_USE_EXTRA_PAD);
        return;
    }


    // Check for presence of button, place one if there isn't one
    m_ControlManager.Clear();
    switch (m_State)
    {
    case STATE_EXTRA_PAD_ON:
    case STATE_EXTRA_PAD_OFF:
        SetControlButton(CONTROL_ID_USE_EXTRA_PAD_ON);
        SetControlButton(CONTROL_ID_USE_EXTRA_PAD_OFF);
        SetControlButton(CONTROL_ID_CONNECT_EXTRA_PAD);
        SetControlButton(CONTROL_ID_EXTRA_PAD_CALIBRATE);
        SetControlButton(CONTROL_ID_SAVE);
        SetControlButton(CONTROL_ID_RETURN_EXTRA_PAD);
        SetControlButton(CONTROL_ID_RETURN_SELECT_DATA);
        break;

    case STATE_ASK_TO_CONFIRM:
        SetControlButton(CONTROL_ID_YES);
        break;

    case STATE_CONNECTING:
        SetControlButton(CONTROL_ID_CANCEL);

        // When there is sleep processing, because the thread stops,
        // start detection sequentially
        StartDetectExtraPadThread();

        // Display for at least 1 second
        if((nn::os::Tick::GetSystemCurrent() - m_StartDetectionTick).ToTimeSpan().GetSeconds() >= 1)
        {
            // When detecting
            if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
            {
                // Stop detection
                StopDetectExtraPadThread();
                m_State = STATE_CONNECTION_SUCCESS;
            }
            // Detection timeout
            if((nn::os::Tick::GetSystemCurrent() - m_StartDetectionTick).ToTimeSpan().GetSeconds() >= 15)
            {
                // Stop detection
                StopDetectExtraPadThread();
                m_State = STATE_CONNECTION_FAILURE;
            }
        }
        break;

    case STATE_CONNECTION_SUCCESS:
    case STATE_CONNECTION_FAILURE:
    case STATE_DISCONNECT:
        SetControlButton(CONTROL_ID_OK);
        break;
    }

    m_Counter++;
    m_ControlManager.Update();
}



// Confirm presence of control button, place one if there isn't one.
void SceneOptionMenu::SetControlButton(s32 id)
{
    if(!m_ControlManager.FindBy(id))
    {
        switch(id)
        {
        case CONTROL_ID_USE_EXTRA_PAD_ON:
            m_ControlManager.Add(new scene::Button(CONTROL_ID_USE_EXTRA_PAD_ON,
                BUTTON_X, BUTTON_Y_TOP,
                BUTTON_WIDTH, BUTTON_HEIGHT, L"On"));
            break;

        case CONTROL_ID_USE_EXTRA_PAD_OFF:
            m_ControlManager.Add(new scene::Button(CONTROL_ID_USE_EXTRA_PAD_OFF,
                BUTTON_X+BUTTON_WIDTH+BUTTON_SPACE, BUTTON_Y_TOP,
                BUTTON_WIDTH, BUTTON_HEIGHT, L"Off"));
            break;

        case CONTROL_ID_CONNECT_EXTRA_PAD:
            m_ControlManager.Add(new scene::Button(CONTROL_ID_CONNECT_EXTRA_PAD,
                BUTTON_X, BUTTON_Y_TOP + BUTTON_HEIGHT + BUTTON_SPACE,
                2*BUTTON_WIDTH+BUTTON_SPACE, BUTTON_HEIGHT, L"Connect"));
            break;

        case CONTROL_ID_EXTRA_PAD_CALIBRATE:
            m_ControlManager.Add(new scene::Button(CONTROL_ID_EXTRA_PAD_CALIBRATE,
                BUTTON_X, BUTTON_Y_TOP + 2*BUTTON_HEIGHT + 2*BUTTON_SPACE,
                2*BUTTON_WIDTH+BUTTON_SPACE, BUTTON_HEIGHT, L"Calibrate"));
            break;

        case CONTROL_ID_SAVE:
            m_ControlManager.Add(new scene::Button(CONTROL_ID_SAVE,
                BUTTON_X, BUTTON_Y_TOP + 3.5*BUTTON_HEIGHT + 3.5*BUTTON_SPACE,
                2*BUTTON_WIDTH+BUTTON_SPACE, BUTTON_HEIGHT, L"Save"));
            break;

        case CONTROL_ID_RETURN_EXTRA_PAD:
            m_ControlManager.Add(new scene::Button(CONTROL_ID_RETURN_EXTRA_PAD, 60, 189,
                198, 25, L"Return to play"));
            break;

        case CONTROL_ID_RETURN_SELECT_DATA:
            m_ControlManager.Add(new scene::Button(CONTROL_ID_RETURN_SELECT_DATA, 60, 218,
                198, 20, L"Select save data"));
            break;

        case CONTROL_ID_OK:
            m_ControlManager.Add(new scene::Button(CONTROL_ID_OK, 218, 208, 100, 30, L"OK"));
            break;

        case CONTROL_ID_CANCEL:
            m_ControlManager.Add(new scene::Button(CONTROL_ID_CANCEL, 2, 208, 100, 30, L"Cancel"));
            break;

        case CONTROL_ID_YES:
        case CONTROL_ID_NO:
            scene::AddYesNoButtons(&m_ControlManager, CONTROL_ID_NO, CONTROL_ID_YES, L"No", L"Yes");
            break;
        }
    }
}

// Render the scene.
void SceneOptionMenu::Draw()
{
    // Application
    Application* pApp = Application::GetInstance();
    // Renderer
    demo::RenderSystemDrawing* pRenderSystem = pApp->GetRenderSystem();
    // Text writer
    scene::TextWriter* pTextWriter = pApp->GetTextWriter();

    f32 displayHeight = 200; // Text rendering area
    f32 y;                   // Text display Y coordinate
    f32 textHeight;          // Character height
    f32 textWidth;           // Width of one character
    f32 textLeft  = 10;      // Left edge of one character
    f32 lineNum;             // Number of lines in message
    u32 lineTop = 100;       // Y coordinate of first line
    u32 lineHeight;          // Space between lines
    f32 defaultFontScale = pTextWriter->GetFontScale(); // Default font scale

    // Upper screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY0);
    pRenderSystem->Clear();

    pTextWriter->BeginRender(NN_GX_DISPLAY0);
    {
        scene::DrawTitle(pTextWriter);

        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(2, 20, L"[Scene] Option menu");


        // Display the various statuses
        pTextWriter->SetFontScale(defaultFontScale*0.8f);
        lineHeight = pTextWriter->CalculateStringHeight(L"");

        // Save data to use
        // Name of selected save data
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Selected save data");
        pTextWriter->Printf(200-textWidth, lineTop, L"Selected save data");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": ");
        textWidth = pTextWriter->CalculateStringWidth(L": ");
        pTextWriter->Printf(200+textWidth, lineTop, SaveData::SAVEDATA_FILENAME[pApp->m_SelectedSaveData]);
        lineTop += lineHeight;

        // Current total play time
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Total play time");
        pTextWriter->Printf(200-textWidth, lineTop, L"Total play time");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": ");
        textWidth = pTextWriter->CalculateStringWidth(L": ");
        // Convert ticks to play time display (hours:minutes:seconds)
        s64 playTimeSeconds = (pApp->m_saveDataSet.playTimeTick + nn::os::Tick::GetSystemCurrent() - pApp->m_saveData.startTick).ToTimeSpan().GetSeconds();
        s64 playTimeHours   = playTimeSeconds / 3600;
        playTimeSeconds -= playTimeHours * 3600;
        s64 playTimeMinutes = playTimeSeconds / 60;
        playTimeSeconds -= playTimeMinutes * 60;
        pTextWriter->Printf(200+textWidth, lineTop, L"%4lld:%02lld:%02lld",playTimeHours, playTimeMinutes, playTimeSeconds);
        lineTop += lineHeight;

        // Circle Pad Pro use settings
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Extra pad setting");
        pTextWriter->Printf(200-textWidth, lineTop, L"Extra pad setting");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": %s",
            pApp->m_saveDataSet.isUseExtraPad ? "On" : "Off");
        lineTop += lineHeight;

        lineTop += lineHeight; // Internal parameter boundary

        // Circle Pad Pro connection state
        if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
        {
            pTextWriter->SetTextColor(127, 127, 127);
            textWidth = pTextWriter->CalculateStringWidth(L"Extra pad connection state");
            pTextWriter->Printf(200-textWidth, lineTop, L"Extra pad connection state");
            pTextWriter->SetTextColor(255, 255, 255);
            pTextWriter->Printf(200, lineTop, L": Connected");
            lineTop += lineHeight;

            pTextWriter->SetTextColor(127, 127, 127);
            textWidth = pTextWriter->CalculateStringWidth(L"Battery level");
            pTextWriter->Printf(200-textWidth, lineTop, L"Battery level");
            pTextWriter->SetTextColor(255, 255, 255);
            pTextWriter->Printf(200, lineTop, L": %s",
                m_ExtraPadStatus.batteryLevel ? "Normal" : "Low");
        }
        else
        {
            pTextWriter->SetTextColor(127, 127, 127);
            textWidth = pTextWriter->CalculateStringWidth(L"Extra pad connection state");
            pTextWriter->Printf(200-textWidth, lineTop, L"Extra pad connection state");
            pTextWriter->SetTextColor(255, 255, 255);
            pTextWriter->Printf(200, lineTop, L": Not connected");
            lineTop += lineHeight;
        }
        lineTop += lineHeight;


        lineTop += lineHeight; // Boundary with processing determination flag

        // Display presence of thread
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Detect extra pad thread flag");
        pTextWriter->Printf(200-textWidth, lineTop, L"Detect extra pad thread flag");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": %s",
            IsAliveDetectExtraPadThread() ? "true" : "false");
        lineTop += lineHeight;

        // First connection since starting application
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Extra pad first connection flag");
        pTextWriter->Printf(200-textWidth, lineTop, L"Extra pad first connection flag");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": %s",
        pApp->m_IsFirstConnect ? "true" : "false");
        lineTop += lineHeight;

    }
    pTextWriter->EndRender();
    pRenderSystem->SwapBuffers();


    // Lower screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY1);
    pRenderSystem->Clear();

    pTextWriter->BeginRender(NN_GX_DISPLAY1);
    {
        pRenderSystem->SetAutoDepth(true);

        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(m_Titele);
        pTextWriter->Printf((320 - textWidth)/2, 2, m_Titele);
        pTextWriter->SetFontScale(defaultFontScale);

        textHeight = pTextWriter->CalculateStringHeight(L"");

        // Configure display messages by state
        switch(m_State)
        {
        case STATE_EXTRA_PAD_ON:
        case STATE_EXTRA_PAD_OFF:
            pTextWriter->SetTextColor(127, 127, 127);
            textWidth = pTextWriter->CalculateStringWidth(L"Extra pad setting");
            pTextWriter->Printf(160-textWidth, BUTTON_Y_TOP, L"Extra pad setting");

//            if(m_State == STATE_EXTRA_PAD_ON)
//            {
                textWidth = pTextWriter->CalculateStringWidth(L"Connect extra pad");
                pTextWriter->Printf(160-textWidth, BUTTON_Y_TOP + BUTTON_HEIGHT + BUTTON_SPACE, L"Connect extra pad");

                textWidth = pTextWriter->CalculateStringWidth(L"Extra pad calibration");
                pTextWriter->Printf(160-textWidth, BUTTON_Y_TOP + 2*BUTTON_HEIGHT + 2*BUTTON_SPACE, L"Extra pad calibration");
//            }

            textWidth = pTextWriter->CalculateStringWidth(L"Save current setting");
            pTextWriter->Printf(160-textWidth, BUTTON_Y_TOP + 3.5*BUTTON_HEIGHT + 3.5*BUTTON_SPACE, L"Save current setting");
            break;
        case STATE_DISCONNECT:
            pTextWriter->SetTextColor(255, 255, 255);
            lineNum = sizeof(MESSAGES_DISCONNECT)/sizeof(MESSAGES_DISCONNECT[0]);
            y = (displayHeight-BUTTON_Y_TOP-textHeight*lineNum)/2 +BUTTON_Y_TOP;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_DISCONNECT[i]);
                y += textHeight;
            }
            break;

        case STATE_ASK_TO_CONFIRM:
            pTextWriter->SetTextColor(255, 255, 255);
            lineNum = sizeof(MESSAGES_ASK_TO_CONFIRM)/sizeof(MESSAGES_ASK_TO_CONFIRM[0]);
            y = (displayHeight-BUTTON_Y_TOP-textHeight*lineNum)/2 +BUTTON_Y_TOP;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_ASK_TO_CONFIRM[i]);
                y += textHeight;
            }
            break;

        case STATE_CONNECTING:
            pTextWriter->SetTextColor(255, 255, 255);
            lineNum = sizeof(MESSAGES_CONNECTING)/sizeof(MESSAGES_CONNECTING[0]);
            y = (displayHeight-BUTTON_Y_TOP-textHeight*lineNum)/2 +BUTTON_Y_TOP;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_CONNECTING[i]);
                y += textHeight;
            }
            scene::DrawBusyIcon(pTextWriter, 294, 2, m_Counter);
            break;

        case STATE_CONNECTION_SUCCESS:
            pTextWriter->SetTextColor(0, 255, 0);
            lineNum = sizeof(MESSAGES_CONNECTION_SUCCESS)/sizeof(MESSAGES_CONNECTION_SUCCESS[0]);
            y = (displayHeight-BUTTON_Y_TOP-textHeight*lineNum)/2 +BUTTON_Y_TOP;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_CONNECTION_SUCCESS[i]);
                y += textHeight;
            }
            break;

        case STATE_CONNECTION_FAILURE:
            pTextWriter->SetTextColor(255, 0, 0);
            lineNum = sizeof(MESSAGES_CONNECTION_FAILURE)/sizeof(MESSAGES_CONNECTION_FAILURE[0]);
            y = (displayHeight-BUTTON_Y_TOP-textHeight*lineNum)/2 +BUTTON_Y_TOP;
            for(int i=0; i<lineNum; i++)
            {
                pTextWriter->Printf(textLeft, y, MESSAGES_CONNECTION_FAILURE[i]);
                y += textHeight;
            }
            break;
        }

        m_ControlManager.Draw();
    }
    pTextWriter->EndRender();


    // Border depends on Circle Pad Pro enabled setting
    if(m_State == STATE_EXTRA_PAD_ON ||m_State == STATE_EXTRA_PAD_OFF)
    {
        f32 x1, y1, x2, y2;

        if(pApp->m_saveDataSet.isUseExtraPad)
        {
            x1 = BUTTON_X-2;
            y1 = BUTTON_Y_TOP-2;
            x2 = BUTTON_X+BUTTON_WIDTH+2;
            y2 = BUTTON_Y_TOP+BUTTON_HEIGHT+2;
        }
        else
        {
            x1 = BUTTON_X-2 +BUTTON_SPACE+BUTTON_WIDTH;
            y1 = BUTTON_Y_TOP-2;
            x2 = BUTTON_X+BUTTON_WIDTH+2 +BUTTON_SPACE+BUTTON_WIDTH;
            y2 = BUTTON_Y_TOP+BUTTON_HEIGHT+2;
        }
        pRenderSystem->SetColor( 45/225.f, 116/225.f, 174/225.f, 1.0f);
        pRenderSystem->SetLineWidth(3);

        pRenderSystem->DrawLine(x1  , y1  ,x2  , y1  );
        pRenderSystem->DrawLine(x1  , y2-1,x2  , y2-1);
        pRenderSystem->DrawLine(x1  , y1  ,x1  , y2  );
        pRenderSystem->DrawLine(x2-1, y1  ,x2-1, y2  );
    }

//    // Display in gray when disabled
//    if(m_State == STATE_EXTRA_PAD_OFF)
//    {
//        f32 x1, x2, y1, y2, y3, y4;
//
//        x1 = BUTTON_X;
//        x2 = x1 + 2 * BUTTON_WIDTH + BUTTON_SPACE;
//
//        y1 = BUTTON_Y_TOP + BUTTON_HEIGHT + BUTTON_SPACE;
//        y2 = y1 + BUTTON_HEIGHT;
//        y3 = y2 + BUTTON_SPACE;
//        y4 = y3 + BUTTON_HEIGHT;
//
//        pRenderSystem->SetColor(1.0f, 1.0f, 1.0f);
//        pRenderSystem->SetLineWidth(2);
//
//        // Horizontal line
//        pRenderSystem->DrawLine(x1  , y1  ,x2  , y1  );
//        pRenderSystem->DrawLine(x1  , y2  ,x2  , y2  );
//        pRenderSystem->DrawLine(x1  , y3  ,x2  , y3  );
//        pRenderSystem->DrawLine(x1  , y4  ,x2  , y4  );
//
//        // Vertical line
//        pRenderSystem->DrawLine(x1  , y1  ,x1  , y2  );
//        pRenderSystem->DrawLine(x2  , y1  ,x2  , y2  );
//        pRenderSystem->DrawLine(x1  , y1  ,x1  , y2  );
//        pRenderSystem->DrawLine(x2-1, y1  ,x2-1, y2  ); //
//    }


    pRenderSystem->SwapBuffers();
    pRenderSystem->WaitVsync(NN_GX_DISPLAY_BOTH);
}


// Event callback
void SceneOptionMenu::MyControlEventCallback(scene::ControlEvent event, scene::ControlBase* pControl, s16 x, s16 y, void* pParam)
{
    NN_UNUSED_VAR(x);
    NN_UNUSED_VAR(y);


    if (event == scene::CONTROL_EVENT_TOUCH)
    {
        // Scene
        SceneOptionMenu* pScene = (SceneOptionMenu*)pParam;

        // Application
        Application* pApp = Application::GetInstance();

        // Save selected save data
        switch(pControl->GetId())
        {
        case CONTROL_ID_USE_EXTRA_PAD_ON:
            // When changing setting from disabled
            if(!pApp->m_saveDataSet.isUseExtraPad)
            {
                pScene->m_State = STATE_ASK_TO_CONFIRM;
            }
            // Change Circle Pad Pro enabled setting
            pApp->m_saveDataSet.isUseExtraPad = true;
            break;

        case CONTROL_ID_USE_EXTRA_PAD_OFF:
            // When changing setting from enabled
            if(pApp->m_saveDataSet.isUseExtraPad)
            {
                if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
                {
                    pScene->m_State = STATE_DISCONNECT;
                }
            }
            else
            {
                pScene->m_State = STATE_EXTRA_PAD_OFF;
            }
            // Change Circle Pad Pro enabled setting
            pApp->m_saveDataSet.isUseExtraPad = false;
            StopDetectExtraPadThread();
            nn::hid::ExtraPad::StopSampling();
            break;

        case CONTROL_ID_CONNECT_EXTRA_PAD:
        case CONTROL_ID_YES:
            // Start detecting
            StartDetectExtraPadThread();
            pScene->m_State = STATE_CONNECTING;
            // Get tick from when detection started
            pScene->m_StartDetectionTick = nn::os::Tick::GetSystemCurrent();
            break;

        case CONTROL_ID_EXTRA_PAD_CALIBRATE:
            // Call Circle Pad Pro calibration applet

            // Stop sampling before launching this applet
            StopDetectExtraPadThread();
            nn::hid::ExtraPad::StopSampling();

            // ExtraPad settings structure
            nn::extrapad::Parameter extraPadParam;
            nn::applet::AppletWakeupState wstate;

            // Initialize settings struct
            nn::extrapad::InitializeConfig( &extraPadParam.config );

            // Start the applet
            nn::extrapad::StartExtraPadApplet( &wstate, &extraPadParam );

            // Recover the GPU register settings
            nngxUpdateState( NN_GX_STATE_ALL );
            nngxValidateState( NN_GX_STATE_ALL,GL_TRUE );

            // Start detecting
            StartDetectExtraPadThread();

            break;

        case CONTROL_ID_SAVE:
            // Clear flag for determining new data
            pApp->m_saveDataSet.isNewSaveData = false;

            // Update play time
            pApp->m_saveDataSet.playTimeTick += nn::os::Tick::GetSystemCurrent() - pApp->m_saveData.startTick;
            pApp->m_saveData.WriteSaveData(&(pApp->m_saveDataSet), pApp->m_SelectedSaveData);
            pApp->m_saveData.startTick = nn::os::Tick::GetSystemCurrent();
            break;

        case CONTROL_ID_RETURN_EXTRA_PAD:
            // Stop detection
            StopDetectExtraPadThread();
            // To scene using Circle Pad Pro
            pScene->SetNextSceneId(SCENE_ID_EP_USE_EXTRA_PAD);
            break;

        case CONTROL_ID_RETURN_SELECT_DATA:
            // Stop detection
            StopDetectExtraPadThread();
            // To save data selection scene
            pScene->SetNextSceneId(SCENE_ID_EP_SELECT_SAVE_DATA);
            break;

        case CONTROL_ID_NO:              // Cancel connection confirmation after enabling
        case CONTROL_ID_CANCEL:          // Cancel connection confirmation from Connect
            StopDetectExtraPadThread();  // Stop detection
        case CONTROL_ID_OK:
            // Circle Pad Pro use settings
            if(pApp->m_saveDataSet.isUseExtraPad)
            {
                pScene->m_State = STATE_EXTRA_PAD_ON;
            }
            else
            {
                pScene->m_State = STATE_EXTRA_PAD_OFF;
            }
            break;

        }
    }

}}}}}

