﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SceneUseExtraPad.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/


#include <nn.h>
#include "SceneUseExtraPad.h"

//////////////////////////////////////////////////////////////////////////////////////////
// Constant Definitions
//////////////////////////////////////////////////////////////////////////////////////////

namespace
{
    enum State
    {
        STATE_DETECT_EXTRA_PAD,    // Circle Pad Pro detected
        STATE_UNDETECT_EXTRA_PAD,  // Circle Pad Pro not detected

        STATE_MAX
    };
}

//////////////////////////////////////////////////////////////////////////////////////////

namespace SampleDemos {
namespace hid {
namespace ExtraPad {
namespace AppRecommendedFlow {

// Initialize a scene.
bool SceneUseExtraPad::Initialize()
{
    // Application
    Application* pApp = Application::GetInstance();

    m_ExtraPadReadLen = 0;
    m_PadReadLen      = 0;

    m_IsLowBattery = false;

    m_MessageType = MESSAGE_NONE;
    m_MessageProperty.isInitialized = false;
    m_MessageProperty.counter = 0;

    // Process branching depending on whether Circle Pad Pro is enabled
    if(pApp->m_saveDataSet.isUseExtraPad)
    {
        // Detect connection events
        m_ConnectionEvent.Initialize(false); // Manual reset setting
        // Register event
        nn::hid::ExtraPad::RegisterConnectionEvent(&m_ConnectionEvent);

        // Start detecting
        StartDetectExtraPadThread();

        // Connection
        if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
        {
            m_State = STATE_DETECT_EXTRA_PAD;   // Detected
            StopDetectExtraPadThread();
        }
        else
        {
            m_State = STATE_UNDETECT_EXTRA_PAD; // Not detected
        }
    }
    else
    {
        m_State = STATE_UNDETECT_EXTRA_PAD; // Not detected
        StopDetectExtraPadThread();
        nn::hid::ExtraPad::StopSampling();
    }
    return true;
}


// Finalize the scene.
void SceneUseExtraPad::Finalize()
{
    // Application
    Application* pApp = Application::GetInstance();

    // If Circle Pad Pro enabled
    if(pApp->m_saveDataSet.isUseExtraPad)
    {
        // Event detection
        nn::hid::ExtraPad::UnregisterConnectionEvent();
        m_ConnectionEvent.Finalize();
    }
}


// Update the scene.
void SceneUseExtraPad::Update()
{
    // Application
    Application* pApp = Application::GetInstance();

    // Get input from device
    while(!m_ExtraPadReader.ReadLatest(&m_ExtraPadStatus))
    {
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(10));
    }
    m_ExtraPadReader.Read(m_ExtraPadStatusArray,&m_ExtraPadReadLen,32);
    m_PadReader.ReadLatest(&m_PadStatus);
    m_PadReader.Read(m_PadStatusArray,&m_PadReadLen,32);

    // To settings changing scene after pressing START Button
    if(m_ExtraPadStatus.release & (nn::hid::BUTTON_START))
    {
        // Scene
        SceneUseExtraPad* pScene = (SceneUseExtraPad*)this;

        // To Options menu scene
        pScene->SetNextSceneId(SCENE_ID_EP_OPTION_MENU);
        return;
    }


    // Update state of object being operated on
    UpdateObjectState();


    // Process branching depending on whether Circle Pad Pro is enabled
    if(pApp->m_saveDataSet.isUseExtraPad)
    {
        // When there is sleep processing, because the thread stops,
        // start detection sequentially
        StartDetectExtraPadThread();

        // When state has changed
        if(m_ConnectionEvent.TryWait())
        {
            // Connection
            if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
            {
                StopDetectExtraPadThread(); // Stop detection thread
                m_State = STATE_DETECT_EXTRA_PAD; // Detection state
                m_MessageType = MESSAGE_CONNECT;  // Connection message
                m_MessageProperty.counter = 0;
                m_MessageProperty.displayY = -m_MessageProperty.textHeight;

                // When first detecting after launch
                if(pApp->m_IsFirstConnect)
                {
                    pApp->m_IsFirstConnect = false; // Clear flag for first-time detection after launch

                    // Battery level is low
                    if(m_ExtraPadStatus.batteryLevel == 0)
                    {
                        m_IsLowBattery = true;
                    }
                }
            }

            //Disconnect
            else if((nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_NO_CONNECTION)
                || (nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_STOPPED))
            {
                m_State = STATE_UNDETECT_EXTRA_PAD; // Undetected state
                if(m_MessageType != MESSAGE_DISCONNECT)
                {
                    m_MessageType = MESSAGE_DISCONNECT; // Disconnection message
                    m_MessageProperty.counter = 0;
                    m_MessageProperty.displayY = -m_MessageProperty.textHeight;
                }

                // Start detecting
                StartDetectExtraPadThread();
            }
        }
    }


    // Update message
    switch(m_MessageType)
    {
    case MESSAGE_DISCONNECT:
    case MESSAGE_CONNECT:
        // Lower display position
        if(m_MessageProperty.counter < 170)
        {
            if(m_MessageProperty.displayY <= 0)
            {
                m_MessageProperty.displayY++;
            }
        }
        // Raise display position
        else if(m_MessageProperty.counter >= 170)
        {
            m_MessageProperty.displayY--;

            // Stop displaying message when moving out of screen
            if(m_MessageProperty.displayY <= -m_MessageProperty.textHeight)
            {

                // To low battery warning if flag is set
                if(m_IsLowBattery)
                {
                    m_IsLowBattery = false;
                    m_MessageType = MESSAGE_LOW_BATTERY;  // Connection message
                    m_MessageProperty.counter = 0;
                    m_MessageProperty.displayY = -m_MessageProperty.textHeight * 2;
                    break;
                }
                else
                {
                    m_MessageType = MESSAGE_NONE;
                    m_MessageProperty.counter = 0;
                    break;
                }
            }
        }
        m_MessageProperty.counter++;
        break;

    case MESSAGE_LOW_BATTERY:
        // Lower display position
        if(m_MessageProperty.counter < 200)
        {
            if(m_MessageProperty.displayY <= 0)
            {
                m_MessageProperty.displayY++;
            }
        }
        // Raise display position
        else if(m_MessageProperty.counter >= 200)
        {
            m_MessageProperty.displayY--;

            // Stop displaying message when moving out of screen
            if(m_MessageProperty.displayY <= -m_MessageProperty.textHeight*2)
            {
                m_MessageType = MESSAGE_NONE;
                m_MessageProperty.counter = 0;
                break;
            }
        }
        m_MessageProperty.counter++;
        break;
    }

    m_ControlManager.Update();
}


// Updates state of the object being operated on.
void SceneUseExtraPad::UpdateObjectState()
{
    // Application
    Application* pApp = Application::GetInstance();

    // Get input from device
    while(!m_ExtraPadReader.ReadLatest(&m_ExtraPadStatus))
    {
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(10));
    }
    m_ExtraPadReader.Read(m_ExtraPadStatusArray,&m_ExtraPadReadLen,32);
    m_PadReader.ReadLatest(&m_PadStatus);
    m_PadReader.Read(m_PadStatusArray,&m_PadReadLen,32);

    //// Update object information
    // Position (X)
    if(nn::math::FAbs(m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.stick.x)) < 0.8f)
    {
        pApp->m_saveDataSet.objectPositionX += 2*m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.stick.x);
        if(pApp->m_saveDataSet.objectPositionX < 0 )
        {
            pApp->m_saveDataSet.objectPositionX = 0;
        }
        if(320 < pApp->m_saveDataSet.objectPositionX)
        {
            pApp->m_saveDataSet.objectPositionX = 320;
        }
    }
    else if(nn::math::FAbs(m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.stick.x)) >= 0.8f)
    {
        pApp->m_saveDataSet.objectPositionX += 3*m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.stick.x);
        if(pApp->m_saveDataSet.objectPositionX < 0 )
        {
            pApp->m_saveDataSet.objectPositionX = 0;
        }
        if(320 < pApp->m_saveDataSet.objectPositionX)
        {
            pApp->m_saveDataSet.objectPositionX = 320;
        }
    }

    // Position (Y)
    if(nn::math::FAbs(m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.stick.y)) < 0.8f)
    {
        pApp->m_saveDataSet.objectPositionY -= 2*m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.stick.y);
        if(pApp->m_saveDataSet.objectPositionY < 0 )
        {
            pApp->m_saveDataSet.objectPositionY = 0;
        }
        if(240 < pApp->m_saveDataSet.objectPositionY)
        {
            pApp->m_saveDataSet.objectPositionY = 240;
        }
    }
    else if(nn::math::FAbs(m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.stick.y)) >= 0.8f)
    {
        pApp->m_saveDataSet.objectPositionY -= 3*m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.stick.y);
        if(pApp->m_saveDataSet.objectPositionY < 0 )
        {
            pApp->m_saveDataSet.objectPositionY = 0;
        }
        if(240 < pApp->m_saveDataSet.objectPositionY)
        {
            pApp->m_saveDataSet.objectPositionY = 240;
        }
    }

    // Angle (Circle Pad Pro)
    if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
    {
        if(nn::math::FAbs(m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.extraStick.x)) < 0.8f)
        {
            pApp->m_saveDataSet.objectDegree += 5*m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.extraStick.x);
        }
        else if(nn::math::FAbs(m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.extraStick.x)) >= 0.8f)
        {
            pApp->m_saveDataSet.objectDegree += 10*m_ExtraPadReader.NormalizeExtraStick(m_ExtraPadStatus.extraStick.x);
        }
        // Keep angle's absolute value within the range of -360 to 360
        pApp->m_saveDataSet.objectDegree = nn::math::FMod(pApp->m_saveDataSet.objectDegree, 360);
    }

    // Angle (+Control Pad)
    if(m_ExtraPadStatus.hold & (nn::hid::BUTTON_LEFT))
    {
        pApp->m_saveDataSet.objectDegree -= 7.5f;
        // Keep angle's absolute value within the range of -360 to 360
        pApp->m_saveDataSet.objectDegree = nn::math::FMod(pApp->m_saveDataSet.objectDegree, 360);
    }
    else if(m_ExtraPadStatus.hold & (nn::hid::BUTTON_RIGHT))
    {
        pApp->m_saveDataSet.objectDegree += 7.5f;
        // Keep angle's absolute value within the range of -360 to 360
        pApp->m_saveDataSet.objectDegree = nn::math::FMod(pApp->m_saveDataSet.objectDegree, 360);
    }

}


// Render the scene.
void SceneUseExtraPad::Draw()
{
    // Application
    Application* pApp = Application::GetInstance();
    // Renderer
    demo::RenderSystemDrawing* pRenderSystem = pApp->GetRenderSystem();

    // Text writer
    scene::TextWriter* pTextWriter = pApp->GetTextWriter();

    f32 textWidth;       // Width of one character
    f32 textHeight;      // Character height
    u32 lineTop = 100;   // Y coordinate of first line
    u32 lineHeight;      // Space between lines
    f32 defaultFontScale = pTextWriter->GetFontScale(); // Default font scale

    //// Upper screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY0);
    pRenderSystem->Clear();

    pTextWriter->BeginRender(NN_GX_DISPLAY0);
    {
        scene::DrawTitle(pTextWriter);

        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(2, 20, L"[Scene] Use extra pad");

        // Display the various statuses
        pTextWriter->SetFontScale(defaultFontScale*0.8f);
        lineHeight = pTextWriter->CalculateStringHeight(L"");

        // Save data to use
        // Name of selected save data
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Selected save data");
        pTextWriter->Printf(200-textWidth, lineTop, L"Selected save data");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": ");
        textWidth = pTextWriter->CalculateStringWidth(L": ");
        pTextWriter->Printf(200+textWidth, lineTop, SaveData::SAVEDATA_FILENAME[pApp->m_SelectedSaveData]);
        lineTop += lineHeight;

        // Current total play time
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Total play time");
        pTextWriter->Printf(200-textWidth, lineTop, L"Total play time");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": ");
        textWidth = pTextWriter->CalculateStringWidth(L": ");
        // Convert ticks to play time display (hours:minutes:seconds)
        s64 playTimeSeconds = (pApp->m_saveDataSet.playTimeTick + nn::os::Tick::GetSystemCurrent() - pApp->m_saveData.startTick).ToTimeSpan().GetSeconds();
        s64 playTimeHours   = playTimeSeconds / 3600;
        playTimeSeconds -= playTimeHours * 3600;
        s64 playTimeMinutes = playTimeSeconds / 60;
        playTimeSeconds -= playTimeMinutes * 60;
        pTextWriter->Printf(200+textWidth, lineTop, L"%4lld:%02lld:%02lld",playTimeHours, playTimeMinutes, playTimeSeconds);
        lineTop += lineHeight;

        // Circle Pad Pro use settings
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Extra pad setting");
        pTextWriter->Printf(200-textWidth, lineTop, L"Extra pad setting");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": %s",
            pApp->m_saveDataSet.isUseExtraPad ? "On" : "Off");
        lineTop += lineHeight;

        lineTop += lineHeight; // Internal parameter boundary

        // Circle Pad Pro connection state
        if(nn::hid::ExtraPad::GetConnectionState() == nn::hid::ExtraPad::CONNECTION_STATE_CONNECTED)
        {
            pTextWriter->SetTextColor(127, 127, 127);
            textWidth = pTextWriter->CalculateStringWidth(L"Extra pad connection state");
            pTextWriter->Printf(200-textWidth, lineTop, L"Extra pad connection state");
            pTextWriter->SetTextColor(255, 255, 255);
            pTextWriter->Printf(200, lineTop, L": Connected");
            lineTop += lineHeight;

            pTextWriter->SetTextColor(127, 127, 127);
            textWidth = pTextWriter->CalculateStringWidth(L"Battery level");
            pTextWriter->Printf(200-textWidth, lineTop, L"Battery level");
            pTextWriter->SetTextColor(255, 255, 255);
            pTextWriter->Printf(200, lineTop, L": %s",
                m_ExtraPadStatus.batteryLevel ? "Normal" : "Low");
        }
        else
        {
            pTextWriter->SetTextColor(127, 127, 127);
            textWidth = pTextWriter->CalculateStringWidth(L"Extra pad connection state");
            pTextWriter->Printf(200-textWidth, lineTop, L"Extra pad connection state");
            pTextWriter->SetTextColor(255, 255, 255);
            pTextWriter->Printf(200, lineTop, L": Not connected");
            lineTop += lineHeight;
        }
        lineTop += lineHeight;


        lineTop += lineHeight; // Boundary with processing determination flag

        // Display presence of thread
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Detect extra pad thread flag");
        pTextWriter->Printf(200-textWidth, lineTop, L"Detect extra pad thread flag");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": %s",
            IsAliveDetectExtraPadThread() ? "true" : "false");
        lineTop += lineHeight;

        // First connection since starting application
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Extra pad first connection flag");
        pTextWriter->Printf(200-textWidth, lineTop, L"Extra pad first connection flag");
        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(200, lineTop, L": %s",
        pApp->m_IsFirstConnect ? "true" : "false");
        lineTop += lineHeight;
    }

    pTextWriter->SetFontScale(defaultFontScale); // Reset font scale
    pTextWriter->EndRender();
    pRenderSystem->SwapBuffers();




    //// Lower screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY1);
    pRenderSystem->Clear();

    // The object itself
    pRenderSystem->SetColor(255,255,255);
    pRenderSystem->SetPointSize(15);
    pRenderSystem->DrawPoint(pApp->m_saveDataSet.objectPositionX, pApp->m_saveDataSet.objectPositionY);

    // Line
    pRenderSystem->SetColor(255,0,0);
    pRenderSystem->SetLineWidth(5);
    pRenderSystem->DrawLine(pApp->m_saveDataSet.objectPositionX, pApp->m_saveDataSet.objectPositionY,
        pApp->m_saveDataSet.objectPositionX + pApp->m_saveDataSet.objectLineLength * nn::math::SinDeg(pApp->m_saveDataSet.objectDegree),
        pApp->m_saveDataSet.objectPositionY - pApp->m_saveDataSet.objectLineLength * nn::math::CosDeg(pApp->m_saveDataSet.objectDegree));

    pTextWriter->BeginRender(NN_GX_DISPLAY1);
    {
        // LR Buttons
        pTextWriter->SetTextColor(100, 100, 100);
        pTextWriter->Printf(10,25, L"L");
        textWidth = pTextWriter->CalculateStringWidth(L"R");
        pTextWriter->Printf(310-textWidth,25, L"R");

        pTextWriter->SetTextColor(255, 255, 255);
        if(m_ExtraPadStatus.hold  &  nn::hid::BUTTON_L){pTextWriter->Printf(10,25, L"L");}
        if(m_ExtraPadStatus.hold  &  nn::hid::BUTTON_R)
        {
            textWidth = pTextWriter->CalculateStringWidth(L"R");
            pTextWriter->Printf(310-textWidth,25, L"R");
        }

        if(m_ExtraPadStatus.isConnected) // When using Circle Pad
        {
            // ZL, ZR Buttons
            pTextWriter->SetTextColor(100, 100, 100);
            pTextWriter->Printf(5, 0, L"ZL");
            textWidth = pTextWriter->CalculateStringWidth(L"ZR");
            pTextWriter->Printf(315-textWidth, 0, L"ZR");

            pTextWriter->SetTextColor(255, 255, 255);
            if(m_ExtraPadStatus.hold  &  nn::hid::BUTTON_ZL){pTextWriter->Printf(5, 0, L"ZL");}
            if(m_ExtraPadStatus.hold  &  nn::hid::BUTTON_ZR)
            {
                textWidth = pTextWriter->CalculateStringWidth(L"ZR");
                pTextWriter->Printf(315-textWidth, 0, L"ZR");
            }
        }

        // START Button description
        pTextWriter->SetTextColor(127, 127, 127);
        textWidth = pTextWriter->CalculateStringWidth(L"Press START to open OPTIONS");
        textHeight = pTextWriter->CalculateStringHeight(L"Press START to open OPTIONS");
        pTextWriter->Printf((320 - textWidth)/2, 240-textHeight, L"Press START to open OPTIONS");


        //// State change message
        pTextWriter->SetFontScale(defaultFontScale);
        // Initialization
        if(!m_MessageProperty.isInitialized)
        {
            // Configure display start Y coordinate from character height
            m_MessageProperty.textHeight = pTextWriter->CalculateStringHeight(L"");
            m_MessageProperty.isInitialized = true;
        }

        switch(m_MessageType)
        {
        case MESSAGE_DISCONNECT:
            pTextWriter->SetTextColor(255, 0, 0); // Red
            textWidth = pTextWriter->CalculateStringWidth(L"Circle Pad Pro removed.");
            pTextWriter->Printf((320 - textWidth)/2, m_MessageProperty.displayY, L"Circle Pad Pro removed.");
            break;

        case MESSAGE_CONNECT:
            pTextWriter->SetTextColor(0, 255, 0); // Green
            textWidth = pTextWriter->CalculateStringWidth(L"Circle Pad Pro connected.");
            pTextWriter->Printf((320 - textWidth)/2, m_MessageProperty.displayY, L"Circle Pad Pro connected.");
            break;

        case MESSAGE_LOW_BATTERY:
            pTextWriter->SetTextColor(255, 255, 0); // Yellow
            textWidth = pTextWriter->CalculateStringWidth(L"The Circle Pad Pro battery is low.");
            pTextWriter->Printf((320 - textWidth)/2, m_MessageProperty.displayY, L"The Circle Pad Pro battery is low.");
            textWidth = pTextWriter->CalculateStringWidth(L"Replace the battery.");
            pTextWriter->Printf((320 - textWidth)/2, m_MessageProperty.displayY + m_MessageProperty.textHeight, L"Replace the battery.");
            break;
        }

        pTextWriter->SetFontScale(defaultFontScale);

        // Update control manager
        m_ControlManager.Draw();
    }
    pTextWriter->EndRender();

    pRenderSystem->SwapBuffers();
    pRenderSystem->WaitVsync(NN_GX_DISPLAY_BOTH);
}

}}}}
