﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     extra_pad.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/hid/CTR/hid_ExtraPadReader.h>
#include <nn/hid/CTR/hid_ExtraPadStatus.h>
#include <nn/extrapad.h> //Used to call the Nintendo 3DS Circle Pad Pro calibration applet
#include <nn/math.h>
#include <nn/applet.h>

#include "demo.h"
#include "applet.h"

//Displays input of +Control Pad emulation by the Circle Pad (R).
//#define DISPLAY_EXTRA_PAD_KEY_EMULATION

namespace 
{
    u8 extraPadWorkingMemory[4096] NN_ATTRIBUTE_ALIGN(4096);
    struct TextData
    {
        bit32 button;
        int x, y;
        char* text;
    };
}

void nnMain()
{
    NN_LOG("ExtraPad demo Start\n");

    const u32 GX_HEAP_SIZE = 0x800000;

    TransitionHandler::Initialize();
    if(TransitionHandler::IsExitRequired())
    {
        TransitionHandler::Finalize();
        nn::applet::CloseApplication();
        return;
    }


    // Initialize hid
    nn::hid::Initialize();
    nn::hid::CTR::ExtraPad::Initialize(extraPadWorkingMemory);

    // Class that gets information from the Circle Pad Pro and the game pad on the CTR system
    nn::hid::CTR::ExtraPadReader extraPadReader;
    nn::hid::CTR::ExtraPadStatus extraPadStatus;
    nn::hid::CTR::ExtraPadStatus extraPadStatusArray[32];
    s32 extraPadReadLen = 0;

    // Class that gets information from the game pad on the CTR system (PadReader does not return button information when the Circle Pad Pro is connected)
    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;
    nn::hid::PadStatus padStatusArray[32];
    s32 padReadLen = 0;


    // Initialization for graphics
    nn::fnd::ExpHeap appHeap(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    uptr heapForGx = reinterpret_cast<uptr>(appHeap.Allocate(GX_HEAP_SIZE));
    demo::RenderSystemDrawing renderSystem;
    renderSystem.Initialize(heapForGx, GX_HEAP_SIZE);


    // After this, the application itself handles sleep
    TransitionHandler::EnableSleep();

    while(1)
    {

        if(!extraPadReader.ReadLatest(&extraPadStatus))
        {
            nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(10));
            continue;
        }

        extraPadReader.Read(extraPadStatusArray,&extraPadReadLen,32);
        padReader.Read(padStatusArray,&padReadLen,32);
        padReader.ReadLatest(&padStatus);


        // CONNECTION_STATE_STOPPED means that the Circle Pad Pro has been disconnected for some external reason (this makes no distinction between physical disconnection, battery depletion, Sleep Mode, and other various reasons).
        if(nn::hid::CTR::ExtraPad::GetConnectionState() == nn::hid::CTR::ExtraPad::CONNECTION_STATE_STOPPED)
        {
            NN_LOG("ExtraPad - Disconnected .\n");
            //A call to StopSampling changes the internal state from STOPPED to NO_CONNECTION.
            nn::hid::CTR::ExtraPad::StopSampling();
        }

        //Switch between stopping and starting sampling
        if(extraPadStatus.trigger & nn::hid::BUTTON_A)
        {
            //Connect if sampling is not in progress, or disconnect and re-connect if sampling is in progress.
            nn::Result result = nn::hid::CTR::ExtraPad::StartSampling(0, 8);
            if(result.IsFailure())
            {
                NN_LOG("Connect Failure\n");
                NN_DBG_PRINT_RESULT(result);
                continue;
            }
            else
            {
                NN_LOG("ExtraPad - ON\n");
            }
        }
        else if(extraPadStatus.trigger & nn::hid::BUTTON_B)
        {
            nn::hid::CTR::ExtraPad::StopSampling();
            NN_LOG("ExtraPad - OFF\n");
        }
        
        
        //------------------------------------------------------
        // Jump to the Circle Pad Pro calibration applet.
        //------------------------------------------------------
        if ( extraPadStatus.trigger & nn::hid::BUTTON_START )
        {
            //Stop sampling
            nn::hid::CTR::ExtraPad::StopSampling();
            
            // ExtraPad settings structure
            nn::extrapad::Parameter extraPadParam;
            nn::applet::AppletWakeupState wstate;

            // Initialize settings struct
            nn::extrapad::InitializeConfig( &extraPadParam.config );

            // Start the applet
            nn::extrapad::StartExtraPadApplet( &wstate, &extraPadParam );

            // Recover the GPU register settings
            nngxUpdateState( NN_GX_STATE_ALL );
            nngxValidateState( NN_GX_STATE_ALL,GL_TRUE );

            NN_LOG( "Return Code : %d\n", extraPadParam.config.returnCode );
        }



        //Render the upper screen
        renderSystem.SetRenderTarget(NN_GX_DISPLAY0);
        renderSystem.Clear();

        renderSystem.SetColor(1.f,1.f,1.f);
        renderSystem.DrawText(0,0,"Extra Pad Demo");

        renderSystem.DrawText(0,20,"PadReader      : (x,y) = (%4d , %4d) Len = %d ",padStatus.stick.x,padStatus.stick.y,padReadLen);
        renderSystem.DrawText(0,30,"ExtraPadReader : (x,y) = (%4d , %4d) Len = %d",extraPadStatus.stick.x,extraPadStatus.stick.y,extraPadReadLen);
        renderSystem.DrawText(0,40,"                 (x,y) = (%4d , %4d) "        ,extraPadStatus.extraStick.x,extraPadStatus.extraStick.y);
        renderSystem.SetColor(0.5f,0.5f,0.5f);
        renderSystem.DrawText(0,50,"        Button : A B X Y L R ZL ZR");
        renderSystem.SetColor(1.f,1.f,1.f);
        renderSystem.DrawText(0,50,"        Button : %s %s %s %s %s %s %s %s",
                            extraPadStatus.hold  &  nn::hid::BUTTON_A ? "A" : " ",
                            extraPadStatus.hold  &  nn::hid::BUTTON_B ? "B" : " ",
                            extraPadStatus.hold  &  nn::hid::BUTTON_X ? "X" : " ",
                            extraPadStatus.hold  &  nn::hid::BUTTON_Y ? "Y" : " ",
                            extraPadStatus.hold  &  nn::hid::BUTTON_L ? "L" : " ",
                            extraPadStatus.hold  &  nn::hid::BUTTON_R ? "R" : " ",
                            extraPadStatus.hold  &  nn::hid::BUTTON_ZL ? "ZL" : "  ",
                            extraPadStatus.hold  &  nn::hid::BUTTON_ZR ? "ZR" : "  ");
        renderSystem.DrawText(0,60,"                 Connected = %s,  Battery = %d",extraPadStatus.isConnected?"ON ":"OFF",extraPadStatus.batteryLevel);

        renderSystem.DrawText(100,100,"Push [A] to Start Sampling");
        renderSystem.DrawText(100,110,"Push [B] to Stop Sampling");

        renderSystem.DrawText(0,130,"        Pad-stick     ExtPad-stick  ExtPad-extra");
        for(s32 i=0;i<extraPadReadLen;i++)
        {
            renderSystem.DrawText(0,140+i*10,
                "(x,y) = (%4d , %4d) (%4d , %4d) (%4d , %4d)",
                padStatusArray[i].stick.x,padStatusArray[i].stick.y,
                extraPadStatusArray[i].stick.x,extraPadStatusArray[i].stick.y,
                extraPadStatusArray[i].extraStick.x,extraPadStatusArray[i].extraStick.y
            );
        }

        renderSystem.SwapBuffers();

        //Render the lower screen
        renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
        renderSystem.Clear();

#ifdef DISPLAY_EXTRA_PAD_KEY_EMULATION
        TextData emulationButtons[] = {
            {nn::hid::BUTTON_EMULATION_LEFT, 10, nn::gx::DISPLAY1_WIDTH / 2 - 25, "LL"},
            {nn::hid::BUTTON_EMULATION_RIGHT, nn::gx::DISPLAY1_HEIGHT - 35, nn::gx::DISPLAY1_WIDTH / 2 - 25, "LR"},
            {nn::hid::BUTTON_EMULATION_UP, nn::gx::DISPLAY1_HEIGHT / 2 - 25, 10, "LU"},
            {nn::hid::BUTTON_EMULATION_DOWN, nn::gx::DISPLAY1_HEIGHT / 2 - 25, nn::gx::DISPLAY1_WIDTH - 10, "LD"}
        };
        TextData emulationRButtons[] = {
            {nn::hid::BUTTON_EMULATION_R_LEFT, 10, nn::gx::DISPLAY1_WIDTH / 2 + 5, "RL"},
            {nn::hid::BUTTON_EMULATION_R_RIGHT, nn::gx::DISPLAY1_HEIGHT - 35, nn::gx::DISPLAY1_WIDTH / 2 + 5, "RR"},
            {nn::hid::BUTTON_EMULATION_R_UP, nn::gx::DISPLAY1_HEIGHT / 2 + 5, 10, "RU"},
            {nn::hid::BUTTON_EMULATION_R_DOWN, nn::gx::DISPLAY1_HEIGHT / 2 + 5, nn::gx::DISPLAY1_WIDTH - 10, "RD"}
        };

        renderSystem.SetColor(0.5f,0.5f,0.5f);
        for(s32 i = 0; i < 4; i++)
        {
            TextData dat = emulationButtons[i];
            TextData datR = emulationRButtons[i];
            renderSystem.DrawText(dat.x, dat.y, " %s ", dat.text);
            renderSystem.DrawText(datR.x, datR.y, " %s ", datR.text);
        }
#endif
        // Block rendering
        renderSystem.SetColor(1.f,0.f,0.f);
        renderSystem.FillRectangle(
            160 + extraPadReader.NormalizeExtraStick(extraPadStatus.stick.x)*110 - 5,
            120 - extraPadReader.NormalizeExtraStick(extraPadStatus.stick.y)*110 - 5,
            10,
            10
        );

#ifdef DISPLAY_EXTRA_PAD_KEY_EMULATION
        for(s32 i = 0; i < 4; i++)
        {
            TextData dat = emulationButtons[i];
            renderSystem.DrawText(dat.x, dat.y, "%s%s%s",
                    extraPadStatus.trigger & dat.button ? "+" : " ",
                    extraPadStatus.hold & dat.button ? dat.text : "  ",
                    extraPadStatus.release & dat.button ? "-" : " ");
        }
#endif

        // Block rendering
        if(extraPadStatus.isConnected)
        {
            renderSystem.SetColor(0.f,0.f,1.f);
            renderSystem.FillRectangle(
                160 + extraPadReader.NormalizeExtraStick(extraPadStatus.extraStick.x)*110 - 5,
                120 - extraPadReader.NormalizeExtraStick(extraPadStatus.extraStick.y)*110 - 5,
                10,
                10
            );
        }

#ifdef DISPLAY_EXTRA_PAD_KEY_EMULATION
        for(s32 i = 0; i < 4; i++)
        {
            TextData dat = emulationRButtons[i];
            renderSystem.DrawText(dat.x, dat.y, "%s%s%s",
                    extraPadStatus.trigger & dat.button ? "+" : " ",
                    extraPadStatus.hold & dat.button ? dat.text : "  ",
                    extraPadStatus.release & dat.button ? "-" : " ");
        }
#endif

        renderSystem.SwapBuffers();

        renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

        // Run sleep, HOME Menu, and POWER Menu transitions.
        TransitionHandler::Process();
        // Determine whether there has been an exit notification.
        if (TransitionHandler::IsExitRequired())
        {
            break;
        }
    }

    renderSystem.Finalize();
    appHeap.Free(reinterpret_cast<void*>(heapForGx));
    appHeap.Finalize();

    //Stop processing when we exit
    nn::hid::CTR::ExtraPad::StopSampling();
    nn::hid::CTR::ExtraPad::Finalize();
    nn::hid::Finalize();

    TransitionHandler::Finalize();

    nn::applet::CloseApplication();

}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
