﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gyroscope.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/math.h>
#include <nn/applet.h>

#include "demo.h"
#include "gx.h"


//------------------------------------------------------------


void nnMain()
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    NN_LOG("Gyroscope demo Start\n");

    const u32 GX_HEAP_SIZE = 0x800000;

    // Initial posture
    const nn::hid::Direction dir(nn::math::MTX33::Identity());

    // Initialize hid
    nn::hid::Initialize();
    nn::hid::PadReader padReader;
    nn::hid::AccelerometerReader accReader;

#if 1
    // Specify AccelerometerReader to use for accelerometer correction
    nn::hid::GyroscopeReader gyroReader(&accReader);
#else
    // Makes AccelerometerReader the default for accelerometer correction
    nn::hid::GyroscopeReader gyroReader;
#endif
    // Prepare RenderSystem
    nn::fnd::ExpHeap appHeap(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    uptr heapForGx = reinterpret_cast<uptr>(appHeap.Allocate(GX_HEAP_SIZE));
    demo::RenderSystemDrawing renderSystem;
    renderSystem.Initialize(heapForGx, GX_HEAP_SIZE);

    // Prepare to render the upper screen
    GraphicsDemo gxDemo;
    gxDemo.Initialize(&renderSystem,&gyroReader,&accReader);
    gxDemo.Start();

    nn::hid::PadStatus padStatus;

    while(1)
    {
        padReader.ReadLatest(&padStatus);

        //Render the upper screen
        renderSystem.SetRenderTarget(NN_GX_DISPLAY0);
        renderSystem.Clear();

        gxDemo.DrawUpperFrame();

        renderSystem.SwapBuffers();

        //Render the lower screen
        renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
        renderSystem.Clear();

        gxDemo.DrawLowerFrame();
        renderSystem.DrawText(170,  0, "<Command>");
        renderSystem.DrawText(170, 10, " A: Reset All");
        renderSystem.DrawText(170, 20, " B: Reset Dir");
        renderSystem.DrawText(170, 30, " X: Acc");
        renderSystem.DrawText(170, 40, "    Revise %s", gyroReader.IsEnableAccRevise()?"On":"Off");
        renderSystem.DrawText(170, 50, " Y: ZeroDrift");
        renderSystem.DrawText(170, 60, "    Revise %s", gyroReader.IsEnableZeroDrift()?"On":"Off");
        renderSystem.DrawText(170, 70, " L: Acc Offset");
        renderSystem.DrawText(170, 80, "           %s", accReader.IsEnableOffset()?"On":"Off");
        renderSystem.DrawText(170, 90, " R: Axis Rotation");
        renderSystem.DrawText(170,100, "           %s", gyroReader.IsEnableAxisRotation()?"On":"Off");

        renderSystem.SwapBuffers();

        // Reset all settings
        if(padStatus.trigger == (nn::hid::BUTTON_A))
        {
            gyroReader.Reset();
            accReader.ResetAxisRotationMatrix();
            accReader.ResetOffset();
        }

        // Reset posture
        if(padStatus.trigger == (nn::hid::BUTTON_B))
        {
            gyroReader.SetDirection(dir);
        }

        // Toggles accelerometer correction on/off
        if(padStatus.trigger == (nn::hid::BUTTON_X))
        {
            if(gyroReader.IsEnableAccRevise())
            {
                gyroReader.DisableAccRevise();
            }
            else
            {
                gyroReader.EnableAccRevise();
            }
        }

        // Toggles zero-point drift on/off
        if(padStatus.trigger == (nn::hid::BUTTON_Y))
        {
            if(gyroReader.IsEnableZeroDrift())
            {
                gyroReader.DisableZeroDrift();
            }
            else
            {
                gyroReader.EnableZeroDrift();
            }
        }

        // Accelerometer offset setting
        if(padStatus.trigger == (nn::hid::BUTTON_L))
        {
            if(accReader.IsEnableOffset())
            {
                accReader.DisableOffset();
            }
            else
            {
                nn::hid::CTR::AccelerometerStatus baseStatus;

                // Turn off all change options in order to obtain standard gravity
                accReader.DisableOffset();
                accReader.DisableAxisRotation();
                gyroReader.DisableAxisRotation();

                if(accReader.ReadLatest(&baseStatus))
                {
                    // Offset setting so that baseStatus becomes the horizontal state
                    accReader.SetOffsetFromBaseStatus(baseStatus);
                }
                accReader.EnableOffset();
            }
        }

        // Axial rotation setting
        if(padStatus.trigger == (nn::hid::BUTTON_R))
        {
            if(gyroReader.IsEnableAxisRotation())
            {
                gyroReader.DisableAxisRotation();
                accReader.DisableAxisRotation();
            }else
            {
                // Set the rotational matrix to convert so it is as if the current accelerometer state is the horizontal state
                accReader.DisableAxisRotation();
                nn::hid::CTR::AccelerometerStatus accStatus;
                if(accReader.ReadLatest(&accStatus))
                {
                    // Rotational matrix to perform axial rotation by internal angle with the vector product of the following two axes: current gravitational vector indicated by the accelerometer and the ideal gravitational vector
                    nn::math::MTX34 rotMtx;

                    nn::math::VEC3 accVec(accStatus.x,accStatus.y,accStatus.z);
                    nn::math::VEC3 gVec(0,-1,0);
                    nn::math::VEC3 cross;
                    nn::math::VEC3Cross(&cross,accVec,gVec);
                    rotMtx.SetupRotate(cross,nn::math::AcosRad(accVec.Dot(gVec)/(accVec.Length() * gVec.Length())));

                    accReader.SetAxisRotationMatrix(rotMtx);
                    gyroReader.SetAxisRotationMatrix(rotMtx);
                    accReader.EnableAxisRotation();
                    gyroReader.EnableAxisRotation();
                    gyroReader.SetDirection(dir);
                }
            }
        }

        renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
