﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     hiodemo_Main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/assert.h>
#include <nn/hio.h>
#include <cstring>
#include <string.h>

void FileIoSample();
void SerialIoSample();
void SharedMemorySample();
void EnvironmentSample();

namespace 
{
    const s32 DEVICE_MEMORY_SIZE = 32 * 1024 * 1024;    // Device memory size to use

    const s32 SERIALIO_CH = 0;
    const s32 SHAREDMEMORY_CH = 1;
}

void FileIoSample()
{
    nn::Result result;

    //=======================================================================
    // Load and write file
    //=======================================================================
    nn::hio::CTR::HostFile file;

    const char sampleFileName[] = TEST_DIRECTORY "/sample.txt";
    const char sampleText[] = "Hello World!";
    const s32 bufferSize = 1024;
    char readBuffer[bufferSize];
    size_t writeSize = 0;
    size_t readSize = 0;

    NN_LOG("Create file '%s'.\n", sampleFileName);
    result = file.Open(sampleFileName,
        nn::hio::CTR::HostFile::ACCESS_MODE_READ_WRITE,
        nn::hio::HostFile::OPEN_DISP_CREATE_ALWAYS);
    if ( result.IsFailure() )
    {
        NN_LOG(" -> Failed to create file.\n");
    }
    else{

        NN_LOG("Write to file.\n");
        NN_LOG("data: '%s'\n", sampleText);
        result = file.Write(&writeSize, sampleText, strlen(sampleText)+1);
        if ( result.IsFailure() )
        {
            NN_LOG(" -> Failed to write file.\n");
        }
        NN_LOG("-> %d bytes written.\n", writeSize);

        NN_LOG("Seek to head.\n");
        s64 position;
        result = file.Seek(&position, 0, nn::hio::CTR::HostFile::SEEK_TYPE_SET);
        if ( result.IsFailure() )
        {
            NN_LOG(" -> Failed to seek.\n");
        }
        NN_LOG(" -> seek to 0x%016llx\n", position);

        NN_LOG("Read file.\n");
        result = file.Read(&readSize, readBuffer, bufferSize);
        if ( result.IsFailure() )
        {
            NN_LOG(" -> Failed to read.\n");
        }
        NN_LOG(" -> %d bytes read.\n", readSize);
        NN_LOG(" -> data: '%s'\n", readBuffer);

        file.Close();
    }

    //=======================================================================
    // Change file name
    //=======================================================================
    const char renameFileName[] = TEST_DIRECTORY "/sample-rename.txt";

    NN_LOG("File rename\n    to: '%s'\n  from: '%s'\n", renameFileName, sampleFileName);
    result = nn::hio::CTR::Rename(renameFileName, sampleFileName);
    if ( result.IsFailure() )
    {
        NN_LOG("Fail to rename\n");
    }

    //=======================================================================
    // Delete file
    //=======================================================================
    NN_LOG("File delete ''\n", renameFileName);
    result = nn::hio::CTR::DeleteFile(renameFileName);
    if ( result.IsFailure() )
    {
        NN_LOG("Fail to delete file.\n");
    }

    //=======================================================================
    // Create directory
    //=======================================================================
#ifndef NN_SWITCH_DISABLE_DEBUG_PRINT
    const char sampleDirectoryName[] = TEST_DIRECTORY "/sampledirectory";
#endif
    const wchar_t sampleDirectoryNameW[] = TEST_DIRECTORY_W L"/sampledirectory";

    NN_LOG("Directory create '%s'\n", sampleDirectoryName);
    result = nn::hio::CTR::CreateDirectory(sampleDirectoryNameW);
    if ( result.IsFailure() )
    {
        NN_LOG("Fail to create directory.\n");
    }

    //=======================================================================
    // Delete directory
    //=======================================================================
    NN_LOG("Directory delete '%s'\n", sampleDirectoryName);
    result = nn::hio::CTR::DeleteDirectory(sampleDirectoryNameW);
    if ( result.IsFailure() )
    {
        NN_LOG("Fail to delete directory.\n");
    }

    //=======================================================================
    // Entry enumeration for directory
    //=======================================================================
    nn::hio::CTR::HostDirectoryA directory;

    NN_LOG("Open directory '%s'\n", TEST_DIRECTORY);
    result = directory.Open(TEST_DIRECTORY);
    if ( result.IsFailure() )
    {
        NN_LOG("Fail to open directory.\n");
    }
    else
    {
        NN_LOG("Directory Entries\n");
        nn::hio::CTR::DirectoryEntryA directoryEntry;

        result = directory.GetNextEntry(&directoryEntry);

        while ( result.IsSuccess() )
        {
            NN_LOG("  filename: '%s'\n", directoryEntry.filename);
            NN_LOG("    size:        %lld bytes\n", directoryEntry.fileSize);
            NN_LOG("    attributes:  %08x\n", directoryEntry.fileAttributes);
            NN_LOG("    create:      %04d-%02d-%02d(%d) %02d:%02d:%02d.%03d\n",
                directoryEntry.creationTime.year,
                directoryEntry.creationTime.month,
                directoryEntry.creationTime.day,
                directoryEntry.creationTime.dayofweek,
                directoryEntry.creationTime.hour,
                directoryEntry.creationTime.minute,
                directoryEntry.creationTime.second,
                directoryEntry.creationTime.msec);
            NN_LOG("    last access: %04d-%02d-%02d(%d) %02d:%02d:%02d.%03d\n",
                directoryEntry.lastAccessTime.year,
                directoryEntry.lastAccessTime.month,
                directoryEntry.lastAccessTime.day,
                directoryEntry.lastAccessTime.dayofweek,
                directoryEntry.lastAccessTime.hour,
                directoryEntry.lastAccessTime.minute,
                directoryEntry.lastAccessTime.second,
                directoryEntry.lastAccessTime.msec);
            NN_LOG("    last write:  %04d-%02d-%02d(%d) %02d:%02d:%02d.%03d\n",
                directoryEntry.lastWriteTime.year,
                directoryEntry.lastWriteTime.month,
                directoryEntry.lastWriteTime.day,
                directoryEntry.lastWriteTime.dayofweek,
                directoryEntry.lastWriteTime.hour,
                directoryEntry.lastWriteTime.minute,
                directoryEntry.lastWriteTime.second,
                directoryEntry.lastWriteTime.msec);

            result = directory.GetNextEntry(&directoryEntry);
        }
    }

}

void EnvironmentSample()
{
    nn::Result result;

    const size_t  BUFFER_SIZE = 1024;
    char buffer[BUFFER_SIZE];
    const char NAME[] = "CTRSDK_ROOT";

    //=======================================================================
    // Get environment variables
    //=======================================================================
    NN_LOG("GetEnvironmentVariable\n");
    nn::hio::CTR::GetEnvironmentVariable(buffer, sizeof(buffer), NAME);
    NN_LOG("env %s = %s", NAME, buffer);

    //=======================================================================
    // Get current directory
    //=======================================================================
    NN_LOG("GetCurrentDirectory\n");
    nn::hio::CTR::GetCurrentDirectory(buffer,sizeof(buffer));
    NN_LOG("current directory: %s\n",buffer);

    //=======================================================================
    // Current directory settings
    //=======================================================================
    NN_LOG("SetCurrentDirectory\n");
    nn::hio::CTR::SetCurrentDirectory(TEST_DIRECTORY);
    nn::hio::CTR::GetCurrentDirectory(buffer,sizeof(buffer));
    NN_LOG("current directory: %s\n",buffer);
}

void SerialIoSample()
{
    nn::Result result;
    char buffer[1024];
    const char sampleString[] = "Hello!!";
    const size_t sampleLength = sizeof(sampleString);

    //=======================================================================
    // Connect to the PC-side program and send and receive data one at time.
    //=======================================================================

    NN_LOG("start.\n");
    nn::hio::CTR::SerialChannel serial;

    // Open channel.
    result = serial.Open(SERIALIO_CH,
                         reinterpret_cast<void*>(
                             nn::os::GetDeviceMemoryAddress()
                             + nn::hio::CTR::WORKMEMORY_SIZE * (SERIALIO_CH + 1)));
    if ( result.IsFailure() )
    {
        NN_LOG("serial.Open() ERROR(code:%08x)\n",result.GetPrintableBits());
        serial.Close();
        return;
    }
    NN_LOG("open.\n");

    // Connect to PC-side program.
    // Wait until connected.
    result = serial.Connect();
    if ( result.IsFailure() )
    {
        NN_LOG("serial.Connect() ERROR(code:%08x)\n",result.GetPrintableBits());
        serial.Close();        
        return;
    }
    NN_LOG("connect.\n");

    // Receive data from the PC.
    size_t readsize;
    result = serial.Read(&readsize, buffer, sampleLength);
    if ( result.IsFailure() )
    {
        NN_LOG("serial.Read() ERROR(code:%08x)\n",result.GetPrintableBits());
        serial.Disconnect();
        serial.Close();        
        return;
    }
    NN_LOG("read size: %d\n", readsize);
    NN_LOG("read data: [%s]\n", buffer);

    // Send data to the PC-side program.
    size_t writesize;
    result = serial.Write(&writesize, sampleString, sampleLength);
    if ( result.IsFailure() )
    {
        NN_LOG("serial.Write() ERROR(code:%08x)\n",result.GetPrintableBits());
        serial.Disconnect();
        serial.Close();        
        return;
    }

    // Wait until PC-side program receives the data.
    result = serial.Flush();
    if ( result.IsFailure() )
    {
        NN_LOG("serial.Flush() ERROR(code:%08x)\n",result.GetPrintableBits());
        serial.Disconnect();
        serial.Close();        
        return;
    }
    NN_LOG("write\n");
    NN_LOG("write size: %d\n", writesize);

    // Disconnect from PC.
    result = serial.Disconnect();
    if ( result.IsFailure() )
    {
        NN_LOG("serial.Disconnect() ERROR(code:%08x)\n",result.GetPrintableBits());
        serial.Close();
        return;
    }
    NN_LOG("disconnect.\n");

    // Close the channel.
    result = serial.Close();
    if ( result.IsFailure() )
    {
        NN_LOG("serial.Close() ERROR(code:%08x)\n",result.GetPrintableBits());
        return;
    }
    NN_LOG("close.\n");
}

void SharedMemorySample()
{
    nn::Result result;
    const int BUFFER_SIZE = 1024;
    char buffer[BUFFER_SIZE];
    const char sendString[] = "Ping!!";
    const size_t sendLength = sizeof(sendString);
    const char receiveString[] = "Pong!!";
    const size_t receiveLength = sizeof(receiveString);

    for ( int i = 0; i < sizeof(buffer); i++ )
    {
        buffer[i] = 0;
    }

    //=======================================================================
    // Send and receive data one at time to the PC-side program through the shared memory.
    //=======================================================================

    // Viewed from the CTR-side, the communication order is send, receive, send.
    // The operation on the shared memory can no longer be done if the PC side closes the shared memory channel, so communications are ended at the stage where the PC side confirms reception.
    // 

    nn::hio::CTR::SharedMemoryChannel sharedMemoryChannel;

    result = sharedMemoryChannel.Open(SHAREDMEMORY_CH,
                         reinterpret_cast<void*>(
                             nn::os::GetDeviceMemoryAddress()
                             + nn::hio::CTR::WORKMEMORY_SIZE * (SHAREDMEMORY_CH + 1)));
    if ( result.IsFailure() )
    {
        NN_LOG("sharedMemoryChannel.Open() ERROR(code:%08x)\n",result.GetPrintableBits());
        return;
    }
    
#ifndef NN_SWITCH_DISABLE_DEBUG_PRINT
    size_t size = sharedMemoryChannel.GetSpaceSize();
#endif
    NN_LOG("shared memory size: %d\n", size);

    sharedMemoryChannel.Write(sendString, sendLength, 0);
    if ( result.IsFailure() )
    {
        NN_LOG("sharedMemoryChannel.Write() ERROR(code:%08x)\n",result.GetPrintableBits());
        sharedMemoryChannel.Close();
        return;
    }
    NN_LOG("write data: '%s'\n", sendString);
    
    while( strcmp(receiveString, buffer) )
    {
        result = sharedMemoryChannel.Read(buffer, receiveLength, 0);
        if ( result.IsFailure() )
        {
            NN_LOG("sharedMemoryChannel.Read() ERROR(code:%08x)\n",result.GetPrintableBits());
            sharedMemoryChannel.Close();
            return;
        }
    }
    NN_LOG("read data: '%s'\n", buffer);

    sharedMemoryChannel.Write(sendString, sendLength, 0);
    if ( result.IsFailure() )
    {
        NN_LOG("sharedMemoryChannel.Write() ERROR(code:%08x)\n",result.GetPrintableBits());
        sharedMemoryChannel.Close();
        return;
    }
    NN_LOG("write data: '%s'\n", sendString);

    sharedMemoryChannel.Close();
    if ( result.IsFailure() )
    {
        NN_LOG("sharedMemoryChannel.Close() ERROR(code:%08x)\n",result.GetPrintableBits());
        return;
    }
}

extern "C" void nnMain(void)
{
    nn::Result result;

    // Initialize 32 MB contiguous-memory
    // The memory passed to HostIO must be continuous memory, so DeviceMemory is used.
    // 
    NN_LOG("SetDeviceMemorySize.\n", __LINE__);
    nn::os::SetDeviceMemorySize( DEVICE_MEMORY_SIZE );

    // Gets the start of contiguous memory
    // Use this as the HostIO buffer
    NN_LOG("GetDeviceMemoryAddress.\n", __LINE__);
    uptr deviceMemory = nn::os::GetDeviceMemoryAddress();

    // Initialize HostIO
    // Pass the device memory address as the HostIO buffer.
    NN_LOG("HostIO Library Initialize.\n", __LINE__);
    nn::hio::CTR::Initialize(reinterpret_cast<void*>(deviceMemory));

    // HostIO library version display
    // The PC side library must either be the same version as the CTR side library or a newer version library than the CTR side library version.
    // 
    NN_LOG("HostIO Library Version: %d\n", nn::hio::CTR::GetVersion());

    // Current directory and environment variable sample
    NN_LOG("===== Environment Variable Sample =====\n");
    EnvironmentSample();

    // File input/output sample
    NN_LOG("===== File IO Sample =====\n");
    FileIoSample();

    // Serial Communication Sample
    NN_LOG("===== Serial IO Sample =====\n");
    SerialIoSample();

    // Shared memory sample
    NN_LOG("===== Shared Memory Sample =====\n");
    SharedMemorySample();

    nn::hio::CTR::Finalize();
    NN_LOG("HostIO Sample Finished.\n");

    while(1)    {}
}

