/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     HioDemoPCCpp.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <windows.h>
#include <devhio_host.h>
#include <devhio_sio.h>
#include <devhio_smem.h>
#include <CDevHIO.h>
#include <stdio.h>
#include <string.h>

namespace {
    const int SERIALIO_CH = 0;
    const int SHAREDMEMORY_CH = 1;
    ::CDEVHIO s_HostIo;
}

void SerialIoSample();
void SharedMemorySample();

void SerialIoSample()
{
    int ret;
    char buffer[1024];
    const char sampleString[] = "Hello!!";
    const size_t sampleLength = sizeof(sampleString);

    //=======================================================================
    // Connect to the CTR-side program and send and receive data one at time.
    //=======================================================================

    // Open channel.
    ret = s_HostIo.OpenSio(SERIALIO_CH);
    if ( ret < 0 )
    {
        printf("OpenSio() ERROR(code:%08x)\n", ret);
        return;
    }
    printf("open\n");

    // Wait for connection to the CTR-side program.
    ret = s_HostIo.WaitSio(SERIALIO_CH, DEVHIO_SIO_ATTR_NONE);
    if ( ret != 1 )
    {
        printf("WaitSio() ERROR(code:%08x)\n", ret);
        s_HostIo.CloseSio(SERIALIO_CH);
        return;
    }

    // Send data to the CTR-side program.
    int writesize = s_HostIo.WriteSio(SERIALIO_CH, sampleString, sampleLength, DEVHIO_SIO_ATTR_NONE);
    if ( writesize < 0 )
    {
        printf("WriteSio() ERROR(code:%08x)\n", writesize);
        s_HostIo.DisconnectSio(SERIALIO_CH);
        s_HostIo.CloseSio(SERIALIO_CH);
        return;
    }
    printf("write\n");
    printf("write size: %d\n", writesize);
    ret = s_HostIo.FlushSio(SERIALIO_CH);
    if ( ret < 0 )
    {
        printf("FlushSio() ERROR(code:%08x)\n", ret);
        s_HostIo.CloseSio(SERIALIO_CH);
        return;
    }
    printf("flush\n");

    // Receives data from the CTR-side.
    int readsize = s_HostIo.ReadSio(SERIALIO_CH, buffer, sampleLength, DEVHIO_SIO_ATTR_NONE);
    if ( readsize < 0 )
    {
        printf("ReadSio() ERROR(code:%08x)\n", readsize);
        s_HostIo.DisconnectSio(SERIALIO_CH);
        s_HostIo.CloseSio(SERIALIO_CH);
        return;
    }
    printf("read size: %d\n", readsize);
    printf("read data: '%s'\n", buffer);

    // Disconnect, and close the channel.
    ret = s_HostIo.DisconnectSio(SERIALIO_CH);
    if ( ret < 0 )
    {
        printf("DisconnectSio() ERROR(code:%08x)\n", ret);
        s_HostIo.CloseSio(SERIALIO_CH);
        return;
    }
    printf("disconnect.\n");
    ret = s_HostIo.CloseSio(SERIALIO_CH);
    if ( ret < 0 )
    {
        printf("CloseSio() ERROR(code:%08x)\n", ret);
        return;
    }
    printf("close.\n");
}

void SharedMemorySample()
{
    int ret;
    const int BUFFER_SIZE = 1024;
    char buffer[BUFFER_SIZE];
    const char receiveString[] = "Ping!!";
    const size_t receiveLength = sizeof(receiveString);
    const char sendString[] = "Pong!!";
    const size_t sendLength = sizeof(sendString);

    for ( int i = 0; i < sizeof(buffer); i++ )
    {
        buffer[i] = 0;
    }

    //=======================================================================
    // Send and receive data to the CTR-side program through shared memory.
    //=======================================================================

    // Viewed from the PC-side, the communication process order is receive, send, receive.
    // Because operations to shared memory cannot be done if PC-side closes the shared memory channel, the communications are terminated at the stage when it is confirmed that the PC-side has received the data.
    // 

    ret = s_HostIo.OpenSmem(SHAREDMEMORY_CH);
    if ( ret < 0 )
    {
        printf("OpenSmem() ERROR(code:%08x)\n", ret);
        return;
    }

    int size = s_HostIo.GetSmemSize(SHAREDMEMORY_CH);
    printf("shared memory size: %d\n", size);

    while( strcmp(buffer, receiveString) ){
        ret = s_HostIo.ReadSmem(SHAREDMEMORY_CH, 0, buffer, receiveLength);
        if ( ret < 0 )
        {
            printf("ReadSmem() ERROR(code:%08x)\n", ret);
            s_HostIo.CloseSmem(SHAREDMEMORY_CH);
            return;
        }
    }
    printf("read data: '%s'\n", buffer);

    ret = s_HostIo.WriteSmem(SHAREDMEMORY_CH, 0, sendString, sendLength);
    s_HostIo.ReadSmem(SHAREDMEMORY_CH, 0, buffer, receiveLength);
    printf("read data: '%s'\n", buffer);

    if ( ret < 0 )
    {
        printf("WriteSmem() ERROR(code:%08x)\n", ret);
        s_HostIo.CloseSmem(SHAREDMEMORY_CH);
        return;
    }
    printf("write data: '%s'\n", sendString);

    while( strcmp(buffer, receiveString) ){
        ret = s_HostIo.ReadSmem(SHAREDMEMORY_CH, 0, buffer, receiveLength);
        if ( ret < 0 )
        {
            printf("ReadSmem() ERROR(code:%08x)\n", ret);
            s_HostIo.CloseSmem(SHAREDMEMORY_CH);
            return;
        }
    }
    printf("read data: '%s'\n", buffer);

    ret = s_HostIo.CloseSmem(SHAREDMEMORY_CH);
    if ( ret < 0 )
    {
        printf("CloseSmem() ERROR(code:%08x)\n", ret);
        return;
    }
}

int main()
{
    unsigned long id;

    // HostIO library version display
    // The PC-side library must either be the same version as the CTR-side library or a newer version than the CTR-side library.
    // 
    printf("HostIO Library Version: %d\n", s_HostIo.GetVersion());

    // Search for and connect to debugger
    printf("START\n");
    if (s_HostIo.Find(&id, 0) <= 0) {
	    printf("NO DEVICES FOUND\n");
	    return -1;
    }
    if (s_HostIo.Connect(id) < 0) {
        printf("CONNECT ERROR\n");
        return -1;
    }

    // Serial Communication Sample
    printf("===== Serial IO Sample =====\n");
    SerialIoSample();

    // Shared memory sample
    printf("===== Shared Memory Sample =====\n");
    SharedMemorySample();

    printf("END\n");
    s_HostIo.Disconnect();

    return 0;
}
