/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     communicator.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>

#include <string.h> //for memcpy

#include "demo.h"
#include "ir_util.h"

namespace 
{
    struct InputData{
        s16 akX;
        s16 akY;
        s16 tpX;
        s16 tpY;
    };

    //Application's unique ID
    const bit32 UNIQUE_ID = 0xFF000;
    //Arbitrarily specified ID matched to the communication scene
    const bit8 SUB_ID = 0x01;
    //Passphrase. For security, general terms should be avoided
    const char PASSPHRASE[] = "this is sample passphrase.";

    // All events
    nn::os::Event s_ReceiveEvent;
    nn::os::Event s_SendEvent;
    nn::os::Event s_ConnectEvent;

    // System sleep
    nn::os::LightEvent s_AwakeEvent;

    char* s_ConnectionRoleText[] =
    {
        "None",
        "Require",
        "Wait"
    };
}

/*
 * Sleep support
 */
// Callback to a Sleep request
nn::applet::AppletQueryReply mySleepQueryCallback(uptr arg)
{
    NN_UNUSED_VAR(arg);

    NN_LOG("Sleep Request.\n");
    s_AwakeEvent.ClearSignal();

    return nn::applet::CTR::IsActive() ? nn::applet::CTR::REPLY_LATER: nn::applet::CTR::REPLY_ACCEPT;
}

// Callback called when waking from Sleep
void myAwakeCallback(uptr arg)
{
    NN_UNUSED_VAR(arg);
    NN_LOG("Awake.\n");
    s_AwakeEvent.Signal();
}

// This is the ID authentication process after connection
nn::Result ConfirmID()
{
    nn::Result result;
    
    // Get whether this is the requesting side or the responding side during connection.
    nn::ir::ConnectionRole role = nn::ir::Communicator::GetConnectionRole();

    // Generate communication ID.
    // When creating a trial version, set the isDemo parameter to true to prevent communicating with the production version.
#ifndef NN_IR_COMMUNICATOR_DEMO_MODE
    bit32 communicationId = nn::ir::Communicator::CreateCommunicationId(UNIQUE_ID);
#else
    bit32 communicationId = nn::ir::Communicator::CreateCommunicationId(UNIQUE_ID, true);
#endif

    switch(role)
    {
    case nn::ir::CONNECTION_ROLE_TO_REQUIRE:
        {
            // For the request side, RequireToConfirmID is executed here.
            NN_LOG("Require Confirm id..\n");

            // A wait is inserted because RequireToConfirmID must be executed after ReplyToConfirm.
            // 
            nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(500));
            result = nn::ir::Communicator::RequireToConfirmId(SUB_ID, communicationId, PASSPHRASE, sizeof(PASSPHRASE));
        }
        break;
    case nn::ir::CONNECTION_ROLE_TO_WAIT:
        {
            // For the responding side, WaitToConfirmID is executed here.
            NN_LOG("Wait Confirm id..\n");

            // Waits (1000 msec) until a RequireToConfirmId is received from the partner.
            result = nn::ir::Communicator::WaitToConfirmId(SUB_ID, communicationId, PASSPHRASE, sizeof(PASSPHRASE),nn::fnd::TimeSpan::FromMilliSeconds(1000));
        }
        break;
    default:
        // Disconnected by the partner, so becomes ROLE_NONE
        break;
    }

    return result;
}

// Communication process after authentication
void Communicate(nn::hid::PadStatus* padStatus,nn::hid::TouchPanelStatus* tpStatus, nn::fnd::ExpHeap* appHeap)
{
    //Get whether this is the requesting side or the responding side during connection.
    nn::ir::ConnectionRole role = nn::ir::Communicator::GetConnectionRole();

    //Data format to use for transmissions
    InputData data;

    nn::Result result = nn::ResultSuccess();
    s32 usedSize = -1;
    size_t receiveSize = 0;

    //Create buffer for sending/receiving
    size_t bufferSize = nn::ir::Communicator::CalculateBufferSizeToCommunicate(sizeof(InputData));

    //Must be 4-byte aligned
    bit8* buffer = reinterpret_cast<bit8*>(appHeap->Allocate(bufferSize,4));
    ::memset(buffer,0,bufferSize);
    
    //If this is the requesting side during connection, send the Circle Pad and touch panel coordinates
    //If this is the responding side during connection, receive and overwrite
    switch(role)
    {
    case nn::ir::CONNECTION_ROLE_TO_REQUIRE:
        //Confirm remaining send data; if there is none remaining, send
        nn::ir::Communicator::GetSendSizeFreeAndUsed(NULL,NULL,NULL,&usedSize);
        if(usedSize == 0)
        {
            //Create send data
            data.akX = padStatus->stick.x;
            data.akY = padStatus->stick.y;
            data.tpX = tpStatus->touch?tpStatus->x:-1;
            data.tpY = tpStatus->touch?tpStatus->y:-1;

            ::memcpy(buffer,reinterpret_cast<void*>(&data),sizeof(InputData));
            
            s_SendEvent.ClearSignal();
            result = nn::ir::Communicator::Send(buffer,sizeof(InputData),bufferSize,false);
            if(result.IsFailure())
            {
                if(result == nn::ir::ResultNotConnected())
                {
                    // It is possible that the connection was disconnected by a disconnection request from the partner
                }

                NN_LOG("send failed\n");
            }

            // Send is performed asynchronously. When it is necessary to notify that sending has completed, use a send complete event.
            // 
        }
        break;
    case nn::ir::CONNECTION_ROLE_TO_WAIT:
        //Performs receiving processes while confirming remaining receive data. Receiving is repeated while there is some data remaining.

        // First, checks if data is being received
        if(!s_ReceiveEvent.Wait(0)){ break; }

        // Process received data
        do
        {
            s_ReceiveEvent.ClearSignal();
            result = nn::ir::Communicator::Receive(buffer,bufferSize,&receiveSize,&usedSize);
            if(result.IsFailure())
            {
                receiveSize = 0; 

                if(result == nn::ir::ResultInvalidData())
                { 
                    // Data was received, but the result of the verification was that the data was invalid.
                    // Because the data is removed from the receive buffer at this time, continue to process subsequent packets (if they exist).
                    //
                }
                else if(result == nn::ir::ResultBufferInsufficient())
                { 
                    // The buffer size is insufficient.
                    // At this time, it cannot be determined if the data is valid because the data verification process cannot be executed.
                    // 

                    // To receive this data, confirm the receive data size with the GetNextReceiveDataSize function, and prepare a buffer of the required size.
                    // 

                    // With this demo, it is probable that the received data is invalid since the required buffer to receive the expected data is prepared.
                    // 

                    // Data remains in the receive buffer.
                    // Since this data is unnecessary, destroy it and continue to process subsequent packets.
                    nn::ir::Communicator::DropNextReceiveData(&usedSize);
                }
                else
                {
                    // Other
                    // Exit the loop
                    break;
                }
            }

        }while(usedSize > 0);

        //When data was received, overwrite the Circle Pad and touch panel coordinates with that data.
        if(receiveSize == sizeof(InputData))
        {
            ::std::memcpy(reinterpret_cast<void*>(&data),buffer,receiveSize);
            
            tpStatus->touch = (data.tpX >= 0);
            if(tpStatus->touch){
                tpStatus->x = data.tpX;
                tpStatus->y = data.tpY;
            }

            padStatus->stick.x = data.akX;
            padStatus->stick.y = data.akY;
        }
        break;
    default:
        // Disconnected by the partner, so becomes ROLE_NONE
        break;
    }
    
    appHeap->Free(buffer);
}

// Main function
void nnMain()
{
    static u8 IrBuff[4096] NN_ATTRIBUTE_ALIGN(4096);
    nn::Result result;

    //Set sleep callback, sleep recovery callback
    s_AwakeEvent.Initialize(false);
    nn::applet::SetSleepQueryCallback(mySleepQueryCallback);
    nn::applet::SetAwakeCallback(myAwakeCallback);
    nn::applet::Enable(true);

    NN_LOG("ir/Communicator demo Start\n");

    const u32 GX_HEAP_SIZE = 0x800000;

    // Initialize hid
    nn::hid::Initialize();

    // Class to obtain system Circle Pad information
    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;
    nn::hid::TouchPanelReader tpReader;
    nn::hid::TouchPanelStatus tpStatus;

    // Initialization for graphics
    nn::fnd::ExpHeap appHeap(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    uptr heapForGx = reinterpret_cast<uptr>(appHeap.Allocate(GX_HEAP_SIZE));
    demo::RenderSystemDrawing renderSystem;
    renderSystem.Initialize(heapForGx, GX_HEAP_SIZE);


    const nn::fnd::TimeSpan waitTime = nn::fnd::TimeSpan::FromMilliSeconds(5*1000);
    nn::ir::BaudRate baudrate = nn::ir::BAUD_RATE_115200;

    //---------------------------
    //Setting of the size of the buffer for sending/receiving
    //---------------------------
    //A maximum of one InputData is saved in the receive buffer, but it has a bit extra to take noise into consideration
    //For ConfirmId, the minimum required size must be specified
    size_t receiveBufferSize     = nn::ir::Communicator::GetPacketSize(sizeof(InputData))*2;
    size_t receiveManagementSize = nn::ir::Communicator::GetManagementSize(1)*2;   //This demo essentially stores the equivalent of one InputData, but extra is added to be on the safe side
    //These also have extra to be on the safe side
    //In the same way, for ConfirmId, the minimum required size must be specified
    size_t sendBufferSize        = nn::ir::Communicator::GetPacketSize(sizeof(InputData))*2;
    size_t sendManagementSize    = nn::ir::Communicator::GetManagementSize(1)*2;

    //The total of the management region, send/receive buffer region, and reserved region must be smaller than the given buffer size
    NN_ASSERT(nn::ir::Communicator::GetReservedSize() + sendBufferSize + sendManagementSize + receiveBufferSize + receiveManagementSize <= 4096);

    result = nn::ir::Communicator::Initialize(
        IrBuff,4096,
        receiveBufferSize,
        receiveManagementSize,
        sendBufferSize,
        sendManagementSize,
        baudrate);

    NN_UTIL_PANIC_IF_FAILED(result);

    NN_LOG("Set baudrate: %d\n", BAUD_RATE_VALUES[GetBaudRateIndex(baudrate)]);

    // Get all events
    // Send completion event
    nn::ir::Communicator::GetSendEvent(&s_SendEvent);
    // Receive completion event
    nn::ir::Communicator::GetReceiveEvent(&s_ReceiveEvent);
    // Connection state event
    nn::ir::Communicator::GetConnectionStatusEvent(&s_ConnectEvent);

    while(1)
    {
        //Sleep support
        if ( nn::applet::IsExpectedToReplySleepQuery() )
        {
            // Due to Sleep, the Ir library transitions to a stopped state (CONNECTION_STATUS_STOPPED) and returns to the state immediately after Initialize.
            // 
            // If in a connected state, that connection is broken, and all data currently being sent and all remaining receive data are discarded.
            // 

            // Connection state events are not signaled.
            
            // Disconnection notification is not sent to the communication partner. If necessary, execute the Disconnect function before Sleep to notify the partner of the disconnection.
            // 
            
            s_AwakeEvent.ClearSignal();
            // Sleep response
            nn::applet::ReplySleepQuery(nn::applet::REPLY_ACCEPT);
            // Wait for wakeup
            s_AwakeEvent.Wait();

            // Processing after restart
            nn::gx::StartLcdDisplay();
        }

        nn::ir::ConnectionStatus status = nn::ir::Communicator::GetConnectionStatus();

        if(!padReader.ReadLatest(&padStatus))
        {
            continue;
        }
        if(!tpReader.ReadLatest(&tpStatus))
        {
            continue;
        }

        //----------------
        //Key operations
        //A: Connection Request
        //Y: Connection Response
        //X: Automatic Connection
        //B: Disconnect
        //L/R: Change baud rate
        //----------------
        if(padStatus.trigger & nn::hid::BUTTON_A)
        {
            // Request partner for connection
            NN_LOG("RequireConnection\n");
            result = nn::ir::Communicator::RequireConnection();
        }
        else if(padStatus.trigger & nn::hid::BUTTON_Y)
        {
            // Wait for connection request
            NN_LOG("WaitConnection\n");
            result = nn::ir::Communicator::WaitConnection(waitTime);
        }
        else if(padStatus.trigger & nn::hid::BUTTON_X)
        {
            // Start automatic connection
            NN_LOG("AutoConnection\n");
            nn::ir::Communicator::AutoConnection();
        }
        else if(padStatus.trigger & nn::hid::BUTTON_B)
        {
            // If currently connecting, cancel request/wait
            // If in a connected state, send disconnect request to partner and enter a disconnected state
            NN_LOG("Disconnect\n");
            nn::ir::Communicator::Disconnect();
        }
        else if(padStatus.trigger & nn::hid::BUTTON_L)
        {
            if(status == nn::ir::CONNECTION_STATUS_STOPPED)
            {
                nn::ir::Communicator::Finalize();
                baudrate = GetNextSmallerBaudRate(baudrate);
                result = nn::ir::Communicator::Initialize(
                    IrBuff,4096,
                    receiveBufferSize,
                    receiveManagementSize,
                    sendBufferSize,
                    sendManagementSize,
                    baudrate);
                // Get all events
                // Send completion event
                nn::ir::Communicator::GetSendEvent(&s_SendEvent);
                // Receive completion event
                nn::ir::Communicator::GetReceiveEvent(&s_ReceiveEvent);
                // Connection state event
                nn::ir::Communicator::GetConnectionStatusEvent(&s_ConnectEvent);
                
                NN_LOG("Set baudrate: %d\n", BAUD_RATE_VALUES[GetBaudRateIndex(baudrate)]);
            }
        }
        else if(padStatus.trigger & nn::hid::BUTTON_R)
        {
            if(status == nn::ir::CONNECTION_STATUS_STOPPED)
            {
                nn::ir::Communicator::Finalize();
                baudrate = GetNextBiggerBaudRate(baudrate);
                result = nn::ir::Communicator::Initialize(
                    IrBuff,4096,
                    receiveBufferSize,
                    receiveManagementSize,
                    sendBufferSize,
                    sendManagementSize,
                    baudrate);
                // Get all events
                // Send completion event
                nn::ir::Communicator::GetSendEvent(&s_SendEvent);
                // Receive completion event
                nn::ir::Communicator::GetReceiveEvent(&s_ReceiveEvent);
                // Connection state event
                nn::ir::Communicator::GetConnectionStatusEvent(&s_ConnectEvent);
                
                NN_LOG("Set baudrate: %d\n", BAUD_RATE_VALUES[GetBaudRateIndex(baudrate)]);
            }
        }

        // Monitor state changes.
        // An event is signaled under the following conditions.
        // 
        // 1. When a connection is established with ***Connection
        // 2. Disconnect process completes
        // 3. Connection cut due to Disconnect from partner
        // 4. When an IR failure (nn::ir::ResultFatalError) occurs
        if(s_ConnectEvent.Wait(0))
        {
            switch(nn::ir::Communicator::GetConnectionStatus())
            {
                case nn::ir::CONNECTION_STATUS_CONNECTED:
                    {
                        NN_LOG("Connected\n");
                    }
                    break;
                case nn::ir::CONNECTION_STATUS_STOPPED:
                    {
                        NN_LOG("Disconnected\n");
                    }
                    break;
                case nn::ir::CONNECTION_STATUS_FATAL_ERROR:
                    {
                        // The IR may be malfunctioning. This state continues until the power to the system is cut.
                        // Recommended to terminate the IR library and notify the user that IR may be malfunctioning
                        nn::ir::Communicator::Finalize();

                        NN_PANIC("IR Fatal Error occurred.\n");
                    }
                    break;
            }
        }
        // Process that is implemented only when in a connected state.
        if( nn::ir::Communicator::GetConnectionStatus() == nn::ir::CONNECTION_STATUS_CONNECTED)
        {
            // Execute confirmation process if the ID confirmation process with the connection partner has yet to be performed
            if(!nn::ir::Communicator::IsConfirmedId())
            {
                //----------------
                //ID confirmation process
                //----------------
                result = ConfirmID();

                if(result.IsSuccess())
                {
                    NN_LOG("Comfirm Success.\n");
                }
                else
                {
                    NN_LOG("Comfirm Failure.\n");
                    nn::ir::Communicator::Disconnect();
                }
            }
            else
            {
                // Execute communication because ID confirmation finished
                
                //----------------
                //Communication processing
                //----------------
                Communicate(&padStatus,&tpStatus,&appHeap);
            }

        }

        /* After this, render processing */

        //Render the upper screen
        renderSystem.SetRenderTarget(NN_GX_DISPLAY0);
        renderSystem.Clear();

        renderSystem.SetColor(1.f,1.f,1.f);
        renderSystem.DrawText(0,0,"Communicator Demo");
        renderSystem.DrawText(0,10,"[STATUS]  %s",GetStatusLabel(nn::ir::Communicator::GetConnectionStatus()));
        renderSystem.DrawText(0,20,"[Confirm] %s",nn::ir::Communicator::IsConfirmedId()?"true":"false");
        renderSystem.DrawText(0,30,"[Role] %s", s_ConnectionRoleText[nn::ir::Communicator::GetConnectionRole()]);
        renderSystem.DrawText(0,40,"[BaudRate] L < % 5d > R",GetBaudRateValue(baudrate));
        renderSystem.DrawText(0,50,"[API] A:RequireConnection");
        renderSystem.DrawText(0,60,"      Y:WaitConnection");
        renderSystem.DrawText(0,70,"      X:Auto");
        renderSystem.DrawText(0,80,"      B:Disconnect");
        renderSystem.SwapBuffers();

        //Render the lower screen
        renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
        renderSystem.Clear();

        renderSystem.SetColor(1.f,0.f,0.f);
        renderSystem.FillRectangle(
            110 * padReader.NormalizeStick(padStatus.stick.x) + 155,
           -110 * padReader.NormalizeStick(padStatus.stick.y) + 120,
            5,5
        );
        if(tpStatus.touch)
        {
            renderSystem.SetColor(0.f,1.f,0.f);
            renderSystem.FillRectangle(
                tpStatus.x,
                tpStatus.y,
                5,5
            );
        }

        renderSystem.SwapBuffers();
        renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
    }
}

