﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     jpeg_MultiPicture.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/types.h>
#include <nn/dbg.h>

#include <nn/os/os_Memory.h>
#include <nn/os/os_MemoryTypes.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include <nn/fs.h>
#include <nn/applet.h>

nn::fnd::ExpHeap s_AppHeap;

void Initialize(void);

void Finalize(void);


// Buffers for JPEG codex tests, etc.
#define WIDTH 640
#define HEIGHT 480
#include <nn/jpeg.h>
#include <string.h>

const char s_SampleDateTime[nn::jpeg::DATE_TIME_SIZE] = "2000:01:01 00:00:00";

// The original image for encoding must be a minimum of 4 bytes and, depending on the format, may require a larger alignment size.
// Since it is loaded from ROM, the CPU cache line size unit is used.
u16 encodeSrcYuvBuffer[2][WIDTH * HEIGHT] NN_ATTRIBUTE_ALIGN(32);

// For this demo, the encoding result size is hard-coded to be 256 KB or less. Alignment is freely set.
u8 encodeDstBuffer[256 * 1024];

// Since it is loaded from ROM, the CPU cache line size unit is used.
u8 encodeVerifyBuffer[256 * 1024] NN_ATTRIBUTE_ALIGN(32);
u8 decodeSrcBuffer[256 * 1024] NN_ATTRIBUTE_ALIGN(32);

// The encoding result image must be a minimum of 4 bytes and, depending on the format, may require a larger alignment size.
// Aligned at 128 bytes assuming referencing from the GPU as the texture data.
u16 decodeDstRgbBuffer[2][WIDTH * HEIGHT] NN_ATTRIBUTE_ALIGN(128);


void JpegMultiPictureEncode()
{
    // Original image size is hard-coded.
    const u32 width = WIDTH;
    const u32 height = HEIGHT;
    // Two images are stored in MP format.
    const u32 numImages = 2;

    // Load the original image for encoding.
    // Two pieces of 640x480 pixel image data (either existing or as if captured from the camera) in the YUV format the same as the results of camera capture are loaded from ROM.
    // 
    // 
    nn::fs::FileReader yuvReaderL(L"rom:/sample2l.yuv");
    size_t sizeL = yuvReaderL.GetSize();
    s32 readL = yuvReaderL.Read(encodeSrcYuvBuffer[0], sizeL);
    yuvReaderL.Finalize();
    if(readL != sizeL)
    {
        NN_PANIC("not expected\n");
    }

    nn::fs::FileReader yuvReaderR(L"rom:/sample2r.yuv");
    size_t sizeR = yuvReaderR.GetSize();
    s32 readR = yuvReaderR.Read(encodeSrcYuvBuffer[1], sizeR);
    yuvReaderR.Finalize();
    if(readR != sizeR)
    {
        NN_PANIC("not expected\n");
    }

    // Example of Encoder Use:
    // The process flow is basically the same as when the MP format is not used.
    // However, the number of stored images must be added to the arguments for GetWorkBufferSize and Initialize function.

    // 1. Generates the encoder object.
    nn::jpeg::JpegMpEncoder encoder;

    // 2. Allocates a work buffer for the encoder object.
    // Four-byte alignment is required.
    // Since MP format is used, specify the number of stored images.
    size_t encoderWorkSize = encoder.GetWorkBufferSize(numImages);
    void* encoderWork = s_AppHeap.Allocate(encoderWorkSize, 4);
    NN_ASSERT(encoderWork);

    // 3. Initializes the encoder object.
    // Since MP format is used, specify the number of stored images.
    bool init NN_IS_UNUSED_VAR = encoder.Initialize(encoderWork, encoderWorkSize, numImages);
    NN_ASSERT(init);

    // 4. Allocates output buffer.
    // encodeDstBuffer is used this time.

    // In this demo, the date and time have a fixed value so that the encoding results are fixed.
    encoder.SetDateTime(s_SampleDateTime);

    // 5. Calls the encoding function.
    size_t encodeResult = encoder.StartMpEncoderLR(encodeDstBuffer,
                                                   sizeof(encodeDstBuffer),
                                                   encodeSrcYuvBuffer[0],
                                                   encodeSrcYuvBuffer[1],
                                                   width,
                                                   height,
                                                   80,
                                                   nn::jpeg::PIXEL_SAMPLING_YUV422,
                                                   nn::jpeg::PIXEL_FORMAT_YUYV8,
                                                   false,
                                                   false);

    // 6. Terminates and destroys the encoder object.
    encoder.Finalize();

    // 7. Releases the work buffer for the encoder object.
    s_AppHeap.Free(encoderWork);

    // 8. Releases the output buffer once it is no longer required.
    // Do not release this time.

    // Encode result verification.
    nn::fs::FileReader mpoReader(L"rom:/sample2_2x1.mpo");
    size_t size = mpoReader.GetSize();
    size_t read NN_IS_UNUSED_VAR = mpoReader.Read(encodeVerifyBuffer, size);
    mpoReader.Finalize();
    NN_ASSERT(read == size);

    NN_LOG("MP encodeResult 0x%x, expected 0x%x\n", encodeResult, size);
    if(encodeResult != size)
    {
        NN_PANIC("not expected\n");
    }
    NN_ASSERT(!memcmp(encodeVerifyBuffer, encodeDstBuffer, size));
}

void JpegMultiPictureDecode()
{
    u8* src = decodeSrcBuffer;
    size_t srcSize;
    // Original image size is hard-coded.
    const u32 width = WIDTH;
    const u32 height = HEIGHT;
    // Load the original image for decoding.
    // Load 640x480 pixel JPEG data.
    nn::fs::FileReader mpoReader(L"rom:/sample2_2x1.mpo");
    srcSize = mpoReader.GetSize();
    size_t read NN_IS_UNUSED_VAR = mpoReader.Read(src, srcSize);
    mpoReader.Finalize();
    NN_ASSERT(read == srcSize);

    // Example of Decoder Use:
    // The process flow is basically the same as when MP format is not used except for the acquisition of MP format attached data.
    // 

    // 1. Generates the decoder object.
    nn::jpeg::JpegMpDecoder decoder;

    // 2. Allocates a work buffer for the decoder object.
    // Four-byte alignment is required.
    size_t decoderWorkSize = decoder.GetWorkBufferSize();
    void* decoderWork = s_AppHeap.Allocate(decoderWorkSize, 4);
    NN_ASSERT(decoderWork);

    // 3. Initializes the decoder object.
    // Unlike for the MP encoder, it is unnecessary to specify the number of images when using MP format.
    // It may become necessary in the future.
    bool init NN_IS_UNUSED_VAR = decoder.Initialize(decoderWork, decoderWorkSize);
    NN_ASSERT(init);

    // 4. Allocates output buffer.
    // This time a sufficiently sized decodeDstYuvBuffer is allocated in advance.
    const nn::jpeg::PixelFormat dstPixelFormat = nn::jpeg::PIXEL_FORMAT_CTR_RGB565_BLOCK8;
    size_t dstBufferSize = decoder.GetDstBufferSize(width, height, dstPixelFormat);

    // 5. In order to acquire the MP format attached data, the MP index IFD is obtained first.
    nn::jpeg::MpIndex index;
    bool result NN_IS_UNUSED_VAR;
    result = decoder.GetMpIndex(&index, src, srcSize);
    NN_ASSERT(result);

    // 6. Gets the number of recorded images using the obtained MP index IFD.
    u32 numImages = 0;
    result = decoder.GetMpNumberOfImages(&numImages, &index);
    NN_ASSERT(result);
    NN_ASSERT(numImages == 2);

    // 7. Gets the MP entry.
    nn::jpeg::MpEntry entry[2];
    result = decoder.GetMpEntry(&entry[0], &index, 0);
    NN_ASSERT(result);
    result = decoder.GetMpEntry(&entry[1], &index, 1);
    NN_ASSERT(result);

    // 8. Gets the management data per individual image type using the obtained MP entry.
    u32 type[2] NN_IS_UNUSED_VAR;
    type[0] = decoder.GetMpImageType(&entry[0]);
    type[1] = decoder.GetMpImageType(&entry[1]);
    NN_ASSERT(type[0] == (nn::jpeg::MP_TYPE_FLAG_REPRESENTATIVE_IMAGE | nn::jpeg::MP_TYPE_CODE_MULTI_VIEW_DISPARITY_IMAGE));
    NN_ASSERT(type[1] == nn::jpeg::MP_TYPE_CODE_MULTI_VIEW_DISPARITY_IMAGE);

    // 9. Gets the data size and offset for the individual images.
    size_t imageSize[2];
    imageSize[0] = decoder.GetMpImageSize(&entry[0]);
    imageSize[1] = decoder.GetMpImageSize(&entry[1]);

    size_t imageOffset[2];
    imageOffset[0] = decoder.GetMpImageOffset(&entry[0]);
    imageOffset[1] = decoder.GetMpImageOffset(&entry[1]);

    // 10. Calls the decoding function.
    u32 loop;
    for (loop = 0; loop < 2; loop++)
    {
        // Check whether the data size and offset specify an appropriate region.
        NN_ASSERT(imageSize[loop] && (imageSize[loop] < srcSize));
        NN_ASSERT(imageOffset[loop] < srcSize);
        NN_ASSERT((imageOffset[loop] + imageSize[loop]) <= srcSize);

        NN_LOG("[%u] offset 0x%x, size 0x%x\n", loop, imageOffset[loop], imageSize[loop]);
        size_t decodeResult = decoder.StartJpegDecoder(decodeDstRgbBuffer[loop],
                                                       sizeof(decodeDstRgbBuffer[loop]),
                                                       src + imageOffset[loop],
                                                       imageSize[loop],
                                                       width,
                                                       height,
                                                       dstPixelFormat,
                                                       false);
        // decodeResult is the number of bytes of the decoded image and is non-zero if successful.
        // The result varies depending on the size of the decoded image.
        NN_LOG("decodeResult 0x%x, expected 0x%x\n", decodeResult, dstBufferSize);
        if(decodeResult != dstBufferSize)
        {
            NN_PANIC("not expected\n");
        }
    }

    // 11. Terminates and destroys the decoder object.
    decoder.Finalize();

    // 12. Releases the work buffer for the decoder object.
    s_AppHeap.Free(decoderWork);

    // 13. Releases the output buffer once it is no longer required.
    // Do not release this time.
}

void JpegMultiPictureEncode2()
{
    // Original image size is hard-coded.
    const u32 width = WIDTH;
    const u32 height = HEIGHT;
    // Two images are stored in MP format.
    const u32 numImages = 2;

    // Load the original image for encoding.
    // Two pieces of 640x480 pixel image data (as if captured from the camera) in the YUV format the same as the results of camera capture are loaded from ROM.
    // 
    // 
    nn::fs::FileReader yuvReaderL(L"rom:/sample2l.yuv");
    size_t sizeL = yuvReaderL.GetSize();
    s32 readL = yuvReaderL.Read(encodeSrcYuvBuffer[0], sizeL);
    yuvReaderL.Finalize();
    if(readL != sizeL)
    {
        NN_PANIC("not expected\n");
    }

    nn::fs::FileReader yuvReaderR(L"rom:/sample2r.yuv");
    size_t sizeR = yuvReaderR.GetSize();
    s32 readR = yuvReaderR.Read(encodeSrcYuvBuffer[1], sizeR);
    yuvReaderR.Finalize();
    if(readR != sizeR)
    {
        NN_PANIC("not expected\n");
    }

    // Example of Encoder Use:
    // The process flow is basically the same as when the MP format is not used.
    // However, the number of stored images must be added to the arguments for GetWorkBufferSize and Initialize function.

    // 1. Generates the encoder object.
    nn::jpeg::JpegMpEncoder encoder;

    // 2. Allocates a work buffer for the encoder object.
    // Four-byte alignment is required.
    // Since MP format is used, specify the number of stored images.
    size_t encoderWorkSize = encoder.GetWorkBufferSize(numImages);
    void* encoderWork = s_AppHeap.Allocate(encoderWorkSize, 4);
    NN_ASSERT(encoderWork);

    // 3. Initializes the encoder object.
    // Since MP format is used, specify the number of stored images.
    bool init NN_IS_UNUSED_VAR = encoder.Initialize(encoderWork, encoderWorkSize, numImages);
    NN_ASSERT(init);

    // 4. Allocates output buffer.
    // encodeDstBuffer is used this time.

    // In this demo, the date and time have a fixed value so that the encoding results are fixed.
    encoder.SetDateTime(s_SampleDateTime);

    // 5. Calls the encoding function.
    // Encodes the first image.
    const u32 quality = 80;
    const nn::jpeg::PixelSampling dstPixelSampling = nn::jpeg::PIXEL_SAMPLING_YUV422;
    const nn::jpeg::PixelFormat   srcPixelFormat   = nn::jpeg::PIXEL_FORMAT_YUYV8;
    size_t encodeResult = encoder.StartMpEncoderFirst(encodeDstBuffer,
                                                      sizeof(encodeDstBuffer),
                                                      encodeSrcYuvBuffer[0],
                                                      width,
                                                      height,
                                                      quality,
                                                      dstPixelSampling,
                                                      srcPixelFormat,
                                                      false,
                                                      numImages);
    NN_ASSERT(encodeResult);

    if (encodeResult)
    {
        // In this demo, the date and time have a fixed value so that the encoding results are fixed.
        encoder.SetDateTime(s_SampleDateTime);

        // Encodes the second image.
        encodeResult = encoder.StartMpEncoderNext(encodeSrcYuvBuffer[1],
                                                  width,
                                                  height,
                                                  quality,
                                                  dstPixelSampling,
                                                  srcPixelFormat,
                                                  false);
    }

    // 6. Terminates and destroys the encoder object.
    encoder.Finalize();

    // 7. Releases the work buffer for the encoder object.
    s_AppHeap.Free(encoderWork);

    // 8. Releases the output buffer once it is no longer required.
    // Do not release this time.

    // Encode result verification.
    // This example has the same result as when using StartMpEncoderLR function.
    nn::fs::FileReader mpoReader(L"rom:/sample2_2x1.mpo");
    size_t size = mpoReader.GetSize();
    size_t read NN_IS_UNUSED_VAR = mpoReader.Read(encodeVerifyBuffer, size);
    mpoReader.Finalize();
    NN_ASSERT(read == size);

    NN_LOG("MP encodeResult 0x%x, expected 0x%x\n", encodeResult, size);
    if(encodeResult != size)
    {
        NN_PANIC("not expected\n");
    }
    NN_ASSERT(!memcmp(encodeVerifyBuffer, encodeDstBuffer, size));
}

void Initialize(void)
{
    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    // Create heap    
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), 
        nn::os::GetDeviceMemorySize() );
}

void Finalize(void)
{
    s_AppHeap.Finalize();    
}

void nnMain( void )
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    Initialize();

    // JPEG encoding/decoding test (not shown onscreen).
    JpegMultiPictureEncode();
    JpegMultiPictureDecode();

    // One image at a time is encoded in MP format.
    JpegMultiPictureEncode2();

    // If it gets all the way to here, the test is successful.
    NN_LOG("Demo finished.\n");

    // Finalize();
    while(1)
    {
    }
}
