﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     jpeg_Simple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/types.h>
#include <nn/dbg.h>

#include <nn/os/os_Memory.h>
#include <nn/os/os_MemoryTypes.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include "demo.h"

demo::RenderSystemDrawing s_RenderSystem;
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;

namespace 
{
    const int GL_MEMORY_SIZE = 0x800000;
}

// Texture0
GLuint s_BmpTexture0Id = 0;
u32 s_Bmp0Width = 0;
u32 s_Bmp0Height = 0;
u32 s_Texture0Width = 0;
u32 s_Texture0Height = 0;


void Initialize(void);
void LoadTexture0(void);
void DeleteTexture0(void);

void Finalize(void);
void DrawFrame(void);
void DrawDisplay0(void);


// Buffers for JPEG codex tests, etc.
#define WIDTH 256
#define HEIGHT 128
#define Y2R_RGB_BYTES 2
#define Y2R_YUV_BYTES 2
#include <nn/jpeg.h>
#include <string.h>

const char s_SampleDateTime[nn::jpeg::DATE_TIME_SIZE] = "2000:01:01 00:00:00";

u8 rgbBufferJpeg[2 * WIDTH * HEIGHT * Y2R_RGB_BYTES] NN_ATTRIBUTE_ALIGN(64);

// The original image for encoding must be a minimum of 4 bytes and, depending on the format, may require a larger alignment size.
// Since it is loaded from ROM, the CPU cache line size unit is used.
u8 encodeSrcYuvBuffer[2 * WIDTH * HEIGHT * Y2R_YUV_BYTES] NN_ATTRIBUTE_ALIGN(32);

// For this demo, the encoding result size is hard-coded to be 256 KB or less. Alignment is freely set.
u8 encodeDstBuffer[256 * 1024];

// Since it is loaded from ROM, the CPU cache line size unit is used.
u8 encodeVerifyBuffer[256 * 1024] NN_ATTRIBUTE_ALIGN(32);
u8 decodeSrcBuffer[256 * 1024] NN_ATTRIBUTE_ALIGN(32);

// The encoded image must be a minimum of 4 bytes and, depending on the format, may require a larger alignment size.
// With this demo, since it is also the input buffer to the Y2R conversion hardware, the CPU cache line size unit is used.
u8 decodeDstYuvBuffer[2 * WIDTH * HEIGHT * Y2R_YUV_BYTES] NN_ATTRIBUTE_ALIGN(32);


void JpegEncode()
{
    // Original image size is hard-coded.
    const u32 width = WIDTH;
    const u32 height = HEIGHT;
    // Load the original image for encoding.
    // Loads from ROM 256x128 pixel images in a YUV format that is the same as results captured with the camera (images that seem as if they were captured from the camera).
    // 
    // 
    nn::fs::FileReader yuvReader(L"rom:/sample1.yuv");
    size_t size = yuvReader.GetSize();
    s32 read = yuvReader.Read(encodeSrcYuvBuffer, size);
    yuvReader.Finalize();
    if(read != size)
    {
        NN_PANIC("Unexpected\n");
    }

    // Example of Encoder Use:
    // 1. Generates the encoder object.
    // 2. Allocates a work buffer for the encoder object.
    // 3. Initializes the encoder object.
    // 4. Allocates output buffer.
    // 5. Calls the encoding function.
    // 6. Terminates and destroys the encoder object.
    // 7. Releases the work buffer for the encoder object.
    // 8. Releases the output buffer once it is no longer required.

    // Allocating and releasing the buffer does not have to follow this exactly.
    // For example, step 4 can be performed before 3.

    // 1. Generates the encoder object.
    nn::jpeg::JpegMpEncoder encoder;

    // 2. Allocates a work buffer for the encoder object.
    // Four-byte alignment is required.
    size_t encoderWorkSize = encoder.GetWorkBufferSize();
    void* encoderWork = s_AppHeap.Allocate(encoderWorkSize, 4);

    // 3. Initializes the encoder object.
    encoder.Initialize(encoderWork, encoderWorkSize);

    // 4. Allocates output buffer.
    // encodeDstBuffer is used this time.

    // In this demo, the date and time have a fixed value so that the encoding results are fixed.
    encoder.SetDateTime(s_SampleDateTime);

    // 5. Calls the encoding function.
    const nn::jpeg::PixelSampling dstPixelSampling = nn::jpeg::PIXEL_SAMPLING_YUV422;
    size_t encodeResult = encoder.StartJpegEncoder(encodeDstBuffer,
                                                   sizeof(encodeDstBuffer),
                                                   encodeSrcYuvBuffer,
                                                   width,
                                                   height,
                                                   80,
                                                   dstPixelSampling,
                                                   nn::jpeg::PIXEL_FORMAT_YUYV8,
                                                   false);

    // 6. Terminates and destroys the encoder object.
    encoder.Finalize();

    // 7. Releases the work buffer for the encoder object.
    s_AppHeap.Free(encoderWork);

    // 8. Releases the output buffer once it is no longer required.
    // Do not release this time.

    // Encode result verification.
    nn::fs::FileReader jpgReader((dstPixelSampling == nn::jpeg::PIXEL_SAMPLING_YUV420)    ? L"rom:/sample1_2x2.jpg"
                                 : ((dstPixelSampling == nn::jpeg::PIXEL_SAMPLING_YUV422) ? L"rom:/sample1_2x1.jpg"
                                    : /* PIXEL_SAMPLING_YUV444 */                                L"rom:/sample1_1x1.jpg"));
    size = jpgReader.GetSize();
    read = jpgReader.Read(encodeVerifyBuffer, size);
    jpgReader.Finalize();
    NN_ASSERT(read == size);

    NN_LOG("encodeResult 0x%x, expected 0x%x\n", encodeResult, size);
    if(encodeResult != size)
    {
        NN_PANIC("Unexpected\n");
    }
    NN_ASSERT(!memcmp(encodeVerifyBuffer, encodeDstBuffer, size));
}

void JpegDecode()
{
    // Original image size is hard-coded.
    const u32 width = WIDTH;
    const u32 height = HEIGHT;
    // Load the original image for decoding.
    // Load 256x128 pixel JPEG data.
    nn::fs::FileReader jpgReader(L"rom:/sample1_2x1.jpg");
    size_t size = jpgReader.GetSize();
    s32 read = jpgReader.Read(decodeSrcBuffer, size);
    jpgReader.Finalize();
    if(read != size)
    {
        NN_PANIC("Unexpected\n");
    }

    // Example of Decoder Use:
    // 1. Generates the decoder object.
    // 2. Allocates a work buffer for the decoder object.
    // 3. Initializes the decoder object.
    // 4. Allocates output buffer.
    // 5. Calls the decoding function.
    // 6. Terminates and destroys the decoder object.
    // 7. Releases the work buffer for the decoder object.
    // 8. Releases the output buffer once it is no longer required.

    // Allocating and releasing the buffer does not have to follow this exactly.
    // For example, step 4 can be performed before 3.

    // 1. Generates the decoder object.
    nn::jpeg::JpegMpDecoder decoder;

    // 2. Allocates a work buffer for the decoder object.
    // Four-byte alignment is required.
    size_t decoderWorkSize = decoder.GetWorkBufferSize();
    void* decoderWork = s_AppHeap.Allocate(decoderWorkSize, 4);

    // 3. Initializes the decoder object.
    decoder.Initialize(decoderWork, decoderWorkSize);

    // 4. Allocates output buffer.
    // This time a sufficiently sized rgbBuffer is allocated in advance.
    const nn::jpeg::PixelFormat dstPixelFormat = nn::jpeg::PIXEL_FORMAT_CTR_RGB565_BLOCK8;
    size_t dstBufferSize = decoder.GetDstBufferSize(width, height, dstPixelFormat);

    // 5. Calls the decoding function.
    size_t decodeResult = decoder.StartJpegDecoder(rgbBufferJpeg,
                                                   dstBufferSize,
                                                   decodeSrcBuffer,
                                                   size,
                                                   width,
                                                   height,
                                                   dstPixelFormat,
                                                   false);

    // Sets the size of the decoding result.
    //s_Bmp0Width = s_Texture0Width = decoder.GetWidth();
    //s_Bmp0Height = s_Texture0Height = decoder.GetHeight();
    s_Bmp0Width = s_Texture0Width = WIDTH;
    s_Bmp0Height = s_Texture0Height = HEIGHT;

    // 6. Terminates and destroys the decoder object.
    decoder.Finalize();

    // 7. Releases the work buffer for the decoder object.
    s_AppHeap.Free(decoderWork);

    // 8. Releases the output buffer once it is no longer required.
    // Do not release this time.

    // decodeResult is the number of bytes of the decoded image and is non-zero if successful.
    // The result varies depending on the size of the decoded image.
    NN_LOG("decodeResult 0x%x, expected 0x%x\n", decodeResult, dstBufferSize);
    if(decodeResult != dstBufferSize)
    {
        NN_PANIC("Unexpected\n");
    }
}


void Initialize(void)
{
    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    // Create heap    
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), 
        nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(GL_MEMORY_SIZE));
       

    // JPEG encoding/decoding test
    JpegEncode();
    JpegDecode();

    // Allocates 8 MB memory in FCRAM for graphics    
    s_RenderSystem.Initialize(s_HeapForGx, GL_MEMORY_SIZE);

    LoadTexture0();
}

void LoadTexture0(void)
{
    if ( s_BmpTexture0Id != 0 )
    {
        DeleteTexture0();
    }

    GLenum target = GL_TEXTURE_2D;
    GLenum internalFormat = GL_RGB_NATIVE_DMP; 
    GLenum format = GL_RGB_NATIVE_DMP;
    GLenum type = GL_UNSIGNED_SHORT_5_6_5;
    GLuint textureId = 0;

    s_RenderSystem.GenerateTexture(target, internalFormat, WIDTH, HEIGHT,
        format, type, rgbBufferJpeg, textureId);

    if ( textureId != 0 )
    {
        s_BmpTexture0Id = textureId;
        NN_LOG("  Create texture (id = %d)\n", textureId);
    }
}

void DeleteTexture0(void)
{        
    if ( s_BmpTexture0Id != 0 )
    {
        bool flag = s_RenderSystem.DeleteTexture(s_BmpTexture0Id);
        if ( flag )
        {
            NN_LOG("  Delete texture. (id = %d)\n", s_BmpTexture0Id);
            s_BmpTexture0Id = 0;
        }
        else
        {
            NN_TPANIC_("  Failed to delete texture. (id = %d)\n", s_BmpTexture0Id);
        }
    }
}

void Finalize(void)
{
    DeleteTexture0();
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));

    s_AppHeap.Finalize();    
}

void DrawFrame(void)
{
    DrawDisplay0();
  
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();
    
    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void DrawDisplay0(void)
{
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, 0.0f, 1.0f, 0.0f, 1.0f);
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);    
    s_RenderSystem.Clear();

    f32 windowCoordinateX = 0.0f;
    f32 windowCoordinateY = 0.0f;
    f32 rectangleWidth = 400.0f;
    f32 rectangleHeight = 240.0f;
    s_RenderSystem.FillTexturedRectangle(s_BmpTexture0Id,
        windowCoordinateX, windowCoordinateY,
        rectangleWidth, rectangleHeight,
        s_Bmp0Width, s_Bmp0Height,
        s_Texture0Width, s_Texture0Height);

   s_RenderSystem.SwapBuffers(); 
}

void nnMain( void )
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    Initialize();
    
    while ( 1 )
    {
        DrawFrame();
    }

    // Finalize();
}
