﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     jpeg_StopDecoder.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/jpeg.h>
#include <nn/applet.h>

nn::fnd::ExpHeap s_AppHeap;

void Initialize(void);

void Finalize(void);


// Buffers for JPEG codex tests, etc.
#define WIDTH 256
#define HEIGHT 128

nn::jpeg::JpegMpDecoder s_Decoder;
size_t s_DecoderWorkSize;
void* s_pDecoderWork;
nn::os::Thread s_Thread;
nn::os::StackMemoryBlock s_ThreadStack(4096);
volatile size_t s_DecodeResult;

// Since it is loaded from ROM, the CPU cache line size unit is used.
u8 s_DecodeSrcBuffer[256 * 1024] NN_ATTRIBUTE_ALIGN(32);
size_t s_DecodeSrcSize;

// The encoding result image must be a minimum of 4 bytes and, depending on the format, may require a larger alignment size.
// Aligned at 128 bytes assuming referencing from the GPU as the texture data.
u16 s_DecodeDstRgbBuffer[WIDTH * HEIGHT] NN_ATTRIBUTE_ALIGN(128);

void InitializeJpegDecoder()
{
    // Load the original image for decoding.
    nn::fs::FileReader jpgReader(L"rom:/sample1_2x2.jpg");
    s_DecodeSrcSize = jpgReader.GetSize();
    size_t read NN_IS_UNUSED_VAR = jpgReader.Read(s_DecodeSrcBuffer, s_DecodeSrcSize);
    jpgReader.Finalize();
    NN_ASSERT(read == s_DecodeSrcSize);

    // 1. Allocates a work buffer for the decoder object.
    // Four-byte alignment is required.
    s_DecoderWorkSize = s_Decoder.GetWorkBufferSize();
    s_pDecoderWork = s_AppHeap.Allocate(s_DecoderWorkSize, 4);
    NN_ASSERT(s_pDecoderWork);

    // 2. Initializes the decoder object.
    bool init NN_IS_UNUSED_VAR = s_Decoder.Initialize(s_pDecoderWork, s_DecoderWorkSize);
    NN_ASSERT(init);
}

void DecoderThreadFunc()
{
    // Original image size is hard-coded.
    const u32 width = WIDTH;
    const u32 height = HEIGHT;

    // 3. Allocates output buffer.
    // This time a sufficiently sized s_DecodeDstRgbBuffer is allocated in advance.
    const nn::jpeg::PixelFormat dstPixelFormat = nn::jpeg::PIXEL_FORMAT_CTR_RGB565;
    size_t dstBufferSize NN_IS_UNUSED_VAR = s_Decoder.GetDstBufferSize(width, height, dstPixelFormat);
    NN_ASSERT(dstBufferSize <= sizeof(s_DecodeDstRgbBuffer));

    // 4. Calls the decoding function.
    // The return value is checked with the main thread.
    s_DecodeResult = s_Decoder.StartJpegDecoder(s_DecodeDstRgbBuffer,
                                                sizeof(s_DecodeDstRgbBuffer),
                                                s_DecodeSrcBuffer,
                                                s_DecodeSrcSize,
                                                width,
                                                height,
                                                dstPixelFormat,
                                                false);
}

void FinalizeJpegDecoder()
{
    // 5. Terminates the decoder object.
    // Do not destroy this time.
    s_Decoder.Finalize();

    // 6. Releases the work buffer for the decoder object.
    s_AppHeap.Free(s_pDecoderWork);

    // 7. Releases the output buffer once it is no longer required.
    // Do not release this time.
}

void Initialize(void)
{
    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    // Create heap    
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), 
        nn::os::GetDeviceMemorySize() );
}

void Finalize(void)
{
    s_AppHeap.Finalize();    
}

void nnMain( void )
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    const u32 MAX_LOOPS = 3;
    // Changes the main thread standby time to confirm the effectiveness of the decoding abort feature.
    // 1 us: The abort request is ignored because the decode thread cannot reach the internal main processing of the decoding function. Decoding is successful.
    // 1 ms: The decoding function is performing the internal main process. An abort request is received, so decoding fails. GetLastError function returns nn::jpeg::JPEG_DECODER_ERROR_STOPPED (-10).
    // 1 s: The decoding has already succeeded. The abort request is ignored.
    // The expected result may not be obtained because of the JPEG data and slight differences in the timing of the execution environment.
    static const u32 waitTable[MAX_LOOPS] =
    {
        1,           // 1 us - decodeResult 0x10000, errorCode 0
        1000,        // 1 ms - decodeResult 0x0, errorCode -10
        1000 * 1000  // 1 s  - decodeResult 0x10000, errorCode 0
    };
    u32 loop;
    s32 errorCode NN_IS_UNUSED_VAR;

    Initialize();

    // JPEG decode abort feature test (not shown onscreen).
    // Initializes the decoder object.
    InitializeJpegDecoder();

    for (loop = 0; loop < MAX_LOOPS; loop++)
    {
        // Generates a thread for decoding, and starts decoding in the background.
        s_Thread.Start(DecoderThreadFunc, s_ThreadStack, nn::os::DEFAULT_THREAD_PRIORITY + 1);

        // Find the good request timing during the decoding and then request to abort the decode process.
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMicroSeconds(waitTable[loop]));

        // Perform the decode stop request repeatedly in intervals (form example in each game frame) until it is confirmed that either the decode stopped or finished regardless of failure or success.
        // 
        // The abort request is cleared once from the decoding function called by the decode thread.
        // For that reason, the abort request may be ignored if the request is performed just once.
        while (true)
        {
            s_Decoder.StopDecoder();

            // The sample exemplifies the case of performing the abort request only once and that the request may be ignored.
            // To repeat the abort request, change #if 1 to #if 0.
#if 1
            break;
#else
            if (!s_Thread.IsAlive())
            {
                break;
            }

            nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(100));
#endif
        }

        // Waits for the decode thread to complete and then destroys it.
        s_Thread.Join();
        s_Thread.Finalize();

        // Gets the decode function error code.
        errorCode = s_Decoder.GetLastError();

        NN_LOG("loop[%u] decodeResult 0x%x, errorCode %d\n", loop, s_DecodeResult, errorCode);
    }

    // Finalizes the decoder object.
    FinalizeJpegDecoder();

    // If it gets all the way to here, the test is successful.
    NN_LOG("Demo finished.\n");

    // Finalize();
    while(1)
    {
    }
}
