/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     os_AlarmDemo.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/os.h>

/*!
    @addtogroup Demo Demo
    @{
*/
/*!
    @addtogroup os_Demo OS
    @{
*/
/*!
    @defgroup os_AlarmDemo Alarm
    
    Demo program for alarm-related functions.
    
    @{
*/

/*!
  @brief        Alarm handler used in AlarmDemo. Performs the following processing.

                When not canceled, outputs a string. 
                When canceled, outputs a string, and generates a cancel completion event. 

  @param[in]    pEvent      Event that indicates cancelation is complete. 
  @param[in]    cancelled   Variable that indicates whether it was canceled.

  @return  None.
*/
void SimpleAlarmHandler(nn::os::Event* pEvent, bool cancelled)
{
    if(cancelled)
    {
        NN_LOG(" Cancelled\n");
        pEvent->Signal();
    }
    else
    {
        NN_LOG(" SimpleAlarmHandler\n");
    }
}

/*!
  @brief        This is an alarm demo.

                It sets alarms and runs alarm handlers. 

  @return  None.
*/
void AlarmDemo1()
{
    NN_LOG("-- Alarm Demo1 --\n");

    nn::os::Alarm alarm;
    nn::os::Event cancelEvent(false);

    // Alarm initialization
    alarm.Initialize();

    // Sets the one-shot alarm for 1 second
    alarm.SetOneShot(nn::fnd::TimeSpan::FromSeconds(1), SimpleAlarmHandler, &cancelEvent);

    NN_LOG("SimpleAlarmHandler will be executed in 1 sec.\n");

    // Wait until alarm can be set
    while(!alarm.CanSet()){ nn::os::Thread::Yield(); }

    // Set periodic alarm (initially 100 milliseconds, then 200 milliseconds)
    alarm.SetPeriodic(nn::fnd::TimeSpan::FromMilliSeconds(100), nn::fnd::TimeSpan::FromMilliSeconds(200), SimpleAlarmHandler, &cancelEvent);

    // 1 second sleep
    NN_LOG("Sleep 1 sec.\n");
    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromSeconds(1));

    // Cancel the alarm
    alarm.Cancel();

    // Wait until cancelation has completed
    cancelEvent.Wait();

    alarm.Finalize();
}

/*!
  @brief        This is an alarm demo. 

                Initializes the alarm system and runs the alarm demo.

  @return  None.
*/
void AlarmDemo()
{
    NN_LOG("*** Alarm Demo starts. ***\n");

    nn::os::InitializeAlarmSystem();
    AlarmDemo1();

    NN_LOG("*** End of Alarm Demo ***\n");
}


/*!
    @}
*/

/*!
    @}
*/

/*!
    @}
*/

