﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_DataClient.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#include "rdt_DataClient.h"
#include "rdt_SenderUtil.h"
#include "rdt_ReceiverUtil.h"
#include "Common.h"

namespace
{
    u8* rdtReceiveData;
    size_t rdtReceiveDataSize = 0;

    static u8 rdtCommand[RDT_COMMAND_SIZE];     // For sending commands
    static u8 rdtCommandBuf[RDT_COMMAND_SIZE];  // For receiving commands

    ReceiverUtil*    m_pCommandReceiverUtil;
    SenderUtil*      m_pCommandSenderUtil;
    ReceiverUtil*      m_pDataReceiverUtil;


    bool                m_Initialized;
    s32                 m_CancelTimeoutCount = 3*60;

}

RdtClient::RdtClient()
{
    m_pDataReceiver = new nn::rdt::Receiver();
    m_pCommandSender = new nn::rdt::Sender();
    m_pCommandReceiver = new nn::rdt::Receiver();
    m_pCommandReceiverUtil = new ReceiverUtil();
    m_pCommandSenderUtil = new SenderUtil();
    m_pDataReceiverUtil = new ReceiverUtil();
    m_Initialized = false;
    m_State = STATE_CLOSE;
}

RdtClient::~RdtClient()
{
    delete m_pDataReceiver;
    delete m_pCommandSender;
    delete m_pCommandReceiver;
    delete m_pCommandReceiverUtil;
    delete m_pCommandSenderUtil;
    delete m_pDataReceiverUtil;
}

nn::Result RdtClient::Initialize(nn::rdt::ReceiverConfig& dataReceiverConfig, nn::rdt::SenderConfig& commandSenderConfig, nn::rdt::ReceiverConfig& commandReceiverConfig)
{
    nn::Result result = nn::ResultSuccess();
    m_State = STATE_CLOSE;
    if(!m_Initialized)
    {
        // Initialize for rdt::Receiver and rdt::Sender
         
        result = m_pDataReceiver->Initialize(dataReceiverConfig);
        if(result.IsFailure())
        {
            return result;
        }

        result = m_pCommandSender->Initialize(commandSenderConfig);
        if(result.IsFailure())
        {
            return result;
        }
        
        result = m_pCommandReceiver->Initialize(commandReceiverConfig);
        if(result.IsFailure())
        {
            return result;
        }
        
        m_pDataReceiverUtil->Initialize(m_pDataReceiver);
        m_pCommandSenderUtil->Initialize(m_pCommandSender);
        m_pCommandReceiverUtil->Initialize(m_pCommandReceiver);
   
        
        m_Initialized = true;
    }
    return result;
}

void RdtClient::Finalize()
{
    if(m_Initialized)
    {
        // Finalize for rdt::Receiver and rdt::Sender
        m_pDataReceiver->Finalize();
        m_pCommandReceiver->Finalize();
        m_pCommandSender->Finalize();
        m_Initialized = false;
    }
}

void RdtClient::SetReceiveBuf(u8* data, size_t size)
{
    rdtReceiveData = data;
    rdtReceiveDataSize = size;
}


nn::Result RdtClient::Start()
{
    nn::Result result = nn::ResultSuccess();

    // Set the receive buffer
    m_pCommandReceiverUtil->SetReceiveBuf(rdtCommandBuf, sizeof(rdtCommandBuf));
    m_pDataReceiverUtil->SetReceiveBuf(rdtReceiveData, rdtReceiveDataSize);

    ChangeState(STATE_WAIT);

    return result;
}

bool RdtClient::IsFinished()
{
    return (m_State == STATE_FINISHED);
}

bool RdtClient::IsCanceled()
{
    return (m_State == STATE_CANCEL_CLOSED || m_State == STATE_ERROR);
}


bool RdtClient::IsReqCommand(u8* data)
{
    // Check whether the received packet is a Request.
    if(MEMCMP(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_REQ, sizeof(RDT_DEMO_HEADER_REQ)) == 0)
    {
        return true;
    }
    return false;
}

bool RdtClient::IsAckCommand(u8* data)
{
    // Check whether the received packet is an Ack.
    if(MEMCMP(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_ACK, sizeof(RDT_DEMO_HEADER_ACK)) == 0)
    {
        return true;
    }
    return false;
}

bool RdtClient::IsCancelCommand(u8* data)
{
    // Check whether the received packet is a Cancel.
    if(MEMCMP(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_CANCEL, sizeof(RDT_DEMO_HEADER_CANCEL)) == 0)
    {
        return true;
    }
    return false;
}

void RdtClient::SetCommandReq(u8* data)
{
    MEMCPY(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_REQ, sizeof(RDT_DEMO_HEADER_REQ));
}

void RdtClient::SetCommandAck(u8* data)
{
    MEMCPY(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_ACK, sizeof(RDT_DEMO_HEADER_ACK));
}

void RdtClient::SetCommandCancel(u8* data)
{
    // Create a Cancel packet
    MEMCPY(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_CANCEL, sizeof(RDT_DEMO_HEADER_CANCEL));
}

nn::Result RdtClient::Process()
{
    nn::Result result = nn::ResultSuccess();
    
    // Call the Process function for Sender and Receiver
    
    result = m_pCommandReceiverUtil->Process();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [CLIENT] %s(%d): CommandReceiver result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    result = m_pCommandSenderUtil->Process();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [CLIENT] %s(%d): CommandSender result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    result = m_pDataReceiverUtil->Process();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [CLIENT] %s(%d): DataReceiver result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    //DEBUG_PRINTF("  [CLIENT]  DatRecv:%d ComSend:%d ComRecv:%d\n",
    //    m_pDataReceiver->GetStatus(), m_pCommandSender->GetStatus(), m_pCommandReceiver->GetStatus());
    return result;
}

nn::Result RdtClient::UpdateRdt()
{
    nn::Result result = nn::ResultSuccess();
    
    // Processes of Sender and Receiver

    result = m_pCommandReceiverUtil->Update();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [CLIENT] %s(%d): CommandReceiver result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    result = m_pCommandSenderUtil->Update();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [CLIENT] %s(%d): CommandSender result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    result = m_pDataReceiverUtil->Update();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [CLIENT] %s(%d): DataReceiver result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    //DEBUG_PRINTF("  [CLIENT]  DatRecv:%d ComSend:%d ComRecv:%d\n",
    //    m_pDataReceiver->GetStatus(), m_pCommandSender->GetStatus(), m_pCommandReceiver->GetStatus());
    return result;
}

nn::Result RdtClient::Update()
{
    nn::Result result = nn::ResultSuccess();
    
    if(IsCanceled())
    {
        DEBUG_PRINTF("  [CLIENT]  ... Canceled \n");
        return result;
    }
    
    if(m_State != STATE_CANCEL_RECEIVED && m_State != STATE_CANCEL_SEND && m_State != STATE_CANCEL_CLOSED)
    {
        if(m_pCommandReceiverUtil->IsReceiveFinished())
        {
            if(IsCancelCommand(rdtCommandBuf))
            {
                DEBUG_PRINTF("  [CLIENT]  %s: receive CANCEL command \n",__FUNCTION__);
                ChangeState(STATE_CANCEL_RECEIVED);
            }
            else
            {
                // The Client receives nothing other than Cancel, so reopen the port if Cancel not received.
                m_pCommandReceiverUtil->Open(); 
            }
        }
    }

    switch(m_State)
    {
    case STATE_CLOSE:
        break;
    case STATE_WAIT:
        {
            // Wait for sending of Request packet to complete.
            if(m_pCommandSenderUtil->IsSendFinished())
            {
                ChangeState(STATE_OPEN);
            }
        }
        break;
    case STATE_OPEN:
        {
            // Wait for reception of data to complete.
            if(m_pDataReceiverUtil->IsReceiveFinished())
            {
                ChangeState(STATE_WAITING_FINISHED);
            }
        }
        break;
    case STATE_WAITING_FINISHED:
        {
            // Wait for sending of Ack packet to complete.
            if(m_pCommandSenderUtil->IsSendFinished())
            {
                ChangeState(STATE_FINISHED);
            }
        }
        break;
    case STATE_CANCEL_SEND:
        {
            // Wait for sending of Cancel packet to complete.
            if(m_pCommandSenderUtil->IsSendFinished())
            {
                ChangeState(STATE_CANCEL_CLOSED);
                    break;
            }
            m_CancelTimeoutCount --;
            if(m_CancelTimeoutCount == 0)
            {
                m_CancelTimeoutCount = 3*60;
                ChangeState(STATE_CANCEL_CLOSED);
            }
        }
        break;
    case STATE_CANCEL_RECEIVED:
        {
            // Provide some time to close the data port after Cancel has been received.
            m_CancelTimeoutCount--;
            if(m_CancelTimeoutCount == 0)
            {
                m_CancelTimeoutCount = 2*60;
                ChangeState(STATE_CANCEL_CLOSED);
            }
        }
        break;
        
    default:
        break;
    }
    
    result = UpdateRdt();
    return result;

}

void RdtClient::Cancel()
{
    ChangeState(STATE_CANCEL_SEND);
}

void RdtClient::ForceClose()
{
    // Call the Cancel function for Sender and Receiver. (Do not send Cancel packet.)
    m_pDataReceiverUtil->Cancel();
    m_pCommandReceiverUtil->Cancel();
    m_pCommandSenderUtil->Cancel();
    ChangeState(STATE_CANCEL_CLOSED);
}

void RdtClient::initClose()
{
}

void RdtClient::initWait()
{
    // Sending of Request packet.
    SetCommandReq(rdtCommand);
    m_pCommandSenderUtil->SetSendData(rdtCommand, sizeof(rdtCommand));
    m_pCommandSenderUtil->Open();
    // Open port to receive commands.
    m_pCommandReceiverUtil->Open();

}

void RdtClient::initOpen()
{
    m_pCommandSenderUtil->Close();
    //Open port to receive data.
    m_pDataReceiverUtil->Open();
}

void RdtClient::initWaitingFinished()
{
    m_pDataReceiverUtil->Close();
    // Sending of Ack packet.
    SetCommandAck(rdtCommand);
    m_pCommandSenderUtil->SetSendData(rdtCommand, sizeof(rdtCommand));
    m_pCommandSenderUtil->Open();
}

void RdtClient::initFinished()
{
}

void RdtClient::initCancelSend()
{
    m_CancelTimeoutCount = 3*60;    // Provide a little time for processes like the sending of the Cancel packet.
    // Sending of Cancel packet
    SetCommandCancel(rdtCommand);
    m_pCommandSenderUtil->SetSendData(rdtCommand, sizeof(rdtCommand));
    m_pCommandSenderUtil->Open();
}

void RdtClient::initCancelReceived()
{
    m_CancelTimeoutCount = 2*60;    // Provide a little time for processes like Cancel.
    // Call Cancel function for Receiver that receives data.
    m_pDataReceiverUtil->Cancel();
}

void RdtClient::initCancelClosed()
{
    m_pDataReceiverUtil->Close();
    m_pCommandReceiverUtil->Close();
    m_pCommandSenderUtil->Close();
}

void RdtClient::initError()
{
    m_pDataReceiverUtil->Close();
    m_pCommandReceiverUtil->Close();
    m_pCommandSenderUtil->Close();
}

void RdtClient::initExit()
{
    m_pDataReceiverUtil->Close();
    m_pCommandReceiverUtil->Close();
    m_pCommandSenderUtil->Close();
}


void RdtClient::ChangeState(ClientState state)
{
    if(m_State == state){ return; }
    DEBUG_PRINTF("  [CLIENT] ChangeState %d --> %d\n", m_State, state);
    m_State = state;
    static void (*initFunc[STATE_MAX])() = 
    {
        &initClose,
        &initWait,
        &initOpen,
        &initWaitingFinished,
        &initFinished,
        &initCancelSend,
        &initCancelReceived,
        &initCancelClosed,
        &initError,
        &initExit
    };
    
    initFunc[m_State]();
}


