﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_DataClient.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#ifndef NN_SAMPLE_RDT_DEMO_RDT_DATA_CLIENT_H_
#define NN_SAMPLE_RDT_DEMO_RDT_DATA_CLIENT_H_
#include <nn.h>
#include <nn/rdt.h>

/* RdtClient 
    - Receiver for receiving data
    - Sender for sending commands
    - Receiver for receiving commands
    
        STATE_WAIT
            The Sender for sending commands sends REQUEST to Server.
        STATE_OPEN
            The Receiver for receiving data receives data from the Server.
        STATE_WAITING_FINISHED
            The Sender for sending commands sends ACK to Server.
        STATE_FINISHED
            In this state, Ack has finished being sent and data has finished being received

        STATE_CANCEL_SEND
            The Sender for sending commands sends a CANCEL command.
        STATE_CANCEL_RECEIVED
            The Receiver for receiving commands receives the CANCEL command.
            The Sender for sending data is closed by the Cancel function.
        STATE_CANCEL_CLOSED
            The cancel process ends and the Sender and Receiver are stopped.

        Note: When the Sender::Cancel function and the Receiver::Cancel function are called, a reset packet is sent and the Sender and Receiver themselves transition to the CLOSED state, but the Cancel may not get conveyed to the communication partner, due to lost packets or the state of the partner.

           
           In this demo, the port for sending commands is used for Cancel notifications, but even this is not for certain, so communication is interrupted on a timeout in the upper level.
           

*/

class RdtClient
{
public:
    RdtClient();
    ~RdtClient();

    nn::Result Initialize(nn::rdt::ReceiverConfig& dataReceiverConfig, nn::rdt::SenderConfig& commandSenderConfig, nn::rdt::ReceiverConfig& commandReceiverConfig);
    void Finalize();
    
    /* Set a buffer for storing the data received by the Client via the data port. */
    void    SetReceiveBuf(u8* data, size_t size);
    
    /* Start Client processes. Transition to STATE_WAIT. */
    nn::Result  Start();

    /* Stop Client processes. Transition to STATE_CANCEL_SEND. */
    void        Cancel();
    void        ForceClose();

    /* Call the Process functions for Sender and Receiver. */
    nn::Result  Process();

    /* Perform the Sender and Receiver processes and transition the state. */
    nn::Result  Update();
    
    /* True when stopped due to a cancel or error. */
    bool IsFinished();
    bool IsCanceled();

private:
    enum ClientState
    {
        STATE_CLOSE,
        STATE_WAIT,
        STATE_OPEN,
        STATE_WAITING_FINISHED,
        STATE_FINISHED,
        STATE_CANCEL_SEND,
        STATE_CANCEL_RECEIVED,
        STATE_CANCEL_CLOSED,
        STATE_ERROR,
        STATE_EXIT,
        
        STATE_MAX
    };
    bool IsReqCommand(u8* data);
    bool IsAckCommand(u8* data);
    bool IsCancelCommand(u8* data);
    static void SetCommandReq(u8* data);
    static void SetCommandAck(u8* data);
    static void SetCommandCancel(u8* data);

    static void initClose();
    static void initWait();
    static void initOpen();
    static void initWaitingFinished();
    static void initFinished();
    static void initCancelSend();
    static void initCancelReceived();
    static void initCancelClosed();
    static void initError();
    static void initExit();
    void ChangeState(enum ClientState state);
    
    nn::Result  UpdateRdt();

    nn::rdt::Receiver*  m_pDataReceiver;
    nn::rdt::Sender*    m_pCommandSender;
    nn::rdt::Receiver*  m_pCommandReceiver;

    enum ClientState    m_State;


    NN_PADDING3;


};

#endif

