﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_DataServer.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#include "rdt_DataServer.h"
#include "rdt_SenderUtil.h"
#include "rdt_ReceiverUtil.h"
#include "Common.h"

namespace
{
    u8* rdtSendData;
    size_t rdtSendDataSize = 0;

    static u8 rdtCommand[RDT_COMMAND_SIZE];
    static u8 rdtCommandBuf[RDT_COMMAND_SIZE];

    ReceiverUtil*    m_pCommandReceiverUtil;
    SenderUtil*      m_pCommandSenderUtil;
    SenderUtil*      m_pDataSenderUtil;


    bool                m_Initialized;
    s32                 m_CancelTimeoutCount = 3*60;
}

RdtServer::RdtServer()
{
    m_pCommandSender = new nn::rdt::Sender();
    m_pCommandReceiver = new nn::rdt::Receiver();
    m_pDataSender = new nn::rdt::Sender();
    m_pCommandReceiverUtil = new ReceiverUtil();
    m_pCommandSenderUtil = new SenderUtil();
    m_pDataSenderUtil = new SenderUtil();
    m_Initialized = false;
    m_State = STATE_CLOSE;
}

RdtServer::~RdtServer()
{
    delete m_pCommandReceiver;
    delete m_pCommandSender;
    delete m_pDataSender;
    delete m_pCommandReceiverUtil;
    delete m_pCommandSenderUtil;
    delete m_pDataSenderUtil;
}

nn::Result RdtServer::Initialize(nn::rdt::SenderConfig& dataSenderConfig, nn::rdt::SenderConfig& commandSenderConfig, nn::rdt::ReceiverConfig& commandReceiverConfig)
{
    nn::Result result = nn::ResultSuccess();
    m_State = STATE_CLOSE;
    if(!m_Initialized)
    {
        // Initialize for rdt::Receiver and rdt::Sender
         
        result = m_pDataSender->Initialize(dataSenderConfig);
        if(result.IsFailure())
        {
            return result;
        }

        result = m_pCommandSender->Initialize(commandSenderConfig);
        if(result.IsFailure())
        {
            return result;
        }
        
        result = m_pCommandReceiver->Initialize(commandReceiverConfig);
        if(result.IsFailure())
        {
            return result;
        }
   
        m_pDataSenderUtil->Initialize(m_pDataSender);
        m_pCommandSenderUtil->Initialize(m_pCommandSender);
        m_pCommandReceiverUtil->Initialize(m_pCommandReceiver);
   
        
        m_Initialized = true;
    }
    return result;
}

void RdtServer::Finalize()
{
    if(m_Initialized)
    {
        // Finalize for rdt::Receiver and rdt::Sender
        m_pDataSender->Finalize();
        m_pCommandReceiver->Finalize();
        m_pCommandSender->Finalize();
        m_Initialized = false;
    }
}

nn::Result RdtServer::Start()
{
    nn::Result result = nn::ResultSuccess();

    // Set send data
    m_pCommandReceiverUtil->SetReceiveBuf(rdtCommandBuf, sizeof(rdtCommandBuf));
    m_pDataSenderUtil->SetSendData(rdtSendData, rdtSendDataSize);

    ChangeState(STATE_WAIT);

    return result;
}

void RdtServer::SetSendData(u8* data, size_t size)
{
    rdtSendData = data; 
    rdtSendDataSize = size;
}

bool RdtServer::IsFinished()
{
    return (m_State == STATE_FINISHED);
}

bool RdtServer::IsCanceled()
{
    return (m_State == STATE_CANCEL_CLOSED || m_State == STATE_ERROR);
}

bool RdtServer::IsReqCommand(u8* data)
{
    // Check whether the received packet is a Request.
    if(MEMCMP(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_REQ, sizeof(RDT_DEMO_HEADER_REQ)) == 0)
    {
        return true;
    }
    return false;
}

bool RdtServer::IsAckCommand(u8* data)
{
    // Check whether the received packet is an Ack.
    if(MEMCMP(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_ACK, sizeof(RDT_DEMO_HEADER_ACK)) == 0)
    {
        return true;
    }
    return false;
}

bool RdtServer::IsCancelCommand(u8* data)
{
    // Check whether the received packet is a Cancel.
    if(MEMCMP(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_CANCEL, sizeof(RDT_DEMO_HEADER_CANCEL)) == 0)
    {
        return true;
    }
    return false;
}

void RdtServer::SetCommandCancel(u8* data)
{
    // Create a Cancel packet
    MEMCPY(reinterpret_cast<u16*>(data), &RDT_DEMO_HEADER_CANCEL, sizeof(RDT_DEMO_HEADER_CANCEL));
}

nn::Result RdtServer::Process()
{
    nn::Result result = nn::ResultSuccess();

    // Call the Process function for Sender and Receiver
    
    result = m_pCommandReceiverUtil->Process();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [SERVER] %s(%d): CommandReceiver result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());

        ChangeState(STATE_ERROR);
        return result;
    }
    result = m_pCommandSenderUtil->Process();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [SERVER] %s(%d): CommandSender result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    result = m_pDataSenderUtil->Process();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [SERVER] %s(%d): DataSender result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    
    //DEBUG_PRINTF("  [SERVER]  DatSend:%d ComSend:%d ComRecv:%d\n",
    //    m_pDataSender->GetStatus(), m_pCommandSender->GetStatus(), m_pCommandReceiver->GetStatus());
    return result;
}

nn::Result RdtServer::UpdateRdt()
{
    nn::Result result = nn::ResultSuccess();

    // Processes of Sender and Receiver

    result = m_pCommandReceiverUtil->Update();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [SERVER] %s(%d): CommandReceiver result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    result = m_pCommandSenderUtil->Update();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [SERVER] %s(%d): CommandSender result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    result = m_pDataSenderUtil->Update();
    if(result.IsFailure())
    {
        DEBUG_PRINTF("  [SERVER] %s(%d): DataSender result:0x%08x\n", __FUNCTION__,__LINE__,result.GetPrintableBits());
        ChangeState(STATE_ERROR);
        return result;
    }
    
    return result;
}

nn::Result RdtServer::Update()
{
    nn::Result result = nn::ResultSuccess();
    if(IsCanceled())
    {
        DEBUG_PRINTF("  [SERVER]  ... Canceled \n");
        return result;
    }

    
    if(m_State != STATE_CANCEL_RECEIVED && m_State != STATE_CANCEL_SEND && m_State != STATE_CANCEL_CLOSED)
    {
        if(m_pCommandReceiverUtil->IsReceiveFinished())
        {
            if(IsCancelCommand(rdtCommandBuf))
            {
                DEBUG_PRINTF("  [SERVER]  %s: receive CANCEL command \n",__FUNCTION__);
                ChangeState(STATE_CANCEL_RECEIVED);
            }
        }
    }

    switch (m_State)
    {
    case STATE_WAIT:
        {
            // Wait for completion of reception of Request packet from Client.
            if(m_pCommandReceiverUtil->IsReceiveFinished())
            {
                if(IsReqCommand(rdtCommandBuf))
                {
                    ChangeState(STATE_OPEN);
                }
                else
                {
                    m_pCommandReceiverUtil->Open();
                }
            }
        }
        break;
    case STATE_OPEN:
        {
            // Wait for completion of sending of data.
            if(m_pDataSenderUtil->IsSendFinished())
            {
                ChangeState(STATE_WAITING_FINISHED);
            }
        }
        break;  
    case STATE_WAITING_FINISHED:
        {
            // Wait for completion of reception of Ack packet from Client.
            
            if(m_pCommandReceiverUtil->IsReceiveFinished())
            {
                if(IsAckCommand(rdtCommandBuf))
                {
                    //ChangeState(STATE_WAIT);      // After sending, revert to wait state.
                    ChangeState(STATE_FINISHED);    // End after sending once.
                }
                else
                {
                    m_pCommandReceiverUtil->Open();
                }
            }
        }
        break;
    case STATE_CANCEL_SEND:
        // Wait for sending of Cancel packet to complete.
        {
            if(m_pCommandSenderUtil->IsSendFinished())
            {
                ChangeState(STATE_CANCEL_CLOSED);
                break;
            }
            m_CancelTimeoutCount --;
            if(m_CancelTimeoutCount == 0)
            {
                m_CancelTimeoutCount = 3*60;
                ChangeState(STATE_CANCEL_CLOSED);
            }
        }
        break;
    case STATE_CANCEL_RECEIVED:
        {
            // Provide some time to close the data port after Cancel has been received.
            m_CancelTimeoutCount--;
            if(m_CancelTimeoutCount == 0)
            {
            m_CancelTimeoutCount = 2*60;
                ChangeState(STATE_CANCEL_CLOSED);
            }
        }
        break;
        
    default:
        break;

    }
    
    result = UpdateRdt();
    
    return result;
}

void RdtServer::Cancel()
{
    ChangeState(STATE_CANCEL_SEND);
}

void RdtServer::ForceClose()
{
    // Call the Cancel function for Sender and Receiver. (Do not send Cancel packet.)
    m_pDataSenderUtil->Cancel();
    m_pCommandReceiverUtil->Cancel();
    m_pCommandSenderUtil->Cancel();
    ChangeState(STATE_CANCEL_CLOSED);
}


void RdtServer::initClose()
{
}

void RdtServer::initWait()
{
    // Open port to receive commands.
    m_pCommandReceiverUtil->Open();
}

void RdtServer::initOpen()
{
    m_pCommandReceiverUtil->Open();
    // Sending Data
    m_pDataSenderUtil->Open();
}

void RdtServer::initWaitingFinished()
{
    m_pDataSenderUtil->Close();
}

void RdtServer::initFinished()
{
    // Open port to receive commands.
    m_pCommandReceiverUtil->Open();
}


void RdtServer::initCancelSend()
{
    m_CancelTimeoutCount = 3*60;    // Provide a little time for processes like the sending of the Cancel packet.
    // Sending of Cancel packet
    SetCommandCancel(rdtCommand);
    m_pCommandSenderUtil->SetSendData(rdtCommand, sizeof(rdtCommand));
    m_pCommandSenderUtil->Open();
}

void RdtServer::initCancelReceived()
{
    m_CancelTimeoutCount = 2*60;    // Provide a little time for processes like Cancel.
    // Cancel function of Sender for sending data.
    m_pDataSenderUtil->Cancel();
}

void RdtServer::initCancelClosed()
{
    
}


void RdtServer::initError()
{
    m_pDataSenderUtil->Close();
    m_pCommandReceiverUtil->Close();
    m_pCommandSenderUtil->Close();
}

void RdtServer::initExit()
{
    m_pDataSenderUtil->Close();
    m_pCommandReceiverUtil->Close();
    m_pCommandSenderUtil->Close();
}

void RdtServer::ChangeState(ServerState state)
{
    if(m_State == state){ return; }
    DEBUG_PRINTF("  [SERVER]  ChangeState %d --> %d\n", m_State, state);
    m_State = state;
    static void (*initFunc[STATE_MAX])() =
    {
        &initClose,
        &initWait,
        &initOpen,
        &initWaitingFinished,
        &initFinished,
        &initCancelSend,
        &initCancelReceived,
        &initCancelClosed,
        &initError,
        &initExit,
    };
    
    initFunc[m_State]();
}


