﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_Main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

/*
    RDT Data Transmission Demo
    
    In communication in a single session between an RDT Sender and Receiver, there is no way for the Sender to know for certain that the Receiver got the sent data.
    
    (Even though the Sender should have completed sending, it is possible for the Receiver to cancel just before then.)
    
    In this demo, ports have been prepared for the sending and receiving of data, and for the sending and receiving of commands. So the sending side can learn whether data has been sent successfully by following a procedure with these steps: Client sends request → Server sends data → Client sends Ack.
    
   

    If these series of steps is not successful, it means that communication has failed at some point.
    It is also possible for communications to bog down due to a misalignment of the Server/Client sequence, caused by such factors as lost and canceled packets.
    In particular, when Sender/Receiver is STATE_CLOSED, the reset packet is not received when the communication partner aborts using the Cancel function so there is no way of knowing that the partner has stopped communication.
    
    A timeout has been set in this demo, so if the series of steps does not complete within that time you know that communication has failed.

*/


#include <nn.h>
#include <nn/Result.h>
#include <nn/applet.h>
#include <nn/uds.h>
#include <nn/rdt.h>

#include <nn/types.h>
#include <nn/hid.h>
#include "Common.h"
#include "rdt_UdsUtil.h"
#include "rdt_DataClient.h"
#include "rdt_DataServer.h"


enum RdtDemoState
{
    STATE_INIT,
    STATE_UDS_SELECT,
    STATE_UDS_MASTER_CONNECTING,
    STATE_UDS_CLIENT_SCANNING,
    STATE_UDS_CLIENT_CONNECTING,
    STATE_SELECT_SERVER_CLIENT,
    STATE_SERVER,
    STATE_CLIENT,
    STATE_CLIENT_RECEIVING,
    STATE_EXIT,
    
    STATE_MAX
};

namespace
{
    // Buffer that will be used by RDT Sender and Receiver.
    static u8 rdtServerDataSendWorkBuf[nn::rdt::Sender::SENDER_WORKBUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtServerCommandRecvWorkBuf[nn::rdt::Receiver::RECEIVER_WORKBUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtServerCommandSendWorkBuf[nn::rdt::Sender::SENDER_WORKBUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtClientDataRecvWorkBuf[nn::rdt::Receiver::RECEIVER_WORKBUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtClientCommandSendWorkBuf[nn::rdt::Sender::SENDER_WORKBUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtClientCommandRecvWorkBuf[nn::rdt::Receiver::RECEIVER_WORKBUF_SIZE] NN_ATTRIBUTE_ALIGN(8);

    static u8 rdtServerDataSendBuf[RDT_BUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtServerCommandSendBuf[RDT_BUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtServerCommandRecvBuf[RDT_BUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtClientDataRecvBuf[RDT_BUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtClientCommandSendBuf[RDT_BUF_SIZE] NN_ATTRIBUTE_ALIGN(8);
    static u8 rdtClientCommandRecvBuf[RDT_BUF_SIZE] NN_ATTRIBUTE_ALIGN(8);

    // Buffer for storing the data received by the Client.
    static u8 clientReceiveDatabuf[RDT_SENDDATA_SIZE];
    
    // The data that will be sent by the Server.
    static u8 serverSendDatabuf[RDT_SENDDATA_SIZE];

    RdtDemoState rdtDemoState = STATE_INIT;

    static s32 s_TimeoutCount = -1;
}

nn::Result SetupServer(RdtServer* server, u16 peerId)
{
    nn::Result result;
    // For sending data
    nn::rdt::SenderConfig cfg_d_s;
    cfg_d_s.pWorkBuf    = rdtServerDataSendWorkBuf;
    cfg_d_s.pSendBuf    = rdtServerDataSendBuf;
    cfg_d_s.sendBufSize = sizeof(rdtServerDataSendBuf);
    cfg_d_s.nodeId      = peerId;  
    cfg_d_s.port        = RDT_DATA_PORT_NUMBER;
    
    // For sending commands
    nn::rdt::SenderConfig cfg_c_s;
    cfg_c_s.pWorkBuf    = rdtServerCommandSendWorkBuf;
    cfg_c_s.pSendBuf    = rdtServerCommandSendBuf;
    cfg_c_s.sendBufSize = sizeof(rdtServerCommandSendBuf);
    cfg_c_s.nodeId      = peerId;  
    cfg_c_s.port        = RDT_COMMAND_PORT_NUMBER_S2C;
    
    // For receiving commands
    nn::rdt::ReceiverConfig cfg_c_r;
    cfg_c_r.pWorkBuf    = rdtServerCommandRecvWorkBuf;
    cfg_c_r.pRecvBuf    = rdtServerCommandRecvBuf;
    cfg_c_r.recvBufSize = sizeof(rdtServerCommandRecvBuf);
    cfg_c_r.nodeId      = peerId;  
    cfg_c_r.port        = RDT_COMMAND_PORT_NUMBER_C2S;

    
    result = server->Initialize(cfg_d_s, cfg_c_s, cfg_c_r);
    if(result.IsSuccess())
    {
        DEBUG_PRINTF("[RDT_DEMO]server.SetReceiveBuf %d byte!!\n", sizeof(serverSendDatabuf));
        server->SetSendData(serverSendDatabuf, sizeof(serverSendDatabuf));
    }
    else
    {
        DEBUG_PRINTF("[RDT_DEMO]server.Initialize Failed!!\n");
        return result;
    }
    return result;
}

nn::Result SetupClient(RdtClient* client, u16 peerId)
{
    nn::Result result;
    // For receiving data
    nn::rdt::ReceiverConfig cfg_d_r;
    cfg_d_r.pWorkBuf    = rdtClientDataRecvWorkBuf;
    cfg_d_r.pRecvBuf    = rdtClientDataRecvBuf;
    cfg_d_r.recvBufSize = sizeof(rdtClientDataRecvBuf);
    cfg_d_r.nodeId      = peerId;  
    cfg_d_r.port        = RDT_DATA_PORT_NUMBER;
    
    // For sending commands
    nn::rdt::SenderConfig cfg_c_s;
    cfg_c_s.pWorkBuf    = rdtClientCommandSendWorkBuf;
    cfg_c_s.pSendBuf    = rdtClientCommandSendBuf;
    cfg_c_s.sendBufSize = sizeof(rdtClientCommandSendBuf);
    cfg_c_s.nodeId      = peerId;  
    cfg_c_s.port        = RDT_COMMAND_PORT_NUMBER_C2S;
    
    // For receiving commands
    nn::rdt::ReceiverConfig cfg_c_r;
    cfg_c_r.pWorkBuf    = rdtClientCommandRecvWorkBuf;
    cfg_c_r.pRecvBuf    = rdtClientCommandRecvBuf;
    cfg_c_r.recvBufSize = sizeof(rdtClientCommandRecvBuf);
    cfg_c_r.nodeId      = peerId;  
    cfg_c_r.port        = RDT_COMMAND_PORT_NUMBER_S2C;

    
    result = client->Initialize(cfg_d_r, cfg_c_s, cfg_c_r);
    if(result.IsSuccess())
    {
        DEBUG_PRINTF("[RDT_DEMO]client.SetReceiveBuf %d byte!!\n", sizeof(clientReceiveDatabuf));
        client->SetReceiveBuf(clientReceiveDatabuf, sizeof(clientReceiveDatabuf));
    }
    else
    {
        DEBUG_PRINTF("[RDT_DEMO]client.Initialize Failed!!\n");
        return result;
    }
    return result;
}

bool IsUdsDisconnected()
{
    return (GetPeerNodeId() == 0);
}

void SetTimeout(u32 timeout_frame)
{
    s_TimeoutCount = timeout_frame;
}

bool CheckTimeout()
{
    if(s_TimeoutCount > 0) s_TimeoutCount-- ;
    return (s_TimeoutCount == 0) ;
}


#define MODE_SELECT_STR     "<MODE SELECT> Press A or B button. (A: Server  B: Client  Y: Exit)\n"

void nnMain( void )
{
    NN_LOG("START\n");

    nn::Result result;

    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();


   // Initialize hid
    result = nn::hid::Initialize();
    NN_UTIL_PANIC_IF_FAILED(result);

    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;
    bool ret = false;

    RdtClient client;
    RdtServer server;
    u16 peerId = 0;

    bool loop = true;
    
    while(loop)
    {
        padReader.ReadLatest(&padStatus);
        
        switch(rdtDemoState)
        {
        case STATE_INIT:
            {
                ret = setupUds();
                NN_ASSERT(ret);

                rdtDemoState = STATE_UDS_SELECT;
#ifdef SUPPORT_UDS_CLIENT
                DEBUG_PRINTF("[RDT_DEMO] Select UDS (A: Master  B: Client)\n");
#endif
            }
            break;
            
        case STATE_UDS_SELECT:
            {
#ifdef SUPPORT_UDS_CLIENT
                // A for UDS Master, and B for UDS Client.

                if(PUSHED_A(padStatus))
                {
                    DEBUG_PRINTF("[RDT_DEMO]  -> Select: UDS Master!\n");
                    ret = createNetwork();
                    if(ret)
                    {
                        StartMonitorThread();
                        rdtDemoState = STATE_UDS_MASTER_CONNECTING;
                    }
                    NN_ASSERT(ret);
                }
                else if(PUSHED_B(padStatus))
                {
                    DEBUG_PRINTF("[RDT_DEMO]  -> Select: UDS Client!\n");
                    rdtDemoState = STATE_UDS_CLIENT_SCANNING;
                }
                else if(PUSHED_Y(padStatus))
                {
                    loop = false;
                    DEBUG_PRINTF("[RDT_DEMO]  -> UDS: Disconnect!\n");
                }
#else
                    DEBUG_PRINTF("[RDT_DEMO]  Start: UDS Master!\n");
                    ret = createNetwork();
                    if(ret)
                    {
                        StartMonitorThread();
                        rdtDemoState = STATE_UDS_MASTER_CONNECTING;
                    }
#endif
            }
            break;
        case STATE_UDS_MASTER_CONNECTING:
            {
                // Becomes the UDS Master and waits for connection.

                peerId = GetPeerNodeId();
                if(peerId != 0)
                {
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
            }
            break;
        case STATE_UDS_CLIENT_SCANNING:
            {
#ifdef SUPPORT_UDS_CLIENT
                // Becomes the UDS Client and scans and connects.

                if(connectNetwork())
                {
                    // Start the UDS monitoring  thread.
                    StartMonitorThread();
                    rdtDemoState = STATE_UDS_CLIENT_CONNECTING;
                }
#endif
            }
            break;              
        case STATE_UDS_CLIENT_CONNECTING:
            {
#ifdef SUPPORT_UDS_CLIENT
                // Wait for connection with UDS Master.

                peerId = GetPeerNodeId();
                if(peerId != 0)
                {
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
#endif
            }
            break;
        case STATE_SELECT_SERVER_CLIENT:
            {
                // Select SERVER/CLIENT.
                
                if(IsUdsDisconnected())     // UDS connection monitoring.
                {
                    DEBUG_PRINTF("[RDT_DEMO] UDS: Disconnected\n");
                    loop = false;
                    break;
                }

                if(PUSHED_A(padStatus))
                {
                    // Select SERVER.
                    DEBUG_PRINTF("[RDT_DEMO]  -> Select: Server! (X: Cancel)\n");
                    result = SetupServer(&server, peerId);
                    if(result.IsFailure())
                    {
                        break;
                    }

                    result = server.Start();    // Start SERVER processes.
                    if(result.IsFailure())
                    {
                        break;
                    }
                    
                    SetTimeout(15*60);  // Timeout after 15 seconds including wait time (900 frames).
                                        // This is enough time for transferring 512KB of data plus the exchange of control packets.
                    rdtDemoState = STATE_SERVER;
                }
                else if(PUSHED_B(padStatus))
                {
                    // Select CLIENT
                    DEBUG_PRINTF("[RDT_DEMO]  -> Select: Client! (X: Cancel)\n");
                    result = SetupClient(&client, peerId);
                    if(result.IsFailure())
                    {
                        break;
                    }
                    rdtDemoState = STATE_CLIENT;
                }
                else if(PUSHED_Y(padStatus))
                {
                    loop = false;
                    DEBUG_PRINTF("[RDT_DEMO]  -> UDS: Disconnect!\n");
                    break;
                }
            }
            break;
        case STATE_SERVER:
            {
                if(IsUdsDisconnected())     // UDS connection monitoring.
                {
                    DEBUG_PRINTF("[RDT_DEMO] UDS: Disconnected\n");
                    server.Finalize();
                    loop = false;
                    break;
                }

                if(CheckTimeout())          // Timeout monitoring.
                {
                    server.Cancel();
                    SetTimeout(15*60);
                    DEBUG_PRINTF("[RDT_DEMO]  -> Server: Timeout Cancel !\n");
                }

                if(PUSHED_X(padStatus))
                {
                    server.Cancel();
                    DEBUG_PRINTF("[RDT_DEMO]  -> Server: Cancel !\n");
                }
                if(PUSHED_R(padStatus))
                {
                    server.ForceClose();
                    DEBUG_PRINTF("[RDT_DEMO]  -> Server: ForceClose !\n");
                }

                
                result = server.Update();      // RDT actual processing.
                if(result.IsFailure())
                {
                    // RDT processing error.
                    DEBUG_PRINTF("[RDT_DEMO]SERVER ERROR... \n");
                    server.Cancel();
                    server.Finalize();
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
                
                result = server.Process();      // RDT actual processing.
                if(result.IsFailure())
                {
                    // RDT processing error.
                    DEBUG_PRINTF("[RDT_DEMO]SERVER ERROR... \n");
                    server.Cancel();
                    server.Finalize();
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
                if(server.IsFinished())
                {
                    // Send completed!
                   server.Finalize();
                    DEBUG_PRINTF("[RDT_DEMO]RDT DEMO - SERVER - Send Data Success.\n");
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }

                if(server.IsCanceled())
                {
                    server.Finalize();
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }

            }
            break;
        case STATE_CLIENT:
            {
                if(IsUdsDisconnected())     // UDS connection monitoring.
                {
                    DEBUG_PRINTF("[RDT_DEMO] UDS: Disconnected\n");
                    client.Finalize();
                    loop = false;
                    break;
                }

                result = client.Update();      // RDT processing
                if(result.IsFailure())
                {
                    // RDT processing error.
                    DEBUG_PRINTF("[RDT_DEMO]CLIENT ERROR... \n");
                    client.Cancel();
                    client.Finalize();
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
                
                result = client.Process();      // RDT actual processing.
                if(result.IsFailure())
                {
                    // RDT processing error.
                    DEBUG_PRINTF("[RDT_DEMO]CLIENT ERROR... \n");
                    client.Cancel();
                    client.Finalize();
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
                
                //if(PUSHED_A(padStatus))
                {
                    DEBUG_PRINTF("[RDT_DEMO]  -> ReceiveRequest!\n");
                    result = client.Start();    // Start CLIENT processes.
                    if(result.IsFailure())
                    {
                        break;
                    }
                    SetTimeout(10*60);  // Timeout after 10 seconds (600 frames).
                    rdtDemoState = STATE_CLIENT_RECEIVING;
                }
            }
            break;
        case STATE_CLIENT_RECEIVING:
            {
                if(IsUdsDisconnected())     // UDS connection monitoring.
                {
                    DEBUG_PRINTF("[RDT_DEMO] UDS: Disconnected\n");
                    client.Finalize();
                    loop = false;
                    break;
                }
                
                if(CheckTimeout())          // Timeout monitoring.
                {
                    client.Cancel();
                    SetTimeout(10*60);
                    DEBUG_PRINTF("[RDT_DEMO]  -> Client: Timeout Cancel !\n");
                }


                if(PUSHED_X(padStatus))
                {
                    client.Cancel();
                    DEBUG_PRINTF("[RDT_DEMO]  -> Client: Cancel !\n");
                }

                if(PUSHED_R(padStatus))
                {
                    client.ForceClose();
                    DEBUG_PRINTF("[RDT_DEMO]  -> Client: ForceClose !\n");
                }

                result = client.Update();      // RDT processing
                if(result.IsFailure())
                {
                    // RDT processing error.
                    DEBUG_PRINTF("[RDT_DEMO]CLIENT ERROR... \n");
                    client.Cancel();
                    client.Finalize();
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
                
                result = client.Process();      // RDT actual processing.
                if(result.IsFailure())
                {
                    // RDT processing error.
                    DEBUG_PRINTF("[RDT_DEMO]CLIENT ERROR... \n");
                    client.Cancel();
                    client.Finalize();
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
                
                if(client.IsFinished())
                {
                    // Receive completed!
                    client.Finalize();
                    DEBUG_PRINTF("[RDT_DEMO]RDT DEMO - CLIENT - Receive Data Success.\n");
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
                if(client.IsCanceled())
                {
                    client.Finalize();
                    DEBUG_PRINTF(MODE_SELECT_STR);
                    rdtDemoState = STATE_SELECT_SERVER_CLIENT;
                }
            }
            break;
            
        case STATE_EXIT:
            break;
        }
        
        MSLEEP(16);     // Assume 1 frame.
    }
    DEBUG_PRINTF("[RDT_DEMO]Exit Loop --> Finalize\n");
    StopMonitorThread();
    cleanupUds();
    DEBUG_PRINTF("[RDT_DEMO]DEMO END\n");
    
    while(1)
    {
    }
}

