﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_ReceiverUtil.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#include "rdt_ReceiverUtil.h"
#include "Common.h"

namespace
{
    bool    m_IsInitialized = false;
}


void ReceiverUtil::Initialize(nn::rdt::Receiver* receiver)
{
    m_pReceiver = receiver;
    m_ReceiverUtilState = RECEIVER_CLOSE;
    m_IsInitialized = true;
}

void ReceiverUtil::SetReceiveBuf(u8* data, size_t size)
{
    m_recvData = data;
    m_recvDataSize = size;
}

void ReceiverUtil::Open(s32 timeout)
{
    m_ReceiverUtilState = RECEIVER_OPENING;
    m_Timeout = timeout;
}

void ReceiverUtil::Close()
{
    m_ReceiverUtilState = RECEIVER_CLOSE;
}

bool ReceiverUtil::IsReceiveFinished()
{
    return (m_ReceiverUtilState == RECEIVER_FINISHED);
}

bool ReceiverUtil::IsClosed()
{
    return (m_ReceiverUtilState == RECEIVER_CLOSE);
}

nn::Result ReceiverUtil::Process()
{
    if(!m_IsInitialized)
    {
        return nn::rdt::ResultNotInitialized();
    }
    
    nn::Result result;
    for(int i=0; i<REPEAT_PROCESS_FUNCTION; i++)
    {
        result = m_pReceiver->Process();
        if(result.IsFailure())
        {
            return result;
        }
    }
    return nn::ResultSuccess();
}


nn::Result ReceiverUtil::Update()
{
    if(!m_IsInitialized)
    {
        return nn::rdt::ResultNotInitialized();
    }
    
    nn::Result result = nn::ResultSuccess();
    bool finished, success;
    switch(m_ReceiverUtilState)
    {
        case RECEIVER_CLOSE:
            if(m_pReceiver->GetStatus() != nn::rdt::RECEIVER_STATE_CLOSED)
            {
                DEBUG_PRINTF("    [Receiver] CloseRequest -> Cancel()\n");
                m_pReceiver->Cancel();
            }
            break;
        case RECEIVER_OPENING:
            result = Ready( &finished, &success, m_Timeout);
            if(result.IsSuccess())
            {
                if(finished)
                {
                    if(success) { m_ReceiverUtilState = RECEIVER_OPENED; }
                    else        { m_ReceiverUtilState = RECEIVER_CLOSE; }
                }
            }
            break;

        case RECEIVER_OPENED:
            result = Receive(m_recvData, m_recvDataSize, &finished, &success, m_Timeout);
            if(result.IsSuccess())
            {
                if(finished)
                {
                    if(success) { m_ReceiverUtilState = RECEIVER_WAITING_FINISHED; }
                    else        { m_ReceiverUtilState = RECEIVER_CLOSE; }
                }
            }
            break;
        
        case RECEIVER_WAITING_FINISHED:
            result = Finish(&finished, &success, m_Timeout);
            if(result.IsSuccess())
            {
                if(finished)
                {
                    if(success) { m_ReceiverUtilState = RECEIVER_FINISHED; }
                    else        { m_ReceiverUtilState = RECEIVER_CLOSE; }
                }
            }
            break;
        default:
            break;
    }
    if(result.IsFailure())
    {
        DEBUG_PRINTF("    [Receiver] %s: Failed state: %d  result: 0x%08x \n",__FUNCTION__, m_ReceiverUtilState, result.GetPrintableBits());
        return result;
    }

    return result;
}


nn::Result ReceiverUtil::Ready(bool* finished, bool* success, s32 timeout)
{
    nn::Result result = nn::ResultSuccess();
    static s32 loop = 0;
    *finished = false;
    *success = false;
    nn::rdt::ReceiverState state = m_pReceiver->GetStatus();
    if(state==nn::rdt::RECEIVER_STATE_CLOSED)
    {
        result = m_pReceiver->Wait();
        if( result.IsSuccess() )
        {
            m_TotalRecv = 0;
            *success = true;
        }
        loop = 0;
    }
    else if(state==nn::rdt::RECEIVER_STATE_OPENED)
    {
        MEMSET(m_recvData, 0, m_recvDataSize);
        *finished = true;
        *success = true;
        loop = 0;
    }
    else if(timeout > 0)
    {
        loop++;
        if(loop > timeout)  // TIMEOUT
        {
            DEBUG_PRINTF("    [Receiver] %s TIMEOUT CANCEL\n", __FUNCTION__);
            m_pReceiver->Cancel();
            loop = 0;
            *finished = true;
            *success = false;
        }
    }

    return result;
}

nn::Result ReceiverUtil::Receive(u8* dataBuf, size_t dataBufSize, bool* finished, bool* success, s32 timeout)
{
    size_t recvedSize = 0;
    static s32 loop = 0;
    
    *finished = false;
    *success = false;
    nn::Result result = m_pReceiver->Receive(dataBuf+m_TotalRecv, &recvedSize, dataBufSize - m_TotalRecv);
    if(result.IsSuccess())
    {
        m_TotalRecv += recvedSize;
        if(recvedSize != 0)
        {
            DEBUG_PRINTF("    [Receiver]Receiver::Receive %d bytes %d/%d\n", recvedSize, m_TotalRecv, dataBufSize);
        }
        // Determine completion of reception.
        if(m_pReceiver->GetStatus()==nn::rdt::RECEIVER_STATE_FINISHED)
        {
            if(recvedSize == 0)
            {
                DEBUG_PRINTF("    [Receiver]Receive Finish! %d bytes \n", m_TotalRecv);
                result = m_pReceiver->Close();
                if(result.IsSuccess())
                {
                    *finished = true;
                    *success = true;
                }
            }
        }
        else if(m_pReceiver->GetStatus()==nn::rdt::RECEIVER_STATE_OPENED && timeout > 0)
        {
            loop = (recvedSize == 0) ? loop + 1 : 0 ;
            
            if(loop > timeout)  // TIMEOUT
            {
                DEBUG_PRINTF("    [Receiver] %s TIMEOUT CANCEL\n", __FUNCTION__);
                m_pReceiver->Cancel();
                loop = 0;
                *finished = true;
                *success = false;
            }
        }
    }

    return result;

}



nn::Result ReceiverUtil::Finish(bool* finished, bool* success, s32 timeout)
{
    nn::Result result = nn::ResultSuccess();
    static s32 loop = 0;
    
    *finished = false;
    *success = false;
    if(m_pReceiver->GetStatus() == nn::rdt::RECEIVER_STATE_CLOSED)
    {
        *finished = true;
        *success = true;
        loop = 0;
    }
    else if(timeout > 0)
    {
        loop++;
        if(loop > timeout)  // TIMEOUT
        {
            DEBUG_PRINTF("    [Receiver] %s TIMEOUT CANCEL\n", __FUNCTION__);
            m_pReceiver->Cancel();
            loop = 0;
            *finished = true;
            *success = false;
        }
    }
    return result;
}

void ReceiverUtil::Cancel()
{
    m_pReceiver->Cancel();
    m_ReceiverUtilState = RECEIVER_CLOSE;
    DEBUG_PRINTF("    [Receiver] Cancel\n");
}

