﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_SenderUtil.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#include <nn/uds.h>
#include "rdt_SenderUtil.h"
#include "Common.h"

nn::Result SenderUtil::Initialize(nn::rdt::Sender* sender)
{
    m_pSender = sender;
    m_SenderUtilState = SENDER_CLOSE;
    return nn::ResultSuccess();
}

void SenderUtil::SetSendData(u8* data, size_t size)
{
    m_sendData = data;
    m_sendDataSize = size;
}

void SenderUtil::Open(s32 timeout)
{
    m_SenderUtilState = SENDER_OPENING;
    m_Timeout = timeout;
}

void SenderUtil::Close()
{
    m_SenderUtilState = SENDER_CLOSE;
}

bool SenderUtil::IsSendFinished()
{
    return (m_SenderUtilState == SENDER_FINISHED);
}

bool SenderUtil::IsClosed()
{
    return (m_SenderUtilState == SENDER_CLOSE);
}

nn::Result SenderUtil::Process()
{
    nn::Result result;
    for(int i=0; i<REPEAT_PROCESS_FUNCTION; i++)
    {
        result = m_pSender->Process();
        if(result.IsFailure() && result != nn::uds::ResultBufferIsFull())
        {
            DEBUG_PRINTF("    [Sender]  %s: Failed state: %d  result: 0x%08x \n",__FUNCTION__, m_SenderUtilState, result.GetPrintableBits());
            return result;
        }
    }
    return  nn::ResultSuccess();
}

nn::Result SenderUtil::Update()
{
    bool finished, success;
    nn::Result result = nn::ResultSuccess();
    switch(m_SenderUtilState)
    {
        case SENDER_CLOSE:
            if(m_pSender->GetStatus() != nn::rdt::SENDER_STATE_CLOSED)
            {
                DEBUG_PRINTF("    [Sender]  CloseRequest -> Cancel()\n");
                m_pSender->Cancel();
            }
            break;
        case SENDER_OPENING:
            result = Ready(&finished, &success, m_Timeout);
            if(result.IsSuccess())
            {
                if(finished)
                {
                    if(success) { m_SenderUtilState = SENDER_OPENED; }
                    else        { m_SenderUtilState = SENDER_CLOSE;  }
                }
            }
            break;

        case SENDER_OPENED:
            result = Send(m_sendData, m_sendDataSize, &finished, &success, m_Timeout);
            if(result.IsSuccess())
            {
                if(finished)
                {
                    if(success) { m_SenderUtilState = SENDER_WAITING_FINISHED; }
                    else        { m_SenderUtilState = SENDER_CLOSE; }
                }
            }
            break;
        
        case SENDER_WAITING_FINISHED:
            result = Finish(&finished, &success, m_Timeout);
            if(result.IsSuccess())
            {
                if(finished)
                {
                    if(success) { m_SenderUtilState = SENDER_FINISHED; }
                    else        { m_SenderUtilState = SENDER_CLOSE; }
                }
            }
            break;
        default:
            break;
    }
    
    if(result.IsFailure())
    {
        DEBUG_PRINTF("    [Sender]  %s: Failed state: %d  result: 0x%08x \n",__FUNCTION__, m_SenderUtilState, result.GetPrintableBits());
        return result;
    }

    return result;
}




nn::Result SenderUtil::Ready(bool* finished, bool* success, s32 timeout)
{
    nn::Result result = nn::ResultSuccess();
    static s32 loop = 0;
    *finished = false;
    *success = false;
    nn::rdt::SenderState status = m_pSender->GetStatus();
    if(status==nn::rdt::SENDER_STATE_CLOSED)
    {
        result = m_pSender->Open();
        if(result.IsSuccess())
        {
            m_TotalSent = 0;
            *success = true;
        }
        loop = 0;
    }
    else if(status==nn::rdt::SENDER_STATE_OPENED)
    {
        // If STATE_OPENED, preparations for sending have completed.
        loop = 0;
        *finished = true;
        *success = true;
    }
    else if(timeout > 0)
    {
        loop++;
        if(loop > timeout)  // TIMEOUT
        {
            DEBUG_PRINTF("    [Sender]  %s TIMEOUT CANCEL\n", __FUNCTION__);
            m_pSender->Cancel();
            loop = 0;
            *finished = true;
            *success = false;
        }
    }
    
    return result;
}

#define SENT_UNIT_SIZE (4*1024)
nn::Result SenderUtil::Send( u8* data, size_t dataSize, bool* finished, bool* success, s32 timeout)
{
    nn::Result result = nn::ResultSuccess();
    size_t sz;
    static s32 loop = 0;
    *finished = false;
    *success = false;
    if(m_pSender->GetStatus() == nn::rdt::SENDER_STATE_OPENED)
    {
        if(m_TotalSent < dataSize)
        {
            sz = (SENT_UNIT_SIZE < (dataSize - m_TotalSent))? SENT_UNIT_SIZE : (dataSize - m_TotalSent);
            nn::Result sendResult = m_pSender->Send(data + m_TotalSent, sz);
            if(sendResult.IsSuccess())
            {
                m_TotalSent += sz ;
                DEBUG_PRINTF("    [Sender]  Sender::Send %4d bytes %d/%d\n", sz, m_TotalSent, dataSize);
                loop = 0;
            }
            else
            {
                if(sendResult == nn::rdt::ResultSendBufferIsNotAvailable())
                {
                    // Do nothing if the send buffer is full (i.e., retry).
                }
                else
                {
                    result = sendResult;
                }
            }
        }
        
        if(timeout > 0)
        {
            loop++;
            if(loop > timeout)  // TIMEOUT
            {
                DEBUG_PRINTF("    [Sender]  %s TIMEOUT CANCEL\n", __FUNCTION__);
                m_pSender->Cancel();
                loop = 0;
                *finished = true;
                *success = false;
            }
        }
        
        if(m_TotalSent >= dataSize )
        {
            // Data sending completed -> Close function
            result = m_pSender->Close();
            if(result.IsSuccess())
            {
                *finished = true;
                *success = true;
            }
        }
    }
    else if(m_pSender->GetStatus() == nn::rdt::SENDER_STATE_CLOSED)
    {
        result = nn::rdt::ResultUntimelyFunctionCall();
    }
    else if(timeout > 0)
    {
        loop++;
        if(loop > timeout)  // TIMEOUT
        {
            DEBUG_PRINTF("    [Sender]  %s TIMEOUT CANCEL\n", __FUNCTION__);
            m_pSender->Cancel();
            loop = 0;
            *finished = true;
            *success = false;
        }
    }
    return result;
}

nn::Result SenderUtil::Finish(bool* finished, bool* success, s32 timeout)
{
    static s32 loop = 0;
    *finished = false;
    *success = false;
    if(m_pSender->GetStatus() == nn::rdt::SENDER_STATE_CLOSED)
    {
        // If STATE_CLOSED, sending has completed.
        *finished = true;
        *success = true;
        loop = 0;
    }
    else if(timeout > 0)
    {
        loop++;
        if(loop > timeout)  // TIMEOUT
        {
            DEBUG_PRINTF("    [Sender]  %s TIMEOUT CANCEL\n", __FUNCTION__);
            m_pSender->Cancel();
            loop = 0;
            *finished = true;
            *success = false;
        }
    }
    return nn::ResultSuccess();
}


void SenderUtil::Cancel()
{
    m_pSender->Cancel();
    DEBUG_PRINTF("    [Sender]  Cancel\n");
}


