﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_UdsUtil.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/


#include "rdt_UdsUtil.h"
#include "Common.h"

#include <nn.h>
#include <nn/hid.h>
#include <nn/uds.h>


namespace
{
    const u32 UNIQUE_ID_RDT_BASIC = 0xF7FFF;  // This value must differ between RDT sample demos.

    const char PASSPHRASE[] = "CTR-SDK RDT sample";
    const size_t PASS_LENGTH = sizeof(PASSPHRASE) - 1;


    nn::os::StackBuffer<4096>    s_MonitorStackBuffer; //For communications monitor thread
    static nn::os::Thread s_MonitorThread;
    bool s_IsActiveMonitorThread = false;
    bool s_IsAliveMonitorThread = false;
    
    u16 s_NodeIdList[nn::uds::NODE_MAX];

    u16 s_MyNodeId = 0;
    u16 s_PeerNodeId = 0;

}  // End of anonymous namespace


static nn::os::Event statusUpdateEvent;
const u8 c_subId = 0x20;    // ID used when dividing communications (for match-ups or exchange) in the application.
const u8 c_maxEntry = 2;    // Maximum number of connections (Set to "2" as this sample uses one-to-one communication.)
#ifdef SUPPORT_UDS_CLIENT
static bit8 s_ScanBuffer[1024 * 8];
#endif
static bit32 s_localId = 0; // The current value is set during initialization.

static bit8 s_UdsReceiveBuffer[4096*10] NN_ATTRIBUTE_ALIGN(4096);  //Receive buffer for system. Must be a multiple of 4096 because the buffer is 4096-byte aligned.

bool setupUds(void)
{
    nn::Result result = nn::uds::Initialize( &statusUpdateEvent, s_UdsReceiveBuffer, sizeof(s_UdsReceiveBuffer) );
    if(result.IsFailure())
    {
        DEBUG_PRINTF("nn::uds::Initialize() failed.\n");
        return false;
    }
    else
    {
        s_localId = nn::uds::CreateLocalCommunicationId(UNIQUE_ID_RDT_BASIC);
        return true;
    }
}


void cleanupUds(void)
{
    nn::uds::Finalize();
}

#ifdef SUPPORT_UDS_CLIENT
// Try to connect to existing UDS network.
// false, if failed. Expected to call again after waiting.
bool connectNetwork(void)
{
    nn::Result result;

    //printNetworkInfo(s_ScanBuffer, sizeof(s_ScanBuffer), c_subId);
    
    result = nn::uds::Scan( s_ScanBuffer, sizeof(s_ScanBuffer), c_subId, s_localId );
    if(result.IsFailure())
    {
        DEBUG_PRINTF("nn::uds::Scan() failed.\n");
        return false;
    }

    nn::uds::ScanResultReader networkList(s_ScanBuffer);


    //Connect to the first network found
    if( 0 != networkList.GetCount() )
    {
        nn::uds::NetworkDescriptionReader netDescReader = networkList.GetNextDescription();
        nn::uds::NetworkDescription netDesc;
        result = netDescReader.GetNetworkDescription(&netDesc);
        NN_UTIL_PANIC_IF_FAILED(result);

        DEBUG_PRINTF("Channel: %d\n", netDesc.GetChannel());

        result = nn::uds::ConnectNetwork( netDesc, nn::uds::CONNECT_AS_CLIENT, PASSPHRASE, PASS_LENGTH );
        if( result.IsFailure() )
        {
            //Display the reason when connection fails
            DEBUG_PRINTF("ConnectNetwork() Failed:\n");
            return false;
        }
        else
        {
            //Go to next process as connection was successful
            return true;
        }
    }
    else
    {
        DEBUG_PRINTF("Network not found.\n");
        return false;
    }
}
#endif

bool createNetwork(void)
{
    DEBUG_PRINTF("nn::uds::CreateNetwork()\n");
    nn::Result result = nn::uds::CreateNetwork( c_subId, c_maxEntry, s_localId, PASSPHRASE, PASS_LENGTH);

    if(result.IsFailure())
    {
        DEBUG_PRINTF("nn::uds::CreateNetwork() failed.\n");
        return false;
    }
    else
    {
        return true;
    }
}

#ifdef SUPPORT_UDS_CLIENT

// Scan nearby and print information about detected networks.
void printNetworkInfo(void *pScanBuf, size_t bufSize, u8 subId)
{
    nn::Result result;

    result = nn::uds::Scan( pScanBuf, bufSize, subId, s_localId );
    NN_UTIL_PANIC_IF_FAILED(result);
    nn::uds::ScanResultReader networkList(pScanBuf);

    u32 networks = networkList.GetCount();
    DEBUG_PRINTF("Detected networks: %d\n", networks);
    for(u32 i=0; i<networks; ++i)
    {
        nn::uds::NetworkDescriptionReader reader = networkList.GetNextDescription();
        nn::uds::NetworkDescription netDesc;
        result = reader.GetNetworkDescription(&netDesc);
        NN_UTIL_PANIC_IF_FAILED(result);
        
        const bit8* bssid = netDesc.GetBssid();
        NN_UNUSED_VAR(bssid);
        
        DEBUG_PRINTF("--- Network information: %d ---\n", i);
        DEBUG_PRINTF("  GameID:      %08X,\n", netDesc.GetGameId());
        DEBUG_PRINTF("  SubID:       %02X,\n", netDesc.GetSubId());
        DEBUG_PRINTF("  TemporaryID: %08X\n",  netDesc.GetTemporaryId());
        DEBUG_PRINTF("  Channel:     %02d\n",  netDesc.GetChannel());
        DEBUG_PRINTF("  MaxEntry:    %02d\n",  netDesc.GetMaxEntry());
        DEBUG_PRINTF("  MAC address: %02x %02x %02x %02x %02x %02x\n", bssid[0], bssid[1], bssid[2], bssid[3], bssid[4], bssid[5]);
        DEBUG_PRINTF("\n");
    }
}
#endif

u16 GetPeerNodeId()
{
    return s_PeerNodeId;
}

u16 GetMyNodeId()
{
    return s_MyNodeId;
}

static void ConnectionMonitorThread()
{
    nn::Result result;
    MEMSET(s_NodeIdList, 0x00, sizeof(s_NodeIdList));
    while(s_IsAliveMonitorThread)
    {
        statusUpdateEvent.Wait();

        nn::uds::ConnectionStatus connectionStatus;
        result = nn::uds::GetConnectionStatus(&connectionStatus);
        if( result.IsFailure() )
        {
            DEBUG_PRINTF("[UDS] GetConnectionStatus Failed.\n");            // Would not usually occur when UDS is initialized
            break;
        }

        if( connectionStatus.nowState != nn::uds::STATE_MASTER
#ifdef SUPPORT_UDS_CLIENT
         && connectionStatus.nowState != nn::uds::STATE_CLIENT
         && connectionStatus.nowState != nn::uds::STATE_SPECTATOR
#endif
        )
        {
            DEBUG_PRINTF("[UDS] Leave Network (Reason:%d)\n", connectionStatus.disconnectReason);
            break;
        }
        else
        {
            s_MyNodeId = connectionStatus.myNodeId;

            for(int i = 0;i < nn::uds::NODE_MAX; i++) //List all nodes
            {
                if( connectionStatus.updateNodeBitmap&(0x0001 << i ) )
                {
                    if( connectionStatus.nodeIdList[i] != 0 ) //New connection
                    {
                        nn::uds::NodeInformation nodeInfo;
                        result = nn::uds::GetNodeInformation( &nodeInfo, connectionStatus.nodeIdList[i] );
                        if( result.IsSuccess() )
                        {
                            if(connectionStatus.nodeIdList[i] != s_MyNodeId)
                            {
                                s_PeerNodeId = connectionStatus.nodeIdList[i];
                                DEBUG_PRINTF("[UDS] Connect! ID:%d\n", connectionStatus.nodeIdList[i]);
                            }
                            s_NodeIdList[i] = connectionStatus.nodeIdList[i];
                        }
                        else
                        {
                            DEBUG_PRINTF("[UDS] Connect Unknwon Node\n");
                        }
                    }
                    else //Disconnect
                    {
                        DEBUG_PRINTF("[UDS] Disconnect! ID:%d\n", s_NodeIdList[i]);
                        if(s_PeerNodeId == s_NodeIdList[i])
                        {
                            s_PeerNodeId = 0;
                        }
                        s_NodeIdList[i] = 0;
                    }
                }
            }
        }
    }
    s_PeerNodeId = 0;
    s_MyNodeId = 0;
}

void StartMonitorThread()
{
    s_IsAliveMonitorThread = true;
    s_MonitorThread.Start( ConnectionMonitorThread,
                s_MonitorStackBuffer);
    s_IsActiveMonitorThread = true;
}

void StopMonitorThread()
{
    if(s_IsAliveMonitorThread )
    {
        s_IsAliveMonitorThread = false;
        statusUpdateEvent.Signal();
    }
    if(s_IsActiveMonitorThread)
    {
        s_MonitorThread.Join();
        s_MonitorThread.Finalize();
        s_IsActiveMonitorThread = false;
    }
}
