﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47456 $
 *---------------------------------------------------------------------------*/

#include <nn.h>

#include <nn/fnd.h>
#include <nn/os.h>
#include <nn/fs.h>
#include <nn/ro.h>
#include <nn/applet.h>

extern "C" NN_DLL_IMPORT void Test4();


extern "C" NN_DLL_EXPORT void Test3()
{
    if( nn::os::Tick::GetSystemCurrent() > 0 )
    {
        NN_LOG("App:    throw from Test3\n");
        throw new std::exception();
    }
}

void nnMain()
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    NN_LOG("App: nnMain\n");

    NN_LOG("App: initialize\n");

    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    nn::fs::FileInputStream f;

    nn::os::MemoryBlock mb;
    mb.Initialize(8 * 1024 * 1024);
    bit8* pHeap = reinterpret_cast<bit8*>(mb.GetAddress());


    // Load crs
    NN_LOG("App: load crs\n");
    s64 crsSize;
    {
        f.Initialize(L"rom:/static.crs");
        crsSize = f.GetSize();
        f.Read(pHeap, crsSize);
        f.Finalize();
    }
    void* pRs = pHeap;
    pHeap += nn::math::RoundUp(crsSize, 0x1000);


    // Load crr
    NN_LOG("App: load crr\n");
    s64 crrSize;
    {
        f.Initialize(L"rom:/.crr/static.crr");
        crrSize = f.GetSize();
        f.Read(pHeap, crrSize);
        f.Finalize();
    }
    void* pRr = pHeap;
    pHeap += crrSize;


    // Load from cro file (stored on the heap)
    NN_LOG("App: load cro\n", pHeap);
    s64 croSize;
    {
        f.Initialize(L"rom:/Module.cro");
        croSize = f.GetSize();
        f.Read(pHeap, croSize);
        f.Finalize();
    }
    void* pRo = pHeap;
    pHeap += nn::math::RoundUp(croSize, 0x1000);


    // cro size information debug output
    nn::ro::SizeInfo sizeInfo;
    nn::ro::GetSizeInfo(&sizeInfo, pRo);
    void* pBuffer = reinterpret_cast<void*>(sizeInfo.fix1End);
    pHeap += nn::math::RoundUp(sizeInfo.bufferSize, 0x1000);

    const size_t crsSize4k = nn::math::RoundUp(crsSize, 0x1000);
    const size_t croSize4k = nn::math::RoundUp(croSize, 0x1000);

    
    // Initialize DLL system
    NN_LOG("App:    ro initialize\n");
    nn::ro::Initialize(pRs, crsSize4k);

    NN_LOG("App:    register crr\n");
    nn::ro::RegistrationList* pRl = nn::ro::RegisterList(pRr, crrSize);
    NN_NULL_ASSERT(pRl);
    {
        // Load cro
        NN_LOG("App:    ro load\n");
        nn::ro::Module* pModule1 = nn::ro::LoadModule(pRo, croSize4k, pBuffer, sizeInfo.bufferSize);
        NN_NULL_ASSERT(pModule1);

        // Call routine for initializing cro
        NN_LOG("App:    ro initialize\n");
        pModule1->DoInitialize();

        // Get function pointer to function on cro
        NN_LOG("App:    ro get\n");
        void (*pFunc1)() = pModule1->GetPointer<void (*)()>("Test");
        NN_NULL_ASSERT(pFunc1);

        // Call function from the function pointer obtained
        NN_LOG("App:    pFunc = %08x\n", pFunc1);
        try
        {
            (pFunc1)();
        }
        catch(std::exception* p)
        {
            NN_LOG("App:    exception occured while calling pFunc1\n");
            delete p;
        }

        try
        {
            Test3();
        }
        catch(std::exception* p)
        {
            NN_LOG("App:    exception occured while calling Test3\n");
            delete p;
        }

        try
        {
            Test4();
        }
        catch(std::exception* p)
        {
            NN_LOG("App:    exception occured while calling Test4\n");
            delete p;
        }


        // Call routine for finalizing cro
        NN_LOG("App:    ro finalize\n");
        pModule1->DoFinalize();

        // Unload cro (in .bss)
        NN_LOG("App:    ro unload\n");
        pModule1->Unload();
    }
    // Finalize DLL system
    NN_LOG("App:    unregister crr\n");
    pRl->Unregister();
    NN_LOG("App:    ro finalize\n");
    nn::ro::Finalize();

    NN_LOG("App: end\n");
}
