﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     3DSurround.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/snd.h>

#include "3DSurround.h"

#include "WaveFileReader.h"

namespace
{
    const int FILE_NUM = 2;
    const char* FILE_NAME[FILE_NUM] =
    {
        "rom:/kart_title.32.wav",  // BGM
        "rom:/idle.wav",           // SE
    };

    // Voice
    nn::snd::Voice* s_pVoice[FILE_NUM];
}

void SurroundPlayer::OnInitialize()
{
    // Initialize DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadDefaultComponent());
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

    m_CriticalSection.Initialize();
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i] = NULL;
    }

    m_OutputMode = nn::snd::OUTPUT_MODE_STEREO;
    m_SpeakerPosition = nn::snd::SURROUND_SPEAKER_POSITION_SQUARE;
    m_SurroundDepth = NN_SND_SURROUND_DEPTH_MAX;
    m_Pan0 = 0;
    m_Pan1 = 0;

    // Start sound thread
    m_SoundThreadFlag = true;
    m_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        reinterpret_cast<uptr>(this),
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );
}

void SurroundPlayer::OnFinalize()
{
    // Destroy the sound thread
    m_SoundThreadFlag = false;
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    m_CriticalSection.Finalize();

    // Shut down the DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();
}

void SurroundPlayer::OnUpdate(nn::hid::PadStatus& padStatus)
{
    // PAN
    if (padStatus.trigger & nn::hid::BUTTON_RIGHT)
    {
        m_Pan0 += 0.2f;
        m_Pan0 = nn::math::Min(m_Pan0, 1.0f);
        SetMixParam(s_pVoice[1]);
    }
    if (padStatus.trigger & nn::hid::BUTTON_LEFT)
    {
        m_Pan0 -= 0.2f;
        m_Pan0 = nn::math::Max(m_Pan0, -1.0f);
        SetMixParam(s_pVoice[1]);
    }
    // Surround PAN
    if (padStatus.trigger & nn::hid::BUTTON_UP)
    {
        m_Pan1 += 0.2f;
        m_Pan1 = nn::math::Min(m_Pan1, 1.0f);
        SetMixParam(s_pVoice[1]);
    }
    if (padStatus.trigger & nn::hid::BUTTON_DOWN)
    {
        m_Pan1 -= 0.2f;
        m_Pan1 = nn::math::Max(m_Pan1, -1.0f);
        SetMixParam(s_pVoice[1]);
    }
    // Surround depth
    if (padStatus.trigger & nn::hid::BUTTON_R)
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        m_SurroundDepth += 0.1f;
        m_SurroundDepth = nn::math::Min(m_SurroundDepth, NN_SND_SURROUND_DEPTH_MAX);
        nn::snd::SetSurroundDepth(m_SurroundDepth);
    }
    if (padStatus.trigger & nn::hid::BUTTON_L)
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        m_SurroundDepth -= 0.1f;
        m_SurroundDepth = nn::math::Max(m_SurroundDepth, NN_SND_SURROUND_DEPTH_MIN);
        nn::snd::SetSurroundDepth(m_SurroundDepth);
    }
    // Sound output mode
    if (padStatus.trigger & nn::hid::BUTTON_A)
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        if (m_OutputMode == nn::snd::OUTPUT_MODE_MONO)
        {
            m_OutputMode = nn::snd::OUTPUT_MODE_STEREO;
        }
        else if (m_OutputMode == nn::snd::OUTPUT_MODE_STEREO)
        {
            m_OutputMode = nn::snd::OUTPUT_MODE_3DSURROUND;
        }
        else if (m_OutputMode == nn::snd::OUTPUT_MODE_3DSURROUND)
        {
            m_OutputMode = nn::snd::OUTPUT_MODE_MONO;
        }
        nn::snd::SetSoundOutputMode(m_OutputMode);
    }
    // Virtual speaker positions
    if (padStatus.trigger & nn::hid::BUTTON_B)
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        if (m_SpeakerPosition == nn::snd::SURROUND_SPEAKER_POSITION_SQUARE)
        {
            m_SpeakerPosition = nn::snd::SURROUND_SPEAKER_POSITION_WIDE;
        }
        else if (m_SpeakerPosition == nn::snd::SURROUND_SPEAKER_POSITION_WIDE)
        {
            m_SpeakerPosition = nn::snd::SURROUND_SPEAKER_POSITION_SQUARE;
        }
        nn::snd::SetSurroundSpeakerPosition(m_SpeakerPosition);
    }
}

void SurroundPlayer::OnDraw(demo::RenderSystemDrawing& renderSystem)
{
    int line = 1;

    renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    renderSystem.Clear();

    renderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    renderSystem.DrawText(10, 10 * line++, "CTR SampleDemos - snd/3DSurround");
    line++;

    renderSystem.DrawText(10, 10 * line++, "[A]          Change output mode");
    renderSystem.DrawText(10, 10 * line++, "[B]          Change speaker position");
    renderSystem.DrawText(10, 10 * line++, "[L/R]        Change surround depth");
    renderSystem.DrawText(10, 10 * line++, "[Left/Right] Change pan");
    renderSystem.DrawText(10, 10 * line++, "[Up/Down]    Change surround pan");
    line++;

    static const char* outputMode[] = { "mono", "stereo", "surround" };
    renderSystem.DrawText(10, 10 * line++, "output mode   : %s", outputMode[m_OutputMode]);
    static const char* speakerPos[] = { "square", "wide" };
    renderSystem.DrawText(10, 10 * line++, "speaker pos   : %s", speakerPos[m_SpeakerPosition]);
    renderSystem.DrawText(10, 10 * line++, "surround depth: %6.2f", m_SurroundDepth);
    line++;

    renderSystem.DrawText(10, 10 * line++, "pan           : %6.2f", m_Pan0);
    renderSystem.DrawText(10, 10 * line++, "surround pan  : %6.2f", m_Pan1);

    renderSystem.SwapBuffers();
    renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void SurroundPlayer::SoundThreadFuncImpl()
{
    nn::snd::SetSoundOutputMode(m_OutputMode);
    nn::snd::SetSurroundSpeakerPosition(m_SpeakerPosition);
    nn::snd::SetSurroundDepth(m_SurroundDepth);

    NN_LOG("Loading wav files...\n");

    u8* pMemory[FILE_NUM];
    nn::snd::WaveBuffer waveBuffer[FILE_NUM];
    // Open the file
    for (int i = 0; i < FILE_NUM; i++)
    {
        WaveFileReader reader(FILE_NAME[i]);
        int dataSize = reader.GetDataSize();
        int channelCount = reader.GetChannelCount();
        int sampleRate = reader.GetSampleRate();
        int quantumBits = reader.GetQuantumBits();

        // Get and check device memory
        pMemory[i] = reinterpret_cast<u8*>(AllocateMemory(dataSize, 32));
        NN_ASSERT(pMemory[i]);

        // Load wave data and invalidate cache
        reader.ReadData(pMemory[i], dataSize);
        nn::snd::FlushDataCache(reinterpret_cast<uptr>(pMemory[i]), dataSize);

        nn::snd::Voice* pVoice = nn::snd::AllocVoice(128, NULL, NULL);
        NN_ASSERT(pVoice);
        s_pVoice[i] = pVoice;

        nn::snd::SampleFormat format = (quantumBits == 8) ?
            nn::snd::SAMPLE_FORMAT_PCM8 : nn::snd::SAMPLE_FORMAT_PCM16;

        pVoice->SetChannelCount(channelCount);
        pVoice->SetSampleFormat(format);

        nn::snd::InitializeWaveBuffer(&waveBuffer[i]);
        waveBuffer[i].bufferAddress = pMemory[i];
        waveBuffer[i].sampleLength  = nn::snd::GetSampleLength(dataSize, format, channelCount);
        waveBuffer[i].loopFlag      = true;

        // Set volume
        if (i == 0)
        {   // Initial file is BGM
            pVoice->SetFrontBypassFlag(true);  // Apply the front bypass (3D processing on rear only)
            pVoice->SetVolume(0.25f);
        }
        else
        {   // Second file is SE
            pVoice->SetFrontBypassFlag(false);  // 3D processing for all channels
            SetMixParam(pVoice);
            pVoice->SetVolume(0.75f);
        }

        // Set pitch
        pVoice->SetSampleRate(sampleRate);
        pVoice->SetPitch(1.0f);

        // Add buffer
        pVoice->AppendWaveBuffer(&waveBuffer[i]);
    }

    // Start playback
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i]->SetState(nn::snd::Voice::STATE_PLAY);
    }

    // The sound thread stops automatically during Sleep Mode
    // (It waits in nn::snd::WaitForDspSync).
    // However, if strict synchronization with graphics is required, the application must control it.
    // 
    while (m_SoundThreadFlag)
    {
        nn::snd::WaitForDspSync();      // Wait to receive data from DSP

        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        nn::snd::SendParameterToDsp();  // Send parameters to the DSP.
    }

    // End playback
    for (int i = 0; i < FILE_NUM; i++)
    {
        if (s_pVoice[i])
        {
            nn::snd::FreeVoice(s_pVoice[i]);
        }
        FreeMemory(pMemory[i]);
    }
}

void SurroundPlayer::SetMixParam(nn::snd::Voice* pVoice)
{
    NN_ASSERT(pVoice);
    f32 left = (1.0f - m_Pan0) * 0.5f;
    f32 rght = (1.0f + m_Pan0) * 0.5f;
    f32 frnt = (1.0f + m_Pan1) * 0.5f;
    f32 rear = (1.0f - m_Pan1) * 0.5f;
    nn::snd::MixParam mix;
    mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 1.0f * left * frnt;
    mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 1.0f * rght * frnt;
    mix.mainBus[nn::snd::CHANNEL_INDEX_REAR_LEFT  ] = 1.0f * left * rear;
    mix.mainBus[nn::snd::CHANNEL_INDEX_REAR_RIGHT ] = 1.0f * rght * rear;
//     NN_LOG(
//         "pan = %5.3f, surround pan = %5.3f\n"
//         "mix param: [front] (left, right) = (%5.3f, %5.3f)\n"
//         "           [rear ] (left, right) = (%5.3f, %5.3f)\n",
//         m_Pan0, m_Pan1,
//         mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ],
//         mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT],
//         mix.mainBus[nn::snd::CHANNEL_INDEX_REAR_LEFT  ],
//         mix.mainBus[nn::snd::CHANNEL_INDEX_REAR_RIGHT ]
//     );
    nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
    pVoice->SetMixParam(mix);
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
