﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     AdpcmEncDec.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47435 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/fs.h>
#include <nn/fnd.h>
#include <nn/snd.h>
#include <nn/applet.h>

#include <string>
#include "WaveFileReader.h"

namespace
{
//     const char INPUT_FILE_NAME[] = "rom:/fanfare_16bit_fs32kHz_1ch.wav";
    const char INPUT_FILE_NAME[] = "rom:/sine.wav";

    const size_t DEVICE_MEMORY_SIZE = (32 * 1024 * 1024);    // Device memory size to use

    const int LOOP_START_BYTE = 32;  // Multiple of 8 (because ADPCM is 8 bytes per frame)
    const int LOOP_START = LOOP_START_BYTE * 14 / 8;
    const int LOOP_PADDING_LEN = 128;

    // Sound thread related
    const int SOUND_THREAD_PRIORITY = 2;
    const int SOUND_THREAD_STACK_SIZE = 4096;
    nn::os::Thread threadSound;
    bool threadSoundFlag;
    void SoundThreadFunc(uptr arg)
    {
        NN_UNUSED_VAR(arg);

        // Prepare device memory
        nn::fnd::ExpHeap expHeap;
        expHeap.Initialize(nn::os::GetDeviceMemoryAddress(), DEVICE_MEMORY_SIZE );

        // Set sound output to stereo
        nn::snd::SetSoundOutputMode(nn::snd::OUTPUT_MODE_STEREO);

        // -----------------------------------------------------------------------
        // Encode
        NN_LOG("[demo] start encoding pcm data...\n");

        // Open the file
        WaveFileReader reader(INPUT_FILE_NAME);
        u32 size = reader.GetDataSize();
        s16* pPcmBuffer = reinterpret_cast<s16*>(expHeap.Allocate(size + LOOP_PADDING_LEN * sizeof(s16), 32));
        NN_ASSERT(pPcmBuffer);
        reader.ReadData(pPcmBuffer, reader.GetDataSize());

        nn::snd::SampleFormat format = (reader.GetQuantumBits() == 8) ?
            nn::snd::SAMPLE_FORMAT_PCM8 : nn::snd::SAMPLE_FORMAT_PCM16;
        s32 nSamples = nn::snd::GetSampleLength(size, format, reader.GetChannelCount());
        const int LOOP_LEN = nSamples;  // Loop playback of entire waveform
        nSamples += LOOP_PADDING_LEN;

        // Copy loop start to end of waveform
        ::std::memcpy(pPcmBuffer + LOOP_LEN, pPcmBuffer, LOOP_PADDING_LEN * sizeof(s16));

        // Prepare the buffer
        s32 nSizeOut = nn::snd::GetAdpcmOutputBufferSize(nSamples);
        u8* pAdpcmBuffer = reinterpret_cast<u8*>(expHeap.Allocate(nSizeOut + sizeof(nn::snd::DspsndAdpcmHeader), 32));
        NN_ASSERT(pAdpcmBuffer);
        nn::snd::DspsndAdpcmHeader* pHeader = reinterpret_cast<nn::snd::DspsndAdpcmHeader*>(pAdpcmBuffer);
        u8* pBody = pAdpcmBuffer + sizeof(nn::snd::DspsndAdpcmHeader);
        NN_ASSERT((reinterpret_cast<uptr>(pBody) & 0x3) == 0);  // Must be four-byte aligned.

        // Encode
        nn::snd::EncodeAdpcmData(pPcmBuffer, pBody, nSamples, reader.GetSampleRate(), LOOP_START, LOOP_LEN + LOOP_START, pHeader);
        NN_LOG("[demo] finished encoding! nSamples(%d) LS(%d) LLEN(%d) LE(%d)\n",
                nSamples, LOOP_START, LOOP_LEN, LOOP_LEN + LOOP_START );

        // -----------------------------------------------------------------------
        // Playback
        NN_LOG("[demo] now start playing adpcm data...\n");

        nn::snd::FlushDataCache(reinterpret_cast<uptr>(pBody), nSizeOut);

        nn::snd::AdpcmParam   param;
        nn::snd::AdpcmContext context, context2;

        nn::snd::Voice* pVoice = nn::snd::AllocVoice(128, NULL, NULL);
        NN_ASSERT(pVoice);

        pVoice->SetChannelCount(1);
        pVoice->SetSampleFormat(nn::snd::SAMPLE_FORMAT_ADPCM);

        // Get ADPCM arguments
        ::std::memcpy(&param, pHeader->state.a, sizeof(param));
        pVoice->SetAdpcmParam(param);

        // Get ADPCM context information
        const nn::snd::DspsndAdpcmState& state = pHeader->state;
        context.yn1 = state.yn1;
        context.yn2 = state.yn2;
        context.pred_scale = state.pred_scale;

        context2.yn1 = state.loop_yn1;
        context2.yn2 = state.loop_yn2;
        context.pred_scale = state.loop_pred_scale;
        NN_LOG("yn1(%d) yn2(%d) ps(%d) LOOP yn1(%d) yn2(%d) ps(%d)\n",
                state.yn1, state.yn2, state.pred_scale,
                state.loop_yn1, state.loop_yn2, state.loop_pred_scale );

        // Add buffer
        nn::snd::WaveBuffer waveBuffer;
        nn::snd::WaveBuffer waveBuffer2;

        nn::snd::InitializeWaveBuffer(&waveBuffer);
        waveBuffer.bufferAddress = pBody;
        waveBuffer.sampleLength  = LOOP_LEN + LOOP_START;
        waveBuffer.loopFlag      = false;
        waveBuffer.pAdpcmContext = &context;
        pVoice->AppendWaveBuffer(&waveBuffer);

        nn::snd::InitializeWaveBuffer(&waveBuffer2);
        waveBuffer2.bufferAddress = pBody + LOOP_START_BYTE;
        waveBuffer2.sampleLength = LOOP_LEN;
        waveBuffer2.loopFlag = true;
        waveBuffer2.pAdpcmContext = &context2;
        pVoice->AppendWaveBuffer(&waveBuffer2);


        pVoice->SetSampleRate(pHeader->sample_rate);
        pVoice->SetState(nn::snd::Voice::STATE_PLAY);

        threadSoundFlag = true;
        while (true)
        {
            nn::snd::WaitForDspSync();
            if (!threadSoundFlag)
            {
                // Fade-out
                f32 volume = pVoice->GetVolume();
                if (volume <= 0.0f)
                {
                    nn::snd::SendParameterToDsp();
                    break;
                }
                pVoice->SetVolume(volume - 0.01f);
            }
            nn::snd::SendParameterToDsp();
        }

        nn::snd::FreeVoice(pVoice);

        // -----------------------------------------------------------------------
        // Decode
        NN_LOG("[demo] start decoding adpcm data...\n");
        nn::snd::AdpcmParam   adpcmParam;
        nn::snd::AdpcmContext adpcmContext;
        ::std::memcpy(&adpcmParam, pHeader->state.a, sizeof(nn::snd::AdpcmParam));
        ::std::memset(&adpcmContext, 0, sizeof(nn::snd::AdpcmContext));
        nn::snd::DecodeAdpcmData(pBody, pPcmBuffer, adpcmParam, adpcmContext, nSamples);
        NN_LOG("[demo] finished decoding!\n");

        // Repeat just LOOP_LEN from LOOP_START
        ::std::memcpy(pPcmBuffer, pPcmBuffer + LOOP_START, LOOP_LEN * sizeof(s16));

        // -----------------------------------------------------------------------
        // Playback
        NN_LOG("[demo] now start playing decoded pcm data...\n");

        nn::snd::FlushDataCache(reinterpret_cast<uptr>(pPcmBuffer), nSamples * sizeof(s16));

        pVoice = nn::snd::AllocVoice(128, NULL, NULL);
        NN_ASSERT(pVoice);

        pVoice->SetChannelCount(1);
        pVoice->SetSampleFormat(nn::snd::SAMPLE_FORMAT_PCM16);

        // Add buffer
        nn::snd::InitializeWaveBuffer(&waveBuffer);
        waveBuffer.bufferAddress = pPcmBuffer;
        waveBuffer.sampleLength  = LOOP_LEN;
        waveBuffer.loopFlag      = true;
        pVoice->AppendWaveBuffer(&waveBuffer);

        pVoice->SetVolume(0.0f);
        pVoice->SetSampleRate(pHeader->sample_rate);
        pVoice->SetState(nn::snd::Voice::STATE_PLAY);

        threadSoundFlag = true;
        bool fadeIn = true;
        while (threadSoundFlag)
        {
            nn::snd::WaitForDspSync();
            if (fadeIn)
            {
                // Fade-in
                f32 volume = pVoice->GetVolume();
                volume = nn::math::Min(1.0f, volume + 0.01f);
                pVoice->SetVolume(volume);
                if (volume == 1.0f)
                {
                    fadeIn = false;
                }
            }
            nn::snd::SendParameterToDsp();
        }
    }
}

void nnMain()
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    NN_LOG("Demo Start\n");

    nn::Result result;

    // fs initialization
    nn::fs::Initialize();

    // Initialize hid
    result = nn::hid::Initialize();
    NN_UTIL_PANIC_IF_FAILED(result);

    // Prepare pad
    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    // Initialize DSP and SND libraries
    result = nn::dsp::Initialize();
    NN_UTIL_PANIC_IF_FAILED(result);
    result = nn::dsp::LoadDefaultComponent();
    NN_UTIL_PANIC_IF_FAILED(result);
    result = nn::snd::Initialize();
    NN_UTIL_PANIC_IF_FAILED(result);

    threadSound.StartUsingAutoStack(SoundThreadFunc, NULL, SOUND_THREAD_STACK_SIZE, SOUND_THREAD_PRIORITY);

    NN_LOG("[demo] press button A to start adpcm decoder test\n");
    while (true)
    {
        padReader.ReadLatest(&padStatus);
        if(padStatus.trigger & nn::hid::BUTTON_A)
        {
            threadSoundFlag = false;  // Complete encoding, start decoding
            break;
        }
    }

    while (true) {}
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
