﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     AuxBusSample.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/snd.h>

#include "AuxBusSample.h"

#include "WaveFileReader.h"

namespace
{
    const int FILE_NUM = 1;
    const char* FILE_NAME[FILE_NUM] =
    {
        "rom:/shutter.wav",
    };

    // Voice
    nn::snd::Voice* s_pVoice[FILE_NUM];

    // AUX_BUS_A: Simple delay effect
    const int AUX_DELAY_BUFFER_LENGTH = NN_SND_SAMPLES_PER_FRAME * 80;
    s32 s_DelayBuffer[2][AUX_DELAY_BUFFER_LENGTH];
    s32 s_DelayBufferPosition = 0;
    void AuxUserCallbackFunctionA(nn::snd::AuxBusData* data, s32 sampleLength, uptr userData)
    {
        NN_UNUSED_VAR(userData);

        s32* frontLeft  = data->frontLeft;
        s32* frontRight = data->frontRight;

        s32* frontLeftDelay  = s_DelayBuffer[0] + s_DelayBufferPosition;
        s32* frontRightDelay = s_DelayBuffer[1] + s_DelayBufferPosition;

        for (int i = 0; i < sampleLength; i++)
        {
            frontLeftDelay [i] = frontLeft [i];
            frontRightDelay[i] = frontRight[i];
        }
        s_DelayBufferPosition += sampleLength;

        if (s_DelayBufferPosition >= AUX_DELAY_BUFFER_LENGTH)
        {
            s_DelayBufferPosition = 0;
        }

        frontLeftDelay  = s_DelayBuffer[0] + s_DelayBufferPosition;
        frontRightDelay = s_DelayBuffer[1] + s_DelayBufferPosition;
        for (int i = 0; i < sampleLength; i++)
        {
            frontLeft [i] = frontLeftDelay [i];
            frontRight[i] = frontRightDelay[i];
        }
    }

    // AUX_BUS_B: Write a sine wave
    u16 s_Index = 0;
    void AuxUserCallbackFunctionB(nn::snd::AuxBusData* data, s32 sampleLength, uptr userData)
    {
        NN_UNUSED_VAR(userData);

        s32* frontLeft  = data->frontLeft;
        s32* frontRight = data->frontRight;

        for (int i = 0; i < sampleLength; i++)
        {
            frontLeft [i] = (32767.0f * nn::math::SinFIdx(4*s_Index));
            frontRight[i] = (32767.0f * nn::math::SinFIdx(4*s_Index));
            s_Index++;
        }
    }
}

void AuxBusSample::OnInitialize()
{
    // Initialize DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadDefaultComponent());
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

    m_CriticalSection.Initialize();
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i] = NULL;
    }

    // Start sound thread
    m_SoundThreadFlag = true;
    m_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        reinterpret_cast<uptr>(this),
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );
}

void AuxBusSample::OnFinalize()
{
    // Destroy the sound thread
    m_SoundThreadFlag = false;
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    m_CriticalSection.Finalize();

    // Shut down the DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();
}

void AuxBusSample::OnUpdate(nn::hid::PadStatus& padStatus)
{
    // A: Add a callback to AuxBusA
    if (padStatus.trigger & nn::hid::BUTTON_A)
    {
        nn::snd::AuxCallback callback;
        uptr arg;
        nn::snd::GetAuxCallback(nn::snd::AUX_BUS_A, &callback, &arg);
        if (callback)
        {
            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            nn::snd::ClearAuxCallback(nn::snd::AUX_BUS_A);
        }
        else
        {
            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            nn::snd::RegisterAuxCallback(nn::snd::AUX_BUS_A, AuxUserCallbackFunctionA, NULL);
        }
    }
    // B: Add a callback to AuxBusB
    if (padStatus.trigger & nn::hid::BUTTON_B)
    {
        nn::snd::AuxCallback callback;
        uptr arg;
        nn::snd::GetAuxCallback(nn::snd::AUX_BUS_B, &callback, &arg);
        if (callback)
        {
            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            nn::snd::ClearAuxCallback(nn::snd::AUX_BUS_B);
        }
        else
        {
            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            nn::snd::RegisterAuxCallback(nn::snd::AUX_BUS_B, AuxUserCallbackFunctionB, NULL);
        }
    }
}

void AuxBusSample::OnDraw(demo::RenderSystemDrawing& renderSystem)
{
    int line = 1;

    renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    renderSystem.Clear();

    renderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    renderSystem.DrawText(10, 10 * line++, "CTR SampleDemos - snd/AuxBusSample");
    line++;

    renderSystem.DrawText(10, 10 * line++, "[A]          Enable/disable callback");
    renderSystem.DrawText(10, 10 * line++, "                 for AuxBusA");
    renderSystem.DrawText(10, 10 * line++, "[B]          Enable/disable callback");
    renderSystem.DrawText(10, 10 * line++, "                 for AuxBusB");
    line++;

    nn::snd::AuxCallback callback;
    uptr arg;
    nn::snd::GetAuxCallback(nn::snd::AUX_BUS_A, &callback, &arg);
    renderSystem.DrawText(10, 10 * line++, "AuxBusA (delay) : %s",
                          callback ? "On" : "Off");
    nn::snd::GetAuxCallback(nn::snd::AUX_BUS_B, &callback, &arg);
    renderSystem.DrawText(10, 10 * line++, "AuxBusB (sine)  : %s",
                          callback ? "On" : "Off");

    renderSystem.SwapBuffers();
    renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void AuxBusSample::SoundThreadFuncImpl()
{
    // Set sound output to stereo
    nn::snd::SetSoundOutputMode(nn::snd::OUTPUT_MODE_STEREO);

    NN_LOG("Loading wav files...\n");

    u8* pMemory[FILE_NUM];
    nn::snd::WaveBuffer waveBuffer[FILE_NUM];
    // Open the file
    for (int i = 0; i < FILE_NUM; i++)
    {
        WaveFileReader reader(FILE_NAME[i]);
        int dataSize = reader.GetDataSize();
        int channelCount = reader.GetChannelCount();
        int sampleRate = reader.GetSampleRate();
        int quantumBits = reader.GetQuantumBits();

        // Get and check device memory
        pMemory[i] = reinterpret_cast<u8*>(AllocateMemory(dataSize, 32));
        NN_ASSERT(pMemory[i]);

        // Load wave data and invalidate cache
        reader.ReadData(pMemory[i], dataSize);
        nn::snd::FlushDataCache(reinterpret_cast<uptr>(pMemory[i]), dataSize);

        nn::snd::Voice* pVoice = nn::snd::AllocVoice(128, NULL, NULL);
        NN_ASSERT(pVoice);
        s_pVoice[i] = pVoice;

        nn::snd::SampleFormat format = (quantumBits == 8) ?
            nn::snd::SAMPLE_FORMAT_PCM8 : nn::snd::SAMPLE_FORMAT_PCM16;

        pVoice->SetChannelCount(channelCount);
        pVoice->SetSampleFormat(format);

        nn::snd::InitializeWaveBuffer(&waveBuffer[i]);
        waveBuffer[i].bufferAddress = pMemory[i];
        waveBuffer[i].sampleLength  = nn::snd::GetSampleLength(dataSize, format, channelCount);
        waveBuffer[i].loopFlag      = true;

        // Set volume
        nn::snd::MixParam mix;
        mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.707f;
        mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.707f;
        mix.auxBusA[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.707f * 0.25f;
        mix.auxBusA[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.707f * 0.25f;
        mix.auxBusB[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.707f * 0.25f;
        mix.auxBusB[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.707f * 0.25f;
        pVoice->SetMixParam(mix);
        pVoice->SetVolume(1.0f);

        // Set pitch
        pVoice->SetSampleRate(sampleRate);
        pVoice->SetPitch(1.0f);

        // Add buffer
        pVoice->AppendWaveBuffer(&waveBuffer[i]);
    }

    // Start playback
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i]->SetState(nn::snd::Voice::STATE_PLAY);
    }

    // The sound thread stops automatically during Sleep Mode
    // (It waits in nn::snd::WaitForDspSync).
    // However, if strict synchronization with graphics is required, the application must control it.
    // 
    while (m_SoundThreadFlag)
    {
        nn::snd::WaitForDspSync();      // Wait to receive data from DSP

        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        nn::snd::SendParameterToDsp();  // Send parameters to the DSP.
    }

    // End playback
    for (int i = 0; i < FILE_NUM; i++)
    {
        if (s_pVoice[i])
        {
            nn::snd::FreeVoice(s_pVoice[i]);
        }
        FreeMemory(pMemory[i]);
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
