﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     BcwavPlayer.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/snd.h>

#include "BcwavPlayer.h"

namespace
{
    // File
    const int FILE_NUM = 3;
    const char* FILE_NAME[FILE_NUM] =
    {
        "rom:/yoshi_fs16kHz_1ch.pcm8.bcwav",
        "rom:/fanfare_fs32kHz_1ch.pcm16.bcwav",
        "rom:/sin440.dspadpcm.bcwav",
    };

    // Voice
    int s_CurrentVoice;
    nn::snd::Voice* s_pVoice[FILE_NUM];

    // Buffer
    struct MyWaveBuffer : public nn::snd::WaveBuffer
    {
    public:
        void Dump()
        {
            NN_LOG("[WaveBuffer::Dump] bufAddr(%p) len(%6d|%08X) loop(%d)",
                   bufferAddress, sampleLength, sampleLength, loopFlag);

            if (pAdpcmContext == NULL)
            {
                NN_LOG("\n");
            }
            else
            {
                NN_LOG(" ps(%04X) yn1(%04X) yn2(%04X)\n",
                       pAdpcmContext->pred_scale,
                       pAdpcmContext->yn1, pAdpcmContext->yn2);
            }
        }
    };
    void* s_pBcwavBuffer[FILE_NUM];
    MyWaveBuffer s_WaveBuffer[FILE_NUM][2];  // Allocate two for stereo sound
}

void BcwavPlayer::OnInitialize()
{
    // Initialize DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadDefaultComponent());
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

    m_CriticalSection.Initialize();
    s_CurrentVoice = 0;
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i] = NULL;
    }

    // Allocate memory and read data
    for (int i = 0; i < FILE_NUM; i++)
    {
        nn::fs::FileReader fileReader = nn::fs::FileReader(FILE_NAME[i]);
        size_t fileSize = fileReader.GetSize();
        s_pBcwavBuffer[i] = reinterpret_cast<u8*>(AllocateMemory(fileSize, 32));
        size_t size = fileReader.Read(s_pBcwavBuffer[i], fileSize);
        NN_LOG("[%s] fileSize(%d) loadSize(%d) fileSize(%d)\n", FILE_NAME[i], fileSize, size, fileSize);
        fileReader.Finalize();

        nn::snd::FlushDataCache(reinterpret_cast<uptr>(s_pBcwavBuffer[i]), size);
    }

    // Start sound thread
    m_SoundThreadFlag = true;
    m_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        reinterpret_cast<uptr>(this),
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );
}

void BcwavPlayer::OnFinalize()
{
    // Destroy the sound thread
    m_SoundThreadFlag = false;
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    // Free memory for the data
    for (int i = 0; i < FILE_NUM; i++)
    {
        if (s_pBcwavBuffer[i])
        {
            FreeMemory(s_pBcwavBuffer[i]);
        }
    }

    m_CriticalSection.Finalize();

    // Shut down the DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();
}

void BcwavPlayer::OnUpdate(nn::hid::PadStatus& padStatus)
{
    int voice = s_CurrentVoice;
    nn::snd::Voice* pVoice = s_pVoice[voice];

    if (padStatus.trigger & nn::hid::BUTTON_A)
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);

        if (pVoice == NULL)
        {
            pVoice = nn::snd::AllocVoice(128, NULL, NULL);
            NN_ASSERT(pVoice);

            pVoice->SetupBcwav(reinterpret_cast<uptr>(s_pBcwavBuffer[voice]),
                               &s_WaveBuffer[voice][0], &s_WaveBuffer[voice][1]);

            s_WaveBuffer[voice][0].Dump();
            s_WaveBuffer[voice][1].Dump();

            nn::snd::MixParam mix;
            mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.5f;
            mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.5f;
            pVoice->SetMixParam(mix);
            pVoice->SetState(nn::snd::Voice::STATE_PLAY);
        }
        else
        {
            nn::snd::FreeVoice(pVoice);
            pVoice = NULL;
        }
        s_pVoice[voice] = pVoice;
    }
    if (padStatus.trigger & nn::hid::BUTTON_UP)
    {
        voice--;
        if (voice < 0) voice = 0;
    }
    if (padStatus.trigger & nn::hid::BUTTON_DOWN)
    {
        voice++;
        if (voice >= FILE_NUM) voice = FILE_NUM - 1;
    }

    s_CurrentVoice = voice;
}

void BcwavPlayer::OnDraw(demo::RenderSystemDrawing& renderSystem)
{
    renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    renderSystem.Clear();

    renderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    int line = 1;
    renderSystem.DrawText(10, 10 * line++, "CTR SampleDemos - snd/BcwavPlayer");
    line++;

    renderSystem.DrawText(10, 10 * line++, "[A]          Play/Pause");
    renderSystem.DrawText(10, 10 * line++, "[Up/Down]    Change voice index");
    line++;

    for (int i = 0; i < FILE_NUM; i++)
    {
        renderSystem.DrawText(
            10, 10 * line++, "%s Voice[%d]: (%s)",
            ((i == s_CurrentVoice) ? ">" : " "), i,
            (s_pVoice[i] == NULL ? " stop" : " play")
        );
    }

    renderSystem.SwapBuffers();
    renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void BcwavPlayer::SoundThreadFuncImpl()
{
    // Set sound output to stereo
    nn::snd::SetSoundOutputMode(nn::snd::OUTPUT_MODE_STEREO);

    // The sound thread stops automatically during Sleep Mode
    // (It waits in nn::snd::WaitForDspSync).
    // However, if strict synchronization with graphics is required, the application must control it.
    // 
    while (m_SoundThreadFlag)
    {
        nn::snd::WaitForDspSync();      // Wait for data from DSP

        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        nn::snd::SendParameterToDsp();  // Send arguments to DSP
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
