﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     EffectSample.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/snd.h>

#include "EffectSample.h"

namespace
{
    // File
    const int FILE_NUM = 1;
    const char* FILE_NAME[FILE_NUM] =
    {
        "rom:/fanfare_fs32kHz_1ch.pcm16.bcwav",
    };

    // Voice
    int s_CurrentVoice;
    nn::snd::Voice* s_pVoice[FILE_NUM];

    // Buffer
    void* s_pBcwavBuffer[FILE_NUM];
    nn::snd::WaveBuffer s_WaveBuffer[FILE_NUM][2];  // Allocate two for stereo sound

    // Effect index
    int s_EffectType = EffectSample::EFFECT_TYPE_NONE;
}

void EffectSample::OnInitialize()
{
    // Initialize DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadDefaultComponent());
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

    m_CriticalSection.Initialize();
    s_CurrentVoice = 0;
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i] = NULL;
    }

    // Allocate memory and read data
    for (int i = 0; i < FILE_NUM; i++)
    {
        nn::fs::FileReader fileReader = nn::fs::FileReader(FILE_NAME[i]);
        size_t fileSize = fileReader.GetSize();
        s_pBcwavBuffer[i] = reinterpret_cast<u8*>(AllocateMemory(fileSize, 32));
        size_t size = fileReader.Read(s_pBcwavBuffer[i], fileSize);
        NN_LOG("[%s] fileSize(%d) loadSize(%d) fileSize(%d)\n", FILE_NAME[i], fileSize, size, fileSize);
        fileReader.Finalize();

        nn::snd::FlushDataCache(reinterpret_cast<uptr>(s_pBcwavBuffer[i]), size);
    }

    // Start sound thread
    m_SoundThreadFlag = true;
    m_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        reinterpret_cast<uptr>(this),
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );
}

void EffectSample::OnFinalize()
{
    // Destroy the sound thread
    m_SoundThreadFlag = false;
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    // Free memory for the data
    for (int i = 0; i < FILE_NUM; i++)
    {
        if (s_pBcwavBuffer[i])
        {
            FreeMemory(s_pBcwavBuffer[i]);
        }
    }
    FinalizeEffect();

    m_CriticalSection.Finalize();

    // Shut down the DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();
}

void EffectSample::OnUpdate(nn::hid::PadStatus& padStatus)
{
    int voice = s_CurrentVoice;
    nn::snd::Voice* pVoice = s_pVoice[voice];

    if (padStatus.trigger & nn::hid::BUTTON_A)
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);

        if (pVoice == NULL)
        {
            pVoice = nn::snd::AllocVoice(128, NULL, NULL);
            NN_ASSERT(pVoice);

            pVoice->SetupBcwav(reinterpret_cast<uptr>(s_pBcwavBuffer[voice]),
                               &s_WaveBuffer[voice][0], &s_WaveBuffer[voice][1]);

            nn::snd::MixParam mix;
            mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.5f;
            mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.5f;
            mix.auxBusA[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.3f;
            mix.auxBusA[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.3f;
            pVoice->SetMixParam(mix);
            pVoice->SetState(nn::snd::Voice::STATE_PLAY);
        }
        else
        {
            nn::snd::FreeVoice(pVoice);
            pVoice = NULL;
        }
        s_pVoice[voice] = pVoice;
    }
    if (padStatus.trigger & nn::hid::BUTTON_UP)
    {
        voice--;
        if (voice < 0) voice = 0;
    }
    if (padStatus.trigger & nn::hid::BUTTON_DOWN)
    {
        voice++;
        if (voice >= FILE_NUM) voice = FILE_NUM - 1;
    }
    if (padStatus.trigger & nn::hid::BUTTON_X)
    {
        FinalizeEffect();
        s_EffectType++;
        if (s_EffectType >= EFFECT_TYPE_NUM)
        {
            s_EffectType = 0;
        }
        InitializeEffect(static_cast<EffectType>(s_EffectType));
    }

    s_CurrentVoice = voice;
}

void EffectSample::OnDraw(demo::RenderSystemDrawing& renderSystem)
{
    int line = 1;

    renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    renderSystem.Clear();

    renderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    renderSystem.DrawText(10, 10 * line++, "CTR SampleDemos - snd/EffectSample");
    line++;

    renderSystem.DrawText(10, 10 * line++, "[A]          Play/Pause");
    renderSystem.DrawText(10, 10 * line++, "[X]          Change effect type");
    if (FILE_NUM > 1)
    {
        renderSystem.DrawText(10, 10 * line++, "[Up/Down]    Change voice index");
    }
    line++;

    renderSystem.DrawText(10, 10 * line++, "Effect Type: %s", GetEffectTypeString());
    line++;

    for (int i = 0; i < FILE_NUM; i++)
    {
        renderSystem.DrawText(
            10, 10 * line++, "%s Voice[%d]: (%s)",
            ((i == s_CurrentVoice) ? ">" : " "), i,
            (s_pVoice[i] == NULL ? " stop" : " play")
        );
    }

    renderSystem.SwapBuffers();
    renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void EffectSample::SoundThreadFuncImpl()
{
    // Set sound output to stereo
    nn::snd::SetSoundOutputMode(nn::snd::OUTPUT_MODE_STEREO);

    // The sound thread stops automatically during Sleep Mode
    // (It waits in nn::snd::WaitForDspSync).
    // However, if strict synchronization with graphics is required, the application must control it.
    // 
    while (m_SoundThreadFlag)
    {
        nn::snd::WaitForDspSync();      // Wait for data from DSP

        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        nn::snd::SendParameterToDsp();  // Send arguments to DSP
    }
}

void EffectSample::InitializeEffect(EffectType type)
{
    NN_ASSERT(m_FxType == EFFECT_TYPE_NONE);
    m_FxType = type;
    switch (m_FxType)
    {
    case EFFECT_TYPE_NONE:
        break;
    case EFFECT_TYPE_CPU_DELAY:
        {
            m_FxDelayParam.m_DelayTime = 300;
            // Set other arguments as needed.
            m_FxDelay.SetParam(m_FxDelayParam);
            size_t delayBufSize = m_FxDelay.GetRequiredMemSize();
            m_FxWorkBuffer = reinterpret_cast<uptr>(AllocateMemory(delayBufSize, 32));
            m_FxDelay.AssignWorkBuffer(m_FxWorkBuffer, delayBufSize);

            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            nn::snd::SetEffect(nn::snd::AUX_BUS_A, &m_FxDelay);
        }
        break;
    case EFFECT_TYPE_CPU_REVERB:
        {
            m_FxReverbParam.m_EarlyReflectionTime = 400;
            // Set other arguments as needed.
            m_FxReverb.SetParam(m_FxReverbParam);
            size_t delayBufSize = m_FxReverb.GetRequiredMemSize();
            m_FxWorkBuffer = reinterpret_cast<uptr>(AllocateMemory(delayBufSize, 32));
            m_FxReverb.AssignWorkBuffer(m_FxWorkBuffer, delayBufSize);

            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            nn::snd::SetEffect( nn::snd::AUX_BUS_A, &m_FxReverb );
        }
        break;
    case EFFECT_TYPE_DSP_DELAY:
        {
            bool ret;
            NN_UNUSED_VAR(ret);

            // Prepare the delay buffer
            size_t bufferSize = nn::snd::DspFxDelay::GetRequiredMemorySize(m_DspFxDelayParam);
            m_FxWorkBuffer = reinterpret_cast<uptr>(AllocateMemory(bufferSize, 32));

            // Initialize the delay
            ret = m_DspFxDelay.Initialize(m_FxWorkBuffer, bufferSize);
            NN_ASSERT(ret);

            // Associate the delay with AUX_BUS_A
            ret = m_DspFxDelay.Attach(nn::snd::AUX_BUS_A);
            NN_ASSERT(ret);

            // Set the delay parameters
            ret = m_DspFxDelay.SetParam(m_DspFxDelayParam);
            NN_ASSERT(ret);

            // Enable the delay
            ret = m_DspFxDelay.Enable();
            NN_ASSERT(ret);
        }
        break;
    case EFFECT_TYPE_DSP_REVERB:
        {
            bool ret;
            NN_UNUSED_VAR(ret);

            // Prepare the reverb buffer
            size_t bufferSize = nn::snd::DspFxReverb::GetRequiredMemorySize(m_DspFxReverbParam);
            m_FxWorkBuffer = reinterpret_cast<uptr>(AllocateMemory(bufferSize, 32));

            // Initialize reverb
            ret = m_DspFxReverb.Initialize(m_FxWorkBuffer, bufferSize);
            NN_ASSERT(ret);

            // Associate the reverb with AUX_BUS_A
            ret = m_DspFxReverb.Attach(nn::snd::AUX_BUS_A);
            NN_ASSERT(ret);

            // Set the reverb parameters
            ret = m_DspFxReverb.SetParam(m_DspFxReverbParam);
            NN_ASSERT(ret);

            // Initialize the reverb
            ret = m_DspFxReverb.Enable();
            NN_ASSERT(ret);
        }
        break;
    }
}

void EffectSample::FinalizeEffect()
{
    switch ( m_FxType )
    {
    case EFFECT_TYPE_NONE:
        break;
    case EFFECT_TYPE_CPU_DELAY:
        {
            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            nn::snd::ClearEffect(nn::snd::AUX_BUS_A);
        }
        m_FxDelay.ReleaseWorkBuffer();
        break;
    case EFFECT_TYPE_CPU_REVERB:
        {
            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            nn::snd::ClearEffect(nn::snd::AUX_BUS_A);
        }
        m_FxReverb.ReleaseWorkBuffer();
        break;
    case EFFECT_TYPE_DSP_DELAY:
        {
            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            m_DspFxDelay.Finalize();
        }
        // Wait until memory can be freed
        while (m_DspFxDelay.IsBufferInUse()) {}
        break;
    case EFFECT_TYPE_DSP_REVERB:
        {
            nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
            m_DspFxReverb.Finalize();
        }
        // Wait until memory can be freed
        while (m_DspFxReverb.IsBufferInUse()) {}
        break;
    }
    m_FxType = EFFECT_TYPE_NONE;

    if (m_FxWorkBuffer)
    {
        FreeMemory(reinterpret_cast<void*>(m_FxWorkBuffer));
        m_FxWorkBuffer = NULL;
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
