﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     OutputCapture.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/hio.h>
#include <nn/snd.h>

#include "OutputCapture.h"
#include "WaveFileWriter.h"

namespace
{
    uptr s_MemoryForHio;
    s16* s_pCaptureBuffer;
    nn::snd::OutputCapture s_Capture;

    // Output HostIO
    bool                s_IsDataWrote = false;
    const s32           HIO_OUTPUT_LENGTH = 2048;
    s16                 s_HioBuffer[NN_SND_SAMPLES_PER_FRAME*2*HIO_OUTPUT_LENGTH];
    s16*                s_pHioBuffer = s_HioBuffer;
    const s16*          s_pHioBufferEnd = s_HioBuffer + sizeof(s_HioBuffer) / sizeof(s_HioBuffer[0]);
    void OpenAndWriteToHostFile()
    {
        // Only write to a file once
        if (s_IsDataWrote)
        {
            return;
        }

        const char FILE_NAME[] = TEST_DIRECTORY OUTPUT_FILE;

        NN_LOG("Create file [%s]\n", FILE_NAME );
        HostIoWaveFileWriter writer(FILE_NAME);

        // Output format settings
        writer.SetChannelCount(2);
        writer.SetSampleRate(NN_SND_HW_I2S_CLOCK_32KHZ);
        writer.SetQuantumBits(16);

        // Output
        writer.WriteData(s_HioBuffer, sizeof(s_HioBuffer));

        s_IsDataWrote = true;
    }

    // Generate sine wave
    void MakeSineWave(s16* p, s32 n)
    {
        for (s32 i = 0; i < n; i++)
        {
            p[i] = (32767.0f * nn::math::SinFIdx(8.0f*i));
        }
    }
}

void OutputCapture::OnInitialize()
{
    // Initialize HostIO
    s_MemoryForHio = reinterpret_cast<uptr>(AllocateMemory(nn::hio::WORKMEMORY_SIZE));
    nn::hio::Initialize(reinterpret_cast<void*>(s_MemoryForHio));
    NN_LOG("HIO Lib ver: %d\n", nn::hio::GetVersion());

    // Initialize DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadDefaultComponent());
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

    // Initialize OutputCapture
    size_t captureBufferSize = nn::snd::OutputCapture::GetRequiredMemorySize(16);
    s_pCaptureBuffer = reinterpret_cast<s16*>(AllocateMemory(captureBufferSize));
    NN_ASSERT(s_pCaptureBuffer);
    s_Capture.Initialize(s_pCaptureBuffer, captureBufferSize);
    nn::snd::SetOutputCapture(&s_Capture);
    s_Capture.Enable();

    // Start sound thread
    m_SoundThreadFlag = true;
    m_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        reinterpret_cast<uptr>(this),
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );
}

void OutputCapture::OnFinalize()
{
    // Destroy the sound thread
    m_SoundThreadFlag = false;
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    // Finalize capture
    s_Capture.Finalize();
    if (s_pCaptureBuffer)
    {
        FreeMemory(s_pCaptureBuffer);
        s_pCaptureBuffer = NULL;
    }

    // Shut down the DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();

    // File output
    OpenAndWriteToHostFile();

    // Finalize HostIO
    nn::hio::Finalize();
    if (s_MemoryForHio)
    {
        FreeMemory(reinterpret_cast<void*>(s_MemoryForHio));
        s_MemoryForHio = NULL;
    }
}

void OutputCapture::OnUpdate(nn::hid::PadStatus& padStatus)
{
    NN_UNUSED_VAR(padStatus);
}

void OutputCapture::OnDraw(demo::RenderSystemDrawing& renderSystem)
{
    int line = 1;

    renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    renderSystem.Clear();

    renderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    renderSystem.DrawText(10, 10 * line++, "CTR SampleDemos - snd/OutputCapture");
    line++;

    while (s_pHioBuffer < s_pHioBufferEnd)
    {
        s32 samples = s_Capture.Read(s_pHioBuffer, NN_SND_SAMPLES_PER_FRAME);
        if (samples == 0)                     // Exit if no data can be obtained
        {
            break;
        }
        s_pHioBuffer += samples * 2;          // Advance the pointer by the amount that was read (Note that stereo is always used)
        if (s_pHioBuffer >= s_pHioBufferEnd)  // If the data input buffer is full, exit
        {
            break;
        }
    }
    if (s_pHioBuffer < s_pHioBufferEnd)
    {
        renderSystem.DrawText(10, 10 * line++, " capturing sound output...");
    }
    else
    {
        // File output
        OpenAndWriteToHostFile();
        renderSystem.DrawText(10, 10 * line++, " capture done!");
    }

    renderSystem.SwapBuffers();
    renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void OutputCapture::SoundThreadFuncImpl()
{
    // Prepare waveform
    const s32 SAMPLE_LENGTH = 2048;
    s16* pSoundData = reinterpret_cast<s16*>(AllocateMemory(SAMPLE_LENGTH * sizeof(s16), 32));
    NN_ASSERT(pSoundData);
    MakeSineWave(pSoundData, SAMPLE_LENGTH);
    nn::snd::FlushDataCache(reinterpret_cast<uptr>(pSoundData), SAMPLE_LENGTH * sizeof(s16));

    // Initialize Voice
    nn::snd::Voice* pVoice = nn::snd::AllocVoice(128, NULL, NULL);
    NN_ASSERT(pVoice);

    pVoice->SetChannelCount(1);
    pVoice->SetSampleFormat(nn::snd::SAMPLE_FORMAT_PCM16);

    nn::snd::MixParam mix;
    mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT]  = 0.707f;
    mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.707f;
    pVoice->SetMixParam(mix);
    pVoice->SetVolume(1.0f);

    pVoice->SetSampleRate(NN_SND_HW_I2S_CLOCK_32KHZ);
    pVoice->SetPitch(1.0f);

    nn::snd::WaveBuffer waveBuffer;
    nn::snd::InitializeWaveBuffer(&waveBuffer);
    waveBuffer.bufferAddress = pSoundData;
    waveBuffer.sampleLength  = SAMPLE_LENGTH;
    waveBuffer.loopFlag = true;
    pVoice->AppendWaveBuffer(&waveBuffer);

    pVoice->SetState(nn::snd::Voice::STATE_PLAY);

    // The sound thread stops automatically during Sleep Mode
    // (It waits in nn::snd::WaitForDspSync).
    // However, if strict synchronization with graphics is required, the application must control it.
    // 
    while (m_SoundThreadFlag)
    {
        nn::snd::WaitForDspSync();      // Wait to receive data from DSP

        nn::snd::SendParameterToDsp();  // Send parameters to the DSP.
    }

    // Release Voice
    nn::snd::FreeVoice(pVoice);
    pVoice = NULL;

    // Release the waveform buffer
    if (pSoundData)
    {
        FreeMemory(pSoundData);
        pSoundData = NULL;
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
