﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SimplePlayer.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/snd.h>

#include "SimplePlayer.h"

#include "WaveFileReader.h"
#include "WaveFileWriter.h"

namespace
{
    const int FILE_NUM = 4;
    const char* FILE_NAME[FILE_NUM] =
    {
        "rom:/yoshi_16bit_fs16kHz_1ch.wav",
        "rom:/wihaho_16bit_fs22kHz_1ch.wav",
        "rom:/fanfare_16bit_fs32kHz_1ch.wav",
        "rom:/L-sin440Hz_R-sin660Hz_8bit_fs44kHz_2ch.wav",
    };

    // Voice
    int s_CurrentVoice;
    nn::snd::Voice* s_pVoice[FILE_NUM];
}

void SimplePlayer::OnInitialize()
{
    // Initialize DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadDefaultComponent());
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

    m_CriticalSection.Initialize();
    s_CurrentVoice = 0;
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i] = NULL;
    }

    // Start sound thread
    m_SoundThreadFlag = true;
    m_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        reinterpret_cast<uptr>(this),
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );
}

void SimplePlayer::OnFinalize()
{
    // Destroy the sound thread
    m_SoundThreadFlag = false;
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    m_CriticalSection.Finalize();

    // Shut down the DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();
}

void SimplePlayer::OnUpdate(nn::hid::PadStatus& padStatus)
{
    static const f32 DELTA_PITCH = 0.1f;

    int voice = s_CurrentVoice;
    nn::snd::Voice* pVoice = s_pVoice[voice];

    if (padStatus.trigger & nn::hid::BUTTON_A)     // Voice playback and stop
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        if (pVoice->GetState() == nn::snd::Voice::STATE_PAUSE)
        {
            pVoice->SetState(nn::snd::Voice::STATE_PLAY);
        }
        else // if (pVoice->GetState() == nn::snd::Voice::STATE_PLAY)
        {
            pVoice->SetState(nn::snd::Voice::STATE_PAUSE);
        }
    }
    if (padStatus.trigger & nn::hid::BUTTON_UP)    // voice selection
    {
        voice--;
        if (voice < 0) voice = 0;
    }
    if (padStatus.trigger & nn::hid::BUTTON_DOWN)  // voice selection
    {
        voice++;
        if (voice >= FILE_NUM) voice = FILE_NUM - 1;
    }
    if (padStatus.trigger & nn::hid::BUTTON_RIGHT) // Change pitch
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        pVoice->SetPitch(pVoice->GetPitch() + DELTA_PITCH);
    }
    if (padStatus.trigger & nn::hid::BUTTON_LEFT)  // Change pitch
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        pVoice->SetPitch(nn::math::Max(pVoice->GetPitch() - DELTA_PITCH, DELTA_PITCH));
    }

    s_CurrentVoice = voice;
}

void SimplePlayer::OnDraw(demo::RenderSystemDrawing& renderSystem)
{
    int line = 1;

    renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    renderSystem.Clear();

    renderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    renderSystem.DrawText(10, 10 * line++, "CTR SampleDemos - snd/SimplePlayer");
    line++;

    renderSystem.DrawText(10, 10 * line++, "[A]          Play/Pause");
    renderSystem.DrawText(10, 10 * line++, "[Up/Down]    Change voice index");
    renderSystem.DrawText(10, 10 * line++, "[Left/Right] Change pitch");
    line++;

    for (int i = 0; i < FILE_NUM; i++)
    {
        if (s_pVoice[i] != NULL)
        {
            renderSystem.DrawText(
                10, 10 * line++, "%s Voice[%d]: (%s) pitch = %5.2f",
                ((i == s_CurrentVoice) ? ">" : " "), i,
                s_pVoice[i]->GetState() == nn::snd::Voice::STATE_PLAY ? " play" : "pause",
                s_pVoice[i]->GetPitch()
            );
        }
    }

    renderSystem.SwapBuffers();
    renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void SimplePlayer::SoundThreadFuncImpl()
{
    // Set sound output to stereo
    nn::snd::SetSoundOutputMode(nn::snd::OUTPUT_MODE_STEREO);

    // Set master volume
    nn::snd::SetMasterVolume(1.0);

    NN_LOG("Loading wav files...\n");

    u8* pMemory[FILE_NUM];
    nn::snd::WaveBuffer waveBuffer[FILE_NUM];
    // Open the file
    for (int i = 0; i < FILE_NUM; i++)
    {
        WaveFileReader reader(FILE_NAME[i]);
        int dataSize = reader.GetDataSize();
        int channelCount = reader.GetChannelCount();
        int sampleRate = reader.GetSampleRate();
        int quantumBits = reader.GetQuantumBits();

        // Get and check device memory
        pMemory[i] = reinterpret_cast<u8*>(AllocateMemory(dataSize, 32));
        NN_ASSERT(pMemory[i]);

        // Load wave data and invalidate cache
        reader.ReadData(pMemory[i], dataSize);
        nn::snd::FlushDataCache(reinterpret_cast<uptr>(pMemory[i]), dataSize);

        nn::snd::Voice* pVoice = nn::snd::AllocVoice(128, NULL, NULL);
        NN_ASSERT(pVoice);
        s_pVoice[i] = pVoice;

        nn::snd::SampleFormat format = (quantumBits == 8) ?
            nn::snd::SAMPLE_FORMAT_PCM8 : nn::snd::SAMPLE_FORMAT_PCM16;

        pVoice->SetChannelCount(channelCount);
        pVoice->SetSampleFormat(format);

        nn::snd::InitializeWaveBuffer(&waveBuffer[i]);
        waveBuffer[i].bufferAddress = pMemory[i];
        waveBuffer[i].sampleLength  = nn::snd::GetSampleLength(dataSize, format, channelCount);
        waveBuffer[i].loopFlag      = true;

        // Set volume
        nn::snd::MixParam mix;
        mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.707f; // Main volume (L)
        mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.707f; // Main volume (R)
        pVoice->SetMixParam(mix);
        pVoice->SetVolume(1.0f);

        // Set pitch
        pVoice->SetSampleRate(sampleRate);
        pVoice->SetPitch(1.0f);

        // Add buffer
        pVoice->AppendWaveBuffer(&waveBuffer[i]);
    }

    // Start playback
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i]->SetState(nn::snd::Voice::STATE_PLAY);
    }

    // The sound thread stops automatically during Sleep Mode
    // (It waits in nn::snd::WaitForDspSync).
    // However, if strict synchronization with graphics is required, the application must control it.
    // 
    while (m_SoundThreadFlag)
    {
        nn::snd::WaitForDspSync();      // Wait to receive data from DSP

        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        nn::snd::SendParameterToDsp();  // Send parameters to the DSP.
    }

    // End playback
    for (int i = 0; i < FILE_NUM; i++)
    {
        if (s_pVoice[i])
        {
            nn::snd::FreeVoice(s_pVoice[i]);
        }
        FreeMemory(pMemory[i]);
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
