﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     StreamPlayer.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/snd.h>

#include "StreamPlayer.h"

namespace
{
    // The number and size of buffers to use for stream playback
    const int WAVE_BUFFER_NUM = 4;
    const size_t WAVE_BUFFER_SIZE = 2048;

    // Communication queue between the sound thread and the file-reading thread
    nn::os::BlockingQueue s_Queue;
    const size_t QUEUE_BUFFER_SIZE = WAVE_BUFFER_NUM;
    uptr s_QueueBuffer[QUEUE_BUFFER_SIZE];

    // File-related
    const char FILE_NAME[] = "rom:/kart_title.32.dspadpcm.bcwav";
    const nn::snd::Bcwav::WaveInfo* pWaveInfo = NULL;
    const nn::snd::Bcwav::DspAdpcmInfo* pAdpcmInfo = NULL;
    nn::fs::FileReader s_FileReader;
    size_t s_HeaderSize;

    // Get the loop information and sample length that can be obtained with the next read operation
    s32 s_Position = 0;
    s32 GetAvailableSampleLength(s32 size, bool& isLooped)
    {
        // Determine whether this has looped
        if (s_Position == pWaveInfo->loopEndFrame)
        {
            s_Position = pWaveInfo->loopStartFrame;
            isLooped = true;
        }
        else
        {
            isLooped = false;
        }
        // Convert the requested buffer length into samples
        s32 nSamplesRequired = nn::snd::GetSampleLength(size, nn::snd::SAMPLE_FORMAT_ADPCM, 1);
        s32 ret;
        if (nSamplesRequired < pWaveInfo->loopEndFrame - s_Position)
        {
            ret = nSamplesRequired;
        }
        else
        {
            ret = pWaveInfo->loopEndFrame - s_Position;
        }
        s_Position += ret;
        return ret;
    }

    // File-reading thread
    const int FILE_READ_THREAD_PRIORITY = 4;
    const size_t FILE_READ_THREAD_STACK_SIZE = 4096;
    nn::os::Thread s_FileReadThread;
    bool s_FileReadThreadFlag;
    void FreadThreadFunc(uptr arg)
    {
        NN_UNUSED_VAR(arg);
        s32 offset = 0;
        while (s_FileReadThreadFlag)
        {
            // Check the queue every 5 ms
            nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(5));
            uptr address;
            while (s_Queue.TryDequeue(&address))
            {
                // Load file
                void* pBuffer = reinterpret_cast<void*>(address);
                size_t size = WAVE_BUFFER_SIZE;
                size_t sizeRead = s_FileReader.Read(pBuffer, size);
                offset += sizeRead;
                nn::snd::FlushDataCache(reinterpret_cast<uptr>(pBuffer), size);
                // If we have read to the end of the file, jump to the loop position
                if (offset >= nn::snd::Bcwav::FrameToByte(pWaveInfo->encoding, pWaveInfo->loopEndFrame))
                {
                    offset = nn::snd::Bcwav::FrameToByte(pWaveInfo->encoding, pWaveInfo->loopStartFrame);
                    s_FileReader.Seek(s_HeaderSize + offset, nn::fs::POSITION_BASE_BEGIN);
                }
            }
        }
    }
}

void StreamPlayer::OnInitialize()
{
    // Initialize DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadDefaultComponent());
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

    m_CriticalSection.Initialize();

    // Start sound thread
    m_SoundThreadFlag = true;
    m_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        reinterpret_cast<uptr>(this),
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );
}

void StreamPlayer::OnFinalize()
{
    // Destroy the sound thread
    m_SoundThreadFlag = false;
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    m_CriticalSection.Finalize();

    // Shut down the DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();
}

void StreamPlayer::OnUpdate(nn::hid::PadStatus& padStatus)
{
    NN_UNUSED_VAR(padStatus);
}

void StreamPlayer::OnDraw(demo::RenderSystemDrawing& renderSystem)
{
    int line = 1;

    renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    renderSystem.Clear();

    renderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    renderSystem.DrawText(10, 10 * line++, "CTR SampleDemos - snd/StreamPlayer");
    line++;

    renderSystem.SwapBuffers();
    renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void StreamPlayer::SoundThreadFuncImpl()
{
    // Set sound output to stereo
    nn::snd::SetSoundOutputMode(nn::snd::OUTPUT_MODE_STEREO);

    // Set master volume
    nn::snd::SetMasterVolume(1.0);

    void* pMemory[WAVE_BUFFER_NUM];
    for (s32 i = 0; i < WAVE_BUFFER_NUM; i++)
    {   // Prepare data buffer and generate data
        pMemory[i] = reinterpret_cast<void*>(AllocateMemory(WAVE_BUFFER_SIZE, 32));
        NN_ASSERT(pMemory[i]);
    }
    s_FileReader.Initialize(FILE_NAME);
    // Configure file reads to be high-priority so that sound doesn't break up during stream playback
    s_FileReader.SetPriority(nn::fs::PRIORITY_APP_REALTIME);

    // For now, read 1 KB
    void* pBcwavHeader = reinterpret_cast<void*>(AllocateMemory(0x400, 32));
    s_FileReader.Read(pBcwavHeader, 0x400);

    const nn::snd::Bcwav::WaveInfo& info = nn::snd::Bcwav::GetWaveInfo(pBcwavHeader);
    pWaveInfo = &info;

    s32 channelCount = nn::snd::Bcwav::GetChannelCount(pBcwavHeader);
    NN_ASSERT(channelCount == 1);

    if (info.encoding == nn::snd::Bcwav::ENCODING_DSP_ADPCM)
    {
        const nn::snd::Bcwav::DspAdpcmInfo* adpcm[2];
        for (s32 ch = 0; ch < channelCount; ch++)
        {
            adpcm[ch] = nn::snd::Bcwav::GetDspAdpcmInfo(pBcwavHeader, ch);
            if (ch == 0)
            {
                pAdpcmInfo = adpcm[ch];
            }
        }
    }

    // Obtain the offset from the start of the file to the data
    const void* pBcwavBody = nn::snd::Bcwav::GetWave(pBcwavHeader, nn::snd::Bcwav::CHANNEL_INDEX_L);
    s_HeaderSize = reinterpret_cast<size_t>(pBcwavBody) - reinterpret_cast<size_t>(pBcwavHeader);

    if (pBcwavHeader)
    {
        FreeMemory(pBcwavHeader);
        pBcwavHeader = NULL;
    }

    // Initialize file position
    s_FileReader.Seek(s_HeaderSize, nn::fs::POSITION_BASE_BEGIN);

    // Initialize the file-reading thread
    s_Queue.Initialize(s_QueueBuffer, QUEUE_BUFFER_SIZE);
    s_FileReadThreadFlag = true;
    s_FileReadThread.StartUsingAutoStack(
        FreadThreadFunc,
        NULL,
        FILE_READ_THREAD_STACK_SIZE,
        FILE_READ_THREAD_PRIORITY
    );

    // Initialize Voice
    nn::snd::Voice* pVoice;
    pVoice = nn::snd::AllocVoice(128, NULL, NULL);
    NN_ASSERT(pVoice);

    // Set volume
    nn::snd::MixParam mix;
    mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.707f; // Main volume (L)
    mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.707f; // Main volume (R)
    pVoice->SetMixParam(mix);
    pVoice->SetVolume(1.0f);

    if (pAdpcmInfo)
    {
        pVoice->SetAdpcmParam(pAdpcmInfo->param);
    }

    pVoice->SetChannelCount(1);
    pVoice->SetSampleFormat(static_cast<nn::snd::SampleFormat>(pWaveInfo->encoding));
    pVoice->SetSampleRate(pWaveInfo->sampleRate);
    pVoice->SetPitch(1.0);

    // Add buffer
    nn::snd::WaveBuffer waveBuffer[WAVE_BUFFER_NUM];
    for (s32 i = 0; i < WAVE_BUFFER_NUM; i++)
    {
        // File read request
        s_Queue.Enqueue(reinterpret_cast<uptr>(pMemory[i]));

        bool isLooped;
        nn::snd::InitializeWaveBuffer(&waveBuffer[i]);
        waveBuffer[i].bufferAddress = pMemory[i];
        waveBuffer[i].sampleLength  = GetAvailableSampleLength(WAVE_BUFFER_SIZE, isLooped);
        waveBuffer[i].loopFlag  = false;
        if (i == 0)
        {
            waveBuffer[i].pAdpcmContext = &pAdpcmInfo->context;
        }
        else if (isLooped)
        {
            waveBuffer[i].pAdpcmContext = &pAdpcmInfo->loopContext;
        }
        else
        {
            waveBuffer[i].pAdpcmContext = NULL;
        }
        pVoice->AppendWaveBuffer(&waveBuffer[i]);
    }
    // Wait until the queue for file read requests is empty
    {
        uptr tmp;
        while (s_Queue.TryGetFront(&tmp))
        {
            nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(5));
        }
    }

    pVoice->SetState(nn::snd::Voice::STATE_PLAY);

    // The sound thread is signaled approximately once every 5 milliseconds. When you use heavy processing, you should call only AppendWaveBuffer here and handle main processing in a separate thread.
    // 
    //  Otherwise, sound will stop.
    s32 currentBufferId = 0;
    while (m_SoundThreadFlag)
    {
        nn::snd::WaitForDspSync();  // Wait to receive data from DSP

        // If the buffer has finished playing, load sound again and register
        if (pVoice != NULL &&
            pVoice->GetState() == nn::snd::Voice::STATE_PLAY &&
            waveBuffer[currentBufferId].status == nn::snd::WaveBuffer::STATUS_DONE)
        {
            s32 id = currentBufferId;
            nn::snd::WaveBuffer* pBuf = &waveBuffer[id];

            s_Queue.Enqueue(reinterpret_cast<uptr>(pMemory[id]));

            // Add buffer again
            bool isLooped;
            nn::snd::InitializeWaveBuffer(pBuf);
            pBuf->bufferAddress = pMemory[id];
            pBuf->sampleLength  = GetAvailableSampleLength(WAVE_BUFFER_SIZE, isLooped);
            pBuf->loopFlag  = false;
            if (isLooped)
            {
                pBuf->pAdpcmContext = &pAdpcmInfo->loopContext;
            }
            else
            {
                pBuf->pAdpcmContext = NULL;
            }
            pVoice->AppendWaveBuffer(pBuf);

            if (++currentBufferId == WAVE_BUFFER_NUM)
            {
                currentBufferId = 0;
            }
        }

        nn::snd::SendParameterToDsp();  // Send parameters to the DSP.
    }

    // Release the Voice
    if (pVoice)
    {
        nn::snd::FreeVoice(pVoice);
        pVoice = NULL;
    }

    // Shut down the file-reading thread
    s_FileReadThreadFlag = false;
    s_FileReadThread.Join();
    s_FileReadThread.Finalize();
    s_Queue.Finalize();

    // Memory deallocation
    for (s32 i = 0; i < WAVE_BUFFER_NUM; i++)
    {
        if (pMemory[i])
        {
            FreeMemory(pMemory[i]);
        }
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
