/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     WaveBufferOperator.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/snd.h>

#include "WaveBufferOperator.h"

#include "WaveFileReader.h"

namespace
{
    const int FILE_NUM = 12;
    const char* FILE_NAME[FILE_NUM] =
    {
        "rom:/261.wav", "rom:/277.wav", "rom:/293.wav", "rom:/311.wav",
        "rom:/329.wav", "rom:/349.wav", "rom:/369.wav", "rom:/391.wav",
        "rom:/415.wav", "rom:/440.wav", "rom:/466.wav", "rom:/493.wav",
    };

    nn::snd::Voice* s_pVoice = NULL;
    const int WAVE_BUFFER_NUM = FILE_NUM;
    nn::snd::WaveBuffer s_WaveBuffer[WAVE_BUFFER_NUM];

    int s_WaveBufferIndex = 0;
    int UpdateWaveBufferIndex(int index, int delta)
    {
        if ((index = index + delta) >= WAVE_BUFFER_NUM)
        {
            index -= WAVE_BUFFER_NUM;
        }
        NN_ASSERT(index >= 0 && index < WAVE_BUFFER_NUM);
        return index;
    }
}

void WaveBufferOperator::OnInitialize()
{
    // Initialize DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadDefaultComponent());
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

    m_CriticalSection.Initialize();

    // Start sound thread
    m_SoundThreadFlag = true;
    m_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        reinterpret_cast<uptr>(this),
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );
}

void WaveBufferOperator::OnFinalize()
{
    // Destroy the sound thread
    m_SoundThreadFlag = false;
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    m_CriticalSection.Finalize();

    // Shut down the DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();
}

void WaveBufferOperator::OnUpdate(nn::hid::PadStatus& padStatus)
{
    if (padStatus.trigger & nn::hid::BUTTON_A)
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        if (padStatus.hold & nn::hid::BUTTON_R)
        {
            // Cut sampleLength in half (valid only when loop playback is disabled)
            if (s_WaveBuffer[s_WaveBufferIndex].loopFlag == false)
            {
                for (int i = 0; i < WAVE_BUFFER_NUM; i++)
                {
                    s_WaveBuffer[i].sampleLength /= 2;
                    s_pVoice->UpdateWaveBuffer(&s_WaveBuffer[i]);
                }
            }
        }
        else
        {
            // Reverse loopFlag
            int index = s_WaveBufferIndex;
            if (padStatus.hold & nn::hid::BUTTON_L)
            {
                // Set the next WaveBuffer as the one to operate on
                index = UpdateWaveBufferIndex(index, 1);
            }
            s_WaveBuffer[index].loopFlag = !s_WaveBuffer[index].loopFlag;
            s_pVoice->UpdateWaveBuffer(&s_WaveBuffer[index]);
        }
    }
    if (padStatus.trigger & nn::hid::BUTTON_B)
    {
        // Call DeleteWaveBuffer six times, starting with the current WaveBuffer and once on each consecutive buffer
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        int index = s_WaveBufferIndex;
        for (int i = 0; i < 6; i++)
        {
            s_pVoice->DeleteWaveBuffer(&s_WaveBuffer[index]);
            index = UpdateWaveBufferIndex(index, 1);
        }
    }
    if (padStatus.trigger & nn::hid::BUTTON_X)
    {
        // Call DeleteWaveBuffer four times, starting with the next WaveBuffer skipping every other buffer
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        int index = UpdateWaveBufferIndex(s_WaveBufferIndex, 1);
        for (int i = 0; i < 4; i++)
        {
            s_pVoice->DeleteWaveBuffer(&s_WaveBuffer[index]);
            index = UpdateWaveBufferIndex(index, 2);
        }
    }
    if (padStatus.trigger & nn::hid::BUTTON_Y)
    {
        // Call DeleteWaveBuffer three times, starting with the WaveBuffer after next and skipping two buffers at a time
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        int index = UpdateWaveBufferIndex(s_WaveBufferIndex, 2);
        for (int i = 0; i < 3; i++)
        {
            s_pVoice->DeleteWaveBuffer(&s_WaveBuffer[index]);
            index = UpdateWaveBufferIndex(index, 3);
        }
    }
}

void WaveBufferOperator::OnDraw(demo::RenderSystemDrawing& renderSystem)
{
    int line = 1;

    renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    renderSystem.Clear();

    renderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    renderSystem.DrawText(10, 10 * line++, "CTR SampleDemos - snd/WaveBufferOperator");
    line++;

    renderSystem.DrawText(10, 10 * line++, "[A]   Change loopFlag of the current");
    renderSystem.DrawText(10, 10 * line++, "      WaveBuffer");
    renderSystem.DrawText(10, 10 * line++, "[L+A] Change loopFlag of the next");
    renderSystem.DrawText(10, 10 * line++, "      WaveBuffer");
    renderSystem.DrawText(10, 10 * line++, "[R+A] Halve sampleLength of all");
    renderSystem.DrawText(10, 10 * line++, "      WaveBuffers (valid when loopFlag");
    renderSystem.DrawText(10, 10 * line++, "      of the current buffer is false)");
    renderSystem.DrawText(10, 10 * line++, "[B]   Delete WaveBuffer (pattern 1)");
    renderSystem.DrawText(10, 10 * line++, "[X]   Delete WaveBuffer (pattern 2)");
    renderSystem.DrawText(10, 10 * line++, "[Y]   Delete WaveBuffer (pattern 3)");
    line++;

    const char waveBufferStatus[] = { 'F', 'W', 'P', 'D', 'T' };
    for (int i = 0; i < WAVE_BUFFER_NUM; i++)
    {
        renderSystem.DrawText(10, 10 * line, "status: ");
        renderSystem.DrawText(74 + 10 * i, 10 * line,
                              "%c", waveBufferStatus[s_WaveBuffer[i].status]);
        renderSystem.DrawText(10, 10 * (line + 1), "  loop: ");
        renderSystem.DrawText(74 + 10 * i, 10 * (line + 1),
                              "%c", s_WaveBuffer[i].loopFlag ? 't' : 'f');
    }
    line += 2;

    renderSystem.DrawText(10, 10 * line++, "status 'F': FREE");
    renderSystem.DrawText(10, 10 * line++, "status 'W': WAIT");
    renderSystem.DrawText(10, 10 * line++, "status 'P': PLAY");
    renderSystem.DrawText(10, 10 * line++, "status 'D': DONE");
    renderSystem.DrawText(10, 10 * line++, "status 'T': TO_BE_DELETED");

    renderSystem.SwapBuffers();
    renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void WaveBufferOperator::SoundThreadFuncImpl()
{
    // Set sound output to stereo
    nn::snd::SetSoundOutputMode(nn::snd::OUTPUT_MODE_STEREO);

    // Set master volume
    nn::snd::SetMasterVolume(1.0);

    NN_LOG("Loading wav files...\n");

    u8* pMemory[FILE_NUM];

    int size = 0;
    int quantumBits = 0;
    int channels = 0;
    int sampleRate = 0;

    // Open the file
    for (int i = 0; i < FILE_NUM; i++)
    {
        WaveFileReader reader(FILE_NAME[i]);

        if (i == 0)
        {
            size = reader.GetDataSize();
            quantumBits = reader.GetQuantumBits();
            channels = reader.GetChannelCount();
            sampleRate = reader.GetSampleRate();
        }
        else
        {
            NN_ASSERT(size == reader.GetDataSize());
            NN_ASSERT(quantumBits == reader.GetQuantumBits());
            NN_ASSERT(channels == reader.GetChannelCount());
            NN_ASSERT(sampleRate == reader.GetSampleRate());
        }

        pMemory[i] = reinterpret_cast<u8*>(AllocateMemory(size, 32));
        NN_ASSERT(pMemory[i]);

        // Load wave data and invalidate cache
        reader.ReadData(pMemory[i], size);
        nn::snd::FlushDataCache(reinterpret_cast<uptr>(pMemory[i]), size);
    }

    // Initialize a Voice and add a WaveBuffer
    s_pVoice = nn::snd::AllocVoice(128, NULL, NULL);
    NN_ASSERT(s_pVoice);

    nn::snd::SampleFormat format = (quantumBits == 8) ?
        nn::snd::SAMPLE_FORMAT_PCM8 : nn::snd::SAMPLE_FORMAT_PCM16;

    s_pVoice->SetChannelCount(channels);
    s_pVoice->SetSampleFormat(format);

    // Set volume
    nn::snd::MixParam mix;
    mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.707f; // Main volume (L)
    mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.707f; // Main volume (R)
    s_pVoice->SetMixParam(mix);
    s_pVoice->SetVolume(1.0f);

    // Set pitch
    s_pVoice->SetSampleRate(sampleRate);
    s_pVoice->SetPitch(1.0f);

    // Add buffer
    for (int i = 0; i < WAVE_BUFFER_NUM; i++)
    {
        nn::snd::InitializeWaveBuffer(&s_WaveBuffer[i]);
        s_WaveBuffer[i].bufferAddress = pMemory[i];
        s_WaveBuffer[i].sampleLength = nn::snd::GetSampleLength(size, format, channels);
        s_WaveBuffer[i].loopFlag = false;
        s_pVoice->AppendWaveBuffer(&s_WaveBuffer[i]);
    }

    // Start playback
    s_pVoice->SetState(nn::snd::Voice::STATE_PLAY);

    // The sound thread stops automatically during Sleep Mode
    // (It waits in nn::snd::WaitForDspSync).
    // However, if strict synchronization with graphics is required, the application must control it.
    // 
    while (m_SoundThreadFlag)
    {
        nn::snd::WaitForDspSync();      // Wait to receive data from DSP

        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        while (s_WaveBuffer[s_WaveBufferIndex].status == nn::snd::WaveBuffer::STATUS_DONE)
        {
            int i = s_WaveBufferIndex;
            nn::snd::InitializeWaveBuffer(&s_WaveBuffer[i]);
            s_WaveBuffer[i].bufferAddress = pMemory[i];
            s_WaveBuffer[i].sampleLength = nn::snd::GetSampleLength(size, format, channels);
            s_WaveBuffer[i].loopFlag = false;
            s_pVoice->AppendWaveBuffer(&s_WaveBuffer[i]);
            s_WaveBufferIndex = UpdateWaveBufferIndex(s_WaveBufferIndex, 1);
        }
        nn::snd::SendParameterToDsp();  // Send parameters to the DSP.
    }

    // End playback
    if (s_pVoice)
    {
        nn::snd::FreeVoice(s_pVoice);
    }
    for (int i = 0; i < FILE_NUM; i++)
    {
        if (pMemory[i])
        {
            FreeMemory(pMemory[i]);
        }
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
