/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SndDemoBase.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47435 $
 *---------------------------------------------------------------------------*/

#include <nn/fs.h>

#include "SndDemoBase.h"
#include "applet.h"

void SndDemoBase::Initialize()
{
    // Applet-related initialization
    // Sleep Mode is rejected automatically until TransitionHandler::EnableSleep is called.
    TransitionHandler::Initialize();
}

void SndDemoBase::Finalize()
{
    // Applet-related finalization
    TransitionHandler::Finalize();

    // Finalize the application. Calls to nn::applet::CloseApplication do not return.
    nn::applet::CloseApplication();
}

void SndDemoBase::InitializeInternal()
{
    // fs initialization
    nn::fs::Initialize();
#ifndef SND_DEMO_BASE_NO_ROMFS
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
#endif // SND_DEMO_BASE_NO_ROMFS

    // Initialize hid
    NN_UTIL_PANIC_IF_FAILED(nn::hid::Initialize());
    m_pPadReader = new (m_PadReaderBuffer) nn::hid::PadReader();

    // Initialize device memory
    m_DeviceHeap.Initialize(
        nn::os::GetDeviceMemoryAddress(),
        DEVICE_MEMORY_SIZE
    );

    // Initialize the RenderSystem
    m_HeapAddressForGx = reinterpret_cast<uptr>(m_DeviceHeap.Allocate(MEMORY_SIZE_FOR_GX));
    NN_NULL_ASSERT(m_HeapAddressForGx);
    m_RenderSystem.Initialize(m_HeapAddressForGx, MEMORY_SIZE_FOR_GX);

    OnInitialize();
}

void SndDemoBase::FinalizeInternal()
{
    OnFinalize();

    // Finalize the RenderSystem
    m_RenderSystem.Finalize();
    FreeMemory(reinterpret_cast<void*>(m_HeapAddressForGx));

    // Free device memory
    m_DeviceHeap.Finalize();

    // Finalize the HID library
    nn::hid::Finalize();
}

void SndDemoBase::Start()
{
    // We need to check for exit notifications here.
    if (!TransitionHandler::IsExitRequired())
    {
        // Initialization
        InitializeInternal();

        // After this, the application itself handles Sleep Mode.
        TransitionHandler::EnableSleep();

        // Main loop
        while (true)
        {
            // Pad processing
            nn::hid::PadStatus padStatus;
            m_pPadReader->ReadLatest(&padStatus);
            OnUpdate(padStatus);

            // Rendering
            OnDraw(m_RenderSystem);

            // Sleep Mode, HOME Menu, and POWER Menu transitions
            TransitionHandler::Process();
            // Determine whether there has been an exit notification.
            if (TransitionHandler::IsExitRequired())
            {
                break;
            }
        }

        // End processing
        FinalizeInternal();
    }
}

void SndDemoBase::SoundThreadFunc(uptr arg)
{
    SndDemoBase* pApp = reinterpret_cast<SndDemoBase*>(arg);
    NN_ASSERT(pApp);
    pApp->SoundThreadFuncImpl();
}

void* SndDemoBase::AllocateMemory(size_t size, int alignment)
{
    return m_DeviceHeap.Allocate(size, alignment);
}

void SndDemoBase::FreeMemory(void* p)
{
    if (p)
    {
        m_DeviceHeap.Free(p);
    }
}
