﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/init.h>
#include <nn/os.h>
#include <nn/ac.h>
#include <nn/fs.h>
#include <nn/math.h>
#include <nn/applet.h>
#include <nn/socket_Initialize.h>
#include <cstdio>

namespace {
    nn::Result InitializeNetwork(void)
    {
        nn::Result result;
        nn::ac::Config config;

        result = nn::ac::Initialize();
        NN_UTIL_PANIC_IF_FAILED(result);

        // Create parameters for connection request
        result = nn::ac::CreateDefaultConfig( &config );
        if (result.IsFailure())
        {
            NN_UTIL_PANIC_IF_FAILED(nn::ac::Finalize());
            return result;
        }

        // Invalidate connection test for debugging
        nn::ac::DebugSetNetworkArea( &config, nn::ac::NETWORK_AREA_LAN );

        // Issue connection request
        result = nn::ac::Connect( config );
        if (result.IsFailure())
        {
            NN_UTIL_PANIC_IF_FAILED(nn::ac::Finalize());
            return result;
        }

        return nn::ResultSuccess();
    }

    nn::Result FinalizeNetwork(void)
    {
        nn::Result result;

        // Create parameters for connection request
        result = nn::ac::Close();
        NN_UTIL_RETURN_IF_FAILED(result);

        result = nn::ac::Finalize();
        NN_UTIL_PANIC_IF_FAILED(result);

        return nn::ResultSuccess();
    }

#ifndef NN_SWITCH_DISABLE_DEBUG_PRINT
    const char* GetAddressString(nn::socket::InAddr addr)
    {
        static char buffer[16];
        u8* p = reinterpret_cast<u8*>(&addr);
        std::snprintf(buffer, sizeof(buffer), "%d.%d.%d.%d", p[0], p[1], p[2], p[3]);
        return buffer;
    }
#endif
}

extern "C" void nnMain( void )
{
    nn::fs::Initialize();

    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    nn::Result result;

    do
    {
        NN_LOG("Initializing the network.\n");

        // Initialize network connection using network settings written on system
        result = InitializeNetwork();
        NN_UTIL_PANIC_IF_FAILED(result);

        {
            NN_LOG("Initializing socket..\n");

            // Use socket API from a single thread
            const s32    sessionCount           = 1;
            // Allocate 64 KB as send/receive buffer for the socket
            const size_t bufferSizeForSockets   = 65536;
            // Find the work size required by the socket library
            const size_t workSizeForLibrary     = nn::socket::GetRequiredMemorySize(bufferSizeForSockets, sessionCount);

            // Allocate working memory and align to 4 KB.
            u8* pWorkMemory = new u8[workSizeForLibrary + 4096];
            uptr workMemoryAddress = nn::math::RoundUp<uptr>(reinterpret_cast<uptr>(pWorkMemory), 4096);

            // Initialize socket library
            result = nn::socket::Initialize(workMemoryAddress, workSizeForLibrary, bufferSizeForSockets, sessionCount);
            NN_UTIL_PANIC_IF_FAILED(result);

            {
                s32 ret;
                nn::socket::InAddr addr, netmask;
                ret = nn::socket::GetPrimaryAddress(reinterpret_cast<u8*>(&addr), reinterpret_cast<u8*>(&netmask));
                NN_ASSERT(ret == 0);
                NN_LOG("host       : %s\n", GetAddressString(addr));
                NN_LOG("netmask    : %s\n", GetAddressString(netmask));

                nn::socket::InAddr dns1, dns2;
                ret = nn::socket::GetResolver(reinterpret_cast<u8*>(&dns1), reinterpret_cast<u8*>(&dns2));
                if (ret == 0)
                {
                    NN_LOG("dns1       : %s\n", GetAddressString(dns1));
                    NN_LOG("dns2       : %s\n", GetAddressString(dns2));
                }

                nn::socket::InAddr gateway;
                ret = nn::socket::GetDefaultGateway(reinterpret_cast<u8*>(&gateway));
                if (ret == 0)
                {
                    NN_LOG("gateway    : %s\n", GetAddressString(gateway));
                }

                NN_LOG("\n");
#ifndef NN_SWITCH_DISABLE_DEBUG_PRINT
                nn::socket::DumpRoutingTable();
#endif
                NN_UNUSED_VAR(ret);
            }

            NN_LOG("Finalizing socket..\n");
            // Finalize socket library
            result = nn::socket::Finalize();
            NN_UTIL_PANIC_IF_FAILED(result);

            // If the Finalize function is called on the socket library, the working region can be reused.
            delete [] pWorkMemory;
        }

        NN_LOG("Finalizing network.\n");
        result = FinalizeNetwork();
        NN_UTIL_PANIC_IF_FAILED(result);

    } while(false);
}

