﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/types.h>
#include <nn/dbg.h>

#include <nn/os/os_Memory.h>
#include <nn/os/os_MemoryTypes.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include <nn/gx.h>
#include <nn/fs.h>

#include "demo.h"

#include <nn/uds.h>
#include "uds.h"

/* Please see man pages for details



*/

// uds.cpp global variables declared as extern
extern volatile bool isDetectError;

namespace 
{
    const int GL_MEMORY_SIZE = 0x800000;
    
    enum COMMAND_TYPE
    {
        SCAN,
        CREATE_NETWORK,
        CONNECT_AS_CLIENT,
        CONNECT_AS_SPECTATOR,
        EJECT_ALL_CLIENTS,
        DESTROY_NETWORK,
        DISCONNECT_NETWORK,
        DISALLOW_TO_CONNECT,
        ALLOW_TO_CONNECT,
        EJECT_SPECTATOR,
        ALLOW_TO_SPECTATE,
        COMMAND_MAX
    };

    const char* COMMAND_NAME[COMMAND_MAX] =
    {
        "SCAN >",
        "< CREATE NETWORK >",
        "< CONNECT AS CLIENT >",
        "< CONNECT AS SPECTATOR >",
        "< EJECT ALL CLIENTS >",
        "< DESTROY NETWORK >",
        "< DISCONNECT NETWORK >",
        "< DISALLOW_TO_CONNECT >",
        "< ALLOW_TO_CONNECT >",
        "< EJECT_SPECTATOR >",
        "< ALLOW_TO_SPECTATE"
    };
   

    ComMode s_Mode = NOT_INITIALIZED;
    
    demo::RenderSystemDrawing s_RenderSystem;
    nn::fnd::ExpHeap s_AppHeap;
    uptr s_HeapForGx;
    
    NodeData                    s_NodeData[nn::uds::NODE_MAX];
    nn::uds::NetworkDescription s_NetDesc[SCAN_MAX];
    nn::uds::NodeInformation    s_MasterInfo[SCAN_MAX];

    nn::os::StackBuffer<4096>    s_MonitorStackBuffer; //For communications monitor thread
    nn::os::StackBuffer<4096>    s_MiscStackBuffer;    //For initialization and scan

    static nn::os::Thread s_MonitorThread;
    static nn::os::Thread s_InitializeThread;
    static nn::os::Thread s_ScanThread;

    nn::os::LightEvent s_AwakeEvent;         // Wake events
  
}

nn::applet::AppletQueryReply mySleepQueryCallback(uptr arg)
{
    NN_UNUSED_VAR(arg);

    // Although this demo enters Sleep Mode without finalizing the UDS library, we normally recommend that you finalize the UDS library beforehand.
    // 
    // If the system enters Sleep Mode before the library has been finalized, UDS library features will not be usable until Finalize has been called once after the system recovers from Sleep Mode.
    // 

    return nn::applet::CTR::REPLY_LATER;
}

void myAwakeCallback(uptr arg)
{
    NN_UNUSED_VAR(arg);
    NN_LOG("AWAKE\n");
    s_AwakeEvent.Signal();
}

void nnMain( void )
{
    NN_LOG("START\n");
    NN_LOG("X + Y + L + R : End Demo\n");
    nn::Result result;

    //Sleep support
    nn::applet::SetSleepQueryCallback( mySleepQueryCallback, 0 );
    nn::applet::SetAwakeCallback( myAwakeCallback, 0 );
    nn::applet::Enable();
    s_AwakeEvent.Initialize(false);
    
   // Initialize hid
    result = nn::hid::Initialize();
    NN_UTIL_PANIC_IF_FAILED(result);

    // Create heap    
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), 
                        nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(GL_MEMORY_SIZE));
   
    // Allocates 8 MB memory in FCRAM for graphics    
    s_RenderSystem.Initialize(s_HeapForGx, GL_MEMORY_SIZE);
    
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, 0.0f, 0.0f, 0.0f, 0.0f);
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY1, 0.0f, 0.0f, 0.0f, 0.0f);

    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;
    
    int command = 0; //Commands
    int targetNetwork = 0;
    int foundNetwork = 0;

    u16 myNodeId = 0;
    int sendError = 0;

    while(true)
    {
        s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
        s_RenderSystem.Clear();
        s_RenderSystem.SetFontSize(8);
        s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f);
        
        padReader.ReadLatest(&padStatus);
        if(padStatus.hold == (nn::hid::BUTTON_X | nn::hid::BUTTON_Y | nn::hid::BUTTON_L | nn::hid::BUTTON_R))
        {
            break;
        }
        switch( s_Mode )
        {
        case NOT_INITIALIZED:
            {
                s_RenderSystem.DrawText(0, 0, "A: Initialize uds Library %s", isDetectError?"(ERROR)":"" );
                if( padStatus.trigger & nn::hid::BUTTON_A )
                {
                    s_InitializeThread.Start( UdsInitializeThread, NULL, s_MiscStackBuffer );
                    s_Mode = INITIALIZING;
                }
            }break;
            
        case INITIALIZING:
            {
                s_RenderSystem.DrawText(0, 0, "Initializing uds Library...");
                if ( !s_InitializeThread.IsAlive() )
                {
                    s_InitializeThread.Join();
                    s_InitializeThread.Finalize();
                    
                    if( isDetectError )
                    {
                        s_Mode = NOT_INITIALIZED;
                    }
                    else
                    {
                        s_Mode = DISCONNECTED;
                    }
                }
            }break;
            
        case ERROR:  // If the system enters sleep or WiFi OFF mode during this, UDS must be initialized again.
            {
                s_RenderSystem.DrawText(0, 0, "A: Finalize uds Library");
                if( padStatus.trigger & nn::hid::BUTTON_A )
                {
                    //Blocks until processing is complete, but in this demo, in the interest of readability, it runs from the main thread.
                    nn::uds::Finalize();
                    isDetectError = false;
                    s_Mode = NOT_INITIALIZED;
                }
            }break;

        case SCANNING:
            {
                s_RenderSystem.DrawText(0, 0, "Scanning UDS Networks...");
                if ( !s_ScanThread.IsAlive() )
                {
                    s_ScanThread.Join();
                    s_ScanThread.Finalize();
                    
                    //Gets the screen display data for the scan results
                    GetNetDescList(&foundNetwork, s_NetDesc, s_MasterInfo );

                    if( isDetectError )
                    {
                        s_Mode = ERROR;
                    }
                    else
                    {
                        s_Mode = DISCONNECTED;
                    }
                }
            }break;

        default:
            {
                s_RenderSystem.DrawText(0, 0, "UP/DOWN: SELECT NETWORK, LEFT/RIGHT: SELECT MODE");
                s_RenderSystem.DrawText(0, 180, "%s", COMMAND_NAME[command]);
            
                if( padStatus.trigger & nn::hid::BUTTON_RIGHT )
                {
                    if( command < COMMAND_MAX-1 ) command++;
                }
                if( padStatus.trigger & nn::hid::BUTTON_LEFT )
                {
                    if ( command > 0) command--;
                }
                if( padStatus.trigger & nn::hid::BUTTON_UP )
                {
                    if( targetNetwork > 0 ) targetNetwork--;
                }
                if( padStatus.trigger & nn::hid::BUTTON_DOWN )
                {
                    if( targetNetwork < foundNetwork -1 ) targetNetwork++;
                }

                if( padStatus.trigger & nn::hid::BUTTON_A )    //Run the selected command
                {
                    switch(command)
                    {
                    case SCAN:  //Scan
                        if( s_Mode==DISCONNECTED )
                        {
                            s_Mode = SCANNING;
                            s_ScanThread.Start( UdsScanThread, NULL, s_MiscStackBuffer );
                        }break;
                        
                    case CREATE_NETWORK:    // Start operation as Master
                        if( s_Mode==DISCONNECTED )
                        {
                            s_Mode = MASTER;
                            ConnectParam param;
                                param.mode = MASTER;
                                param.targetNum = 0;
                                
                            s_MonitorThread.Start( ConnectionMonitorThread, param, s_MonitorStackBuffer, CONNECTION_MONITOR_THREAD_PRIORITY );
                        }break;
                        
                    case CONNECT_AS_CLIENT:
                        if( s_Mode==DISCONNECTED )
                        {
                            s_Mode = CLIENT;
                            ConnectParam param;
                                param.mode = CLIENT;
                                param.targetNum = targetNetwork;
                            
                            s_MonitorThread.Start( ConnectionMonitorThread, param, s_MonitorStackBuffer, CONNECTION_MONITOR_THREAD_PRIORITY );
                        }break;
                        
                    case CONNECT_AS_SPECTATOR:
                        if( s_Mode== DISCONNECTED )
                        {
                            s_Mode = SPECTATOR;
                            ConnectParam param;
                                param.mode = SPECTATOR;
                                param.targetNum = targetNetwork;
                            
                            s_MonitorThread.Start( ConnectionMonitorThread, param, s_MonitorStackBuffer, CONNECTION_MONITOR_THREAD_PRIORITY );
                        }break;
                        
                    case EJECT_ALL_CLIENTS:
                        if(s_Mode == MASTER)
                        {
                            //Blocks until processing is complete, but in this demo, in the interest of readability, it runs from the main thread.
                            result = nn::uds::EjectClient( nn::uds::BROADCAST_NODE_ID );
                            if( result.IsFailure() )
                            {
                                NN_LOG("EjectClient Error\n");
                            }
                        }break;
                        
                    case DESTROY_NETWORK:
                        if( s_Mode==MASTER)
                        {
                            //Blocks until processing is complete, but in this demo, in the interest of readability, it runs from the main thread.
                            result = nn::uds::DestroyNetwork();
                            if( result.IsFailure() )
                            {
                                NN_LOG("DestroyNetwork Error\n");
                            }
                        }break;
                        
                    case DISCONNECT_NETWORK:
                        if( s_Mode==CLIENT | s_Mode==SPECTATOR)
                        {
                            //Blocks until processing is complete, but in this demo, in the interest of readability, it runs from the main thread.
                            result = nn::uds::DisconnectNetwork();
                            if( result.IsFailure() )
                            {
                                NN_LOG("DisconnectNetwork Error\n");
                            }
                        }
                        break;
                        
                    case DISALLOW_TO_CONNECT:
                        if( s_Mode == MASTER )
                        {
                            //Blocks until processing is complete, but in this demo, in the interest of readability, it runs from the main thread.
                            result = nn::uds::DisallowToConnect(false); //In this demo, devices that connect once are permitted to connect again.
                            if( result.IsFailure() )
                            {
                                NN_LOG("DisallowToConnect Error\n");
                            }
                        }
                        break;
                        
                    case ALLOW_TO_CONNECT:
                        if( s_Mode == MASTER )
                        {
                            //Blocks until processing is complete, but in this demo, in the interest of readability, it runs from the main thread.
                            result = nn::uds::AllowToConnect();
                            if( result.IsFailure() )
                            {
                                NN_LOG("AllowToConnect Error\n");
                            }
                        }
                        break;
                        
                    case EJECT_SPECTATOR:
                        if( s_Mode == MASTER )
                        {
                            //Blocks until processing is complete, but in this demo, in the interest of readability, it runs from the main thread.
                            result = nn::uds::EjectSpectator();
                            if( result.IsFailure() )
                            {
                                NN_LOG("EjectSpectator Error\n");
                            }
                        }
                        break;
                        
                    case ALLOW_TO_SPECTATE:
                        if( s_Mode == MASTER )
                        {
                            //Blocks until processing is complete, but in this demo, in the interest of readability, it runs from the main thread.
                            result = nn::uds::AllowToSpectate();
                            if( result.IsFailure() )
                            {
                                NN_LOG("AllowToSpectate Error\n");
                            }
                        }
                        break;
                        
                    }
                }
            }
        }
        
        if( s_Mode== DISCONNECTED )
        {
            for(int i=0;i<foundNetwork;i++)
            {
                if( targetNetwork == i )
                {
                    s_RenderSystem.SetColor(1.0f, 0.0f, 0.0f);
                }
                else
                {
                    s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f);
                }

                char name[11] = "\0";
                if( s_MasterInfo[i].userName.isNgUserName == true )
                {
                    // If the user name is a profanity word, change to ***********
                    std::memset(name, '*', 11);
                }
                else
                {
                    std::wcstombs(name, s_MasterInfo[i].userName.userName, 11);
                }
            
                s_RenderSystem.DrawText(0, (i*3+2)*8, "No.%2d: ID: %08X-%02X-%08X", i, s_NetDesc[i].GetLocalCommunicationId(), s_NetDesc[i].GetSubId(), s_NetDesc[i].GetTemporaryId() );
                s_RenderSystem.DrawText(0, (i*3+3)*8, "     : Ch: %02d, Entry: %02d/%02d Spectate:%s", s_NetDesc[i].GetChannel(), s_NetDesc[i].GetNowEntry(), s_NetDesc[i].GetMaxEntry(), s_NetDesc[i].CanConnectAsSpectator()?"OK":"NG" );
                s_RenderSystem.DrawText(0, (i*3+4)*8, "     : NAME: %s ", name );
            }
            s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f);
        }
        
        s_RenderSystem.SwapBuffers();

        //Display the communication status
        s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
        s_RenderSystem.Clear();        

        if( s_Mode == MASTER || s_Mode == CLIENT || s_Mode == SPECTATOR )
        {
            //Confirm the connection status
            if( !s_MonitorThread.IsAlive() )
            {
                s_MonitorThread.Join();
                s_MonitorThread.Finalize();
                
                myNodeId = 0;
                sendError = 0;
                if( isDetectError )
                {
                    s_Mode = ERROR;
                }
                else
                {
                    s_Mode = DISCONNECTED;
                }
            }
            else
            {
                //Communication in progress, so display the communication status
                GetResult( &myNodeId, &sendError, s_NodeData );

                s_RenderSystem.DrawText(0, 180, "MyNodeID: %02d, LINK: %d", myNodeId, nn::uds::GetLinkLevel() );
                s_RenderSystem.DrawText(0, 188, "SendError:%04d",sendError);
                
                for(int i=0;i<nn::uds::NODE_MAX;i++)
                {
                    if( s_NodeData[i].nodeId == 0 ) continue; //Skip unconnected nodes
                
                    if ( s_NodeData[i].nodeId == myNodeId ) //Do not take local node data
                    {
                        s_RenderSystem.SetColor(0.0f, 1.0f, 1.0f);
                        s_RenderSystem.DrawText(0, (i+1) * 8, "%02d MYSELF   ----------- ----------- [%s]", s_NodeData[i].nodeId, s_NodeData[i].name);
                        s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f);
                    }
                    else
                    {
                        s_RenderSystem.DrawText(0, (i+1) * 8, "%02d %08X %04.3fKbps %02.2f/%02.2f [%s]", s_NodeData[i].nodeId, s_NodeData[i].latestCount, 
                                                    s_NodeData[i].lapResult.receivedSize * 8.0/1000, s_NodeData[i].lapResult.errorNum*100.0/s_NodeData[i].lapResult.receivedTime,
                                                    s_NodeData[i].totalResult.errorNum*100.0/s_NodeData[i].totalResult.receivedTime, s_NodeData[i].name);
                    }
                }
            }
        }
        
        s_RenderSystem.SwapBuffers();
        s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

        //Sleep support
        if ( nn::applet::IsExpectedToReplySleepQuery() )
        {
            nn::applet::ReplySleepQuery(nn::applet::REPLY_ACCEPT);
            s_AwakeEvent.Wait();

            // Processing after restart
            nn::gx::StartLcdDisplay();
        }
        
        //HOME Button support
        if ( nn::applet::IsExpectedToProcessHomeButton() )
        {
        
            // While the HOME Menu is being displayed, communications can be maintained on another thread.
            // However, the processing capacity of the application will drop below its normal level, so we recommend limiting the amount of communication when displaying the HOME Menu.
            // 

            nn::applet::ProcessHomeButton();
            AppletWakeupState r = nn::applet::WaitForStarting();
            
            // Recover the GPU register settings
            nngxUpdateState(NN_GX_STATE_ALL);
            nngxValidateState(NN_GX_STATE_ALL,GL_TRUE);
            
            if ( r == nn::applet::WAKEUP_BY_CANCEL )
            {
                break;
            }
        }
        
        //Support for POWER Button
        if ( nn::applet::IsExpectedToProcessPowerButton() )
        {
            nn::applet::ProcessPowerButton();
            nn::applet::WaitForStarting();
            if ( nn::applet::IsExpectedToCloseApplication() )
            {
              break;
            }
        }
    }

    if(s_Mode==MASTER)nn::uds::DestroyNetwork();
    if(s_Mode==CLIENT|s_Mode==SPECTATOR)nn::uds::DisconnectNetwork();

    nn::uds::Finalize();
    nn::hid::Finalize();

    nn::applet::DisableSleep();
    s_AwakeEvent.Finalize();
    nn::applet::SetSleepQueryCallback(NULL, 0);
    nn::applet::SetAwakeCallback(NULL, 0);
    nn::applet::CloseApplication();
}
