﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     uds.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/Result.h>
#include <nn/uds.h>

#include "uds.h"
#include "uniqueId_List.h"

/* Please see man pages for details 
    
              
*/



volatile bool isDetectError = false;

namespace
{

    nn::os::Event s_StatusUpdateEvent;
    nn::os::Event s_DisconnectEvent; //For detection of when the ResultThread terminates
    nn::os::CriticalSection s_ResultLock = nn::WithInitialize();

    NodeData s_NodeData[nn::uds::NODE_MAX];          // for rendering

    nn::os::StackBuffer<4096>    s_SendStackBuffer;
    nn::os::StackBuffer<4096>    s_ReceiveStackBuffer;
    nn::os::StackBuffer<4096>    s_ResultStackBuffer;

    bit8 s_UdsReceiveBuffer[4096*20] NN_ATTRIBUTE_ALIGN(4096);  //Receive buffer for system. Must be a multiple of 4096 because the buffer is 4096-byte aligned.

    int s_SendError = 0;
    u16 s_MyNodeId  = 0;

    int s_ResultLoopCounter = 0;

    int s_FoundNetworkNum = 0;
    nn::uds::NetworkDescription s_NetDesc[SCAN_MAX];
    nn::uds::NodeInformation    s_MasterInfo[SCAN_MAX];

    nn::os::Thread s_SendThread;
    nn::os::Thread s_RecvThread;
    nn::os::Thread s_ResultThread;
}


void UdsSendThread(uptr param);
void UdsRecvThread(uptr param);

//Initialization thread
void UdsInitializeThread(uptr param)
{
NN_UNUSED_VAR( param );

    nn::Result result = nn::uds::Initialize( &s_StatusUpdateEvent, s_UdsReceiveBuffer, sizeof(s_UdsReceiveBuffer) );
    if( result.IsFailure() )
    {
        isDetectError = true;
        if( result == nn::uds::ResultAlreadyOccupiedWirelessDevice() )
        {
            NN_LOG(" Other Network Service is Already Occupying Wireless Device.\n"); //Application is already performing other communication.
        }
        else if( result == nn::uds::ResultWirelessOff() )
        {
            NN_LOG(" Wireless Off Mode Now.\n"); // Was in wireless OFF mode
        }
        else
        {
            NN_DBG_PRINT_RESULT(result);
        }
    }
    else
    {
        isDetectError = false;
    }
}

//Scan thread
void UdsScanThread(uptr param)
{
NN_UNUSED_VAR( param );
    static bit8 s_ScanBuffer[2048];

    s_FoundNetworkNum = 0;
    std::memset(s_NetDesc, 0x00, sizeof(s_NetDesc));
    std::memset(s_MasterInfo, 0x00, sizeof(s_MasterInfo));

    nn::Result result = nn::uds::Scan( s_ScanBuffer, sizeof(s_ScanBuffer), SUB_ID, nn::uds::CreateLocalCommunicationId(nn::uds::CTR::SampleDemos::UNIQUE_ID_SAMPLEDEMOS_UDS_COM_DEMO1) );
    if( result.IsSuccess() )
    {
        nn::uds::ScanResultReader networkList(s_ScanBuffer);

        //Get the top 5 networks
        s_FoundNetworkNum = networkList.GetCount();
        if( s_FoundNetworkNum > SCAN_MAX )
        {
            s_FoundNetworkNum = SCAN_MAX;
        }
        for( int i= 0;i < s_FoundNetworkNum; i++ )
        {
            nn::uds::NetworkDescriptionReader netDescReader = networkList.GetNextDescription();
            result = netDescReader.GetNetworkDescription(&s_NetDesc[i]);
            if( result.IsSuccess() )
            {
                static nn::uds::NodeInformation s_NodeInfo[nn::uds::NODE_MAX]; //Use static so that stack is not used up
                result = netDescReader.GetNodeInformationList(s_NodeInfo);
                if( result.IsSuccess() )
                {
                    std::memcpy(&s_MasterInfo[i], &s_NodeInfo[0], sizeof(nn::uds::NodeInformation));
                }
            }
        }
    }
    else
    {
        isDetectError = true;
    }
}

void GetNetDescList(int* pNum, nn::uds::NetworkDescription netDesc[SCAN_MAX], nn::uds::NodeInformation masterInfo[SCAN_MAX] )
{
    *pNum = s_FoundNetworkNum;
    std::memcpy( netDesc, s_NetDesc, sizeof(s_NetDesc) );
    std::memcpy( masterInfo, s_MasterInfo, sizeof(s_MasterInfo) );
}

//Send thread
void UdsSendThread(uptr param)
{
NN_UNUSED_VAR( param );

    nn::uds::EndpointDescriptor ed;
    nn::Result result = nn::uds::CreateEndpoint(&ed);
    NN_UTIL_PANIC_IF_FAILED(result);

    //Send at fixed intervals
    nn::os::Timer timer;
    timer.Initialize(false);
    timer.StartPeriodic( nn::fnd::TimeSpan::FromMilliSeconds(10), nn::fnd::TimeSpan::FromMilliSeconds(SEND_INTERVAL) );

    static u32 counter[PACKET_SIZE/sizeof(u32) +1];
    counter[0] = 0;

    while(true)
    {
        timer.Wait();
        result = nn::uds::SendTo(ed, counter, PACKET_SIZE, nn::uds::BROADCAST_NODE_ID, 1, nn::uds::NO_WAIT);

        if( result.IsFailure() )
        {
            if( result == nn::uds::ResultBufferIsFull() )
            {
                s_SendError++;
            }
            else
            {
                break;
            }
        }
        else
        {
            //If SendTo function fails, do not treat as packet loss
            counter[0]++;
        }
    }
    timer.Stop();
    timer.Finalize();
    //Disconnect
    NN_LOG("DestroyEndpoint\n");
    result = nn::uds::DestroyEndpoint(&ed);
    NN_UTIL_PANIC_IF_FAILED(result);

    NN_LOG("%s is Finish!\n", __func__);
}

//Receive thread
void UdsRecvThread(uptr param)
{
NN_UNUSED_VAR( param );

    nn::uds::EndpointDescriptor ed;
    nn::Result result;
    result = nn::uds::CreateEndpoint(&ed);
    NN_UTIL_PANIC_IF_FAILED(result);

    result = nn::uds::Attach(&ed, nn::uds::BROADCAST_NODE_ID, 1);
    NN_UTIL_PANIC_IF_FAILED(result);

    while(true)
    {
        u16    srcNodeId;
        size_t recvedSize;
        static u32 counter[PACKET_SIZE/sizeof(u32)+1];

        result = nn::uds::ReceiveFrom(ed, counter, &recvedSize, &srcNodeId, sizeof(counter));
        if( result.IsFailure() )
        {
            break;
        }

        s_ResultLock.Enter();
        //Search for data storage location
        int index = -1;
        for( int i=0; i< nn::uds::NODE_MAX; i++ )
        {
            if( s_NodeData[i].nodeId == srcNodeId )
            {
                index = i;
                break;
            }
        }
        if( index < 0 )
        {
            NN_LOG("Receive from Unknown Node (%02d)\n", srcNodeId);
            s_ResultLock.Leave();
            continue;
        }

        //Data storage
        //first time
        if( s_NodeData[index].latestCount == 0 )
        {
            s_NodeData[index].startCount = counter[0];
            s_NodeData[index].startTick  = nn::os::Tick::GetSystemCurrent();
        }
        else //Check for packet loss by comparing counter values starting from the second time
        {
            s_NodeData[index].errorCount += counter[0] - (s_NodeData[index].latestCount+1);
        }
        s_NodeData[index].receivedSize += recvedSize;
        s_NodeData[index].latestCount  =  counter[0];
        s_ResultLock.Leave();
    }

    NN_LOG("DestroyEndpoint\n");
    result = nn::uds::DestroyEndpoint(&ed);
    NN_UTIL_PANIC_IF_FAILED(result);

    NN_LOG("%s is Finish!\n", __func__);
}

//Communication status monitor thread
void ResultMoniterThread(uptr param)
{
NN_UNUSED_VAR( param );

    //Collect wrap data at fixed intervals
    nn::os::Timer timer;
    timer.Initialize(false);
    timer.StartPeriodic( nn::fnd::TimeSpan::FromSeconds(1), nn::fnd::TimeSpan::FromSeconds(1) );

    bool isExit =false;
    s_ResultLoopCounter = 0;
    while(!isExit)
    {
        nn::os::WaitObject* waitEvents[] ={ &s_DisconnectEvent, &timer };
        switch( nn::os::WaitObject::WaitAny(waitEvents, sizeof(waitEvents)/sizeof(nn::os::WaitObject*) ) )
        {
        case 0:
            isExit = true;
            break;

        case 1:
            {
                s_ResultLock.Enter();

                for(int i=0; i<nn::uds::NODE_MAX; i++)
                {
                    if( s_NodeData[i].nodeId == 0 ) continue; //not connected

                    s_NodeData[i].lapResult.receivedSize = s_NodeData[i].receivedSize - s_NodeData[i].totalResult.receivedSize;
                    s_NodeData[i].lapResult.receivedTime = ( s_NodeData[i].latestCount - s_NodeData[i].startCount ) - s_NodeData[i].totalResult.receivedTime;
                    s_NodeData[i].lapResult.errorNum     = s_NodeData[i].errorCount  - s_NodeData[i].totalResult.errorNum;

                    s_NodeData[i].totalResult.receivedSize = s_NodeData[i].receivedSize;
                    s_NodeData[i].totalResult.receivedTime = s_NodeData[i].latestCount - s_NodeData[i].startCount;
                    s_NodeData[i].totalResult.errorNum = s_NodeData[i].errorCount;
/*
                    NN_LOG("%02d %08X %04.3fKbps %02.2f/%02.2f [%s]\n", s_NodeData[i].nodeId, s_NodeData[i].latestCount,
                            s_NodeData[i].lapResult.receivedSize * 8.0/1000, s_NodeData[i].lapResult.errorNum*100.0/s_NodeData[i].lapResult.receivedTime,
                            s_NodeData[i].totalResult.errorNum*100.0/s_NodeData[i].totalResult.receivedTime, s_NodeData[i].name);
*/
                }
                s_ResultLock.Leave();
            }
        }
        s_ResultLoopCounter++;
    }

    timer.Stop();
    timer.Finalize();
    NN_LOG("%s is Finish!\n", __func__);
}

void GetResult(u16* pMyNodeId, int* pSendError, NodeData nodeData[nn::uds::NODE_MAX])
{
    static int s_PrevCount = 0;
    if( s_PrevCount != s_ResultLoopCounter) //Avoid needless copying
    {
        s_ResultLock.Enter();
            *pMyNodeId = s_MyNodeId;
            *pSendError = s_SendError;
            std::memcpy( nodeData, s_NodeData, sizeof(s_NodeData) );
        s_ResultLock.Leave();
    }
    s_PrevCount = s_ResultLoopCounter;
}

//Connection status monitor thread
void ConnectionMonitorThread(ConnectParam param)
{
    nn::Result result;

    s_MyNodeId  = 0;
    s_SendError = 0;
    std::memset(s_NodeData, 0x00, sizeof(s_NodeData));

    if( param.mode == MASTER )
    {
        result = nn::uds::CreateNetwork( SUB_ID, 12, nn::uds::CreateLocalCommunicationId(nn::uds::CTR::SampleDemos::UNIQUE_ID_SAMPLEDEMOS_UDS_COM_DEMO1), PASSPHRASE, sizeof(PASSPHRASE) );
        if( result.IsFailure() )
        {
            NN_LOG("CreateNetwork Error\n");
        }
    }
    else
    {
        result = nn::uds::ConnectNetwork( s_NetDesc[param.targetNum], (param.mode==CLIENT)? nn::uds::CONNECT_AS_CLIENT:nn::uds::CONNECT_AS_SPECTATOR, PASSPHRASE, sizeof(PASSPHRASE) );
        if ( result.IsFailure() )
        {
            if( result == nn::uds::CTR::ResultNotFoundNetwork() )
            {
                NN_LOG( "ConnectNetwork Error (NotFoundNetwork)\n" );
            }
            else if( result == nn::uds::CTR::ResultAlreadyNetworkIsFull() )
            {
                NN_LOG( "ConnectNetwork Error (AlreadyNetworkIsFull)\n" );
            }
            else if( result == nn::uds::CTR::ResultDeniedFromMaster() )
            {
                NN_LOG( "ConnectNetwork Error (DeniedFromMaster)\n" );
            }
            else if( result == nn::uds::CTR::ResultConnectionTimeout() )
            {
                NN_LOG( "ConnectNetwork Error (ConnectionTimeout)\n" );
            }
            else if( result == nn::uds::CTR::ResultNotAuthorized() )
            {
                NN_LOG( "ConnectNetwork Error (NotAuthorized)\n" );
            }
            else
            {
                NN_LOG( "ConnectNetwork Error (other)\n" );
                isDetectError = true;
            }
        }
    }

    if ( result.IsFailure() )
    {
        return;
    }
    else
    {
        if( param.mode == MASTER || param.mode == CLIENT )
        {
            s_SendThread.Start( UdsSendThread, NULL, s_SendStackBuffer, SEND_THREAD_PRIORITY );
        }
    }

    //Start threads
    s_DisconnectEvent.Initialize(false);

    s_RecvThread.Start( UdsRecvThread, NULL, s_ReceiveStackBuffer, RECEIVE_THREAD_PRIORITY );
    s_ResultThread.Start( ResultMoniterThread, NULL, s_ResultStackBuffer, RESULT_MONITOR_THREAD_PRIORITY );

    NN_LOG( "Enter MonitorLoop\n" );
    //Monitor connection status
    while(true)
    {
        NN_LOG( "s_StatusUpdateEvent Wait\n" );
        s_StatusUpdateEvent.Wait();
        NN_LOG( "s_StatusUpdateEvent Signal\n" );

        nn::uds::ConnectionStatus connectionStatus;
        result = nn::uds::GetConnectionStatus(&connectionStatus);
        if( result.IsFailure() )
        {
            NN_LOG("GetConnectionStatus Failed.\n");            // Would not usually occur when UDS is initialized
            isDetectError = true;
            break;
        }

        if( connectionStatus.nowState != nn::uds::STATE_MASTER && connectionStatus.nowState != nn::uds::STATE_CLIENT && connectionStatus.nowState != nn::uds::STATE_SPECTATOR )
        {
            s_DisconnectEvent.Signal();
            NN_LOG("Leave Network (Reason:%d)\n", connectionStatus.disconnectReason);
            if( connectionStatus.disconnectReason == nn::uds::REQUEST_FROM_SYSTEM )
            {
                //A disconnection requested by the system is treated as an error
                isDetectError = true;
            }
            break;
        }
        else
        {
            s_MyNodeId = connectionStatus.myNodeId;

            for(int i = 0;i < nn::uds::NODE_MAX; i++) //List all nodes
            {
                if( connectionStatus.updateNodeBitmap&(0x0001 << i ) )
                {
                    if( connectionStatus.nodeIdList[i] != 0 ) //New connection
                    {
                        nn::uds::NodeInformation nodeInfo;
                        result = nn::uds::GetNodeInformation( &nodeInfo, connectionStatus.nodeIdList[i] );
                        if( result.IsSuccess() )
                        {
                            //Get name
                             char name[11];
                            if( nodeInfo.userName.isNgUserName == true )
                            {
                                // If the user name is a profanity word, change to ***********
                                std::memset(name, '*', 11);
                            }
                            else
                            {
                                std::wcstombs(name, nodeInfo.userName.userName, 11);
                            }
                            s_ResultLock.Enter();
                            s_NodeData[i].nodeId = connectionStatus.nodeIdList[i];
                            std::memcpy(s_NodeData[i].name, name, sizeof(name));
                            s_ResultLock.Leave();
                            NN_LOG("Connect! ID:%d, name:%s\n", connectionStatus.nodeIdList[i], name);
                        }
                        else
                        {
                            NN_LOG("Connect Unknwon Node\n");
                        }
                    }
                    else //Disconnect
                    {
                        NN_LOG("Disconnect! ID:%d\n", s_NodeData[i].nodeId);
                        s_ResultLock.Enter();
                        std::memset(&s_NodeData[i], 0x00, sizeof(NodeData));
                        s_ResultLock.Leave();
                    }
                }
            }
        }
    }

    //Terminate threads
    if( param.mode == MASTER || param.mode == CLIENT )
    {
        s_SendThread.Join();
        s_SendThread.Finalize();
    }

    s_RecvThread.Join();
    s_ResultThread.Join();
    s_RecvThread.Finalize();
    s_ResultThread.Finalize();
    s_DisconnectEvent.Finalize();

    return;
}
