﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     uds_Main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/Result.h>
#include <nn/hid.h>
#include <nn/uds.h>
#include <nn/applet.h>
#include "uniqueId_List.h"

/* Please see man pages for details 
    
              
              
*/

namespace
{
    bit8 s_ScanBuffer[2048];    //Scan buffer. About 1 KB is needed for each network.
    bit8 s_UdsReceiveBuffer[4096*20] NN_ATTRIBUTE_ALIGN(4096);  //Receive buffer for system. Must be a multiple of 4096 because the buffer is 4096-byte aligned.

    bool isMaster = false;
    bool isSpectator = false;

    const u8 SUB_ID = 0x01;                // ID used when dividing communications (for match-ups or exchange) in the application.
    const u8 MAX_ENTRY = 5;                // Maximum number of connections
    const char PASSPHRASE[] = "CTR_SDK_SAMPLE";     // Passphrase. Uses this string to generate an encryption key for communication routes.
    const int PACKET_SIZE = nn::uds::UDS_PACKET_PAYLOAD_MAX_SIZE;  //Size of the data sent and received in this demo.

    u16 s_NodeList[MAX_ENTRY];             // Array used to display disconnected nodes

    u32 s_SendBuffer[ PACKET_SIZE / sizeof(u32) ];
    u32 s_ReceiveBuffer[ PACKET_SIZE / sizeof(u32) ];


}

//Send thread
void SendThread(uptr param)
{
NN_UNUSED_VAR( param );

    nn::uds::EndpointDescriptor ed;
    nn::Result result = nn::uds::CreateEndpoint(&ed); //Generate Endpoint
    NN_UTIL_PANIC_IF_FAILED(result);

    //Send at fixed intervals
    nn::os::Timer timer;
    timer.Initialize(false);
    timer.StartPeriodic( nn::fnd::TimeSpan::FromMilliSeconds(10), nn::fnd::TimeSpan::FromMilliSeconds(50) );

    s_SendBuffer[0] = 0;

    bool isExit = false;
    while(isExit == false)
    {
        timer.Wait();

        result = nn::uds::SendTo(ed, s_SendBuffer, sizeof(s_SendBuffer), nn::uds::BROADCAST_NODE_ID, 1, nn::uds::NO_WAIT);
        if( result.IsFailure() )
        {
            if( result == nn::uds::ResultBufferIsFull() )
            {
                // When send/receive processing is slow and the low-layer send buffer is full
                NN_LOG(" nn::uds::SendTo is Failed (BUSY)\n");
            }
            else if( result == nn::uds::ResultInvalidState() )
            {
                // When disconnected from the network
                NN_LOG(" nn::uds::SendTo is Failed (INVALID_STATE)\n");
                isExit = true;
            }
            else if ( result == nn::uds::ResultWirelessOff() )
            {
                    // When in wireless OFF mode
                    NN_LOG(" nn::uds::SendTo is Failed (WIFI_OFF)\n");
                    isExit = true;
            }
            else
            {
                NN_PANIC(" nn::uds::SendTo is Failed (OTHER)");
            }
        }
        s_SendBuffer[0]++; //Use the buffer's starting address as a counter
    }
    timer.Stop();
    timer.Finalize();

    //Disconnect
    NN_LOG("DestroyEndpoint\n");
    result = nn::uds::DestroyEndpoint(&ed);
    NN_UTIL_PANIC_IF_FAILED(result);

    NN_LOG("%s is Finish!\n", __func__);
}

//Receive thread
void RecvThread(uptr param)
{
NN_UNUSED_VAR( param );

    nn::uds::EndpointDescriptor ed;
    nn::Result result;
    result = nn::uds::CreateEndpoint(&ed);  //Generate Endpoint
    NN_UTIL_PANIC_IF_FAILED(result);

    result = nn::uds::Attach(&ed, nn::uds::BROADCAST_NODE_ID, 1 );  //Endpoint receive settings. Always run this before receiving data.
    NN_UTIL_PANIC_IF_FAILED(result);

    while(true)
    {
        u16    srcNodeId;
        size_t recvedSize;

        result = nn::uds::ReceiveFrom(ed, s_ReceiveBuffer, &recvedSize, &srcNodeId, sizeof(s_ReceiveBuffer));
        if( result.IsFailure() )
        {
            break;
        }
        NN_LOG("Recv From:%02u, counter=%08x, size=%d Byte\n", srcNodeId, s_ReceiveBuffer[0], recvedSize);
    }

    NN_LOG("DestroyEndpoint\n");
    result = nn::uds::DestroyEndpoint(&ed);
    NN_UTIL_PANIC_IF_FAILED(result);

    NN_LOG("%s is Finish!\n", __func__);
}


void nnMain( void )
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    NN_LOG("START\n");

    nn::Result result;

    //Use HID to change the operation mode
    {
        // Initialize hid
        result = nn::hid::Initialize();
        NN_UTIL_PANIC_IF_FAILED(result);

        NN_LOG("MODE SELECT (A:Master  B:Client, Y:Spectator)\n");

        // Prepare pad
        nn::hid::PadReader padReader;
        nn::hid::PadStatus padStatus;

        // Wait for HID input to stabilize
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromSeconds(1));
        while(true)
        {
            // Get input
            padReader.ReadLatest(&padStatus);

            if(padStatus.hold & nn::hid::BUTTON_A)
            {
                isMaster = true;
                NN_LOG("-> Select: Master!\n");
                break;
            }

            if(padStatus.hold & nn::hid::BUTTON_B)
            {
                isMaster   = false;
                isSpectator = false;
                NN_LOG("-> Select: Client!\n");
                break;
            }

            if(padStatus.hold & nn::hid::BUTTON_Y)
            {
                isMaster   = false;
                isSpectator = true;
                NN_LOG("-> Select: Spectator!\n");
                break;
            }

            nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(50));
        }
        nn::hid::Finalize(); //To reduce processing that is not related to UDS, this demo finalizes the hid library here.
    }

    nn::os::Event statusUpdateEvent;
    result = nn::uds::Initialize( &statusUpdateEvent, s_UdsReceiveBuffer, sizeof(s_UdsReceiveBuffer) );
    if( result == nn::uds::ResultAlreadyOccupiedWirelessDevice() )
    {
        // If UDS is in other communications, the initialization fails
        NN_PANIC(" Other Network Service is Already Occupying Wireless Device.");
    }
    else if( result == nn::uds::ResultWirelessOff() )
    {
        // Was in wireless OFF mode
        NN_PANIC(" Wireless Off Mode Now.");
    }
    NN_UTIL_PANIC_IF_FAILED(result);

    if( isMaster )
    {
        //Build the network
        result = nn::uds::CreateNetwork( SUB_ID, MAX_ENTRY, nn::uds::CreateLocalCommunicationId(nn::uds::CTR::SampleDemos::UNIQUE_ID_SAMPLEDEMOS_UDS_SIMPLE), PASSPHRASE, sizeof(PASSPHRASE) );
        NN_UTIL_PANIC_IF_FAILED(result);
    }
    else // Client/Spectator initially searches until it finds a network
    {
        while(true)
        {
            result = nn::uds::Scan( s_ScanBuffer, sizeof(s_ScanBuffer), SUB_ID, nn::uds::CreateLocalCommunicationId(nn::uds::CTR::SampleDemos::UNIQUE_ID_SAMPLEDEMOS_UDS_SIMPLE) );
            NN_UTIL_PANIC_IF_FAILED(result);

            //Analyze scan results
            {
                nn::uds::ScanResultReader networkList(s_ScanBuffer);

                //Connect to the first network found
                if( 0 != networkList.GetCount() )
                {
                    nn::uds::NetworkDescriptionReader netDescReader = networkList.GetNextDescription();
                    nn::uds::NetworkDescription netDesc;
                    result = netDescReader.GetNetworkDescription(&netDesc);
                    NN_UTIL_PANIC_IF_FAILED(result);

                    result = nn::uds::ConnectNetwork( netDesc, isSpectator? nn::uds::CONNECT_AS_SPECTATOR:nn::uds::CONNECT_AS_CLIENT, PASSPHRASE, sizeof(PASSPHRASE) );

                    if( result.IsFailure() )
                    {
                        if( result == nn::uds::ResultNotFoundNetwork() )
                        {
                            // The target network was not found
                            NN_LOG(" nn::uds::ConnectNetwork is Failed ( Not Found )\n");
                        }
                        else if( result == nn::uds::ResultAlreadyNetworkIsFull() )
                        {
                            // The number of connected devices is already at the maximum
                            NN_LOG(" nn::uds::ConnectNetwork is Failed ( Max Connection )\n");
                        }
                        else if( result == nn::uds::ResultDeniedFromMaster() )
                        {
                            // Connection to Master refused
                            NN_LOG(" nn::uds::ConnectNetwork is Failed ( Denied )\n");
                        }
                        else if( result == nn::uds::ResultConnectionTimeout() )
                        {
                            // Target network found, but there was no response from Master
                            NN_LOG(" nn::uds::ConnectNetwork is Failed ( Timeout )\n");
                        }
                        else if( result == nn::uds::ResultWirelessOff() )
                        {
                            // Entered wireless OFF mode
                            NN_LOG(" nn::uds::ConnectNetwork is Failed ( Wireless Off )\n");
                        }
                        else
                        {
                            //This demo should not generate state errors, and therefore it is designed to Panic.
                            NN_PANIC(" nn::uds::ConnectNetwork is Failed (Other)");
                        }
                    }
                    else
                    {
                        break; //Go to next process as connection was successful
                    }
                }
            }
        }
    }

    //Connection is complete. Set the send/receive thread.
    nn::os::Thread sendThread;
    nn::os::Thread recvThread;
    sendThread.StartUsingAutoStack ( SendThread, NULL, 4096 );
    recvThread.StartUsingAutoStack ( RecvThread, NULL, 4096 );

    //Monitor connection status
    while(true)
    {
        // statusUpdateEvent is signaled if the network status has changed (there was a disconnection that included local node).
        statusUpdateEvent.Wait();
        NN_LOG("Signal!\n");
        nn::uds::ConnectionStatus connectionStatus;
        result = nn::uds::GetConnectionStatus(&connectionStatus);
        NN_UTIL_PANIC_IF_FAILED(result);

        if( connectionStatus.nowState ) //Check connection status
        {
            NN_LOG("State is Update!(Now=%d, NodeId=%d)\n", connectionStatus.nowState, connectionStatus.myNodeId);

            if( connectionStatus.nowState != nn::uds::STATE_MASTER && connectionStatus.nowState != nn::uds::STATE_CLIENT )
            {
                // The Master can only leave the network when DestroyNetwork has been run.
                NN_LOG("Destruct or Disconnect! Reason = %d\n", connectionStatus.disconnectReason);
                break;
            }
        }
        for(int i = 0;i < MAX_ENTRY; i++) //List all nodes
        {
            if( connectionStatus.updateNodeBitmap&(0x0001 << i ))
            {
                //Change occurred
                if( connectionStatus.nodeIdList[i] != 0 )
                {
                    if( connectionStatus.nodeIdList[i] != connectionStatus.myNodeId )
                    {
                        //New connection
                        nn::uds::NodeInformation nodeInfo;
                        NN_LOG("Connect New Node!: %d\n", connectionStatus.nodeIdList[i]);
                        result = nn::uds::GetNodeInformation( &nodeInfo, connectionStatus.nodeIdList[i] );
                        if( result.IsFailure() )
                        {
                            continue;
                        }
                        //Display name
                        if( nodeInfo.userName.isNgUserName == false )
                        {
                            char name[11];
                            std::wcstombs(name, nodeInfo.userName.userName, 11);
                            NN_LOG( "  Name: %s\n", name);
                        }
                    }
                }
                else
                {
                    // If a change occurred and NodeID=0, it means that all previously connected nodes were disconnected.
                    NN_LOG("Disconnect Node!: %d\n", s_NodeList[i]);
                }

                s_NodeList[i] = connectionStatus.nodeIdList[i];
            }
        }

        //Get link level
        {
            nn::uds::LinkLevel linkLevel;
            nn::uds::GetLinkLevel(&linkLevel);

            NN_LOG("LinkLevel = %d\n", linkLevel);
        }
    }

    sendThread.Join();
    recvThread.Join();
    sendThread.Finalize();
    recvThread.Finalize();

    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(200));
    NN_LOG("END\n");

    nn::uds::Finalize();
    NN_LOG("Finalize!\n");
}
